/* This file is part of Cloudy and is copyright (C) 1978-2005 by Gary J. Ferland.
 * For conditions of distribution and use, see copyright notice in license.txt */
/*CO_step fills in matrix for heavy elements molecular routines */
/* 
 * change following to TRUE to turn on debug statements 
 *
 * Nick Abel between July and October of 2003 assisted Dr. Ferland in improving the heavy element 
 * molecular network in Cloudy. Before this routine would predict negative abundances if 
 * the fraction of carbon in the form of molecules came close to 100%. A reorganizing of 
 * the reaction network detected several bugs.  Treatment of "coupled reactions",
 * in which both densities in the reaction rate were being predicted by Cloudy, were also 
 * added.  Due to these improvements, Cloudy can now perform calculations
 * where 100% of the carbon is in the form of CO without predicting negative abundances
 *
 * Additional changes were made in November of 2003 so that our reaction 
 * network would include all reactions from the TH85 paper.  This involved 
 * adding silicon to the chemical network.  Also the reaction rates were
 * labeled to make identification with the reaction easier and the matrix 
 * elements of atomic C, O, and Si are now done in a loop, which makes 
 * the addition of future chemical species (like N or S) easy.
 * */
#include "cddefines.h"
#include "physconst.h"
#include "taulines.h"
#include "secondaries.h"
#include "dense.h"
#include "ionbal.h"
#include "thermal.h"
#include "grainvar.h"
#include "hmi.h"
#include "conv.h"
#include "trace.h"
#include "rfield.h"
#include "phycon.h"
#include "doppvel.h"
#include "rt.h"
#include "coolheavy.h"
#include "timesc.h"
#include "lapack.h"
#include "mole.h"
/*lint -e778 constant expression evaluatess to 0 in operation '-' */
/*=================================================================*/


/*CO_step fills in matrix for heavy elements molecular routines */
void CO_step(
	/* set true if we found neg pops */
	int *lgNegPop, 
	/* set true if we tried to compute the pops, but some were zero */
	int *lgZerPop )
{

	/* this routine works with the heavy molecules, of which there are NUM_HEAVY_MOLEC,
	 * plus C and O, making a total of NUM_COMOLE_CALC*/
	int32 ipiv[NUM_COMOLE_CALC], merror;
	long int i, j, k,
		nelem , ion , ion2;
	double 
	C_CH2_CH_CH_1,
	C_COP_CO_CP_1,
	C_H2OP_OH_CHP_1,
	C_H3OP_HCOP_H2_1,
	C_HCOP_CO_CHP_1,
	C_O_CO_nu_1,
	C_O2_CO_O_1,
	C_O2P_O2_CP_1,
	C_O2P_COP_O_1,
	C_OH_O_CH_1,
	C_OH_CO_H_1,
	C_OHP_O_CHP_1,
	C_SiOP_SiP_CO_1,
	CP_OH_CO_HP_1,
	CP_CH_CHP_C_1,
	CP_SiO_SiP_CO_1,
	CP_CH2_CH2P_C_1,
	CP_H2O_HCOP_H_1,
	CP_O_COP_nu_1,
	CP_O2_COP_O_1,
	CP_O2_CO_OP_1,
	CP_OH_COP_H_1,
	O_CH_OH_C_1,
	O_CH_HCOP_e_1,
	O_CH_CO_H_1,
	O_CHP_COP_H_1,
	O_CH2_OH_CH_1,
	O_CH2_CO_H_H_1,
	O_CH2_CO_H2_1,
	O_CH2P_HCOP_H_1,
	O_CH3P_HCOP_H2_1,
	O_COP_CO_OP_1,
	O_H2O_OH_OH_1,
	O_H2OP_O2P_H2_1,
	O_O_O2_nu_1,
	O_OH_O2_H_1,
	O_OHP_O2P_H_1,
	O_Si_SiO_nu_1,
	O_SiP_SiOP_nu_1,
	O_SiH_SiO_H_1,
	O_SiH2P_SiOHP_H_1,
	O_SiOP_O2_SiP_1,
	OP_CH_O_CHP_1,
	OP_CH_COP_H_1,
	OP_CH2_O_CH2P_1,
	OP_CO_COP_O_1,
	OP_H2O_H2OP_O_1,
	OP_O2_O2P_O_1,
	OP_OH_O2P_H_1,
	OP_OH_OHP_O_1,
	Si_CHP_SiP_CH_1,
	Si_H2OP_SiP_H2O_1,
	Si_OH_SiO_H_1,
	Si_O2P_O2_SiP_1,
	SiP_H2O_SiOHP_H_1,
	SiP_OH_SiOP_H_1,
	SiP_O2_SiOP_O_1,
	CH_COP_HCOP_C_1,
	CH_COP_CO_CHP_1,
	CH_H2OP_H2O_CHP_1,
	CH_H2OP_OH_CH2P_1,
	CH_H3OP_H2O_CH2P_1,
	CH_HCOP_CO_CH2P_1,
	CH_O2_CO_OH_1,
	CH_O2P_O2_CHP_1,
	CH_O2P_HCOP_O_1,
	CH_OHP_OH_CHP_1,
	CH_OHP_O_CH2P_1,
	CH_SiOP_HCOP_Si_1,
	CHP_H2O_H3OP_C_1,
	CHP_OH_COP_H2_1,
	CHP_H2O_HCOP_H2_1,
	CHP_O2_HCOP_O_1,
	CHP_O2_COP_OH_1,
	CH2_COP_HCOP_CH_1,
	CH2_COP_CO_CH2P_1,
	CH2_H2OP_H2O_CH2P_1,
	CH2_H2OP_OH_CH3P_1,
	CH2_H3OP_H2O_CH3P_1,
	CH2_HCOP_CO_CH3P_1,
	CH2_O2_CO_H2O_1,
	CH2_O2P_O2_CH2P_1,
	CH2_OH_H2O_CH_1,
	CH2_OHP_OH_CH2P_1,
	CH2_OHP_O_CH3P_1,
	CH2P_O2_HCOP_OH_1,
	COP_O2_O2P_CO_1,
	H2O_COP_CO_H2OP_1,
	H2O_COP_HCOP_OH_1,
	H2O_HCOP_CO_H3OP_1,
	H2OP_CO_HCOP_OH_1,
	H2OP_H2O_H3OP_OH_1,
	H2OP_O2_O2P_H2O_1,
	H3OP_SiH_SiH2P_H2O_1,
	H3OP_SiO_SiOHP_H2O_1,
	HCOP_SiH_SiH2P_CO_1,
	HCOP_SiO_SiOHP_CO_1,
	OH_COP_HCOP_O_1,
	OH_COP_CO_OHP_1,
	OH_H2OP_H3OP_O_1,
	OH_HCOP_CO_H2OP_1,
	OH_OH_H2O_O_1,
	OHP_CO_HCOP_O_1,
	OHP_H2O_H3OP_O_1,
	OHP_H2O_H2OP_OH_1,
	OHP_O2_O2P_OH_1,
	OHP_OH_H2OP_O_1,
	OHP_SiH_SiH2P_O_1,
	OHP_SiO_SiOHP_O_1,
	C_CH2_CH_CH_2,
	C_COP_CO_CP_2,
	C_H2OP_OH_CHP_2,
	C_H3OP_HCOP_H2_2,
	C_HCOP_CO_CHP_2,
	C_O_CO_nu_2,
	C_O2_CO_O_2,
	C_O2P_O2_CP_2,
	C_O2P_COP_O_2,
	C_OH_O_CH_2,
	C_OH_CO_H_2,
	C_OHP_O_CHP_2,
	C_SiOP_SiP_CO_2,
	CP_OH_CO_HP_2,
	CP_CH_CHP_C_2,
	CP_SiO_SiP_CO_2,
	CP_CH2_CH2P_C_2,
	CP_H2O_HCOP_H_2,
	CP_O_COP_nu_2,
	CP_O2_COP_O_2,
	CP_O2_CO_OP_2,
	CP_OH_COP_H_2,
	O_CH_OH_C_2,
	O_CH_HCOP_e_2,
	O_CH_CO_H_2,
	O_CHP_COP_H_2,
	O_CH2_OH_CH_2,
	O_CH2_CO_H_H_2,
	O_CH2_CO_H2_2,
	O_CH2P_HCOP_H_2,
	O_CH3P_HCOP_H2_2,
	O_COP_CO_OP_2,
	O_H2O_OH_OH_2,
	O_H2OP_O2P_H2_2,
	O_O_O2_nu_2,
	O_OH_O2_H_2,
	O_OHP_O2P_H_2,
	O_Si_SiO_nu_2,
	O_SiP_SiOP_nu_2,
	O_SiH_SiO_H_2,
	O_SiH2P_SiOHP_H_2,
	O_SiOP_O2_SiP_2,
	OP_CH_O_CHP_2,
	OP_CH_COP_H_2,
	OP_CH2_O_CH2P_2,
	OP_CO_COP_O_2,
	OP_H2O_H2OP_O_2,
	OP_O2_O2P_O_2,
	OP_OH_O2P_H_2,
	OP_OH_OHP_O_2,
	Si_CHP_SiP_CH_2,
	Si_H2OP_SiP_H2O_2,
	Si_OH_SiO_H_2,
	Si_O2P_O2_SiP_2,
	SiP_H2O_SiOHP_H_2,
	SiP_OH_SiOP_H_2,
	SiP_O2_SiOP_O_2,
	CH_COP_HCOP_C_2,
	CH_COP_CO_CHP_2,
	CH_H2OP_H2O_CHP_2,
	CH_H2OP_OH_CH2P_2,
	CH_H3OP_H2O_CH2P_2,
	CH_HCOP_CO_CH2P_2,
	CH_O2_CO_OH_2,
	CH_O2P_O2_CHP_2,
	CH_O2P_HCOP_O_2,
	CH_OHP_OH_CHP_2,
	CH_OHP_O_CH2P_2,
	CH_SiOP_HCOP_Si_2,
	CHP_H2O_H3OP_C_2,
	CHP_OH_COP_H2_2,	
	CHP_H2O_HCOP_H2_2,
	CHP_O2_HCOP_O_2,
	CHP_O2_COP_OH_2,
	CH2_COP_HCOP_CH_2,
	CH2_COP_CO_CH2P_2,
	CH2_H2OP_H2O_CH2P_2,
	CH2_H2OP_OH_CH3P_2,
	CH2_H3OP_H2O_CH3P_2,
	CH2_HCOP_CO_CH3P_2,
	CH2_O2_CO_H2O_2,
	CH2_O2P_O2_CH2P_2,
	CH2_OH_H2O_CH_2,
	CH2_OHP_OH_CH2P_2,
	CH2_OHP_O_CH3P_2,
	CH2P_O2_HCOP_OH_2,
	COP_O2_O2P_CO_2,
	H2O_COP_CO_H2OP_2,
	H2O_COP_HCOP_OH_2,
	H2O_HCOP_CO_H3OP_2,
	H2OP_CO_HCOP_OH_2,
	H2OP_H2O_H3OP_OH_2,
	H2OP_O2_O2P_H2O_2,
	H3OP_SiH_SiH2P_H2O_2,
	H3OP_SiO_SiOHP_H2O_2,
	HCOP_SiH_SiH2P_CO_2,
	HCOP_SiO_SiOHP_CO_2,
	OH_COP_HCOP_O_2,
	OH_COP_CO_OHP_2,
	OH_H2OP_H3OP_O_2,
	OH_HCOP_CO_H2OP_2,
	OH_OH_H2O_O_2,
	OHP_CO_HCOP_O_2,
	OHP_H2O_H3OP_O_2,
	OHP_H2O_H2OP_OH_2,
	OHP_O2_O2P_OH_2,
	OHP_OH_H2OP_O_2,
	OHP_SiH_SiH2P_O_2,
	OHP_SiO_SiOHP_O_2,
	H_CH_C_H_H,
	H_OH_O_H_H,
	H_H2O_OH_H_H,
	H_COP_CO_HP,
	H_CH_C_H2,
	H_CHP_CP_H2,
	H_CH2_CH_H2,
	H_CH3P_CH2P_H2,
	H_OH_O_H2,
	H_H2O_OH_H2,
	H_O2_OH_O,
	H_O2_O_O_H,
	H_C_CH_nu,
	H_CP_CHP_nu,
	H_OH_H2O_nu,
	H_O_OH_nu,
	Hminus_CH_CH2_e,
	Hminus_OH_H2O_e,
	Hminus_O_OH_e,
	Hminus_HCOP_CO_H2,
	Hminus_H3OP_H2O_H2,
	Hminus_H3OP_OH_H2_H,
	HP_CH_CHP_H,
	HP_CH2_CH2P_H,
	HP_H2O_H2OP_H,
	HP_O2_O2P_H,
	HP_OH_OHP_H,
	HP_SiO_SiOP_H,
	HP_CH2_CHP_H2,
	HP_SiH_SiP_H2,
	H2_C_CH_H,
	H2_CP_CHP_H,
	H2_CH_CH2_H,
	H2_CHP_CH2P_H,
	H2_OH_H2O_H,
	H2_OHP_H2OP_H,
	H2_H2OP_H3OP_H,
	H2_COP_HCOP_H,
	H2_O_OH_H,
	H2_OP_OHP_H,
	H2_SiOP_SiOHP_H,
	H2_CH_C_H2_H,
	H2_OH_O_H2_H,
	H2_CH2P_CH3P_H,
	H2_H2O_OH_H2_H,
	H2_O2_O_O_H2,
	H2_O2_OH_OH,
	H2_C_CH2_nu,
	H2_CP_CH2P_nu,
	H2_SiP_SiH2P_nu,
	H2P_C_CHP_H,
	H2P_CH_CH2P_H,
	H2P_CH2_CH3P_H,
	H2P_OH_H2OP_H,
	H2P_H2O_H3OP_H,
	H2P_CO_HCOP_H,
	H2P_O_OHP_H,
	H2P_CH_CHP_H2,
	H2P_CH2_CH2P_H2,
	H2P_CO_COP_H2,
	H2P_H2O_H2OP_H2,
	H2P_O2_O2P_H2,
	H2P_OH_OHP_H2,
	H3P_C_CHP_H2,
	H3P_CH_CH2P_H2,
	H3P_CH2_CH3P_H2,
	H3P_OH_H2OP_H2,
	H3P_H2O_H3OP_H2,
	H3P_CO_HCOP_H2,
	H3P_O_OHP_H2,
	H3P_SiH_SiH2P_H2,
	H3P_SiO_SiOHP_H2,
	HeP_CO_OP_C_He,
	HeP_CO_O_CP_He,
	HeP_CH_CHP_He,
	HeP_H2O_H2OP_He,
	HeP_O2_O2P_He,
	HeP_CH_CP_He_H,
	HeP_CH2_CHP_He_H,
	HeP_OH_OP_He_H,
	HeP_H2O_OHP_He_H,
	HeP_SiH_SiP_He_H,
	HeP_H2O_OH_He_HP,
	HeP_CH2_CP_He_H2,
	HeP_O2_OP_O_He,
	HeP_SiO_SiP_O_He,
	HeP_SiO_Si_OP_He,
	crnu_CH_C_H,
	crnu_CHP_CP_H,
	crnu_CO_O_C,
	crnu_H2O_OH_H,
	crnu_O2_O2P_e,
	crnu_O2_O_O,
	crnu_OH_O_H,
	crnu_SiH_Si_H,
	crnu_SiO_Si_O,
	CRP_CO_COP_e,
	nu_CH_CHP_e,
	nu_CH_C_H,
	nu_CHP_CP_H,
	nu_CH2_CH2P_e,
	nu_CH2_CH_H,
	nu_CH2P_CHP_H,
	nu_CH3P_CH2P_H,
	nu_CH3P_CHP_H2,
	nu_CO_O_C,
	nu_H2O_H2OP_e,
	nu_H2O_OH_H,
	nu_O2_O2P_e,
	nu_O2_O_O,
	nu_OH_OHP_e,
	nu_OH_O_H,
	nu_OHP_O_HP,
	nu_SiH_Si_H,
	nu_SiO_Si_O,
	e_CHP_C_H,
	e_CH2P_CH_H,
	e_CH2P_C_H_H,
	e_CH2P_C_H2,
	e_CH3P_C_H2_H,
	e_CH3P_CH2_H,
	e_CH3P_CH_H_H,
	e_CH3P_CH_H2,
	e_CH3P_CH3_nu,
	e_COP_O_C,
	e_H2OP_OH_H,
	e_H2OP_O_H_H,
	e_H2OP_O_H2,
	e_H3OP_H2O_H,
	e_H3OP_OH_H_H,
	e_H3OP_OH_H2,
	e_H3OP_O_H2_H,
	e_HCOP_CO_H,
	e_O2P_O_O,
	e_OHP_O_H,
	e_SiH2P_SiH_H,
	e_SiH2P_Si_H_H,
	e_SiH2P_Si_H2,
	e_SiOP_Si_O,
	e_SiOHP_SiO_H,
	e_SiOHP_Si_OH,
	co_e_c_o_e,
	h2s_ch_ch2_h,
	h2s_o_oh_h,
	h2s_oh_h2o_h,
	h2s_o2_twooh,
	h2s_c_ch_h,
	h2s_cp_chp_h,

	/* These reactions were put in by NPA so that we can have EXACTLY the same 
	   reaction network as the others in Leiden */

	H2s_CH_C_H2_H, 
	H2s_OH_O_H2_H, 
	H2s_H2O_OH_H2_H, 
	H2s_O2_O_O_H2, 
	H2s_CH2_CH3_H,  
	H2s_CH_CH3_nu,
	H2s_CH3_CH4_H,

	H_CH3_CH2_H2,
	H_CH4P_CH3P_H2,
	H_CH5P_CH4P_H2,
	H2_CH2_CH3_H,
	H2_CH_CH3_nu,
	H2_CH3_CH4_H,
	H2_CH4P_CH5P_H,
	H2_CH3P_CH5P_nu,
	H2P_CH4_CH3P_H2,
	H2P_CH4_CH4P_H2,
	H2P_CH4_CH5P_H,
	H3P_CH3_CH4P_H2,
	H3P_CH4_CH5P_H2,
	HeP_CH3_CHP_He,
	HeP_CH4_CHP_He,
	HeP_CH4_CH2P_He,
	HeP_CH4_CH3_He,
	HeP_CH4_CH3P_He,
	HeP_CH4_CH4P_He,
	Hminus_CH2_CH3_e,
	Hminus_CH3_CH4_e,
	HP_CH3_CH3P_H,
	HP_CH4_CH3P_H2,
	HP_CH4_CH4P_H,
	nu_CH3_CH3P_e,
	nu_CH3_CH2_H,
	nu_CH3_CH_H2,
	crnu_CH3_CH3P_e,
	crnu_CH3_CH2_H,
	crnu_CH3_CH_H2,
	nu_CH4_CH3_H,
	nu_CH4_CH2_H2,
	nu_CH4_CH_H2,
	crnu_CH4_CH2_H2,
	e_CH5P_CH3_H2,
	e_CH5P_CH4_H,
	e_CH4P_CH3_H,
	e_CH4P_CH2_H_H,
	C_CH5P_CH4_CHP_1, 
	C_CH5P_CH4_CHP_2, 
	O_CH4_OH_CH3_1, 
	O_CH4_OH_CH3_2, 
	O_CH4P_OH_CH3P_1, 
	O_CH4P_OH_CH3P_2, 
	O_CH5P_H3OP_CH2_1, 
	O_CH5P_H3OP_CH2_2, 
	OP_CH4_OH_CH3P_1, 
	OP_CH4_OH_CH3P_2, 
	OP_CH4_CH4P_O_1, 
	OP_CH4_CH4P_O_2, 
	CH4_CH_CH3_CH2_1, 
	CH4_CH_CH3_CH2_2, 
	CH5P_CH_CH4_CH2P_1, 
	CH5P_CH_CH4_CH2P_2, 
	CH2_CH2_CH3_CH_1, 
	CH2_CH2_CH3_CH_2, 
	CH4_CH2_CH3_CH3_1, 
	CH4_CH2_CH3_CH3_2, 
	OH_CH2_O_CH3_1, 
	OH_CH2_O_CH3_2, 
	CH5P_CH2_CH4_CH3P_1, 
	CH5P_CH2_CH4_CH3P_2, 
	OH_CH3_CH4_O_1, 
	OH_CH3_CH4_O_2, 
	OH_CH3_H2O_CH2_1, 
	OH_CH3_H2O_CH2_2, 
	H2O_CH3_OH_CH4_1, 
	H2O_CH3_OH_CH4_2, 
	CH3_CH3_CH4_CH2_1, 
	CH3_CH3_CH4_CH2_2, 
	OH_CH4_H2O_CH3_1, 
	OH_CH4_H2O_CH3_2, 
	OHP_CH4_CH5P_O_1, 
	OHP_CH4_CH5P_O_2, 
	OHP_CH4_H3OP_CH2_1, 
	OHP_CH4_H3OP_CH2_2, 
	H2OP_CH4_H3OP_CH3_1, 
	H2OP_CH4_H3OP_CH3_2, 
	COP_CH4_HCOP_CH3_1, 
	COP_CH4_HCOP_CH3_2, 
	COP_CH4_CO_CH4P_1, 
	COP_CH4_CO_CH4P_2, 
	CH4_CH4P_CH5P_CH3_1, 
	CH4_CH4P_CH5P_CH3_2, 
	H2O_CH4P_H3OP_CH3_1, 
	H2O_CH4P_H3OP_CH3_2, 
	CO_CH4P_HCOP_CH3_1, 
	CO_CH4P_HCOP_CH3_2, 
	O2_CH4P_O2P_CH4_1, 
	O2_CH4P_O2P_CH4_2, 
	H2O_CH5P_H3OP_CH4_1, 
	H2O_CH5P_H3OP_CH4_2, 
	CO_CH5P_HCOP_CH4_1, 
	CO_CH5P_HCOP_CH4_2, 
	CH5P_OH_H2OP_CH4_1, 
	CH5P_OH_H2OP_CH4_2, 


	P_C_CH2,
	P_C_COP,
	P_C_H2OP,
	P_C_H3OP,
	P_C_HCOP,
	P_C_O,
	P_C_O2,
	P_C_O2P,
	P_C_OH,
	P_C_OHP,
	P_C_SiOP,
	P_CP_OH,
	P_CP_CH,
	P_CP_SiO,
	P_CP_CH2,
	P_CP_H2O,
	P_CP_O,
	P_CP_O2,
	P_O_CH,
	P_O_CHP,
	P_O_CH2,
	P_O_CH2P,
	P_O_CH3P,
	P_O_COP,
	P_O_H2O,
	P_O_H2OP,
	P_O_O,
	P_O_OH,
	P_O_OHP,
	P_O_Si,
	P_O_SiP,
	P_O_SiH,
	P_O_SiH2P,
	P_O_SiOP,
	P_OP_CH,
	P_OP_CH2,
	P_OP_CO,
	P_OP_H2O,
	P_OP_O2,
	P_OP_OH,
	P_Si_CHP,
	P_Si_H2OP,
	P_Si_OH,
	P_Si_O2P,
	P_SiP_H2O,
	P_SiP_OH,
	P_SiP_O2,
	P_CH_COP,
	P_CH_H2OP,
	P_CH_H3OP,
	P_CH_HCOP,
	P_CH_O2,
	P_CH_O2P,
	P_CH_OHP,
	P_CH_SiOP,
	P_CHP_H2O,
	P_CHP_O2,
	P_CHP_OH,
	P_CH2_COP,
	P_CH2_H2OP,
	P_CH2_H3OP,
	P_CH2_HCOP,
	P_CH2_O2,
	P_CH2_O2P,
	P_CH2_OH,
	P_CH2_OHP,
	P_CH2P_O2,
	P_COP_O2,
	P_H2O_COP,
	P_H2O_HCOP,
	P_H2OP_CO,
	P_H2OP_H2O,
	P_H2OP_O2,
	P_H3OP_SiH,
	P_H3OP_SiO,
	P_HCOP_SiH,
	P_HCOP_SiO,
	P_OH_COP,
	P_OH_H2OP,
	P_OH_HCOP,
	P_OH_OH,
	P_OHP_CO,
	P_OHP_H2O,
	P_OHP_O2,
	P_OHP_OH,
	P_OHP_SiH,
	P_OHP_SiO,

	/* new products */
	P_C_CH5P,
	P_O_CH4,
	P_O_CH4P,
	P_O_CH5P,
	P_OP_CH4,
	P_CH4_CH,
	P_CH5P_CH,
	P_CH2_CH2,
	P_CH4_CH2,
	P_OH_CH2,
	P_CH5P_CH2,
	P_OH_CH3,
	P_H2O_CH3,
	P_CH3_CH3,
	P_OH_CH4,
	P_OHP_CH4,
	P_H2OP_CH4,
	P_COP_CH4,
	P_CH4_CH4P,
	P_H2O_CH4P,
	P_CO_CH4P,
	P_O2_CH4P,
	P_H2O_CH5P,
	P_CO_CH5P,
	P_CH5P_OH,

	b_C_CH2_CH_CH,
	b_C_COP_CO_CP,
	b_C_H2OP_OH_CHP,
	b_C_H3OP_HCOP_H2,
	b_C_HCOP_CO_CHP,
	b_C_O_CO_nu,
	b_C_O2_CO_O,
	b_C_O2P_O2_CP,
	b_C_O2P_COP_O,
	b_C_OH_O_CH,
	b_C_OH_CO_H,
	b_C_OHP_O_CHP,
	b_C_SiOP_SiP_CO,
	b_CP_OH_CO_HP,
	b_CP_CH_CHP_C,
	b_CP_SiO_SiP_CO,
	b_CP_CH2_CH2P_C,
	b_CP_H2O_HCOP_H,
	b_CP_O_COP_nu,
	b_CP_O2_COP_O,
	b_CP_O2_CO_OP,
	b_CP_OH_COP_H,
	b_O_CH_OH_C,
	b_O_CH_HCOP_e,
	b_O_CH_CO_H,
	b_O_CHP_COP_H,
	b_O_CH2_OH_CH,
	b_O_CH2_CO_H_H,
	b_O_CH2_CO_H2,
	b_O_CH2P_HCOP_H,
	b_O_CH3P_HCOP_H2,
	b_O_COP_CO_OP,
	b_O_H2O_OH_OH,
	b_O_H2OP_O2P_H2,
	b_O_O_O2_nu,
	b_O_OH_O2_H,
	b_O_OHP_O2P_H,
	b_O_Si_SiO_nu,
	b_O_SiP_SiOP_nu,
	b_O_SiH_SiO_H,
	b_O_SiH2P_SiOHP_H,
	b_O_SiOP_O2_SiP,
	b_OP_CH_O_CHP,
	b_OP_CH_COP_H,
	b_OP_CH2_O_CH2P,
	b_OP_CO_COP_O,
	b_OP_H2O_H2OP_O,
	b_OP_O2_O2P_O,
	b_OP_OH_O2P_H,
	b_OP_OH_OHP_O,
	b_Si_CHP_SiP_CH,
	b_Si_H2OP_SiP_H2O,
	b_Si_OH_SiO_H,
	b_Si_O2P_O2_SiP,
	b_SiP_H2O_SiOHP_H,
	b_SiP_OH_SiOP_H,
	b_SiP_O2_SiOP_O,
	b_CH_COP_HCOP_C,
	b_CH_COP_CO_CHP,
	b_CH_H2OP_H2O_CHP,
	b_CH_H2OP_OH_CH2P,
	b_CH_H3OP_H2O_CH2P,
	b_CH_HCOP_CO_CH2P,
	b_CH_O2_CO_OH,
	b_CH_O2P_O2_CHP,
	b_CH_O2P_HCOP_O,
	b_CH_OHP_OH_CHP,
	b_CH_OHP_O_CH2P,
	b_CH_SiOP_HCOP_Si,
	b_CHP_H2O_H3OP_C,
	b_CHP_OH_COP_H2,
	b_CHP_H2O_HCOP_H2,
	b_CHP_O2_HCOP_O,
	b_CHP_O2_COP_OH,
	b_CH2_COP_HCOP_CH,
	b_CH2_COP_CO_CH2P,
	b_CH2_H2OP_H2O_CH2P,
	b_CH2_H2OP_OH_CH3P,
	b_CH2_H3OP_H2O_CH3P,
	b_CH2_HCOP_CO_CH3P,
	b_CH2_O2_CO_H2O,
	b_CH2_O2P_O2_CH2P,
	b_CH2_OH_H2O_CH,
	b_CH2_OHP_OH_CH2P,
	b_CH2_OHP_O_CH3P,
	b_CH2P_O2_HCOP_OH,
	b_COP_O2_O2P_CO,
	b_H2O_COP_CO_H2OP,
	b_H2O_COP_HCOP_OH,
	b_H2O_HCOP_CO_H3OP,
	b_H2OP_CO_HCOP_OH,
	b_H2OP_H2O_H3OP_OH,
	b_H2OP_O2_O2P_H2O,
	b_H3OP_SiH_SiH2P_H2O,
	b_H3OP_SiO_SiOHP_H2O,
	b_HCOP_SiH_SiH2P_CO,
	b_HCOP_SiO_SiOHP_CO,
	b_OH_COP_HCOP_O,
	b_OH_COP_CO_OHP,
	b_OH_H2OP_H3OP_O,
	b_OH_HCOP_CO_H2OP,
	b_OH_OH_H2O_O,
	b_OHP_CO_HCOP_O,
	b_OHP_H2O_H3OP_O,
	b_OHP_H2O_H2OP_OH,
	b_OHP_O2_O2P_OH,
	b_OHP_OH_H2OP_O,
	b_OHP_SiH_SiH2P_O,
	b_OHP_SiO_SiOHP_O,
	b_C_CH5P_CH4_CHP,
	b_O_CH4_OH_CH3,
	b_O_CH4P_OH_CH3P,
	b_O_CH5P_H3OP_CH2,
	b_OP_CH4_OH_CH3P,
	b_OP_CH4_CH4P_O,
	b_CH4_CH_CH3_CH2,
	b_CH5P_CH_CH4_CH2P,
	b_CH2_CH2_CH3_CH,
	b_CH4_CH2_CH3_CH3,
	b_OH_CH2_O_CH3,
	b_CH5P_CH2_CH4_CH3P,
	b_OH_CH3_CH4_O,
	b_OH_CH3_H2O_CH2,
	b_H2O_CH3_OH_CH4,
	b_CH3_CH3_CH4_CH2,
	b_OH_CH4_H2O_CH3,
	b_OHP_CH4_CH5P_O,
	b_OHP_CH4_H3OP_CH2,
	b_H2OP_CH4_H3OP_CH3,
	b_COP_CH4_HCOP_CH3,
	b_COP_CH4_CO_CH4P,
	b_CH4_CH4P_CH5P_CH3,
	b_H2O_CH4P_H3OP_CH3,
	b_CO_CH4P_HCOP_CH3,
	b_O2_CH4P_O2P_CH4,
	b_H2O_CH5P_H3OP_CH4,
	b_CO_CH5P_HCOP_CH4,
	b_CH5P_OH_H2OP_CH4,
	csav[NUM_COMOLE_CALC][NUM_COMOLE_CALC],
	esc_co,
	C_NH_N_CH_1,          
	C_NH_CN_H_1,         
	C_N2_CN_N_1,      
	C_NO_CN_O_1,      
	C_NO_CO_N_1,     
	C_HS_S_CH_1,    
	C_HS_CS_H_1,   
	C_OCN_CO_CN_1, 
	C_NS_S_CN_1,   
	C_NS_CS_N_1,   
	C_SO_S_CO_1,   
	C_S2_CS_S_1,       
	CH_N_NH_C_1,       
	CH_N_CN_H_1,       
	CH_N2_HCN_N_1,     
	CH_NO_HCN_O_1,     
	CH_NO_CN_OH_1,     
	CH_NO_OCN_H_1,     
	CH_HNO_NO_CH2_1,   
	CH_S_HS_C_1,       
	CH_S_CS_H_1,       
	N_NH_N2_H_1,       
	N_CH3_HCN_H_H_1,   
	N_CH3_HCN_H2_1,    
	N_OH_O_NH_1,       
	N_OH_NO_H_1,       
	N_CN_N2_C_1,       
	N_SIH_SIN_H_1,     
	N_NO_N2_O_1,       
	N_HNO_NO_NH_1,     
	N_HNO_N2O_H_1,     
	N_O2_NO_O_1,       
	N_HS_NS_H_1,       
	N_HS_S_NH_1,       
	N_CS_S_CN_1,       
	N_NS_S_N2_1,       
	N_SO_NS_O_1,       
	N_SO_S_NO_1,       
	N_S2_NS_S_1,       
	CH2_CN_HCN_CH_1,   
	CH2_NO_HCN_OH_1,   
	CH2_HNO_NO_CH3_1,  
	NH_NH_N2_H_H_1,    
	NH_NH_N2_H2_1,     
	NH_O_OH_N_1,       
	NH_O_NO_H_1,       
	NH_OH_HNO_H_1,     
	NH_OH_NH2_O_1,     
	NH_OH_H2O_N_1,     
	NH_H2O_OH_NH2_1,   
	NH_CN_HCN_N_1,     
	NH_NO_N2O_H_1,     
	NH_NO_N2_O_H_1,    
	NH_NO_N2_OH_1,     
	NH_S_HS_N_1,       
	NH_S_NS_H_1,       
	NH_NO2_HNO_NO_1,   
	NH_NO2_N2O_OH_1,   
	CH3_NH3_CH4_NH2_1, 
	CH3_CN_HCN_CH2_1,  
	CH3_HNO_NO_CH4_1,  
	O_NH2_OH_NH_1,     
	O_NH2_NO_H2_1,     
	O_NH3_OH_NH2_1,    
	O_CN_NO_C_1,       
	O_CN_CO_N_1,       
	O_HCN_CN_OH_1,     
	O_HCN_CO_NH_1,     
	O_HCN_OCN_H_1,     
	O_N2_NO_N_1,       
	O_NO_O2_N_1,       
	O_HNO_NO_OH_1,     
	O_HNO_O2_NH_1,  
	O_HNO_NO2_H_1, 
	O_HS_S_OH_1,  
	O_HS_SO_H_1, 
	O_OCN_NO_CO_1,
	O_OCN_O2_CN_1,
	O_SIN_NO_SI_1,
	O_SIN_SIO_N_1,
	O_N2O_NO_NO_1,
	O_N2O_O2_N2_1,
	O_CS_S_CO_1,  
	O_CS_SO_C_1,  
	O_NS_S_NO_1,  
	O_NS_SO_N_1,  
	O_SO_S_O2_1,  
	O_OCS_SO_CO_1,
	O_S2_SO_S_1,     
	NH2_OH_NH3_O_1,  
	NH2_OH_H2O_NH_1, 
	NH2_NO_N2_OH_H_1,
	NH2_NO_N2_H2O_1, 
	CH4_CN_HCN_CH3_1,
	OH_NH3_H2O_NH2_1,
	OH_CN_HCN_O_1,   
	OH_CN_OCN_H_1,     
	OH_HCN_CN_H2O_1,   
	OH_HCN_CO_NH2_1,   
	OH_NO_NO2_H_1,     
	OH_S_HS_O_1,       
	OH_S_SO_H_1,       
	OH_N2O_HNO_NO_1,   
	OH_CS_OCS_H_1,     
	NH3_CN_HCN_NH2_1,  
	CN_NO_N2_CO_1,     
	CN_NO_OCN_N_1,     
	CN_HNO_NO_HCN_1,   
	CN_O2_NO_CO_1,     
	CN_S_NS_C_1,       
	CN_S_CS_N_1,       
	CO_HS_OCS_H_1,     
	N2_O2_N2O_O_1,     
	NO_NO_O2_N2_1,     
	NO_NO_N2O_O_1,     
	NO_HNO_N2O_OH_1,   
	NO_S_NS_O_1,       
	NO_S_SO_N_1,       
	O2_S_SO_O_1,       
	S_SO_S2_O_1,       
	C_NHP_N_CHP_1,     
	CP_NH_CNP_H_1,     
	CP_NH2_HCNP_H_1,   
	C_NH2P_NH_CHP_1,   
	CP_NH3_HCNP_H2_1,  
	C_NH3P_NH_CH2P_1,  
	C_HCNP_CN_CHP_1,   
	C_HNOP_NO_CHP_1,   
	CP_HS_CSP_H_1,     
	C_HSP_CSP_H_1,     
	CP_OCN_COP_CN_1,   
	CP_NS_CSP_N_1,     
	CP_SO_SP_CO_1,     
	CP_SO_S_COP_1,     
	CP_SO_CSP_O_1,     
	CP_OCS_CSP_CO_1,   
	CHP_N_CNP_H_1,     
	CH_NP_CNP_H_1,     
	CH_NHP_CH2P_N_1,   
	CHP_NH_CNP_H2_1,   
	CHP_NH2_HCNP_H2_1, 
	CH_NH2P_NH_CH2P_1, 
	CHP_NH3_NH4P_C_1,  
	CH_NH3P_NH4P_C_1,  
	CH_HCNP_CN_CH2P_1, 
	CH_HNOP_NO_CH2P_1, 
	CHP_S_HSP_C_1,     
	CHP_S_CSP_H_1,     
	CH_SP_CSP_H_1,     
	CH_HSP_S_CH2P_1,   
	CHP_OCS_HCSP_CO_1, 
	N_CH2P_HCNP_H_1,   
	NP_NH_N2P_H_1,     
	N_NHP_N2P_H_1,     
	NP_CH4_HCNP_H2_H_1,
	N_OHP_NOP_H_1,     
	NP_NH3_NH2P_NH_1,  
	N_H2OP_HNOP_H_1,   
	NP_CO_NOP_C_1,     
	NP_NO_N2P_O_1,     
	NP_O2_NO_OP_1,     
	NP_O2_NOP_O_1,     
	N_O2P_NOP_O_1,    
	N_HSP_NSP_H_1,    
	N_SIOP_NOP_SI_1,  
	N_SIOP_NO_SIP_1,  
	N_SOP_NSP_O_1,    
	NP_OCS_SP_CO_N_1, 
	NP_OCS_CSP_NO_1,  
	CH2_NHP_CH3P_N_1, 
	CH2_NH2P_CH3P_NH_1,
	CH2P_NH3_NH4P_CH_1, 
	CH2_NH3P_NH2_CH3P_1,
	CH2_HCNP_CN_CH3P_1, 
	CH2_HNOP_NO_CH3P_1, 
	CH2P_S_HCSP_H_1,    
	CH2_SP_HCSP_H_1,    
	NHP_NH_NH2P_N_1,    
	NHP_O_OHP_N_1,      
	NH_OP_NOP_H_1,      
	NHP_NH2_NH3P_N_1,   
	NH_NH2P_NH3P_N_1,   
	NHP_OH_H2OP_N_1,    
	NH_OHP_NH2P_O_1,    
	NHP_NH3_NH4P_N_1,   
	NH_NH3P_NH4P_N_1,   
	NH_CH5P_CH4_NH2P_1, 
	NHP_H2O_NH3P_O_1,   
	NHP_H2O_H3OP_N_1,   
	NHP_H2O_HNOP_H2_1,    
	NH_H2OP_H3OP_N_1,     
	NHP_H2O_OH_NH2P_1,    
	NHP_CN_HCNP_N_1,      
	NH_HCNP_CN_NH2P_1,    
	NHP_CO_HCOP_N_1,      
	NH_COP_HCOP_N_1,      
	NHP_CO_OCNP_H_1,      
	NH_SIP_SINP_H_1,      
	NH_HCOP_CO_NH2P_1,    
	NH_HNOP_NO_NH2P_1,    
	NH_O2P_HNOP_O_1,      
	NHP_O2_NOP_OH_1,      
	NH_O2P_NO2P_H_1,      
	NHP_S_HSP_N_1,        
	NHP_S_NSP_H_1,        
	NH_SP_NSP_H_1,        
	CH3P_NH3_NH4P_CH2_1,  
	CH3P_S_HCSP_H2_1,     
	O_NH2P_HNOP_H_1,      
	O_NH3P_HNOP_H2_1,     
	OP_CN_NOP_C_1,        
	OP_HCN_COP_NH_1,      
	OP_HCN_NOP_CH_1,      
	OP_HCN_HCOP_N_1,      
	OP_N2_NOP_N_1,        
	O_N2P_NOP_N_1,        
	O_HNOP_NO2P_H_1,      
	O_HSP_SP_OH_1,       
	O_HSP_SOP_H_1,       
	O_SINP_SIOP_N_1,     
	OP_N2O_NOP_NO_1,     
	O_CSP_S_COP_1,       
	O_HCSP_S_HCOP_1,    
	O_HCSP_OCSP_H_1,    
	OP_NO2_O2_NOP_1,    
	O_NSP_S_NOP_1,      
	NH2P_NH2_NH3P_NH_1, 
	NH2_OHP_NH3P_O_1,   
	NH2P_OH_H2OP_NH_1,  
	NH2P_NH3_NH4P_NH_1, 
	NH2_NH3P_NH4P_NH_1,
	NH2_CH5P_NH3P_CH4_1, 
	NH2P_H2O_NH3P_OH_1,  
	NH2P_H2O_H3OP_NH_1,  
	NH2_H2OP_NH3P_OH_1,  
	NH2P_H2O_NH4P_O_1,   
	NH2_H3OP_H2O_NH3P_1, 
	NH2_HCNP_CN_NH3P_1,  
	NH2_COP_HCOP_NH_1,   
	NH2_HCOP_CO_NH3P_1,  
	NH2_HNOP_NO_NH3P_1,  
	NH2P_O2_HNOP_OH_1,   
	NH2P_S_HSP_NH_1,     
	CH4P_NH3_NH4P_CH3_1, 
	CH4_NH3P_NH4P_CH3_1, 
	CH4_N2P_N2_CH2P_H2_1,
	CH4_N2P_N2_CH3P_H_1, 
	CH4_HNOP_NO_CH5P_1,   
	CH4_SP_HCSP_H2_H_1,   
	CH4_CSP_HCSP_CH3_1,   
	OHP_NH3_NH4P_O_1,     
	OH_NH3P_NH4P_O_1,     
	OHP_CN_HCNP_O_1,      
	OH_HCNP_CN_H2OP_1,    
	OHP_NO_HNOP_O_1,      
	OH_HNOP_NO_H2OP_1,    
	OHP_S_HSP_O_1,        
	OHP_S_SOP_H_1,        
	OH_SP_SOP_H_1,        
	NH3P_NH3_NH4P_NH2_1,  
	NH3_CH5P_NH4P_CH4_1,  
	NH3P_H2O_NH4P_OH_1,   
	NH3_H2OP_NH4P_OH_1,   
	NH3_H3OP_NH4P_H2O_1,  
	NH3_COP_HCOP_NH2_1,   
	NH3_HCOP_CO_NH4P_1,   
	NH3_HNOP_NO_NH4P_1,   
	NH3_HSP_S_NH4P_1,     
	NH3_HCSP_CS_NH4P_1,   
	CH5P_S_HSP_CH4_1,     
	H2O_CNP_HCNP_OH_1,    
	H2O_CNP_HCOP_NH_1,    
	H2O_HCNP_CN_H3OP_1,   
	H2O_HNOP_NO_H3OP_1,   
	H2OP_S_HSP_OH_1,      
	H2O_HSP_S_H3OP_1,     
	H3OP_CS_HCSP_H2O_1,   
	CNP_NO_OCNP_N_1,      
	CN_HNOP_NO_HCNP_1,    
	CNP_O2_NOP_CO_1,      
	CNP_O2_OCNP_O_1,      
	HCNP_CO_HCOP_CN_1,    
	HCNP_S_HSP_CN_1,      
	CO_HNOP_NO_HCOP_1,    
	N2P_OCS_SP_N2_CO_1,   
	HCOP_S_HSP_CO_1,      
	HCOP_CS_HCSP_CO_1,    
	HNOP_S_HSP_NO_1,      
	O2_SP_SOP_O_1,        
	O2P_S_SOP_O_1,        
	O2_CSP_OCSP_O_1,      
	S_SIOP_SO_SIP_1,      
	CP_NH3_NH3P_C_1,     
	C_CNP_CN_CP_1,       
	C_N2P_N2_CP_1,      
	CP_NO_NOP_C_1,      
	CP_SIN_SINP_C_1,    
	C_CSP_CS_CP_1,      
	CP_NS_NSP_C_1,      
	CP_SO_SOP_C_1,      
	CP_OCS_OCSP_C_1,    
	CH_NH2P_NH2_CHP_1,  
	CHP_NH3_NH3P_CH_1,  
	CH_CNP_CN_CHP_1,    
	CH_N2P_N2_CHP_1,    
	CHP_NO_NOP_CH_1,    
	NP_NH_NHP_N_1,      
	NP_NH2_NH2P_N_1,    
	NP_NH3_NH3P_N_1,    
	NP_CN_CNP_N_1,      
	NP_HCN_HCNP_N_1,    
	N_N2P_N2_NP_1,      
	NP_NO_NOP_N_1,      
	NP_OCS_OCSP_N_1,    
	CH2_NH2P_NH2_CH2P_1,
	CH2_CNP_CN_CH2P_1,  
	CH2_N2P_N2_CH2P_1,  
	CH2P_NO_NOP_CH2_1,  
	NH_OP_O_NHP_1,      
	NH_OHP_OH_NHP_1,    
	NHP_NH3_NH3P_NH_1,  
	NHP_H2O_H2OP_NH_1,  
	NH_CNP_CN_NHP_1,    
	NH_COP_CO_NHP_1,    
	NH_N2P_N2_NHP_1,    
	NHP_NO_NOP_NH_1,    
	NHP_O2_O2P_NH_1,    
	NHP_S_SP_NH_1,      
	CH3P_NO_NOP_CH3_1,  
	OP_NH2_NH2P_O_1,    
	OP_NH3_NH3P_O_1,    
	O_CNP_CN_OP_1,      
	O_HCNP_HCN_OP_1,    
	O_N2P_N2_OP_1,      
	OP_NO_NOP_O_1,      
	OP_OCS_OCSP_O_1,    
	NH2_OHP_OH_NH2P_1,  
	NH2P_NH3_NH3P_NH2_1,
	NH2_H2OP_H2O_NH2P_1,  
	NH2_CNP_CN_NH2P_1,    
	NH2_COP_CO_NH2P_1,    
	NH2_N2P_N2_NH2P_1,    
	NH2P_NO_NOP_NH2_1,    
	NH2_O2P_O2_NH2P_1,    
	NH2P_S_SP_NH2_1,      
	CH4P_NH3_NH3P_CH4_1,  
	CH4P_OCS_OCSP_CH4_1,  
	OHP_NH3_NH3P_OH_1,    
	OH_CNP_CN_OHP_1,      
	OH_N2P_N2_OHP_1,      
	OHP_NO_NOP_OH_1,      
	NH3_H2OP_H2O_NH3P_1,  
	NH3_HCNP_HCN_NH3P_1,  
	NH3_COP_CO_NH3P_1,    
	NH3_N2P_N2_NH3P_1,    
	NH3P_SI_SIP_NH3_1,    
	NH3_O2P_O2_NH3P_1,    
	NH3_SP_S_NH3P_1,      
	NH3_HSP_HS_NH3P_1,    
	NH3_SOP_SO_NH3P_1,    
	H2O_HCNP_HCN_H2OP_1,  
	H2O_N2P_N2_H2OP_1,    
	H2OP_NO_NOP_H2O_1,    
	CNP_HCN_HCNP_CN_1,    
	CNP_CO_COP_CN_1,      
	CN_N2P_N2_CNP_1,      
	CNP_NO_NOP_CN_1,      
	CNP_O2_O2P_CN_1,      
	CNP_S_SP_CN_1,        
	HCN_COP_CO_HCNP_1,    
	HCN_N2P_N2_HCNP_1,    
	HCNP_NO_NOP_HCN_1,    
	HCNP_O2_O2P_HCN_1,    
	HCNP_S_SP_HCN_1,      
	CO_N2P_N2_COP_1,      
	COP_NO_NOP_CO_1,      
	N2P_NO_NOP_N2_1,      
	N2P_O2_O2P_N2_1,      
	N2P_S_SP_N2_1,        
	SI_NOP_NO_SIP_1,      
	SI_HSP_HS_SIP_1,      
	SI_CSP_CS_SIP_1,      
	NO_HNOP_HNO_NOP_1,    
	NO_O2P_O2_NOP_1,      
	NO_SP_S_NOP_1,        
	NO_HSP_HS_NOP_1,      
	NO_SIOP_SIO_NOP_1,    
	NO_S2P_S2_NOP_1,      
	O2P_NO2_NO2P_O2_1,    
	S_HSP_HS_SP_1,        
	C_N_CN_NU_1,          
	C_S_CS_NU_1,          
	CP_S_CSP_NU_1,        
	NP_N_N2P_NU_1,        
	CO_S_OCS_NU_1,        
	CH_NP_N_CHP_1,        
	CHP_S_SP_CH_1,        
	NP_CH2_CH2P_N_1,      
	NP_CH4_CH4P_N_1,      
	NP_OH_OHP_N_1,        
	NP_H2O_H2OP_N_1,      
	NP_CO_COP_N_1,        
	NP_O2_O2P_N_1,        
	OHP_S_SP_OH_1,        
	H2OP_S_SP_H2O_1,      
	COP_S_SP_CO_1,        
	SI_SP_S_SIP_1,        
	O2P_S_SP_O2_1,        
	C_NH_N_CH_2,          
	C_NH_CN_H_2,         
	C_N2_CN_N_2,      
	C_NO_CN_O_2,      
	C_NO_CO_N_2,     
	C_HS_S_CH_2,    
	C_HS_CS_H_2,   
	C_OCN_CO_CN_2, 
	C_NS_S_CN_2,   
	C_NS_CS_N_2,   
	C_SO_S_CO_2,   
	C_S2_CS_S_2,       
	CH_N_NH_C_2,       
	CH_N_CN_H_2,       
	CH_N2_HCN_N_2,     
	CH_NO_HCN_O_2,     
	CH_NO_CN_OH_2,     
	CH_NO_OCN_H_2,     
	CH_HNO_NO_CH2_2,   
	CH_S_HS_C_2,       
	CH_S_CS_H_2,       
	N_NH_N2_H_2,       
	N_CH3_HCN_H_H_2,   
	N_CH3_HCN_H2_2,    
	N_OH_O_NH_2,       
	N_OH_NO_H_2,       
	N_CN_N2_C_2,       
	N_SIH_SIN_H_2,     
	N_NO_N2_O_2,       
	N_HNO_NO_NH_2,     
	N_HNO_N2O_H_2,     
	N_O2_NO_O_2,       
	N_HS_NS_H_2,       
	N_HS_S_NH_2,       
	N_CS_S_CN_2,       
	N_NS_S_N2_2,       
	N_SO_NS_O_2,       
	N_SO_S_NO_2,       
	N_S2_NS_S_2,       
	CH2_CN_HCN_CH_2,   
	CH2_NO_HCN_OH_2,   
	CH2_HNO_NO_CH3_2,  
	NH_NH_N2_H_H_2,    
	NH_NH_N2_H2_2,     
	NH_O_OH_N_2,       
	NH_O_NO_H_2,       
	NH_OH_HNO_H_2,     
	NH_OH_NH2_O_2,     
	NH_OH_H2O_N_2,     
	NH_H2O_OH_NH2_2,   
	NH_CN_HCN_N_2,     
	NH_NO_N2O_H_2,     
	NH_NO_N2_O_H_2,    
	NH_NO_N2_OH_2,     
	NH_S_HS_N_2,       
	NH_S_NS_H_2,       
	NH_NO2_HNO_NO_2,   
	NH_NO2_N2O_OH_2,   
	CH3_NH3_CH4_NH2_2, 
	CH3_CN_HCN_CH2_2,  
	CH3_HNO_NO_CH4_2,  
	O_NH2_OH_NH_2,     
	O_NH2_NO_H2_2,     
	O_NH3_OH_NH2_2,    
	O_CN_NO_C_2,       
	O_CN_CO_N_2,       
	O_HCN_CN_OH_2,     
	O_HCN_CO_NH_2,     
	O_HCN_OCN_H_2,     
	O_N2_NO_N_2,       
	O_NO_O2_N_2,       
	O_HNO_NO_OH_2,     
	O_HNO_O2_NH_2,  
	O_HNO_NO2_H_2, 
	O_HS_S_OH_2,  
	O_HS_SO_H_2, 
	O_OCN_NO_CO_2,
	O_OCN_O2_CN_2,
	O_SIN_NO_SI_2,
	O_SIN_SIO_N_2,
	O_N2O_NO_NO_2,
	O_N2O_O2_N2_2,
	O_CS_S_CO_2,  
	O_CS_SO_C_2,  
	O_NS_S_NO_2,  
	O_NS_SO_N_2,  
	O_SO_S_O2_2,  
	O_OCS_SO_CO_2,
	O_S2_SO_S_2,     
	NH2_OH_NH3_O_2,  
	NH2_OH_H2O_NH_2, 
	NH2_NO_N2_OH_H_2,
	NH2_NO_N2_H2O_2, 
	CH4_CN_HCN_CH3_2,
	OH_NH3_H2O_NH2_2,
	OH_CN_HCN_O_2,   
	OH_CN_OCN_H_2,     
	OH_HCN_CN_H2O_2,   
	OH_HCN_CO_NH2_2,   
	OH_NO_NO2_H_2,     
	OH_S_HS_O_2,       
	OH_S_SO_H_2,       
	OH_N2O_HNO_NO_2,   
	OH_CS_OCS_H_2,     
	NH3_CN_HCN_NH2_2,  
	CN_NO_N2_CO_2,     
	CN_NO_OCN_N_2,     
	CN_HNO_NO_HCN_2,   
	CN_O2_NO_CO_2,     
	CN_S_NS_C_2,       
	CN_S_CS_N_2,       
	CO_HS_OCS_H_2,     
	N2_O2_N2O_O_2,     
	NO_NO_O2_N2_2,     
	NO_NO_N2O_O_2,     
	NO_HNO_N2O_OH_2,   
	NO_S_NS_O_2,       
	NO_S_SO_N_2,       
	O2_S_SO_O_2,       
	S_SO_S2_O_2,       
	C_NHP_N_CHP_2,     
	CP_NH_CNP_H_2,     
	CP_NH2_HCNP_H_2,   
	C_NH2P_NH_CHP_2,   
	CP_NH3_HCNP_H2_2,  
	C_NH3P_NH_CH2P_2,  
	C_HCNP_CN_CHP_2,   
	C_HNOP_NO_CHP_2,   
	CP_HS_CSP_H_2,     
	C_HSP_CSP_H_2,     
	CP_OCN_COP_CN_2,   
	CP_NS_CSP_N_2,     
	CP_SO_SP_CO_2,     
	CP_SO_S_COP_2,     
	CP_SO_CSP_O_2,     
	CP_OCS_CSP_CO_2,   
	CHP_N_CNP_H_2,     
	CH_NP_CNP_H_2,     
	CH_NHP_CH2P_N_2,   
	CHP_NH_CNP_H2_2,   
	CHP_NH2_HCNP_H2_2, 
	CH_NH2P_NH_CH2P_2, 
	CHP_NH3_NH4P_C_2,  
	CH_NH3P_NH4P_C_2,  
	CH_HCNP_CN_CH2P_2, 
	CH_HNOP_NO_CH2P_2, 
	CHP_S_HSP_C_2,     
	CHP_S_CSP_H_2,     
	CH_SP_CSP_H_2,     
	CH_HSP_S_CH2P_2,   
	CHP_OCS_HCSP_CO_2, 
	N_CH2P_HCNP_H_2,   
	NP_NH_N2P_H_2,     
	N_NHP_N2P_H_2,     
	NP_CH4_HCNP_H2_H_2,
	N_OHP_NOP_H_2,     
	NP_NH3_NH2P_NH_2,  
	N_H2OP_HNOP_H_2,   
	NP_CO_NOP_C_2,     
	NP_NO_N2P_O_2,     
	NP_O2_NO_OP_2,     
	NP_O2_NOP_O_2,     
	N_O2P_NOP_O_2,    
	N_HSP_NSP_H_2,    
	N_SIOP_NOP_SI_2,  
	N_SIOP_NO_SIP_2,  
	N_SOP_NSP_O_2,    
	NP_OCS_SP_CO_N_2, 
	NP_OCS_CSP_NO_2,  
	CH2_NHP_CH3P_N_2, 
	CH2_NH2P_CH3P_NH_2,
	CH2P_NH3_NH4P_CH_2, 
	CH2_NH3P_NH2_CH3P_2,
	CH2_HCNP_CN_CH3P_2, 
	CH2_HNOP_NO_CH3P_2, 
	CH2P_S_HCSP_H_2,    
	CH2_SP_HCSP_H_2,    
	NHP_NH_NH2P_N_2,    
	NHP_O_OHP_N_2,      
	NH_OP_NOP_H_2,      
	NHP_NH2_NH3P_N_2,   
	NH_NH2P_NH3P_N_2,   
	NHP_OH_H2OP_N_2,    
	NH_OHP_NH2P_O_2,    
	NHP_NH3_NH4P_N_2,   
	NH_NH3P_NH4P_N_2,   
	NH_CH5P_CH4_NH2P_2, 
	NHP_H2O_NH3P_O_2,   
	NHP_H2O_H3OP_N_2,   
	NHP_H2O_HNOP_H2_2,    
	NH_H2OP_H3OP_N_2,     
	NHP_H2O_OH_NH2P_2,    
	NHP_CN_HCNP_N_2,      
	NH_HCNP_CN_NH2P_2,    
	NHP_CO_HCOP_N_2,      
	NH_COP_HCOP_N_2,      
	NHP_CO_OCNP_H_2,      
	NH_SIP_SINP_H_2,      
	NH_HCOP_CO_NH2P_2,    
	NH_HNOP_NO_NH2P_2,    
	NH_O2P_HNOP_O_2,      
	NHP_O2_NOP_OH_2,      
	NH_O2P_NO2P_H_2,      
	NHP_S_HSP_N_2,        
	NHP_S_NSP_H_2,        
	NH_SP_NSP_H_2,        
	CH3P_NH3_NH4P_CH2_2,  
	CH3P_S_HCSP_H2_2,     
	O_NH2P_HNOP_H_2,      
	O_NH3P_HNOP_H2_2,     
	OP_CN_NOP_C_2,        
	OP_HCN_COP_NH_2,      
	OP_HCN_NOP_CH_2,      
	OP_HCN_HCOP_N_2,      
	OP_N2_NOP_N_2,        
	O_N2P_NOP_N_2,        
	O_HNOP_NO2P_H_2,      
	O_HSP_SP_OH_2,       
	O_HSP_SOP_H_2,       
	O_SINP_SIOP_N_2,     
	OP_N2O_NOP_NO_2,     
	O_CSP_S_COP_2,       
	O_HCSP_S_HCOP_2,    
	O_HCSP_OCSP_H_2,    
	OP_NO2_O2_NOP_2,    
	O_NSP_S_NOP_2,      
	NH2P_NH2_NH3P_NH_2, 
	NH2_OHP_NH3P_O_2,   
	NH2P_OH_H2OP_NH_2,  
	NH2P_NH3_NH4P_NH_2, 
	NH2_NH3P_NH4P_NH_2,
	NH2_CH5P_NH3P_CH4_2, 
	NH2P_H2O_NH3P_OH_2,  
	NH2P_H2O_H3OP_NH_2,  
	NH2_H2OP_NH3P_OH_2,  
	NH2P_H2O_NH4P_O_2,   
	NH2_H3OP_H2O_NH3P_2, 
	NH2_HCNP_CN_NH3P_2,  
	NH2_COP_HCOP_NH_2,   
	NH2_HCOP_CO_NH3P_2,  
	NH2_HNOP_NO_NH3P_2,  
	NH2P_O2_HNOP_OH_2,   
	NH2P_S_HSP_NH_2,     
	CH4P_NH3_NH4P_CH3_2, 
	CH4_NH3P_NH4P_CH3_2, 
	CH4_N2P_N2_CH2P_H2_2,
	CH4_N2P_N2_CH3P_H_2, 
	CH4_HNOP_NO_CH5P_2,   
	CH4_SP_HCSP_H2_H_2,   
	CH4_CSP_HCSP_CH3_2,   
	OHP_NH3_NH4P_O_2,     
	OH_NH3P_NH4P_O_2,     
	OHP_CN_HCNP_O_2,      
	OH_HCNP_CN_H2OP_2,    
	OHP_NO_HNOP_O_2,      
	OH_HNOP_NO_H2OP_2,    
	OHP_S_HSP_O_2,        
	OHP_S_SOP_H_2,        
	OH_SP_SOP_H_2,        
	NH3P_NH3_NH4P_NH2_2,  
	NH3_CH5P_NH4P_CH4_2,  
	NH3P_H2O_NH4P_OH_2,   
	NH3_H2OP_NH4P_OH_2,   
	NH3_H3OP_NH4P_H2O_2,  
	NH3_COP_HCOP_NH2_2,   
	NH3_HCOP_CO_NH4P_2,   
	NH3_HNOP_NO_NH4P_2,   
	NH3_HSP_S_NH4P_2,     
	NH3_HCSP_CS_NH4P_2,   
	CH5P_S_HSP_CH4_2,     
	H2O_CNP_HCNP_OH_2,    
	H2O_CNP_HCOP_NH_2,    
	H2O_HCNP_CN_H3OP_2,   
	H2O_HNOP_NO_H3OP_2,   
	H2OP_S_HSP_OH_2,      
	H2O_HSP_S_H3OP_2,     
	H3OP_CS_HCSP_H2O_2,   
	CNP_NO_OCNP_N_2,      
	CN_HNOP_NO_HCNP_2,    
	CNP_O2_NOP_CO_2,      
	CNP_O2_OCNP_O_2,      
	HCNP_CO_HCOP_CN_2,    
	HCNP_S_HSP_CN_2,      
	CO_HNOP_NO_HCOP_2,    
	N2P_OCS_SP_N2_CO_2,   
	HCOP_S_HSP_CO_2,      
	HCOP_CS_HCSP_CO_2,    
	HNOP_S_HSP_NO_2,      
	O2_SP_SOP_O_2,        
	O2P_S_SOP_O_2,        
	O2_CSP_OCSP_O_2,      
	S_SIOP_SO_SIP_2,      
	CP_NH3_NH3P_C_2,     
	C_CNP_CN_CP_2,       
	C_N2P_N2_CP_2,      
	CP_NO_NOP_C_2,      
	CP_SIN_SINP_C_2,    
	C_CSP_CS_CP_2,      
	CP_NS_NSP_C_2,      
	CP_SO_SOP_C_2,      
	CP_OCS_OCSP_C_2,    
	CH_NH2P_NH2_CHP_2,  
	CHP_NH3_NH3P_CH_2,  
	CH_CNP_CN_CHP_2,    
	CH_N2P_N2_CHP_2, 
	CH_NP_N_CHP_2,
	CHP_NO_NOP_CH_2,    
	NP_NH_NHP_N_2,      
	NP_NH2_NH2P_N_2,    
	NP_NH3_NH3P_N_2,    
	NP_CN_CNP_N_2,      
	NP_HCN_HCNP_N_2,    
	N_N2P_N2_NP_2,      
	NP_NO_NOP_N_2,      
	NP_OCS_OCSP_N_2,    
	CH2_NH2P_NH2_CH2P_2,
	CH2_CNP_CN_CH2P_2,  
	CH2_N2P_N2_CH2P_2,  
	CH2P_NO_NOP_CH2_2,  
	NH_OP_O_NHP_2,      
	NH_OHP_OH_NHP_2,    
	NHP_NH3_NH3P_NH_2,  
	NHP_H2O_H2OP_NH_2,  
	NH_CNP_CN_NHP_2,    
	NH_COP_CO_NHP_2,    
	NH_N2P_N2_NHP_2,    
	NHP_NO_NOP_NH_2,    
	NHP_O2_O2P_NH_2,    
	NHP_S_SP_NH_2,      
	CH3P_NO_NOP_CH3_2,  
	OP_NH2_NH2P_O_2,    
	OP_NH3_NH3P_O_2,    
	O_CNP_CN_OP_2,      
	O_HCNP_HCN_OP_2,    
	O_N2P_N2_OP_2,      
	OP_NO_NOP_O_2,      
	OP_OCS_OCSP_O_2,    
	NH2_OHP_OH_NH2P_2,  
	NH2P_NH3_NH3P_NH2_2,
	NH2_H2OP_H2O_NH2P_2,  
	NH2_CNP_CN_NH2P_2,    
	NH2_COP_CO_NH2P_2,    
	NH2_N2P_N2_NH2P_2,    
	NH2P_NO_NOP_NH2_2,    
	NH2_O2P_O2_NH2P_2,    
	NH2P_S_SP_NH2_2,      
	CH4P_NH3_NH3P_CH4_2,  
	CH4P_OCS_OCSP_CH4_2,  
	OHP_NH3_NH3P_OH_2,    
	OH_CNP_CN_OHP_2,      
	OH_N2P_N2_OHP_2,      
	OHP_NO_NOP_OH_2,      
	NH3_H2OP_H2O_NH3P_2,  
	NH3_HCNP_HCN_NH3P_2,  
	NH3_COP_CO_NH3P_2,    
	NH3_N2P_N2_NH3P_2,    
	NH3P_SI_SIP_NH3_2,    
	NH3_O2P_O2_NH3P_2,    
	NH3_SP_S_NH3P_2,      
	NH3_HSP_HS_NH3P_2,    
	NH3_SOP_SO_NH3P_2,    
	H2O_HCNP_HCN_H2OP_2,  
	H2O_N2P_N2_H2OP_2,    
	H2OP_NO_NOP_H2O_2,    
	CNP_HCN_HCNP_CN_2,    
	CNP_CO_COP_CN_2,      
	CN_N2P_N2_CNP_2,      
	CNP_NO_NOP_CN_2,      
	CNP_O2_O2P_CN_2,      
	CNP_S_SP_CN_2,        
	HCN_COP_CO_HCNP_2,    
	HCN_N2P_N2_HCNP_2,    
	HCNP_NO_NOP_HCN_2,    
	HCNP_O2_O2P_HCN_2,    
	HCNP_S_SP_HCN_2,      
	CO_N2P_N2_COP_2,      
	COP_NO_NOP_CO_2,      
	N2P_NO_NOP_N2_2,      
	N2P_O2_O2P_N2_2,      
	N2P_S_SP_N2_2,        
	SI_NOP_NO_SIP_2,      
	SI_HSP_HS_SIP_2,      
	SI_CSP_CS_SIP_2,      
	NO_HNOP_HNO_NOP_2,    
	NO_O2P_O2_NOP_2,      
	NO_SP_S_NOP_2,        
	NO_HSP_HS_NOP_2,      
	NO_SIOP_SIO_NOP_2,    
	NO_S2P_S2_NOP_2,      
	O2P_NO2_NO2P_O2_2,    
	S_HSP_HS_SP_2,        
	C_N_CN_NU_2,          
	C_S_CS_NU_2,          
	CP_S_CSP_NU_2,        
	NP_N_N2P_NU_2,        
	CO_S_OCS_NU_2,        
	CHP_S_SP_CH_2,        
	NP_CH2_CH2P_N_2,      
	NP_CH4_CH4P_N_2,      
	NP_OH_OHP_N_2,        
	NP_H2O_H2OP_N_2,      
	NP_CO_COP_N_2,        
	NP_O2_O2P_N_2,        
	OHP_S_SP_OH_2,        
	H2OP_S_SP_H2O_2,      
	COP_S_SP_CO_2,        
	SI_SP_S_SIP_2,        
	O2P_S_SP_O2_2, 
	H2_N_NH_H          , 
	H2_NH_NH2_H         ,
	H2_NH2_NH3_H        ,
	H2_CN_HCN_H         ,
	HP_HNO_NOP_H2       ,
	HP_HS_SP_H2         ,
	H_HSP_SP_H2         ,
	HP_OCS_HSP_CO       ,
	H2P_N_NHP_H         ,
	H2_NP_NHP_H         ,
	H2_NHP_N_H3P        ,
	H2P_NH_NH2P_H       ,
	H2_NHP_NH2P_H       ,
	H2_NH2P_NH3P_H      ,
	H2_NH3P_NH4P_H      ,
	H2P_CN_HCNP_H      ,
	H2_CNP_HCNP_H       ,
	H2P_NO_HNOP_H       ,
	H2_SP_HSP_H         ,
	H2_CSP_HCSP_H       ,
	H2_NO2P_NOP_H2O     ,
	H3P_NH_NH2P_H2      ,
	H3P_NH2_NH3P_H2     ,
	H3P_NH3_NH4P_H2     ,
	H3P_CN_HCNP_H2      ,
	H3P_NO_HNOP_H2      ,
	H3P_S_HSP_H2        ,
	H3P_CS_HCSP_H2      ,
	H3P_NO2_NOP_OH_H2   ,
	HeP_NH_NP_He_H      ,
	HeP_NH2_NP_He_H2    ,
	HeP_NH2_NHP_He_H    ,
	HeP_NH3_NHP_He_H2   ,
	HeP_NH3_NH2P_He_H   ,
	HeP_CN_N_CP_He      ,
	HeP_CN_NP_C_He      ,
	HeP_HCN_N_CHP_He    ,
	HeP_HCN_NP_CH_He    ,
	HeP_HCN_N_CP_He_H   ,
	HeP_HCN_CNP_He_H    ,
	HeP_N2_NP_N_He      ,
	HeP_NO_OP_N_He      ,
	HeP_NO_O_NP_He      ,
	HeP_HNO_NOP_He_H    ,
	HeP_HNO_NO_He_HP    ,
	HeP_HS_SP_He_H      ,
	HeP_OCN_CN_OP_He    ,
	HeP_OCN_CNP_O_He    ,
	HeP_SIN_SIP_N_He    ,
	HeP_N2O_N2_OP_He    ,
	HeP_N2O_N2P_O_He    ,
	HeP_N2O_NO_NP_He    ,
	HeP_N2O_NOP_N_He,
	HeP_CS_SP_C_He  ,
	HeP_CS_S_CP_He  ,
	HeP_NS_S_NP_He  ,
	HeP_NS_SP_N_He  ,
	HeP_SO_S_OP_He  ,
	HeP_SO_SP_O_He  ,
	HeP_OCS_SP_CO_He,
	HeP_OCS_S_COP_He,
	HeP_OCS_CSP_O_He,
	HeP_OCS_CS_OP_He,
	HeP_S2_SP_S_He  ,
	HP_NH_NHP_H ,
	HP_NH2_NH2P_H   ,
	HP_NH3_NH3P_H   ,
	H_CNP_CN_HP ,
	HP_HCN_HCNP_H   ,
	H_HCNP_HCN_HP   ,
	H_N2P_N2_HP ,
	HP_NO_NOP_H ,
	HP_HS_HSP_H ,
	HP_SIN_SINP_H   ,
	HP_CS_CSP_H ,
	HP_NS_NSP_H ,
	HP_SO_SOP_H ,
	HP_OCS_OCSP_H   ,
	HP_S2_S2P_H ,
	H2P_NH_NHP_H2   ,
	H2P_NH2_NH2P_H2 ,
	H2P_NH3_NH3P_H2 ,
	H2P_CN_CNP_H2   ,
	H2P_HCN_HCNP_H2 ,
	H2P_NO_NOP_H2   ,
	HeP_NH3_NH3P_He ,
	HeP_N2_N2P_He   ,
	NHP_e_N_H ,
	NH2P_e_N_H_H  ,
	NH2P_e_NH_H   ,
	NH3P_e_NH_H_H ,
	NH3P_e_NH2_H  ,
	NH4P_e_NH2_H_H,
	NH4P_e_NH2_H2 ,
	NH4P_e_NH3_H  ,
	CNP_e_N_C ,
	HCNP_e_CN_H   ,
	N2P_e_N_N ,
	NOP_e_O_N ,
	HNOP_e_NO_H   ,
	HSP_e_S_H ,
	OCNP_e_CO_N   ,
	SINP_e_SI_N   ,
	CSP_e_S_C ,
	HCSP_e_CS_H   ,
	NO2P_e_NO_O   ,
	NSP_e_S_N ,
	SOP_e_S_O ,
	OCSP_e_S_CO   ,
	OCSP_e_CS_O   ,
	S2P_e_S_S ,
	Hminus_N_NH_e ,
	Hminus_NH_NH2_e   ,
	Hminus_NH2_NH3_e  ,
	Hminus_CN_HCN_e   ,
	Hminus_NH4P_NH3_H2,
	Hminus_NP_N_H ,
	NH_NU_NHP_e   ,
	NH_NU_N_H ,
	NH2_NU_NH_H   ,
	NH2_NU_NH2P_e ,
	NH3_NU_NH3P_e ,
	NH3_NU_NH_H2  ,
	NH3_NU_NH2_H  ,
	CN_NU_N_C ,
	HCN_NU_CN_H   ,
	N2_NU_N_N ,
	NO_NU_O_N ,
	NO_NU_NOP_e   ,
	HNO_NU_NO_H   ,
	HS_NU_S_H ,
	HSP_NU_SP_H   ,
	OCN_NU_CN_O   ,
	CS_NU_S_C ,
	CSP_NU_SP_C   ,
	CS_NU_CSP_e   ,
	NO2_NU_NO_O   ,
	NS_NU_S_N ,
	SO_NU_S_O ,
	SO_NU_SOP_e   ,
	OCS_NU_S_CO   ,
	OCS_NU_OCSP_e ,
	NH_CRP_NHP_e,
	NH_CRP_N_H   ,     
	NH2_CRP_NH_H  ,    
	NH2_CRP_NH2P_e ,   
	NH3_CRP_NH3P_e  ,  
	NH3_CRP_NH_H2    , 
	NH3_CRP_NH2_H     ,
	CN_CRP_N_C        ,
	HCN_CRP_CN_H      ,
	N2_CRP_N_N        ,
	NO_CRP_O_N        ,
	NO_CRP_NOP_e      ,
	HNO_CRP_NO_H      ,
	HS_CRP_S_H        ,
	OCN_CRP_CN_O      ,
	CS_CRP_S_C        ,
	CS_CRP_CSP_e      ,
	NO2_CRP_NO_O      ,
	NS_CRP_S_N        ,
	SO_CRP_S_O        ,
	SO_CRP_SOP_e      ,
	OCS_CRP_S_CO      ,
	OCS_CRP_OCSP_e    ,
	b_C_NH_N_CH ,  
	b_C_NH_CN_H , 
	b_C_N2_CN_N ,  
	b_C_NO_CN_O ,  
	b_C_NO_CO_N , 
	b_C_HS_S_CH ,
	b_C_HS_CS_H ,   
	b_C_OCN_CO_CN , 
	b_C_NS_S_CN ,   
	b_C_NS_CS_N ,   
	b_C_SO_S_CO ,   
	b_C_S2_CS_S ,   
	b_CH_N_NH_C ,   
	b_CH_N_CN_H ,   
	b_CH_N2_HCN_N , 
	b_CH_NO_HCN_O , 
	b_CH_NO_CN_OH , 
	b_CH_NO_OCN_H , 
	b_CH_HNO_NO_CH2 ,   
	b_CH_S_HS_C ,   
	b_CH_S_CS_H ,   
	b_N_NH_N2_H ,   
	b_N_CH3_HCN_H_H ,   
	b_N_CH3_HCN_H2 ,
	b_N_OH_O_NH ,   
	b_N_OH_NO_H ,   
	b_N_CN_N2_C ,   
	b_N_SIH_SIN_H , 
	b_N_NO_N2_O ,   
	b_N_HNO_NO_NH , 
	b_N_HNO_N2O_H , 
	b_N_O2_NO_O ,   
	b_N_HS_NS_H ,   
	b_N_HS_S_NH ,   
	b_N_CS_S_CN ,   
	b_N_NS_S_N2 ,   
	b_N_SO_NS_O ,   
	b_N_SO_S_NO ,   
	b_N_S2_NS_S ,   
	b_CH2_CN_HCN_CH ,   
	b_CH2_NO_HCN_OH ,   
	b_CH2_HNO_NO_CH3 ,  
	b_NH_NH_N2_H_H ,
	b_NH_NH_N2_H2 , 
	b_NH_O_OH_N ,   
	b_NH_O_NO_H ,   
	b_NH_OH_HNO_H , 
	b_NH_OH_NH2_O , 
	b_NH_OH_H2O_N , 
	b_NH_H2O_OH_NH2 ,   
	b_NH_CN_HCN_N , 
	b_NH_NO_N2O_H , 
	b_NH_NO_N2_O_H ,
	b_NH_NO_N2_OH , 
	b_NH_S_HS_N ,       
	b_NH_S_NS_H ,       
	b_NH_NO2_HNO_NO ,   
	b_NH_NO2_N2O_OH ,   
	b_CH3_NH3_CH4_NH2 , 
	b_CH3_CN_HCN_CH2 ,  
	b_CH3_HNO_NO_CH4 ,  
	b_O_NH2_OH_NH ,     
	b_O_NH2_NO_H2 ,     
	b_O_NH3_OH_NH2 ,    
	b_O_CN_NO_C ,       
	b_O_CN_CO_N ,       
	b_O_HCN_CN_OH ,     
	b_O_HCN_CO_NH ,     
	b_O_HCN_OCN_H ,     
	b_O_N2_NO_N ,       
	b_O_NO_O2_N ,       
	b_O_HNO_NO_OH ,     
	b_O_HNO_O2_NH ,  
	b_O_HNO_NO2_H , 
	b_O_HS_S_OH ,  
	b_O_HS_SO_H , 
	b_O_OCN_NO_CO ,
	b_O_OCN_O2_CN ,
	b_O_SIN_NO_SI ,
	b_O_SIN_SIO_N ,
	b_O_N2O_NO_NO ,
	b_O_N2O_O2_N2 ,
	b_O_CS_S_CO ,  
	b_O_CS_SO_C ,  
	b_O_NS_S_NO ,  
	b_O_NS_SO_N ,  
	b_O_SO_S_O2 ,  
	b_O_OCS_SO_CO ,
	b_O_S2_SO_S ,     
	b_NH2_OH_NH3_O ,  
	b_NH2_OH_H2O_NH , 
	b_NH2_NO_N2_OH_H ,
	b_NH2_NO_N2_H2O , 
	b_CH4_CN_HCN_CH3 ,
	b_OH_NH3_H2O_NH2 ,
	b_OH_CN_HCN_O ,   
	b_OH_CN_OCN_H ,     
	b_OH_HCN_CN_H2O ,   
	b_OH_HCN_CO_NH2 ,   
	b_OH_NO_NO2_H ,     
	b_OH_S_HS_O ,       
	b_OH_S_SO_H ,       
	b_OH_N2O_HNO_NO ,   
	b_OH_CS_OCS_H ,     
	b_NH3_CN_HCN_NH2 ,  
	b_CN_NO_N2_CO ,     
	b_CN_NO_OCN_N ,     
	b_CN_HNO_NO_HCN ,   
	b_CN_O2_NO_CO ,     
	b_CN_S_NS_C ,       
	b_CN_S_CS_N ,       
	b_CO_HS_OCS_H ,     
	b_N2_O2_N2O_O ,     
	b_NO_NO_O2_N2 ,     
	b_NO_NO_N2O_O ,     
	b_NO_HNO_N2O_OH ,   
	b_NO_S_NS_O ,       
	b_NO_S_SO_N ,       
	b_O2_S_SO_O ,       
	b_S_SO_S2_O ,       
	b_C_NHP_N_CHP ,     
	b_CP_NH_CNP_H ,     
	b_CP_NH2_HCNP_H ,   
	b_C_NH2P_NH_CHP ,   
	b_CP_NH3_HCNP_H2 ,  
	b_C_NH3P_NH_CH2P ,  
	b_C_HCNP_CN_CHP ,   
	b_C_HNOP_NO_CHP ,   
	b_CP_HS_CSP_H ,     
	b_C_HSP_CSP_H ,     
	b_CP_OCN_COP_CN ,   
	b_CP_NS_CSP_N ,     
	b_CP_SO_SP_CO ,     
	b_CP_SO_S_COP ,     
	b_CP_SO_CSP_O ,     
	b_CP_OCS_CSP_CO ,   
	b_CHP_N_CNP_H ,     
	b_CH_NP_CNP_H ,     
	b_CH_NHP_CH2P_N ,   
	b_CHP_NH_CNP_H2 ,   
	b_CHP_NH2_HCNP_H2 , 
	b_CH_NH2P_NH_CH2P , 
	b_CHP_NH3_NH4P_C ,  
	b_CH_NH3P_NH4P_C ,  
	b_CH_HCNP_CN_CH2P , 
	b_CH_HNOP_NO_CH2P , 
	b_CHP_S_HSP_C ,     
	b_CHP_S_CSP_H ,     
	b_CH_SP_CSP_H ,     
	b_CH_HSP_S_CH2P ,   
	b_CHP_OCS_HCSP_CO , 
	b_N_CH2P_HCNP_H ,   
	b_NP_NH_N2P_H ,     
	b_N_NHP_N2P_H ,     
	b_NP_CH4_HCNP_H2_H ,
	b_N_OHP_NOP_H ,     
	b_NP_NH3_NH2P_NH ,  
	b_N_H2OP_HNOP_H ,   
	b_NP_CO_NOP_C ,     
	b_NP_NO_N2P_O ,     
	b_NP_O2_NO_OP ,     
	b_NP_O2_NOP_O ,     
	b_N_O2P_NOP_O ,    
	b_N_HSP_NSP_H ,    
	b_N_SIOP_NOP_SI ,  
	b_N_SIOP_NO_SIP ,  
	b_N_SOP_NSP_O ,    
	b_NP_OCS_SP_CO_N , 
	b_NP_OCS_CSP_NO ,  
	b_CH2_NHP_CH3P_N , 
	b_CH2_NH2P_CH3P_NH ,
	b_CH2P_NH3_NH4P_CH , 
	b_CH2_NH3P_NH2_CH3P ,
	b_CH2_HCNP_CN_CH3P , 
	b_CH2_HNOP_NO_CH3P , 
	b_CH2P_S_HCSP_H ,    
	b_CH2_SP_HCSP_H ,    
	b_NHP_NH_NH2P_N ,    
	b_NHP_O_OHP_N ,      
	b_NH_OP_NOP_H ,      
	b_NHP_NH2_NH3P_N ,   
	b_NH_NH2P_NH3P_N ,   
	b_NHP_OH_H2OP_N ,    
	b_NH_OHP_NH2P_O ,    
	b_NHP_NH3_NH4P_N ,   
	b_NH_NH3P_NH4P_N ,   
	b_NH_CH5P_CH4_NH2P , 
	b_NHP_H2O_NH3P_O ,   
	b_NHP_H2O_H3OP_N ,   
	b_NHP_H2O_HNOP_H2 ,    
	b_NH_H2OP_H3OP_N ,     
	b_NHP_H2O_OH_NH2P ,    
	b_NHP_CN_HCNP_N ,      
	b_NH_HCNP_CN_NH2P ,    
	b_NHP_CO_HCOP_N ,      
	b_NH_COP_HCOP_N ,      
	b_NHP_CO_OCNP_H ,      
	b_NH_SIP_SINP_H ,      
	b_NH_HCOP_CO_NH2P ,    
	b_NH_HNOP_NO_NH2P ,    
	b_NH_O2P_HNOP_O ,      
	b_NHP_O2_NOP_OH ,      
	b_NH_O2P_NO2P_H ,      
	b_NHP_S_HSP_N ,        
	b_NHP_S_NSP_H ,        
	b_NH_SP_NSP_H ,        
	b_CH3P_NH3_NH4P_CH2 ,  
	b_CH3P_S_HCSP_H2 ,     
	b_O_NH2P_HNOP_H ,      
	b_O_NH3P_HNOP_H2 ,     
	b_OP_CN_NOP_C ,        
	b_OP_HCN_COP_NH ,      
	b_OP_HCN_NOP_CH ,      
	b_OP_HCN_HCOP_N ,      
	b_OP_N2_NOP_N ,        
	b_O_N2P_NOP_N ,        
	b_O_HNOP_NO2P_H ,      
	b_O_HSP_SP_OH ,       
	b_O_HSP_SOP_H ,       
	b_O_SINP_SIOP_N ,     
	b_OP_N2O_NOP_NO ,     
	b_O_CSP_S_COP ,       
	b_O_HCSP_S_HCOP ,    
	b_O_HCSP_OCSP_H ,    
	b_OP_NO2_O2_NOP ,    
	b_O_NSP_S_NOP ,      
	b_NH2P_NH2_NH3P_NH , 
	b_NH2_OHP_NH3P_O ,   
	b_NH2P_OH_H2OP_NH ,  
	b_NH2P_NH3_NH4P_NH , 
	b_NH2_NH3P_NH4P_NH ,
	b_NH2_CH5P_NH3P_CH4 , 
	b_NH2P_H2O_NH3P_OH ,  
	b_NH2P_H2O_H3OP_NH ,  
	b_NH2_H2OP_NH3P_OH ,  
	b_NH2P_H2O_NH4P_O ,   
	b_NH2_H3OP_H2O_NH3P , 
	b_NH2_HCNP_CN_NH3P ,  
	b_NH2_COP_HCOP_NH ,   
	b_NH2_HCOP_CO_NH3P ,  
	b_NH2_HNOP_NO_NH3P ,  
	b_NH2P_O2_HNOP_OH ,   
	b_NH2P_S_HSP_NH ,     
	b_CH4P_NH3_NH4P_CH3 , 
	b_CH4_NH3P_NH4P_CH3 , 
	b_CH4_N2P_N2_CH2P_H2 ,
	b_CH4_N2P_N2_CH3P_H , 
	b_CH4_HNOP_NO_CH5P ,   
	b_CH4_SP_HCSP_H2_H ,   
	b_CH4_CSP_HCSP_CH3 ,   
	b_OHP_NH3_NH4P_O ,     
	b_OH_NH3P_NH4P_O ,     
	b_OHP_CN_HCNP_O ,      
	b_OH_HCNP_CN_H2OP ,    
	b_OHP_NO_HNOP_O ,      
	b_OH_HNOP_NO_H2OP ,    
	b_OHP_S_HSP_O ,        
	b_OHP_S_SOP_H ,        
	b_OH_SP_SOP_H ,        
	b_NH3P_NH3_NH4P_NH2 ,  
	b_NH3_CH5P_NH4P_CH4 ,  
	b_NH3P_H2O_NH4P_OH ,   
	b_NH3_H2OP_NH4P_OH ,   
	b_NH3_H3OP_NH4P_H2O ,  
	b_NH3_COP_HCOP_NH2 ,   
	b_NH3_HCOP_CO_NH4P ,   
	b_NH3_HNOP_NO_NH4P ,   
	b_NH3_HSP_S_NH4P ,     
	b_NH3_HCSP_CS_NH4P ,   
	b_CH5P_S_HSP_CH4 ,     
	b_H2O_CNP_HCNP_OH ,    
	b_H2O_CNP_HCOP_NH ,    
	b_H2O_HCNP_CN_H3OP ,   
	b_H2O_HNOP_NO_H3OP ,   
	b_H2OP_S_HSP_OH ,      
	b_H2O_HSP_S_H3OP ,     
	b_H3OP_CS_HCSP_H2O ,   
	b_CNP_NO_OCNP_N ,      
	b_CN_HNOP_NO_HCNP ,    
	b_CNP_O2_NOP_CO ,      
	b_CNP_O2_OCNP_O ,      
	b_HCNP_CO_HCOP_CN ,    
	b_HCNP_S_HSP_CN ,      
	b_CO_HNOP_NO_HCOP ,    
	b_N2P_OCS_SP_N2_CO ,   
	b_HCOP_S_HSP_CO ,      
	b_HCOP_CS_HCSP_CO ,    
	b_HNOP_S_HSP_NO ,      
	b_O2_SP_SOP_O ,        
	b_O2P_S_SOP_O ,        
	b_O2_CSP_OCSP_O ,      
	b_S_SIOP_SO_SIP ,      
	b_CP_NH3_NH3P_C ,     
	b_C_CNP_CN_CP ,       
	b_C_N2P_N2_CP ,      
	b_CP_NO_NOP_C ,      
	b_CP_SIN_SINP_C ,    
	b_C_CSP_CS_CP ,      
	b_CP_NS_NSP_C ,      
	b_CP_SO_SOP_C ,      
	b_CP_OCS_OCSP_C ,    
	b_CH_NH2P_NH2_CHP ,  
	b_CHP_NH3_NH3P_CH ,  
	b_CH_CNP_CN_CHP ,    
	b_CH_N2P_N2_CHP ,    
	b_CHP_NO_NOP_CH ,    
	b_NP_NH_NHP_N ,      
	b_NP_NH2_NH2P_N ,    
	b_NP_NH3_NH3P_N ,    
	b_NP_CN_CNP_N ,      
	b_NP_HCN_HCNP_N ,    
	b_N_N2P_N2_NP ,      
	b_NP_NO_NOP_N ,      
	b_NP_OCS_OCSP_N ,    
	b_CH2_NH2P_NH2_CH2P ,
	b_CH2_CNP_CN_CH2P ,  
	b_CH2_N2P_N2_CH2P ,  
	b_CH2P_NO_NOP_CH2 ,  
	b_NH_OP_O_NHP ,      
	b_NH_OHP_OH_NHP ,    
	b_NHP_NH3_NH3P_NH ,  
	b_NHP_H2O_H2OP_NH ,  
	b_NH_CNP_CN_NHP ,    
	b_NH_COP_CO_NHP ,    
	b_NH_N2P_N2_NHP ,    
	b_NHP_NO_NOP_NH ,    
	b_NHP_O2_O2P_NH ,    
	b_NHP_S_SP_NH ,      
	b_CH3P_NO_NOP_CH3 ,  
	b_OP_NH2_NH2P_O ,    
	b_OP_NH3_NH3P_O ,    
	b_O_CNP_CN_OP ,      
	b_O_HCNP_HCN_OP ,    
	b_O_N2P_N2_OP ,      
	b_OP_NO_NOP_O ,      
	b_OP_OCS_OCSP_O ,    
	b_NH2_OHP_OH_NH2P ,  
	b_NH2P_NH3_NH3P_NH2 ,
	b_NH2_H2OP_H2O_NH2P ,  
	b_NH2_CNP_CN_NH2P ,    
	b_NH2_COP_CO_NH2P ,    
	b_NH2_N2P_N2_NH2P ,    
	b_NH2P_NO_NOP_NH2 ,    
	b_NH2_O2P_O2_NH2P ,    
	b_NH2P_S_SP_NH2 ,      
	b_CH4P_NH3_NH3P_CH4 ,  
	b_CH4P_OCS_OCSP_CH4 ,  
	b_OHP_NH3_NH3P_OH ,    
	b_OH_CNP_CN_OHP ,      
	b_OH_N2P_N2_OHP ,      
	b_OHP_NO_NOP_OH ,      
	b_NH3_H2OP_H2O_NH3P ,  
	b_NH3_HCNP_HCN_NH3P ,  
	b_NH3_COP_CO_NH3P ,    
	b_NH3_N2P_N2_NH3P ,    
	b_NH3P_SI_SIP_NH3 ,    
	b_NH3_O2P_O2_NH3P ,    
	b_NH3_SP_S_NH3P ,      
	b_NH3_HSP_HS_NH3P ,    
	b_NH3_SOP_SO_NH3P ,    
	b_H2O_HCNP_HCN_H2OP ,  
	b_H2O_N2P_N2_H2OP ,    
	b_H2OP_NO_NOP_H2O ,    
	b_CNP_HCN_HCNP_CN ,    
	b_CNP_CO_COP_CN ,      
	b_CN_N2P_N2_CNP ,      
	b_CNP_NO_NOP_CN ,      
	b_CNP_O2_O2P_CN ,      
	b_CNP_S_SP_CN ,        
	b_HCN_COP_CO_HCNP ,    
	b_HCN_N2P_N2_HCNP ,    
	b_HCNP_NO_NOP_HCN ,    
	b_HCNP_O2_O2P_HCN ,    
	b_HCNP_S_SP_HCN ,      
	b_CO_N2P_N2_COP ,      
	b_COP_NO_NOP_CO ,      
	b_N2P_NO_NOP_N2 ,      
	b_N2P_O2_O2P_N2 ,      
	b_N2P_S_SP_N2 ,        
	b_SI_NOP_NO_SIP ,      
	b_SI_HSP_HS_SIP ,      
	b_SI_CSP_CS_SIP ,      
	b_NO_HNOP_HNO_NOP ,    
	b_NO_O2P_O2_NOP ,      
	b_NO_SP_S_NOP ,        
	b_NO_HSP_HS_NOP ,      
	b_NO_SIOP_SIO_NOP ,    
	b_NO_S2P_S2_NOP ,      
	b_O2P_NO2_NO2P_O2 ,    
	b_S_HSP_HS_SP ,        
	b_C_N_CN_NU ,          
	b_C_S_CS_NU ,          
	b_CP_S_CSP_NU ,        
	b_NP_N_N2P_NU ,        
	b_CO_S_OCS_NU ,        
	b_CH_NP_N_CHP ,        
	b_CHP_S_SP_CH ,        
	b_NP_CH2_CH2P_N ,      
	b_NP_CH4_CH4P_N ,      
	b_NP_OH_OHP_N ,        
	b_NP_H2O_H2OP_N ,      
	b_NP_CO_COP_N ,        
	b_NP_O2_O2P_N ,        
	b_OHP_S_SP_OH ,        
	b_H2OP_S_SP_H2O ,      
	b_COP_S_SP_CO ,        
	b_SI_SP_S_SIP ,        
	b_O2P_S_SP_O2 ,  
	b_C_CH_C2_H,
	b_C_CN_C2_N,
	b_C_CO_C2_O,
	b_C_CS_S_C2,
	b_N_C2_CN_C,
	b_O_C2_CO_C,
	b_C2_O2_CO_CO,
	b_C2_S_CS_C,
	b_NHP_C2_HCNP_C,
	b_OP_C2_COP_C,
	b_C2_O2P_COP_CO,
	b_C2_SP_CSP_C,
	b_C_C2P_C2_CP,
	b_CH_C2P_C2_CHP,
	b_NP_C2_C2P_N,
	b_CH2_C2P_C2_CH2P,
	b_OP_C2_C2P_O,
	b_NH2_C2P_C2_NH2P,
	b_OHP_C2_C2P_OH,
	b_OH_C2P_C2_OHP,
	b_H2OP_C2_C2P_H2O,
	b_C2_CNP_CN_C2P,
	b_C2_COP_CO_C2P,
	b_C2_N2P_N2_C2P,
	b_C2P_NO_NOP_C2,
	b_C2_O2P_O2_C2P,
	b_C2P_S_SP_C2,
	b_C_C_C2_nu,
	b_CP_CH_C2P_H,
	b_C_CHP_C2P_H,
	b_CHP_CH_C2P_H2,
	b_N_C2P_CN_CP,
	b_O_C2P_COP_C,
	b_C2P_O2_COP_CO,
	b_C2P_S_CSP_C,
	b_CP_C_C2P_nu,
	P_C_NH       ,      
	P_C_N2       ,    
	P_C_NO       ,   
	P_C_HS       ,  
	P_C_OCN       ,
	P_C_NS       ,
	P_C_SO       ,
	P_C_S2       ,
	P_C_NH3P       ,   
	P_C_HCNP       ,   
	P_C_HNOP       ,   
	P_CH_N       ,     
	P_CH_N2       ,    
	P_CH_NO       ,    
	P_CH_HNO       ,   
	P_CH_S       ,     
	P_N_NH       ,     
	P_N_CH3       ,    
	P_N_OH       ,     
	P_N_CN       ,     
	P_N_SIH       ,    
	P_N_NO       ,     
	P_N_HNO       ,    
	P_N_O2       ,    
	P_N_HS       ,     
	P_N_CS       ,        
	P_N_NS       ,       
	P_N_SO       ,      
	P_N_S2       ,     
	P_CH2_CN       ,  
	P_CH2_NO       , 
	P_CH2_HNO       ,     
	P_NH_NH       ,     
	P_NH_O       ,     
	P_NH_OH       ,   
	P_NH_H2O       , 
	P_NH_CN       , 
	P_NH_NO       ,        
	P_NH_S       ,        
	P_NH_NO2       ,     
	P_CH3_NH3       ,   
	P_CH3_CN       ,   
	P_CH3_HNO       , 
	P_O_NH2       ,      
	P_O_NH3       ,     
	P_O_CN       ,     
	P_O_HCN       ,   
	P_O_N2       ,   
	P_O_NO       ,  
	P_O_HNO       ,
	P_O_HS       ,        
	P_O_OCN       ,      
	P_O_SIN       ,     
	P_O_N2O       ,    
	P_O_CS       ,    
	P_O_NS       ,  
	P_O_SO       , 
	P_O_OCS       ,        
	P_O_S2       ,        
	P_NH2_OH       ,     
	P_NH2_NO       ,    
	P_CH4_CN       ,   
	P_OH_NH3       ,  
	P_OH_CN       ,  
	P_OH_HCN       ,       
	P_OH_NO       ,       
	P_OH_S       ,      
	P_OH_N2O       ,   
	P_OH_CS       ,   
	P_NH3_CN       , 
	P_CN_NO       , 
	P_CN_HNO       ,       
	P_CN_O2       ,       
	P_CN_S       ,       
	P_CO_HS       ,     
	P_N2_O2       ,    
	P_NO_NO       ,   
	P_NO_HNO       , 
	P_NO_S       ,  
	P_O2_S       , 
	P_S_SO       ,        
	P_C_NHP       ,      
	P_CP_NH       ,     
	P_CP_NH2       ,   
	P_C_NH2P       ,  
	P_CP_NH3       , 
	P_CP_HS       , 
	P_C_HSP       ,
	P_CP_OCN       ,       
	P_CP_NS       ,       
	P_CP_SO       ,      
	P_CP_OCS       ,    
	P_CHP_N       ,    
	P_CH_NP       ,   
	P_CH_NHP       , 
	P_CHP_NH       ,
	P_CHP_NH2       ,        
	P_CH_NH2P       ,       
	P_CHP_NH3       ,      
	P_CH_NH3P       ,     
	P_CH_HCNP       ,    
	P_CH_HNOP       ,   
	P_CHP_S       ,    
	P_CH_SP       ,   
	P_CH_HSP       , 
	P_CHP_OCS       ,       
	P_N_CH2P       ,       
	P_NP_NH       ,       
	P_N_NHP       ,      
	P_NP_CH4       ,    
	P_N_OHP       ,    
	P_NP_NH3       ,  
	P_N_H2OP       , 
	P_NP_CO       , 
	P_NP_NO       ,        
	P_NP_O2       ,       
	P_N_O2P       ,      
	P_N_HSP       ,     
	P_N_SIOP       ,   
	P_N_SOP       ,   
	P_NP_OCS       , 
	P_CH2_NHP       ,        
	P_CH2_NH2P       ,      
	P_CH2P_NH3       ,     
	P_CH2_NH3P       ,    
	P_CH2_HCNP       ,   
	P_CH2_HNOP       ,  
	P_CH2P_S       ,   
	P_CH2_SP       ,  
	P_NHP_NH       , 
	P_NHP_O       , 
	P_NH_OP       ,
	P_NHP_NH2       ,        
	P_NH_NH2P       ,       
	P_NHP_OH       ,       
	P_NH_OHP       ,      
	P_NHP_NH3       ,    
	P_NH_NH3P       ,   
	P_NH_CH5P       ,  
	P_NHP_H2O       , 
	P_NH_H2OP       ,
	P_NHP_CN       ,        
	P_NH_HCNP       ,      
	P_NH_COP       ,      
	P_NHP_CO       ,     
	P_NH_SIP       ,    
	P_NH_HCOP       ,  
	P_NH_HNOP       , 
	P_NHP_O2       , 
	P_NH_O2P       ,
	P_NHP_S       ,
	P_NH_SP       ,        
	P_CH3P_NH3       ,   
	P_CH3P_S       ,    
	P_O_NH2P       ,   
	P_O_NH3P       ,  
	P_OP_CN       ,  
	P_OP_HCN       ,       
	P_OP_N2       ,       
	P_O_N2P       ,      
	P_O_HNOP       ,    
	P_O_HSP       ,    
	P_O_SINP       ,  
	P_OP_N2O       , 
	P_O_CSP       , 
	P_O_HCSP       ,        
	P_OP_NO2       ,       
	P_O_NSP       ,       
	P_NH2P_NH2       ,   
	P_NH2_OHP       ,   
	P_NH2P_OH       ,  
	P_NH2P_NH3       ,        
	P_NH2_NH3P       ,       
	P_NH2_CH5P       ,      
	P_NH2P_H2O       ,     
	P_NH2_H2OP       ,    
	P_NH2_H3OP       ,  
	P_NH2_HCNP       , 
	P_NH2_COP       , 
	P_NH2_HCOP       ,       
	P_NH2_HNOP       ,      
	P_NH2P_O2       ,      
	P_NH2P_S       ,      
	P_CH4P_NH3       ,   
	P_CH4_NH3P       ,  
	P_CH4_N2P       ,  
	P_CH4_HNOP       ,
	P_CH4_SP       , 
	P_CH4_CSP       ,        
	P_OHP_NH3       ,       
	P_OH_NH3P       ,      
	P_OHP_CN       ,      
	P_OH_HCNP       ,    
	P_OHP_NO       ,    
	P_OH_HNOP       ,  
	P_OHP_S       ,   
	P_OH_SP       ,  
	P_NH3P_NH3       ,        
	P_NH3_CH5P       ,       
	P_NH3P_H2O       ,      
	P_NH3_H2OP       ,     
	P_NH3_H3OP       ,    
	P_NH3_COP       ,    
	P_NH3_HCOP       ,  
	P_NH3_HNOP       , 
	P_NH3_HSP       , 
	P_NH3_HCSP       ,        
	P_CH5P_S       ,        
	P_H2O_CNP       ,       
	P_H2O_HCNP       ,        
	P_H2O_HNOP       ,       
	P_H2OP_S       ,        
	P_H2O_HSP       ,      
	P_H3OP_CS       ,     
	P_CNP_NO       ,     
	P_CN_HNOP       ,   
	P_CNP_O2       ,   
	P_HCNP_CO       , 
	P_HCNP_S       , 
	P_CO_HNOP       ,        
	P_N2P_OCS       ,       
	P_HCOP_S       ,       
	P_HCOP_CS       ,     
	P_HNOP_S       ,     
	P_O2_SP       ,     
	P_O2P_S       ,    
	P_O2_CSP       ,  
	P_S_SIOP       , 
	P_C_CNP       ,        
	P_C_N2P       ,       
	P_CP_NO       ,      
	P_CP_SIN       ,    
	P_C_CSP       ,    
	P_CH_CNP       ,       
	P_CH_N2P       ,      
	P_CHP_NO       ,     
	P_NP_NH2       ,   
	P_NP_CN       ,  
	P_NP_HCN       ,
	P_N_N2P       ,        
	P_CH2_CNP       ,  
	P_CH2_N2P       , 
	P_CH2P_NO       ,
	P_NH_CNP       ,   
	P_NH_N2P       , 
	P_NHP_NO       ,
	P_CH3P_NO       ,     
	P_OP_NH2       ,     
	P_OP_NH3       ,    
	P_O_CNP       ,    
	P_O_HCNP       ,  
	P_OP_NO       , 
	P_OP_OCS       ,        
	P_NH2_CNP       ,   
	P_NH2_N2P       , 
	P_NH2P_NO       ,        
	P_NH2_O2P       ,       
	P_CH4P_OCS       ,   
	P_OH_CNP       ,   
	P_OH_N2P       ,  
	P_NH3_HCNP       ,       
	P_NH3_N2P       ,      
	P_NH3P_SI       ,     
	P_NH3_O2P       ,   
	P_NH3_SP       ,   
	P_NH3_SOP       ,
	P_H2O_N2P       ,       
	P_H2OP_NO       ,      
	P_CNP_HCN       ,     
	P_CNP_CO       ,     
	P_CN_N2P       ,    
	P_CNP_S       ,  
	P_HCN_COP       ,        
	P_HCN_N2P       ,       
	P_HCNP_NO       ,      
	P_HCNP_O2       ,     
	P_CO_N2P       ,    
	P_COP_NO       ,   
	P_N2P_NO       ,  
	P_N2P_O2       , 
	P_N2P_S       , 
	P_SI_NOP       ,        
	P_SI_HSP       ,       
	P_SI_CSP       ,      
	P_NO_HNOP       ,    
	P_NO_O2P       ,    
	P_NO_SP       ,    
	P_NO_HSP       ,  
	P_NO_SIOP       ,
	P_NO_S2P       ,        
	P_O2P_NO2       ,      
	P_S_HSP       ,       
	P_C_N       ,        
	P_C_S       ,       
	P_CP_S       ,     
	P_NP_N       ,    
	P_CO_S       ,   
	P_NP_CH2       ,      
	P_NP_OH       ,     
	P_NP_H2O       ,   
	P_COP_S       ,      
	P_SI_SP       ,
	P_C_CH,
	P_C_CN,
	P_C_CO,
	P_C_CS,
	P_N_C2,
	P_O_C2,
	P_C2_O2,
	P_C2_S,
	P_NHP_C2,
	P_OP_C2,
	P_C2_O2P,
	P_C2_SP,
	P_C_C2P,
	P_CH_C2P,
	P_NP_C2,
	P_CH2_C2P,
	P_NH2_C2P,
	P_OHP_C2,
	P_OH_C2P,
	P_H2OP_C2,
	P_C2_CNP,
	P_C2_COP,
	P_C2_N2P,
	P_C2P_NO,
	P_C2P_S,
	P_C_C,
	P_C_CHP,
	P_CHP_CH,
	P_N_C2P,
	P_O_C2P,
	P_C2P_O2,
	P_CP_C,
	/* C2 and C2+ chemistry */
	H_C2_CH_C,
	C_CH_C2_H_1,
	C_CN_C2_N_1,
	C_CO_C2_O_1,
	C_CS_S_C2_1,
	N_C2_CN_C_1,
	O_C2_CO_C_1,
	C2_O2_CO_CO_1,
	C2_S_CS_C_1,
	HeP_C2_CP_C_He,
	NHP_C2_HCNP_C_1,
	OP_C2_COP_C_1,
	C2_O2P_COP_CO_1,
	C2_SP_CSP_C_1,
	HP_C2_C2P_H,
	H2P_C2_C2P_H2,
	HeP_C2_C2P_He,
	C_C2P_C2_CP_1,
	CH_C2P_C2_CHP_1,
	NP_C2_C2P_N_1,
	CH2_C2P_C2_CH2P_1,
	OP_C2_C2P_O_1,
	NH2_C2P_C2_NH2P_1,
	OHP_C2_C2P_OH_1,
	OH_C2P_C2_OHP_1,
	H2OP_C2_C2P_H2O_1,
	C2_CNP_CN_C2P_1,
	C2_COP_CO_C2P_1,
	C2_N2P_N2_C2P_1,
	C2P_NO_NOP_C2_1,
	C2_O2P_O2_C2P_1,
	C2P_S_SP_C2_1,
	C_C_C2_nu_1,
	C2_nu_C_C,
	C2_nu_C2P_e,
	C2_cr_C_C,
	CP_CH_C2P_H_1,
	C_CHP_C2P_H_1,
	CHP_CH_C2P_H2_1,
	N_C2P_CN_CP_1,
	O_C2P_COP_C_1,
	C2P_O2_COP_CO_1,
	C2P_S_CSP_C_1,
	C2P_e_C_C,
	CP_C_C2P_nu_1,
	C2P_nu_CP_C,
	C_CH_C2_H_2,
	C_CN_C2_N_2,
	C_CO_C2_O_2,
	C_CS_S_C2_2,
	N_C2_CN_C_2,
	O_C2_CO_C_2,
	C2_O2_CO_CO_2,
	C2_S_CS_C_2,
	NHP_C2_HCNP_C_2,
	OP_C2_COP_C_2,
	C2_O2P_COP_CO_2,
	C2_SP_CSP_C_2,
	C_C2P_C2_CP_2,
	CH_C2P_C2_CHP_2,
	NP_C2_C2P_N_2,
	CH2_C2P_C2_CH2P_2,
	OP_C2_C2P_O_2,
	NH2_C2P_C2_NH2P_2,
	OHP_C2_C2P_OH_2,
	OH_C2P_C2_OHP_2,
	H2OP_C2_C2P_H2O_2,
	C2_CNP_CN_C2P_2,
	C2_COP_CO_C2P_2,
	C2_N2P_N2_C2P_2,
	C2P_NO_NOP_C2_2,
	C2_O2P_O2_C2P_2,
	C2P_S_SP_C2_2,
	C_C_C2_nu_2,
	CP_CH_C2P_H_2,
	C_CHP_C2P_H_2,
	CHP_CH_C2P_H2_2,
	N_C2P_CN_CP_2,
	O_C2P_COP_C_2,
	C2P_O2_COP_CO_2,
	C2P_S_CSP_C_2,
	CP_C_C2P_nu_2,
	O_CCl_Cl_CO_1,
	/* Adding Chlorine chemistry */
	O_CCl_ClO_C_1,
	O_ClO_Cl_O2_1,
	H2_ClP_HClP_H,
	H2_HClP_H2ClP_H,
	H3P_Cl_HClP_H2,
	H3P_HCl_H2ClP_H2,
	HeP_HCl_ClP_He_H,
	HeP_CCl_Cl_CP_He,
	HeP_ClO_ClP_O_He,
	CP_HCl_CClP_H_1,
	CH3P_HCl_H2CClP_H2_1,
	H2O_H2ClP_HCl_H3OP_1,
	CO_H2ClP_HCl_HCOP_1,
	HP_HCl_HClP_H,
	CP_CCl_CClP_C_1,
	CP_ClO_ClOP_C_1,
	O2_ClP_Cl_O2P_1,
	HClP_e_Cl_H,
	H2ClP_e_Cl_H_H,
	H2ClP_e_HCl_H,
	CClP_e_Cl_C,
	H2CClP_e_CCl_H_H,
	ClOP_e_Cl_O,
	HCl_NU_Cl_H,
	CCl_NU_Cl_C,
	ClO_NU_Cl_O,
	HCl_crnu_Cl_H,
	CCl_crnu_Cl_C,
	ClO_crnu_Cl_O,
	O_CCl_Cl_CO_2,
	O_CCl_ClO_C_2,
	O_ClO_Cl_O2_2,
	CP_HCl_CClP_H_2,
	CH3P_HCl_H2CClP_H2_2,
	H2O_H2ClP_HCl_H3OP_2,
	CO_H2ClP_HCl_HCOP_2,
	CP_CCl_CClP_C_2,
	CP_ClO_ClOP_C_2,
	O2_ClP_Cl_O2P_2,
	b_O_CCl_Cl_CO,
	b_O_CCl_ClO_C,
	b_O_ClO_Cl_O2,
	b_CP_HCl_CClP_H,
	b_CH3P_HCl_H2CClP_H2,
	b_H2O_H2ClP_HCl_H3OP,
	b_CO_H2ClP_HCl_HCOP,
	b_CP_CCl_CClP_C,
	b_CP_ClO_ClOP_C,
	b_O2_ClP_Cl_O2P,
	P_O_CCl,
	P_O_ClO,
	P_CP_HCl,
	P_CH3P_HCl,
	P_H2O_H2ClP,
	P_CO_H2ClP,
	P_CP_CCl,
	P_CP_ClO,
	P_O2_ClP,
	CP_NH_CHP_N_1,
	CP_NH_CHP_N_2,
	b_CP_NH_CHP_N ,
	HeII, HII, HI,H2g, H2p, H3p, ele,CH,
	CHP,OH,OHP,O2,CO,COP,H2O,H2OP,O2P,H3OP,CH2P,
	CH2,HCOP,CH3P,SIH2P,SIH, 
	SIO,SIOP,CH3,CH4,CH4P,CH5P,N2,
	N2P,NO,NOP,S2,S2P,OCN,NH,
	NHP,NH2,NH2P,NH3,NH3P,CN,
	CNP,HCN,HCNP,
	HNO,HNOP,HS,HSP,CS,CSP,NO2,
	NS,NSP,SO,SOP,SIN,SINP,
	N2O,HCSP,OCS,C2,C2P,CP,OP,SIP,NP,SP,ATC,ATO, 
	ATSI,ATN,ATS,crnu_rate, TH85_field, H2s, Hm,
	ClO,HCl,CCl,H2ClP,ClP;
	
	double
		amat[NUM_COMOLE_CALC][NUM_COMOLE_CALC], 
		b[NUM_COMOLE_CALC], 
		c[NUM_COMOLE_CALC + 1][NUM_COMOLE_CALC + 1],
		matrix_value, 
		co_denominator;
	int ns, nd;


	float abundan;
	
	double number_of_sites, 
			binding_energy, 
			exponent, 
			vib_freq, 
			kevap, 
			den_times_area;
	float cartot_mol, oxytot_mol , cartot_ion, oxytot_ion,  siltot_ion,
		 nitrotot_ion,  sulptot_ion, chlotot_ion;

#	ifdef DEBUG_FUN
	fputs( "<+>CO_step()\n", debug_fp );
#	endif

	*lgNegPop = FALSE; 
	*lgZerPop = FALSE;

	/* dissociation energies;
	 * CO 11.09, CO+ 8.3eV, CH+ 3.8eV
	 * LTE population of CO+ */

	/* zero out array used for formation rates */
	for ( i=0; i < NUM_COMOLE_CALC; i++ )
	{
		b[i] = 0.;
		for ( j=0; j < NUM_COMOLE_CALC; j++ )
		{
			c[j][i] = 0.;
		}
	}

	/* carbon conservation, add up all ion forms of carbon that
	 * are not explictly included in the matrix.  The chemistry
	 * netowrk works with CO, C^0, and C^+ */
	cartot_mol = dense.xMolecules[ipCARBON] + co.hevmol[ipATC] + co.hevmol[ipCP];
	oxytot_mol = dense.xMolecules[ipOXYGEN] + co.hevmol[ipATO] + co.hevmol[ipOP];

	ASSERT( cartot_mol >= 0. && oxytot_mol >= 0.);
	
	/* >>chng 05 mar 22, NA, define  common densities with varibles that take up less memory */
	
	CH	= co.hevmol[ipCH];
	CHP	= co.hevmol[ipCHP];
	OH	= co.hevmol[ipOH];
	OHP	= co.hevmol[ipOHP];
	O2  = co.hevmol[ipO2];
	CO	= co.hevmol[ipCO];
	COP	= co.hevmol[ipCOP];
	H2O	= co.hevmol[ipH2O];
	H2OP = co.hevmol[ipH2OP];
	O2P	= co.hevmol[ipO2P];
	H3OP = co.hevmol[ipH3OP];
	CH2P = co.hevmol[ipCH2P];
	CH2 = co.hevmol[ipCH2];
	HCOP = co.hevmol[ipHCOP];
	CH3P = co.hevmol[ipCH3P];
	SIH2P = co.hevmol[ipSIH2P];
	SIH   = co.hevmol[ipSIH];
	SIO   = co.hevmol[ipSIO];
	SIOP  = co.hevmol[ipSIOP];
	CH3   = co.hevmol[ipCH3];
	CH4   = co.hevmol[ipCH4];
	CH4P  = co.hevmol[ipCH4P];
	CH5P  = co.hevmol[ipCH5P];
	N2	= co.hevmol[ipN2];
	N2P	= co.hevmol[ipN2P];
	NO	= co.hevmol[ipNO];
	NOP	= co.hevmol[ipNOP];
	S2  = co.hevmol[ipS2];
	S2P	= co.hevmol[ipS2P];
	OCN	= co.hevmol[ipOCN];
	NH	= co.hevmol[ipNH];
	NHP	= co.hevmol[ipNHP];
	NH2	= co.hevmol[ipNH2];
	NH2P = co.hevmol[ipNH2P];
	NH3 = co.hevmol[ipNH3];
	NH3P = co.hevmol[ipNH3P];
	CN  = co.hevmol[ipCN];
	CNP = co.hevmol[ipCNP];
	HCN = co.hevmol[ipHCN];
	HCNP = co.hevmol[ipHCNP];
	HNO = co.hevmol[ipHNO];
	HNOP = co.hevmol[ipHNOP];
	HS  = co.hevmol[ipHS];
	HSP = co.hevmol[ipHSP];
	CS  = co.hevmol[ipCS];
	CSP	= co.hevmol[ipCSP];
	NO2	= co.hevmol[ipNO2];
	NS	= co.hevmol[ipNS];
	NSP = co.hevmol[ipNSP];
	SO	= co.hevmol[ipSO];
	SOP	= co.hevmol[ipSOP];
	SIN	= co.hevmol[ipSIN];
	SINP=	co.hevmol[ipSINP];
	N2O	=co.hevmol[ipN2O];
	HCSP=	co.hevmol[ipHCSP];
	OCS	=co.hevmol[ipOCS];
	C2	= co.hevmol[ipC2];
	C2P	= co.hevmol[ipC2P];
	CP   = co.hevmol[ipCP];
	OP    = co.hevmol[ipOP];
	SIP	= co.hevmol[ipSIP];
	NP  = co.hevmol[ipNP];
	SP  = co.hevmol[ipSP];
	ATC = co.hevmol[ipATC];
	ATO = co.hevmol[ipATO];
	ATSI= co.hevmol[ipATSI];
	ATN = co.hevmol[ipATN];
	ATS = co.hevmol[ipATS];
	HeII = dense.xIonDense[ipHELIUM][1];
	HII = dense.xIonDense[ipHYDROGEN][1];
	HI = dense.xIonDense[ipHYDROGEN][0];
	H2g = hmi.Hmolec[ipMH2g];
	H2p = hmi.Hmolec[ipMH2p];
	H3p = hmi.Hmolec[ipMH3p];
	ele = dense.eden;
	crnu_rate = secondaries.csupra[ipHYDROGEN][0];
	TH85_field = hmi.UV_Cont_rel2_Habing_TH85_depth/1.66f;
	H2s = hmi.Hmolec[ipMH2s];
	Hm = hmi.Hmolec[ipMHm];
	CCl = co.hevmol[ipCCl];        
	ClO = co.hevmol[ipClO];     
	HCl = co.hevmol[ipHCl];     
	CCl = co.hevmol[ipCCl];   
	H2ClP = co.hevmol[ipH2ClP];
	ClP = co.hevmol[ipClP];


	/* now see how much is in the gas phase and ions */

	
	/******************************************************************************
	*																			  *
	*																			  *
	*						THE REACTION NETWORK								  *
	*																			  *
	*******************************************************************************

	This part of the code sets of the matrix elements used to predict molecular abundances 
	for all "heavy element" molecules.  This means molecules that contain either a carbon, 
	oxygen, or a silicon atom.  H2 formation and destruction is not treated here, it is 
	done separately in h2.c and hmole.c.

	The network of reactions include every reaction in the appendix of the TH85 paper
	(Tables 5-14, see references below) although many of the rates come from the later
	paper by Hollenbach and McKee (1989--Tables 3,4, and 5 and henceforth referred to 
	as HM89).  Other reactions not included in the TH85 paper but presented in other
	papers are also part of our reaction network.  These include cosmic ray 
	photodissociation reactions and many neutral-neutral reactions given in the HM89 paper.
	  
	Each term is put in with its reference so that a table of reactions can be put together. 
	In order to put a reaction in, one should follow the following format. */
	
	
	/* >>refer <tab> molecule <tab> reactant1_+_reactant2_=>_product1_product2 <tab> reference 

	  
	Here the underscore denotes spacing and the <tab> means places where the tab key should
	be struck.  In the future more reactions will be inserted, so it is a good idea to 
	continue putting reactions into Cloudy in this format so future tables can be made easily  
	by running the perl script extract_atomic_data.pl.  In the future, if one wants to insert
	a new molecule into the network, three files must be modified
	
	1.  The reactions must be put into this program (co.c)
	2.  The new molecule must be defined and the molecule variables updated in co.h
	3.  pressuretotal.c must be adjusted to account for the new species
	
	Additionally, if the new molecule has as a constituent an element that is not 
	a part of the chemical network, the elements ionization must account for the amount
	of that element in molecules.  This is done in the series of files labeled ion(element).c
	Examples are ioncarbo.c and ionsilic.c.  The adjusting of the new elements file will
	work by exact analogy with the previously mentioned examples.*/

	/* Define all the products of two densities that comole tries to update here, these are used in coupled
	 * reactions.  The purpose of doing this is to assure that the code always uses the same sequence in multiplying
	 * two densities, if this is not done then the code can have problems with precision when 100% CO is reached */

	/*****************************************************************************************
	*																					     *
	*  Certain reactions require special attention because one of						     *
	*  the densities that we multiply by in the square matrix A is a density we are also     *
	*  trying to solve for.  This non-linearity has to be dealt with in order to converge on *
	*  an accurate solution.  We will use a scheme outlined by E. Muller on pg. 421-423 of   *
	*  "Computational Methods for Astrophysical Fluid Flows".  For a reaction rate of the    *
	*  form:																				 *
	*																						 *
	*                    Rate = (+/-)k*X(new)Y(new)                                          *
	*																						 *
	*  where X(new) and Y(new) are the densities that we are solving for we can approximate  *
	*  this product by assuming that the difference between the old and new densities are    *
	*  small.  Defining this difference as:													 *
	*																						 *
	*					deltax = X(new) - X(old)											 *													
	*					deltay = Y(new) - Y(old)											 *
	*																						 *
	*  and then keeping terms to first order in delta gives the rate as:					 *
	*																						 *
	*                  Rate = (+/-)k*[X(new)Y(old) + X(old)Y(new) - X(old)Y(old)]		     *
	*																						 *
	*  Setting the rate = 0 for a time steady solution then gives:							 *
	*																						 *
	*				(+/-)k*X(old)Y(old) = (+/-)k*[X(new)Y(old) + X(old)Y(new)]				 *
	*																						 *
	*  For the matrix equation:																 *
	*																						 *
	*								Ax = B													 *	
	*																						 *
	*  The new densities get put into the solution column matrix x, the product of old		 * 
	*  density times k gets put into the square matrix A, and the product of k times the	 *
	*  old densities gets put intot the column vector B (here labeled as b[i])			 *
	*																						 *
	*****************************************************************************************/  
	
	P_C_CH2 = ATC*CH2;
	P_C_COP = ATC*COP;
	P_C_H2OP= ATC*H2OP;
	P_C_H3OP= ATC*H3OP;
	P_C_HCOP= ATC*HCOP;
	P_C_O= ATC*ATO;
	P_C_O2= ATC*O2;
	P_C_O2P= ATC*O2P;
	P_C_OH= ATC*OH;
	P_C_OHP= ATC*OHP;
	P_C_SiOP= ATC*SIOP;
	P_CP_OH= CP*OH;
	P_CP_CH = CP*CH;
	P_CP_SiO = CP*SIO;
	P_CP_CH2 = CP*CH2;
	P_CP_H2O = CP*H2O;
	P_CP_O = CP*ATO;
	P_CP_O2 = CP*O2;
	P_CP_OH = CP*OH;
	P_O_CH = ATO*CH;
	P_O_CHP = ATO*CHP;
	P_O_CH2 = ATO*CH2;
	P_O_CH2P = ATO*CH2P; 
	P_O_CH3P = ATO*CH3P;
	P_O_COP = ATO*COP;
	P_O_H2O = ATO*H2O;
	P_O_H2OP = ATO*H2OP;
	P_O_O = ATO*ATO;
	P_O_OH = ATO*OH;
	P_O_OHP = ATO*OHP;
	P_O_Si = ATO*ATSI;
	P_O_SiP = ATO*SIP;
	P_O_SiH = ATO*SIH;
	P_O_SiH2P = ATO*SIH2P;
	P_O_SiOP = ATO*SIOP;
	P_OP_CH = OP*CH;
	P_OP_CH2 = OP*CH2;
	P_OP_CO = OP*CO;
	P_OP_H2O = OP*H2O;
	P_OP_O2 = OP*O2;
	P_OP_OH = OP*OH;
	P_Si_CHP = ATSI*CHP;
	P_Si_H2OP = ATSI*H2OP;
	P_Si_OH = ATSI*OH;
	P_Si_O2P = ATSI*O2P;
	P_SiP_H2O = SIP*H2O;
	P_SiP_OH = SIP*OH;
	P_SiP_O2 = SIP*O2;
	P_CH_COP = CH*COP;
	P_CH_H2OP = CH*H2OP;
	P_CH_H3OP = CH*H3OP;
	P_CH_HCOP = CH*HCOP;
	P_CH_O2 = CH*O2;
	P_CH_O2P = CH*O2P;
	P_CH_OHP = CH*OHP;
	P_CH_SiOP = CH*SIOP;
	P_CHP_H2O = CHP*H2O;
	P_CHP_OH = CHP*OH;
	P_CHP_O2 = CHP*O2;
	P_CH2_COP = CH2*COP;
	P_CH2_H2OP = CH2*H2OP;
	P_CH2_H3OP = CH2*H3OP;
	P_CH2_HCOP = CH2*HCOP;
	P_CH2_O2 = CH2*O2;
	P_CH2_O2P = CH2*O2P;
	P_CH2_OH = CH2*OH;
	P_CH2_OHP = CH2*OHP;
	P_CH2P_O2 = CH2P*O2;
	P_COP_O2 = COP*O2;
	P_H2O_COP = H2O*COP;
	P_H2O_HCOP = H2O*HCOP;
	P_H2OP_CO = H2OP*CO;
	P_H2OP_H2O = H2OP*H2O;
	P_H2OP_O2 = H2OP*O2;
	P_H3OP_SiH = H3OP*SIH;
	P_H3OP_SiO = H3OP*SIO;
	P_HCOP_SiH = HCOP*SIH;
	P_HCOP_SiO = HCOP*SIO;
	P_OH_COP = OH*COP;
	P_OH_H2OP = OH*H2OP;
	P_OH_HCOP = OH*HCOP;
	P_OH_OH = OH*OH;
	P_OHP_CO = OHP*CO;
	P_OHP_H2O = OHP*H2O;
	P_OHP_O2 = OHP*O2;
	P_OHP_OH = OHP*OH;
	P_OHP_SiH = OHP*SIH;
	P_OHP_SiO = OHP*SIO;
	P_C_CH = ATC*CH;
	P_C_CN = ATC*CN;
	P_C_CO = ATC*CO;
	P_C_CS = ATC*CS;
	P_N_C2 = ATN*C2;
	P_O_C2 = ATO*C2;
	P_C2_O2 = C2*O2;
	P_C2_S = C2*ATS;
	P_NHP_C2 = NHP*C2;
	P_OP_C2 = OP*C2;
	P_C2_O2P = C2*O2P;
	P_C2_SP = C2*SP;
	P_C_C2P = ATC*C2P;
	P_CH_C2P = CH*C2P;
	P_NP_C2 = NP*C2;
	P_CH2_C2P = CH2*C2P;
	P_OP_C2 = OP*C2;
	P_NH2_C2P = NH2*C2P;
	P_OHP_C2 = OHP*C2;
	P_OH_C2P = OH*C2P;
	P_H2OP_C2 = H2OP*C2;
	P_C2_CNP = C2*CNP;
	P_C2_COP = C2*COP;
	P_C2_N2P = C2*N2P;
	P_C2P_NO = C2P*NO;
	P_C2_O2P = C2*O2P;
	P_C2P_S = C2P*ATS;
	P_C_C = ATC*ATC;
	P_CP_CH = CP*CH;
	P_C_CHP = ATC*CHP;
	P_CHP_CH = CHP*CH;
	P_CH_C2P = CH*C2P;
	P_N_C2P = ATN*C2P;
	P_O_C2P = ATO*C2P;
	P_C2P_O2 = C2P*O2;
	P_C2P_S = C2P*ATS;
	P_CP_C = CP*ATC;

	/* new products */

	P_C_CH5P = ATC*CH5P;
	P_O_CH4 = ATO*CH4;
	P_O_CH4P = ATO*CH4P;
	P_O_CH5P = ATO*CH5P;
	P_OP_CH4 = OP*CH4;
	P_CH4_CH = CH4*CH;
	P_CH5P_CH = CH5P*CH;
	P_CH2_CH2 = CH2*CH2;
	P_CH4_CH2 = CH4*CH2;
	P_OH_CH2 = OH*CH2;
	P_CH5P_CH2 = CH5P*CH2;
	P_OH_CH3 = OH*CH3;
	P_OH_CH3 = OH*CH3;
	P_H2O_CH3 = H2O*CH3;
	P_CH3_CH3 = CH3*CH3;
	P_OH_CH4 = OH*CH4;
	P_OHP_CH4 = OHP*CH4;
	P_H2OP_CH4 = H2OP*CH4;
	P_COP_CH4 = COP*CH4;
	P_CH4_CH4P = CH4*CH4P;
	P_H2O_CH4P = H2O*CH4P;
	P_CO_CH4P = CO*CH4P;
	P_O2_CH4P = O2*CH4P;
	P_H2O_CH5P = H2O*CH5P;
	P_CO_CH5P = CO*CH5P;
	P_CH5P_OH = CH5P*OH;



	 /* ***************************************************************************************************
	  *Here we are defining the reaction rates for all the reactions in the network, the exceptions		  *
	  * will be a few factors such as the CO self shielding term, this will come in later.  Each reaction *
	  * rate is labeled to make the actual reaction self-explanetary.  For example, in the first reaction *
	  * rate:																							  *	
	  *																									  *	
	  *    C_CH2_CH_CH_1																				  *
	  *																									  *
	  *the rate is for the reaction:																	  *
      *																								      *
	  *    C + CH2 -> CH + CH.																			  *
	  *																									  *
	  * All terms that are "coupled" (see below) will have a one or a two at the end.  Any future		  *
	  * reaction (for as long as this system is desired) should be put in this same way.				  *
	  *****************************************************************************************************/


	

	/* Coupled Reactions !!! */

	C_CH2_CH_CH_1		    = HMRATE(2.69e-12,0,23550)*ATC;
	C_COP_CO_CP_1		    = HMRATE(1.1e-10,0,0)*ATC;
	C_H2OP_OH_CHP_1		    = HMRATE(1.1e-9,0,0)*ATC;
	C_H3OP_HCOP_H2_1	    = HMRATE(1.0e-11,0,0)*ATC;
	C_HCOP_CO_CHP_1		    = HMRATE(1.1e-9,0,0)*ATC;
	C_O_CO_nu_1			    = HMRATE(2.1E-19,0,0)*ATC;
	C_O2_CO_O_1			    = HMRATE(3.3e-11,0,0)*ATC;
	C_O2P_O2_CP_1		    = HMRATE(5.2e-11,0,0)*ATC;
	C_O2P_COP_O_1		    = HMRATE(5.2e-11,0,0)*ATC;
	C_OH_O_CH_1			    = HMRATE(2.25e-11,0.5,14800)*ATC;
	C_OH_CO_H_1			    = HMRATE(1.1e-10,0.5,0)*ATC;
	C_OHP_O_CHP_1		    = HMRATE(1.2e-9,0,0)*ATC;
	C_SiOP_SiP_CO_1		    = HMRATE(1.0e-9,0,0)*ATC;
	/* >>chng 05 mar 20, NA, 
	 * The rate for C+ + OH -> CO+ + H or CO + H+ is a power fit to the
	 * data given in Table 1 of 
	>> refer	Dubernet et al. 1992, ApJ, 239, 855 */
	CP_OH_CO_HP_1		    = HMRATE(2.7e-9,-0.3508,0)*CP;
	CP_CH_CHP_C_1		    = HMRATE(3.8e-10,0,0)*CP;
	CP_SiO_SiP_CO_1		    = HMRATE(5.4e-10,0,0)*CP;
	CP_CH2_CH2P_C_1		    = HMRATE(5.2e-10,0,0)*CP;
	CP_H2O_HCOP_H_1		    = HMRATE(9.0e-10,0,0)*CP;
	CP_O_COP_nu_1		    = HMRATE(2.5E-18,0,0)*CP;
	CP_O2_COP_O_1		    = HMRATE(3.8e-10,0,0)*CP;
	CP_O2_CO_OP_1		    = HMRATE(6.2e-10,0,0)*CP;
	/* >>chng 05 mar 20, NA, 
	 * The rate for C+ + OH -> CO+ + H or CO + H+ is a power fit to the
	 * data given in Table 1 of 
	>> refer	Dubernet et al. 1992, ApJ, 239, 855 */

	CP_OH_COP_H_1		    = HMRATE(2.7e-9,-0.3508,0)*CP;
	O_CH_OH_C_1			    = HMRATE(2.52e-11,0,2381)*ATO;
	O_CH_HCOP_e_1		    = HMRATE(2.0e-11,0.44,0)*ATO;
	O_CH_CO_H_1			    = HMRATE(6.6e-11,0,0)*ATO;
	O_CHP_COP_H_1		    = HMRATE(3.5e-10,0,0)*ATO;
	O_CH2_OH_CH_1		    = HMRATE(4.98e-10,0,6000)*ATO;
	O_CH2_CO_H_H_1		    = HMRATE((1e-10)*1.33,0,0)*ATO;
	O_CH2_CO_H2_1		    = HMRATE((1e-11)*8.,0,0)*ATO;
	O_CH2P_HCOP_H_1		    = HMRATE((1e-10)*7.5,0,0)*ATO;
	O_CH3P_HCOP_H2_1	    = HMRATE(4.0e-10,0,0)*ATO;
	O_COP_CO_OP_1		    = HMRATE((1e-10)*1.4,0,0)*ATO;
	O_H2O_OH_OH_1		    = HMRATE((1e-11)*1.85,0.95,8571)*ATO;
	O_H2OP_O2P_H2_1		    = HMRATE(4.0e-11,0,0)*ATO;
	O_O_O2_nu_1			    = HMRATE(4.9E-20,1.58,0)*ATO;
	O_OH_O2_H_1			    = HMRATE((1e-11)*4.34,-0.5,30)*ATO;
	O_OHP_O2P_H_1		    = HMRATE((1e-10)*7.1,0,0)*ATO;
	O_Si_SiO_nu_1		    = HMRATE(5.52E-18,0.31,0)*ATO;
	O_SiP_SiOP_nu_1		    = HMRATE((1e-17)*1.,0,0)*ATO;
	O_SiH_SiO_H_1		    = HMRATE(4.0e-11,0.5,0)*ATO;
	O_SiH2P_SiOHP_H_1	    = HMRATE((1e-10)*6.3,0,0)*ATO;
	O_SiOP_O2_SiP_1		    = HMRATE(2.0e-10,0,0)*ATO;
	OP_CH_O_CHP_1		    = HMRATE(3.5e-10,0,0)*OP;
	OP_CH_COP_H_1		    = HMRATE(3.5e-10,0,0)*OP;
	OP_CH2_O_CH2P_1		    = HMRATE((1e-10)*9.7,0,0)*OP;
	OP_CO_COP_O_1		    = HMRATE((1e-12)*4.9,0.5,4580)*OP;
	OP_H2O_H2OP_O_1		    = HMRATE((1e-9)*3.2,0,0)*OP;
	OP_O2_O2P_O_1		    = HMRATE((1e-11)*1.9,0,0)*OP;
	OP_OH_O2P_H_1		    = HMRATE((1e-10)*3.6,0,0)*OP;
	OP_OH_OHP_O_1		    = HMRATE((1e-10)*3.6,0,0)*OP;
	Si_CHP_SiP_CH_1		    = HMRATE(2.0e-10,0,0)*ATSI;
	Si_H2OP_SiP_H2O_1	    = HMRATE(3.0e-9,0,0)*ATSI;
	Si_OH_SiO_H_1		    = HMRATE(2.0e-10,0.5,0)*ATSI;
	Si_O2P_O2_SiP_1		    = HMRATE((1e-9)*1.6,0,0)*ATSI;
	SiP_H2O_SiOHP_H_1	    = HMRATE((1e-10)*2.3,0,0)*SIP;
	SiP_OH_SiOP_H_1		    = HMRATE((1e-10)*6.3,0,0)*SIP;
	SiP_O2_SiOP_O_1		    = HMRATE((1e-13)*1.,0,0)*SIP;
	CH_COP_HCOP_C_1		    = HMRATE((1e-10)*3.2,0,0)*CH;
	CH_COP_CO_CHP_1		    = HMRATE((1e-10)*3.2,0,0)*CH;
	CH_H2OP_H2O_CHP_1	    = HMRATE((1e-10)*3.4,0,0)*CH;
	CH_H2OP_OH_CH2P_1	    = HMRATE((1e-10)*3.4,0,0)*CH;	
	CH_H3OP_H2O_CH2P_1	    = HMRATE((1e-10)*6.8,0,0)*CH;
	CH_HCOP_CO_CH2P_1	    = HMRATE((1e-10)*6.3,0,0)*CH;
	CH_O2_CO_OH_1		    = HMRATE((1e-11)*2.6,0,0)*CH;
	CH_O2P_O2_CHP_1		    = HMRATE((1e-10)*3.1,0,0)*CH;
	CH_O2P_HCOP_O_1		    = HMRATE((1e-10)*3.1,0,0)*CH;
	CH_OHP_OH_CHP_1		    = HMRATE(3.5e-10,0,0)*CH;
	CH_OHP_O_CH2P_1		    = HMRATE(3.5e-10,0,0)*CH;
	CH_SiOP_HCOP_Si_1	    = HMRATE((1e-10)*5.9,0,0)*CH;
	CHP_H2O_H3OP_C_1	    = HMRATE((1e-10)*5.8,0,0)*CHP;
	CHP_H2O_HCOP_H2_1	    = HMRATE((1e-9)*2.9,0,0)*CHP;
	CHP_OH_COP_H2_1	        = HMRATE(7.5e-10,0,0)*CHP;
	CHP_O2_HCOP_O_1		    = HMRATE((1e-10)*9.7,0,0)*CHP;
	CHP_O2_COP_OH_1		    = HMRATE(1.0e-11,0,0)*CHP;
	CH2_COP_HCOP_CH_1	    = HMRATE((1e-10)*4.3,0,0)*CH2;
	CH2_COP_CO_CH2P_1	    = HMRATE((1e-10)*4.3,0,0)*CH2;
	CH2_H2OP_H2O_CH2P_1     = HMRATE((1e-10)*4.7,0,0)*CH2;
	CH2_H2OP_OH_CH3P_1	    = HMRATE((1e-10)*4.7,0,0)*CH2;
	CH2_H3OP_H2O_CH3P_1     = HMRATE((1e-10)*9.4,0,0)*CH2;
	CH2_HCOP_CO_CH3P_1	    = HMRATE((1e-10)*8.6,0,0)*CH2;
	CH2_O2_CO_H2O_1		    = HMRATE((1e-10)*2.48,-3.3,1443)*CH2;
	CH2_O2P_O2_CH2P_1	    = HMRATE((1e-10)*4.3,0,0)*CH2;
	CH2_OH_H2O_CH_1		    = HMRATE((1e-11)*1.44,0.5,3000)*CH2;
	CH2_OHP_OH_CH2P_1	    = HMRATE((1e-10)*4.8,0,0)*CH2;
	CH2_OHP_O_CH3P_1	    = HMRATE((1e-10)*4.8,0,0)*CH2;
	CH2P_O2_HCOP_OH_1	    = HMRATE((1e-10)*9.1,0,0)*CH2P;
	COP_O2_O2P_CO_1		    = HMRATE((1e-10)*1.2,0,0)*COP;
	H2O_COP_CO_H2OP_1	    = HMRATE((1e-9)*1.72,0,0)*H2O;
	H2O_COP_HCOP_OH_1	    = HMRATE((1e-10)*8.84,0,0)*H2O;
	H2O_HCOP_CO_H3OP_1	    = HMRATE((1e-9)*2.5,0,0)*H2O;
	H2OP_CO_HCOP_OH_1	    = HMRATE(5.0e-10,0,0)*H2OP;
	H2OP_H2O_H3OP_OH_1	    = HMRATE((1e-9)*2.1,0,0)*H2OP;
	H2OP_O2_O2P_H2O_1	    = HMRATE((1e-10)*4.6,0,0)*H2OP;
	H3OP_SiH_SiH2P_H2O_1    = HMRATE((1e-10)*9.7,0,0)*H3OP;
	H3OP_SiO_SiOHP_H2O_1    = HMRATE(2.0e-9,0,0)*H3OP;
	HCOP_SiH_SiH2P_CO_1     = HMRATE((1e-10)*8.7,0,0)*HCOP;
	HCOP_SiO_SiOHP_CO_1     = HMRATE((1e-10)*7.9,0,0)*HCOP;
	OH_COP_HCOP_O_1		    = HMRATE((1e-10)*3.1,0,0)*OH;
	OH_COP_CO_OHP_1		    = HMRATE((1e-10)*3.1,0,0)*OH;
	OH_H2OP_H3OP_O_1	    = HMRATE((1e-10)*6.9,0,0)*OH;
	OH_HCOP_CO_H2OP_1	    = HMRATE(6.2e-10,0,0)*OH;
	OH_OH_H2O_O_1		    = HMRATE((1e-12)*1.65,1.14,50)*OH;
	OHP_CO_HCOP_O_1		    = HMRATE((1e-9)*1.05,0,0)*OHP;

	OHP_H2O_H3OP_O_1	    = HMRATE(1.3e-9,0,0)*OHP;
	OHP_H2O_H2OP_OH_1	    = HMRATE((1e-9)*1.59,0,0)*OHP;
	OHP_O2_O2P_OH_1		    = HMRATE((1e-10)*5.9,0,0)*OHP;
	OHP_OH_H2OP_O_1		    = HMRATE(7.0e-10,0,0)*OHP;
	OHP_SiH_SiH2P_O_1	    = HMRATE(1.0e-9,0,0)*OHP;
	OHP_SiO_SiOHP_O_1	    = HMRATE((1e-10)*9.4,0,0)*OHP;
	
	C_CH2_CH_CH_2		    = HMRATE(2.69e-12,0,23550)*CH2;
	C_COP_CO_CP_2		    = HMRATE(1.1e-10,0,0)*COP;
	C_H2OP_OH_CHP_2		    = HMRATE(1.1e-9,0,0)*H2OP;
	C_H3OP_HCOP_H2_2	    = HMRATE(1.0e-11,0,0)*H3OP;
	C_HCOP_CO_CHP_2		    = HMRATE(1.1e-9,0,0)*HCOP;
	C_O_CO_nu_2			    = HMRATE(2.1E-19,0,0)*ATO;
	C_O2_CO_O_2			    = HMRATE(3.3e-11,0,0)*O2;
	C_O2P_O2_CP_2		    = HMRATE(5.2e-11,0,0)*O2P;
	C_O2P_COP_O_2		    = HMRATE(5.2e-11,0,0)*O2P;
	C_OH_O_CH_2			    = HMRATE(2.25e-11,0.5,14800)*OH;
	C_OH_CO_H_2			    = HMRATE(1.1e-10,0.5,0)*OH;
	C_OHP_O_CHP_2		    = HMRATE(1.2e-9,0,0)*OHP;
	C_SiOP_SiP_CO_2		    = HMRATE(1.0e-9,0,0)*SIOP;
	/* >>chng 05 mar 20, NA, 
     * The rate for C+ + OH -> CO+ + H or CO + H+ is a power fit to the
	   data given in Table 1 of 
	>> refer	Dubernet et al. 1992, ApJ, 239, 855 */
	CP_OH_CO_HP_2		    = HMRATE(2.7e-9,-0.3508,0)*OH;
	CP_CH_CHP_C_2		    = HMRATE(3.8e-10,0,0)*CH;
	CP_SiO_SiP_CO_2		    = HMRATE(5.4e-10,0,0)*SIO;
	CP_CH2_CH2P_C_2		    = HMRATE(5.2e-10,0,0)*CH2;
	CP_H2O_HCOP_H_2		    = HMRATE(9.0e-10,0,0)*H2O;
	CP_O_COP_nu_2		    = HMRATE(2.5E-18,0,0)*ATO;
	CP_O2_COP_O_2		    = HMRATE(3.8e-10,0,0)*O2;
	CP_O2_CO_OP_2		    = HMRATE(6.2e-10,0,0)*O2;
	/* >>chng 05 mar 20, NA, 
	 * The rate for C+ + OH -> CO+ + H or CO + H+ is a power fit to the
	   data given in Table 1 of 
	>> refer	Dubernet et al. 1992, ApJ, 239, 855 */
	CP_OH_COP_H_2		    = HMRATE(2.7e-9,-0.3508,0)*OH;
	O_CH_OH_C_2			    = HMRATE(2.52e-11,0,2381)*CH;
	O_CH_HCOP_e_2		    = HMRATE(2.0e-11,0.44,0)*CH;
	O_CH_CO_H_2			    = HMRATE(6.6e-11,0,0)*CH;
	O_CHP_COP_H_2		    = HMRATE(3.5e-10,0,0)*CHP;
	O_CH2_OH_CH_2		    = HMRATE(4.98e-10,0,6000)*CH2;
	O_CH2_CO_H_H_2		    = HMRATE((1e-10)*1.33,0,0)*CH2;
	O_CH2_CO_H2_2		    = HMRATE((1e-11)*8.,0,0)*CH2;
	O_CH2P_HCOP_H_2		    = HMRATE((1e-10)*7.5,0,0)*CH2P;
	O_CH3P_HCOP_H2_2	    = HMRATE(4.0e-10,0,0)*CH3P;
	O_COP_CO_OP_2		    = HMRATE((1e-10)*1.4,0,0)*COP;
	O_H2O_OH_OH_2		    = HMRATE((1e-11)*1.85,0.95,8571)*H2O;
	O_H2OP_O2P_H2_2		    = HMRATE(4.0e-11,0,0)*H2OP;
	O_O_O2_nu_2			    = HMRATE(4.9E-20,1.58,0)*ATO;
	O_OH_O2_H_2			    = HMRATE((1e-11)*4.34,-0.5,30)*OH;
	O_OHP_O2P_H_2		    = HMRATE((1e-10)*7.1,0,0)*OHP;
	O_Si_SiO_nu_2		    = HMRATE(5.52E-18,0.31,0)*ATSI;
	O_SiP_SiOP_nu_2		    = HMRATE((1e-17)*1.,0,0)*SIP;
	O_SiH_SiO_H_2		    = HMRATE(4.0e-11,0.5,0)*SIH;
	O_SiH2P_SiOHP_H_2	    = HMRATE((1e-10)*6.3,0,0)*SIH2P;
	O_SiOP_O2_SiP_2		    = HMRATE(2.0e-10,0,0)*SIOP;
	OP_CH_O_CHP_2		    = HMRATE(3.5e-10,0,0)*CH;
	OP_CH_COP_H_2		    = HMRATE(3.5e-10,0,0)*CH;
	OP_CH2_O_CH2P_2		    = HMRATE((1e-10)*9.7,0,0)*CH2;
	OP_CO_COP_O_2		    = HMRATE((1e-12)*4.9,0.5,4580)*CO;
	OP_H2O_H2OP_O_2		    = HMRATE((1e-9)*3.2,0,0)*H2O;
	OP_O2_O2P_O_2		    = HMRATE((1e-11)*1.9,0,0)*O2;
	OP_OH_O2P_H_2		    = HMRATE((1e-10)*3.6,0,0)*OH;
	OP_OH_OHP_O_2		    = HMRATE((1e-10)*3.6,0,0)*OH;
	Si_CHP_SiP_CH_2		    = HMRATE(2.0e-10,0,0)*CHP;
	Si_H2OP_SiP_H2O_2	    = HMRATE(3.0e-9,0,0)*H2OP;
	Si_OH_SiO_H_2		    = HMRATE(2.0e-10,0.5,0)*OH;
	Si_O2P_O2_SiP_2		    = HMRATE((1e-9)*1.6,0,0)*O2P;
	SiP_H2O_SiOHP_H_2	    = HMRATE((1e-10)*2.3,0,0)*H2O;
	SiP_OH_SiOP_H_2		    = HMRATE((1e-10)*6.3,0,0)*OH;
	SiP_O2_SiOP_O_2		    = HMRATE((1e-13)*1.,0,0)*O2;
	CH_COP_HCOP_C_2		    = HMRATE((1e-10)*3.2,0,0)*COP;
	CH_COP_CO_CHP_2		    = HMRATE((1e-10)*3.2,0,0)*COP;
	CH_H2OP_H2O_CHP_2	    = HMRATE((1e-10)*3.4,0,0)*H2OP;
	CH_H2OP_OH_CH2P_2	    = HMRATE((1e-10)*3.4,0,0)*H2OP;
	CH_H3OP_H2O_CH2P_2	    = HMRATE((1e-10)*6.8,0,0)*H3OP;
	CH_HCOP_CO_CH2P_2	    = HMRATE((1e-10)*6.3,0,0)*HCOP;
	CH_O2_CO_OH_2		    = HMRATE((1e-11)*2.6,0,0)*O2;
	CH_O2P_O2_CHP_2		    = HMRATE((1e-10)*3.1,0,0)*O2P;
	CH_O2P_HCOP_O_2		    = HMRATE((1e-10)*3.1,0,0)*O2P;
	CH_OHP_OH_CHP_2		    = HMRATE(3.5e-10,0,0)*OHP;
	CH_OHP_O_CH2P_2		    = HMRATE(3.5e-10,0,0)*OHP;
	CH_SiOP_HCOP_Si_2	    = HMRATE((1e-10)*5.9,0,0)*SIOP;
	CHP_H2O_H3OP_C_2	    = HMRATE((1e-10)*5.8,0,0)*H2O;
	CHP_H2O_HCOP_H2_2	    = HMRATE((1e-9)*2.9,0,0)*H2O;
	CHP_OH_COP_H2_2	        = HMRATE(7.5e-10,0,0)*OH;
	CHP_O2_HCOP_O_2		    = HMRATE((1e-10)*9.7,0,0)*O2;
	CHP_O2_COP_OH_2		    = HMRATE(1.0e-11,0,0)*O2;
	CH2_COP_HCOP_CH_2	    = HMRATE((1e-10)*4.3,0,0)*COP;
	CH2_COP_CO_CH2P_2	    = HMRATE((1e-10)*4.3,0,0)*COP;
	CH2_H2OP_H2O_CH2P_2     = HMRATE((1e-10)*4.7,0,0)*H2OP;
	CH2_H2OP_OH_CH3P_2	    = HMRATE((1e-10)*4.7,0,0)*H2OP;
	CH2_H3OP_H2O_CH3P_2     = HMRATE((1e-10)*9.4,0,0)*H3OP;
	CH2_HCOP_CO_CH3P_2	    = HMRATE((1e-10)*8.6,0,0)*HCOP;
	CH2_O2_CO_H2O_2		    = HMRATE((1e-10)*2.48,-3.3,1443)*O2;
	CH2_O2P_O2_CH2P_2	    = HMRATE((1e-10)*4.3,0,0)*O2P;
	CH2_OH_H2O_CH_2		    = HMRATE((1e-11)*1.44,0.5,3000)*OH;
	CH2_OHP_OH_CH2P_2	    = HMRATE((1e-10)*4.8,0,0)*OHP;
	CH2_OHP_O_CH3P_2	    = HMRATE((1e-10)*4.8,0,0)*OHP;
	CH2P_O2_HCOP_OH_2	    = HMRATE((1e-10)*9.1,0,0)*O2;
	COP_O2_O2P_CO_2		    = HMRATE((1e-10)*1.2,0,0)*O2;
	H2O_COP_CO_H2OP_2	    = HMRATE((1e-9)*1.72,0,0)*COP;
	H2O_COP_HCOP_OH_2	    = HMRATE((1e-10)*8.84,0,0)*COP;
	H2O_HCOP_CO_H3OP_2	    = HMRATE((1e-9)*2.5,0,0)*HCOP;
	H2OP_CO_HCOP_OH_2	    = HMRATE(5.0e-10,0,0)*CO;
	H2OP_H2O_H3OP_OH_2	    = HMRATE((1e-9)*2.1,0,0)*H2O;
	H2OP_O2_O2P_H2O_2	    = HMRATE((1e-10)*4.6,0,0)*O2;
	H3OP_SiH_SiH2P_H2O_2    = HMRATE((1e-10)*9.7,0,0)*SIH;
	H3OP_SiO_SiOHP_H2O_2    = HMRATE(2.0e-9,0,0)*SIO;
	HCOP_SiH_SiH2P_CO_2     = HMRATE((1e-10)*8.7,0,0)*SIH;
	HCOP_SiO_SiOHP_CO_2     = HMRATE((1e-10)*7.9,0,0)*SIO;
	OH_COP_HCOP_O_2		    = HMRATE((1e-10)*3.1,0,0)*COP;
	OH_COP_CO_OHP_2		    = HMRATE((1e-10)*3.1,0,0)*COP;
	OH_H2OP_H3OP_O_2	    = HMRATE((1e-10)*6.9,0,0)*H2OP;
	OH_HCOP_CO_H2OP_2	    = HMRATE(6.2e-10,0,0)*HCOP;
	OH_OH_H2O_O_2		    = HMRATE((1e-12)*1.65,1.14,50)*OH;
	OHP_CO_HCOP_O_2		    = HMRATE((1e-9)*1.05,0,0)*CO;
	OHP_H2O_H3OP_O_2	    = HMRATE(1.3e-9,0,0)*H2O;
	OHP_H2O_H2OP_OH_2	    = HMRATE((1e-9)*1.59,0,0)*H2O;
	OHP_O2_O2P_OH_2		    = HMRATE((1e-10)*5.9,0,0)*O2;
	OHP_OH_H2OP_O_2		    = HMRATE(7.0e-10,0,0)*OH;
	OHP_SiH_SiH2P_O_2	    = HMRATE(1.0e-9,0,0)*SIH;
	OHP_SiO_SiOHP_O_2	    = HMRATE((1e-10)*9.4,0,0)*SIO;
	C_CH5P_CH4_CHP_1	    = HMRATE(1.2e-9,0,0)*CH5P;
	C_CH5P_CH4_CHP_2	    = HMRATE(1.2e-9,0,0)*ATC;
	O_CH4_OH_CH3_1		    = HMRATE((1e-12)*2.29,2.2,3820)*CH4;
	O_CH4_OH_CH3_2		    = HMRATE((1e-12)*2.29,2.2,3820)*ATO;
	O_CH4P_OH_CH3P_1	    = HMRATE(1.0e-9,0,0)*CH4P;
	O_CH4P_OH_CH3P_2	    = HMRATE(1.0e-9,0,0)*ATO;
	O_CH5P_H3OP_CH2_1	    = HMRATE((1e-10)*2.2,0,0)*CH5P;
	O_CH5P_H3OP_CH2_2	    = HMRATE((1e-10)*2.2,0,0)*ATO;
	OP_CH4_OH_CH3P_1	    = HMRATE(1.1e-10,0,0)*CH4;
	OP_CH4_OH_CH3P_2	    = HMRATE(1.1e-10,0,0)*OP;
	OP_CH4_CH4P_O_1		    = HMRATE((1e-10)*8.9,0,0)*CH4;
	OP_CH4_CH4P_O_2		    = HMRATE((1e-10)*8.9,0,0)*OP;
	CH4_CH_CH3_CH2_1	    = HMRATE((1e-11)*2.28,0.7,3000)*CH;
	CH4_CH_CH3_CH2_2	    = HMRATE((1e-11)*2.28,0.7,3000)*CH4;
	CH5P_CH_CH4_CH2P_1      = HMRATE((1e-10)*6.9,0,0)*CH;
	CH5P_CH_CH4_CH2P_2      = HMRATE((1e-10)*6.9,0,0)*CH5P;
	CH2_CH2_CH3_CH_1	    = HMRATE(4.0e-10,0,5000)*CH2;
	CH2_CH2_CH3_CH_2	    = HMRATE(4.0e-10,0,5000)*CH2;
	CH4_CH2_CH3_CH3_1	    = HMRATE((1e-12)*7.13,0,5050)*CH2;
	CH4_CH2_CH3_CH3_2	    = HMRATE((1e-12)*7.13,0,5050)*CH4;
	OH_CH2_O_CH3_1		    = HMRATE((1e-11)*1.44,0.5,3000)*CH2;
	OH_CH2_O_CH3_2		    = HMRATE((1e-11)*1.44,0.5,3000)*OH;
	CH5P_CH2_CH4_CH3P_1     = HMRATE((1e-10)*9.6,0,0)*CH2;
	CH5P_CH2_CH4_CH3P_2     = HMRATE((1e-10)*9.6,0,0)*CH5P;
	OH_CH3_CH4_O_1		    = HMRATE((1e-14)*3.27,2.2,2240)*CH3;
	OH_CH3_CH4_O_2		    = HMRATE((1e-14)*3.27,2.2,2240)*OH;
	OH_CH3_H2O_CH2_1	    = HMRATE((1e-10)*1.2,0,1400)*CH3;
	OH_CH3_H2O_CH2_2	    = HMRATE((1e-10)*1.2,0,1400)*OH;
	H2O_CH3_OH_CH4_1	    = HMRATE((1e-15)*2.3,3.47,6681)*CH3;
	H2O_CH3_OH_CH4_2	    = HMRATE((1e-15)*2.3,3.47,6681)*H2O;
	CH3_CH3_CH4_CH2_1	    = HMRATE((1e-12)*7.13,0,5052)*CH3;
	CH3_CH3_CH4_CH2_2	    = HMRATE((1e-12)*7.13,0,5052)*CH3;
	OH_CH4_H2O_CH3_1	    = HMRATE((1e-13)*3.77,2.42,1162)*CH4;
	OH_CH4_H2O_CH3_2	    = HMRATE((1e-13)*3.77,2.42,1162)*OH;
	OHP_CH4_CH5P_O_1        = HMRATE((1e-10)*1.95,0,0)*CH4;
	OHP_CH4_CH5P_O_2        = HMRATE((1e-10)*1.95,0,0)*OHP;
	OHP_CH4_H3OP_CH2_1      = HMRATE((1e-9)*1.31,0,0)*CH4;
	OHP_CH4_H3OP_CH2_2      = HMRATE((1e-9)*1.31,0,0)*OHP;
	H2OP_CH4_H3OP_CH3_1     = HMRATE((1e-9)*1.4,0,0)*CH4;
	H2OP_CH4_H3OP_CH3_2     = HMRATE((1e-9)*1.4,0,0)*H2OP;
	COP_CH4_HCOP_CH3_1      = HMRATE((1e-10)*4.55,0,0)*CH4;
	COP_CH4_HCOP_CH3_2      = HMRATE((1e-10)*4.55,0,0)*COP;
	COP_CH4_CO_CH4P_1	    = HMRATE((1e-10)*7.93,0,0)*CH4;
	COP_CH4_CO_CH4P_2	    = HMRATE((1e-10)*7.93,0,0)*COP;
	CH4_CH4P_CH5P_CH3_1     = HMRATE((1e-9)*1.5,0,0)*CH4P;
	CH4_CH4P_CH5P_CH3_2     = HMRATE((1e-9)*1.5,0,0)*CH4;
	H2O_CH4P_H3OP_CH3_1     = HMRATE((1e-9)*2.6,0,0)*CH4P;
	H2O_CH4P_H3OP_CH3_2     = HMRATE((1e-9)*2.6,0,0)*H2O;
	CO_CH4P_HCOP_CH3_1      = HMRATE((1e-9)*1.4,0,0)*CH4P;
	CO_CH4P_HCOP_CH3_2      = HMRATE((1e-9)*1.4,0,0)*CO;
	O2_CH4P_O2P_CH4_1	    = HMRATE(4.0e-10,0,0)*CH4P;
	O2_CH4P_O2P_CH4_2	    = HMRATE(4.0e-10,0,0)*O2;
	H2O_CH5P_H3OP_CH4_1     = HMRATE((1e-9)*3.7,0,0)*CH5P;
	H2O_CH5P_H3OP_CH4_2     = HMRATE((1e-9)*3.7,0,0)*H2O;
	CO_CH5P_HCOP_CH4_1      = HMRATE(1.0e-9,0,0)*CH5P;
	CO_CH5P_HCOP_CH4_2      = HMRATE(1.0e-9,0,0)*CO;
	CH5P_OH_H2OP_CH4_1      = HMRATE(7.0e-10,0,0)*OH;
	CH5P_OH_H2OP_CH4_2      = HMRATE(7.0e-10,0,0)*CH5P;
	H_CH_C_H_H			    = HMRATE(6.0e-9,0,40200)*HI;
	H_OH_O_H_H			    = HMRATE(6.0e-9,0,50900)*HI;
	H_H2O_OH_H_H		    = HMRATE((1e-9)*5.8,0,52900)*HI;
	H_COP_CO_HP			    = HMRATE((1e-10)*7.5,0,0)*HI;
	H_CH_C_H2			    = HMRATE((1e-11)*2.7,0.38,0)*HI;
	H_CHP_CP_H2			    = HMRATE((1e-10)*7.5,0,0)*HI;
	H_CH2_CH_H2			    = HMRATE((1e-11)*6.64,0,0)*HI;
	H_CH3P_CH2P_H2		    = HMRATE(7.0e-10,0,10560)*HI;
	H_OH_O_H2			    = HMRATE(0.0000000000000699,2.8,1950)*HI;
	H_H2O_OH_H2			    = HMRATE((1e-11)*1.59,1.2,9610)*HI;
	H_O2_OH_O			    = HMRATE((1e-10)*2.61,0,8156)*HI;
	H_O2_O_O_H			    = HMRATE(6.0e-9,0,52300)*HI;
	H_C_CH_nu               = HMRATE((1e-17)*1.,0,0)*HI;
	H_CP_CHP_nu             = HMRATE((1e-17)*1.7,0,0)*HI;
	H_OH_H2O_nu             = HMRATE(5.26E-18,-5.22,90)*HI;
	H_O_OH_nu               = HMRATE(9.9E-19,-0.38,0)*HI;
	Hminus_CH_CH2_e         = HMRATE(1.0e-10,0,0)*Hm;
	Hminus_OH_H2O_e         = HMRATE(1.0e-10,0,0)*Hm;
	Hminus_O_OH_e           = HMRATE(1.0e-9,0,0)*Hm;
	Hminus_HCOP_CO_H2       = HMRATE(0.00000023,-0.5,0)*Hm;
	Hminus_H3OP_H2O_H2      = HMRATE(0.00000023,-0.5,0)*Hm;
	Hminus_H3OP_OH_H2_H     = HMRATE(0.00000023,-0.5,0)*Hm;
	HP_CH_CHP_H             = HMRATE((1e-9)*1.9,0,0)*HII;
	HP_CH2_CH2P_H           = HMRATE((1e-9)*1.4,0,0)*HII;
	HP_H2O_H2OP_H           = HMRATE((1e-9)*6.9,0,0)*HII;
	HP_O2_O2P_H             = HMRATE(2.0e-9,0,0)*HII;
	HP_OH_OHP_H             = HMRATE((1e-9)*2.1,0,0)*HII;
	HP_SiO_SiOP_H           = HMRATE((1e-9)*3.3,0,0)*HII;
	HP_CH2_CHP_H2           = HMRATE((1e-9)*1.4,0,0)*HII;
	HP_SiH_SiP_H2           = HMRATE((1e-9)*1.7,0,0)*HII;
	H2_CHP_CH2P_H           = HMRATE(1.2e-9,0,0)*H2g;
	H2_CH2P_CH3P_H          = HMRATE((1e-9)*1.6,0,0)*H2g;
	H2_OHP_H2OP_H           = HMRATE((1e-9)*1.01,0,0)*H2g;
	H2_H2OP_H3OP_H          = HMRATE((1e-10)*6.4,0,0)*H2g;
	H2_COP_HCOP_H           = HMRATE((1e-9)*1.8,0,0)*H2g;
	H2_OP_OHP_H             = HMRATE((1e-9)*1.7,0,0)*H2g;
	H2_SiOP_SiOHP_H         = HMRATE((1e-10)*3.2,0,0)*H2g;
	H2_C_CH2_nu             = HMRATE((1e-17)*1.,0,0)*H2g;
	H2_CP_CH2P_nu           = HMRATE(0.0000000000000004,-0.2,0)*H2g;
	H2_SiP_SiH2P_nu         = HMRATE(0.000000000000000003,0,0)*H2g;
	H2_C_CH_H               = HMRATE((1e-10)*6.64,0,11700)*H2g;
	H2_CP_CHP_H             = HMRATE(1.0e-10,0,4640)*H2g;
	H2_CH_CH2_H             = HMRATE((1e-10)*5.46,0,1943)*H2g;
	H2_OH_H2O_H             = HMRATE((1e-12)*2.05,1.52,1736)*H2g;
	H2_O_OH_H               = HMRATE((1e-13)*3.14,2.7,3150)*H2g;
	H2_CH_C_H2_H            = HMRATE(6.0e-9,0,40200)*H2g;
	H2_OH_O_H2_H            = HMRATE(6.0e-9,0,50900)*H2g;
	H2_H2O_OH_H2_H          = HMRATE((1e-9)*5.8,0,52900)*H2g;
	H2_O2_O_O_H2            = HMRATE(6.0e-9,0,52300)*H2g;
	H2_O2_OH_OH             = HMRATE((1e-10)*3.16,0,21890)*H2g;
	H2_C_CH_H               = HMRATE((1e-10)*6.64,0,11700)*H2g;
	H2s_CH_C_H2_H           = HMRATE(6.0e-9,0,0)*H2s*hmi.lgLeiden_Keep_ipMH2s;
	H2s_OH_O_H2_H           = HMRATE(6.0e-9,0,0)*H2s*hmi.lgLeiden_Keep_ipMH2s;
	H2s_H2O_OH_H2_H         = HMRATE((1e-9)*5.8,0,0)*H2s*hmi.lgLeiden_Keep_ipMH2s;
	H2s_O2_O_O_H2           = HMRATE(6.0e-9,0,0)*H2s*hmi.lgLeiden_Keep_ipMH2s;
	H2P_C_CHP_H             = HMRATE((1e-9)*2.4,0,0)*H2p;
	H2P_CH_CH2P_H           = HMRATE((1e-10)*7.1,0,0)*H2p;
	H2P_CH2_CH3P_H          = HMRATE(1.0e-9,0,0)*H2p;
	H2P_OH_H2OP_H           = HMRATE((1e-10)*7.6,0,0)*H2p;
	H2P_H2O_H3OP_H          = HMRATE((1e-9)*3.4,0,0)*H2p;
	H2P_CO_HCOP_H           = HMRATE((1e-9)*2.16,0,0)*H2p;
	H2P_O_OHP_H             = HMRATE((1e-9)*1.5,0,0)*H2p;
	H2P_CH_CHP_H2           = HMRATE((1e-10)*7.1,0,0)*H2p;
	H2P_CH2_CH2P_H2         = HMRATE(1.0e-9,0,0)*H2p;
	H2P_CO_COP_H2           = HMRATE((1e-10)*6.4,0,0)*H2p;
	H2P_H2O_H2OP_H2         = HMRATE((1e-9)*3.9,0,0)*H2p;
	H2P_O2_O2P_H2           = HMRATE((1e-10)*8.,0,0)*H2p;
	H2P_OH_OHP_H2           = HMRATE((1e-10)*7.6,0,0)*H2p;
	H3P_C_CHP_H2            = HMRATE(2.0e-9,0,0)*H3p;
	H3P_CH_CH2P_H2          = HMRATE(1.2e-9,0,0)*H3p;
	H3P_CH2_CH3P_H2         = HMRATE((1e-9)*1.7,0,0)*H3p;
	H3P_OH_H2OP_H2          = HMRATE(1.3e-9,0,0)*H3p;
	H3P_H2O_H3OP_H2         = HMRATE((1e-9)*5.9,0,0)*H3p;
	H3P_CO_HCOP_H2          = HMRATE((1e-9)*1.7,0,0)*H3p;
	H3P_O_OHP_H2            = HMRATE((1e-10)*8.,0,0)*H3p;
	H3P_SiH_SiH2P_H2        = HMRATE(2.0e-9,0,0)*H3p;
	H3P_SiO_SiOHP_H2        = HMRATE(2.0e-9,0,0)*H3p;
	HeP_CO_OP_C_He          = HMRATE(0.00000000000000014,-0.5,0)*HeII;
	HeP_CO_O_CP_He          = HMRATE((1e-9)*1.6,0,0)*HeII;
	HeP_CH_CHP_He           = HMRATE(5.0e-10,0,0)*HeII;
	HeP_H2O_H2OP_He         = HMRATE((1e-11)*6.05,0,0)*HeII;
	HeP_O2_O2P_He           = HMRATE(3.3e-11,0,0)*HeII;
	HeP_CH_CP_He_H          = HMRATE(1.1e-9,0,0)*HeII;
	HeP_CH2_CHP_He_H        = HMRATE((1e-10)*7.5,0,0)*HeII;
	HeP_OH_OP_He_H          = HMRATE(1.1e-9,0,0)*HeII;
	HeP_H2O_OHP_He_H        = HMRATE((1e-10)*2.86,0,0)*HeII;
	HeP_SiH_SiP_He_H        = HMRATE((1e-9)*1.8,0,0)*HeII;
	HeP_H2O_OH_He_HP        = HMRATE((1e-10)*2.04,0,0)*HeII;
	HeP_CH2_CP_He_H2        = HMRATE((1e-10)*7.5,0,0)*HeII;
	HeP_O2_OP_O_He          = HMRATE(1.0e-9,0,0)*HeII;
	HeP_SiO_SiP_O_He        = HMRATE((1e-10)*8.6,0,0)*HeII;
	HeP_SiO_Si_OP_He        = HMRATE((1e-10)*8.6,0,0)*HeII;
	crnu_CH_C_H             = crnu_rate * 2. * 756;
	crnu_CHP_CP_H           = crnu_rate * 2. * 183;
	crnu_CO_O_C             = crnu_rate * 2. * 10;
	crnu_H2O_OH_H           = crnu_rate * 2. * 979;
	crnu_O2_O2P_e           = crnu_rate * 2. *88;
	crnu_O2_O_O             = crnu_rate * 2. *730;
	crnu_OH_O_H             = crnu_rate * 2. *522;
	crnu_SiH_Si_H           = crnu_rate * 2. *500;
	crnu_SiO_Si_O           = crnu_rate * 2. *500;
	CRP_CO_COP_e            = 0.000000000000000039;
	nu_CO_O_C               = HMRATE(2.0e-10,0,0);
	e_CHP_C_H               = HMRATE(0.00000015,-0.42,0)*ele;
	e_CH2P_CH_H             = HMRATE(0.00000016,-0.6,0)*ele;
	e_CH2P_C_H_H            = HMRATE(0.000000403,-0.6,0)*ele;
	e_CH2P_C_H2             = HMRATE(0.0000000768,-0.6,0)*ele;
	e_CH3P_C_H2_H           = HMRATE(0.000000105,-0.5,0)*ele;
	e_CH3P_CH2_H            = HMRATE(0.00000014,-0.5,0)*ele;
	e_CH3P_CH_H_H           = HMRATE(0.000000056,-0.5,0)*ele;
	e_CH3P_CH_H2            = HMRATE(0.000000049,-0.5,0)*ele;
	e_CH3P_CH3_nu           = HMRATE(1.1e-10,-0.5,0)*ele;
	e_COP_O_C               = HMRATE(0.0000002,-0.48,0)*ele;
	e_H2OP_OH_H             = HMRATE(0.0000000792,-0.5,0)*ele;
	e_H2OP_O_H_H            = HMRATE(0.000000245,-0.5,0)*ele;
	e_H2OP_O_H2             = HMRATE(0.000000036,-0.5,0)*ele;
	e_H3OP_H2O_H            = HMRATE(0.000000108,-0.5,0)*ele;
	e_H3OP_OH_H_H           = HMRATE(0.000000258,-0.5,0)*ele;
	e_H3OP_OH_H2            = HMRATE(0.0000000645,-0.5,0)*ele;
	e_H3OP_O_H2_H           = HMRATE((1e-9)*5.59,-0.5,0)*ele;
	e_HCOP_CO_H             = HMRATE(0.00000011,-1,0)*ele;
	e_O2P_O_O               = HMRATE(0.000000195,-0.7,0)*ele;
	e_OHP_O_H               = HMRATE(0.0000000375,-0.5,0)*ele;
	e_SiH2P_SiH_H           = HMRATE(0.00000015,-0.5,0)*ele;
	e_SiH2P_Si_H_H          = HMRATE(0.0000002,-0.5,0)*ele;
	e_SiH2P_Si_H2           = HMRATE(0.00000015,-0.5,0)*ele;
	e_SiOP_Si_O             = HMRATE(0.0000002,-0.5,0)*ele;
	e_SiOHP_SiO_H           = HMRATE(0.00000015,-0.5,0)*ele;
	e_SiOHP_Si_OH           = HMRATE(0.00000015,-0.5,0)*ele;
	h2s_ch_ch2_h            = HMRATE((1e-10)*5.46,0, 0. )*H2s*hmi.lgLeiden_Keep_ipMH2s;
	h2s_o_oh_h              = HMRATE((1e-13)*3.14,2.7, 0. )*H2s*hmi.lgLeiden_Keep_ipMH2s;
	h2s_oh_h2o_h            = HMRATE((1e-12)*2.05,1.52, 0. )*H2s*hmi.lgLeiden_Keep_ipMH2s;
	h2s_o2_twooh            = HMRATE((1e-10)*3.16,0, 0. )*H2s*hmi.lgLeiden_Keep_ipMH2s;
	h2s_c_ch_h              = HMRATE((1e-10)*6.64,0, 0. )*H2s*hmi.lgLeiden_Keep_ipMH2s;
	h2s_cp_chp_h            = HMRATE(1.0e-10,0, 0. )*H2s*hmi.lgLeiden_Keep_ipMH2s;
	H_CH3_CH2_H2            = HMRATE(1.0e-10,0,7600)*HI;
	H_CH4P_CH3P_H2          = HMRATE(1.0e-11,0,0)*HI;
	H_CH5P_CH4P_H2          = HMRATE(2.0e-11,0,0)*HI;
	H2_CH2_CH3_H            = HMRATE((1e-11)*5.18,0.17,6400)*H2g;
	H2_CH_CH3_nu            = HMRATE(5.09E-18,-0.71,11.6)*H2g;
	H2_CH3_CH4_H            = HMRATE(0.0000000000000686,2.74,4740)*H2g;
	H2_CH4P_CH5P_H          = HMRATE(3.3e-11,0,0)*H2g;
	H2_CH3P_CH5P_nu         = HMRATE(0.000000000000013,-1,0)*H2g;
	H2s_CH2_CH3_H           = HMRATE((1e-11)*5.18,0.17,0)*H2s*hmi.lgLeiden_Keep_ipMH2s;
	H2s_CH_CH3_nu           = HMRATE(5.09E-18,-0.71,0)*H2s*hmi.lgLeiden_Keep_ipMH2s;
	H2s_CH3_CH4_H           = HMRATE(0.0000000000000686,2.74,0)*H2s*hmi.lgLeiden_Keep_ipMH2s;
	H2P_CH4_CH3P_H2         = HMRATE((1e-9)*2.3,0,0)*H2p;
	H2P_CH4_CH4P_H2         = HMRATE((1e-9)*1.4,0,0)*H2p;
	H2P_CH4_CH5P_H          = HMRATE((1e-10)*1.14,0,0)*H2p;
	H3P_CH3_CH4P_H2         = HMRATE((1e-9)*2.1,0,0)*H3p;
	H3P_CH4_CH5P_H2         = HMRATE((1e-9)*2.4,0,0)*H3p;
	HeP_CH3_CHP_He          = HMRATE((1e-9)*1.8,0,0)*HeII;
	HeP_CH4_CHP_He          = HMRATE((1e-10)*2.4,0,0)*HeII;
	HeP_CH4_CH2P_He         = HMRATE((1e-10)*9.5,0,0)*HeII;
	HeP_CH4_CH3_He          = HMRATE((1e-10)*4.8,0,0)*HeII;
	HeP_CH4_CH3P_He         = HMRATE((1e-11)*8.5,0,0)*HeII;
	HeP_CH4_CH4P_He         = HMRATE((1e-11)*5.1,0,0)*HeII;
	Hminus_CH2_CH3_e        = HMRATE(1.0e-9,0,0)*Hm;
	Hminus_CH3_CH4_e        = HMRATE(1.0e-9,0,0)*Hm;
	HP_CH3_CH3P_H           = HMRATE((1e-9)*3.4,0,0)*HII;
	HP_CH4_CH3P_H2          = HMRATE((1e-9)*2.3,0,0)*HII;
	HP_CH4_CH4P_H           = HMRATE((1e-9)*1.5,0,0)*HII;
	crnu_CH3_CH3P_e         = crnu_rate * 2.*500;
	crnu_CH3_CH2_H          = crnu_rate * 2.*500;
	crnu_CH3_CH_H2          = crnu_rate * 2.*500;
	crnu_CH4_CH2_H2         = crnu_rate * 2.*2272;
	e_CH5P_CH3_H2           = HMRATE(0.00000055,-0.3,0)*ele;
	e_CH5P_CH4_H            = HMRATE(0.00000055,-0.3,0)*ele;
	e_CH4P_CH3_H            = HMRATE(0.000000175,-0.5,0)*ele;
	e_CH4P_CH2_H_H			= HMRATE(0.000000175,-0.5,0)*ele;
	b_C_CH2_CH_CH	    = HMRATE(2.69e-12,0,23550)*P_C_CH2;
	b_C_COP_CO_CP	    = HMRATE(1.1e-10,0,0)*P_C_COP;
	b_C_H2OP_OH_CHP	    = HMRATE(1.1e-9,0,0)*P_C_H2OP;
	b_C_H3OP_HCOP_H2     = HMRATE(1.0e-11,0,0)*P_C_H3OP;
	b_C_HCOP_CO_CHP	    = HMRATE(1.1e-9,0,0)*P_C_HCOP;
	b_C_O_CO_nu		    = HMRATE(2.1E-19,0,0)*P_C_O;
	b_C_O2_CO_O		    = HMRATE(3.3e-11,0,0)*P_C_O2;
	b_C_O2P_O2_CP	    = HMRATE(5.2e-11,0,0)*P_C_O2P;
	b_C_O2P_COP_O	    = HMRATE(5.2e-11,0,0)*P_C_O2P;
	b_C_OH_O_CH		    = HMRATE(2.25e-11,0.5,14800)*P_C_OH;
	b_C_OH_CO_H		    = HMRATE(1.1e-10,0.5,0)*P_C_OH;
	b_C_OHP_O_CHP	    = HMRATE(1.2e-9,0,0)*P_C_OHP;
	b_C_SiOP_SiP_CO	    = HMRATE(1.0e-9,0,0)*P_C_SiOP;
	/* >>chng 05 mar 20, NA, 
	 * The rate for C+ + OH -> CO+ + H or CO + H+ is a power fit to the
	   data given in Table 1 of 
	>> refer	Dubernet et al. 1992, ApJ, 239, 855 */
	b_CP_OH_CO_HP	    = HMRATE(2.7e-9,-0.3508,0)*P_CP_OH;
	b_CP_CH_CHP_C	    = HMRATE(3.8e-10,0,0)*P_CP_CH;
	b_CP_SiO_SiP_CO	    = HMRATE(5.4e-10,0,0)*P_CP_SiO;
	b_CP_CH2_CH2P_C	    = HMRATE(5.2e-10,0,0)*P_CP_CH2;
	b_CP_H2O_HCOP_H	    = HMRATE(9.0e-10,0,0)*P_CP_H2O;
	b_CP_O_COP_nu	    = HMRATE(2.5E-18,0,0)*P_CP_O;
	b_CP_O2_COP_O	    = HMRATE(3.8e-10,0,0)*P_CP_O2;
	b_CP_O2_CO_OP	    = HMRATE(6.2e-10,0,0)*P_CP_O2;
	/* >>chng 05 mar 20, NA, 
	 * The rate for C+ + OH -> CO+ + H or CO + H+ is a power fit to the
	   data given in Table 1 of 
	>> refer	Dubernet et al. 1992, ApJ, 239, 855 */
	b_CP_OH_COP_H	    = HMRATE(2.7e-9,-0.3508,0)*P_CP_OH;
	b_O_CH_OH_C		    = HMRATE(2.52e-11,0,2381)*P_O_CH;
	b_O_CH_HCOP_e	    = HMRATE(2.0e-11,0.44,0)*P_O_CH;
	b_O_CH_CO_H		    = HMRATE(6.6e-11,0,0)*P_O_CH;
	b_O_CHP_COP_H	    = HMRATE(3.5e-10,0,0)*P_O_CHP;
	b_O_CH2_OH_CH	    = HMRATE(4.98e-10,0,6000)*P_O_CH2;
	b_O_CH2_CO_H_H	    = HMRATE((1e-10)*1.33,0,0)*P_O_CH2;
	b_O_CH2_CO_H2	    = HMRATE((1e-11)*8.,0,0)*P_O_CH2;
	b_O_CH2P_HCOP_H	    = HMRATE((1e-10)*7.5,0,0)*P_O_CH2P;
	b_O_CH3P_HCOP_H2     = HMRATE(4.0e-10,0,0)*P_O_CH3P;
	b_O_COP_CO_OP	    = HMRATE((1e-10)*1.4,0,0)*P_O_COP;
	b_O_H2O_OH_OH	    = HMRATE((1e-11)*1.85,0.95,8571)*P_O_H2O;
	b_O_H2OP_O2P_H2	    = HMRATE(4.0e-11,0,0)*P_O_H2OP;
	b_O_O_O2_nu		    = HMRATE(4.9E-20,1.58,0)*P_O_O;
	b_O_OH_O2_H		    = HMRATE((1e-11)*4.34,-0.5,30)*P_O_OH;
	b_O_OHP_O2P_H	    = HMRATE((1e-10)*7.1,0,0)*P_O_OHP;
	b_O_Si_SiO_nu	    = HMRATE(5.52E-18,0.31,0)*P_O_Si;
	b_O_SiP_SiOP_nu	    = HMRATE((1e-17)*1.,0,0)*P_O_SiP;
	b_O_SiH_SiO_H	    = HMRATE(4.0e-11,0.5,0)*P_O_SiH;
	b_O_SiH2P_SiOHP_H    = HMRATE((1e-10)*6.3,0,0)*P_O_SiH2P;
	b_O_SiOP_O2_SiP	    = HMRATE(2.0e-10,0,0)*P_O_SiOP;
	b_OP_CH_O_CHP	    = HMRATE(3.5e-10,0,0)*P_OP_CH;
	b_OP_CH_COP_H	    = HMRATE(3.5e-10,0,0)*P_OP_CH;
	b_OP_CH2_O_CH2P	    = HMRATE((1e-10)*9.7,0,0)*P_OP_CH2;
	b_OP_CO_COP_O	    = HMRATE((1e-12)*4.9,0.5,4580)*P_OP_CO;
	b_OP_H2O_H2OP_O	    = HMRATE((1e-9)*3.2,0,0)*P_OP_H2O;
	b_OP_O2_O2P_O	    = HMRATE((1e-11)*1.9,0,0)*P_OP_O2;
	b_OP_OH_O2P_H	    = HMRATE((1e-10)*3.6,0,0)*P_OP_OH;
	b_OP_OH_OHP_O	    = HMRATE((1e-10)*3.6,0,0)*P_OP_OH;
	b_Si_CHP_SiP_CH	    = HMRATE(2.0e-10,0,0)*P_Si_CHP;
	b_Si_H2OP_SiP_H2O    = HMRATE(3.0e-9,0,0)*P_Si_H2OP;
	b_Si_OH_SiO_H	    = HMRATE(2.0e-10,0.5,0)*P_Si_OH;
	b_Si_O2P_O2_SiP	    = HMRATE((1e-9)*1.6,0,0)*P_Si_O2P;
	b_SiP_H2O_SiOHP_H    = HMRATE((1e-10)*2.3,0,0)*P_SiP_H2O;
	b_SiP_OH_SiOP_H	    = HMRATE((1e-10)*6.3,0,0)*P_SiP_OH;
	b_SiP_O2_SiOP_O	    = HMRATE((1e-13)*1.,0,0)*P_SiP_O2;
	b_CH_COP_HCOP_C	    = HMRATE((1e-10)*3.2,0,0)*P_CH_COP;
	b_CH_COP_CO_CHP	    = HMRATE((1e-10)*3.2,0,0)*P_CH_COP;
	b_CH_H2OP_H2O_CHP    = HMRATE((1e-10)*3.4,0,0)*P_CH_H2OP;
	b_CH_H2OP_OH_CH2P    = HMRATE((1e-10)*3.4,0,0)*P_CH_H2OP;
	b_CH_H3OP_H2O_CH2P   = HMRATE((1e-10)*6.8,0,0)*P_CH_H3OP;
	b_CH_HCOP_CO_CH2P    = HMRATE((1e-10)*6.3,0,0)*P_CH_HCOP;
	b_CH_O2_CO_OH        = HMRATE((1e-11)*2.6,0,0)*P_CH_O2;
	b_CH_O2P_O2_CHP      = HMRATE((1e-10)*3.1,0,0)*P_CH_O2P;
	b_CH_O2P_HCOP_O      = HMRATE((1e-10)*3.1,0,0)*P_CH_O2P;
	b_CH_OHP_OH_CHP      = HMRATE(3.5e-10,0,0)*P_CH_OHP;
	b_CH_OHP_O_CH2P      = HMRATE(3.5e-10,0,0)*P_CH_OHP;
	b_CH_SiOP_HCOP_Si    = HMRATE((1e-10)*5.9,0,0)*P_CH_SiOP;
	b_CHP_H2O_H3OP_C     = HMRATE((1e-10)*5.8,0,0)*P_CHP_H2O;
	b_CHP_H2O_HCOP_H2    = HMRATE((1e-9)*2.9,0,0)*P_CHP_H2O;
	b_CHP_OH_COP_H2      = HMRATE(7.5e-10,0,0)*P_CHP_OH;	
	b_CHP_O2_HCOP_O      = HMRATE((1e-10)*9.7,0,0)*P_CHP_O2;
	b_CHP_O2_COP_OH      = HMRATE(1.0e-11,0,0)*P_CHP_O2;
	b_CH2_COP_HCOP_CH    = HMRATE((1e-10)*4.3,0,0)*P_CH2_COP;
	b_CH2_COP_CO_CH2P    = HMRATE((1e-10)*4.3,0,0)*P_CH2_COP;
	b_CH2_H2OP_H2O_CH2P  = HMRATE((1e-10)*4.7,0,0)*P_CH2_H2OP;
	b_CH2_H2OP_OH_CH3P   = HMRATE((1e-10)*4.7,0,0)*P_CH2_H2OP;
	b_CH2_H3OP_H2O_CH3P  = HMRATE((1e-10)*9.4,0,0)*P_CH2_H3OP;
	b_CH2_HCOP_CO_CH3P   = HMRATE((1e-10)*8.6,0,0)*P_CH2_HCOP;
	b_CH2_O2_CO_H2O      = HMRATE((1e-10)*2.48,-3.3,1443)*P_CH2_O2;
	b_CH2_O2P_O2_CH2P    = HMRATE((1e-10)*4.3,0,0)*P_CH2_O2P;
	b_CH2_OH_H2O_CH      = HMRATE((1e-11)*1.44,0.5,3000)*P_CH2_OH;
	b_CH2_OHP_OH_CH2P    = HMRATE((1e-10)*4.8,0,0)*P_CH2_OHP;
	b_CH2_OHP_O_CH3P     = HMRATE((1e-10)*4.8,0,0)*P_CH2_OHP;
	b_CH2P_O2_HCOP_OH    = HMRATE((1e-10)*9.1,0,0)*P_CH2P_O2;
	b_COP_O2_O2P_CO      = HMRATE((1e-10)*1.2,0,0)*P_COP_O2;
	b_H2O_COP_CO_H2OP    = HMRATE((1e-9)*1.72,0,0)*P_H2O_COP;
	b_H2O_COP_HCOP_OH    = HMRATE((1e-10)*8.84,0,0)*P_H2O_COP;
	b_H2O_HCOP_CO_H3OP   = HMRATE((1e-9)*2.5,0,0)*P_H2O_HCOP;
	b_H2OP_CO_HCOP_OH    = HMRATE(5.0e-10,0,0)*P_H2OP_CO;
	b_H2OP_H2O_H3OP_OH   = HMRATE((1e-9)*2.1,0,0)*P_H2OP_H2O;
	b_H2OP_O2_O2P_H2O    = HMRATE((1e-10)*4.6,0,0)*P_H2OP_O2;
	b_H3OP_SiH_SiH2P_H2O = HMRATE((1e-10)*9.7,0,0)*P_H3OP_SiH;
	b_H3OP_SiO_SiOHP_H2O = HMRATE(2.0e-9,0,0)*P_H3OP_SiO;
	b_HCOP_SiH_SiH2P_CO  = HMRATE((1e-10)*8.7,0,0)*P_HCOP_SiH;
	b_HCOP_SiO_SiOHP_CO  = HMRATE((1e-10)*7.9,0,0)*P_HCOP_SiO;
	b_OH_COP_HCOP_O      = HMRATE((1e-10)*3.1,0,0)*P_OH_COP;
	b_OH_COP_CO_OHP      = HMRATE((1e-10)*3.1,0,0)*P_OH_COP;
	b_OH_H2OP_H3OP_O     = HMRATE((1e-10)*6.9,0,0)*P_OH_H2OP;
	b_OH_HCOP_CO_H2OP    = HMRATE(6.2e-10,0,0)*P_OH_HCOP;
	b_OH_OH_H2O_O        = HMRATE((1e-12)*1.65,1.14,50)*P_OH_OH;
	b_OHP_CO_HCOP_O      = HMRATE((1e-9)*1.05,0,0)*P_OHP_CO;
	b_OHP_H2O_H3OP_O     = HMRATE(1.3e-9,0,0)*P_OHP_H2O;
	b_OHP_H2O_H2OP_OH    = HMRATE((1e-9)*1.59,0,0)*P_OHP_H2O;
	b_OHP_O2_O2P_OH      = HMRATE((1e-10)*5.9,0,0)*P_OHP_O2;
	b_OHP_OH_H2OP_O      = HMRATE(7.0e-10,0,0)*P_OHP_OH;
	b_OHP_SiH_SiH2P_O    = HMRATE(1.0e-9,0,0)*P_OHP_SiH;
	b_OHP_SiO_SiOHP_O    = HMRATE((1e-10)*9.4,0,0)*P_OHP_SiO;

	/* new bs */
	b_C_CH5P_CH4_CHP     = HMRATE(1.2e-9,0,0)*P_C_CH5P;
	b_O_CH4_OH_CH3       = HMRATE((1e-12)*2.29,2.2,3820)*P_O_CH4;
	b_O_CH4P_OH_CH3P     = HMRATE(1.0e-9,0,0)*P_O_CH4P;
	b_O_CH5P_H3OP_CH2    = HMRATE((1e-10)*2.2,0,0)*P_O_CH5P;
	b_OP_CH4_OH_CH3P     = HMRATE(1.1e-10,0,0)*P_OP_CH4;
	b_OP_CH4_CH4P_O      = HMRATE((1e-10)*8.9,0,0)*P_OP_CH4;
	b_CH4_CH_CH3_CH2     = HMRATE((1e-11)*2.28,0.7,3000)*P_CH4_CH;
	b_CH5P_CH_CH4_CH2P   = HMRATE((1e-10)*6.9,0,0)*P_CH5P_CH;
	b_CH2_CH2_CH3_CH     = HMRATE(4.0e-10,0,5000)*P_CH2_CH2;
	b_CH4_CH2_CH3_CH3    = HMRATE((1e-12)*7.13,0,5050)*P_CH4_CH2;
	b_OH_CH2_O_CH3       = HMRATE((1e-11)*1.44,0.5,3000)*P_OH_CH2;
	b_CH5P_CH2_CH4_CH3P  = HMRATE((1e-10)*9.6,0,0)*P_CH5P_CH2;
	b_OH_CH3_CH4_O       = HMRATE((1e-14)*3.27,2.2,2240)*P_OH_CH3;
	b_OH_CH3_H2O_CH2     = HMRATE((1e-10)*1.2,0,1400)*P_OH_CH3;
	b_H2O_CH3_OH_CH4     = HMRATE((1e-15)*2.3,3.47,6681)*P_H2O_CH3;
	b_CH3_CH3_CH4_CH2    = HMRATE((1e-12)*7.13,0,5052)*P_CH3_CH3;
	b_OH_CH4_H2O_CH3     = HMRATE((1e-13)*3.77,2.42,1162)*P_OH_CH4;
	b_OHP_CH4_CH5P_O     = HMRATE((1e-10)*1.95,0,0)*P_OHP_CH4;
	b_OHP_CH4_H3OP_CH2   = HMRATE((1e-9)*1.31,0,0)*P_OHP_CH4;
	b_H2OP_CH4_H3OP_CH3  = HMRATE((1e-9)*1.4,0,0)*P_H2OP_CH4;
	b_COP_CH4_HCOP_CH3   = HMRATE((1e-10)*4.55,0,0)*P_COP_CH4;
	b_COP_CH4_CO_CH4P    = HMRATE((1e-10)*7.93,0,0)*P_COP_CH4;
	b_CH4_CH4P_CH5P_CH3  = HMRATE((1e-9)*1.5,0,0)*P_CH4_CH4P;
	b_H2O_CH4P_H3OP_CH3  = HMRATE((1e-9)*2.6,0,0)*P_H2O_CH4P;
	b_CO_CH4P_HCOP_CH3   = HMRATE((1e-9)*1.4,0,0)*P_CO_CH4P;
	b_O2_CH4P_O2P_CH4    = HMRATE(4.0e-10,0,0)*P_O2_CH4P;
	b_H2O_CH5P_H3OP_CH4  = HMRATE((1e-9)*3.7,0,0)*P_H2O_CH5P;
	b_CO_CH5P_HCOP_CH4   = HMRATE(1.0e-9,0,0)*P_CO_CH5P;
	b_CH5P_OH_H2OP_CH4   = HMRATE(7.0e-10,0,0)*P_CH5P_OH;
	b_C_CH_C2_H = HMRATE((1e-11)*1.73,0.5,0)*P_C_CH;
	b_C_CN_C2_N = HMRATE(4.98e-10,0,18116)*P_C_CN;
	b_C_CO_C2_O = HMRATE(2.94e-11,0.5,58025)*P_C_CO;
	b_C_CS_S_C2 = HMRATE((1e-11)*1.44,0.5,20435)*P_C_CS;
	b_N_C2_CN_C = HMRATE((1e-11)*5.,0.5,0)*P_N_C2;
	b_O_C2_CO_C = HMRATE((1e-11)*5.,0.5,0)*P_O_C2;
	b_C2_O2_CO_CO = HMRATE((1e-12)*1.83,0,0)*P_C2_O2;
	b_C2_S_CS_C = HMRATE((1e-11)*1.73,0.5,0)*P_C2_S;
	b_NHP_C2_HCNP_C = HMRATE((1e-10)*4.9,0,0)*P_NHP_C2;
	b_OP_C2_COP_C = HMRATE((1e-10)*4.8,0,0)*P_OP_C2;
	b_C2_O2P_COP_CO = HMRATE((1e-10)*4.1,0,0)*P_C2_O2P;
	b_C2_SP_CSP_C = HMRATE((1e-10)*8.1,0,0)*P_C2_SP;
	b_C_C2P_C2_CP = HMRATE(1.1e-10,0,0)*P_C_C2P;
	b_CH_C2P_C2_CHP = HMRATE((1e-10)*3.2,0,0)*P_CH_C2P;
	b_NP_C2_C2P_N = HMRATE(1.0e-9,0,0)*P_NP_C2;
	b_CH2_C2P_C2_CH2P = HMRATE((1e-10)*4.5,0,0)*P_CH2_C2P;
	b_OP_C2_C2P_O = HMRATE((1e-10)*4.8,0,0)*P_OP_C2;
	b_NH2_C2P_C2_NH2P = HMRATE((1e-10)*4.6,0,0)*P_NH2_C2P;
	b_OHP_C2_C2P_OH = HMRATE((1e-10)*4.8,0,0)*P_OHP_C2;
	b_OH_C2P_C2_OHP = HMRATE((1e-10)*6.5,0,0)*P_OH_C2P;
	b_H2OP_C2_C2P_H2O = HMRATE((1e-10)*4.7,0,0)*P_H2OP_C2;
	b_C2_CNP_CN_C2P = HMRATE((1e-10)*8.5,0,0)*P_C2_CNP;
	b_C2_COP_CO_C2P = HMRATE((1e-10)*8.4,0,0)*P_C2_COP;
	b_C2_N2P_N2_C2P = HMRATE((1e-10)*8.4,0,0)*P_C2_N2P;
	b_C2P_NO_NOP_C2 = HMRATE((1e-10)*3.4,0,0)*P_C2P_NO;
	b_C2_O2P_O2_C2P = HMRATE((1e-10)*4.1,0,0)*P_C2_O2P;
	b_C2P_S_SP_C2 = HMRATE((1e-10)*5.8,0,0)*P_C2P_S;
	b_C_C_C2_nu = HMRATE(4.36E-18,0.35,161.3)*P_C_C;
	b_CP_CH_C2P_H = HMRATE(3.8e-10,0,0)*P_CP_CH;
	b_C_CHP_C2P_H = HMRATE(1.2e-9,0,0)*P_C_CHP;
	b_CHP_CH_C2P_H2 = HMRATE((1e-10)*7.4,0,0)*P_CHP_CH;
	b_N_C2P_CN_CP = HMRATE(4.0e-11,0,0)*P_N_C2P;
	b_O_C2P_COP_C = HMRATE((1e-10)*3.1,0,0)*P_O_C2P;
	b_C2P_O2_COP_CO = HMRATE((1e-10)*8.,0,0)*P_C2P_O2;
	b_C2P_S_CSP_C = HMRATE((1e-10)*5.8,0,0)*P_C2P_S;
	b_CP_C_C2P_nu = HMRATE(4.01E-18,0.17,101.5)*P_CP_C;
	H_C2_CH_C = HMRATE((1e-10)*4.67,0.5,30450)*HI;
	C_CH_C2_H_1 = HMRATE((1e-11)*1.73,0.5,0)*CH;
	C_CN_C2_N_1 = HMRATE(4.98e-10,0,18116)*CN;
	C_CO_C2_O_1 = HMRATE(2.94e-11,0.5,58025)*CO;
	C_CS_S_C2_1 = HMRATE((1e-11)*1.44,0.5,20435)*CS;
	N_C2_CN_C_1 = HMRATE((1e-11)*5.,0.5,0)*C2;
	O_C2_CO_C_1 = HMRATE((1e-11)*5.,0.5,0)*C2;
	C2_O2_CO_CO_1 = HMRATE((1e-12)*1.83,0,0)*O2;
	C2_S_CS_C_1 = HMRATE((1e-11)*1.73,0.5,0)*ATS;
	HeP_C2_CP_C_He = HMRATE((1e-9)*1.6,0,0)*HeII;
	NHP_C2_HCNP_C_1 = HMRATE((1e-10)*4.9,0,0)*C2;
	OP_C2_COP_C_1 = HMRATE((1e-10)*4.8,0,0)*C2;
	C2_O2P_COP_CO_1 = HMRATE((1e-10)*4.1,0,0)*O2P;
	C2_SP_CSP_C_1 = HMRATE((1e-10)*8.1,0,0)*SP;
	HP_C2_C2P_H = HMRATE((1e-9)*3.1,0,0)*HII;
	H2P_C2_C2P_H2 = HMRATE(1.1e-9,0,0)*H2p;
	HeP_C2_C2P_He = HMRATE(5.0e-10,0,0)*HeII;
	C_C2P_C2_CP_1 = HMRATE(1.1e-10,0,0)*C2P;
	CH_C2P_C2_CHP_1 = HMRATE((1e-10)*3.2,0,0)*C2P;
	NP_C2_C2P_N_1 = HMRATE(1.0e-9,0,0)*C2;
	CH2_C2P_C2_CH2P_1 = HMRATE((1e-10)*4.5,0,0)*C2P;
	OP_C2_C2P_O_1 = HMRATE((1e-10)*4.8,0,0)*C2;
	NH2_C2P_C2_NH2P_1 = HMRATE((1e-10)*4.6,0,0)*C2P;
	OHP_C2_C2P_OH_1 = HMRATE((1e-10)*4.8,0,0)*C2;
	OH_C2P_C2_OHP_1 = HMRATE((1e-10)*6.5,0,0)*C2P;
	H2OP_C2_C2P_H2O_1 = HMRATE((1e-10)*4.7,0,0)*C2;
	C2_CNP_CN_C2P_1 = HMRATE((1e-10)*8.5,0,0)*CNP;
	C2_COP_CO_C2P_1 = HMRATE((1e-10)*8.4,0,0)*COP;
	C2_N2P_N2_C2P_1 = HMRATE((1e-10)*8.4,0,0)*N2P;
	C2P_NO_NOP_C2_1 = HMRATE((1e-10)*3.4,0,0)*NO;
	C2_O2P_O2_C2P_1 = HMRATE((1e-10)*4.1,0,0)*O2P;
	C2P_S_SP_C2_1 = HMRATE((1e-10)*5.8,0,0)*ATS;
	C_C_C2_nu_1 = HMRATE(4.36E-18,0.35,161.3)*ATC;
	C2_nu_C_C = HMRATE((1e-10)*1.5,0,0)*(TH85_field);
	C2_nu_C2P_e = HMRATE((1e-10)*4.1,0,0)*(TH85_field);
	C2_cr_C_C = crnu_rate * 2. *500;
	CP_CH_C2P_H_1 = HMRATE(3.8e-10,0,0)*CH;
	C_CHP_C2P_H_1 = HMRATE(1.2e-9,0,0)*CHP;
	CHP_CH_C2P_H2_1 = HMRATE((1e-10)*7.4,0,0)*CH;
	N_C2P_CN_CP_1 = HMRATE(4.0e-11,0,0)*C2P;
	O_C2P_COP_C_1 = HMRATE((1e-10)*3.1,0,0)*C2P;
	C2P_O2_COP_CO_1 = HMRATE((1e-10)*8.,0,0)*O2;
	C2P_S_CSP_C_1 = HMRATE((1e-10)*5.8,0,0)*ATS;
	C2P_e_C_C = HMRATE(0.0000003,-0.5,0)*ele;
	CP_C_C2P_nu_1 = HMRATE(4.01E-18,0.17,101.5)*ATC;
	/* >>chng 05 mar 21, had mult by den of C2P */
	C2P_nu_CP_C = HMRATE(1.0e-11,0,0)*(TH85_field);
	C_CH_C2_H_2 = HMRATE((1e-11)*1.73,0.5,0)*ATC;
	C_CN_C2_N_2 = HMRATE(4.98e-10,0,18116)*ATC;
	C_CO_C2_O_2 = HMRATE(2.94e-11,0.5,58025)*ATC;
	C_CS_S_C2_2 = HMRATE((1e-11)*1.44,0.5,20435)*ATC;
	N_C2_CN_C_2 = HMRATE((1e-11)*5.,0.5,0)*ATN;
	O_C2_CO_C_2 = HMRATE((1e-11)*5.,0.5,0)*ATO;
	C2_O2_CO_CO_2 = HMRATE((1e-12)*1.83,0,0)*C2;
	C2_S_CS_C_2 = HMRATE((1e-11)*1.73,0.5,0)*C2;
	NHP_C2_HCNP_C_2 = HMRATE((1e-10)*4.9,0,0)*NHP;
	OP_C2_COP_C_2 = HMRATE((1e-10)*4.8,0,0)*OP;
	C2_O2P_COP_CO_2 = HMRATE((1e-10)*4.1,0,0)*C2;
	C2_SP_CSP_C_2 = HMRATE((1e-10)*8.1,0,0)*C2;
	C_C2P_C2_CP_2 = HMRATE(1.1e-10,0,0)*ATC;
	CH_C2P_C2_CHP_2 = HMRATE((1e-10)*3.2,0,0)*CH;
	NP_C2_C2P_N_2 = HMRATE(1.0e-9,0,0)*NP;
	CH2_C2P_C2_CH2P_2 = HMRATE((1e-10)*4.5,0,0)*CH2;
	OP_C2_C2P_O_2 = HMRATE((1e-10)*4.8,0,0)*OP;
	NH2_C2P_C2_NH2P_2 = HMRATE((1e-10)*4.6,0,0)*NH2;
	OHP_C2_C2P_OH_2 = HMRATE((1e-10)*4.8,0,0)*OHP;
	OH_C2P_C2_OHP_2 = HMRATE((1e-10)*6.5,0,0)*OH;
	H2OP_C2_C2P_H2O_2 = HMRATE((1e-10)*4.7,0,0)*H2OP;
	C2_CNP_CN_C2P_2 = HMRATE((1e-10)*8.5,0,0)*C2;
	C2_COP_CO_C2P_2 = HMRATE((1e-10)*8.4,0,0)*C2;
	C2_N2P_N2_C2P_2 = HMRATE((1e-10)*8.4,0,0)*C2;
	C2P_NO_NOP_C2_2 = HMRATE((1e-10)*3.4,0,0)*C2P;
	C2_O2P_O2_C2P_2 = HMRATE((1e-10)*4.1,0,0)*C2;
	C2P_S_SP_C2_2 = HMRATE((1e-10)*5.8,0,0)*C2P;
	C_C_C2_nu_2 = HMRATE(4.36E-18,0.35,161.3)*ATC;
	CP_CH_C2P_H_2 = HMRATE(3.8e-10,0,0)*CP;
	C_CHP_C2P_H_2 = HMRATE(1.2e-9,0,0)*ATC;
	CHP_CH_C2P_H2_2 = HMRATE((1e-10)*7.4,0,0)*CHP;
	N_C2P_CN_CP_2 = HMRATE(4.0e-11,0,0)*ATN;
	O_C2P_COP_C_2 = HMRATE((1e-10)*3.1,0,0)*ATO;
	C2P_O2_COP_CO_2 = HMRATE((1e-10)*8.,0,0)*C2P;
	C2P_S_CSP_C_2 = HMRATE((1e-10)*5.8,0,0)*C2P;
	CP_C_C2P_nu_2 = HMRATE(4.01E-18,0.17,101.5)*CP;

	/* >>chng 02 jul 04 -- The following are dissociation rates for various molecular species
	For right now we will calculate this rate by the standard form:
	   
			(alpha)*Go*exp(-Beta*AV)


    when the command "set Leiden hack UMIST rates" is used.  Otherwise we
	will just let cloudy calculate the value of the UV radiation field */

	if (!co.lgUMISTrates)
	{

		nu_CH_CHP_e             = HMRATE((1e-10)*7.6,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(2.8*rfield.extin_mag_V_point))/1.66);
		nu_CH_C_H               = HMRATE((1e-10)*8.6,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(1.2*rfield.extin_mag_V_point))/1.66);
		nu_CHP_CP_H             = HMRATE((1e-10)*2.5,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(2.5*rfield.extin_mag_V_point))/1.66);
		nu_CH2_CH2P_e           = HMRATE(1.0e-9,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(2.3*rfield.extin_mag_V_point))/1.66);
		nu_CH2_CH_H             = HMRATE((1e-10)*7.2,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(1.7*rfield.extin_mag_V_point))/1.66);
		nu_CH2P_CHP_H           = HMRATE((1e-9)*1.7,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(1.7*rfield.extin_mag_V_point))/1.66);
		nu_CH3P_CH2P_H          = HMRATE(1.0e-9,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(1.7*rfield.extin_mag_V_point))/1.66);
		nu_CH3P_CHP_H2          = HMRATE(1.0e-9,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(1.7*rfield.extin_mag_V_point))/1.66);
		nu_CO_O_C               = HMRATE(2.0e-10,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(3.2*rfield.extin_mag_V_point))/1.66);;
		nu_H2O_H2OP_e           = HMRATE(3.3e-11,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(3.9*rfield.extin_mag_V_point))/1.66);
		nu_H2O_OH_H             = HMRATE((1e-10)*5.9,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(1.7*rfield.extin_mag_V_point))/1.66);
		nu_O2_O2P_e             = HMRATE((1e-11)*5.6,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(3.7*rfield.extin_mag_V_point))/1.66);
		nu_O2_O_O               = HMRATE((1e-10)*6.9,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(1.8*rfield.extin_mag_V_point))/1.66);
		nu_OH_OHP_e             = HMRATE((1e-12)*1.6,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(3.1*rfield.extin_mag_V_point))/1.66);
		nu_OH_O_H               = HMRATE(3.5e-10,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(1.7*rfield.extin_mag_V_point))/1.66);
		nu_OHP_O_HP             = HMRATE((1e-12)*1.,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(1.8*rfield.extin_mag_V_point))/1.66);
		nu_SiH_Si_H             = HMRATE((1e-9)*2.8,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(1.1*rfield.extin_mag_V_point))/1.66);
		nu_SiO_Si_O             = HMRATE(1.0e-10,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(2.3*rfield.extin_mag_V_point))/1.66);
		nu_CH3_CH3P_e           = HMRATE(1.0e-10,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(2.1*rfield.extin_mag_V_point))/1.66);
		nu_CH3_CH2_H            = HMRATE((1e-10)*2.5,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(1.9*rfield.extin_mag_V_point))/1.66);
		nu_CH3_CH_H2            = HMRATE((1e-10)*2.5,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(1.9*rfield.extin_mag_V_point))/1.66);
		nu_CH4_CH3_H            = HMRATE((1e-10)*2.2,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(2.2*rfield.extin_mag_V_point))/1.66);
		nu_CH4_CH2_H2           = HMRATE((1e-10)*9.8,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(2.2*rfield.extin_mag_V_point))/1.66);
		nu_CH4_CH_H2            = HMRATE((1e-10)*2.2,0,0)*(hmi.UV_Cont_rel2_Habing_TH85_face*(float)exp(-(2.2*rfield.extin_mag_V_point))/1.66);
		
	}
	/*  This is what will usually be done, calculating the photodissociation rate by calculating the UV
	 * radiation field self-consistently. */
	else
	{
		
		nu_CH_CHP_e             = HMRATE((1e-10)*7.6,0,0)*(TH85_field);
		nu_CH_C_H               = HMRATE((1e-10)*8.6,0,0)*(TH85_field);
		nu_CHP_CP_H             = HMRATE((1e-10)*2.5,0,0)*(TH85_field);
		nu_CH2_CH2P_e           = HMRATE(1.0e-9,0,0)*(TH85_field);
		nu_CH2_CH_H             = HMRATE((1e-10)*7.2,0,0)*(TH85_field);
		nu_CH2P_CHP_H           = HMRATE((1e-9)*1.7,0,0)*(TH85_field);
		nu_CH3P_CH2P_H          = HMRATE(1.0e-9,0,0)*(TH85_field);
		nu_CH3P_CHP_H2          = HMRATE(1.0e-9,0,0)*(TH85_field);
		nu_CO_O_C               = HMRATE(2.0e-10,0,0)*(TH85_field);
		nu_H2O_H2OP_e           = HMRATE(3.3e-11,0,0)*(TH85_field);
		nu_H2O_OH_H             = HMRATE((1e-10)*5.9,0,0)*(TH85_field);
		nu_O2_O2P_e             = HMRATE((1e-11)*5.6,0,0)*(TH85_field);
		nu_O2_O_O               = HMRATE((1e-10)*6.9,0,0)*(TH85_field);
		nu_OH_OHP_e             = HMRATE((1e-12)*1.6,0,0)*(TH85_field);
		nu_OH_O_H               = HMRATE(3.5e-10,0,0)*(TH85_field);
		nu_OHP_O_HP             = HMRATE((1e-12)*1.,0,0)*(TH85_field);
		nu_SiH_Si_H             = HMRATE((1e-9)*2.8,0,0)*(TH85_field);
		nu_SiO_Si_O             = HMRATE(1.0e-10,0,0)*(TH85_field);
		nu_CH3_CH3P_e           = HMRATE(1.0e-10,0,0)*(TH85_field);
		nu_CH3_CH2_H            = HMRATE((1e-10)*2.5,0,0)*(TH85_field);
		nu_CH3_CH_H2            = HMRATE((1e-10)*2.5,0,0)*(TH85_field);
		nu_CH4_CH3_H            = HMRATE((1e-10)*2.2,0,0)*(TH85_field);
		nu_CH4_CH2_H2           = HMRATE((1e-10)*9.8,0,0)*(TH85_field);
		nu_CH4_CH_H2            = HMRATE((1e-10)*2.2,0,0)*(TH85_field);
	}
	C_NH_N_CH_1          = HMRATE((1e-11)*1.73,0.5,4000)*ATC;
	C_NH_CN_H_1          = HMRATE(1.1e-10,0.5,0)*ATC;
	C_N2_CN_N_1          = HMRATE((1e-11)*8.69,0,22600)*ATC;
	C_NO_CN_O_1          = HMRATE((1e-11)*4.8,0,0)*ATC;
	C_NO_CO_N_1          = HMRATE((1e-11)*4.65,0,0)*ATC;
	C_HS_S_CH_1          = HMRATE((1e-11)*1.2,0.58,5880)*ATC;
	C_HS_CS_H_1          = HMRATE(2.0e-11,0,0)*ATC;
	C_OCN_CO_CN_1        = HMRATE(4.0e-11,0.5,0)*ATC;
	C_NS_S_CN_1          = HMRATE(2.0e-11,0.5,0)*ATC;
	C_NS_CS_N_1          = HMRATE((1e-11)*1.73,0.5,4000)*ATC;
	C_SO_S_CO_1          = HMRATE((1e-11)*7.2,0,0)*ATC;
	C_S2_CS_S_1          = HMRATE((1e-11)*1.73,0.5,0)*ATC;
	CH_N_NH_C_1          = HMRATE((1e-11)*3.03,0.65,1207)*CH;
	CH_N_CN_H_1          = HMRATE((1e-10)*1.66,-0.09,0)*CH;
	CH_N2_HCN_N_1        = HMRATE(0.00000000000056,0.88,10128)*CH;
	CH_NO_HCN_O_1        = HMRATE((1e-9)*5.59,0,10814)*CH;
	CH_NO_CN_OH_1        = HMRATE(2.32E-26,0,0)*CH;
	CH_NO_OCN_H_1        = HMRATE(1.13E-25,0,0)*CH;
	CH_HNO_NO_CH2_1      = HMRATE((1e-11)*1.73,0.5,0)*CH;
	CH_S_HS_C_1          = HMRATE((1e-11)*1.73,0.5,4000)*CH;
	CH_S_CS_H_1          = HMRATE((1e-12)*1.1,0,0)*CH;
	N_NH_N2_H_1          = HMRATE((1e-11)*4.98,0,0)*ATN;
	N_CH3_HCN_H_H_1      = HMRATE((1e-13)*3.32,0,0)*ATN;
	N_CH3_HCN_H2_1       = HMRATE((1e-11)*1.3,0.5,0)*ATN;
	N_OH_O_NH_1          = HMRATE((1e-11)*1.88,0.1,10700)*ATN;
	N_OH_NO_H_1          = HMRATE((1e-11)*5.32,-0.25,0)*ATN;
	N_CN_N2_C_1          = HMRATE((1e-10)*3.,0,0)*ATN;
	N_SIH_SIN_H_1        = HMRATE((1e-11)*5.,0.5,0)*ATN;
	N_NO_N2_O_1          = HMRATE((1e-11)*3.75,0,26)*ATN;
	N_HNO_NO_NH_1        = HMRATE((1e-12)*2.94,0.5,1000)*ATN;
	N_HNO_N2O_H_1        = HMRATE((1e-12)*1.43,0.5,1500)*ATN;
	N_O2_NO_O_1          = HMRATE((1e-12)*2.26,0.86,3134)*ATN;
	N_HS_NS_H_1          = HMRATE((1e-11)*1.73,0.5,0)*ATN;
	N_HS_S_NH_1          = HMRATE((1e-11)*1.73,0.5,9060)*ATN;
	N_CS_S_CN_1          = HMRATE((1e-11)*3.8,0.5,1160)*ATN;
	N_NS_S_N2_1          = HMRATE((1e-11)*1.73,0.5,0)*ATN;
	N_SO_NS_O_1          = HMRATE((1e-11)*4.68,0.5,8254)*ATN;
	N_SO_S_NO_1          = HMRATE((1e-11)*1.73,0.5,750)*ATN;
	N_S2_NS_S_1          = HMRATE((1e-11)*1.73,0.5,4000)*ATN;
	CH2_CN_HCN_CH_1      = HMRATE((1e-12)*5.3,0,2500)*CH2;
	CH2_NO_HCN_OH_1      = HMRATE(0.000000000000832,0,1443)*CH2;
	CH2_HNO_NO_CH3_1     = HMRATE((1e-11)*1.73,0.5,0)*CH2;
	NH_NH_N2_H_H_1       = HMRATE((1e-9)*1.16,0,0)*NH;
	NH_NH_N2_H2_1        = HMRATE((1e-11)*1.7,0,0)*NH;
	NH_O_OH_N_1          = HMRATE((1e-11)*1.16,0,0)*NH;
	NH_O_NO_H_1          = HMRATE((1e-10)*1.16,0,0)*NH;
	NH_OH_HNO_H_1        = HMRATE((1e-11)*3.32,0,0)*NH;
	NH_OH_NH2_O_1        = HMRATE((1e-12)*2.93,0.1,5800)*NH;
	NH_OH_H2O_N_1        = HMRATE((1e-12)*3.11,1.2,0)*NH;
	NH_H2O_OH_NH2_1      = HMRATE((1e-12)*1.83,1.6,14090)*NH;
	NH_CN_HCN_N_1        = HMRATE((1e-12)*2.94,0.5,1000)*NH;
	NH_NO_N2O_H_1        = HMRATE((1e-10)*1.16,-1.03,420)*NH;
	NH_NO_N2_O_H_1       = HMRATE((1e-11)*5.,0,0)*NH;
	NH_NO_N2_OH_1        = HMRATE((1e-11)*1.46,-0.58,37)*NH;
	NH_S_HS_N_1          = HMRATE((1e-11)*1.73,0.5,4000)*NH;
	NH_S_NS_H_1          = HMRATE((1e-11)*1.73,0.5,0)*NH;
	NH_NO2_HNO_NO_1      = HMRATE((1e-12)*5.72,0.5,2500)*NH;
	NH_NO2_N2O_OH_1      = HMRATE((1e-13)*1.44,0, 1140)*NH;
	CH3_NH3_CH4_NH2_1    = HMRATE(0.0000000000000955,0,4890)*CH3;
	CH3_CN_HCN_CH2_1     = HMRATE((1e-12)*9.21,0.7,1500)*CH3;
	CH3_HNO_NO_CH4_1     = HMRATE((1e-11)*1.44,0.5,0)*CH3;
	O_NH2_OH_NH_1        = HMRATE((1e-11)*1.39,0,40)*ATO;
	O_NH2_NO_H2_1        = HMRATE((1e-12)*8.3,0,0)*ATO;
	O_NH3_OH_NH2_1       = HMRATE((1e-11)*1.89,0,4003)*ATO;
	O_CN_NO_C_1          = HMRATE((1e-11)*3.81,0.5,14545)*ATO;
	O_CN_CO_N_1          = HMRATE((1e-11)*4.36,0.46,364)*ATO;
	O_HCN_CN_OH_1        = HMRATE((1e-10)*6.21,0,12439)*ATO;
	O_HCN_CO_NH_1        = HMRATE(0.00000000000073,1.14,3742)*ATO;
	O_HCN_OCN_H_1        = HMRATE((1e-12)*1.36,1.38,3693)*ATO;
	O_N2_NO_N_1          = HMRATE((1e-10)*2.51,0,38602)*ATO;
	O_NO_O2_N_1          = HMRATE((1e-11)*1.18,0,20413)*ATO;
	O_HNO_NO_OH_1        = HMRATE((1e-11)*6.,0,0)*ATO;
	O_HNO_O2_NH_1        = HMRATE((1e-12)*2.94,0.5,3500)*ATO;
	O_HNO_NO2_H_1        = HMRATE((1e-12)*1.44,0.5,0)*ATO;
	O_HS_S_OH_1          = HMRATE((1e-11)*1.74,0.67,956)*ATO;
	O_HS_SO_H_1          = HMRATE((1e-10)*2.32,0,0)*ATO;
	O_OCN_NO_CO_1        = HMRATE((1e-11)*9.43,-0.09,100)*ATO;
	O_OCN_O2_CN_1        = HMRATE((1e-10)*4.02,-1.43,3501)*ATO;
	O_SIN_NO_SI_1        = HMRATE((1e-11)*2.5,0.5,0)*ATO;
	O_SIN_SIO_N_1        = HMRATE((1e-11)*5.75,0.1,200)*ATO;
	O_N2O_NO_NO_1        = HMRATE((1e-10)*1.15,0,13400)*ATO;
	O_N2O_O2_N2_1        = HMRATE((1e-10)*1.66,0,14100)*ATO;
	O_CS_S_CO_1          = HMRATE((1e-10)*2.48,-0.65,783)*ATO;
	O_CS_SO_C_1          = HMRATE((1e-11)*4.68,0.5,28940)*ATO;
	O_NS_S_NO_1          = HMRATE((1e-11)*5.,0.5,0)*ATO;
	O_NS_SO_N_1          = HMRATE((1e-11)*1.73,0.5,4000)*ATO;
	O_SO_S_O2_1          = HMRATE(0.00000000000066,0,2760)*ATO;
	O_OCS_SO_CO_1        = HMRATE((1e-11)*1.6,0,2150)*ATO;
	O_S2_SO_S_1          = HMRATE((1e-11)*1.73,0.5,0)*ATO;
	NH2_OH_NH3_O_1       = HMRATE(0.000000000000208,0.76,262)*NH2;
	NH2_OH_H2O_NH_1      = HMRATE((1e-12)*1.5,0,0)*NH2;
	NH2_NO_N2_OH_H_1     = HMRATE((1e-12)*1.49,0,0)*NH2;
	NH2_NO_N2_H2O_1      = HMRATE((1e-11)*4.27,-2.5,331)*NH2;
	CH4_CN_HCN_CH3_1     = HMRATE((1e-12)*3.14,1.53,504)*CH4;
	OH_NH3_H2O_NH2_1     = HMRATE((1e-13)*1.47,2.05,7)*OH;
	OH_CN_HCN_O_1        = HMRATE(1.0e-11,0,1000)*OH;
	OH_CN_OCN_H_1        = HMRATE((1e-11)*7.,0,0)*OH;
	OH_HCN_CN_H2O_1      = HMRATE((1e-13)*1.87,1.5,3887)*OH;
	OH_HCN_CO_NH2_1      = HMRATE((1e-13)*1.07,0,5892)*OH;
	OH_NO_NO2_H_1        = HMRATE((1e-12)*5.2,0,15100)*OH;
	OH_S_HS_O_1          = HMRATE(6.6e-11,0,0)*OH;
	OH_S_SO_H_1          = HMRATE(6.6e-11,0,0)*OH;
	OH_N2O_HNO_NO_1      = HMRATE(1.04E-17,4.33,12623)*OH;
	OH_CS_OCS_H_1        = HMRATE(0.0000000000000939,1.12,800)*OH;
	NH3_CN_HCN_NH2_1     = HMRATE((1e-11)*2.6,-1.1,0)*NH3;
	CN_NO_N2_CO_1        = HMRATE((1e-10)*1.79,0,4040)*CN;
	CN_NO_OCN_N_1        = HMRATE((1e-10)*1.62,0,21205)*CN;
	CN_HNO_NO_HCN_1      = HMRATE((1e-11)*1.5,0.5,0)*CN;
	CN_O2_NO_CO_1        = HMRATE(0.00000000000053,0,0)*CN;
	CN_S_NS_C_1          = HMRATE((1e-11)*5.71,0.5,32010)*CN;
	CN_S_CS_N_1          = HMRATE((1e-11)*1.73,0.5,0)*CN;
	CO_HS_OCS_H_1        = HMRATE(0.0000000000000595,1.12,8330)*CO;
	N2_O2_N2O_O_1        = HMRATE(1.0e-10,0,55200)*N2;
	NO_NO_O2_N2_1        = HMRATE((1e-11)*2.51,0,30653)*NO;
	NO_NO_N2O_O_1        = HMRATE((1e-12)*7.22,0,33155)*NO;
	NO_HNO_N2O_OH_1      = HMRATE((1e-11)*1.41,0,14890)*NO;
	NO_S_NS_O_1          = HMRATE(2.94e-11,0.5,17465)*NO;
	NO_S_SO_N_1          = HMRATE((1e-10)*1.75,0,20200)*NO;
	O2_S_SO_O_1          = HMRATE((1e-12)*2.28,0.52,0)*O2;
	S_SO_S2_O_1          = HMRATE((1e-11)*1.73,0.5,11500)*ATS;
	C_NHP_N_CHP_1        = HMRATE((1e-9)*1.6,0,0)*ATC;
	CP_NH_CNP_H_1        = HMRATE((1e-10)*7.8,0,0)*CP;
	CP_NH2_HCNP_H_1      = HMRATE(1.1e-9,0,0)*CP;
	C_NH2P_NH_CHP_1      = HMRATE(1.2e-9,0,0)*ATC;
	CP_NH3_HCNP_H2_1     = HMRATE((1e-11)*7.,0,0)*CP;
	C_NH3P_NH_CH2P_1     = HMRATE(1.0e-11,0,0)*ATC;
	C_HCNP_CN_CHP_1      = HMRATE(1.1e-9,0,0)*ATC;
	C_HNOP_NO_CHP_1      = HMRATE(1.0e-9,0,0)*ATC;
	CP_HS_CSP_H_1        = HMRATE(1.1e-9,0,0)*CP;
	C_HSP_CSP_H_1        = HMRATE((1e-10)*9.9,0,0)*ATC;
	CP_OCN_COP_CN_1      = HMRATE((1e-9)*3.8,0,0)*CP;
	CP_NS_CSP_N_1        = HMRATE((1e-10)*7.6,0,0)*CP;
	CP_SO_SP_CO_1        = HMRATE((1e-10)*2.6,0,0)*CP;
	CP_SO_S_COP_1        = HMRATE((1e-10)*2.6,0,0)*CP;
	CP_SO_CSP_O_1        = HMRATE((1e-10)*2.6,0,0)*CP;
	CP_OCS_CSP_CO_1      = HMRATE((1e-9)*1.6,0,0)*CP;
	CHP_N_CNP_H_1        = HMRATE((1e-10)*1.9,0,0)*CHP;
	CH_NP_CNP_H_1        = HMRATE((1e-10)*3.6,0,0)*CH;
	CH_NHP_CH2P_N_1      = HMRATE((1e-10)*9.9,0,0)*CH;
	CHP_NH_CNP_H2_1      = HMRATE((1e-10)*7.6,0,0)*CHP;
	CHP_NH2_HCNP_H2_1    = HMRATE(1.1e-9,0,0)*CHP;
	CH_NH2P_NH_CH2P_1    = HMRATE(3.5e-10,0,0)*CH;
	CHP_NH3_NH4P_C_1     = HMRATE((1e-10)*4.05,0,0)*CHP;
	CH_NH3P_NH4P_C_1     = HMRATE((1e-10)*6.9,0,0)*CH;
	CH_HCNP_CN_CH2P_1    = HMRATE((1e-10)*6.3,0,0)*CH;
	CH_HNOP_NO_CH2P_1    = HMRATE(6.2e-10,0,0)*CH;
	CHP_S_HSP_C_1        = HMRATE((1e-10)*4.7,0,0)*CHP;
	CHP_S_CSP_H_1        = HMRATE((1e-10)*4.7,0,0)*CHP;
	CH_SP_CSP_H_1        = HMRATE(6.2e-10,0,0)*CH;
	CH_HSP_S_CH2P_1      = HMRATE((1e-10)*5.8,0,0)*CH;
	CHP_OCS_HCSP_CO_1    = HMRATE((1e-9)*1.05,0,0)*CHP;
	N_CH2P_HCNP_H_1      = HMRATE((1e-10)*2.2,0,0)*ATN;
	NP_NH_N2P_H_1        = HMRATE((1e-10)*3.7,0,0)*NP;
	N_NHP_N2P_H_1        = HMRATE(1.3e-9,0,0)*ATN;
	NP_CH4_HCNP_H2_H_1   = HMRATE((1e-11)*5.6,0,0)*NP;
	N_OHP_NOP_H_1        = HMRATE((1e-10)*8.9,0,0)*ATN;
	NP_NH3_NH2P_NH_1     = HMRATE((1e-10)*2.16,0,0)*NP;
	N_H2OP_HNOP_H_1      = HMRATE((1e-10)*1.9,0,0)*ATN;
	NP_CO_NOP_C_1        = HMRATE((1e-10)*1.45,0,0)*NP;
	NP_NO_N2P_O_1        = HMRATE((1e-11)*7.9,0,0)*NP;
	NP_O2_NO_OP_1        = HMRATE((1e-11)*3.66,0,0)*NP;
	NP_O2_NOP_O_1        = HMRATE((1e-10)*2.63,0,0)*NP;
	N_O2P_NOP_O_1        = HMRATE((1e-10)*1.8,0,0)*ATN;
	N_HSP_NSP_H_1        = HMRATE((1e-10)*7.4,0,0)*ATN;
	N_SIOP_NOP_SI_1      = HMRATE((1e-11)*9.,0,0)*ATN;
	N_SIOP_NO_SIP_1      = HMRATE((1e-10)*2.1,0,0)*ATN;
	N_SOP_NSP_O_1        = HMRATE((1e-11)*5.,0,0)*ATN;
	NP_OCS_SP_CO_N_1     = HMRATE((1e-10)*3.08,0,0)*NP;
	NP_OCS_CSP_NO_1      = HMRATE((1e-11)*7.,0,0)*NP;
	CH2_NHP_CH3P_N_1     = HMRATE((1e-9)*1.4,0,0)*CH2;
	CH2_NH2P_CH3P_NH_1   = HMRATE((1e-10)*4.9,0,0)*CH2;
	CH2P_NH3_NH4P_CH_1   = HMRATE((1e-9)*1.26,0,0)*CH2P;
	CH2_NH3P_NH2_CH3P_1  = HMRATE((1e-10)*9.6,0,0)*CH2;
	CH2_HCNP_CN_CH3P_1   = HMRATE((1e-10)*8.7,0,0)*CH2;
	CH2_HNOP_NO_CH3P_1   = HMRATE((1e-10)*8.6,0,0)*CH2;
	CH2P_S_HCSP_H_1      = HMRATE((1e-9)*1.4,0,0)*CH2P;
	CH2_SP_HCSP_H_1      = HMRATE(1.0e-11,0,0)*CH2;
	NHP_NH_NH2P_N_1      = HMRATE(1.0e-9,0,0)*NHP;
	NHP_O_OHP_N_1        = HMRATE(1.0e-9,0,0)*NHP;
	NH_OP_NOP_H_1        = HMRATE((1e-10)*3.6,0,0)*NH;
	NHP_NH2_NH3P_N_1     = HMRATE((1e-9)*1.5,0,0)*NHP;
	NH_NH2P_NH3P_N_1     = HMRATE((1e-10)*7.3,0,0)*NH;
	NHP_OH_H2OP_N_1      = HMRATE(1.0e-9,0,0)*NHP;
	NH_OHP_NH2P_O_1      = HMRATE((1e-10)*3.6,0,0)*NH;
	NHP_NH3_NH4P_N_1     = HMRATE((1e-10)*6.,0,0)*NHP;
	NH_NH3P_NH4P_N_1     = HMRATE((1e-10)*7.1,0,0)*NH;
	NH_CH5P_CH4_NH2P_1   = HMRATE((1e-10)*7.1,0,0)*NH;
	NHP_H2O_NH3P_O_1     = HMRATE((1e-10)*1.75,0,0)*NHP;
	NHP_H2O_H3OP_N_1     = HMRATE((1e-9)*1.05,0,0)*NHP;
	NHP_H2O_HNOP_H2_1    = HMRATE(3.5e-10,0,0)*NHP;
	NH_H2OP_H3OP_N_1     = HMRATE((1e-10)*7.1,0,0)*NH;
	NHP_H2O_OH_NH2P_1    = HMRATE((1e-10)*8.75,0,0)*NHP;
	NHP_CN_HCNP_N_1      = HMRATE((1e-9)*1.6,0,0)*NHP;
	NH_HCNP_CN_NH2P_1    = HMRATE((1e-10)*6.5,0,0)*NH;
	NHP_CO_HCOP_N_1      = HMRATE((1e-10)*4.41,0,0)*NHP;
	NH_COP_HCOP_N_1      = HMRATE((1e-10)*3.2,0,0)*NH;
	NHP_CO_OCNP_H_1      = HMRATE((1e-10)*5.39,0,0)*NHP;
	NH_SIP_SINP_H_1      = HMRATE(1.0e-9,0,0)*NH;
	NH_HCOP_CO_NH2P_1    = HMRATE((1e-10)*6.4,0,0)*NH;
	NH_HNOP_NO_NH2P_1    = HMRATE((1e-10)*6.3,0,0)*NH;
	NH_O2P_HNOP_O_1      = HMRATE((1e-10)*3.2,0,0)*NH;
	NHP_O2_NOP_OH_1      = HMRATE((1e-10)*2.05,0,0)*NHP;
	NH_O2P_NO2P_H_1      = HMRATE((1e-10)*3.2,0,0)*NH;
	NHP_S_HSP_N_1        = HMRATE((1e-10)*6.9,0,0)*NHP;
	NHP_S_NSP_H_1        = HMRATE((1e-10)*6.9,0,0)*NHP;
	NH_SP_NSP_H_1        = HMRATE((1e-10)*6.3,0,0)*NH;
	CH3P_NH3_NH4P_CH2_1  = HMRATE((1e-10)*3.4,0,0)*CH3P;
	CH3P_S_HCSP_H2_1     = HMRATE((1e-9)*1.4,0,0)*CH3P;
	O_NH2P_HNOP_H_1      = HMRATE((1e-11)*7.2,0,0)*ATO;
	O_NH3P_HNOP_H2_1     = HMRATE(1.0e-11,0,0)*ATO;
	OP_CN_NOP_C_1        = HMRATE(1.0e-9,0,0)*OP;
	OP_HCN_COP_NH_1      = HMRATE(1.2e-9,0,0)*OP;
	OP_HCN_NOP_CH_1      = HMRATE(1.2e-9,0,0)*OP;
	OP_HCN_HCOP_N_1      = HMRATE(1.2e-9,0,0)*OP;
	OP_N2_NOP_N_1        = HMRATE((1e-12)*1.2,0,0)*OP;
	O_N2P_NOP_N_1        = HMRATE((1e-10)*1.3,0,0)*ATO;
	O_HNOP_NO2P_H_1      = HMRATE((1e-12)*1.,0,0)*ATO;
	O_HSP_SP_OH_1        = HMRATE((1e-10)*2.9,0,0)*ATO;
	O_HSP_SOP_H_1        = HMRATE((1e-10)*2.9,0,0)*ATO;
	O_SINP_SIOP_N_1      = HMRATE(1.0e-9,0,0)*ATO;
	OP_N2O_NOP_NO_1      = HMRATE((1e-10)*6.3,0,0)*OP;
	O_CSP_S_COP_1        = HMRATE((1e-11)*6.,0,0)*ATO;
	O_HCSP_S_HCOP_1      = HMRATE((1e-12)*5.,0,0)*ATO;
	O_HCSP_OCSP_H_1      = HMRATE((1e-12)*5.,0,0)*ATO;
	OP_NO2_O2_NOP_1      = HMRATE((1e-10)*8.3,0,0)*OP;
	O_NSP_S_NOP_1        = HMRATE((1e-10)*6.1,0,0)*ATO;
	NH2P_NH2_NH3P_NH_1   = HMRATE(1.0e-9,0,0)*NH2P;
	NH2_OHP_NH3P_O_1     = HMRATE(5.0e-10,0,0)*NH2;
	NH2P_OH_H2OP_NH_1    = HMRATE((1e-10)*7.1,0,0)*NH2P;
	NH2P_NH3_NH4P_NH_1   = HMRATE((1e-9)*1.61,0,0)*NH2P;
	NH2_NH3P_NH4P_NH_1   = HMRATE(1.0e-11,0,0)*NH2;
	NH2_CH5P_NH3P_CH4_1  = HMRATE((1e-10)*9.9,0,0)*NH2;
	NH2P_H2O_NH3P_OH_1   = HMRATE(1.0e-10,0,0)*NH2P;
	NH2P_H2O_H3OP_NH_1   = HMRATE((1e-9)*2.76,0,0)*NH2P;
	NH2_H2OP_NH3P_OH_1   = HMRATE((1e-10)*4.9,0,0)*NH2;
	NH2P_H2O_NH4P_O_1    = HMRATE((1e-10)*1.45,0,0)*NH2P;
	NH2_H3OP_H2O_NH3P_1  = HMRATE((1e-10)*9.7,0,0)*NH2;
	NH2_HCNP_CN_NH3P_1   = HMRATE(9.0e-10,0,0)*NH2;
	NH2_COP_HCOP_NH_1    = HMRATE((1e-10)*4.5,0,0)*NH2;
	NH2_HCOP_CO_NH3P_1   = HMRATE((1e-10)*8.9,0,0)*NH2;
	NH2_HNOP_NO_NH3P_1   = HMRATE((1e-10)*8.8,0,0)*NH2;
	NH2P_O2_HNOP_OH_1    = HMRATE((1e-11)*2.1,0,0)*NH2P;
	NH2P_S_HSP_NH_1      = HMRATE((1e-10)*4.4,0,0)*NH2P;
	CH4P_NH3_NH4P_CH3_1  = HMRATE((1e-9)*1.15,0,0)*CH4P;
	CH4_NH3P_NH4P_CH3_1  = HMRATE((1e-10)*4.8,0,0)*CH4;
	CH4_N2P_N2_CH2P_H2_1 = HMRATE((1e-11)*7.,0,0)*CH4;
	CH4_N2P_N2_CH3P_H_1  = HMRATE((1e-10)*9.3,0,0)*CH4;
	CH4_HNOP_NO_CH5P_1   = HMRATE(1.0e-10,0,0)*CH4;
	CH4_SP_HCSP_H2_H_1   = HMRATE(2.0e-11,0,0)*CH4;
	CH4_CSP_HCSP_CH3_1   = HMRATE(5.0e-10,0,0)*CH4;
	OHP_NH3_NH4P_O_1     = HMRATE(1.2e-9,0,0)*OHP;
	OH_NH3P_NH4P_O_1     = HMRATE(7.0e-10,0,0)*OH;
	OHP_CN_HCNP_O_1      = HMRATE(1.0e-9,0,0)*OHP;
	OH_HCNP_CN_H2OP_1    = HMRATE((1e-10)*6.3,0,0)*OH;
	OHP_NO_HNOP_O_1      = HMRATE((1e-10)*6.11,0,0)*OHP;
	OH_HNOP_NO_H2OP_1    = HMRATE(6.2e-10,0,0)*OH;
	OHP_S_HSP_O_1        = HMRATE((1e-10)*4.3,0,0)*OHP;
	OHP_S_SOP_H_1        = HMRATE((1e-10)*4.3,0,0)*OHP;
	OH_SP_SOP_H_1        = HMRATE((1e-10)*6.1,0,0)*OH;
	NH3P_NH3_NH4P_NH2_1  = HMRATE((1e-9)*2.2,0,0)*NH3P;
	NH3_CH5P_NH4P_CH4_1  = HMRATE((1e-9)*2.5,0,0)*NH3;
	NH3P_H2O_NH4P_OH_1   = HMRATE(1.1e-10,0,0)*NH3P;
	NH3_H2OP_NH4P_OH_1   = HMRATE((1e-10)*9.45,0,0)*NH3;
	NH3_H3OP_NH4P_H2O_1  = HMRATE((1e-9)*2.2,0,0)*NH3;
	NH3_COP_HCOP_NH2_1   = HMRATE((1e-11)*4.12,0,0)*NH3;
	NH3_HCOP_CO_NH4P_1   = HMRATE((1e-9)*2.2,0,0)*NH3;
	NH3_HNOP_NO_NH4P_1   = HMRATE(1.1e-9,0,0)*NH3;
	NH3_HSP_S_NH4P_1     = HMRATE((1e-10)*9.75,0,0)*NH3;
	NH3_HCSP_CS_NH4P_1   = HMRATE(2.0e-9,0,0)*NH3;
	CH5P_S_HSP_CH4_1     = HMRATE(1.3e-9,0,0)*CH5P;
	H2O_CNP_HCNP_OH_1    = HMRATE((1e-9)*1.6,0,0)*H2O;
	H2O_CNP_HCOP_NH_1    = HMRATE((1e-10)*1.6,0,0)*H2O;
	H2O_HCNP_CN_H3OP_1   = HMRATE((1e-9)*1.8,0,0)*H2O;
	H2O_HNOP_NO_H3OP_1   = HMRATE((1e-9)*2.3,0,0)*H2O;
	H2OP_S_HSP_OH_1      = HMRATE((1e-10)*4.3,0,0)*H2OP;
	H2O_HSP_S_H3OP_1     = HMRATE((1e-10)*7.8,0,0)*H2O;
	H3OP_CS_HCSP_H2O_1   = HMRATE(1.0e-9,0,0)*H3OP;
	CNP_NO_OCNP_N_1      = HMRATE((1e-10)*1.9,0,0)*CNP;
	CN_HNOP_NO_HCNP_1    = HMRATE((1e-10)*8.7,0,0)*CN;
	CNP_O2_NOP_CO_1      = HMRATE((1e-11)*8.6,0,0)*CNP;
	CNP_O2_OCNP_O_1      = HMRATE((1e-11)*8.6,0,0)*CNP;
	HCNP_CO_HCOP_CN_1    = HMRATE((1e-10)*1.4,0,0)*HCNP;
	HCNP_S_HSP_CN_1      = HMRATE((1e-10)*5.7,0,0)*HCNP;
	CO_HNOP_NO_HCOP_1    = HMRATE(1.0e-10,0,0)*CO;
	N2P_OCS_SP_N2_CO_1   = HMRATE((1e-9)*1.04,0,0)*N2P;
	HCOP_S_HSP_CO_1      = HMRATE((1e-10)*3.3,0,0)*HCOP;
	HCOP_CS_HCSP_CO_1    = HMRATE(1.2e-9,0,0)*HCOP;
	HNOP_S_HSP_NO_1      = HMRATE(1.1e-9,0,0)*HNOP;
	O2_SP_SOP_O_1        = HMRATE((1e-11)*1.5,0,0)*O2;
	O2P_S_SOP_O_1        = HMRATE(5.4e-10,0,0)*O2P;
	O2_CSP_OCSP_O_1      = HMRATE((1e-10)*1.3,0,0)*O2;
	S_SIOP_SO_SIP_1      = HMRATE(1.0e-9,0,0)*ATS;
	CP_NH3_NH3P_C_1      = HMRATE((1e-10)*5.06,0,0)*CP;
	C_CNP_CN_CP_1        = HMRATE(1.1e-10,0,0)*ATC;
	C_N2P_N2_CP_1        = HMRATE(1.1e-10,0,0)*ATC;
	CP_NO_NOP_C_1        = HMRATE(5.2e-10,0,0)*CP;
	CP_SIN_SINP_C_1      = HMRATE(1.0e-9,0,0)*CP;
	C_CSP_CS_CP_1        = HMRATE((1e-9)*1.6,0,0)*ATC;
	CP_NS_NSP_C_1        = HMRATE((1e-10)*7.6,0,0)*CP;
	CP_SO_SOP_C_1        = HMRATE((1e-10)*2.6,0,0)*CP;
	CP_OCS_OCSP_C_1      = HMRATE(4.0e-10,0,0)*CP;
	CH_NH2P_NH2_CHP_1    = HMRATE(3.5e-10,0,0)*CH;
	CHP_NH3_NH3P_CH_1    = HMRATE((1e-10)*4.59,0,0)*CHP;
	CH_CNP_CN_CHP_1      = HMRATE((1e-10)*6.4,0,0)*CH;
	CH_N2P_N2_CHP_1      = HMRATE((1e-10)*6.3,0,0)*CH;
	CHP_NO_NOP_CH_1      = HMRATE((1e-10)*7.6,0,0)*CHP;
	NP_NH_NHP_N_1        = HMRATE((1e-10)*3.7,0,0)*NP;
	NP_NH2_NH2P_N_1      = HMRATE(1.0e-9,0,0)*NP;
	NP_NH3_NH3P_N_1      = HMRATE((1e-9)*1.97,0,0)*NP;
	NP_CN_CNP_N_1        = HMRATE(1.1e-9,0,0)*NP;
	NP_HCN_HCNP_N_1      = HMRATE(1.2e-9,0,0)*NP;
	N_N2P_N2_NP_1        = HMRATE(1.0e-11,0,0)*ATN;
	NP_NO_NOP_N_1        = HMRATE((1e-10)*4.51,0,0)*NP;
	NP_OCS_OCSP_N_1      = HMRATE((1e-9)*1.02,0,0)*NP;
	CH2_NH2P_NH2_CH2P_1  = HMRATE((1e-10)*4.9,0,0)*CH2;
	CH2_CNP_CN_CH2P_1    = HMRATE((1e-10)*8.8,0,0)*CH2;
	CH2_N2P_N2_CH2P_1    = HMRATE((1e-10)*8.7,0,0)*CH2;
	CH2P_NO_NOP_CH2_1    = HMRATE((1e-10)*4.2,0,0)*CH2P;
	NH_OP_O_NHP_1        = HMRATE((1e-10)*3.6,0,0)*NH;
	NH_OHP_OH_NHP_1      = HMRATE((1e-10)*3.6,0,0)*NH;
	NHP_NH3_NH3P_NH_1    = HMRATE((1e-9)*1.8,0,0)*NHP;
	NHP_H2O_H2OP_NH_1    = HMRATE((1e-9)*1.05,0,0)*NHP;
	NH_CNP_CN_NHP_1      = HMRATE((1e-10)*6.5,0,0)*NH;
	NH_COP_CO_NHP_1      = HMRATE((1e-10)*3.2,0,0)*NH;
	NH_N2P_N2_NHP_1      = HMRATE((1e-10)*6.5,0,0)*NH;
	NHP_NO_NOP_NH_1      = HMRATE((1e-10)*7.12,0,0)*NHP;
	NHP_O2_O2P_NH_1      = HMRATE((1e-10)*4.51,0,0)*NHP;
	NHP_S_SP_NH_1        = HMRATE((1e-10)*6.9,0,0)*NHP;
	CH3P_NO_NOP_CH3_1    = HMRATE(1.0e-9,0,0)*CH3P;
	OP_NH2_NH2P_O_1      = HMRATE(1.0e-9,0,0)*OP;
	OP_NH3_NH3P_O_1      = HMRATE(1.2e-9,0,0)*OP;
	O_CNP_CN_OP_1        = HMRATE((1e-11)*6.5,0,0)*ATO;
	O_HCNP_HCN_OP_1      = HMRATE((1e-11)*6.5,0,0)*ATO;
	O_N2P_N2_OP_1        = HMRATE(1.0e-11,0,0)*ATO;
	OP_NO_NOP_O_1        = HMRATE((1e-12)*1.7,0,0)*OP;
	OP_OCS_OCSP_O_1      = HMRATE((1e-10)*6.5,0,0)*OP;
	NH2_OHP_OH_NH2P_1    = HMRATE(5.0e-10,0,0)*NH2;
	NH2P_NH3_NH3P_NH2_1  = HMRATE((1e-10)*6.9,0,0)*NH2P;
	NH2_H2OP_H2O_NH2P_1  = HMRATE((1e-10)*4.9,0,0)*NH2;
	NH2_CNP_CN_NH2P_1    = HMRATE((1e-10)*9.1,0,0)*NH2;
	NH2_COP_CO_NH2P_1    = HMRATE((1e-10)*4.5,0,0)*NH2;
	NH2_N2P_N2_NH2P_1    = HMRATE((1e-10)*8.9,0,0)*NH2;
	NH2P_NO_NOP_NH2_1    = HMRATE(7.0e-10,0,0)*NH2P;
	NH2_O2P_O2_NH2P_1    = HMRATE((1e-10)*8.7,0,0)*NH2;
	NH2P_S_SP_NH2_1      = HMRATE((1e-10)*4.4,0,0)*NH2P;
	CH4P_NH3_NH3P_CH4_1  = HMRATE((1e-9)*1.65,0,0)*CH4P;
	CH4P_OCS_OCSP_CH4_1  = HMRATE((1e-10)*4.2,0,0)*CH4P;
	OHP_NH3_NH3P_OH_1    = HMRATE(1.2e-9,0,0)*OHP;
	OH_CNP_CN_OHP_1      = HMRATE((1e-10)*6.4,0,0)*OH;
	OH_N2P_N2_OHP_1      = HMRATE((1e-10)*6.3,0,0)*OH;
	OHP_NO_NOP_OH_1      = HMRATE((1e-10)*3.59,0,0)*OHP;
	NH3_H2OP_H2O_NH3P_1  = HMRATE((1e-9)*2.21,0,0)*NH3;
	NH3_HCNP_HCN_NH3P_1  = HMRATE((1e-9)*1.68,0,0)*NH3;
	NH3_COP_CO_NH3P_1    = HMRATE((1e-9)*2.02,0,0)*NH3;
	NH3_N2P_N2_NH3P_1    = HMRATE((1e-9)*1.9,0,0)*NH3;
	NH3P_SI_SIP_NH3_1    = HMRATE((1e-9)*1.9,0,0)*NH3P;
	NH3_O2P_O2_NH3P_1    = HMRATE(2.0e-9,0,0)*NH3;
	NH3_SP_S_NH3P_1      = HMRATE((1e-9)*1.44,0,0)*NH3;
	NH3_HSP_HS_NH3P_1    = HMRATE((1e-10)*5.25,0,0)*NH3;
	NH3_SOP_SO_NH3P_1    = HMRATE(1.3e-9,0,0)*NH3;
	H2O_HCNP_HCN_H2OP_1  = HMRATE((1e-9)*1.8,0,0)*H2O;
	H2O_N2P_N2_H2OP_1    = HMRATE((1e-9)*2.3,0,0)*H2O;
	H2OP_NO_NOP_H2O_1    = HMRATE((1e-10)*2.7,0,0)*H2OP;
	CNP_HCN_HCNP_CN_1    = HMRATE((1e-9)*1.79,0,0)*CNP;
	CNP_CO_COP_CN_1      = HMRATE((1e-10)*6.3,0,0)*CNP;
	CN_N2P_N2_CNP_1      = HMRATE(1.0e-10,0,0)*CN;
	CNP_NO_NOP_CN_1      = HMRATE((1e-10)*5.7,0,0)*CNP;
	CNP_O2_O2P_CN_1      = HMRATE((1e-10)*2.58,0,0)*CNP;
	CNP_S_SP_CN_1        = HMRATE(1.1e-9,0,0)*CNP;
	HCN_COP_CO_HCNP_1    = HMRATE((1e-9)*3.4,0,0)*HCN;
	HCN_N2P_N2_HCNP_1    = HMRATE((1e-10)*3.9,0,0)*HCN;
	HCNP_NO_NOP_HCN_1    = HMRATE((1e-10)*8.1,0,0)*HCNP;
	HCNP_O2_O2P_HCN_1    = HMRATE((1e-10)*3.2,0,0)*HCNP;
	HCNP_S_SP_HCN_1      = HMRATE((1e-10)*5.7,0,0)*HCNP;
	CO_N2P_N2_COP_1      = HMRATE((1e-11)*7.4,0,0)*CO;
	COP_NO_NOP_CO_1      = HMRATE((1e-10)*3.3,0,0)*COP;
	N2P_NO_NOP_N2_1      = HMRATE((1e-10)*4.4,0,0)*N2P;
	N2P_O2_O2P_N2_1      = HMRATE((1e-11)*5.,0,0)*N2P;
	N2P_S_SP_N2_1        = HMRATE(1.1e-9,0,0)*N2P;
	SI_NOP_NO_SIP_1      = HMRATE((1e-9)*1.6,0,0)*ATSI;
	SI_HSP_HS_SIP_1      = HMRATE((1e-9)*1.4,0,0)*ATSI;
	SI_CSP_CS_SIP_1      = HMRATE((1e-10)*1.5,0,0)*ATSI;
	NO_HNOP_HNO_NOP_1    = HMRATE(7.0e-10,0,0)*NO;
	NO_O2P_O2_NOP_1      = HMRATE((1e-10)*4.5,0,0)*NO;
	NO_SP_S_NOP_1        = HMRATE((1e-10)*3.7,0,0)*NO;
	NO_HSP_HS_NOP_1      = HMRATE((1e-10)*4.5,0,0)*NO;
	NO_SIOP_SIO_NOP_1    = HMRATE((1e-10)*7.2,0,0)*NO;
	NO_S2P_S2_NOP_1      = HMRATE((1e-10)*5.1,0,0)*NO;
	O2P_NO2_NO2P_O2_1    = HMRATE((1e-10)*6.6,0,0)*O2P;
	S_HSP_HS_SP_1        = HMRATE((1e-10)*9.7,0,0)*ATS;
	C_N_CN_NU_1          = HMRATE((1e-17)*1.,0,0)*ATC;
	C_S_CS_NU_1          = HMRATE(4.36E-19,0.22,0)*ATC;
	CP_S_CSP_NU_1        = HMRATE(3.07E-19,0.15,0)*CP;
	NP_N_N2P_NU_1        = HMRATE(3.71E-18,0.24,26.1)*NP;
	CO_S_OCS_NU_1        = HMRATE((1e-17)*1.6,-1.5,0)*CO;
	CH_NP_N_CHP_1        = HMRATE((1e-10)*3.6,0,0)*CH;
	CHP_S_SP_CH_1        = HMRATE((1e-10)*4.7,0,0)*CHP;
	NP_CH2_CH2P_N_1      = HMRATE(1.0e-9,0,0)*NP;
	NP_CH4_CH4P_N_1      = HMRATE((1e-11)*2.8,0,0)*NP;
	NP_OH_OHP_N_1        = HMRATE((1e-10)*3.7,0,0)*NP;
	NP_H2O_H2OP_N_1      = HMRATE((1e-9)*2.8,0,0)*NP;
	NP_CO_COP_N_1        = HMRATE((1e-10)*8.25,0,0)*NP;
	NP_O2_O2P_N_1        = HMRATE((1e-10)*3.11,0,0)*NP;
	OP_O2_O2P_O_1        = HMRATE((1e-11)*1.9,0,0)*OP;
	OHP_S_SP_OH_1        = HMRATE((1e-10)*4.3,0,0)*OHP;
	H2OP_S_SP_H2O_1      = HMRATE((1e-10)*4.3,0,0)*H2OP;
	COP_S_SP_CO_1        = HMRATE(1.1e-9,0,0)*COP;
	SI_SP_S_SIP_1        = HMRATE((1e-9)*1.6,0,0)*ATSI;
	O2P_S_SP_O2_1        = HMRATE(5.4e-10,0,0)*O2P;
	C_NH_N_CH_2        = HMRATE((1e-11)*1.73,0.5,4000)*NH;
	C_NH_CN_H_2          = HMRATE(1.1e-10,0.5,0)*NH;
	C_N2_CN_N_2        = HMRATE((1e-11)*8.69,0,22600)*N2;
	C_NO_CN_O_2          = HMRATE((1e-11)*4.8,0,0)*NO;
	C_NO_CO_N_2        = HMRATE((1e-11)*4.65,0,0)*NO;
	C_HS_S_CH_2          = HMRATE((1e-11)*1.2,0.58,5880)*HS;
	C_HS_CS_H_2          = HMRATE(2.0e-11,0,0)*HS;
	C_OCN_CO_CN_2        = HMRATE(4.0e-11,0.5,0)*OCN;
	C_NS_S_CN_2          = HMRATE(2.0e-11,0.5,0)*NS;
	C_NS_CS_N_2        = HMRATE((1e-11)*1.73,0.5,4000)*NS;
	C_SO_S_CO_2          = HMRATE((1e-11)*7.2,0,0)*SO;
	C_S2_CS_S_2          = HMRATE((1e-11)*1.73,0.5,0)*S2;
	CH_N_NH_C_2        = HMRATE((1e-11)*3.03,0.65,1207)*ATN;
	CH_N_CN_H_2        = HMRATE((1e-10)*1.66,-0.09,0)*ATN;
	CH_N2_HCN_N_2      = HMRATE(0.00000000000056,0.88,10128)*N2;
	CH_NO_HCN_O_2        = HMRATE((1e-9)*5.59,0,10814)*NO;
	CH_NO_CN_OH_2        = HMRATE(2.32E-26,0,0)*NO;
	CH_NO_OCN_H_2        = HMRATE(1.13E-25,0,0)*NO;
	CH_HNO_NO_CH2_2      = HMRATE((1e-11)*1.73,0.5,0)*HNO;
	CH_S_HS_C_2          = HMRATE((1e-11)*1.73,0.5,4000)*ATS;
	CH_S_CS_H_2          = HMRATE((1e-12)*1.1,0,0)*ATS;
	N_NH_N2_H_2        = HMRATE((1e-11)*4.98,0,0)*NH;
	N_CH3_HCN_H_H_2    = HMRATE((1e-13)*3.32,0,0)*CH3;
	N_CH3_HCN_H2_2     = HMRATE((1e-11)*1.3,0.5,0)*CH3;
	N_OH_O_NH_2          = HMRATE((1e-11)*1.88,0.1,10700)*OH;
	N_OH_NO_H_2          = HMRATE((1e-11)*5.32,-0.25,0)*OH;
	N_CN_N2_C_2          = HMRATE((1e-10)*3.,0,0)*CN;
	N_SIH_SIN_H_2        = HMRATE((1e-11)*5.,0.5,0)*SIH;
	N_NO_N2_O_2          = HMRATE((1e-11)*3.75,0,26)*NO;
	N_HNO_NO_NH_2        = HMRATE((1e-12)*2.94,0.5,1000)*HNO;
	N_HNO_N2O_H_2        = HMRATE((1e-12)*1.43,0.5,1500)*HNO;
	N_O2_NO_O_2          = HMRATE((1e-12)*2.26,0.86,3134)*O2;
	N_HS_NS_H_2          = HMRATE((1e-11)*1.73,0.5,0)*HS;
	N_HS_S_NH_2          = HMRATE((1e-11)*1.73,0.5,9060)*HS;
	N_CS_S_CN_2          = HMRATE((1e-11)*3.8,0.5,1160)*CS;
	N_NS_S_N2_2          = HMRATE((1e-11)*1.73,0.5,0)*NS;
	N_SO_NS_O_2          = HMRATE((1e-11)*4.68,0.5,8254)*SO;
	N_SO_S_NO_2          = HMRATE((1e-11)*1.73,0.5,750)*SO;
	N_S2_NS_S_2          = HMRATE((1e-11)*1.73,0.5,4000)*S2;
	CH2_CN_HCN_CH_2      = HMRATE((1e-12)*5.3,0,2500)*CN;
	CH2_NO_HCN_OH_2      = HMRATE(0.000000000000832,0,1443)*NO;
	CH2_HNO_NO_CH3_2     = HMRATE((1e-11)*1.73,0.5,0)*HNO;
	NH_NH_N2_H_H_2      = HMRATE((1e-9)*1.16,0,0)*NH;
	NH_NH_N2_H2_2        = HMRATE((1e-11)*1.7,0,0)*NH;
	NH_O_OH_N_2          = HMRATE((1e-11)*1.16,0,0)*ATO;
	NH_O_NO_H_2          = HMRATE((1e-10)*1.16,0,0)*ATO;
	NH_OH_HNO_H_2        = HMRATE((1e-11)*3.32,0,0)*OH;
	NH_OH_NH2_O_2        = HMRATE((1e-12)*2.93,0.1,5800)*OH;
	NH_OH_H2O_N_2        = HMRATE((1e-12)*3.11,1.2,0)*OH;
	NH_H2O_OH_NH2_2      = HMRATE((1e-12)*1.83,1.6,14090)*H2O;
	NH_CN_HCN_N_2        = HMRATE((1e-12)*2.94,0.5,1000)*CN;
	NH_NO_N2O_H_2        = HMRATE((1e-10)*1.16,-1.03,420)*NO;
	NH_NO_N2_O_H_2      = HMRATE((1e-11)*5.,0,0)*NO;
	NH_NO_N2_OH_2        = HMRATE((1e-11)*1.46,-0.58,37)*NO;
	NH_S_HS_N_2          = HMRATE((1e-11)*1.73,0.5,4000)*ATS;
	NH_S_NS_H_2          = HMRATE((1e-11)*1.73,0.5,0)*ATS;
	NH_NO2_HNO_NO_2      = HMRATE((1e-12)*5.72,0.5,2500)*NO2;
	NH_NO2_N2O_OH_2      = HMRATE((1e-13)*1.44,0, 1140)*NO2;
	CH3_NH3_CH4_NH2_2    = HMRATE(0.0000000000000955,0,4890)*NH3;
	CH3_CN_HCN_CH2_2     = HMRATE((1e-12)*9.21,0.7,1500)*CN;
	CH3_HNO_NO_CH4_2     = HMRATE((1e-11)*1.44,0.5,0)*HNO;
	O_NH2_OH_NH_2        = HMRATE((1e-11)*1.39,0,40)*NH2;
	O_NH2_NO_H2_2        = HMRATE((1e-12)*8.3,0,0)*NH2;
	O_NH3_OH_NH2_2       = HMRATE((1e-11)*1.89,0,4003)*NH3;
	O_CN_NO_C_2          = HMRATE((1e-11)*3.81,0.5,14545)*CN;
	O_CN_CO_N_2          = HMRATE((1e-11)*4.36,0.46,364)*CN;
	O_HCN_CN_OH_2        = HMRATE((1e-10)*6.21,0,12439)*HCN;
	O_HCN_CO_NH_2        = HMRATE(0.00000000000073,1.14,3742)*HCN;
	O_HCN_OCN_H_2        = HMRATE((1e-12)*1.36,1.38,3693)*HCN;
	O_N2_NO_N_2          = HMRATE((1e-10)*2.51,0,38602)*N2;
	O_NO_O2_N_2          = HMRATE((1e-11)*1.18,0,20413)*NO;
	O_HNO_NO_OH_2        = HMRATE((1e-11)*6.,0,0)*HNO;
	O_HNO_O2_NH_2        = HMRATE((1e-12)*2.94,0.5,3500)*HNO;
	O_HNO_NO2_H_2        = HMRATE((1e-12)*1.44,0.5,0)*HNO;
	O_HS_S_OH_2          = HMRATE((1e-11)*1.74,0.67,956)*HS;
	O_HS_SO_H_2          = HMRATE((1e-10)*2.32,0,0)*HS;
	O_OCN_NO_CO_2        = HMRATE((1e-11)*9.43,-0.09,100)*OCN;
	O_OCN_O2_CN_2        = HMRATE((1e-10)*4.02,-1.43,3501)*OCN;
	O_SIN_NO_SI_2        = HMRATE((1e-11)*2.5,0.5,0)*SIN;
	O_SIN_SIO_N_2        = HMRATE((1e-11)*5.75,0.1,200)*SIN;
	O_N2O_NO_NO_2        = HMRATE((1e-10)*1.15,0,13400)*N2O;
	O_N2O_O2_N2_2        = HMRATE((1e-10)*1.66,0,14100)*N2O;
	O_CS_S_CO_2          = HMRATE((1e-10)*2.48,-0.65,783)*CS;
	O_CS_SO_C_2          = HMRATE((1e-11)*4.68,0.5,28940)*CS;
	O_NS_S_NO_2          = HMRATE((1e-11)*5.,0.5,0)*NS;
	O_NS_SO_N_2          = HMRATE((1e-11)*1.73,0.5,4000)*NS;
	O_SO_S_O2_2          = HMRATE(0.00000000000066,0,2760)*SO;
	O_OCS_SO_CO_2        = HMRATE((1e-11)*1.6,0,2150)*OCS;
	O_S2_SO_S_2          = HMRATE((1e-11)*1.73,0.5,0)*S2;
	NH2_OH_NH3_O_2       = HMRATE(0.000000000000208,0.76,262)*OH;
	NH2_OH_H2O_NH_2      = HMRATE((1e-12)*1.5,0,0)*OH;
	NH2_NO_N2_OH_H_2    = HMRATE((1e-12)*1.49,0,0)*NO;
	NH2_NO_N2_H2O_2      = HMRATE((1e-11)*4.27,-2.5,331)*NO;
	CH4_CN_HCN_CH3_2     = HMRATE((1e-12)*3.14,1.53,504)*CN;
	OH_NH3_H2O_NH2_2     = HMRATE((1e-13)*1.47,2.05,7)*NH3;
	OH_CN_HCN_O_2        = HMRATE(1.0e-11,0,1000)*CN;
	OH_CN_OCN_H_2        = HMRATE((1e-11)*7.,0,0)*CN;
	OH_HCN_CN_H2O_2      = HMRATE((1e-13)*1.87,1.5,3887)*HCN;
	OH_HCN_CO_NH2_2      = HMRATE((1e-13)*1.07,0,5892)*HCN;
	OH_NO_NO2_H_2        = HMRATE((1e-12)*5.2,0,15100)*NO;
	OH_S_HS_O_2          = HMRATE(6.6e-11,0,0)*ATS;
	OH_S_SO_H_2          = HMRATE(6.6e-11,0,0)*ATS;
	OH_N2O_HNO_NO_2      = HMRATE(1.04E-17,4.33,12623)*N2O;
	OH_CS_OCS_H_2        = HMRATE(0.0000000000000939,1.12,800)*CS;
	NH3_CN_HCN_NH2_2     = HMRATE((1e-11)*2.6,-1.1,0)*CN;
	CN_NO_N2_CO_2        = HMRATE((1e-10)*1.79,0,4040)*NO;
	CN_NO_OCN_N_2        = HMRATE((1e-10)*1.62,0,21205)*NO;
	CN_HNO_NO_HCN_2      = HMRATE((1e-11)*1.5,0.5,0)*HNO;
	CN_O2_NO_CO_2        = HMRATE(0.00000000000053,0,0)*O2;
	CN_S_NS_C_2          = HMRATE((1e-11)*5.71,0.5,32010)*ATS;
	CN_S_CS_N_2          = HMRATE((1e-11)*1.73,0.5,0)*ATS;
	CO_HS_OCS_H_2        = HMRATE(0.0000000000000595,1.12,8330)*HS;
	N2_O2_N2O_O_2        = HMRATE(1.0e-10,0,55200)*O2;
	NO_NO_O2_N2_2        = HMRATE((1e-11)*2.51,0,30653)*NO;
	NO_NO_N2O_O_2        = HMRATE((1e-12)*7.22,0,33155)*NO;
	NO_HNO_N2O_OH_2      = HMRATE((1e-11)*1.41,0,14890)*HNO;
	NO_S_NS_O_2          = HMRATE(2.94e-11,0.5,17465)*ATS;
	NO_S_SO_N_2          = HMRATE((1e-10)*1.75,0,20200)*ATS;
	O2_S_SO_O_2          = HMRATE((1e-12)*2.28,0.52,0)*ATS;
	S_SO_S2_O_2          = HMRATE((1e-11)*1.73,0.5,11500)*SO;
	C_NHP_N_CHP_2        = HMRATE((1e-9)*1.6,0,0)*NHP;
	CP_NH_CNP_H_2        = HMRATE((1e-10)*7.8,0,0)*NH;
	CP_NH2_HCNP_H_2      = HMRATE(1.1e-9,0,0)*NH2;
	C_NH2P_NH_CHP_2      = HMRATE(1.2e-9,0,0)*NH2P;
	CP_NH3_HCNP_H2_2     = HMRATE((1e-11)*7.,0,0)*NH3;
	C_NH3P_NH_CH2P_2     = HMRATE(1.0e-11,0,0)*NH3P;
	C_HCNP_CN_CHP_2      = HMRATE(1.1e-9,0,0)*HCNP;
	C_HNOP_NO_CHP_2      = HMRATE(1.0e-9,0,0)*HNOP;
	CP_HS_CSP_H_2        = HMRATE(1.1e-9,0,0)*HS;
	C_HSP_CSP_H_2        = HMRATE((1e-10)*9.9,0,0)*HSP;
	CP_OCN_COP_CN_2      = HMRATE((1e-9)*3.8,0,0)*OCN;
	CP_NS_CSP_N_2        = HMRATE((1e-10)*7.6,0,0)*NS;
	CP_SO_SP_CO_2        = HMRATE((1e-10)*2.6,0,0)*SO;
	CP_SO_S_COP_2        = HMRATE((1e-10)*2.6,0,0)*SO;
	CP_SO_CSP_O_2        = HMRATE((1e-10)*2.6,0,0)*SO;
	CP_OCS_CSP_CO_2      = HMRATE((1e-9)*1.6,0,0)*OCS;
	CHP_N_CNP_H_2        = HMRATE((1e-10)*1.9,0,0)*ATN;
	CH_NP_N_CHP_2        = HMRATE((1e-10)*3.6,0,0)*NP;
	CH_NP_CNP_H_2        = HMRATE((1e-10)*3.6,0,0)*NP;
	CH_NHP_CH2P_N_2      = HMRATE((1e-10)*9.9,0,0)*NHP;
	CHP_NH_CNP_H2_2      = HMRATE((1e-10)*7.6,0,0)*NH;
	CHP_NH2_HCNP_H2_2    = HMRATE(1.1e-9,0,0)*NH2;
	CH_NH2P_NH_CH2P_2    = HMRATE(3.5e-10,0,0)*NH2P;
	CHP_NH3_NH4P_C_2     = HMRATE((1e-10)*4.05,0,0)*NH3;
	CH_NH3P_NH4P_C_2     = HMRATE((1e-10)*6.9,0,0)*NH3P;
	CH_HCNP_CN_CH2P_2    = HMRATE((1e-10)*6.3,0,0)*HCNP;
	CH_HNOP_NO_CH2P_2    = HMRATE(6.2e-10,0,0)*HNOP;
	CHP_S_HSP_C_2        = HMRATE((1e-10)*4.7,0,0)*ATS;
	CHP_S_CSP_H_2        = HMRATE((1e-10)*4.7,0,0)*ATS;
	CH_SP_CSP_H_2        = HMRATE(6.2e-10,0,0)*SP;
	CH_HSP_S_CH2P_2      = HMRATE((1e-10)*5.8,0,0)*HSP;
	CHP_OCS_HCSP_CO_2    = HMRATE((1e-9)*1.05,0,0)*OCS;
	N_CH2P_HCNP_H_2      = HMRATE((1e-10)*2.2,0,0)*CH2P;
	NP_NH_N2P_H_2        = HMRATE((1e-10)*3.7,0,0)*NH;
	N_NHP_N2P_H_2        = HMRATE(1.3e-9,0,0)*NHP;
	NP_CH4_HCNP_H2_H_2   = HMRATE((1e-11)*5.6,0,0)*CH4;
	N_OHP_NOP_H_2        = HMRATE((1e-10)*8.9,0,0)*OHP;
	NP_NH3_NH2P_NH_2     = HMRATE((1e-10)*2.16,0,0)*NH3;
	N_H2OP_HNOP_H_2      = HMRATE((1e-10)*1.9,0,0)*H2OP;
	NP_CO_NOP_C_2        = HMRATE((1e-10)*1.45,0,0)*CO;
	NP_NO_N2P_O_2        = HMRATE((1e-11)*7.9,0,0)*NO;
	NP_O2_NO_OP_2        = HMRATE((1e-11)*3.66,0,0)*O2;
	NP_O2_NOP_O_2        = HMRATE((1e-10)*2.63,0,0)*O2;
	N_O2P_NOP_O_2        = HMRATE((1e-10)*1.8,0,0)*O2P;
	N_HSP_NSP_H_2        = HMRATE((1e-10)*7.4,0,0)*HSP;
	N_SIOP_NOP_SI_2      = HMRATE((1e-11)*9.,0,0)*SIOP;
	N_SIOP_NO_SIP_2      = HMRATE((1e-10)*2.1,0,0)*SIOP;
	N_SOP_NSP_O_2        = HMRATE((1e-11)*5.,0,0)*SOP;
	NP_OCS_SP_CO_N_2     = HMRATE((1e-10)*3.08,0,0)*OCS;
	NP_OCS_CSP_NO_2      = HMRATE((1e-11)*7.,0,0)*OCS;
	CH2_NHP_CH3P_N_2     = HMRATE((1e-9)*1.4,0,0)*NHP;
	CH2_NH2P_CH3P_NH_2   = HMRATE((1e-10)*4.9,0,0)*NH2P;
	CH2P_NH3_NH4P_CH_2   = HMRATE((1e-9)*1.26,0,0)*NH3;
	CH2_NH3P_NH2_CH3P_2  = HMRATE((1e-10)*9.6,0,0)*NH3P;
	CH2_HCNP_CN_CH3P_2   = HMRATE((1e-10)*8.7,0,0)*HCNP;
	CH2_HNOP_NO_CH3P_2   = HMRATE((1e-10)*8.6,0,0)*HNOP;
	CH2P_S_HCSP_H_2      = HMRATE((1e-9)*1.4,0,0)*ATS;
	CH2_SP_HCSP_H_2      = HMRATE(1.0e-11,0,0)*SP;
	NHP_NH_NH2P_N_2      = HMRATE(1.0e-9,0,0)*NH;
	NHP_O_OHP_N_2        = HMRATE(1.0e-9,0,0)*ATO;
	NH_OP_NOP_H_2        = HMRATE((1e-10)*3.6,0,0)*OP;
	NHP_NH2_NH3P_N_2     = HMRATE((1e-9)*1.5,0,0)*NH2;
	NH_NH2P_NH3P_N_2     = HMRATE((1e-10)*7.3,0,0)*NH2P;
	NHP_OH_H2OP_N_2      = HMRATE(1.0e-9,0,0)*OH;
	NH_OHP_NH2P_O_2      = HMRATE((1e-10)*3.6,0,0)*OHP;
	NHP_NH3_NH4P_N_2     = HMRATE((1e-10)*6.,0,0)*NH3;
	NH_NH3P_NH4P_N_2     = HMRATE((1e-10)*7.1,0,0)*NH3P;
	NH_CH5P_CH4_NH2P_2   = HMRATE((1e-10)*7.1,0,0)*CH5P;
	NHP_H2O_NH3P_O_2     = HMRATE((1e-10)*1.75,0,0)*H2O;
	NHP_H2O_H3OP_N_2     = HMRATE((1e-9)*1.05,0,0)*H2O;
	NHP_H2O_HNOP_H2_2    = HMRATE(3.5e-10,0,0)*H2O;
	NH_H2OP_H3OP_N_2     = HMRATE((1e-10)*7.1,0,0)*H2OP;
	NHP_H2O_OH_NH2P_2    = HMRATE((1e-10)*8.75,0,0)*H2O;
	NHP_CN_HCNP_N_2      = HMRATE((1e-9)*1.6,0,0)*CN;
	NH_HCNP_CN_NH2P_2    = HMRATE((1e-10)*6.5,0,0)*HCNP;
	NHP_CO_HCOP_N_2      = HMRATE((1e-10)*4.41,0,0)*CO;
	NH_COP_HCOP_N_2      = HMRATE((1e-10)*3.2,0,0)*COP;
	NHP_CO_OCNP_H_2      = HMRATE((1e-10)*5.39,0,0)*CO;
	NH_SIP_SINP_H_2      = HMRATE(1.0e-9,0,0)*SIP;
	NH_HCOP_CO_NH2P_2    = HMRATE((1e-10)*6.4,0,0)*HCOP;
	NH_HNOP_NO_NH2P_2    = HMRATE((1e-10)*6.3,0,0)*HNOP;
	NH_O2P_HNOP_O_2      = HMRATE((1e-10)*3.2,0,0)*O2P;
	NHP_O2_NOP_OH_2      = HMRATE((1e-10)*2.05,0,0)*O2;
	NH_O2P_NO2P_H_2      = HMRATE((1e-10)*3.2,0,0)*O2P;
	NHP_S_HSP_N_2        = HMRATE((1e-10)*6.9,0,0)*ATS;
	NHP_S_NSP_H_2        = HMRATE((1e-10)*6.9,0,0)*ATS;
	NH_SP_NSP_H_2        = HMRATE((1e-10)*6.3,0,0)*SP;
	CH3P_NH3_NH4P_CH2_2  = HMRATE((1e-10)*3.4,0,0)*NH3;
	CH3P_S_HCSP_H2_2     = HMRATE((1e-9)*1.4,0,0)*ATS;
	O_NH2P_HNOP_H_2      = HMRATE((1e-11)*7.2,0,0)*NH2P;
	O_NH3P_HNOP_H2_2     = HMRATE(1.0e-11,0,0)*NH3P;
	OP_CN_NOP_C_2        = HMRATE(1.0e-9,0,0)*CN;
	OP_HCN_COP_NH_2      = HMRATE(1.2e-9,0,0)*HCN;
	OP_HCN_NOP_CH_2      = HMRATE(1.2e-9,0,0)*HCN;
	OP_HCN_HCOP_N_2      = HMRATE(1.2e-9,0,0)*HCN;
	OP_N2_NOP_N_2        = HMRATE((1e-12)*1.2,0,0)*N2;
	O_N2P_NOP_N_2        = HMRATE((1e-10)*1.3,0,0)*N2P;
	O_HNOP_NO2P_H_2      = HMRATE((1e-12)*1.,0,0)*HNOP;
	O_HSP_SP_OH_2        = HMRATE((1e-10)*2.9,0,0)*HSP;
	O_HSP_SOP_H_2        = HMRATE((1e-10)*2.9,0,0)*HSP;
	O_SINP_SIOP_N_2      = HMRATE(1.0e-9,0,0)*SINP;
	OP_N2O_NOP_NO_2      = HMRATE((1e-10)*6.3,0,0)*N2O;
	O_CSP_S_COP_2        = HMRATE((1e-11)*6.,0,0)*CSP;
	O_HCSP_S_HCOP_2      = HMRATE((1e-12)*5.,0,0)*HCSP;
	O_HCSP_OCSP_H_2      = HMRATE((1e-12)*5.,0,0)*HCSP;
	OP_NO2_O2_NOP_2      = HMRATE((1e-10)*8.3,0,0)*NO2;
	O_NSP_S_NOP_2        = HMRATE((1e-10)*6.1,0,0)*NSP;
	NH2P_NH2_NH3P_NH_2   = HMRATE(1.0e-9,0,0)*NH2;
	NH2_OHP_NH3P_O_2     = HMRATE(5.0e-10,0,0)*OHP;
	NH2P_OH_H2OP_NH_2    = HMRATE((1e-10)*7.1,0,0)*OH;
	NH2P_NH3_NH4P_NH_2   = HMRATE((1e-9)*1.61,0,0)*NH3;
	NH2_NH3P_NH4P_NH_2   = HMRATE(1.0e-11,0,0)*NH3P;
	NH2_CH5P_NH3P_CH4_2  = HMRATE((1e-10)*9.9,0,0)*CH5P;
	NH2P_H2O_NH3P_OH_2   = HMRATE(1.0e-10,0,0)*H2O;
	NH2P_H2O_H3OP_NH_2   = HMRATE((1e-9)*2.76,0,0)*H2O;
	NH2_H2OP_NH3P_OH_2   = HMRATE((1e-10)*4.9,0,0)*H2OP;
	NH2P_H2O_NH4P_O_2    = HMRATE((1e-10)*1.45,0,0)*H2O;
	NH2_H3OP_H2O_NH3P_2  = HMRATE((1e-10)*9.7,0,0)*H3OP;
	NH2_HCNP_CN_NH3P_2   = HMRATE(9.0e-10,0,0)*HCNP;
	NH2_COP_HCOP_NH_2    = HMRATE((1e-10)*4.5,0,0)*COP;
	NH2_HCOP_CO_NH3P_2   = HMRATE((1e-10)*8.9,0,0)*HCOP;
	NH2_HNOP_NO_NH3P_2   = HMRATE((1e-10)*8.8,0,0)*HNOP;
	NH2P_O2_HNOP_OH_2    = HMRATE((1e-11)*2.1,0,0)*O2;
	NH2P_S_HSP_NH_2      = HMRATE((1e-10)*4.4,0,0)*ATS;
	CH4P_NH3_NH4P_CH3_2  = HMRATE((1e-9)*1.15,0,0)*NH3;
	CH4_NH3P_NH4P_CH3_2  = HMRATE((1e-10)*4.8,0,0)*NH3P;
	CH4_N2P_N2_CH2P_H2_2 = HMRATE((1e-11)*7.,0,0)*N2P;
	CH4_N2P_N2_CH3P_H_2  = HMRATE((1e-10)*9.3,0,0)*N2P;
	CH4_HNOP_NO_CH5P_2   = HMRATE(1.0e-10,0,0)*HNOP;
	CH4_SP_HCSP_H2_H_2   = HMRATE(2.0e-11,0,0)*SP;
	CH4_CSP_HCSP_CH3_2   = HMRATE(5.0e-10,0,0)*CSP;
	OHP_NH3_NH4P_O_2     = HMRATE(1.2e-9,0,0)*NH3;
	OH_NH3P_NH4P_O_2     = HMRATE(7.0e-10,0,0)*NH3P;
	OHP_CN_HCNP_O_2      = HMRATE(1.0e-9,0,0)*CN;
	OH_HCNP_CN_H2OP_2    = HMRATE((1e-10)*6.3,0,0)*HCNP;
	OHP_NO_HNOP_O_2      = HMRATE((1e-10)*6.11,0,0)*NO;
	OH_HNOP_NO_H2OP_2    = HMRATE(6.2e-10,0,0)*HNOP;
	OHP_S_HSP_O_2        = HMRATE((1e-10)*4.3,0,0)*ATS;
	OHP_S_SOP_H_2        = HMRATE((1e-10)*4.3,0,0)*ATS;
	OH_SP_SOP_H_2        = HMRATE((1e-10)*6.1,0,0)*SP;
	NH3P_NH3_NH4P_NH2_2  = HMRATE((1e-9)*2.2,0,0)*NH3;
	NH3_CH5P_NH4P_CH4_2  = HMRATE((1e-9)*2.5,0,0)*CH5P;
	NH3P_H2O_NH4P_OH_2   = HMRATE(1.1e-10,0,0)*H2O;
	NH3_H2OP_NH4P_OH_2   = HMRATE((1e-10)*9.45,0,0)*H2OP;
	NH3_H3OP_NH4P_H2O_2  = HMRATE((1e-9)*2.2,0,0)*H3OP;
	NH3_COP_HCOP_NH2_2   = HMRATE((1e-11)*4.12,0,0)*COP;
	NH3_HCOP_CO_NH4P_2   = HMRATE((1e-9)*2.2,0,0)*HCOP;
	NH3_HNOP_NO_NH4P_2   = HMRATE(1.1e-9,0,0)*HNOP;
	NH3_HSP_S_NH4P_2     = HMRATE((1e-10)*9.75,0,0)*HSP;
	NH3_HCSP_CS_NH4P_2   = HMRATE(2.0e-9,0,0)*HCSP;
	CH5P_S_HSP_CH4_2     = HMRATE(1.3e-9,0,0)*ATS;
	H2O_CNP_HCNP_OH_2    = HMRATE((1e-9)*1.6,0,0)*CNP;
	H2O_CNP_HCOP_NH_2    = HMRATE((1e-10)*1.6,0,0)*CNP;
	H2O_HCNP_CN_H3OP_2   = HMRATE((1e-9)*1.8,0,0)*HCNP;
	H2O_HNOP_NO_H3OP_2   = HMRATE((1e-9)*2.3,0,0)*HNOP;
	H2OP_S_HSP_OH_2      = HMRATE((1e-10)*4.3,0,0)*ATS;
	H2O_HSP_S_H3OP_2     = HMRATE((1e-10)*7.8,0,0)*HSP;
	H3OP_CS_HCSP_H2O_2   = HMRATE(1.0e-9,0,0)*CS;
	CNP_NO_OCNP_N_2      = HMRATE((1e-10)*1.9,0,0)*NO;
	CN_HNOP_NO_HCNP_2    = HMRATE((1e-10)*8.7,0,0)*HNOP;
	CNP_O2_NOP_CO_2      = HMRATE((1e-11)*8.6,0,0)*O2;
	CNP_O2_OCNP_O_2      = HMRATE((1e-11)*8.6,0,0)*O2;
	HCNP_CO_HCOP_CN_2    = HMRATE((1e-10)*1.4,0,0)*CO;
	HCNP_S_HSP_CN_2      = HMRATE((1e-10)*5.7,0,0)*ATS;
	CO_HNOP_NO_HCOP_2    = HMRATE(1.0e-10,0,0)*HNOP;
	N2P_OCS_SP_N2_CO_2   = HMRATE((1e-9)*1.04,0,0)*OCS;
	HCOP_S_HSP_CO_2      = HMRATE((1e-10)*3.3,0,0)*ATS;
	HCOP_CS_HCSP_CO_2    = HMRATE(1.2e-9,0,0)*CS;
	HNOP_S_HSP_NO_2      = HMRATE(1.1e-9,0,0)*ATS;
	O2_SP_SOP_O_2        = HMRATE((1e-11)*1.5,0,0)*SP;
	O2P_S_SOP_O_2        = HMRATE(5.4e-10,0,0)*ATS;
	O2_CSP_OCSP_O_2      = HMRATE((1e-10)*1.3,0,0)*CSP;
	S_SIOP_SO_SIP_2      = HMRATE(1.0e-9,0,0)*SIOP;
	CP_NH3_NH3P_C_2      = HMRATE((1e-10)*5.06,0,0)*NH3;
	C_CNP_CN_CP_2        = HMRATE(1.1e-10,0,0)*CNP;
	C_N2P_N2_CP_2        = HMRATE(1.1e-10,0,0)*N2P;
	CP_NO_NOP_C_2        = HMRATE(5.2e-10,0,0)*NO;
	CP_SIN_SINP_C_2      = HMRATE(1.0e-9,0,0)*SIN;
	C_CSP_CS_CP_2        = HMRATE((1e-9)*1.6,0,0)*CSP;
	CP_NS_NSP_C_2        = HMRATE((1e-10)*7.6,0,0)*NS;
	CP_SO_SOP_C_2        = HMRATE((1e-10)*2.6,0,0)*SO;
	CP_OCS_OCSP_C_2      = HMRATE(4.0e-10,0,0)*OCS;
	CH_NH2P_NH2_CHP_2    = HMRATE(3.5e-10,0,0)*NH2P;
	CHP_NH3_NH3P_CH_2    = HMRATE((1e-10)*4.59,0,0)*NH3;
	CH_CNP_CN_CHP_2      = HMRATE((1e-10)*6.4,0,0)*CNP;
	CH_N2P_N2_CHP_2      = HMRATE((1e-10)*6.3,0,0)*N2P;
	CHP_NO_NOP_CH_2      = HMRATE((1e-10)*7.6,0,0)*NO;
	NP_NH_NHP_N_2        = HMRATE((1e-10)*3.7,0,0)*NH;
	NP_NH2_NH2P_N_2      = HMRATE(1.0e-9,0,0)*NH2;
	NP_NH3_NH3P_N_2      = HMRATE((1e-9)*1.97,0,0)*NH3;
	NP_CN_CNP_N_2        = HMRATE(1.1e-9,0,0)*CN;
	NP_HCN_HCNP_N_2      = HMRATE(1.2e-9,0,0)*HCN;
	N_N2P_N2_NP_2        = HMRATE(1.0e-11,0,0)*N2P;
	NP_NO_NOP_N_2        = HMRATE((1e-10)*4.51,0,0)*NO;
	NP_OCS_OCSP_N_2      = HMRATE((1e-9)*1.02,0,0)*OCS;
	CH2_NH2P_NH2_CH2P_2  = HMRATE((1e-10)*4.9,0,0)*NH2P;
	CH2_CNP_CN_CH2P_2    = HMRATE((1e-10)*8.8,0,0)*CNP;
	CH2_N2P_N2_CH2P_2    = HMRATE((1e-10)*8.7,0,0)*N2P;
	CH2P_NO_NOP_CH2_2    = HMRATE((1e-10)*4.2,0,0)*NO;
	NH_OP_O_NHP_2        = HMRATE((1e-10)*3.6,0,0)*OP;
	NH_OHP_OH_NHP_2      = HMRATE((1e-10)*3.6,0,0)*OHP;
	NHP_NH3_NH3P_NH_2    = HMRATE((1e-9)*1.8,0,0)*NH3;
	NHP_H2O_H2OP_NH_2    = HMRATE((1e-9)*1.05,0,0)*H2O;
	NH_CNP_CN_NHP_2      = HMRATE((1e-10)*6.5,0,0)*CNP;
	NH_COP_CO_NHP_2      = HMRATE((1e-10)*3.2,0,0)*COP;
	NH_N2P_N2_NHP_2      = HMRATE((1e-10)*6.5,0,0)*N2P;
	NHP_NO_NOP_NH_2      = HMRATE((1e-10)*7.12,0,0)*NO;
	NHP_O2_O2P_NH_2      = HMRATE((1e-10)*4.51,0,0)*O2;
	NHP_S_SP_NH_2        = HMRATE((1e-10)*6.9,0,0)*ATS;
	CH3P_NO_NOP_CH3_2    = HMRATE(1.0e-9,0,0)*NO;
	OP_NH2_NH2P_O_2      = HMRATE(1.0e-9,0,0)*NH2;
	OP_NH3_NH3P_O_2      = HMRATE(1.2e-9,0,0)*NH3;
	O_CNP_CN_OP_2        = HMRATE((1e-11)*6.5,0,0)*CNP;
	O_HCNP_HCN_OP_2      = HMRATE((1e-11)*6.5,0,0)*HCNP;
	O_N2P_N2_OP_2        = HMRATE(1.0e-11,0,0)*N2P;
	OP_NO_NOP_O_2        = HMRATE((1e-12)*1.7,0,0)*NO;
	OP_OCS_OCSP_O_2      = HMRATE((1e-10)*6.5,0,0)*OCS;
	NH2_OHP_OH_NH2P_2    = HMRATE(5.0e-10,0,0)*OHP;
	NH2P_NH3_NH3P_NH2_2  = HMRATE((1e-10)*6.9,0,0)*NH3;
	NH2_H2OP_H2O_NH2P_2  = HMRATE((1e-10)*4.9,0,0)*H2OP;
	NH2_CNP_CN_NH2P_2    = HMRATE((1e-10)*9.1,0,0)*CNP;
	NH2_COP_CO_NH2P_2    = HMRATE((1e-10)*4.5,0,0)*COP;
	NH2_N2P_N2_NH2P_2    = HMRATE((1e-10)*8.9,0,0)*N2P;
	NH2P_NO_NOP_NH2_2    = HMRATE(7.0e-10,0,0)*NO;
	NH2_O2P_O2_NH2P_2    = HMRATE((1e-10)*8.7,0,0)*O2P;
	NH2P_S_SP_NH2_2      = HMRATE((1e-10)*4.4,0,0)*ATS;
	CH4P_NH3_NH3P_CH4_2  = HMRATE((1e-9)*1.65,0,0)*NH3;
	CH4P_OCS_OCSP_CH4_2  = HMRATE((1e-10)*4.2,0,0)*OCS;
	OHP_NH3_NH3P_OH_2    = HMRATE(1.2e-9,0,0)*NH3;
	OH_CNP_CN_OHP_2      = HMRATE((1e-10)*6.4,0,0)*CNP;
	OH_N2P_N2_OHP_2      = HMRATE((1e-10)*6.3,0,0)*N2P;
	OHP_NO_NOP_OH_2      = HMRATE((1e-10)*3.59,0,0)*NO;
	NH3_H2OP_H2O_NH3P_2  = HMRATE((1e-9)*2.21,0,0)*H2OP;
	NH3_HCNP_HCN_NH3P_2  = HMRATE((1e-9)*1.68,0,0)*HCNP;
	NH3_COP_CO_NH3P_2    = HMRATE((1e-9)*2.02,0,0)*COP;
	NH3_N2P_N2_NH3P_2    = HMRATE((1e-9)*1.9,0,0)*N2P;
	NH3P_SI_SIP_NH3_2    = HMRATE((1e-9)*1.9,0,0)*ATSI;
	NH3_O2P_O2_NH3P_2    = HMRATE(2.0e-9,0,0)*O2P;
	NH3_SP_S_NH3P_2      = HMRATE((1e-9)*1.44,0,0)*SP;
	NH3_HSP_HS_NH3P_2    = HMRATE((1e-10)*5.25,0,0)*HSP;
	NH3_SOP_SO_NH3P_2    = HMRATE(1.3e-9,0,0)*SOP;
	H2O_HCNP_HCN_H2OP_2  = HMRATE((1e-9)*1.8,0,0)*HCNP;
	H2O_N2P_N2_H2OP_2    = HMRATE((1e-9)*2.3,0,0)*N2P;
	H2OP_NO_NOP_H2O_2    = HMRATE((1e-10)*2.7,0,0)*NO;
	CNP_HCN_HCNP_CN_2    = HMRATE((1e-9)*1.79,0,0)*HCN;
	CNP_CO_COP_CN_2      = HMRATE((1e-10)*6.3,0,0)*CO;
	CN_N2P_N2_CNP_2      = HMRATE(1.0e-10,0,0)*N2P;
	CNP_NO_NOP_CN_2      = HMRATE((1e-10)*5.7,0,0)*NO;
	CNP_O2_O2P_CN_2      = HMRATE((1e-10)*2.58,0,0)*O2;
	CNP_S_SP_CN_2        = HMRATE(1.1e-9,0,0)*ATS;
	HCN_COP_CO_HCNP_2    = HMRATE((1e-9)*3.4,0,0)*COP;
	HCN_N2P_N2_HCNP_2    = HMRATE((1e-10)*3.9,0,0)*N2P;
	HCNP_NO_NOP_HCN_2    = HMRATE((1e-10)*8.1,0,0)*NO;
	HCNP_O2_O2P_HCN_2    = HMRATE((1e-10)*3.2,0,0)*O2;
	HCNP_S_SP_HCN_2      = HMRATE((1e-10)*5.7,0,0)*ATS;
	CO_N2P_N2_COP_2      = HMRATE((1e-11)*7.4,0,0)*N2P;
	COP_NO_NOP_CO_2      = HMRATE((1e-10)*3.3,0,0)*NO;
	N2P_NO_NOP_N2_2      = HMRATE((1e-10)*4.4,0,0)*NO;
	N2P_O2_O2P_N2_2      = HMRATE((1e-11)*5.,0,0)*O2;
	N2P_S_SP_N2_2        = HMRATE(1.1e-9,0,0)*ATS;
	SI_NOP_NO_SIP_2      = HMRATE((1e-9)*1.6,0,0)*NOP;
	SI_HSP_HS_SIP_2      = HMRATE((1e-9)*1.4,0,0)*HSP;
	SI_CSP_CS_SIP_2      = HMRATE((1e-10)*1.5,0,0)*CSP;
	NO_HNOP_HNO_NOP_2    = HMRATE(7.0e-10,0,0)*HNOP;
	NO_O2P_O2_NOP_2      = HMRATE((1e-10)*4.5,0,0)*O2P;
	NO_SP_S_NOP_2        = HMRATE((1e-10)*3.7,0,0)*SP;
	NO_HSP_HS_NOP_2      = HMRATE((1e-10)*4.5,0,0)*HSP;
	NO_SIOP_SIO_NOP_2    = HMRATE((1e-10)*7.2,0,0)*SIOP;
	NO_S2P_S2_NOP_2      = HMRATE((1e-10)*5.1,0,0)*S2P;
	O2P_NO2_NO2P_O2_2    = HMRATE((1e-10)*6.6,0,0)*NO2;
	S_HSP_HS_SP_2        = HMRATE((1e-10)*9.7,0,0)*HSP;
	C_N_CN_NU_2          = HMRATE((1e-17)*1.,0,0)*ATN;
	C_S_CS_NU_2          = HMRATE(4.36E-19,0.22,0)*ATS;
	CP_S_CSP_NU_2        = HMRATE(3.07E-19,0.15,0)*ATS;
	NP_N_N2P_NU_2        = HMRATE(3.71E-18,0.24,26.1)*ATN;
	CO_S_OCS_NU_2        = HMRATE((1e-17)*1.6,-1.5,0)*ATS;
	CHP_S_SP_CH_2        = HMRATE((1e-10)*4.7,0,0)*ATS;
	NP_CH2_CH2P_N_2      = HMRATE(1.0e-9,0,0)*CH2;
	NP_CH4_CH4P_N_2      = HMRATE((1e-11)*2.8,0,0)*CH4;
	NP_OH_OHP_N_2        = HMRATE((1e-10)*3.7,0,0)*OH;
	NP_H2O_H2OP_N_2      = HMRATE((1e-9)*2.8,0,0)*H2O;
	NP_CO_COP_N_2        = HMRATE((1e-10)*8.25,0,0)*CO;
	NP_O2_O2P_N_2        = HMRATE((1e-10)*3.11,0,0)*O2;
	OP_O2_O2P_O_2        = HMRATE((1e-11)*1.9,0,0)*O2;
	OHP_S_SP_OH_2        = HMRATE((1e-10)*4.3,0,0)*ATS;
	H2OP_S_SP_H2O_2      = HMRATE((1e-10)*4.3,0,0)*ATS;
	COP_S_SP_CO_2        = HMRATE(1.1e-9,0,0)*ATS;
	SI_SP_S_SIP_2        = HMRATE((1e-9)*1.6,0,0)*SP;
	O2P_S_SP_O2_2        = HMRATE(5.4e-10,0,0)*ATS;
	H2_N_NH_H           = HMRATE((1e-9)*1.69,0,18095)*H2g;
	H2_NH_NH2_H         = HMRATE((1e-11)*5.96,0,7782)*H2g;
	H2_NH2_NH3_H        = HMRATE((1e-15)*2.05,3.89,1400)*H2g;
	H2_CN_HCN_H         = HMRATE(0.000000000000404,2.87,820)*H2g;
	HP_HNO_NOP_H2       = HMRATE((1e-9)*4.,0,0)*HII;
	HP_HS_SP_H2         = HMRATE((1e-9)*1.6,0,0)*HII;
	H_HSP_SP_H2         = HMRATE(1.1e-10,0,0)*HI;
	HP_OCS_HSP_CO       = HMRATE((1e-9)*2.1,0,0)*HII;
	H2P_N_NHP_H         = HMRATE((1e-9)*1.9,0,0)*H2p;
	H2_NP_NHP_H         = HMRATE(1.0e-9,0,85)*H2g;
	H2_NHP_N_H3P        = HMRATE((1e-10)*2.25,0,0)*H2g;
	H2P_NH_NH2P_H       = HMRATE((1e-10)*7.6,0,0)*H2p;
	H2_NHP_NH2P_H       = HMRATE((1e-9)*1.28,0,0)*H2g;
	H2_NH2P_NH3P_H      = HMRATE((1e-10)*2.7,0,0)*H2g;
	H2_NH3P_NH4P_H      = HMRATE(0.0000000000002,0,0)*H2g;
	H2P_CN_HCNP_H       = HMRATE(1.2e-9,0,0)*H2p;
	H2_CNP_HCNP_H       = HMRATE(1.0e-9,0,0)*H2g;
	H2P_NO_HNOP_H       = HMRATE(1.1e-9,0,0)*H2p;
	H2_SP_HSP_H         = HMRATE(1.1e-10,0,9860)*H2g;
	H2_CSP_HCSP_H       = HMRATE((1e-10)*4.5,0,0)*H2g;
	H2_NO2P_NOP_H2O     = HMRATE((1e-10)*1.5,0,0)*H2g;
	H3P_NH_NH2P_H2      = HMRATE(1.3e-9,0,0)*H3p;
	H3P_NH2_NH3P_H2     = HMRATE((1e-9)*1.8,0,0)*H3p;
	H3P_NH3_NH4P_H2     = HMRATE(2.7e-9,0,0)*H3p;
	H3P_CN_HCNP_H2      = HMRATE(2.0e-9,0,0)*H3p;
	H3P_NO_HNOP_H2      = HMRATE(1.1e-9,0,0)*H3p;
	H3P_S_HSP_H2        = HMRATE((1e-9)*2.6,0,0)*H3p;
	H3P_CS_HCSP_H2      = HMRATE((1e-9)*2.9,0,0)*H3p;
	H3P_NO2_NOP_OH_H2   = HMRATE(7.0e-10,0,0)*H3p;
	HeP_NH_NP_He_H      = HMRATE(1.1e-9,0,0)*HeII;
	HeP_NH2_NP_He_H2    = HMRATE((1e-10)*8.,0,0)*HeII;
	HeP_NH2_NHP_He_H    = HMRATE((1e-10)*8.,0,0)*HeII;
	HeP_NH3_NHP_He_H2   = HMRATE((1e-10)*1.76,0,0)*HeII;
	HeP_NH3_NH2P_He_H   = HMRATE((1e-9)*1.76,0,0)*HeII;
	HeP_CN_N_CP_He      = HMRATE((1e-10)*8.8,0,0)*HeII;
	HeP_CN_NP_C_He      = HMRATE((1e-10)*8.8,0,0)*HeII;
	HeP_HCN_N_CHP_He    = HMRATE((1e-10)*6.51,0,0)*HeII;
	HeP_HCN_NP_CH_He    = HMRATE((1e-10)*2.17,0,0)*HeII;
	HeP_HCN_N_CP_He_H   = HMRATE((1e-10)*7.75,0,0)*HeII;
	HeP_HCN_CNP_He_H    = HMRATE((1e-9)*1.46,0,0)*HeII;
	HeP_N2_NP_N_He      = HMRATE((1e-10)*9.6,0,0)*HeII;
	HeP_NO_OP_N_He      = HMRATE(2.0e-10,0,0)*HeII;
	HeP_NO_O_NP_He      = HMRATE((1e-9)*1.4,0,0)*HeII;
	HeP_HNO_NOP_He_H    = HMRATE(1.0e-9,0,0)*HeII;
	HeP_HNO_NO_He_HP    = HMRATE(1.0e-9,0,0)*HeII;
	HeP_HS_SP_He_H      = HMRATE((1e-9)*1.7,0,0)*HeII;
	HeP_OCN_CN_OP_He    = HMRATE(3.0e-9,0,0)*HeII;
	HeP_OCN_CNP_O_He    = HMRATE(3.0e-9,0,0)*HeII;
	HeP_SIN_SIP_N_He    = HMRATE(2.0e-9,0,0)*HeII;
	HeP_N2O_N2_OP_He    = HMRATE((1e-10)*2.76,0,0)*HeII;
	HeP_N2O_N2P_O_He    = HMRATE((1e-9)*1.24,0,0)*HeII;
	HeP_N2O_NO_NP_He    = HMRATE((1e-10)*3.,0,0)*HeII;
	HeP_N2O_NOP_N_He    = HMRATE((1e-10)*4.83,0,0)*HeII;
	HeP_CS_SP_C_He      = HMRATE(1.3e-9,0,0)*HeII;
	HeP_CS_S_CP_He      = HMRATE(1.3e-9,0,0)*HeII;
	HeP_NS_S_NP_He      = HMRATE(1.2e-9,0,0)*HeII;
	HeP_NS_SP_N_He      = HMRATE(1.2e-9,0,0)*HeII;
	HeP_SO_S_OP_He      = HMRATE((1e-10)*8.3,0,0)*HeII;
	HeP_SO_SP_O_He      = HMRATE((1e-10)*8.3,0,0)*HeII;
	HeP_OCS_SP_CO_He    = HMRATE((1e-10)*7.6,0,0)*HeII;
	HeP_OCS_S_COP_He    = HMRATE((1e-10)*7.6,0,0)*HeII;
	HeP_OCS_CSP_O_He    = HMRATE((1e-10)*7.6,0,0)*HeII;
	HeP_OCS_CS_OP_He    = HMRATE((1e-10)*7.6,0,0)*HeII;
	HeP_S2_SP_S_He      = HMRATE(2.0e-9,0,0)*HeII;
	HP_NH_NHP_H         = HMRATE((1e-9)*2.1,0,0)*HII;
	HP_NH2_NH2P_H       = HMRATE((1e-9)*2.9,0,0)*HII;
	HP_NH3_NH3P_H       = HMRATE(1.1e-9,0,0)*HII;
	H_CNP_CN_HP         = HMRATE((1e-10)*1.9,0,0)*HI;
	HP_HCN_HCNP_H       = HMRATE(0.0000000105,-0.13,0)*HII;
	H_HCNP_HCN_HP       = HMRATE((1e-11)*3.7,0,0)*HI;
	H_N2P_N2_HP         = HMRATE((1e-10)*1.2,0,0)*HI;
	HP_NO_NOP_H         = HMRATE((1e-9)*2.9,0,0)*HII;
	HP_HS_HSP_H         = HMRATE((1e-9)*1.6,0,0)*HII;
	HP_SIN_SINP_H       = HMRATE(3.0e-9,0,0)*HII;
	HP_CS_CSP_H         = HMRATE((1e-9)*4.9,0,0)*HII;
	HP_NS_NSP_H         = HMRATE((1e-9)*4.7,0,0)*HII;
	HP_SO_SOP_H         = HMRATE((1e-9)*3.2,0,0)*HII;
	HP_OCS_OCSP_H       = HMRATE((1e-9)*2.1,0,0)*HII;
	HP_S2_S2P_H         = HMRATE(3.0e-9,0,0)*HII;
	H2P_NH_NHP_H2       = HMRATE((1e-10)*7.6,0,0)*H2p;
	H2P_NH2_NH2P_H2     = HMRATE((1e-9)*2.1,0,0)*H2p;
	H2P_NH3_NH3P_H2     = HMRATE((1e-9)*5.7,0,0)*H2p;
	H2P_CN_CNP_H2       = HMRATE(1.2e-9,0,0)*H2p;
	H2P_HCN_HCNP_H2     = HMRATE(2.7e-9,0,0)*H2p;
	H2P_NO_NOP_H2       = HMRATE(1.1e-9,0,0)*H2p;
	HeP_NH3_NH3P_He     = HMRATE((1e-10)*2.64,0,0)*HeII;
	HeP_N2_N2P_He       = HMRATE((1e-10)*6.4,0,0)*HeII;
	NHP_e_N_H         = HMRATE(0.000000043,-0.5,0)*ele;
	NH2P_e_N_H_H      = HMRATE(0.000000198,-0.5,0)*ele;
	NH2P_e_NH_H       = HMRATE(0.000000102,-0.5,0)*ele;
	NH3P_e_NH_H_H     = HMRATE(0.000000155,-0.5,0)*ele;
	NH3P_e_NH2_H      = HMRATE(0.000000155,-0.5,0)*ele;
	NH4P_e_NH2_H_H    = HMRATE(0.000000286,-0.5,0)*ele;
	NH4P_e_NH2_H2     = HMRATE(0.000000137,-0.5,0)*ele;
	NH4P_e_NH3_H      = HMRATE(0.000000938,-0.5,0)*ele;
	CNP_e_N_C         = HMRATE(0.00000018,-0.5,0)*ele;
	HCNP_e_CN_H       = HMRATE(0.0000002,-0.5,0)*ele;
	N2P_e_N_N         = HMRATE(0.000000036,-0.42,0)*ele;
	NOP_e_O_N         = HMRATE(0.00000043,-0.37,0)*ele;
	HNOP_e_NO_H       = HMRATE(0.0000003,-0.5,0)*ele;
	HSP_e_S_H         = HMRATE(0.0000002,-0.5,0)*ele;
	OCNP_e_CO_N       = HMRATE(0.0000003,-0.5,0)*ele;
	SINP_e_SI_N       = HMRATE(0.0000002,-0.5,0)*ele;
	CSP_e_S_C         = HMRATE(0.0000002,-0.5,0)*ele;
	HCSP_e_CS_H       = HMRATE(0.00000005,-0.75,0)*ele;
	NO2P_e_NO_O       = HMRATE(0.0000003,-0.5,0)*ele;
	NSP_e_S_N         = HMRATE(0.0000002,-0.5,0)*ele;
	SOP_e_S_O         = HMRATE(0.0000002,-0.5,0)*ele;
	OCSP_e_S_CO       = HMRATE(0.00000015,-0.5,0)*ele;
	OCSP_e_CS_O       = HMRATE(0.00000015,-0.5,0)*ele;
	S2P_e_S_S         = HMRATE(0.0000002,-0.5,0)*ele;
	Hminus_N_NH_e       = HMRATE(1.0e-9,0,0)*Hm;
	Hminus_NH_NH2_e     = HMRATE(1.0e-10,0,0)*Hm;
	Hminus_NH2_NH3_e    = HMRATE(1.0e-9,0,0)*Hm;
	Hminus_CN_HCN_e     = HMRATE(1.0e-10,0,0)*Hm;
	Hminus_NH4P_NH3_H2  = HMRATE(0.00000023,-0.5,0)*Hm;
	Hminus_NP_N_H       = HMRATE(0.00000023,-0.5,0)*Hm;
	NH_NU_NHP_e       = HMRATE(1.0e-11,0,0)*(TH85_field);
	NH_NU_N_H         = HMRATE(5.0e-10,0,0)*(TH85_field);
	NH2_NU_NH_H       = HMRATE((1e-10)*3.9,0,0)*(TH85_field);
	NH2_NU_NH2P_e     = HMRATE((1e-10)*1.73,0,0)*(TH85_field);
	NH3_NU_NH3P_e     = HMRATE((1e-10)*1.8,0,0)*(TH85_field);
	NH3_NU_NH_H2      = HMRATE((1e-10)*3.3,0,0)*(TH85_field);
	NH3_NU_NH2_H      = HMRATE(1.1e-9,0,0)*(TH85_field);
	CN_NU_N_C         = HMRATE(1.1e-9,0,0)*(TH85_field);
	HCN_NU_CN_H       = HMRATE(1.3e-9,0,0)*(TH85_field);
	N2_NU_N_N         = HMRATE((1e-10)*2.3,0,0)*(TH85_field);
	NO_NU_O_N         = HMRATE((1e-10)*4.3,0,0)*(TH85_field);
	NO_NU_NOP_e       = HMRATE((1e-10)*2.6,0,0)*(TH85_field);
	HNO_NU_NO_H       = HMRATE((1e-10)*1.7,0,0)*(TH85_field);
	HS_NU_S_H         = HMRATE((1e-10)*9.7,0,0)*(TH85_field);
	HSP_NU_SP_H       = HMRATE((1e-10)*3.,0,0)*(TH85_field);
	OCN_NU_CN_O       = HMRATE(1.0e-11,0,0)*(TH85_field);
	CS_NU_S_C         = HMRATE((1e-10)*9.7,0,0)*(TH85_field);
	CSP_NU_SP_C       = HMRATE(2.0e-10,0,0)*(TH85_field);
	CS_NU_CSP_e       = HMRATE(2.0e-10,0,0)*(TH85_field);
	NO2_NU_NO_O       = HMRATE(1.2e-9,0,0)*(TH85_field);
	NS_NU_S_N         = HMRATE(1.0e-11,0,0)*(TH85_field);
	SO_NU_S_O         = HMRATE((1e-9)*3.7,0,0)*(TH85_field);
	SO_NU_SOP_e       = HMRATE((1e-10)*6.,0,0)*(TH85_field);
	OCS_NU_S_CO       = HMRATE((1e-9)*3.7,0,0)*(TH85_field);
	OCS_NU_OCSP_e     = HMRATE((1e-10)*4.2,0,0)*(TH85_field);
	NH_CRP_NHP_e      = crnu_rate * 500 * 2;
	NH_CRP_N_H        = crnu_rate * 500 * 2;
	NH2_CRP_NH_H      = crnu_rate * 81 * 2;
	NH2_CRP_NH2P_e    = crnu_rate * 610 * 2;
	NH3_CRP_NH3P_e    = crnu_rate * 543 * 2;
	NH3_CRP_NH_H2     = crnu_rate * 541 * 2;
	NH3_CRP_NH2_H     = crnu_rate * 1325 * 2;
	CN_CRP_N_C        = crnu_rate * 10580 * 2;
	HCN_CRP_CN_H      = crnu_rate * 3114 * 2;
	N2_CRP_N_N        = crnu_rate * 50 * 2;
	NO_CRP_O_N        = crnu_rate * 427 * 2;
	NO_CRP_NOP_e      = crnu_rate * 430 * 2;
	HNO_CRP_NO_H      = crnu_rate * 1000 * 2;
	HS_CRP_S_H        = crnu_rate * 500 * 2;
	OCN_CRP_CN_O      = crnu_rate * 1500 * 2;
	CS_CRP_S_C        = crnu_rate * 500 * 2;
	CS_CRP_CSP_e      = crnu_rate * 500 * 2;
	NO2_CRP_NO_O      = crnu_rate * 1500 * 2;
	NS_CRP_S_N        = crnu_rate * 500 * 2;
	SO_CRP_S_O        = crnu_rate * 500 * 2;
	SO_CRP_SOP_e      = crnu_rate * 500 * 2;
	OCS_CRP_S_CO      = crnu_rate * 5360 * 2;
	OCS_CRP_OCSP_e    = crnu_rate * 1444 * 2;
	P_C_NH = ATC*NH;
	P_C_N2 = ATC*N2;
	P_C_NO = ATC*NO;
	P_C_HS = ATC*HS;
	P_C_OCN = ATC*OCN;
	P_C_NS = ATC*NS;
	P_C_SO = ATC*SO;
	P_C_S2 = ATC*S2;
	P_C_NH3P = ATC*NH3P;
	P_C_HCNP = ATC*HCNP;
	P_C_HNOP = ATC*HNOP;
	P_CH_N = CH*ATN;
	P_CH_N2 = CH*N2;
	P_CH_NO = CH*NO;
	P_CH_HNO = CH*HNO;
	P_CH_S = CH*ATS;
	P_N_NH = ATN*NH;
	P_N_CH3 = ATN*CH3;
	P_N_OH = ATN*OH;
	P_N_CN = ATN*CN;
	P_N_SIH = ATN*SIH;
	P_N_NO = ATN*NO;
	P_N_HNO = ATN*HNO;
	P_N_O2 = ATN*O2;
	P_N_HS = ATN*HS;
	P_N_CS = ATN*CS;
	P_N_NS = ATN*NS;
	P_N_SO = ATN*SO;
	P_N_S2 = ATN*S2;
	P_CH2_CN = CH2*CN;
	P_CH2_NO = CH2*NO;
	P_CH2_HNO = CH2*HNO;
	P_NH_NH = NH*NH;
	P_NH_O = NH*ATO;
	P_NH_OH = NH*OH;
	P_NH_H2O = NH*H2O;
	P_NH_CN = NH*CN;
	P_NH_NO = NH*NO;
	P_NH_S = NH*ATS;
	P_NH_NO2 = NH*NO2;
	P_CH3_NH3 = CH3*NH3;
	P_CH3_CN = CH3*CN;
	P_CH3_HNO = CH3*HNO;
	P_O_NH2 = ATO*NH2;
	P_O_NH3 = ATO*NH3;
	P_O_CN = ATO*CN;
	P_O_HCN = ATO*HCN;
	P_O_N2 = ATO*N2;
	P_O_NO = ATO*NO;
	P_O_HNO = ATO*HNO;
	P_O_HS = ATO*HS;
	P_O_OCN = ATO*OCN;
	P_O_SIN = ATO*SIN;
	P_O_N2O = ATO*N2O;
	P_O_CS = ATO*CS;
	P_O_NS = ATO*NS;
	P_O_SO = ATO*SO;
	P_O_OCS = ATO*OCS;
	P_O_S2 = ATO*S2;
	P_NH2_OH = NH2*OH;
	P_NH2_NO = NH2*NO;
	P_CH4_CN = CH4*CN;
	P_OH_NH3 = OH*NH3;
	P_OH_CN = OH*CN;
	P_OH_HCN = OH*HCN;
	P_OH_NO = OH*NO;
	P_OH_S = OH*ATS;
	P_OH_N2O = OH*N2O;
	P_OH_CS = OH*CS;
	P_NH3_CN = NH3*CN;
	P_CN_NO = CN*NO;
	P_CN_HNO = CN*HNO;
	P_CN_O2 = CN*O2;
	P_CN_S = CN*ATS;
	P_CO_HS = CO*HS;
	P_N2_O2 = N2*O2;
	P_NO_NO = NO*NO;
	P_NO_HNO = NO*HNO;
	P_NO_S = NO*ATS;
	P_O2_S = O2*ATS;
	P_S_SO = ATS*SO;
	P_C_NHP = ATC*NHP;
	P_CP_NH = CP*NH;
	P_CP_NH2 = CP*NH2;
	P_C_NH2P = ATC*NH2P;
	P_CP_NH3 = CP*NH3;
	P_CP_HS = CP*HS;
	P_C_HSP = ATC*HSP;
	P_CP_OCN = CP*OCN;
	P_CP_NS = CP*NS;
	P_CP_SO = CP*SO;
	P_CP_OCS = CP*OCS;
	P_CHP_N = CHP*ATN;
	P_CH_NP = CH*NP;
	P_CH_NHP = CH*NHP;
	P_CHP_NH = CHP*NH;
	P_CHP_NH2 = CHP*NH2;
	P_CH_NH2P = CH*NH2P;
	P_CHP_NH3 = CHP*NH3;
	P_CH_NH3P = CH*NH3P;
	P_CH_HCNP = CH*HCNP;
	P_CH_HNOP = CH*HNOP;
	P_CHP_S = CHP*ATS;
	P_CH_SP = CH*SP;
	P_CH_HSP = CH*HSP;
	P_CHP_OCS = CHP*OCS;
	P_N_CH2P = ATN*CH2P;
	P_NP_NH = NP*NH;
	P_N_NHP = ATN*NHP;
	P_NP_CH4 = NP*CH4;
	P_N_OHP = ATN*OHP;
	P_NP_NH3 = NP*NH3;
	P_N_H2OP = ATN*H2OP;
	P_NP_CO = NP*CO;
	P_NP_NO = NP*NO;
	P_NP_O2 = NP*O2;
	P_N_O2P = ATN*O2P;
	P_N_HSP = ATN*HSP;
	P_N_SIOP = ATN*SIOP;
	P_N_SOP = ATN*SOP;
	P_NP_OCS = NP*OCS;
	P_CH2_NHP = CH2*NHP;
	P_CH2_NH2P = CH2*NH2P;
	P_CH2P_NH3 = CH2P*NH3;
	P_CH2_NH3P = CH2*NH3P;
	P_CH2_HCNP = CH2*HCNP;
	P_CH2_HNOP = CH2*HNOP;
	P_CH2P_S = CH2P*ATS;
	P_CH2_SP = CH2*SP;
	P_NHP_NH = NHP*NH;
	P_NHP_O = NHP*ATO;
	P_NH_OP = NH*OP;
	P_NHP_NH2 = NHP*NH2;
	P_NH_NH2P = NH*NH2P;
	P_NHP_OH = NHP*OH;
	P_NH_OHP = NH*OHP;
	P_NHP_NH3 = NHP*NH3;
	P_NH_NH3P = NH*NH3P;
	P_NH_CH5P = NH*CH5P;
	P_NHP_H2O = NHP*H2O;
	P_NH_H2OP = NH*H2OP;
	P_NHP_H2O = NHP*H2O;
	P_NHP_CN = NHP*CN;
	P_NH_HCNP = NH*HCNP;
	P_NH_COP = NH*COP;
	P_NHP_CO = NHP*CO;
	P_NH_SIP = NH*SIP;
	P_NH_HCOP = NH*HCOP;
	P_NH_HNOP = NH*HNOP;
	P_NHP_O2 = NHP*O2;
	P_NH_O2P = NH*O2P;
	P_NHP_S = NHP*ATS;
	P_NH_SP = NH*SP;
	P_CH3P_NH3 = CH3P*NH3;
	P_CH3P_S = CH3P*ATS;
	P_O_NH2P = ATO*NH2P;
	P_O_NH3P = ATO*NH3P;
	P_OP_CN = OP*CN;
	P_OP_HCN = OP*HCN;
	P_OP_N2 = OP*N2;
	P_O_N2P = ATO*N2P;
	P_O_HNOP = ATO*HNOP;
	P_O_HSP = ATO*HSP;
	P_O_SINP = ATO*SINP;
	P_OP_N2O = OP*N2O;
	P_O_CSP = ATO*CSP;
	P_O_HCSP = ATO*HCSP;
	P_OP_NO2 = OP*NO2;
	P_O_NSP = ATO*NSP;
	P_NH2P_NH2 = NH2P*NH2;
	P_NH2_OHP = NH2*OHP;
	P_NH2P_OH = NH2P*OH;
	P_NH2P_NH3 = NH2P*NH3;
	P_NH2_NH3P = NH2*NH3P;
	P_NH2_CH5P = NH2*CH5P;
	P_NH2P_H2O = NH2P*H2O;
	P_NH2_H2OP = NH2*H2OP;
	P_NH2P_H2O = NH2P*H2O;
	P_NH2_H3OP = NH2*H3OP;
	P_NH2_HCNP = NH2*HCNP;
	P_NH2_COP = NH2*COP;
	P_NH2_HCOP = NH2*HCOP;
	P_NH2_HNOP = NH2*HNOP;
	P_NH2P_O2 = NH2P*O2;
	P_NH2P_S = NH2P*ATS;
	P_CH4P_NH3 = CH4P*NH3;
	P_CH4_NH3P = CH4*NH3P;
	P_CH4_N2P = CH4*N2P;
	P_CH4_HNOP = CH4*HNOP;
	P_CH4_SP = CH4*SP;
	P_CH4_CSP = CH4*CSP;
	P_OHP_NH3 = OHP*NH3;
	P_OH_NH3P = OH*NH3P;
	P_OHP_CN = OHP*CN;
	P_OH_HCNP = OH*HCNP;
	P_OHP_NO = OHP*NO;
	P_OH_HNOP = OH*HNOP;
	P_OHP_S = OHP*ATS;
	P_OH_SP = OH*SP;
	P_NH3P_NH3 = NH3P*NH3;
	P_NH3_CH5P = NH3*CH5P;
	P_NH3P_H2O = NH3P*H2O;
	P_NH3_H2OP = NH3*H2OP;
	P_NH3_H3OP = NH3*H3OP;
	P_NH3_COP = NH3*COP;
	P_NH3_HCOP = NH3*HCOP;
	P_NH3_HNOP = NH3*HNOP;
	P_NH3_HSP = NH3*HSP;
	P_NH3_HCSP = NH3*HCSP;
	P_CH5P_S = CH5P*ATS;
	P_H2O_CNP = H2O*CNP;
	P_H2O_HCNP = H2O*HCNP;
	P_H2O_HNOP = H2O*HNOP;
	P_H2OP_S = H2OP*ATS;
	P_H2O_HSP = H2O*HSP;
	P_H3OP_CS = H3OP*CS;
	P_CNP_NO = CNP*NO;
	P_CN_HNOP = CN*HNOP;
	P_CNP_O2 = CNP*O2;
	P_HCNP_CO = HCNP*CO;
	P_HCNP_S = HCNP*ATS;
	P_CO_HNOP = CO*HNOP;
	P_N2P_OCS = N2P*OCS;
	P_HCOP_S = HCOP*ATS;
	P_HCOP_CS = HCOP*CS;
	P_HNOP_S = HNOP*ATS;
	P_O2_SP = O2*SP;
	P_O2P_S = O2P*ATS;
	P_O2_CSP = O2*CSP;
	P_S_SIOP = ATS*SIOP;
	P_CP_NH3 = CP*NH3;
	P_C_CNP = ATC*CNP;
	P_C_N2P = ATC*N2P;
	P_CP_NO = CP*NO;
	P_CP_SIN = CP*SIN;
	P_C_CSP = ATC*CSP;
	P_CP_NS = CP*NS;
	P_CP_SO = CP*SO;
	P_CP_OCS = CP*OCS;
	P_CH_NH2P = CH*NH2P;
	P_CHP_NH3 = CHP*NH3;
	P_CH_CNP = CH*CNP;
	P_CH_N2P = CH*N2P;
	P_CHP_NO = CHP*NO;
	P_NP_NH = NP*NH;
	P_NP_NH2 = NP*NH2;
	P_NP_NH3 = NP*NH3;
	P_NP_CN = NP*CN;
	P_NP_HCN = NP*HCN;
	P_N_N2P = ATN*N2P;
	P_NP_NO = NP*NO;
	P_NP_OCS = NP*OCS;
	P_CH2_NH2P = CH2*NH2P;
	P_CH2_CNP = CH2*CNP;
	P_CH2_N2P = CH2*N2P;
	P_CH2P_NO = CH2P*NO;
	P_NH_OP = NH*OP;
	P_NH_OHP = NH*OHP;
	P_NHP_NH3 = NHP*NH3;
	P_NHP_H2O = NHP*H2O;
	P_NH_CNP = NH*CNP;
	P_NH_COP = NH*COP;
	P_NH_N2P = NH*N2P;
	P_NHP_NO = NHP*NO;
	P_NHP_O2 = NHP*O2;
	P_NHP_S = NHP*ATS;
	P_CH3P_NO = CH3P*NO;
	P_OP_NH2 = OP*NH2;
	P_OP_NH3 = OP*NH3;
	P_O_CNP = ATO*CNP;
	P_O_HCNP = ATO*HCNP;
	P_O_N2P = ATO*N2P;
	P_OP_NO = OP*NO;
	P_OP_OCS = OP*OCS;
	P_NH2_OHP = NH2*OHP;
	P_NH2P_NH3 = NH2P*NH3;
	P_NH2_H2OP = NH2*H2OP;
	P_NH2_CNP = NH2*CNP;
	P_NH2_COP = NH2*COP;
	P_NH2_N2P = NH2*N2P;
	P_NH2P_NO = NH2P*NO;
	P_NH2_O2P = NH2*O2P;
	P_NH2P_S = NH2P*ATS;
	P_CH4P_NH3 = CH4P*NH3;
	P_CH4P_OCS = CH4P*OCS;
	P_OHP_NH3 = OHP*NH3;
	P_OH_CNP = OH*CNP;
	P_OH_N2P = OH*N2P;
	P_OHP_NO = OHP*NO;
	P_NH3_H2OP = NH3*H2OP;
	P_NH3_HCNP = NH3*HCNP;
	P_NH3_COP = NH3*COP;
	P_NH3_N2P = NH3*N2P;
	P_NH3P_SI = NH3P*ATSI;
	P_NH3_O2P = NH3*O2P;
	P_NH3_SP = NH3*SP;
	P_NH3_HSP = NH3*HSP;
	P_NH3_SOP = NH3*SOP;
	P_H2O_HCNP = H2O*HCNP;
	P_H2O_N2P = H2O*N2P;
	P_H2OP_NO = H2OP*NO;
	P_CNP_HCN = CNP*HCN;
	P_CNP_CO = CNP*CO;
	P_CN_N2P = CN*N2P;
	P_CNP_NO = CNP*NO;
	P_CNP_O2 = CNP*O2;
	P_CNP_S = CNP*ATS;
	P_HCN_COP = HCN*COP;
	P_HCN_N2P = HCN*N2P;
	P_HCNP_NO = HCNP*NO;
	P_HCNP_O2 = HCNP*O2;
	P_HCNP_S = HCNP*ATS;
	P_CO_N2P = CO*N2P;
	P_COP_NO = COP*NO;
	P_N2P_NO = N2P*NO;
	P_N2P_O2 = N2P*O2;
	P_N2P_S = N2P*ATS;
	P_SI_NOP = ATSI*NOP;
	P_SI_HSP = ATSI*HSP;
	P_SI_CSP = ATSI*CSP;
	P_NO_HNOP = NO*HNOP;
	P_NO_O2P = NO*O2P;
	P_NO_SP = NO*SP;
	P_NO_HSP = NO*HSP;
	P_NO_SIOP = NO*SIOP;
	P_NO_S2P = NO*S2P;
	P_O2P_NO2 = O2P*NO2;
	P_S_HSP = ATS*HSP;
	P_C_N = ATC*ATN;
	P_C_S = ATC*ATS;
	P_CP_S = CP*ATS;
	P_NP_N = NP*ATN;
	P_CO_S = CO*ATS;
	P_CHP_S = CHP*ATS;
	P_NP_CH2 = NP*CH2;
	P_NP_CH4 = NP*CH4;
	P_NP_OH = NP*OH;
	P_NP_H2O = NP*H2O;
	P_NP_CO = NP*CO;
	P_NP_O2 = NP*O2;
	P_OP_O2 = OP*O2;
	P_OHP_S = OHP*ATS;
	P_H2OP_S = H2OP*ATS;
	P_COP_S = COP*ATS;
	P_SI_SP = ATSI*SP;
	P_O2P_S = O2P*ATS;
	b_C_NH_N_CH          = HMRATE((1e-11)*1.73,0.5,4000)*P_C_NH;
	b_C_NH_CN_H            = HMRATE(1.1e-10,0.5,0)*P_C_NH;
	b_C_N2_CN_N          = HMRATE((1e-11)*8.69,0,22600)*P_C_N2;
	b_C_NO_CN_O            = HMRATE((1e-11)*4.8,0,0)*P_C_NO;
	b_C_NO_CO_N          = HMRATE((1e-11)*4.65,0,0)*P_C_NO;
	b_C_HS_S_CH            = HMRATE((1e-11)*1.2,0.58,5880)*P_C_HS;
	b_C_HS_CS_H            = HMRATE(2.0e-11,0,0)*P_C_HS;
	b_C_OCN_CO_CN          = HMRATE(4.0e-11,0.5,0)*P_C_OCN;
	b_C_NS_S_CN            = HMRATE(2.0e-11,0.5,0)*P_C_NS;
	b_C_NS_CS_N          = HMRATE((1e-11)*1.73,0.5,4000)*P_C_NS;
	b_C_SO_S_CO            = HMRATE((1e-11)*7.2,0,0)*P_C_SO;
	b_C_S2_CS_S            = HMRATE((1e-11)*1.73,0.5,0)*P_C_S2;
	b_CH_N_NH_C          = HMRATE((1e-11)*3.03,0.65,1207)*P_CH_N;
	b_CH_N_CN_H          = HMRATE((1e-10)*1.66,-0.09,0)*P_CH_N;
	b_CH_N2_HCN_N        = HMRATE(0.00000000000056,0.88,10128)*P_CH_N2;
	b_CH_NO_HCN_O          = HMRATE((1e-9)*5.59,0,10814)*P_CH_NO;
	b_CH_NO_CN_OH          = HMRATE(2.32E-26,0,0)*P_CH_NO;
	b_CH_NO_OCN_H          = HMRATE(1.13E-25,0,0)*P_CH_NO;
	b_CH_HNO_NO_CH2        = HMRATE((1e-11)*1.73,0.5,0)*P_CH_HNO;
	b_CH_S_HS_C          = HMRATE((1e-11)*1.73,0.5,4000)*P_CH_S;
	b_CH_S_CS_H          = HMRATE((1e-12)*1.1,0,0)*P_CH_S;
	b_N_NH_N2_H          = HMRATE((1e-11)*4.98,0,0)*P_N_NH;
	b_N_CH3_HCN_H_H      = HMRATE((1e-13)*3.32,0,0)*P_N_CH3;
	b_N_CH3_HCN_H2       = HMRATE((1e-11)*1.3,0.5,0)*P_N_CH3;
	b_N_OH_O_NH          = HMRATE((1e-11)*1.88,0.1,10700)*P_N_OH;
	b_N_OH_NO_H          = HMRATE((1e-11)*5.32,-0.25,0)*P_N_OH;
	b_N_CN_N2_C          = HMRATE((1e-10)*3.,0,0)*P_N_CN;
	b_N_SIH_SIN_H          = HMRATE((1e-11)*5.,0.5,0)*P_N_SIH;
	b_N_NO_N2_O          = HMRATE((1e-11)*3.75,0,26)*P_N_NO;
	b_N_HNO_NO_NH          = HMRATE((1e-12)*2.94,0.5,1000)*P_N_HNO;
	b_N_HNO_N2O_H          = HMRATE((1e-12)*1.43,0.5,1500)*P_N_HNO;
	b_N_O2_NO_O          = HMRATE((1e-12)*2.26,0.86,3134)*P_N_O2;
	b_N_HS_NS_H          = HMRATE((1e-11)*1.73,0.5,0)*P_N_HS;
	b_N_HS_S_NH          = HMRATE((1e-11)*1.73,0.5,9060)*P_N_HS;
	b_N_CS_S_CN          = HMRATE((1e-11)*3.8,0.5,1160)*P_N_CS;
	b_N_NS_S_N2          = HMRATE((1e-11)*1.73,0.5,0)*P_N_NS;
	b_N_SO_NS_O          = HMRATE((1e-11)*4.68,0.5,8254)*P_N_SO;
	b_N_SO_S_NO          = HMRATE((1e-11)*1.73,0.5,750)*P_N_SO;
	b_N_S2_NS_S          = HMRATE((1e-11)*1.73,0.5,4000)*P_N_S2;
	b_CH2_CN_HCN_CH        = HMRATE((1e-12)*5.3,0,2500)*P_CH2_CN;
	b_CH2_NO_HCN_OH        = HMRATE(0.000000000000832,0,1443)*P_CH2_NO;
	b_CH2_HNO_NO_CH3       = HMRATE((1e-11)*1.73,0.5,0)*P_CH2_HNO;
	b_NH_NH_N2_H_H         = HMRATE((1e-9)*1.16,0,0)*P_NH_NH;
	b_NH_NH_N2_H2          = HMRATE((1e-11)*1.7,0,0)*P_NH_NH;
	b_NH_O_OH_N          = HMRATE((1e-11)*1.16,0,0)*P_NH_O;
	b_NH_O_NO_H          = HMRATE((1e-10)*1.16,0,0)*P_NH_O;
	b_NH_OH_HNO_H          = HMRATE((1e-11)*3.32,0,0)*P_NH_OH;
	b_NH_OH_NH2_O          = HMRATE((1e-12)*2.93,0.1,5800)*P_NH_OH;
	b_NH_OH_H2O_N          = HMRATE((1e-12)*3.11,1.2,0)*P_NH_OH;
	b_NH_H2O_OH_NH2          = HMRATE((1e-12)*1.83,1.6,14090)*P_NH_H2O;
	b_NH_CN_HCN_N          = HMRATE((1e-12)*2.94,0.5,1000)*P_NH_CN;
	b_NH_NO_N2O_H          = HMRATE((1e-10)*1.16,-1.03,420)*P_NH_NO;
	b_NH_NO_N2_O_H         = HMRATE((1e-11)*5.,0,0)*P_NH_NO;
	b_NH_NO_N2_OH          = HMRATE((1e-11)*1.46,-0.58,37)*P_NH_NO;
	b_NH_S_HS_N          = HMRATE((1e-11)*1.73,0.5,4000)*P_NH_S;
	b_NH_S_NS_H          = HMRATE((1e-11)*1.73,0.5,0)*P_NH_S;
	b_NH_NO2_HNO_NO          = HMRATE((1e-12)*5.72,0.5,2500)*P_NH_NO2;
	b_NH_NO2_N2O_OH          = HMRATE((1e-13)*1.44,0, 1140)*P_NH_NO2;
	b_CH3_NH3_CH4_NH2          = HMRATE(0.0000000000000955,0,4890)*P_CH3_NH3;
	b_CH3_CN_HCN_CH2          = HMRATE((1e-12)*9.21,0.7,1500)*P_CH3_CN;
	b_CH3_HNO_NO_CH4          = HMRATE((1e-11)*1.44,0.5,0)*P_CH3_HNO;
	b_O_NH2_OH_NH          = HMRATE((1e-11)*1.39,0,40)*P_O_NH2;
	b_O_NH2_NO_H2          = HMRATE((1e-12)*8.3,0,0)*P_O_NH2;
	b_O_NH3_OH_NH2          = HMRATE((1e-11)*1.89,0,4003)*P_O_NH3;
	b_O_CN_NO_C          = HMRATE((1e-11)*3.81,0.5,14545)*P_O_CN;
	b_O_CN_CO_N          = HMRATE((1e-11)*4.36,0.46,364)*P_O_CN;
	b_O_HCN_CN_OH          = HMRATE((1e-10)*6.21,0,12439)*P_O_HCN;
	b_O_HCN_CO_NH          = HMRATE(0.00000000000073,1.14,3742)*P_O_HCN;
	b_O_HCN_OCN_H          = HMRATE((1e-12)*1.36,1.38,3693)*P_O_HCN;
	b_O_N2_NO_N          = HMRATE((1e-10)*2.51,0,38602)*P_O_N2;
	b_O_NO_O2_N          = HMRATE((1e-11)*1.18,0,20413)*P_O_NO;
	b_O_HNO_NO_OH          = HMRATE((1e-11)*6.,0,0)*P_O_HNO;
	b_O_HNO_O2_NH          = HMRATE((1e-12)*2.94,0.5,3500)*P_O_HNO;
	b_O_HNO_NO2_H          = HMRATE((1e-12)*1.44,0.5,0)*P_O_HNO;
	b_O_HS_S_OH          = HMRATE((1e-11)*1.74,0.67,956)*P_O_HS;
	b_O_HS_SO_H          = HMRATE((1e-10)*2.32,0,0)*P_O_HS;
	b_O_OCN_NO_CO          = HMRATE((1e-11)*9.43,-0.09,100)*P_O_OCN;
	b_O_OCN_O2_CN          = HMRATE((1e-10)*4.02,-1.43,3501)*P_O_OCN;
	b_O_SIN_NO_SI          = HMRATE((1e-11)*2.5,0.5,0)*P_O_SIN;
	b_O_SIN_SIO_N          = HMRATE((1e-11)*5.75,0.1,200)*P_O_SIN;
	b_O_N2O_NO_NO          = HMRATE((1e-10)*1.15,0,13400)*P_O_N2O;
	b_O_N2O_O2_N2          = HMRATE((1e-10)*1.66,0,14100)*P_O_N2O;
	b_O_CS_S_CO          = HMRATE((1e-10)*2.48,-0.65,783)*P_O_CS;
	b_O_CS_SO_C          = HMRATE((1e-11)*4.68,0.5,28940)*P_O_CS;
	b_O_NS_S_NO          = HMRATE((1e-11)*5.,0.5,0)*P_O_NS;
	b_O_NS_SO_N          = HMRATE((1e-11)*1.73,0.5,4000)*P_O_NS;
	b_O_SO_S_O2          = HMRATE(0.00000000000066,0,2760)*P_O_SO;
	b_O_OCS_SO_CO          = HMRATE((1e-11)*1.6,0,2150)*P_O_OCS;
	b_O_S2_SO_S          = HMRATE((1e-11)*1.73,0.5,0)*P_O_S2;
	b_NH2_OH_NH3_O          = HMRATE(0.000000000000208,0.76,262)*P_NH2_OH;
	b_NH2_OH_H2O_NH          = HMRATE((1e-12)*1.5,0,0)*P_NH2_OH;
	b_NH2_NO_N2_OH_H         = HMRATE((1e-12)*1.49,0,0)*P_NH2_NO;
	b_NH2_NO_N2_H2O          = HMRATE((1e-11)*4.27,-2.5,331)*P_NH2_NO;
	b_CH4_CN_HCN_CH3          = HMRATE((1e-12)*3.14,1.53,504)*P_CH4_CN;
	b_OH_NH3_H2O_NH2          = HMRATE((1e-13)*1.47,2.05,7)*P_OH_NH3;
	b_OH_CN_HCN_O          = HMRATE(1.0e-11,0,1000)*P_OH_CN;
	b_OH_CN_OCN_H          = HMRATE((1e-11)*7.,0,0)*P_OH_CN;
	b_OH_HCN_CN_H2O          = HMRATE((1e-13)*1.87,1.5,3887)*P_OH_HCN;
	b_OH_HCN_CO_NH2          = HMRATE((1e-13)*1.07,0,5892)*P_OH_HCN;
	b_OH_NO_NO2_H          = HMRATE((1e-12)*5.2,0,15100)*P_OH_NO;
	b_OH_S_HS_O          = HMRATE(6.6e-11,0,0)*P_OH_S;
	b_OH_S_SO_H          = HMRATE(6.6e-11,0,0)*P_OH_S;
	b_OH_N2O_HNO_NO          = HMRATE(1.04E-17,4.33,12623)*P_OH_N2O;
	b_OH_CS_OCS_H          = HMRATE(0.0000000000000939,1.12,800)*P_OH_CS;
	b_NH3_CN_HCN_NH2          = HMRATE((1e-11)*2.6,-1.1,0)*P_NH3_CN;
	b_CN_NO_N2_CO          = HMRATE((1e-10)*1.79,0,4040)*P_CN_NO;
	b_CN_NO_OCN_N          = HMRATE((1e-10)*1.62,0,21205)*P_CN_NO;
	b_CN_HNO_NO_HCN          = HMRATE((1e-11)*1.5,0.5,0)*P_CN_HNO;
	b_CN_O2_NO_CO          = HMRATE(0.00000000000053,0,0)*P_CN_O2;
	b_CN_S_NS_C          = HMRATE((1e-11)*5.71,0.5,32010)*P_CN_S;
	b_CN_S_CS_N          = HMRATE((1e-11)*1.73,0.5,0)*P_CN_S;
	b_CO_HS_OCS_H          = HMRATE(0.0000000000000595,1.12,8330)*P_CO_HS;
	b_N2_O2_N2O_O          = HMRATE(1.0e-10,0,55200)*P_N2_O2;
	b_NO_NO_O2_N2          = HMRATE((1e-11)*2.51,0,30653)*P_NO_NO;
	b_NO_NO_N2O_O          = HMRATE((1e-12)*7.22,0,33155)*P_NO_NO;
	b_NO_HNO_N2O_OH          = HMRATE((1e-11)*1.41,0,14890)*P_NO_HNO;
	b_NO_S_NS_O          = HMRATE(2.94e-11,0.5,17465)*P_NO_S;
	b_NO_S_SO_N          = HMRATE((1e-10)*1.75,0,20200)*P_NO_S;
	b_O2_S_SO_O          = HMRATE((1e-12)*2.28,0.52,0)*P_O2_S;
	b_S_SO_S2_O          = HMRATE((1e-11)*1.73,0.5,11500)*P_S_SO;
	b_C_NHP_N_CHP          = HMRATE((1e-9)*1.6,0,0)*P_C_NHP;
	b_CP_NH_CNP_H          = HMRATE((1e-10)*7.8,0,0)*P_CP_NH;
	b_CP_NH2_HCNP_H          = HMRATE(1.1e-9,0,0)*P_CP_NH2;
	b_C_NH2P_NH_CHP          = HMRATE(1.2e-9,0,0)*P_C_NH2P;
	b_CP_NH3_HCNP_H2          = HMRATE((1e-11)*7.,0,0)*P_CP_NH3;
	b_C_NH3P_NH_CH2P          = HMRATE(1.0e-11,0,0)*P_C_NH3P;
	b_C_HCNP_CN_CHP          = HMRATE(1.1e-9,0,0)*P_C_HCNP;
	b_C_HNOP_NO_CHP          = HMRATE(1.0e-9,0,0)*P_C_HNOP;
	b_CP_HS_CSP_H          = HMRATE(1.1e-9,0,0)*P_CP_HS;
	b_C_HSP_CSP_H          = HMRATE((1e-10)*9.9,0,0)*P_C_HSP;
	b_CP_OCN_COP_CN          = HMRATE((1e-9)*3.8,0,0)*P_CP_OCN;
	b_CP_NS_CSP_N          = HMRATE((1e-10)*7.6,0,0)*P_CP_NS;
	b_CP_SO_SP_CO          = HMRATE((1e-10)*2.6,0,0)*P_CP_SO;
	b_CP_SO_S_COP          = HMRATE((1e-10)*2.6,0,0)*P_CP_SO;
	b_CP_SO_CSP_O          = HMRATE((1e-10)*2.6,0,0)*P_CP_SO;
	b_CP_OCS_CSP_CO          = HMRATE((1e-9)*1.6,0,0)*P_CP_OCS;
	b_CHP_N_CNP_H          = HMRATE((1e-10)*1.9,0,0)*P_CHP_N;
	b_CH_NP_CNP_H          = HMRATE((1e-10)*3.6,0,0)*P_CH_NP;
	b_CH_NHP_CH2P_N          = HMRATE((1e-10)*9.9,0,0)*P_CH_NHP;
	b_CHP_NH_CNP_H2          = HMRATE((1e-10)*7.6,0,0)*P_CHP_NH;
	b_CHP_NH2_HCNP_H2          = HMRATE(1.1e-9,0,0)*P_CHP_NH2;
	b_CH_NH2P_NH_CH2P          = HMRATE(3.5e-10,0,0)*P_CH_NH2P;
	b_CHP_NH3_NH4P_C          = HMRATE((1e-10)*4.05,0,0)*P_CHP_NH3;
	b_CH_NH3P_NH4P_C          = HMRATE((1e-10)*6.9,0,0)*P_CH_NH3P;
	b_CH_HCNP_CN_CH2P          = HMRATE((1e-10)*6.3,0,0)*P_CH_HCNP;
	b_CH_HNOP_NO_CH2P          = HMRATE(6.2e-10,0,0)*P_CH_HNOP;
	b_CHP_S_HSP_C          = HMRATE((1e-10)*4.7,0,0)*P_CHP_S;
	b_CHP_S_CSP_H          = HMRATE((1e-10)*4.7,0,0)*P_CHP_S;
	b_CH_SP_CSP_H          = HMRATE(6.2e-10,0,0)*P_CH_SP;
	b_CH_HSP_S_CH2P          = HMRATE((1e-10)*5.8,0,0)*P_CH_HSP;
	b_CHP_OCS_HCSP_CO          = HMRATE((1e-9)*1.05,0,0)*P_CHP_OCS;
	b_N_CH2P_HCNP_H          = HMRATE((1e-10)*2.2,0,0)*P_N_CH2P;
	b_NP_NH_N2P_H          = HMRATE((1e-10)*3.7,0,0)*P_NP_NH;
	b_N_NHP_N2P_H          = HMRATE(1.3e-9,0,0)*P_N_NHP;
	b_NP_CH4_HCNP_H2_H          = HMRATE((1e-11)*5.6,0,0)*P_NP_CH4;
	b_N_OHP_NOP_H          = HMRATE((1e-10)*8.9,0,0)*P_N_OHP;
	b_NP_NH3_NH2P_NH          = HMRATE((1e-10)*2.16,0,0)*P_NP_NH3;
	b_N_H2OP_HNOP_H          = HMRATE((1e-10)*1.9,0,0)*P_N_H2OP;
	b_NP_CO_NOP_C          = HMRATE((1e-10)*1.45,0,0)*P_NP_CO;
	b_NP_NO_N2P_O          = HMRATE((1e-11)*7.9,0,0)*P_NP_NO;
	b_NP_O2_NO_OP          = HMRATE((1e-11)*3.66,0,0)*P_NP_O2;
	b_NP_O2_NOP_O          = HMRATE((1e-10)*2.63,0,0)*P_NP_O2;
	b_N_O2P_NOP_O          = HMRATE((1e-10)*1.8,0,0)*P_N_O2P;
	b_N_HSP_NSP_H          = HMRATE((1e-10)*7.4,0,0)*P_N_HSP;
	b_N_SIOP_NOP_SI          = HMRATE((1e-11)*9.,0,0)*P_N_SIOP;
	b_N_SIOP_NO_SIP          = HMRATE((1e-10)*2.1,0,0)*P_N_SIOP;
	b_N_SOP_NSP_O          = HMRATE((1e-11)*5.,0,0)*P_N_SOP;
	b_NP_OCS_SP_CO_N          = HMRATE((1e-10)*3.08,0,0)*P_NP_OCS;
	b_NP_OCS_CSP_NO          = HMRATE((1e-11)*7.,0,0)*P_NP_OCS;
	b_CH2_NHP_CH3P_N          = HMRATE((1e-9)*1.4,0,0)*P_CH2_NHP;
	b_CH2_NH2P_CH3P_NH          = HMRATE((1e-10)*4.9,0,0)*P_CH2_NH2P;
	b_CH2P_NH3_NH4P_CH          = HMRATE((1e-9)*1.26,0,0)*P_CH2P_NH3;
	b_CH2_NH3P_NH2_CH3P          = HMRATE((1e-10)*9.6,0,0)*P_CH2_NH3P;
	b_CH2_HCNP_CN_CH3P          = HMRATE((1e-10)*8.7,0,0)*P_CH2_HCNP;
	b_CH2_HNOP_NO_CH3P          = HMRATE((1e-10)*8.6,0,0)*P_CH2_HNOP;
	b_CH2P_S_HCSP_H          = HMRATE((1e-9)*1.4,0,0)*P_CH2P_S;
	b_CH2_SP_HCSP_H          = HMRATE(1.0e-11,0,0)*P_CH2_SP;
	b_NHP_NH_NH2P_N          = HMRATE(1.0e-9,0,0)*P_NHP_NH;
	b_NHP_O_OHP_N          = HMRATE(1.0e-9,0,0)*P_NHP_O;
	b_NH_OP_NOP_H          = HMRATE((1e-10)*3.6,0,0)*P_NH_OP;
	b_NHP_NH2_NH3P_N          = HMRATE((1e-9)*1.5,0,0)*P_NHP_NH2;
	b_NH_NH2P_NH3P_N          = HMRATE((1e-10)*7.3,0,0)*P_NH_NH2P;
	b_NHP_OH_H2OP_N          = HMRATE(1.0e-9,0,0)*P_NHP_OH;
	b_NH_OHP_NH2P_O          = HMRATE((1e-10)*3.6,0,0)*P_NH_OHP;
	b_NHP_NH3_NH4P_N          = HMRATE((1e-10)*6.,0,0)*P_NHP_NH3;
	b_NH_NH3P_NH4P_N          = HMRATE((1e-10)*7.1,0,0)*P_NH_NH3P;
	b_NH_CH5P_CH4_NH2P          = HMRATE((1e-10)*7.1,0,0)*P_NH_CH5P;
	b_NHP_H2O_NH3P_O          = HMRATE((1e-10)*1.75,0,0)*P_NHP_H2O;
	b_NHP_H2O_H3OP_N          = HMRATE((1e-9)*1.05,0,0)*P_NHP_H2O;
	b_NHP_H2O_HNOP_H2          = HMRATE(3.5e-10,0,0)*P_NHP_H2O;
	b_NH_H2OP_H3OP_N          = HMRATE((1e-10)*7.1,0,0)*P_NH_H2OP;
	b_NHP_H2O_OH_NH2P          = HMRATE((1e-10)*8.75,0,0)*P_NHP_H2O;
	b_NHP_CN_HCNP_N          = HMRATE((1e-9)*1.6,0,0)*P_NHP_CN;
	b_NH_HCNP_CN_NH2P          = HMRATE((1e-10)*6.5,0,0)*P_NH_HCNP;
	b_NHP_CO_HCOP_N          = HMRATE((1e-10)*4.41,0,0)*P_NHP_CO;
	b_NH_COP_HCOP_N          = HMRATE((1e-10)*3.2,0,0)*P_NH_COP;
	b_NHP_CO_OCNP_H          = HMRATE((1e-10)*5.39,0,0)*P_NHP_CO;
	b_NH_SIP_SINP_H          = HMRATE(1.0e-9,0,0)*P_NH_SIP;
	b_NH_HCOP_CO_NH2P          = HMRATE((1e-10)*6.4,0,0)*P_NH_HCOP;
	b_NH_HNOP_NO_NH2P          = HMRATE((1e-10)*6.3,0,0)*P_NH_HNOP;
	b_NH_O2P_HNOP_O          = HMRATE((1e-10)*3.2,0,0)*P_NH_O2P;
	b_NHP_O2_NOP_OH          = HMRATE((1e-10)*2.05,0,0)*P_NHP_O2;
	b_NH_O2P_NO2P_H          = HMRATE((1e-10)*3.2,0,0)*P_NH_O2P;
	b_NHP_S_HSP_N          = HMRATE((1e-10)*6.9,0,0)*P_NHP_S;
	b_NHP_S_NSP_H          = HMRATE((1e-10)*6.9,0,0)*P_NHP_S;
	b_NH_SP_NSP_H          = HMRATE((1e-10)*6.3,0,0)*P_NH_SP;
	b_CH3P_NH3_NH4P_CH2          = HMRATE((1e-10)*3.4,0,0)*P_CH3P_NH3;
	b_CH3P_S_HCSP_H2          = HMRATE((1e-9)*1.4,0,0)*P_CH3P_S;
	b_O_NH2P_HNOP_H          = HMRATE((1e-11)*7.2,0,0)*P_O_NH2P;
	b_O_NH3P_HNOP_H2          = HMRATE(1.0e-11,0,0)*P_O_NH3P;
	b_OP_CN_NOP_C          = HMRATE(1.0e-9,0,0)*P_OP_CN;
	b_OP_HCN_COP_NH          = HMRATE(1.2e-9,0,0)*P_OP_HCN;
	b_OP_HCN_NOP_CH          = HMRATE(1.2e-9,0,0)*P_OP_HCN;
	b_OP_HCN_HCOP_N          = HMRATE(1.2e-9,0,0)*P_OP_HCN;
	b_OP_N2_NOP_N          = HMRATE((1e-12)*1.2,0,0)*P_OP_N2;
	b_O_N2P_NOP_N          = HMRATE((1e-10)*1.3,0,0)*P_O_N2P;
	b_O_HNOP_NO2P_H          = HMRATE((1e-12)*1.,0,0)*P_O_HNOP;
	b_O_HSP_SP_OH          = HMRATE((1e-10)*2.9,0,0)*P_O_HSP;
	b_O_HSP_SOP_H          = HMRATE((1e-10)*2.9,0,0)*P_O_HSP;
	b_O_SINP_SIOP_N          = HMRATE(1.0e-9,0,0)*P_O_SINP;
	b_OP_N2O_NOP_NO          = HMRATE((1e-10)*6.3,0,0)*P_OP_N2O;
	b_O_CSP_S_COP          = HMRATE((1e-11)*6.,0,0)*P_O_CSP;
	b_O_HCSP_S_HCOP          = HMRATE((1e-12)*5.,0,0)*P_O_HCSP;
	b_O_HCSP_OCSP_H          = HMRATE((1e-12)*5.,0,0)*P_O_HCSP;
	b_OP_NO2_O2_NOP          = HMRATE((1e-10)*8.3,0,0)*P_OP_NO2;
	b_O_NSP_S_NOP          = HMRATE((1e-10)*6.1,0,0)*P_O_NSP;
	b_NH2P_NH2_NH3P_NH          = HMRATE(1.0e-9,0,0)*P_NH2P_NH2;
	b_NH2_OHP_NH3P_O          = HMRATE(5.0e-10,0,0)*P_NH2_OHP;
	b_NH2P_OH_H2OP_NH          = HMRATE((1e-10)*7.1,0,0)*P_NH2P_OH;
	b_NH2P_NH3_NH4P_NH          = HMRATE((1e-9)*1.61,0,0)*P_NH2P_NH3;
	b_NH2_NH3P_NH4P_NH          = HMRATE(1.0e-11,0,0)*P_NH2_NH3P;
	b_NH2_CH5P_NH3P_CH4          = HMRATE((1e-10)*9.9,0,0)*P_NH2_CH5P;
	b_NH2P_H2O_NH3P_OH          = HMRATE(1.0e-10,0,0)*P_NH2P_H2O;
	b_NH2P_H2O_H3OP_NH          = HMRATE((1e-9)*2.76,0,0)*P_NH2P_H2O;
	b_NH2_H2OP_NH3P_OH          = HMRATE((1e-10)*4.9,0,0)*P_NH2_H2OP;
	b_NH2P_H2O_NH4P_O          = HMRATE((1e-10)*1.45,0,0)*P_NH2P_H2O;
	b_NH2_H3OP_H2O_NH3P          = HMRATE((1e-10)*9.7,0,0)*P_NH2_H3OP;
	b_NH2_HCNP_CN_NH3P          = HMRATE(9.0e-10,0,0)*P_NH2_HCNP;
	b_NH2_COP_HCOP_NH          = HMRATE((1e-10)*4.5,0,0)*P_NH2_COP;
	b_NH2_HCOP_CO_NH3P          = HMRATE((1e-10)*8.9,0,0)*P_NH2_HCOP;
	b_NH2_HNOP_NO_NH3P          = HMRATE((1e-10)*8.8,0,0)*P_NH2_HNOP;
	b_NH2P_O2_HNOP_OH          = HMRATE((1e-11)*2.1,0,0)*P_NH2P_O2;
	b_NH2P_S_HSP_NH          = HMRATE((1e-10)*4.4,0,0)*P_NH2P_S;
	b_CH4P_NH3_NH4P_CH3          = HMRATE((1e-9)*1.15,0,0)*P_CH4P_NH3;
	b_CH4_NH3P_NH4P_CH3          = HMRATE((1e-10)*4.8,0,0)*P_CH4_NH3P;
	b_CH4_N2P_N2_CH2P_H2          = HMRATE((1e-11)*7.,0,0)*P_CH4_N2P;
	b_CH4_N2P_N2_CH3P_H          = HMRATE((1e-10)*9.3,0,0)*P_CH4_N2P;
	b_CH4_HNOP_NO_CH5P          = HMRATE(1.0e-10,0,0)*P_CH4_HNOP;
	b_CH4_SP_HCSP_H2_H          = HMRATE(2.0e-11,0,0)*P_CH4_SP;
	b_CH4_CSP_HCSP_CH3          = HMRATE(5.0e-10,0,0)*P_CH4_CSP;
	b_OHP_NH3_NH4P_O          = HMRATE(1.2e-9,0,0)*P_OHP_NH3;
	b_OH_NH3P_NH4P_O          = HMRATE(7.0e-10,0,0)*P_OH_NH3P;
	b_OHP_CN_HCNP_O          = HMRATE(1.0e-9,0,0)*P_OHP_CN;
	b_OH_HCNP_CN_H2OP          = HMRATE((1e-10)*6.3,0,0)*P_OH_HCNP;
	b_OHP_NO_HNOP_O          = HMRATE((1e-10)*6.11,0,0)*P_OHP_NO;
	b_OH_HNOP_NO_H2OP          = HMRATE(6.2e-10,0,0)*P_OH_HNOP;
	b_OHP_S_HSP_O          = HMRATE((1e-10)*4.3,0,0)*P_OHP_S;
	b_OHP_S_SOP_H          = HMRATE((1e-10)*4.3,0,0)*P_OHP_S;
	b_OH_SP_SOP_H          = HMRATE((1e-10)*6.1,0,0)*P_OH_SP;
	b_NH3P_NH3_NH4P_NH2          = HMRATE((1e-9)*2.2,0,0)*P_NH3P_NH3;
	b_NH3_CH5P_NH4P_CH4          = HMRATE((1e-9)*2.5,0,0)*P_NH3_CH5P;
	b_NH3P_H2O_NH4P_OH          = HMRATE(1.1e-10,0,0)*P_NH3P_H2O;
	b_NH3_H2OP_NH4P_OH          = HMRATE((1e-10)*9.45,0,0)*P_NH3_H2OP;
	b_NH3_H3OP_NH4P_H2O          = HMRATE((1e-9)*2.2,0,0)*P_NH3_H3OP;
	b_NH3_COP_HCOP_NH2          = HMRATE((1e-11)*4.12,0,0)*P_NH3_COP;
	b_NH3_HCOP_CO_NH4P          = HMRATE((1e-9)*2.2,0,0)*P_NH3_HCOP;
	b_NH3_HNOP_NO_NH4P          = HMRATE(1.1e-9,0,0)*P_NH3_HNOP;
	b_NH3_HSP_S_NH4P          = HMRATE((1e-10)*9.75,0,0)*P_NH3_HSP;
	b_NH3_HCSP_CS_NH4P          = HMRATE(2.0e-9,0,0)*P_NH3_HCSP;
	b_CH5P_S_HSP_CH4          = HMRATE(1.3e-9,0,0)*P_CH5P_S;
	b_H2O_CNP_HCNP_OH          = HMRATE((1e-9)*1.6,0,0)*P_H2O_CNP;
	b_H2O_CNP_HCOP_NH          = HMRATE((1e-10)*1.6,0,0)*P_H2O_CNP;
	b_H2O_HCNP_CN_H3OP          = HMRATE((1e-9)*1.8,0,0)*P_H2O_HCNP;
	b_H2O_HNOP_NO_H3OP          = HMRATE((1e-9)*2.3,0,0)*P_H2O_HNOP;
	b_H2OP_S_HSP_OH          = HMRATE((1e-10)*4.3,0,0)*P_H2OP_S;
	b_H2O_HSP_S_H3OP          = HMRATE((1e-10)*7.8,0,0)*P_H2O_HSP;
	b_H3OP_CS_HCSP_H2O          = HMRATE(1.0e-9,0,0)*P_H3OP_CS;
	b_CNP_NO_OCNP_N          = HMRATE((1e-10)*1.9,0,0)*P_CNP_NO;
	b_CN_HNOP_NO_HCNP          = HMRATE((1e-10)*8.7,0,0)*P_CN_HNOP;
	b_CNP_O2_NOP_CO          = HMRATE((1e-11)*8.6,0,0)*P_CNP_O2;
	b_CNP_O2_OCNP_O          = HMRATE((1e-11)*8.6,0,0)*P_CNP_O2;
	b_HCNP_CO_HCOP_CN          = HMRATE((1e-10)*1.4,0,0)*P_HCNP_CO;
	b_HCNP_S_HSP_CN          = HMRATE((1e-10)*5.7,0,0)*P_HCNP_S;
	b_CO_HNOP_NO_HCOP          = HMRATE(1.0e-10,0,0)*P_CO_HNOP;
	b_N2P_OCS_SP_N2_CO          = HMRATE((1e-9)*1.04,0,0)*P_N2P_OCS;
	b_HCOP_S_HSP_CO          = HMRATE((1e-10)*3.3,0,0)*P_HCOP_S;
	b_HCOP_CS_HCSP_CO          = HMRATE(1.2e-9,0,0)*P_HCOP_CS;
	b_HNOP_S_HSP_NO          = HMRATE(1.1e-9,0,0)*P_HNOP_S;
	b_O2_SP_SOP_O          = HMRATE((1e-11)*1.5,0,0)*P_O2_SP;
	b_O2P_S_SOP_O          = HMRATE(5.4e-10,0,0)*P_O2P_S;
	b_O2_CSP_OCSP_O          = HMRATE((1e-10)*1.3,0,0)*P_O2_CSP;
	b_S_SIOP_SO_SIP          = HMRATE(1.0e-9,0,0)*P_S_SIOP;
	b_CP_NH3_NH3P_C          = HMRATE((1e-10)*5.06,0,0)*P_CP_NH3;
	b_C_CNP_CN_CP          = HMRATE(1.1e-10,0,0)*P_C_CNP;
	b_C_N2P_N2_CP          = HMRATE(1.1e-10,0,0)*P_C_N2P;
	b_CP_NO_NOP_C          = HMRATE(5.2e-10,0,0)*P_CP_NO;
	b_CP_SIN_SINP_C          = HMRATE(1.0e-9,0,0)*P_CP_SIN;
	b_C_CSP_CS_CP          = HMRATE((1e-9)*1.6,0,0)*P_C_CSP;
	b_CP_NS_NSP_C          = HMRATE((1e-10)*7.6,0,0)*P_CP_NS;
	b_CP_SO_SOP_C          = HMRATE((1e-10)*2.6,0,0)*P_CP_SO;
	b_CP_OCS_OCSP_C          = HMRATE(4.0e-10,0,0)*P_CP_OCS;
	b_CH_NH2P_NH2_CHP          = HMRATE(3.5e-10,0,0)*P_CH_NH2P;
	b_CHP_NH3_NH3P_CH          = HMRATE((1e-10)*4.59,0,0)*P_CHP_NH3;
	b_CH_CNP_CN_CHP          = HMRATE((1e-10)*6.4,0,0)*P_CH_CNP;
	b_CH_N2P_N2_CHP          = HMRATE((1e-10)*6.3,0,0)*P_CH_N2P;
	b_CHP_NO_NOP_CH          = HMRATE((1e-10)*7.6,0,0)*P_CHP_NO;
	b_NP_NH_NHP_N          = HMRATE((1e-10)*3.7,0,0)*P_NP_NH;
	b_NP_NH2_NH2P_N          = HMRATE(1.0e-9,0,0)*P_NP_NH2;
	b_NP_NH3_NH3P_N          = HMRATE((1e-9)*1.97,0,0)*P_NP_NH3;
	b_NP_CN_CNP_N          = HMRATE(1.1e-9,0,0)*P_NP_CN;
	b_NP_HCN_HCNP_N          = HMRATE(1.2e-9,0,0)*P_NP_HCN;
	b_N_N2P_N2_NP          = HMRATE(1.0e-11,0,0)*P_N_N2P;
	b_NP_NO_NOP_N          = HMRATE((1e-10)*4.51,0,0)*P_NP_NO;
	b_NP_OCS_OCSP_N          = HMRATE((1e-9)*1.02,0,0)*P_NP_OCS;
	b_CH2_NH2P_NH2_CH2P          = HMRATE((1e-10)*4.9,0,0)*P_CH2_NH2P;
	b_CH2_CNP_CN_CH2P          = HMRATE((1e-10)*8.8,0,0)*P_CH2_CNP;
	b_CH2_N2P_N2_CH2P          = HMRATE((1e-10)*8.7,0,0)*P_CH2_N2P;
	b_CH2P_NO_NOP_CH2          = HMRATE((1e-10)*4.2,0,0)*P_CH2P_NO;
	b_NH_OP_O_NHP          = HMRATE((1e-10)*3.6,0,0)*P_NH_OP;
	b_NH_OHP_OH_NHP          = HMRATE((1e-10)*3.6,0,0)*P_NH_OHP;
	b_NHP_NH3_NH3P_NH          = HMRATE((1e-9)*1.8,0,0)*P_NHP_NH3;
	b_NHP_H2O_H2OP_NH          = HMRATE((1e-9)*1.05,0,0)*P_NHP_H2O;
	b_NH_CNP_CN_NHP          = HMRATE((1e-10)*6.5,0,0)*P_NH_CNP;
	b_NH_COP_CO_NHP          = HMRATE((1e-10)*3.2,0,0)*P_NH_COP;
	b_NH_N2P_N2_NHP          = HMRATE((1e-10)*6.5,0,0)*P_NH_N2P;
	b_NHP_NO_NOP_NH          = HMRATE((1e-10)*7.12,0,0)*P_NHP_NO;
	b_NHP_O2_O2P_NH          = HMRATE((1e-10)*4.51,0,0)*P_NHP_O2;
	b_NHP_S_SP_NH          = HMRATE((1e-10)*6.9,0,0)*P_NHP_S;
	b_CH3P_NO_NOP_CH3          = HMRATE(1.0e-9,0,0)*P_CH3P_NO;
	b_OP_NH2_NH2P_O          = HMRATE(1.0e-9,0,0)*P_OP_NH2;
	b_OP_NH3_NH3P_O          = HMRATE(1.2e-9,0,0)*P_OP_NH3;
	b_O_CNP_CN_OP          = HMRATE((1e-11)*6.5,0,0)*P_O_CNP;
	b_O_HCNP_HCN_OP          = HMRATE((1e-11)*6.5,0,0)*P_O_HCNP;
	b_O_N2P_N2_OP          = HMRATE(1.0e-11,0,0)*P_O_N2P;
	b_OP_NO_NOP_O          = HMRATE((1e-12)*1.7,0,0)*P_OP_NO;
	b_OP_OCS_OCSP_O          = HMRATE((1e-10)*6.5,0,0)*P_OP_OCS;
	b_NH2_OHP_OH_NH2P          = HMRATE(5.0e-10,0,0)*P_NH2_OHP;
	b_NH2P_NH3_NH3P_NH2          = HMRATE((1e-10)*6.9,0,0)*P_NH2P_NH3;
	b_NH2_H2OP_H2O_NH2P          = HMRATE((1e-10)*4.9,0,0)*P_NH2_H2OP;
	b_NH2_CNP_CN_NH2P          = HMRATE((1e-10)*9.1,0,0)*P_NH2_CNP;
	b_NH2_COP_CO_NH2P          = HMRATE((1e-10)*4.5,0,0)*P_NH2_COP;
	b_NH2_N2P_N2_NH2P          = HMRATE((1e-10)*8.9,0,0)*P_NH2_N2P;
	b_NH2P_NO_NOP_NH2          = HMRATE(7.0e-10,0,0)*P_NH2P_NO;
	b_NH2_O2P_O2_NH2P          = HMRATE((1e-10)*8.7,0,0)*P_NH2_O2P;
	b_NH2P_S_SP_NH2          = HMRATE((1e-10)*4.4,0,0)*P_NH2P_S;
	b_CH4P_NH3_NH3P_CH4          = HMRATE((1e-9)*1.65,0,0)*P_CH4P_NH3;
	b_CH4P_OCS_OCSP_CH4          = HMRATE((1e-10)*4.2,0,0)*P_CH4P_OCS;
	b_OHP_NH3_NH3P_OH          = HMRATE(1.2e-9,0,0)*P_OHP_NH3;
	b_OH_CNP_CN_OHP          = HMRATE((1e-10)*6.4,0,0)*P_OH_CNP;
	b_OH_N2P_N2_OHP          = HMRATE((1e-10)*6.3,0,0)*P_OH_N2P;
	b_OHP_NO_NOP_OH          = HMRATE((1e-10)*3.59,0,0)*P_OHP_NO;
	b_NH3_H2OP_H2O_NH3P          = HMRATE((1e-9)*2.21,0,0)*P_NH3_H2OP;
	b_NH3_HCNP_HCN_NH3P          = HMRATE((1e-9)*1.68,0,0)*P_NH3_HCNP;
	b_NH3_COP_CO_NH3P          = HMRATE((1e-9)*2.02,0,0)*P_NH3_COP;
	b_NH3_N2P_N2_NH3P          = HMRATE((1e-9)*1.9,0,0)*P_NH3_N2P;
	b_NH3P_SI_SIP_NH3          = HMRATE((1e-9)*1.9,0,0)*P_NH3P_SI;
	b_NH3_O2P_O2_NH3P          = HMRATE(2.0e-9,0,0)*P_NH3_O2P;
	b_NH3_SP_S_NH3P          = HMRATE((1e-9)*1.44,0,0)*P_NH3_SP;
	b_NH3_HSP_HS_NH3P          = HMRATE((1e-10)*5.25,0,0)*P_NH3_HSP;
	b_NH3_SOP_SO_NH3P          = HMRATE(1.3e-9,0,0)*P_NH3_SOP;
	b_H2O_HCNP_HCN_H2OP          = HMRATE((1e-9)*1.8,0,0)*P_H2O_HCNP;
	b_H2O_N2P_N2_H2OP          = HMRATE((1e-9)*2.3,0,0)*P_H2O_N2P;
	b_H2OP_NO_NOP_H2O          = HMRATE((1e-10)*2.7,0,0)*P_H2OP_NO;
	b_CNP_HCN_HCNP_CN          = HMRATE((1e-9)*1.79,0,0)*P_CNP_HCN;
	b_CNP_CO_COP_CN          = HMRATE((1e-10)*6.3,0,0)*P_CNP_CO;
	b_CN_N2P_N2_CNP          = HMRATE(1.0e-10,0,0)*P_CN_N2P;
	b_CNP_NO_NOP_CN          = HMRATE((1e-10)*5.7,0,0)*P_CNP_NO;
	b_CNP_O2_O2P_CN          = HMRATE((1e-10)*2.58,0,0)*P_CNP_O2;
	b_CNP_S_SP_CN          = HMRATE(1.1e-9,0,0)*P_CNP_S;
	b_HCN_COP_CO_HCNP          = HMRATE((1e-9)*3.4,0,0)*P_HCN_COP;
	b_HCN_N2P_N2_HCNP          = HMRATE((1e-10)*3.9,0,0)*P_HCN_N2P;
	b_HCNP_NO_NOP_HCN          = HMRATE((1e-10)*8.1,0,0)*P_HCNP_NO;
	b_HCNP_O2_O2P_HCN          = HMRATE((1e-10)*3.2,0,0)*P_HCNP_O2;
	b_HCNP_S_SP_HCN          = HMRATE((1e-10)*5.7,0,0)*P_HCNP_S;
	b_CO_N2P_N2_COP          = HMRATE((1e-11)*7.4,0,0)*P_CO_N2P;
	b_COP_NO_NOP_CO          = HMRATE((1e-10)*3.3,0,0)*P_COP_NO;
	b_N2P_NO_NOP_N2          = HMRATE((1e-10)*4.4,0,0)*P_N2P_NO;
	b_N2P_O2_O2P_N2          = HMRATE((1e-11)*5.,0,0)*P_N2P_O2;
	b_N2P_S_SP_N2          = HMRATE(1.1e-9,0,0)*P_N2P_S;
	b_SI_NOP_NO_SIP          = HMRATE((1e-9)*1.6,0,0)*P_SI_NOP;
	b_SI_HSP_HS_SIP          = HMRATE((1e-9)*1.4,0,0)*P_SI_HSP;
	b_SI_CSP_CS_SIP          = HMRATE((1e-10)*1.5,0,0)*P_SI_CSP;
	b_NO_HNOP_HNO_NOP          = HMRATE(7.0e-10,0,0)*P_NO_HNOP;
	b_NO_O2P_O2_NOP          = HMRATE((1e-10)*4.5,0,0)*P_NO_O2P;
	b_NO_SP_S_NOP          = HMRATE((1e-10)*3.7,0,0)*P_NO_SP;
	b_NO_HSP_HS_NOP          = HMRATE((1e-10)*4.5,0,0)*P_NO_HSP;
	b_NO_SIOP_SIO_NOP          = HMRATE((1e-10)*7.2,0,0)*P_NO_SIOP;
	b_NO_S2P_S2_NOP          = HMRATE((1e-10)*5.1,0,0)*P_NO_S2P;
	b_O2P_NO2_NO2P_O2          = HMRATE((1e-10)*6.6,0,0)*P_O2P_NO2;
	b_S_HSP_HS_SP          = HMRATE((1e-10)*9.7,0,0)*P_S_HSP;
	b_C_N_CN_NU          = HMRATE((1e-17)*1.,0,0)*P_C_N;
	b_C_S_CS_NU          = HMRATE(4.36E-19,0.22,0)*P_C_S;
	b_CP_S_CSP_NU          = HMRATE(3.07E-19,0.15,0)*P_CP_S;
	b_NP_N_N2P_NU          = HMRATE(3.71E-18,0.24,26.1)*P_NP_N;
	b_CO_S_OCS_NU          = HMRATE((1e-17)*1.6,-1.5,0)*P_CO_S;
	b_CH_NP_N_CHP          = HMRATE((1e-10)*3.6,0,0)*P_CH_NP;
	b_CHP_S_SP_CH          = HMRATE((1e-10)*4.7,0,0)*P_CHP_S;
	b_NP_CH2_CH2P_N          = HMRATE(1.0e-9,0,0)*P_NP_CH2;
	b_NP_CH4_CH4P_N          = HMRATE((1e-11)*2.8,0,0)*P_NP_CH4;
	b_NP_OH_OHP_N          = HMRATE((1e-10)*3.7,0,0)*P_NP_OH;
	b_NP_H2O_H2OP_N          = HMRATE((1e-9)*2.8,0,0)*P_NP_H2O;
	b_NP_CO_COP_N          = HMRATE((1e-10)*8.25,0,0)*P_NP_CO;
	b_NP_O2_O2P_N          = HMRATE((1e-10)*3.11,0,0)*P_NP_O2;
	b_OP_O2_O2P_O          = HMRATE((1e-11)*1.9,0,0)*P_OP_O2;
	b_OHP_S_SP_OH          = HMRATE((1e-10)*4.3,0,0)*P_OHP_S;
	b_H2OP_S_SP_H2O          = HMRATE((1e-10)*4.3,0,0)*P_H2OP_S;
	b_COP_S_SP_CO          = HMRATE(1.1e-9,0,0)*P_COP_S;
	b_SI_SP_S_SIP          = HMRATE((1e-9)*1.6,0,0)*P_SI_SP;
	b_O2P_S_SP_O2          = HMRATE(5.4e-10,0,0)*P_O2P_S;
	O_CCl_Cl_CO_1 = HMRATE(0.0000000000996,0,0)*CCl;
	O_CCl_ClO_C_1 = HMRATE(0.000000000138,0,16050)*CCl;
	O_ClO_Cl_O2_1 = HMRATE(0.000000000038,0,0)*ClO;
	H2_ClP_HClP_H = HMRATE(0.000000001,0,0)*H2g;
	H2_HClP_H2ClP_H = HMRATE(0.0000000013,0,0)*H2g;
	H3P_Cl_HClP_H2 = HMRATE(0.000000001,0,0)*H3p;
	H3P_HCl_H2ClP_H2 = HMRATE(0.0000000038,0,0)*H3p;
	HeP_HCl_ClP_He_H = HMRATE(0.0000000033,0,0)*HeII;
	HeP_CCl_Cl_CP_He = HMRATE(0.0000000033,0,0)*HeII;
	HeP_ClO_ClP_O_He = HMRATE(0.000000001,0,0)*HeII;
	CP_HCl_CClP_H_1 = HMRATE(0.0000000011,0,0)*HCl;
	CH3P_HCl_H2CClP_H2_1 = HMRATE(0.0000000013,0,0)*HCl;
	H2O_H2ClP_HCl_H3OP_1 = HMRATE(0.000000002,0,0)*H2ClP;
	CO_H2ClP_HCl_HCOP_1 = HMRATE(0.00000000078,0,0)*H2ClP;
	HP_HCl_HClP_H = HMRATE(0.0000000033,1,0)*HII;
	CP_CCl_CClP_C_1 = HMRATE(0.000000001,0,0)*CCl;
	CP_ClO_ClOP_C_1 = HMRATE(0.000000001,0,0)*ClO;
	O2_ClP_Cl_O2P_1 = HMRATE(0.00000000046,0,0)*ClP;
	HClP_e_Cl_H = HMRATE(0.0000003,-0.5,0)*ele;
	H2ClP_e_Cl_H_H = HMRATE(0.00000027,-0.5,0)*ele;
	H2ClP_e_HCl_H = HMRATE(0.00000003,-0.5,0)*ele;
	CClP_e_Cl_C = HMRATE(0.0000003,-0.5,0)*ele;
	H2CClP_e_CCl_H_H = HMRATE(0.0000003,-0.5,0)*ele;
	ClOP_e_Cl_O = HMRATE(0.0000002,-0.5,0)*ele;
	HCl_NU_Cl_H = HMRATE(0.00000000011,0,0)*TH85_field;
	CCl_NU_Cl_C = HMRATE(0.0000000001,0,0)*TH85_field;
	ClO_NU_Cl_O = HMRATE(0.0000000001,0,0)*TH85_field;
	HCl_crnu_Cl_H = crnu_rate*2*610;
	CCl_crnu_Cl_C = crnu_rate*2*500;
	ClO_crnu_Cl_O = crnu_rate*2*500;
	O_CCl_Cl_CO_2 = HMRATE(0.0000000000996,0,0)*ATO;
	O_CCl_ClO_C_2 = HMRATE(0.000000000138,0,16050)*ATO;
	O_ClO_Cl_O2_2 = HMRATE(0.000000000038,0,0)*ATO;
	CP_HCl_CClP_H_2 = HMRATE(0.0000000011,0,0)*CP;
	CH3P_HCl_H2CClP_H2_2 = HMRATE(0.0000000013,0,0)*CH3P;
	H2O_H2ClP_HCl_H3OP_2 = HMRATE(0.000000002,0,0)*H2O;
	CO_H2ClP_HCl_HCOP_2 = HMRATE(0.00000000078,0,0)*CO;
	CP_CCl_CClP_C_2 = HMRATE(0.000000001,0,0)*CP;
	CP_ClO_ClOP_C_2 = HMRATE(0.000000001,0,0)*CP;
	O2_ClP_Cl_O2P_2 = HMRATE(0.00000000046,0,0)*O2;
	P_O_CCl = ATO*CCl;
	P_O_ClO = ATO*ClO;
	P_CP_HCl = CP*HCl;
	P_CH3P_HCl = CH3P*HCl;
	P_H2O_H2ClP = H2O*H2ClP;
	P_CO_H2ClP = CO*H2ClP;
	P_CP_CCl = CP*CCl;
	P_CP_ClO = CP*ClO;
	P_O2_ClP = O2*ClP;
	b_O_CCl_Cl_CO = HMRATE(0.0000000000996,0,0)*P_O_CCl;
	b_O_CCl_ClO_C = HMRATE(0.000000000138,0,16050)*P_O_CCl;
	b_O_ClO_Cl_O2 = HMRATE(0.000000000038,0,0)*P_O_ClO;
	b_CP_HCl_CClP_H = HMRATE(0.0000000011,0,0)*P_CP_HCl;
	b_CH3P_HCl_H2CClP_H2 = HMRATE(0.0000000013,0,0)*P_CH3P_HCl;
	b_H2O_H2ClP_HCl_H3OP = HMRATE(0.000000002,0,0)*P_H2O_H2ClP;
	b_CO_H2ClP_HCl_HCOP = HMRATE(0.00000000078,0,0)*P_CO_H2ClP;
	b_CP_CCl_CClP_C = HMRATE(0.000000001,0,0)*P_CP_CCl;
	b_CP_ClO_ClOP_C = HMRATE(0.000000001,0,0)*P_CP_ClO;
	b_O2_ClP_Cl_O2P = HMRATE(0.00000000046,0,0)*P_O2_ClP;

	/* This reaction was not in UMIST, but is in Zsargo et al. 2003 as an 
		important reaction in the equilibrium chemistry of CH */

	CP_NH_CHP_N_1 = 5e-10*CP;
	CP_NH_CHP_N_2 = 5e-10*NH;
	b_CP_NH_CHP_N = 5e-10*P_CP_NH;


	/* This turns on the diffuse cloud chemical rates from Tables 6 & 8 of
	>> refer Federman, S. R. & Zsargo, J. 2003, ApJ, 589, 319
	Without the "set rates federman" command, the default values are
	from UMIST */
	if(co.lgFederman)
	{
		e_COP_O_C = HMRATE(1.0e-7,-0.46,0)*ele;
		e_HCOP_CO_H = HMRATE(2.4e-7,-0.69,0)*ele;
		H2_COP_HCOP_H = HMRATE((1e-9)*1.4,0,0)*H2g;
		e_CH2P_CH_H  = HMRATE(1.25e-7,-0.5,0)*ele;
		e_CH3P_CH_H2 = HMRATE(2.45e-7,-0.5,0)*ele;
		CP_CH_CHP_C_1 = HMRATE(1.5e-9,0,0)*CP;
		CP_CH_CHP_C_2 = HMRATE(1.5e-9,0,0)*CH;
		b_CP_CH_CHP_C = HMRATE(1.5e-9,0,0)*P_CP_CH;
		CP_CH_C2P_H_2 = HMRATE(5.4e-10,0,0)*CP;
		CP_CH_C2P_H_1 = HMRATE(5.4e-10,0,0)*CH;
		b_CP_CH_C2P_H = HMRATE(5.4e-10,0,0)*P_CP_CH;
		O_CH_CO_H_1	= HMRATE(9.5e-11,0.5,0)*ATO;
		O_CH_CO_H_2	= HMRATE(9.5e-11,0.5,0)*CH;
		b_O_CH_CO_H	= HMRATE(9.5e-11,0.5,0)*P_O_CH;
		H2_CH2P_CH3P_H = HMRATE((1e-9)*1.2,0,0)*H2g;
		e_CH2P_C_H2 = HMRATE(1.25e-7,-0.5,0)*ele;
		O_CH3P_HCOP_H2_1 = HMRATE(4.4e-10,0,0)*ATO;
		O_CH3P_HCOP_H2_2 = HMRATE(4.4e-10,0,0)*CH3P;
		b_O_CH3P_HCOP_H2 = HMRATE(4.4e-10,0,0)*P_O_CH3P;
		e_CH3P_CH2_H = HMRATE(7.0e-8,-0.5,0)*ele;
		e_CH3P_C_H2_H = HMRATE(3.5e-8,-0.5,0)*ele;
		H2_CP_CH2P_nu = HMRATE(1.0e-16,0,0)*H2g;
		H_CH_C_H2 = HMRATE((1e-11)*5.0,0.5,0)*HI;
		N_C2_CN_C_1 = HMRATE((1e-11)*1.7,0.5,0)*C2;
		N_C2_CN_C_2 = HMRATE((1e-11)*1.7,0.5,0)*ATN;
		b_N_C2_CN_C = HMRATE((1e-11)*1.7,0.5,0)*P_N_C2;
		CH_N_CN_H_1 = HMRATE((2e-11),0.5,0)*CH;
		CH_N_CN_H_2 = HMRATE((2e-11),0.5,0)*ATN;
		b_CH_N_CN_H = HMRATE((2e-11),0.5,0)*P_CH_N;
		nu_CH_CHP_e = HMRATE((1e-10)*4.2,0,0)*(TH85_field);
		nu_CH_C_H = HMRATE((1e-9)*1.3,0,0)*(TH85_field);
		C2_nu_C_C = HMRATE((1e-10)*2.0,0,0)*(TH85_field);
		CN_NU_N_C = HMRATE(1.0e-10,0,0)*(TH85_field);
		CP_NH_CNP_H_1 = HMRATE((1e-10)*5.0,0,0)*CP;
		CP_NH_CNP_H_2 = HMRATE((1e-10)*5.0,0,0)*NH;
		b_CP_NH_CNP_H = HMRATE((1e-10)*5.0,0,0)*P_CP_NH;




	}

	/* zero out molecular charge transfer rates */
	for(nelem=ipLITHIUM; nelem < LIMELM; ++nelem)
	{
		for(ion=0; ion < nelem+2; ++ion)
		{
			/*zero out the arrays */				
			mole.sink[nelem][ion]     = 0.;
			mole.source[nelem][ion]   = 0.;

			for(ion2=0; ion2< nelem+2; ++ion2)
			{
				mole.xMoleChTrRate[nelem][ion][ion2] = 0.;
			}
		}
	}


	/* >>refer	mol	H	CH	=>	C	H	H	UMIST	*/
	
	c[ipCH][ipCH] -= H_CH_C_H_H;
	c[ipCH][ipATC] += H_CH_C_H_H;

	/* >>refer	mol	H	OH	=>	O	H	H	UMIST	*/

	c[ipOH][ipOH] -= H_OH_O_H_H;
	c[ipOH][ipATO] += H_OH_O_H_H;

	/* >>refer	mol	H	H2O	=>	OH	H	H	UMIST	*/

	c[ipH2O][ipH2O] -= H_H2O_OH_H_H;
	c[ipH2O][ipOH] += H_H2O_OH_H_H;
	
	/* >>refer	mol	H	CO+	=>	CO	H+	UMIST	*/
	
	c[ipCOP][ipCOP] -= H_COP_CO_HP;
	c[ipCOP][ipCO] += H_COP_CO_HP;
	
	/* >>refer	mol	H	CH	=>	C	H2	UMIST	*/
	
	c[ipCH][ipCH] -= H_CH_C_H2;
	c[ipCH][ipATC] += H_CH_C_H2;
	
	/* >>refer	mol	H	CH+	=>	C+	H2	UMIST	*/
	
	c[ipCHP][ipCHP] -= H_CHP_CP_H2;
	c[ipCHP][ipCP]  += H_CHP_CP_H2;
	
	/* >>refer	mol	H	CH2	=>	CH	H2	UMIST	*/
		
	c[ipCH2][ipCH2] -= H_CH2_CH_H2;
	c[ipCH2][ipCH] += H_CH2_CH_H2;
	
	/* >>refer	mol	H	CH3+	=>	CH2+	H2	UMIST	*/
	
	c[ipCH3P][ipCH3P] -= H_CH3P_CH2P_H2;
	c[ipCH3P][ipCH2P] += H_CH3P_CH2P_H2;
	
	/* >>refer	mol	H	OH	=>	O	H2	UMIST	*/
		
	c[ipOH][ipOH] -= H_OH_O_H2;
	c[ipOH][ipATO] += H_OH_O_H2;
	
	/* >>refer	mol	H	H2O	=>	OH	H2	UMIST	*/
		
	c[ipH2O][ipH2O] -= H_H2O_OH_H2;
	c[ipH2O][ipOH]  += H_H2O_OH_H2;
	
	/* >>refer	mol	H	O2	=>	OH	O	UMIST	*/
		
	c[ipO2][ipO2] -= H_O2_OH_O;
	c[ipO2][ipOH] += H_O2_OH_O;
	c[ipO2][ipATO] += H_O2_OH_O;
	
	/* >>refer	mol	H	O2	=>	O	O	H	UMIST	*/
		
	c[ipO2][ipO2] -= H_O2_O_O_H;
	c[ipO2][ipATO] += 2*H_O2_O_O_H;
	
	/* >>refer	mol	H	C	=>	CH	PHOTON	UMIST	*/
	
	c[ipATC][ipCH] += H_C_CH_nu;
	c[ipATC][ipATC] -= H_C_CH_nu;
	
	/* >>refer	mol	H	C+	=>	CH+	PHOTON	UMIST	*/
		
	c[ipCP][ipCP]  -= H_CP_CHP_nu;
	c[ipCP][ipCHP] += H_CP_CHP_nu;
	
	/* >>refer	mol	H	OH	=>	H2O	PHOTON	UMIST	*/
		
	c[ipOH][ipOH]  -= H_OH_H2O_nu;
	c[ipOH][ipH2O] += H_OH_H2O_nu;
	
	/* >>refer	mol	H	O	=>	OH	PHOTON	UMIST	*/
		
	c[ipATO][ipOH] += H_O_OH_nu;
	c[ipATO][ipATO] -= H_O_OH_nu;
	
	/* >>refer	mol	H-	CH	=>	CH2	e	UMIST	*/
	
	c[ipCH][ipCH]  -= Hminus_CH_CH2_e;
	c[ipCH][ipCH2] += Hminus_CH_CH2_e;
	
	/* >>refer	mol	H-	C	=>	CH	e	UMIST	*/
	
	c[ipATC][ipCH]  += Hminus_CH_CH2_e;
	c[ipATC][ipATC]  -= Hminus_CH_CH2_e;
	
	/* >>refer	mol	H-	OH	=>	H2O	e	UMIST	*/
		
	c[ipOH][ipOH]  -= Hminus_OH_H2O_e;
	c[ipOH][ipH2O] += Hminus_OH_H2O_e;
	
	/* >>refer	mol	H-	O	=>	OH	e	UMIST	*/
		
	c[ipATO][ipOH] += Hminus_O_OH_e;
	c[ipATO][ipATO] -= Hminus_O_OH_e;
		
	/* >>refer	mol	H-	HCO+	=>	CO	H2	UMIST	*/
		
	c[ipHCOP][ipHCOP] -= Hminus_HCOP_CO_H2;
	c[ipHCOP][ipCO]   += Hminus_HCOP_CO_H2;
	
	/* >>refer	mol	H-	H3O+	=>	H2O	H2	UMIST	*/
		
	c[ipH3OP][ipH3OP] -= Hminus_H3OP_H2O_H2;
	c[ipH3OP][ipH2O] += Hminus_H3OP_H2O_H2;
	
	/* >>refer	mol	H-	H3O+	=>	OH	H2	H	UMIST	*/
		
	c[ipH3OP][ipH3OP] -= Hminus_H3OP_OH_H2_H;
	c[ipH3OP][ipOH] += Hminus_H3OP_OH_H2_H;
	
	/* >>refer	mol	H+	CH	=>	CH+	H	UMIST	*/
		
	c[ipCH][ipCH] -= HP_CH_CHP_H;
	c[ipCH][ipCHP] += HP_CH_CHP_H;
	
	/* >>refer	mol	H+	CH2	=>	CH2+	H	UMIST	*/
		
	c[ipCH2][ipCH2] -= HP_CH2_CH2P_H;
	c[ipCH2][ipCH2P] += HP_CH2_CH2P_H;
	
	/* >>refer	mol	H+	H2O	=>	H2O+	H	UMIST	*/
		
	c[ipH2O][ipH2O] -= HP_H2O_H2OP_H;
	c[ipH2O][ipH2OP] += HP_H2O_H2OP_H;
	
	/* >>refer	mol	H+	O2	=>	O2+	H	UMIST	*/
	
	c[ipO2][ipO2] -= HP_O2_O2P_H;
	c[ipO2][ipO2P] += HP_O2_O2P_H;
	
	/* >>refer	mol	H+	OH	=>	OH+	H	UMIST	*/
		
	c[ipOH][ipOH] -= HP_OH_OHP_H;
	c[ipOH][ipOHP] += HP_OH_OHP_H;
	
	/* >>refer	mol	H+	SiO	=>	SiO+	H	UMIST	*/
		
	c[ipSIO][ipSIO] -= HP_SiO_SiOP_H;
	c[ipSIO][ipSIOP] += HP_SiO_SiOP_H;
	
	/* >>refer	mol	H+	CH2	=>	CH+	H2	UMIST	*/
		
	c[ipCH2][ipCH2] -= HP_CH2_CHP_H2;
	c[ipCH2][ipCHP] += HP_CH2_CHP_H2;
	
	/* >>refer	mol	H+	SiH	=>	Si+	H2	UMIST	*/
		
	c[ipSIH][ipSIH] -= HP_SiH_SiP_H2;
	c[ipSIH][ipSIP] += HP_SiH_SiP_H2;
	
	/* >>refer	mol	H2	C	=>	CH	H	UMIST	*/
	
	c[ipATC][ipCH] += H2_C_CH_H;
	c[ipATC][ipATC] -= H2_C_CH_H;
	
	/* >>refer	mol	H2	C+	=>	CH+	H	UMIST	*/
		
	c[ipCP][ipCP]  -= H2_CP_CHP_H;
	c[ipCP][ipCHP] += H2_CP_CHP_H;
	
	/* >>refer	mol	H2	CH	=>	CH2	H	UMIST	*/
		
	c[ipCH][ipCH] -= H2_CH_CH2_H;
	c[ipCH][ipCH2] += H2_CH_CH2_H;
	
	/* >>refer	mol	H2	CH+	=>	CH2+	H	UMIST	*/
		
	c[ipCHP][ipCHP] -= H2_CHP_CH2P_H;
	c[ipCHP][ipCH2P] += H2_CHP_CH2P_H;
	
	/* >>refer	mol	H2	OH	=>	H2O	H	UMIST	*/
		
	c[ipOH][ipOH] -= H2_OH_H2O_H;
	c[ipOH][ipH2O] += H2_OH_H2O_H;
	
	/* >>refer	mol	H2	OH+	=>	H2O+	H	UMIST	*/
		
	c[ipOHP][ipOHP] -= H2_OHP_H2OP_H;
	c[ipOHP][ipH2OP] += H2_OHP_H2OP_H;
	
	/* >>refer	mol	H2	H2O+	=>	H3O+	H	UMIST	*/
		
	c[ipH2OP][ipH2OP] -= H2_H2OP_H3OP_H;
	c[ipH2OP][ipH3OP] += H2_H2OP_H3OP_H;
	
	/* >>refer	mol	H2	CO+	=>	HCO+	H	UMIST	*/
		
	c[ipCOP][ipCOP] -= H2_COP_HCOP_H;
	c[ipCOP][ipHCOP] += H2_COP_HCOP_H;
	
	/* >>refer	mol	H2	O	=>	OH	H	UMIST	*/
		
	c[ipATO][ipOH] += H2_O_OH_H;
	c[ipATO][ipATO] -= H2_O_OH_H;

	/* >>refer	mol	H2	CH2+	=>	CH3+	H	*/

	c[ipCH2P][ipCH2P] -= H2_CH2P_CH3P_H;
	c[ipCH2P][ipCH3P] += H2_CH2P_CH3P_H;
	
	/* >>refer	mol	H2	O+	=>	OH+	H	UMIST	*/
		
	c[ipOP][ipOP]  -= H2_OP_OHP_H;
	c[ipOP][ipOHP] += H2_OP_OHP_H;
	
	/* >>refer	mol	H2	SiO+	=>	SiOH+	H	UMIST	*/
	
	c[ipSIOP][ipSIOP] -= H2_SiOP_SiOHP_H;
	c[ipSIOP][ipHOSIP] += H2_SiOP_SiOHP_H;
	
	/* >>refer	mol	H2	CH	=>	C	H2	H	UMIST	*/
		
	c[ipCH][ipCH] -= H2_CH_C_H2_H;
	c[ipCH][ipATC] += H2_CH_C_H2_H;
	
	/* >>refer	mol	H2	OH	=>	O	H2	H	UMIST	*/
	
	c[ipOH][ipOH] -= H2_OH_O_H2_H;
	c[ipOH][ipATO] += H2_OH_O_H2_H;
	
	/* >>refer	mol	H2	H2O	=>	OH	H2	H	UMIST	*/
		
	c[ipH2O][ipH2O] -= H2_H2O_OH_H2_H;
	c[ipH2O][ipOH] += H2_H2O_OH_H2_H;
	
	/* >>refer	mol	H2	O2	=>	O	O	H2	UMIST	*/
		
	c[ipO2][ipO2] -= H2_O2_O_O_H2;
	c[ipO2][ipATO] += 2*H2_O2_O_O_H2;
	
	/* >>refer	mol	H2	O2	=>	OH	OH	UMIST	*/
	
	/**********************************************************************
	*                                                                     *
	*  A note here about the stoic. factors.  For a reaction of the type: *	
	*                                                                     *
	*   aA + bB => cC + dD                                                *
	*                                                                     *	
	*  The rates are related by:                                          *
	*                                                                     *
	*  (1/a)*D[A]/Dt = (1/b)*D[B]/Dt = - (1/c)*D[C]/Dt =  - (1/d)*D[D]/Dt *
	*                                                                     *
	*  (As a reference, look at most Physical Chemistry books)		      *
	*                                                                     *
	*                                                                     *
	***********************************************************************/

	c[ipO2][ipO2] -= H2_O2_OH_OH;
	c[ipO2][ipOH] += 2*H2_O2_OH_OH;
	
	/* >>refer	mol	H2	C	=>	CH2	PHOTON	UMIST	*/
		
	c[ipATC][ipCH2] += H2_C_CH2_nu;
	c[ipATC][ipATC] -= H2_C_CH2_nu;
	
	/* >>refer	mol	H2	C+	=>	CH2+	PHOTON	UMIST	*/
		
	c[ipCP][ipCP]   -= H2_CP_CH2P_nu;
	c[ipCP][ipCH2P] += H2_CP_CH2P_nu;
	
	/* >>refer	mol	H2	Si+	=>	SiH2+	PHOTON	UMIST	*/
		
	c[ipSIP][ipSIP] -= H2_SiP_SiH2P_nu;
	c[ipSIP][ipSIH2P] += H2_SiP_SiH2P_nu;
	
	/* >>refer	mol	H2+	C	=>	CH+	H	UMIST	*/
		
	c[ipATC][ipCHP] += H2P_C_CHP_H;
	c[ipATC][ipATC] -= H2P_C_CHP_H;
	
	/* >>refer	mol	H2+	CH	=>	CH2+	H	UMIST	*/
		
	c[ipCH][ipCH] -= H2P_CH_CH2P_H;
	c[ipCH][ipCH2P] += H2P_CH_CH2P_H;
	
	/* >>refer	mol	H2+	CH2	=>	CH3+	H	UMIST	*/
	
	c[ipCH2][ipCH2] -= H2P_CH2_CH3P_H;
	c[ipCH2][ipCH3P] += H2P_CH2_CH3P_H;
	
	/* >>refer	mol	H2+	OH	=>	H2O+	H	UMIST	*/
		
	c[ipOH][ipOH] -= H2P_OH_H2OP_H;
	c[ipOH][ipH2OP] += H2P_OH_H2OP_H;
	
	/* >>refer	mol	H2+	H2O	=>	H3O+	H	UMIST	*/
		
	c[ipH2O][ipH2O] -= H2P_H2O_H3OP_H;
	c[ipH2O][ipH3OP] += H2P_H2O_H3OP_H;
	
	/* >>refer	mol	H2+	CO	=>	HCO+	H	UMIST	*/
		
	c[ipCO][ipCO] -= H2P_CO_HCOP_H;
	c[ipCO][ipHCOP] += H2P_CO_HCOP_H;
	
	/* >>refer	mol	H2+	O	=>	OH+	H	UMIST	*/
		
	c[ipATO][ipOHP] += H2P_O_OHP_H;
	c[ipATO][ipATO] -= H2P_O_OHP_H;
	
	/* >>refer	mol	H2+	CH	=>	CH+	H2	UMIST	*/
		
	c[ipCH][ipCH] -= H2P_CH_CHP_H2;
	c[ipCH][ipCHP] += H2P_CH_CHP_H2;
	
	/* >>refer	mol	H2+	CH2	=>	CH2+	H2	UMIST	*/
		
	c[ipCH2][ipCH2] -= H2P_CH2_CH2P_H2;
	c[ipCH2][ipCH2P] += H2P_CH2_CH2P_H2;
	
	/* >>refer	mol	H2+	CO	=>	CO+	H2	UMIST	*/
		
	c[ipCO][ipCO] -= H2P_CO_COP_H2;
	c[ipCO][ipCOP] += H2P_CO_COP_H2;
	
	/* >>refer	mol	H2+	H2O	=>	H2O+	H2	UMIST	*/
		
	c[ipH2O][ipH2O] -= H2P_H2O_H2OP_H2;
	c[ipH2O][ipH2OP] += H2P_H2O_H2OP_H2;
	
	/* >>refer	mol	H2+	O2	=>	O2+	H2	UMIST	*/
		
	c[ipO2][ipO2] -= H2P_O2_O2P_H2;
	c[ipO2][ipO2P] += H2P_O2_O2P_H2;
	
	/* >>refer	mol	H2+	OH	=>	OH+	H2	UMIST	*/
		
	c[ipOH][ipOH] -= H2P_OH_OHP_H2;
	c[ipOH][ipOHP] += H2P_OH_OHP_H2;
	
	/* >>refer	mol	H3+	C	=>	CH+	H2	UMIST	*/
		
	c[ipATC][ipCHP] += H3P_C_CHP_H2;
	c[ipATC][ipATC] -= H3P_C_CHP_H2;
	
	/* >>refer	mol	H3+	CH	=>	CH2+	H2	UMIST	*/
		
	c[ipCH][ipCH] -= H3P_CH_CH2P_H2;
	c[ipCH][ipCH2P] += H3P_CH_CH2P_H2;
	
	/* >>refer	mol	H3+	CH2	=>	CH3P	H2	UMIST	*/
		
	c[ipCH2][ipCH2] -= H3P_CH2_CH3P_H2;
	c[ipCH2][ipCH3P] += H3P_CH2_CH3P_H2;
	
	/* >>refer	mol	H3+	OH	=>	H2O+	H2	UMIST	*/
		
	c[ipOH][ipOH] -= H3P_OH_H2OP_H2;
	c[ipOH][ipH2OP] += H3P_OH_H2OP_H2;
	
	/* >>refer	mol	H3+	H2O	=>	H3O+	H2	UMIST	*/
		
	c[ipH2O][ipH2O] -= H3P_H2O_H3OP_H2;
	c[ipH2O][ipH3OP] += H3P_H2O_H3OP_H2;
	
	/* >>refer	mol	H3+	CO	=>	HCO+	H2	UMIST	*/
		
	c[ipCO][ipCO] -= H3P_CO_HCOP_H2;
	c[ipCO][ipHCOP] += H3P_CO_HCOP_H2;
	
	/* >>refer	mol	H3+	O	=>	OH+	H2	UMIST	*/
	
	c[ipATO][ipOHP] += H3P_O_OHP_H2;
	c[ipATO][ipATO] -= H3P_O_OHP_H2;
	
	/* >>refer	mol	H3+	SiH	=>	SiH2+	H2	UMIST	*/
		
	c[ipSIH][ipSIH] -= H3P_SiH_SiH2P_H2;
	c[ipSIH][ipSIH2P] += H3P_SiH_SiH2P_H2;
	
	/* >>refer	mol	H3+	SiO	=>	SiOH+	H2	UMIST	*/
		
	c[ipSIO][ipSIO] -= H3P_SiO_SiOHP_H2;
	c[ipSIO][ipHOSIP] += H3P_SiO_SiOHP_H2;
	
	/* >>refer	mol	He+	CO	=>	O+	C	He	UMIST	*/
		
	c[ipCO][ipCO] -= HeP_CO_OP_C_He;
	c[ipCO][ipOP] += HeP_CO_OP_C_He;
	c[ipCO][ipATC] += HeP_CO_OP_C_He;
	
	/* >>refer	mol	He+	CO	=>	O	C+	He	UMIST	*/
		
	c[ipCO][ipCO] -= HeP_CO_O_CP_He;
	c[ipCO][ipCP] += HeP_CO_O_CP_He;
	c[ipCO][ipATO] += HeP_CO_O_CP_He;

	co.rate_co_hep_o_cp_he = (float)(HMRATE((1e-9)*1.6,0,0));
	
	/* >>refer	mol	He+	CH	=>	CH+	He	UMIST	*/
		
	c[ipCH][ipCH] -= HeP_CH_CHP_He;
	c[ipCH][ipCHP] += HeP_CH_CHP_He;
	
	/* >>refer	mol	He+	H2O	=>	H2O+	He	UMIST	*/
		
	c[ipH2O][ipH2O] -= HeP_H2O_H2OP_He;
	c[ipH2O][ipH2OP] += HeP_H2O_H2OP_He;
	
	/* >>refer	mol	He+	O2	=>	O2+	He	UMIST	*/
		
	c[ipO2][ipO2] -= HeP_O2_O2P_He;
	c[ipO2][ipO2P] += HeP_O2_O2P_He;
	
	/* >>refer	mol	He+	CH	=>	C+	H	He	UMIST	*/
		
	c[ipCH][ipCH] -= HeP_CH_CP_He_H;
	c[ipCH][ipCP] += HeP_CH_CP_He_H;
	
	/* >>refer	mol	He+	CH2	=>	CH+	H	He	UMIST	*/
		
	c[ipCH2][ipCH2] -= HeP_CH2_CHP_He_H;
	c[ipCH2][ipCHP] += HeP_CH2_CHP_He_H;
	
	/* >>refer	mol	He+	OH	=>	O+	H	He	UMIST	*/
		
	c[ipOH][ipOH] -= HeP_OH_OP_He_H;
	c[ipOH][ipOP] += HeP_OH_OP_He_H;
	
	/* >>refer	mol	He+	H2O	=>	OH+	H	He	UMIST	*/
		
	c[ipH2O][ipH2O] -= HeP_H2O_OHP_He_H;
	c[ipH2O][ipOHP] += HeP_H2O_OHP_He_H;
	
	/* >>refer	mol	He+	SiH	=>	Si+	H	He	UMIST	*/
		
	c[ipSIH][ipSIH] -= HeP_SiH_SiP_He_H;
	c[ipSIH][ipSIP] += HeP_SiH_SiP_He_H;
	
	/* >>refer	mol	He+	H2O	=>	OH	H+	He	UMIST	*/
		
	c[ipH2O][ipH2O] -= HeP_H2O_OH_He_HP;
	c[ipH2O][ipOH] += HeP_H2O_OH_He_HP;
		
	/* >>refer	mol	He+	CH2	=>	C+	H2	He	UMIST	*/
	
	c[ipCH2][ipCH2] -= HeP_CH2_CP_He_H2;
	c[ipCH2][ipCP]  += HeP_CH2_CP_He_H2;
	
	/* >>refer	mol	He+	O2	=>	O+	O	He	UMIST	*/
		
	c[ipO2][ipO2] -= HeP_O2_OP_O_He;
	c[ipO2][ipOP] += HeP_O2_OP_O_He;
	c[ipO2][ipATO] += HeP_O2_OP_O_He;
	
	/* >>refer	mol	He+	SiO	=>	Si+	O	He	UMIST	*/
		
	c[ipSIO][ipSIO] -= HeP_SiO_SiP_O_He;
	c[ipSIO][ipSIP] += HeP_SiO_SiP_O_He;
	c[ipSIO][ipATO] += HeP_SiO_SiP_O_He;
	
	/* >>refer	mol	He+	SiO	=>	Si	O+	He	UMIST	*/
		
	c[ipSIO][ipSIO] -= HeP_SiO_Si_OP_He;
	c[ipSIO][ipOP]  += HeP_SiO_Si_OP_He;
	c[ipSIO][ipATSI]  += HeP_SiO_Si_OP_He;
	
	/* >>refer	mol	cr	CH	=>	C	H	UMIST	*/
		
	c[ipCH][ipCH] -= crnu_CH_C_H;
	c[ipCH][ipATC] += crnu_CH_C_H;
	
	/* >>refer	mol	cr	CH+	=>	C+	H	UMIST	*/
		
	c[ipCHP][ipCHP] -= crnu_CHP_CP_H;
	c[ipCHP][ipCP]  += crnu_CHP_CP_H;
	
	/* >>refer	mol	cr	CO	=>	C	O	UMIST	*/
		
	c[ipCO][ipCO] -= crnu_CO_O_C;
	c[ipCO][ipATC] += crnu_CO_O_C;
	c[ipCO][ipATO] += crnu_CO_O_C;
	
	/* >>refer	mol	cr	H2O	=>	OH	H	UMIST	*/
		
	c[ipH2O][ipH2O] -= crnu_H2O_OH_H;
	c[ipH2O][ipOH] += crnu_H2O_OH_H;
	
	/* >>refer	mol	cr	O2	=>	O2+	e	UMIST	*/
		
	c[ipO2][ipO2] -= crnu_O2_O2P_e;
	c[ipO2][ipO2P] += crnu_O2_O2P_e;
	
	/* >>refer	mol	cr	O2	=>	O	O	UMIST	*/
		
	c[ipO2][ipO2] -= crnu_O2_O_O;
	c[ipO2][ipATO] += 2*crnu_O2_O_O;
	
	/* >>refer	mol	cr	OH	=>	O	H	UMIST	*/
		
	c[ipOH][ipOH] -= crnu_OH_O_H;
	c[ipOH][ipATO] += crnu_OH_O_H;
	
	/* >>refer	mol	cr	SiH	=>	Si	H	UMIST	*/
		
	c[ipSIH][ipSIH] -= crnu_SiH_Si_H;
	c[ipSIH][ipATSI] += crnu_SiH_Si_H;
	
	/* >>refer	mol	cr	SiO	=>	Si	O	UMIST	*/
		
	c[ipSIO][ipSIO] -= crnu_SiO_Si_O;
	c[ipSIO][ipATSI] += crnu_SiO_Si_O;
	c[ipSIO][ipATO] += crnu_SiO_Si_O;
	
	/* >>refer	mol	CRP	CO	=>	CO+	e	UMIST	*/
		
	c[ipCO][ipCO] -= CRP_CO_COP_e;
	c[ipCO][ipCOP] += CRP_CO_COP_e;
	
	/* >>refer	mol	PHOTON	CH	=>	CH+	e	UMIST	*/
		
	c[ipCH][ipCH] -= nu_CH_CHP_e;
	c[ipCH][ipCHP] += nu_CH_CHP_e;
	
	/* >>refer	mol	PHOTON	CH	=>	C	H	UMIST	*/
		
	c[ipCH][ipCH] -= nu_CH_C_H;
	c[ipCH][ipATC] += nu_CH_C_H;
	
	/* >>refer	mol	PHOTON	CH+	=>	C+	H	UMIST	*/
		
	c[ipCHP][ipCHP] -= nu_CHP_CP_H;
	c[ipCHP][ipCP]  += nu_CHP_CP_H;
	
	/* >>refer	mol	PHOTON	CH2	=>	CH2+	e	UMIST	*/
		
	c[ipCH2][ipCH2] -= nu_CH2_CH2P_e;
	c[ipCH2][ipCH2P] += nu_CH2_CH2P_e;
	
	/* >>refer	mol	PHOTON	CH2	=>	CH	H	UMIST	*/
		
	c[ipCH2][ipCH2] -= nu_CH2_CH_H;
	c[ipCH2][ipCH] += nu_CH2_CH_H;
	
	/* >>refer	mol	PHOTON	CH2+	=>	CH+	H	UMIST	*/
		
	c[ipCH2P][ipCH2P] -= nu_CH2P_CHP_H;
	c[ipCH2P][ipCHP] += nu_CH2P_CHP_H;
	
	/* >>refer	mol	PHOTON	CH3+	=>	CH2+	H	UMIST	*/
		
	c[ipCH3P][ipCH3P] -= nu_CH3P_CH2P_H;
	c[ipCH3P][ipCH2P] += nu_CH3P_CH2P_H;
	
	/* >>refer	mol	PHOTON	CH3+	=>	CH+	H2	UMIST	*/
		
	c[ipCH3P][ipCH3P] -= nu_CH3P_CHP_H2;
	c[ipCH3P][ipCHP] += nu_CH3P_CHP_H2;
	
	/* >>refer	mol	PHOTON	CO	=>	C	O	UMIST	*/

	/******************************************************************************************
	*	   First define the rate, which is of the form:
	*	   
	*		R = (Ro)*(Go*exp(-3.2Av))*Beta(tau(CO))
	*
	*	   where:
	*
	*	   Ro = 1.67*e-10
	*	   (Go*exp(-3.2Av)) = hmi.UV_Cont_rel2_Habing_TH85_depth
	*	   tauCO = 4.4e-15 * co.hevcol[ipCO] / (DoppVel.doppler[LIMELM]/1e5) /
	*		(1. + phycon.sqrte*0.6019) ; 
	*       tauC = 1.6*e17*N(C)
	*	   Beta(tau(CO)) = esca0k2(esc_co) 
	********************************************************************************************/   
	/* eqn 12 of 
	 * >>refer	CO	dissoc	Hollenbach, D.J., Takahashi, T., & Tielens, A. 1991, ApJ, 377, 192
	 * based on
	 * >>refer	CO	dissoc	Black, J.H., & van Dishoeck, E.F. 1988, ApJ, 334, 771 */
	esc_co = 4.4e-15 * co.hevcol[ipCO] / 
			 /* the line width in km/s */
			 (DoppVel.doppler[LIMELM]/1e5) /
			 /* this term accounts for populations within ground elec state */
			 (1. + phycon.sqrte*0.6019);

	/* hmi.UV_Cont_rel2_Habing_TH85_depth is field relative to Habing background, dimensionless */
	/* >>chng 04 apr 01, move from TH85 to DB96, also correct leading coef to use
	 * UMIST database value */
	/*co.CO_photo_dissoc_rate = 1.67e-10f*hmi.UV_Cont_rel2_Habing_TH85_depth;*/
	co.CO_photo_dissoc_rate = (float)(nu_CO_O_C * esca0k2(esc_co));

	c[ipCO][ipCO]  -= co.CO_photo_dissoc_rate;
	c[ipCO][ipATO] += co.CO_photo_dissoc_rate;
	c[ipCO][ipATC] += co.CO_photo_dissoc_rate;

	/* >>refer	mol	PHOTON	H2O	=>	H2O+	e	UMIST	*/
		
	c[ipH2O][ipH2O] -= nu_H2O_H2OP_e;
	c[ipH2O][ipH2OP] += nu_H2O_H2OP_e;
	
	/* >>refer	mol	PHOTON	H2O	=>	OH	H	UMIST	*/
		
	c[ipH2O][ipH2O] -= nu_H2O_OH_H;
	c[ipH2O][ipOH] += nu_H2O_OH_H;
	
	/* >>refer	mol	PHOTON	O2	=>	O2+	e	UMIST	*/
		
	c[ipO2][ipO2] -= nu_O2_O2P_e;
	c[ipO2][ipO2P] += nu_O2_O2P_e;
	
	/* >>refer	mol	PHOTON	O2	=>	O	O	UMIST	*/
	
	c[ipO2][ipO2] -= nu_O2_O_O;
	c[ipO2][ipATO] += 2*nu_O2_O_O;
	
	/* >>refer	mol	PHOTON	OH	=>	OH+	e	UMIST	*/
		
	c[ipOH][ipOH] -= nu_OH_OHP_e;
	c[ipOH][ipOHP] += nu_OH_OHP_e;
	
	/* >>refer	mol	PHOTON	OH	=>	O	H	UMIST	*/
		

	/* A note here is that approx. 55% of the O produced in this reaction goes into the 1D
	   state, which is the upper level of the 6300 Angstrom line, while 5% forms in the 1S
	   state and can create the 5577 line */
	
	/* >>chng 03 oct 20, following had been deleted so OH formation was always zero,
	 * this is used to get OH photo formation of [OI] 6300
	 *>>refer	OI	photoexcitation	Storzer, H., & Hollenbach, D. 2000, ApJ, 539, 751-759 
	 * discussion on bottom left side of page 752 */

	co.rate_OH_dissoc = (float)(nu_OH_O_H* OH);

	c[ipOH][ipOH] -= nu_OH_O_H;
	c[ipOH][ipATO] += nu_OH_O_H;
	
	/* >>refer	mol	PHOTON	OH+	=>	O	H+	UMIST	*/
		
	c[ipOHP][ipOHP] -= nu_OHP_O_HP;
	c[ipOHP][ipATO] += nu_OHP_O_HP;
	
	/* >>refer	mol	PHOTON	SiH	=>	Si	H	UMIST	*/
		
	c[ipSIH][ipSIH] -= nu_SiH_Si_H;
	c[ipSIH][ipATSI] += nu_SiH_Si_H;
	
	/* >>refer	mol	PHOTON	SiO	=>	Si	O	UMIST	*/
		
	c[ipSIO][ipSIO] -= nu_SiO_Si_O;
	c[ipSIO][ipATO] += nu_SiO_Si_O;
	c[ipSIO][ipATSI] += nu_SiO_Si_O;
	
	/* >>refer	mol	e	CH+	=>	C	H	UMIST	*/
	
	c[ipCHP][ipCHP] -= e_CHP_C_H;
	c[ipCHP][ipATC] += e_CHP_C_H;
	
	/* >>refer	mol	e	CH2+	=>	CH	H	UMIST	*/
		
	c[ipCH2P][ipCH2P] -= e_CH2P_CH_H;
	c[ipCH2P][ipCH] += e_CH2P_CH_H;
	
	/* >>refer	mol	e	CH2+	=>	C	H	H	UMIST	*/
		
	c[ipCH2P][ipCH2P] -= e_CH2P_C_H_H;
	c[ipCH2P][ipATC] += e_CH2P_C_H_H;

	/* >>refer	mol	e	CH2+	=>	C	H2	UMIST	*/
		
	c[ipCH2P][ipCH2P] -= e_CH2P_C_H2;
	c[ipCH2P][ipATC] += e_CH2P_C_H2;
	
	/* >>refer	mol	e	CH3+	=>	C	H2	H	UMIST	*/
		
	c[ipCH3P][ipCH3P] -= e_CH3P_C_H2_H;
	c[ipCH3P][ipATC] += e_CH3P_C_H2_H;
	
	/* >>refer	mol	e	CH3+	=>	CH2	H	UMIST	*/
	
	c[ipCH3P][ipCH3P] -= e_CH3P_CH2_H;
	c[ipCH3P][ipCH2] += e_CH3P_CH2_H;
	
	/* >>refer	mol	e	CH3+	=>	CH	H	H	UMIST	*/
		
	c[ipCH3P][ipCH3P] -= e_CH3P_CH_H_H;
	c[ipCH3P][ipCH] += e_CH3P_CH_H_H;
	
	/* >>refer	mol	e	CH3+	=>	CH	H2	UMIST	*/
		
	c[ipCH3P][ipCH3P] -= e_CH3P_CH_H2;
	c[ipCH3P][ipCH] += e_CH3P_CH_H2;
	
	/* >>refer	mol	e	CH3+	=>	CH3	PHOTON	UMIST	*/
		
	c[ipCH3P][ipCH3P] -= e_CH3P_CH3_nu;
	c[ipCH3P][ipCH3]   += e_CH3P_CH3_nu;

	/* >>refer	mol	e	CO+	=>	C	O	UMIST	*/
	
	c[ipCOP][ipCOP] -= e_COP_O_C;
	c[ipCOP][ipATC] += e_COP_O_C;
	c[ipCOP][ipATO] += e_COP_O_C;
		
	/* >>refer	mol	e	H2O+	=>	OH	H	UMIST	*/
		
	c[ipH2OP][ipH2OP] -= e_H2OP_OH_H;
	c[ipH2OP][ipOH] += e_H2OP_OH_H;
	
	/* >>refer	mol	e	H2O+	=>	O	H	H	UMIST	*/
		
	c[ipH2OP][ipH2OP] -= e_H2OP_O_H_H;
	c[ipH2OP][ipATO] += e_H2OP_O_H_H;
	
	/* >>refer	mol	e	H2O+	=>	O	H2	UMIST	*/
		
	c[ipH2OP][ipH2OP] -= e_H2OP_O_H2;
	c[ipH2OP][ipATO] += e_H2OP_O_H2;
	
	/* >>refer	mol	e	H3O+	=>	H2O	H	UMIST	*/
		
	c[ipH3OP][ipH3OP] -= e_H3OP_H2O_H;
	c[ipH3OP][ipH2O] += e_H3OP_H2O_H;
	
	/* >>refer	mol	e	H3O+	=>	OH	H	H	UMIST	*/
		
	c[ipH3OP][ipH3OP] -= e_H3OP_OH_H_H;
	c[ipH3OP][ipOH] += e_H3OP_OH_H_H;
	
	/* >>refer	mol	e	H3O+	=>	OH	H2	UMIST	*/
		
	c[ipH3OP][ipH3OP] -= e_H3OP_OH_H2;
	c[ipH3OP][ipOH] += e_H3OP_OH_H2;
	
	/* >>refer	mol	e	H3O+	=>	O	H2	H	UMIST	*/
		
	c[ipH3OP][ipH3OP] -= e_H3OP_O_H2_H;
	c[ipH3OP][ipATO] += e_H3OP_O_H2_H;
	
	/* >>refer	mol	e	HCO+	=>	CO	H	UMIST	*/
		
	c[ipHCOP][ipHCOP] -= e_HCOP_CO_H;
	c[ipHCOP][ipCO] += e_HCOP_CO_H;
	
	/* >>refer	mol	e	O2+	=>	O	O	UMIST	*/
	
	c[ipO2P][ipO2P] -= e_O2P_O_O;
	c[ipO2P][ipATO] += 2*e_O2P_O_O;
	
	/* >>refer	mol	e	OH+	=>	O	H	UMIST	*/
	
	c[ipOHP][ipOHP] -= e_OHP_O_H;
	c[ipOHP][ipATO] += e_OHP_O_H;
	
	/* >>refer	mol	e	SiH2+	=>	SiH	H	UMIST	*/
		
	c[ipSIH2P][ipSIH2P] -= e_SiH2P_SiH_H;
	c[ipSIH2P][ipSIH] += e_SiH2P_SiH_H;
	
	/* >>refer	mol	e	SiH2+	=>	Si	H	H	UMIST	*/
		
	c[ipSIH2P][ipSIH2P] -= e_SiH2P_Si_H_H;
	c[ipSIH2P][ipATSI] += e_SiH2P_Si_H_H;
	
	/* >>refer	mol	e	SiH2+	=>	Si	H2	UMIST	*/
		
	c[ipSIH2P][ipSIH2P] -= e_SiH2P_Si_H2;
	c[ipSIH2P][ipATSI] += e_SiH2P_Si_H2;
	
	/* >>refer	mol	e	SiO+	=>	Si	O	UMIST	*/
		
	c[ipSIOP][ipSIOP] -= e_SiOP_Si_O;
	c[ipSIOP][ipATSI] += e_SiOP_Si_O;
	c[ipSIOP][ipATO] += e_SiOP_Si_O;
	
	/* >>refer	mol	e	SiOH+	=>	SiO	H	UMIST	*/
		
	c[ipHOSIP][ipHOSIP] -= e_SiOHP_SiO_H;
	c[ipHOSIP][ipSIO] += e_SiOHP_SiO_H;
	
	/* >>refer	mol	e	SiOH+	=>	Si	OH	UMIST	*/
		
	c[ipHOSIP][ipHOSIP] -= e_SiOHP_Si_OH;
	c[ipHOSIP][ipOH] += e_SiOHP_Si_OH;
	c[ipHOSIP][ipATSI] += e_SiOHP_Si_OH;
	
	/* new simpler reactions for CH3, CH4, CH4+, CH5+ */
	
	/* >>refer mol   H CH3  =>  CH2 H2   UMIST    */
	c[ipCH3][ipCH3]  -=  H_CH3_CH2_H2;
	c[ipCH3][ipCH2]  +=  H_CH3_CH2_H2;
	
	/* >>refer mol   H CH4P  =>  CH3P H2   UMIST    */
	c[ipCH4P][ipCH4P]  -=  H_CH4P_CH3P_H2;
	c[ipCH4P][ipCH3P]  +=  H_CH4P_CH3P_H2;
	
	/* >>refer mol   H CH5P  =>  CH4P H2   UMIST    */
	c[ipCH5P][ipCH5P]  -=  H_CH5P_CH4P_H2;
	c[ipCH5P][ipCH4P]  +=  H_CH5P_CH4P_H2;
	
	/* >>refer mol   H2 CH2  =>  CH3 H   UMIST    */
	c[ipCH2][ipCH2]  -=  H2_CH2_CH3_H;
	c[ipCH2][ipCH3]  +=  H2_CH2_CH3_H;
	
	/* >>refer mol   H2 CH  =>  CH3 nu   UMIST    */
	c[ipCH][ipCH]  -=  H2_CH_CH3_nu;
	c[ipCH][ipCH3]  +=  H2_CH_CH3_nu;
	
	/* >>refer mol   H2 CH3  =>  CH4 H   UMIST    */
	c[ipCH3][ipCH3]  -=  H2_CH3_CH4_H;
	c[ipCH3][ipCH4]  +=  H2_CH3_CH4_H;
	
	/* >>refer mol   H2 CH4P  =>  CH5P H   UMIST    */
	c[ipCH4P][ipCH4P]  -=  H2_CH4P_CH5P_H;
	c[ipCH4P][ipCH5P]  +=  H2_CH4P_CH5P_H;
	
	/* >>refer mol   H2 CH3P  =>  CH5P nu   UMIST    */
	c[ipCH3P][ipCH3P]  -=  H2_CH3P_CH5P_nu;
	c[ipCH3P][ipCH5P]  +=  H2_CH3P_CH5P_nu;
	
	/* >>refer mol   H2P CH4  =>  CH3P H2 H  UMIST    */
	c[ipCH4][ipCH4]  -=  H2P_CH4_CH3P_H2;
	c[ipCH4][ipCH3P]  +=  H2P_CH4_CH3P_H2;
	
	/* >>refer mol   H2P CH4  =>  CH4P H2   UMIST    */
	c[ipCH4][ipCH4]  -=  H2P_CH4_CH4P_H2;
	c[ipCH4][ipCH4P]  +=  H2P_CH4_CH4P_H2;
	
	/* >>refer mol   H2P CH4  =>  CH5P H   UMIST    */
	c[ipCH4][ipCH4]  -=  H2P_CH4_CH5P_H;
	c[ipCH4][ipCH5P]  +=  H2P_CH4_CH5P_H;
	
	/* >>refer mol   H3P CH3  =>  CH4P H2   UMIST    */
	c[ipCH3][ipCH3]  -=  H3P_CH3_CH4P_H2;
	c[ipCH3][ipCH4P]  +=  H3P_CH3_CH4P_H2;
	
	/* >>refer mol   H3P CH4  =>  CH5P H2   UMIST    */
	c[ipCH4][ipCH4]  -=  H3P_CH4_CH5P_H2;
	c[ipCH4][ipCH5P]  +=  H3P_CH4_CH5P_H2;
	
	/* >>refer mol   HeP CH3  =>  CHP He H2  UMIST    */
	c[ipCH3][ipCH3]  -=  HeP_CH3_CHP_He;
	c[ipCH3][ipCHP]  +=  HeP_CH3_CHP_He;
	
	/* >>refer mol   HeP CH4  =>  CHP He H2 H UMIST    */
	c[ipCH4][ipCH4]  -=  HeP_CH4_CHP_He;
	c[ipCH4][ipCHP]  +=  HeP_CH4_CHP_He;
	
	/* >>refer mol   HeP CH4  =>  CH2P He H2  UMIST    */
	c[ipCH4][ipCH4]  -=  HeP_CH4_CH2P_He;
	c[ipCH4][ipCH2P]  +=  HeP_CH4_CH2P_He;
	
	/* >>refer mol   HeP CH4  =>  CH3 He HP  UMIST    */
	c[ipCH4][ipCH4]  -=  HeP_CH4_CH3_He;
	c[ipCH4][ipCH3]  +=  HeP_CH4_CH3_He;
	
	/* >>refer mol   HeP CH4  =>  CH3P He H  UMIST    */
	c[ipCH4][ipCH4]  -=  HeP_CH4_CH3P_He;
	c[ipCH4][ipCH3P]  +=  HeP_CH4_CH3P_He;
	
	/* >>refer mol   HeP CH4  =>  CH4P He   UMIST    */
	c[ipCH4][ipCH4]  -=  HeP_CH4_CH4P_He;
	c[ipCH4][ipCH4P]  +=  HeP_CH4_CH4P_He;
	
	/* >>refer mol   Hminus CH2  =>  CH3 e   UMIST    */
	c[ipCH2][ipCH2]  -=  Hminus_CH2_CH3_e;
	c[ipCH2][ipCH3]  +=  Hminus_CH2_CH3_e;
	
	/* >>refer mol   Hminus CH3  =>  CH4 e   UMIST    */
	c[ipCH3][ipCH3]  -=  Hminus_CH3_CH4_e;
	c[ipCH3][ipCH4]  +=  Hminus_CH3_CH4_e;
	
	/* >>refer mol   HP CH3  =>  CH3P H   UMIST    */
	c[ipCH3][ipCH3]  -=  HP_CH3_CH3P_H;
	c[ipCH3][ipCH3P]  +=  HP_CH3_CH3P_H;
	
	/* >>refer mol   HP CH4  =>  CH3P H2   UMIST    */
	c[ipCH4][ipCH4]  -=  HP_CH4_CH3P_H2;
	c[ipCH4][ipCH3P]  +=  HP_CH4_CH3P_H2;
	
	/* >>refer mol   HP CH4  =>  CH4P H   UMIST    */
	c[ipCH4][ipCH4]  -=  HP_CH4_CH4P_H;
	c[ipCH4][ipCH4P]  +=  HP_CH4_CH4P_H;
	
	/* >>refer mol   nu CH3  =>  CH3P e   UMIST    */
	c[ipCH3][ipCH3]  -=  nu_CH3_CH3P_e;
	c[ipCH3][ipCH3P]  +=  nu_CH3_CH3P_e;
	
	/* >>refer mol   nu CH3  =>  CH2 H   UMIST    */
	c[ipCH3][ipCH3]  -=  nu_CH3_CH2_H;
	c[ipCH3][ipCH2]  +=  nu_CH3_CH2_H;
	
	/* >>refer mol   nu CH3  =>  CH H2   UMIST    */
	c[ipCH3][ipCH3]  -=  nu_CH3_CH_H2;
	c[ipCH3][ipCH]  +=  nu_CH3_CH_H2;
	
	/* >>refer mol   crnu CH3  =>  CH3P e   UMIST    */
	c[ipCH3][ipCH3]  -=  crnu_CH3_CH3P_e;
	c[ipCH3][ipCH3P]  +=  crnu_CH3_CH3P_e;
	
	/* >>refer mol   crnu CH3  =>  CH2 H   UMIST    */
	c[ipCH3][ipCH3]  -=  crnu_CH3_CH2_H;
	c[ipCH3][ipCH2]  +=  crnu_CH3_CH2_H;
	
	/* >>refer mol   crnu CH3  =>  CH H2   UMIST    */
	c[ipCH3][ipCH3]  -=  crnu_CH3_CH_H2;
	c[ipCH3][ipCH]  +=  crnu_CH3_CH_H2;

	/* >>refer	mol	C + crnu => C+	UMIST */
	/* >>chng 04 sep 19, rm this reaction as per NA email */
	/* This reaction should only be used when the UMIST rate file is wanted 
	if(!co.lgUMISTrates)
		c[ipATC][ipCP] += 5e-17* 2.*510 ;*/

	/* >>refer mol   nu CH4  =>  CH3 H   UMIST    */
	c[ipCH4][ipCH4]  -=  nu_CH4_CH3_H;
	c[ipCH4][ipCH3]  +=  nu_CH4_CH3_H;	

	/* >>refer mol   nu CH4  =>  CH2 H2   UMIST    */
	c[ipCH4][ipCH4]  -=  nu_CH4_CH2_H2;
	c[ipCH4][ipCH2]  +=  nu_CH4_CH2_H2;
	
	/* >>refer mol   nu CH4  =>  CH H2 H  UMIST    */
	c[ipCH4][ipCH4]  -=  nu_CH4_CH_H2;
	c[ipCH4][ipCH]  +=  nu_CH4_CH_H2;
	
	/* >>refer mol   crnu CH4  =>  CH2 H2   UMIST    */
	c[ipCH4][ipCH4]  -=  crnu_CH4_CH2_H2;
	c[ipCH4][ipCH2]  +=  crnu_CH4_CH2_H2;
	
	/* >>refer mol   e CH5P  =>  CH3 H2   UMIST    */
	c[ipCH5P][ipCH5P]  -=  e_CH5P_CH3_H2;
	c[ipCH5P][ipCH3]  +=  e_CH5P_CH3_H2;
	
	/* >>refer mol   e CH5P  =>  CH4 H   UMIST    */
	c[ipCH5P][ipCH5P]  -=  e_CH5P_CH4_H;
	c[ipCH5P][ipCH4]  +=  e_CH5P_CH4_H;
	
	/* >>refer mol   e CH4P  =>  CH3 H   UMIST    */
	c[ipCH4P][ipCH4P]  -=  e_CH4P_CH3_H;
	c[ipCH4P][ipCH3]  +=  e_CH4P_CH3_H;

	/* >>refer mol   e CH4P  =>  CH2 H	H   UMIST    */
	c[ipCH4P][ipCH4P]  -=  e_CH4P_CH2_H_H;
	c[ipCH4P][ipCH2]   +=  e_CH4P_CH2_H_H;


 
	/********************************************************************************

									COUPLED REACTIONS

	 ********************************************************************************/

	/* This is the C2 and C2+ chemistry */

	/* >>refer	mol	H		C2	=>		CH		C		UMIST	*/

	c[ipC2][ipC2] -= H_C2_CH_C;
	c[ipC2][ipCH]  += H_C2_CH_C;
	c[ipC2][ipATC]  += H_C2_CH_C;

	/* >>refer	mol	C		CH	=>		C2		H		UMIST	*/
	
	c[ipCH][ipATC] -= C_CH_C2_H_2;
	c[ipATC][ipATC] -= C_CH_C2_H_1;
	b[ipATC]     -= b_C_CH_C2_H;
	
	c[ipCH][ipCH] -= C_CH_C2_H_2;
	c[ipATC][ipCH] -= C_CH_C2_H_1;
	b[ipCH]     -= b_C_CH_C2_H;

	c[ipCH][ipC2]  += C_CH_C2_H_2;
	c[ipATC][ipC2]  += C_CH_C2_H_1;
	b[ipC2]      += b_C_CH_C2_H;

	/* >>refer	mol	C		CN	=>		C2		N		UMIST	*/
	
	c[ipCN][ipCN] -= C_CN_C2_N_2;
	c[ipATC][ipCN] -= C_CN_C2_N_1;
	b[ipCN]     -= b_C_CN_C2_N;
	
	c[ipCN][ipATC] -= C_CN_C2_N_2;
	c[ipATC][ipATC] -= C_CN_C2_N_1;
	b[ipATC]     -= b_C_CN_C2_N;

	c[ipCN][ipC2]  += C_CN_C2_N_2;
	c[ipATC][ipC2]  += C_CN_C2_N_1;
	b[ipC2]      += b_C_CN_C2_N;

	c[ipCN][ipATN]  += C_CN_C2_N_2;
	c[ipATC][ipATN]  += C_CN_C2_N_1;
	b[ipATN]      += b_C_CN_C2_N;

	/* >>refer	mol	C		CO	=>		C2		O		UMIST	*/
	
	c[ipCO][ipCO] -= C_CO_C2_O_2;
	c[ipATC][ipCO] -= C_CO_C2_O_1;
	b[ipCO]     -= b_C_CO_C2_O;
	
	c[ipCO][ipATC] -= C_CO_C2_O_2;
	c[ipATC][ipATC] -= C_CO_C2_O_1;
	b[ipATC]     -= b_C_CO_C2_O;

	c[ipCO][ipC2]  += C_CO_C2_O_2;
	c[ipATC][ipC2]  += C_CO_C2_O_1;
	b[ipC2]      += b_C_CO_C2_O;

	c[ipCO][ipATO]  += C_CO_C2_O_2;
	c[ipATC][ipATO]  += C_CO_C2_O_1;
	b[ipATO]      += b_C_CO_C2_O;

	/* >>refer	mol	C		CS	=>		S		C2		UMIST	*/
	
	c[ipCS][ipCS] -= C_CS_S_C2_2;
	c[ipATC][ipCS] -= C_CS_S_C2_1;
	b[ipCS]     -= b_C_CS_S_C2;
	
	c[ipCS][ipATC] -= C_CS_S_C2_2;
	c[ipATC][ipATC] -= C_CS_S_C2_1;
	b[ipATC]     -= b_C_CS_S_C2;

	c[ipCS][ipATS]  += C_CS_S_C2_2;
	c[ipATC][ipATS]  += C_CS_S_C2_1;
	b[ipATS]      += b_C_CS_S_C2;

	c[ipCS][ipC2]  += C_CS_S_C2_2;
	c[ipATC][ipC2]  += C_CS_S_C2_1;
	b[ipC2]      += b_C_CS_S_C2;

	/* >>refer	mol	N		C2	=>		CN		C		UMIST	*/
	
	c[ipC2][ipC2] -= N_C2_CN_C_2;
	c[ipATN][ipC2] -= N_C2_CN_C_1;
	b[ipC2]     -= b_N_C2_CN_C;
	
	c[ipC2][ipATN] -= N_C2_CN_C_2;
	c[ipATN][ipATN] -= N_C2_CN_C_1;
	b[ipATN]     -= b_N_C2_CN_C;

	c[ipC2][ipCN]  += N_C2_CN_C_2;
	c[ipATN][ipCN]  += N_C2_CN_C_1;
	b[ipCN]      += b_N_C2_CN_C;

	c[ipC2][ipATC]  += N_C2_CN_C_2;
	c[ipATN][ipATC]  += N_C2_CN_C_1;
	b[ipATC]      += b_N_C2_CN_C;

	/* >>refer	mol	O		C2	=>		CO		C		UMIST	*/
	
	c[ipC2][ipC2] -= O_C2_CO_C_2;
	c[ipATO][ipC2] -= O_C2_CO_C_1;
	b[ipC2]     -= b_O_C2_CO_C;
	
	c[ipC2][ipATO] -= O_C2_CO_C_2;
	c[ipATO][ipATO] -= O_C2_CO_C_1;
	b[ipATO]     -= b_O_C2_CO_C;

	c[ipC2][ipCO]  += O_C2_CO_C_2;
	c[ipATO][ipCO]  += O_C2_CO_C_1;
	b[ipCO]      += b_O_C2_CO_C;

	c[ipC2][ipATC]  += O_C2_CO_C_2;
	c[ipATO][ipATC]  += O_C2_CO_C_1;
	b[ipATC]      += b_O_C2_CO_C;

	/* >>refer	mol	C2		O2	=>		CO		CO		UMIST	*/
	
	c[ipC2][ipC2] -= C2_O2_CO_CO_1;
	c[ipO2][ipC2] -= C2_O2_CO_CO_2;
	b[ipC2]     -= b_C2_O2_CO_CO;
	
	c[ipC2][ipO2] -= C2_O2_CO_CO_1;
	c[ipO2][ipO2] -= C2_O2_CO_CO_2;
	b[ipO2]     -= b_C2_O2_CO_CO;

	c[ipC2][ipCO]  += C2_O2_CO_CO_1;
	c[ipO2][ipCO]  += C2_O2_CO_CO_2;
	b[ipCO]      += b_C2_O2_CO_CO;

	c[ipC2][ipCO]  += C2_O2_CO_CO_1;
	c[ipO2][ipCO]  += C2_O2_CO_CO_2;
	b[ipCO]      += b_C2_O2_CO_CO;

	/* >>refer	mol	C2		S	=>		CS		C		UMIST	*/
	
	c[ipC2][ipC2] -= C2_S_CS_C_1;
	c[ipATS][ipC2] -= C2_S_CS_C_2;
	b[ipC2]     -= b_C2_S_CS_C;
	
	c[ipC2][ipATS] -= C2_S_CS_C_1;
	c[ipATS][ipATS] -= C2_S_CS_C_2;
	b[ipATS]     -= b_C2_S_CS_C;

	c[ipC2][ipCS]  += C2_S_CS_C_1;
	c[ipATS][ipCS]  += C2_S_CS_C_2;
	b[ipCS]      += b_C2_S_CS_C;

	c[ipC2][ipATC]  += C2_S_CS_C_1;
	c[ipATS][ipATC]  += C2_S_CS_C_2;
	b[ipATC]      += b_C2_S_CS_C;

	/* >>refer	mol	He+		C2	=>		C+		C	He	UMIST	*/
	
	c[ipC2][ipC2] -= HeP_C2_CP_C_He;
	c[ipC2][ipCP]  += HeP_C2_CP_C_He;
	c[ipC2][ipATC]  += HeP_C2_CP_C_He;

	/* >>refer	mol	NH+		C2	=>		HCN+	C		UMIST	*/
	
	c[ipC2][ipC2] -= NHP_C2_HCNP_C_2;
	c[ipNHP][ipC2] -= NHP_C2_HCNP_C_1;
	b[ipC2]     -= b_NHP_C2_HCNP_C;
	
	c[ipC2][ipNHP] -= NHP_C2_HCNP_C_2;
	c[ipNHP][ipNHP] -= NHP_C2_HCNP_C_1;
	b[ipNHP]     -= b_NHP_C2_HCNP_C;

	c[ipC2][ipHCNP]  += NHP_C2_HCNP_C_2;
	c[ipNHP][ipHCNP]  += NHP_C2_HCNP_C_1;
	b[ipHCNP]      += b_NHP_C2_HCNP_C;

	c[ipC2][ipATC]  += NHP_C2_HCNP_C_2;
	c[ipNHP][ipATC]  += NHP_C2_HCNP_C_1;
	b[ipATC]      += b_NHP_C2_HCNP_C;

	/* >>refer	mol	O+		C2	=>		CO+		C		UMIST	*/
	
	c[ipOP][ipOP] -= OP_C2_COP_C_1;
	c[ipC2][ipOP] -= OP_C2_COP_C_2;
	b[ipOP]     -= b_OP_C2_COP_C;
	
	c[ipOP][ipC2] -= OP_C2_COP_C_1;
	c[ipC2][ipC2] -= OP_C2_COP_C_2;
	b[ipC2]     -= b_OP_C2_COP_C;

	c[ipOP][ipCOP]  += OP_C2_COP_C_1;
	c[ipC2][ipCOP]  += OP_C2_COP_C_2;
	b[ipCOP]      += b_OP_C2_COP_C;

	c[ipOP][ipATC]  += OP_C2_COP_C_1;
	c[ipC2][ipATC]  += OP_C2_COP_C_2;
	b[ipATC]      += b_OP_C2_COP_C;

	/* >>refer	mol	C2		O2+	=>		CO+		CO		UMIST	*/
	
	c[ipC2][ipC2] -= C2_O2P_COP_CO_1;
	c[ipO2P][ipC2] -= C2_O2P_COP_CO_2;
	b[ipC2]     -= b_C2_O2P_COP_CO;
	
	c[ipC2][ipO2P] -= C2_O2P_COP_CO_1;
	c[ipO2P][ipO2P] -= C2_O2P_COP_CO_2;
	b[ipO2P]     -= b_C2_O2P_COP_CO;

	c[ipC2][ipCO]  += C2_O2P_COP_CO_1;
	c[ipO2P][ipCO]  += C2_O2P_COP_CO_2;
	b[ipCO]      += b_C2_O2P_COP_CO;

	c[ipC2][ipCOP]  += C2_O2P_COP_CO_1;
	c[ipO2P][ipCOP]  += C2_O2P_COP_CO_2;
	b[ipCOP]      += b_C2_O2P_COP_CO;

	/* >>refer	mol	C2		S+	=>		CS+		C		UMIST	*/
	
	c[ipC2][ipC2] -= C2_SP_CSP_C_1;
	c[ipSP][ipC2] -= C2_SP_CSP_C_2;
	b[ipC2]     -= b_C2_SP_CSP_C;
	
	c[ipC2][ipSP] -= C2_SP_CSP_C_1;
	c[ipSP][ipSP] -= C2_SP_CSP_C_2;
	b[ipSP]     -= b_C2_SP_CSP_C;

	c[ipC2][ipCSP]  += C2_SP_CSP_C_1;
	c[ipSP][ipCSP]  += C2_SP_CSP_C_2;
	b[ipCSP]      += b_C2_SP_CSP_C;

	c[ipC2][ipATC]  += C2_SP_CSP_C_1;
	c[ipSP][ipATC]  += C2_SP_CSP_C_2;
	b[ipATC]      += b_C2_SP_CSP_C;

	/* >>refer	mol	H+		C2	=>		C2+		H		UMIST	*/
	
	c[ipC2][ipC2] -= HP_C2_C2P_H;
	c[ipC2][ipC2P]  += HP_C2_C2P_H;

	/* >>refer	mol	H2+		C2	=>		C2+		H2		UMIST	*/
	
	c[ipC2][ipC2] -= H2P_C2_C2P_H2;
	c[ipC2][ipC2P]  += H2P_C2_C2P_H2;

	/* >>refer	mol	He+		C2	=>		C2+		He		UMIST	*/
	
	c[ipC2][ipC2] -= HeP_C2_C2P_He;
	c[ipC2][ipC2P]  += HeP_C2_C2P_He;

	/* >>refer	mol	C		C2+	=>		C2		C+		UMIST	*/
	
	c[ipC2P][ipC2P] -= C_C2P_C2_CP_2;
	c[ipATC][ipC2P] -= C_C2P_C2_CP_1;
	b[ipC2P]     -= b_C_C2P_C2_CP;
	
	c[ipC2P][ipATC] -= C_C2P_C2_CP_2;
	c[ipATC][ipATC] -= C_C2P_C2_CP_1;
	b[ipATC]     -= b_C_C2P_C2_CP;

	c[ipC2P][ipC2]  += C_C2P_C2_CP_2;
	c[ipATC][ipC2]  += C_C2P_C2_CP_1;
	b[ipC2]      += b_C_C2P_C2_CP;

	c[ipC2P][ipCP]  += C_C2P_C2_CP_2;
	c[ipATC][ipCP]  += C_C2P_C2_CP_1;
	b[ipCP]      += b_C_C2P_C2_CP;

	/* >>refer	mol	CH		C2+	=>		C2		CH+		UMIST	*/
	
	c[ipC2P][ipC2P] -= CH_C2P_C2_CHP_2;
	c[ipCH][ipC2P] -= CH_C2P_C2_CHP_1;
	b[ipC2P]     -= b_CH_C2P_C2_CHP;
	
	c[ipC2P][ipCH] -= CH_C2P_C2_CHP_2;
	c[ipCH][ipCH] -= CH_C2P_C2_CHP_1;
	b[ipCH]     -= b_CH_C2P_C2_CHP;

	c[ipC2P][ipC2]  += CH_C2P_C2_CHP_2;
	c[ipCH][ipC2]  += CH_C2P_C2_CHP_1;
	b[ipC2]      += b_CH_C2P_C2_CHP;

	c[ipC2P][ipCHP]  += CH_C2P_C2_CHP_2;
	c[ipCH][ipCHP]  += CH_C2P_C2_CHP_1;
	b[ipCHP]      += b_CH_C2P_C2_CHP;

	/* >>refer	mol	N+		C2	=>		C2+		N		UMIST	*/
	
	c[ipC2][ipC2] -= NP_C2_C2P_N_2;
	c[ipNP][ipC2] -= NP_C2_C2P_N_1;
	b[ipC2]     -= b_NP_C2_C2P_N;
	
	c[ipC2][ipNP] -= NP_C2_C2P_N_2;
	c[ipNP][ipNP] -= NP_C2_C2P_N_1;
	b[ipNP]     -= b_NP_C2_C2P_N;

	c[ipC2][ipC2P]  += NP_C2_C2P_N_2;
	c[ipNP][ipC2P]  += NP_C2_C2P_N_1;
	b[ipC2P]      += b_NP_C2_C2P_N;

	c[ipC2][ipATN]  += NP_C2_C2P_N_2;
	c[ipNP][ipATN]  += NP_C2_C2P_N_1;
	b[ipATN]      += b_NP_C2_C2P_N;

	/* >>refer	mol	CH2		C2+	=>		C2		CH2+	UMIST	*/
	
	c[ipC2P][ipC2P] -= CH2_C2P_C2_CH2P_2;
	c[ipCH2][ipC2P] -= CH2_C2P_C2_CH2P_1;
	b[ipC2P]     -= b_CH2_C2P_C2_CH2P;
	
	c[ipC2P][ipCH2] -= CH2_C2P_C2_CH2P_2;
	c[ipCH2][ipCH2] -= CH2_C2P_C2_CH2P_1;
	b[ipCH2]     -= b_CH2_C2P_C2_CH2P;

	c[ipC2P][ipC2]  += CH2_C2P_C2_CH2P_2;
	c[ipCH2][ipC2]  += CH2_C2P_C2_CH2P_1;
	b[ipC2]      += b_CH2_C2P_C2_CH2P;

	c[ipC2P][ipCH2P]  += CH2_C2P_C2_CH2P_2;
	c[ipCH2][ipCH2P]  += CH2_C2P_C2_CH2P_1;
	b[ipCH2P]      += b_CH2_C2P_C2_CH2P;

	/* >>refer	mol	O+		C2	=>		C2+		O		UMIST	*/
	
	c[ipC2][ipC2] -= OP_C2_C2P_O_2;
	c[ipOP][ipC2] -= OP_C2_C2P_O_1;
	b[ipC2]     -= b_OP_C2_C2P_O;
	
	c[ipC2][ipOP] -= OP_C2_C2P_O_2;
	c[ipOP][ipOP] -= OP_C2_C2P_O_1;
	b[ipOP]     -= b_OP_C2_C2P_O;

	c[ipC2][ipC2P]  += OP_C2_C2P_O_2;
	c[ipOP][ipC2P]  += OP_C2_C2P_O_1;
	b[ipC2P]      += b_OP_C2_C2P_O;

	c[ipC2][ipATO]  += OP_C2_C2P_O_2;
	c[ipOP][ipATO]  += OP_C2_C2P_O_1;
	b[ipATO]      += b_OP_C2_C2P_O;

	/* >>refer	mol	NH2		C2+	=>		C2		NH2+	UMIST	*/
	
	c[ipC2P][ipC2P] -= NH2_C2P_C2_NH2P_2;
	c[ipNH2][ipC2P] -= NH2_C2P_C2_NH2P_1;
	b[ipC2P]     -= b_NH2_C2P_C2_NH2P;
	
	c[ipC2P][ipNH2] -= NH2_C2P_C2_NH2P_2;
	c[ipNH2][ipNH2] -= NH2_C2P_C2_NH2P_1;
	b[ipNH2]     -= b_NH2_C2P_C2_NH2P;

	c[ipC2P][ipC2]  += NH2_C2P_C2_NH2P_2;
	c[ipNH2][ipC2]  += NH2_C2P_C2_NH2P_1;
	b[ipC2]      += b_NH2_C2P_C2_NH2P;

	c[ipC2P][ipNH2P]  += NH2_C2P_C2_NH2P_2;
	c[ipNH2][ipNH2P]  += NH2_C2P_C2_NH2P_1;
	b[ipNH2P]      += b_NH2_C2P_C2_NH2P;

	/* >>refer	mol	OH+		C2	=>		C2+		OH		UMIST	*/
	
	c[ipC2][ipC2] -= OHP_C2_C2P_OH_2;
	c[ipOHP][ipC2] -= OHP_C2_C2P_OH_1;
	b[ipC2]     -= b_OHP_C2_C2P_OH;
	
	c[ipC2][ipOHP] -= OHP_C2_C2P_OH_2;
	c[ipOHP][ipOHP] -= OHP_C2_C2P_OH_1;
	b[ipOHP]     -= b_OHP_C2_C2P_OH;

	c[ipC2][ipC2P]  += OHP_C2_C2P_OH_2;
	c[ipOHP][ipC2P]  += OHP_C2_C2P_OH_1;
	b[ipC2P]      += b_OHP_C2_C2P_OH;

	c[ipC2][ipOH]  += OHP_C2_C2P_OH_2;
	c[ipOHP][ipOH]  += OHP_C2_C2P_OH_1;
	b[ipOH]      += b_OHP_C2_C2P_OH;

	/* >>refer	mol	OH		C2+	=>		C2		OH+		UMIST	*/
	
	c[ipOH][ipOH] -= OH_C2P_C2_OHP_1;
	c[ipC2P][ipOH] -= OH_C2P_C2_OHP_2;
	b[ipOH]     -= b_OH_C2P_C2_OHP;
	
	c[ipOH][ipC2P] -= OH_C2P_C2_OHP_1;
	c[ipC2P][ipC2P] -= OH_C2P_C2_OHP_2;
	b[ipC2P]     -= b_OH_C2P_C2_OHP;

	c[ipOH][ipC2]  += OH_C2P_C2_OHP_1;
	c[ipC2P][ipC2]  += OH_C2P_C2_OHP_2;
	b[ipC2]      += b_OH_C2P_C2_OHP;

	c[ipOH][ipOHP]  += OH_C2P_C2_OHP_1;
	c[ipC2P][ipOHP]  += OH_C2P_C2_OHP_2;
	b[ipOHP]      += b_OH_C2P_C2_OHP;

	/* >>refer	mol	H2O+	C2	=>		C2+		H2O		UMIST	*/
	
	c[ipC2][ipC2] -= H2OP_C2_C2P_H2O_2;
	c[ipH2OP][ipC2] -= H2OP_C2_C2P_H2O_1;
	b[ipC2]     -= b_H2OP_C2_C2P_H2O;
	
	c[ipC2][ipH2OP] -= H2OP_C2_C2P_H2O_2;
	c[ipH2OP][ipH2OP] -= H2OP_C2_C2P_H2O_1;
	b[ipH2OP]     -= b_H2OP_C2_C2P_H2O;

	c[ipC2][ipC2P]  += H2OP_C2_C2P_H2O_2;
	c[ipH2OP][ipC2P]  += H2OP_C2_C2P_H2O_1;
	b[ipC2P]      += b_H2OP_C2_C2P_H2O;

	c[ipC2][ipH2O]  += H2OP_C2_C2P_H2O_2;
	c[ipH2OP][ipH2O]  += H2OP_C2_C2P_H2O_1;
	b[ipH2O]      += b_H2OP_C2_C2P_H2O;

	/* >>refer	mol	C2		CN+	=>		CN		C2+		UMIST	*/
	
	c[ipCNP][ipCNP] -= C2_CNP_CN_C2P_2;
	c[ipC2][ipCNP] -= C2_CNP_CN_C2P_1;
	b[ipCNP]     -= b_C2_CNP_CN_C2P;
	
	c[ipCNP][ipC2] -= C2_CNP_CN_C2P_2;
	c[ipC2][ipC2] -= C2_CNP_CN_C2P_1;
	b[ipC2]     -= b_C2_CNP_CN_C2P;

	c[ipCNP][ipCN]  += C2_CNP_CN_C2P_2;
	c[ipC2][ipCN]  += C2_CNP_CN_C2P_1;
	b[ipCN]      += b_C2_CNP_CN_C2P;

	c[ipCNP][ipC2P]  += C2_CNP_CN_C2P_2;
	c[ipC2][ipC2P]  += C2_CNP_CN_C2P_1;
	b[ipC2P]      += b_C2_CNP_CN_C2P;

	/* >>refer	mol	C2		CO+	=>		CO		C2+		UMIST	*/
	
	c[ipCOP][ipCOP] -= C2_COP_CO_C2P_2;
	c[ipC2][ipCOP] -= C2_COP_CO_C2P_1;
	b[ipCOP]     -= b_C2_COP_CO_C2P;
	
	c[ipCOP][ipC2] -= C2_COP_CO_C2P_2;
	c[ipC2][ipC2] -= C2_COP_CO_C2P_1;
	b[ipC2]     -= b_C2_COP_CO_C2P;

	c[ipCOP][ipCO]  += C2_COP_CO_C2P_2;
	c[ipC2][ipCO]  += C2_COP_CO_C2P_1;
	b[ipCO]      += b_C2_COP_CO_C2P;

	c[ipCOP][ipC2P]  += C2_COP_CO_C2P_2;
	c[ipC2][ipC2P]  += C2_COP_CO_C2P_1;
	b[ipC2P]      += b_C2_COP_CO_C2P;

	/* >>refer	mol	C2		N2+	=>		N2		C2+		UMIST	*/
	
	c[ipN2P][ipN2P] -= C2_N2P_N2_C2P_2;
	c[ipC2][ipN2P] -= C2_N2P_N2_C2P_1;
	b[ipN2P]     -= b_C2_N2P_N2_C2P;
	
	c[ipN2P][ipC2] -= C2_N2P_N2_C2P_2;
	c[ipC2][ipC2] -= C2_N2P_N2_C2P_1;
	b[ipC2]     -= b_C2_N2P_N2_C2P;

	c[ipN2P][ipN2]  += C2_N2P_N2_C2P_2;
	c[ipC2][ipN2]  += C2_N2P_N2_C2P_1;
	b[ipN2]      += b_C2_N2P_N2_C2P;

	c[ipN2P][ipC2P]  += C2_N2P_N2_C2P_2;
	c[ipC2][ipC2P]  += C2_N2P_N2_C2P_1;
	b[ipC2P]      += b_C2_N2P_N2_C2P;

	/* >>refer	mol	C2+		NO	=>		NO+		C2		UMIST	*/
	
	c[ipNO][ipNO] -= C2P_NO_NOP_C2_2;
	c[ipC2P][ipNO] -= C2P_NO_NOP_C2_1;
	b[ipNO]     -= b_C2P_NO_NOP_C2;
	
	c[ipNO][ipC2P] -= C2P_NO_NOP_C2_2;
	c[ipC2P][ipC2P] -= C2P_NO_NOP_C2_1;
	b[ipC2P]     -= b_C2P_NO_NOP_C2;

	c[ipNO][ipNOP]  += C2P_NO_NOP_C2_2;
	c[ipC2P][ipNOP]  += C2P_NO_NOP_C2_1;
	b[ipNOP]      += b_C2P_NO_NOP_C2;

	c[ipNO][ipC2]  += C2P_NO_NOP_C2_2;
	c[ipC2P][ipC2]  += C2P_NO_NOP_C2_1;
	b[ipC2]      += b_C2P_NO_NOP_C2;

	/* >>refer	mol	C2		O2+	=>		O2		C2+		UMIST	*/
	
	c[ipO2P][ipO2P] -= C2_O2P_O2_C2P_2;
	c[ipC2][ipO2P] -= C2_O2P_O2_C2P_1;
	b[ipO2P]     -= b_C2_O2P_O2_C2P;
	
	c[ipO2P][ipC2] -= C2_O2P_O2_C2P_2;
	c[ipC2][ipC2] -= C2_O2P_O2_C2P_1;
	b[ipC2]     -= b_C2_O2P_O2_C2P;

	c[ipO2P][ipO2]  += C2_O2P_O2_C2P_2;
	c[ipC2][ipO2]  += C2_O2P_O2_C2P_1;
	b[ipO2]      += b_C2_O2P_O2_C2P;

	c[ipO2P][ipC2P]  += C2_O2P_O2_C2P_2;
	c[ipC2][ipC2P]  += C2_O2P_O2_C2P_1;
	b[ipC2P]      += b_C2_O2P_O2_C2P;

	/* >>refer	mol	C2+		S	=>		S+		C2		UMIST	*/
	
	c[ipATS][ipATS] -= C2P_S_SP_C2_2;
	c[ipC2P][ipATS] -= C2P_S_SP_C2_1;
	b[ipATS]     -= b_C2P_S_SP_C2;
	
	c[ipATS][ipC2P] -= C2P_S_SP_C2_2;
	c[ipC2P][ipC2P] -= C2P_S_SP_C2_1;
	b[ipC2P]     -= b_C2P_S_SP_C2;

	c[ipATS][ipSP]  += C2P_S_SP_C2_2;
	c[ipC2P][ipSP]  += C2P_S_SP_C2_1;
	b[ipSP]      += b_C2P_S_SP_C2;

	c[ipATS][ipC2]  += C2P_S_SP_C2_2;
	c[ipC2P][ipC2]  += C2P_S_SP_C2_1;
	b[ipC2]      += b_C2P_S_SP_C2;

	/* >>refer	mol	C		C	=>		C2		nu		UMIST	*/
	
	c[ipATC][ipATC] -= C_C_C2_nu_2;
	c[ipATC][ipATC] -= C_C_C2_nu_1;
	b[ipATC]     -= b_C_C_C2_nu;
	
	c[ipATC][ipC2]  += 0.5*C_C_C2_nu_2;
	c[ipATC][ipC2]  += 0.5*C_C_C2_nu_1;
	b[ipC2]      += 0.5*b_C_C_C2_nu;

	/* >>refer	mol	C2		nu	=>		C		C		UMIST	*/
	
	c[ipC2][ipC2] -= C2_nu_C_C;
	c[ipC2][ipATC]  += C2_nu_C_C;
	c[ipC2][ipATC]  += C2_nu_C_C;

	/* >>refer	mol	C2		nu	=>		C2+		e		UMIST	*/
	
	c[ipC2][ipC2] -= C2_nu_C2P_e;
	c[ipC2][ipC2P]  += C2_nu_C2P_e;

	/* >>refer	mol	C2		cr	=>		C		C		UMIST	*/
	
	c[ipC2][ipC2] -= C2_cr_C_C;
	c[ipC2][ipATC]  += C2_cr_C_C;
	c[ipC2][ipATC]  += C2_cr_C_C;

	/* >>refer	mol	C+		CH	=>		C2+		H		UMIST	*/
	
	c[ipCH][ipCH] -= CP_CH_C2P_H_2;
	c[ipCP][ipCH] -= CP_CH_C2P_H_1;
	b[ipCH]     -= b_CP_CH_C2P_H;
	
	c[ipCH][ipCP] -= CP_CH_C2P_H_2;
	c[ipCP][ipCP] -= CP_CH_C2P_H_1;
	b[ipCP]     -= b_CP_CH_C2P_H;

	c[ipCH][ipC2P]  += CP_CH_C2P_H_2;
	c[ipCP][ipC2P]  += CP_CH_C2P_H_1;
	b[ipC2P]      += b_CP_CH_C2P_H;

	/* >>refer	mol	C		CH+	=>		C2+		H		UMIST	*/
	
	c[ipCHP][ipCHP] -= C_CHP_C2P_H_2;
	c[ipATC][ipCHP] -= C_CHP_C2P_H_1;
	b[ipCHP]     -= b_C_CHP_C2P_H;
	
	c[ipCHP][ipATC] -= C_CHP_C2P_H_2;
	c[ipATC][ipATC] -= C_CHP_C2P_H_1;
	b[ipATC]     -= b_C_CHP_C2P_H;

	c[ipCHP][ipC2P]  += C_CHP_C2P_H_2;
	c[ipATC][ipC2P]  += C_CHP_C2P_H_1;
	b[ipC2P]      += b_C_CHP_C2P_H;

	/* >>refer	mol	CH+		CH	=>		C2+		H2		UMIST	*/
	
	c[ipCH][ipCH] -= CHP_CH_C2P_H2_2;
	c[ipCHP][ipCH] -= CHP_CH_C2P_H2_1;
	b[ipCH]     -= b_CHP_CH_C2P_H2;
	
	c[ipCH][ipCHP] -= CHP_CH_C2P_H2_2;
	c[ipCHP][ipCHP] -= CHP_CH_C2P_H2_1;
	b[ipCHP]     -= b_CHP_CH_C2P_H2;

	c[ipCH][ipC2P]  += CHP_CH_C2P_H2_2;
	c[ipCHP][ipC2P]  += CHP_CH_C2P_H2_1;
	b[ipC2P]      += b_CHP_CH_C2P_H2;

	/* >>refer	mol	N		C2+	=>		CN		C+		UMIST	*/
	
	c[ipC2P][ipC2P] -= N_C2P_CN_CP_2;
	c[ipATN][ipC2P] -= N_C2P_CN_CP_1;
	b[ipC2P]     -= b_N_C2P_CN_CP;
	
	c[ipC2P][ipATN] -= N_C2P_CN_CP_2;
	c[ipATN][ipATN] -= N_C2P_CN_CP_1;
	b[ipATN]     -= b_N_C2P_CN_CP;

	c[ipC2P][ipCN]  += N_C2P_CN_CP_2;
	c[ipATN][ipCN]  += N_C2P_CN_CP_1;
	b[ipCN]      += b_N_C2P_CN_CP;

	c[ipC2P][ipCP]  += N_C2P_CN_CP_2;
	c[ipATN][ipCP]  += N_C2P_CN_CP_1;
	b[ipCP]      += b_N_C2P_CN_CP;

	/* >>refer	mol	O		C2+	=>		CO+		C		UMIST	*/
	
	c[ipC2P][ipC2P] -= O_C2P_COP_C_2;
	c[ipATO][ipC2P] -= O_C2P_COP_C_1;
	b[ipC2P]     -= b_O_C2P_COP_C;
	
	c[ipC2P][ipATO] -= O_C2P_COP_C_2;
	c[ipATC][ipATO] -= O_C2P_COP_C_1;
	b[ipATO]     -= b_O_C2P_COP_C;

	c[ipC2P][ipCOP]  += O_C2P_COP_C_2;
	c[ipATO][ipCOP]  += O_C2P_COP_C_1;
	b[ipCOP]      += b_O_C2P_COP_C;

	c[ipC2P][ipATC]  += O_C2P_COP_C_2;
	c[ipATO][ipATC]  += O_C2P_COP_C_1;
	b[ipATC]      += b_O_C2P_COP_C;

	/* >>refer	mol	C2+		O2	=>		CO+		CO		UMIST	*/
	
	c[ipO2][ipO2] -= C2P_O2_COP_CO_2;
	c[ipC2P][ipO2] -= C2P_O2_COP_CO_1;
	b[ipO2]     -= b_C2P_O2_COP_CO;
	
	c[ipO2][ipC2P] -= C2P_O2_COP_CO_2;
	c[ipC2P][ipC2P] -= C2P_O2_COP_CO_1;
	b[ipC2P]     -= b_C2P_O2_COP_CO;

	c[ipO2][ipCOP]  += C2P_O2_COP_CO_2;
	c[ipC2P][ipCOP]  += C2P_O2_COP_CO_1;
	b[ipCOP]      += b_C2P_O2_COP_CO;

	c[ipO2][ipCO]  += C2P_O2_COP_CO_2;
	c[ipC2P][ipCO]  += C2P_O2_COP_CO_1;
	b[ipCO]      += b_C2P_O2_COP_CO;

	/* >>refer	mol	C2+		S	=>		CS+		C		UMIST	*/
	
	c[ipATS][ipATS] -= C2P_S_CSP_C_2;
	c[ipC2P][ipATS] -= C2P_S_CSP_C_1;
	b[ipATS]     -= b_C2P_S_CSP_C;
	
	c[ipATS][ipC2P] -= C2P_S_CSP_C_2;
	c[ipC2P][ipC2P] -= C2P_S_CSP_C_1;
	b[ipC2P]     -= b_C2P_S_CSP_C;

	c[ipATS][ipCS]  += C2P_S_CSP_C_2;
	c[ipC2P][ipCS]  += C2P_S_CSP_C_1;
	b[ipCS]      += b_C2P_S_CSP_C;

	c[ipATS][ipATC]  += C2P_S_CSP_C_2;
	c[ipC2P][ipATC]  += C2P_S_CSP_C_1;
	b[ipATC]      += b_C2P_S_CSP_C;

	/* >>refer	mol	C2+		e	=>		C		C		UMIST	*/
	
	c[ipC2P][ipC2P] -= C2P_e_C_C;
	c[ipC2P][ipATC]  += C2P_e_C_C;
	c[ipC2P][ipATC]  += C2P_e_C_C;

	/* >>refer	mol	C+		C	=>		C2+		nu		UMIST	*/
	
	c[ipATC][ipATC] -= CP_C_C2P_nu_2;
	c[ipCP][ipATC] -= CP_C_C2P_nu_1;
	b[ipATC]     -= b_CP_C_C2P_nu;
	
	c[ipATC][ipCP] -= CP_C_C2P_nu_2;
	c[ipCP][ipCP] -= CP_C_C2P_nu_1;
	b[ipCP]     -= b_CP_C_C2P_nu;

	c[ipATC][ipC2P]  += CP_C_C2P_nu_2;
	c[ipCP][ipC2P]  += CP_C_C2P_nu_1;
	b[ipC2P]      += b_CP_C_C2P_nu;

	/* >>refer	mol	C2+		nu	=>		C+		C		UMIST	*/
	
	c[ipC2P][ipC2P] -= C2P_nu_CP_C;
	c[ipC2P][ipCP]  += C2P_nu_CP_C;
	c[ipC2P][ipATC]  += C2P_nu_CP_C;
	
	/* end of C2 and C2+ chemistry */
	
	/* >>refer	mol	C	CH2	=>	CH	CH	UMIST	*/
			
	c[ipCH2][ipCH2] -= C_CH2_CH_CH_1;
	c[ipATC][ipCH2] -= C_CH2_CH_CH_2;
	b[ipCH2]     -= b_C_CH2_CH_CH;
			
	c[ipCH2][ipATC] -= C_CH2_CH_CH_1;
	c[ipATC][ipATC] -= C_CH2_CH_CH_2;
	b[ipATC]     -= b_C_CH2_CH_CH;

	c[ipCH2][ipCH]  += 2*C_CH2_CH_CH_1;
	c[ipATC][ipCH]  += 2*C_CH2_CH_CH_2;
	b[ipCH]      += 2*b_C_CH2_CH_CH;
	
	/* >>refer	mol	C	CO+	=>	CO	C+	UMIST	*/
	
	c[ipCOP][ipCOP] -= C_COP_CO_CP_1;
	c[ipATC][ipCOP] -= C_COP_CO_CP_2;
	b[ipCOP]     -= b_C_COP_CO_CP;
	
	c[ipCOP][ipATC] -= C_COP_CO_CP_1;
	c[ipATC][ipATC] -= C_COP_CO_CP_2;
	b[ipATC]     -= b_C_COP_CO_CP;

	c[ipCOP][ipCO]  += C_COP_CO_CP_1;
	c[ipATC][ipCO]  += C_COP_CO_CP_2;
	b[ipCO]      += b_C_COP_CO_CP;

	c[ipCOP][ipCP]  += C_COP_CO_CP_1;
	c[ipATC][ipCP]  += C_COP_CO_CP_2;
	b[ipCP]      += b_C_COP_CO_CP;
	
	/* >>refer	mol	C	H2O+	=>	OH	CH+	UMIST	*/
		
	c[ipH2OP][ipH2OP] -= C_H2OP_OH_CHP_1;
	c[ipATC][ipH2OP] -= C_H2OP_OH_CHP_2;
	b[ipH2OP]     -= b_C_H2OP_OH_CHP;
		
	c[ipH2OP][ipATC] -= C_H2OP_OH_CHP_1;
	c[ipATC][ipATC] -= C_H2OP_OH_CHP_2;
	b[ipATC]     -= b_C_H2OP_OH_CHP;
	
	c[ipH2OP][ipOH]  += C_H2OP_OH_CHP_1;
	c[ipATC][ipOH]  += C_H2OP_OH_CHP_2;
	b[ipOH]      += b_C_H2OP_OH_CHP;
	
	c[ipH2OP][ipCHP] += C_H2OP_OH_CHP_1;
	c[ipATC][ipCHP] += C_H2OP_OH_CHP_2;
	b[ipCHP]     += b_C_H2OP_OH_CHP;
	
	/* >>refer	mol	C	H3O+	=>	HCO+	H2	UMIST	*/
		
	c[ipH3OP][ipH3OP] -= C_H3OP_HCOP_H2_1;
	c[ipATC][ipH3OP]  -= C_H3OP_HCOP_H2_2;
	b[ipH3OP]      -= b_C_H3OP_HCOP_H2;
		
	c[ipH3OP][ipATC] -= C_H3OP_HCOP_H2_1;
	c[ipATC][ipATC]  -= C_H3OP_HCOP_H2_2;
	b[ipATC]      -= b_C_H3OP_HCOP_H2;
	
	c[ipH3OP][ipHCOP] += C_H3OP_HCOP_H2_1;
	c[ipATC][ipHCOP]  += C_H3OP_HCOP_H2_2;
	b[ipHCOP]      += b_C_H3OP_HCOP_H2;
	
	/* >>refer	mol	C	HCO+	=>	CO	CH+	UMIST	*/
	
	c[ipHCOP][ipHCOP] -= C_HCOP_CO_CHP_1;
	c[ipATC][ipHCOP]  -= C_HCOP_CO_CHP_2;
	b[ipHCOP]      -= b_C_HCOP_CO_CHP;
	
	c[ipHCOP][ipATC] -= C_HCOP_CO_CHP_1;
	c[ipATC][ipATC]  -= C_HCOP_CO_CHP_2;
	b[ipATC]      -= b_C_HCOP_CO_CHP;

	c[ipHCOP][ipCO]   += C_HCOP_CO_CHP_1;
	c[ipATC][ipCO]    += C_HCOP_CO_CHP_2;
	b[ipCO]        += b_C_HCOP_CO_CHP;
	
	c[ipHCOP][ipCHP]  += C_HCOP_CO_CHP_1;
	c[ipATC][ipCHP]   += C_HCOP_CO_CHP_2;
	b[ipCHP]       += b_C_HCOP_CO_CHP;
	
	
	/* >>refer	mol	C	O	=>	CO	PHOTON	UMIST	*/
	
	c[ipATO][ipATO] -= C_O_CO_nu_1;
	c[ipATC][ipATO] -= C_O_CO_nu_2;
	b[ipATO]     -= b_C_O_CO_nu;
	
	c[ipATO][ipATC] -= C_O_CO_nu_1;
	c[ipATC][ipATC] -= C_O_CO_nu_2;
	b[ipATC]     -= b_C_O_CO_nu;
	
	c[ipATO][ipCO] += C_O_CO_nu_1;
	c[ipATC][ipCO] += C_O_CO_nu_2;
	b[ipCO]     += b_C_O_CO_nu;
	
	/* >>refer	mol	C	O2	=>	CO	O	UMIST	*/
	
	c[ipO2][ipO2]  -= C_O2_CO_O_1;
	c[ipATC][ipO2] -= C_O2_CO_O_2;
	b[ipO2]     -= b_C_O2_CO_O;
	
	c[ipO2][ipATC]  -= C_O2_CO_O_1;
	c[ipATC][ipATC] -= C_O2_CO_O_2;
	b[ipATC]     -= b_C_O2_CO_O;
	
	c[ipO2][ipCO]  += C_O2_CO_O_1;
	c[ipATC][ipCO] += C_O2_CO_O_2;
	b[ipCO]     += b_C_O2_CO_O;
	
	c[ipO2][ipATO]  += C_O2_CO_O_1;
	c[ipATC][ipATO] += C_O2_CO_O_2;
	b[ipATO]     += b_C_O2_CO_O;
	
	/* >>refer	mol	C	O2+	=>	O2	C+	UMIST	*/
		
	c[ipO2P][ipO2P] -= C_O2P_O2_CP_1;
	c[ipATC][ipO2P] -= C_O2P_O2_CP_2;
	b[ipO2P]     -= b_C_O2P_O2_CP;
		
	c[ipO2P][ipATC] -= C_O2P_O2_CP_1;
	c[ipATC][ipATC] -= C_O2P_O2_CP_2;
	b[ipATC]     -= b_C_O2P_O2_CP;
	
	c[ipO2P][ipO2]  += C_O2P_O2_CP_1;
	c[ipATC][ipO2]  += C_O2P_O2_CP_2;
	b[ipO2]      += b_C_O2P_O2_CP;
	
	c[ipO2P][ipCP]  += C_O2P_O2_CP_1;
	c[ipATC][ipCP]  += C_O2P_O2_CP_2;
	b[ipCP]      += b_C_O2P_O2_CP;
	
	/* >>refer	mol	C	O2+	=>	CO+	O	UMIST	*/
	
	c[ipO2P][ipO2P] -= C_O2P_COP_O_1;
	c[ipATC][ipO2P] -= C_O2P_COP_O_2;
	b[ipO2P]     -= b_C_O2P_COP_O;
	
	c[ipO2P][ipATC] -= C_O2P_COP_O_1;
	c[ipATC][ipATC] -= C_O2P_COP_O_2;
	b[ipATC]     -= b_C_O2P_COP_O;
	
	c[ipO2P][ipCOP] += C_O2P_COP_O_1;
	c[ipATC][ipCOP] += C_O2P_COP_O_2;
	b[ipCOP]     += b_C_O2P_COP_O;
	
	c[ipO2P][ipATO] += C_O2P_COP_O_1;
	c[ipATC][ipATO] += C_O2P_COP_O_2;
	b[ipATO]     += b_C_O2P_COP_O;
	
	/* >>refer	mol	C	OH	=>	O	CH	UMIST	*/
	
	c[ipOH][ipOH]   -= C_OH_O_CH_1;
	c[ipATC][ipOH]  -= C_OH_O_CH_2;
	b[ipOH]      -= b_C_OH_O_CH;
	
	c[ipOH][ipATC]   -= C_OH_O_CH_1;
	c[ipATC][ipATC]  -= C_OH_O_CH_2;
	b[ipATC]      -= b_C_OH_O_CH;
	
	c[ipOH][ipCH]   += C_OH_O_CH_1;
	c[ipATC][ipCH]  += C_OH_O_CH_2;
	b[ipCH]      += b_C_OH_O_CH;
	
	c[ipOH][ipATO]   += C_OH_O_CH_1;
	c[ipATC][ipATO]  += C_OH_O_CH_2;
	b[ipATO]      += b_C_OH_O_CH;
	
	/* >>refer	mol	C	OH	=>	CO	H	UMIST	*/
	
	c[ipOH][ipOH]  -= C_OH_CO_H_1;
	c[ipATC][ipOH] -= C_OH_CO_H_2;
	b[ipOH]     -= b_C_OH_CO_H;
	
	c[ipOH][ipATC]  -= C_OH_CO_H_1;
	c[ipATC][ipATC] -= C_OH_CO_H_2;
	b[ipATC]     -= b_C_OH_CO_H;
	
	c[ipOH][ipCO]  += C_OH_CO_H_1;
	c[ipATC][ipCO] += C_OH_CO_H_2;
	b[ipCO]     += b_C_OH_CO_H;
	
	/* >>refer	mol	C	OH+	=>	O	CH+	UMIST	*/
		
	c[ipOHP][ipOHP] -= C_OHP_O_CHP_1;
	c[ipATC][ipOHP] -= C_OHP_O_CHP_2;
	b[ipOHP]     -= b_C_OHP_O_CHP;
		
	c[ipOHP][ipATC] -= C_OHP_O_CHP_1;
	c[ipATC][ipATC] -= C_OHP_O_CHP_2;
	b[ipATC]     -= b_C_OHP_O_CHP;
	
	c[ipOHP][ipCHP] += C_OHP_O_CHP_1;
	c[ipATC][ipCHP] += C_OHP_O_CHP_2;
	b[ipCHP]     += b_C_OHP_O_CHP;
	
	c[ipOHP][ipATO] += C_OHP_O_CHP_1;
	c[ipATC][ipATO] += C_OHP_O_CHP_2;
	b[ipATO]     += b_C_OHP_O_CHP;
	
	/* >>refer	mol	C	SiO+	=>	Si+	CO	UMIST	*/
		
	c[ipSIOP][ipSIOP] -= C_SiOP_SiP_CO_1;
	c[ipATC][ipSIOP]  -= C_SiOP_SiP_CO_2;
	b[ipSIOP]      -= b_C_SiOP_SiP_CO;
		
	c[ipSIOP][ipATC] -= C_SiOP_SiP_CO_1;
	c[ipATC][ipATC]  -= C_SiOP_SiP_CO_2;
	b[ipATC]      -= b_C_SiOP_SiP_CO;
	
	c[ipSIOP][ipCO]   += C_SiOP_SiP_CO_1;
	c[ipATC][ipCO]    += C_SiOP_SiP_CO_2;
	b[ipCO]        += b_C_SiOP_SiP_CO;
	
	c[ipSIOP][ipSIP]   += C_SiOP_SiP_CO_1;
	c[ipATC][ipSIP]    += C_SiOP_SiP_CO_2;
	b[ipSIP]        += b_C_SiOP_SiP_CO;
	
	/* >>refer	mol	C+	OH	=>	CO	H+	UMIST	*/
	
	c[ipOH][ipCP]  -= CP_OH_CO_HP_1;
	c[ipCP][ipCP] -= CP_OH_CO_HP_2;
	b[ipCP]     -= b_CP_OH_CO_HP;
	
	c[ipOH][ipOH]  -= CP_OH_CO_HP_1;
	c[ipCP][ipOH] -= CP_OH_CO_HP_2;
	b[ipOH]     -= b_CP_OH_CO_HP;
	
	c[ipOH][ipCO]  += CP_OH_CO_HP_1;
	c[ipCP][ipCO] += CP_OH_CO_HP_2;
	b[ipCO]     += b_CP_OH_CO_HP;
	
	/* >>refer	mol	C+	CH	=>	CH+	C	UMIST	*/
		
	c[ipCH][ipCP]  -= CP_CH_CHP_C_1;
	c[ipCP][ipCP]  -= CP_CH_CHP_C_2;
	b[ipCP]     -= b_CP_CH_CHP_C;
		
	c[ipCH][ipCH]  -= CP_CH_CHP_C_1;
	c[ipCP][ipCH]  -= CP_CH_CHP_C_2;
	b[ipCH]     -= b_CP_CH_CHP_C;
	
	c[ipCH][ipCHP] += CP_CH_CHP_C_1;
	c[ipCP][ipCHP] += CP_CH_CHP_C_2;
	b[ipCHP]    += b_CP_CH_CHP_C;
	
	c[ipCH][ipATC] += CP_CH_CHP_C_1;
	c[ipCP][ipATC] += CP_CH_CHP_C_2;
	b[ipATC]    += b_CP_CH_CHP_C;

	/* >>refer	mol	C+	SiO	=>	Si+	CO	UMIST	*/
		
	c[ipSIO][ipCP] -= CP_SiO_SiP_CO_1;
	c[ipCP][ipCP]  -= CP_SiO_SiP_CO_2;
	b[ipCP]     -= b_CP_SiO_SiP_CO;
		
	c[ipSIO][ipSIO] -= CP_SiO_SiP_CO_1;
	c[ipCP][ipSIO]  -= CP_SiO_SiP_CO_2;
	b[ipSIO]     -= b_CP_SiO_SiP_CO;
	
	c[ipSIO][ipCO]  += CP_SiO_SiP_CO_1;
	c[ipCP][ipCO]   += CP_SiO_SiP_CO_2;
	b[ipCO]      += b_CP_SiO_SiP_CO;
	
	c[ipSIO][ipSIP]  += CP_SiO_SiP_CO_1;
	c[ipCP][ipSIP]   += CP_SiO_SiP_CO_2;
	b[ipSIP]      += b_CP_SiO_SiP_CO;
	
	/* >>refer	mol	C+	CH2	=>	CH2+	C	UMIST	*/
		
	c[ipCH2][ipCP]  -= CP_CH2_CH2P_C_1;
	c[ipCP][ipCP]   -= CP_CH2_CH2P_C_2;
	b[ipCP]      -= b_CP_CH2_CH2P_C;
		
	c[ipCH2][ipCH2]  -= CP_CH2_CH2P_C_1;
	c[ipCP][ipCH2]   -= CP_CH2_CH2P_C_2;
	b[ipCH2]      -= b_CP_CH2_CH2P_C;
	
	c[ipCH2][ipCH2P] += CP_CH2_CH2P_C_1;
	c[ipCP][ipCH2P]  += CP_CH2_CH2P_C_2;
	b[ipCH2P]     += b_CP_CH2_CH2P_C;
	
	c[ipCH2][ipATC] += CP_CH2_CH2P_C_1;
	c[ipCP][ipATC]  += CP_CH2_CH2P_C_2;
	b[ipATC]     += b_CP_CH2_CH2P_C;
	
	/* >>refer	mol	C+	H2O	=>	HCO+	H	UMIST	*/
		
	c[ipH2O][ipCP]  -= CP_H2O_HCOP_H_1;
	c[ipCP][ipCP]   -= CP_H2O_HCOP_H_2;
	b[ipCP]      -= b_CP_H2O_HCOP_H;
		
	c[ipH2O][ipH2O]  -= CP_H2O_HCOP_H_1;
	c[ipCP][ipH2O]   -= CP_H2O_HCOP_H_2;
	b[ipH2O]      -= b_CP_H2O_HCOP_H;
	
	c[ipH2O][ipHCOP] += CP_H2O_HCOP_H_1;
	c[ipCP][ipHCOP]  += CP_H2O_HCOP_H_2;
	b[ipHCOP]     += b_CP_H2O_HCOP_H;
	
	/* >>refer	mol	C+	O	=>	CO+	PHOTON	UMIST	*/
		
	c[ipATO][ipATO] -= CP_O_COP_nu_1;
	c[ipCP][ipATO]  -= CP_O_COP_nu_2;
	b[ipATO]     -= b_CP_O_COP_nu;
		
	c[ipATO][ipCP] -= CP_O_COP_nu_1;
	c[ipCP][ipCP]  -= CP_O_COP_nu_2;
	b[ipCP]     -= b_CP_O_COP_nu;
		
	c[ipATO][ipCOP] += CP_O_COP_nu_1;
	c[ipCP][ipCOP]  += CP_O_COP_nu_2;
	b[ipCOP]     += b_CP_O_COP_nu;
	
	/* >>refer	mol	C+	O2	=>	CO+	O	UMIST	*/
	
	c[ipO2][ipCP]  -= CP_O2_COP_O_1;
	c[ipCP][ipCP]  -= CP_O2_COP_O_2;
	b[ipCP]     -= b_CP_O2_COP_O;
	
	c[ipO2][ipO2]  -= CP_O2_COP_O_1;
	c[ipCP][ipO2]  -= CP_O2_COP_O_2;
	b[ipO2]     -= b_CP_O2_COP_O;
	
	c[ipO2][ipCOP] += CP_O2_COP_O_1;
	c[ipCP][ipCOP] += CP_O2_COP_O_2;
	b[ipCOP]    += b_CP_O2_COP_O;
	
	c[ipO2][ipATO] += CP_O2_COP_O_1;
	c[ipCP][ipATO] += CP_O2_COP_O_2;
	b[ipATO]    += b_CP_O2_COP_O;
	
	/* >>refer	mol	C+	O2	=>	CO	O+	UMIST	*/
		
	c[ipO2][ipCP] -= CP_O2_CO_OP_1;
	c[ipCP][ipCP] -= CP_O2_CO_OP_2;
	b[ipCP]    -= b_CP_O2_CO_OP;
		
	c[ipO2][ipO2] -= CP_O2_CO_OP_1;
	c[ipCP][ipO2] -= CP_O2_CO_OP_2;
	b[ipO2]    -= b_CP_O2_CO_OP;
	
	c[ipO2][ipCO] += CP_O2_CO_OP_1;
	c[ipCP][ipCO] += CP_O2_CO_OP_2;
	b[ipCO]    += b_CP_O2_CO_OP;
	
	c[ipO2][ipOP] += CP_O2_CO_OP_1;
	c[ipCP][ipOP] += CP_O2_CO_OP_2;
	b[ipOP]    += b_CP_O2_CO_OP;
	
	/* >>refer	mol	C+	OH	=>	CO+	H	UMIST	*/
		
	c[ipOH][ipCP]  -= CP_OH_COP_H_1;
	c[ipCP][ipCP]  -= CP_OH_COP_H_2;
	b[ipCP]     -= b_CP_OH_COP_H;
		
	c[ipOH][ipOH]  -= CP_OH_COP_H_1;
	c[ipCP][ipOH]  -= CP_OH_COP_H_2;
	b[ipOH]     -= b_CP_OH_COP_H;
	
	c[ipOH][ipCOP] += CP_OH_COP_H_1;
	c[ipCP][ipCOP] += CP_OH_COP_H_2;
	b[ipCOP]    += b_CP_OH_COP_H;
	
	/* >>refer	mol	O	CH	=>	OH	C	UMIST	*/
		
	c[ipCH][ipCH]  -= O_CH_OH_C_1;
	c[ipATO][ipCH] -= O_CH_OH_C_2;
	b[ipCH]     -= b_O_CH_OH_C;
		
	c[ipCH][ipATO]  -= O_CH_OH_C_1;
	c[ipATO][ipATO] -= O_CH_OH_C_2;
	b[ipATO]     -= b_O_CH_OH_C;
	
	c[ipCH][ipOH]  += O_CH_OH_C_1;
	c[ipATO][ipOH] += O_CH_OH_C_2;
	b[ipOH]     += b_O_CH_OH_C;
	
	c[ipCH][ipATC]  += O_CH_OH_C_1;
	c[ipATO][ipATC] += O_CH_OH_C_2;
	b[ipATC]     += b_O_CH_OH_C;
	
	/* >>refer	mol	O	CH	=>	HCO+	e	UMIST	*/
		
	c[ipCH][ipCH]    -= O_CH_HCOP_e_1;
	c[ipATO][ipCH]   -= O_CH_HCOP_e_2;
	b[ipCH]       -= b_O_CH_HCOP_e;
		
	c[ipCH][ipATO]    -= O_CH_HCOP_e_1;
	c[ipATO][ipATO]   -= O_CH_HCOP_e_2;
	b[ipATO]       -= b_O_CH_HCOP_e;
	
	c[ipCH][ipHCOP]  += O_CH_HCOP_e_1;
	c[ipATO][ipHCOP] += O_CH_HCOP_e_2;
	b[ipHCOP]     += b_O_CH_HCOP_e;
	
	/* >>refer	mol	O	CH	=>	CO	H	UMIST	*/
		
	c[ipCH][ipCH]  -= O_CH_CO_H_1;
	c[ipATO][ipCH] -= O_CH_CO_H_2;
	b[ipCH]     -= b_O_CH_CO_H;
		
	c[ipCH][ipATO]  -= O_CH_CO_H_1;
	c[ipATO][ipATO] -= O_CH_CO_H_2;
	b[ipATO]     -= b_O_CH_CO_H;
	
	c[ipCH][ipCO]  += O_CH_CO_H_1;
	c[ipATO][ipCO] += O_CH_CO_H_2;
	b[ipCO]     += b_O_CH_CO_H;
	
	/* >>refer	mol	O	CH+	=>	CO+	H	UMIST	*/
		
	c[ipCHP][ipCHP] -= O_CHP_COP_H_1;
	c[ipATO][ipCHP] -= O_CHP_COP_H_2;
	b[ipCHP]     -= b_O_CHP_COP_H;
		
	c[ipCHP][ipATO] -= O_CHP_COP_H_1;
	c[ipATO][ipATO] -= O_CHP_COP_H_2;
	b[ipATO]     -= b_O_CHP_COP_H;
	
	c[ipCHP][ipCOP] += O_CHP_COP_H_1;
	c[ipATO][ipCOP] += O_CHP_COP_H_2;
	b[ipCOP]     += b_O_CHP_COP_H;
	
	/* >>refer	mol	O	CH2	=>	OH	CH	UMIST	*/
	
	c[ipCH2][ipCH2] -= O_CH2_OH_CH_1;
	c[ipATO][ipCH2] -= O_CH2_OH_CH_2;
	b[ipCH2]     -= b_O_CH2_OH_CH;
	
	c[ipCH2][ipATO] -= O_CH2_OH_CH_1;
	c[ipATO][ipATO] -= O_CH2_OH_CH_2;
	b[ipATO]     -= b_O_CH2_OH_CH;
	
	c[ipCH2][ipOH]  += O_CH2_OH_CH_1;
	c[ipATO][ipOH]  += O_CH2_OH_CH_2;
	b[ipOH]      += b_O_CH2_OH_CH;
	
	c[ipCH2][ipCH]  += O_CH2_OH_CH_1;
	c[ipATO][ipCH]  += O_CH2_OH_CH_2;
	b[ipCH]      += b_O_CH2_OH_CH;
	
	/* >>refer	mol	O	CH2	=>	CO	H	H	UMIST	*/
	
	c[ipCH2][ipCH2] -= O_CH2_CO_H_H_1;
	c[ipATO][ipCH2] -= O_CH2_CO_H_H_2;
	b[ipCH2]     -= b_O_CH2_CO_H_H;
	
	c[ipCH2][ipATO] -= O_CH2_CO_H_H_1;
	c[ipATO][ipATO] -= O_CH2_CO_H_H_2;
	b[ipATO]     -= b_O_CH2_CO_H_H;
	
	c[ipCH2][ipCO]  += O_CH2_CO_H_H_1;
	c[ipATO][ipCO]  += O_CH2_CO_H_H_2;
	b[ipCO]      += b_O_CH2_CO_H_H;
	
	/* >>refer	mol	O	CH2	=>	CO	H2	UMIST	*/
		
	c[ipCH2][ipCH2] -= O_CH2_CO_H2_1;
	c[ipATO][ipCH2] -= O_CH2_CO_H2_2;
	b[ipCH2]     -= b_O_CH2_CO_H2;
		
	c[ipCH2][ipATO] -= O_CH2_CO_H2_1;
	c[ipATO][ipATO] -= O_CH2_CO_H2_2;
	b[ipATO]     -= b_O_CH2_CO_H2;
	
	c[ipCH2][ipCO]  += O_CH2_CO_H2_1;
	c[ipATO][ipCO]  += O_CH2_CO_H2_2;
	b[ipCO]      += b_O_CH2_CO_H2;
	
	/* >>refer	mol	O	CH2+	=>	HCO+	H	UMIST	*/
	
	c[ipCH2P][ipCH2P] -= O_CH2P_HCOP_H_1;
	c[ipATO][ipCH2P]  -= O_CH2P_HCOP_H_2;
	b[ipCH2P]      -= b_O_CH2P_HCOP_H;
	
	c[ipCH2P][ipATO] -= O_CH2P_HCOP_H_1;
	c[ipATO][ipATO]  -= O_CH2P_HCOP_H_2;
	b[ipATO]      -= b_O_CH2P_HCOP_H;
	
	c[ipCH2P][ipHCOP] += O_CH2P_HCOP_H_1;
	c[ipATO][ipHCOP]  += O_CH2P_HCOP_H_2;
	b[ipHCOP]      += b_O_CH2P_HCOP_H;
	
	/* >>refer	mol	O	CH3+	=>	HCO+	H2	UMIST	*/
		
	c[ipCH3P][ipCH3P] -= O_CH3P_HCOP_H2_1;
	c[ipATO][ipCH3P]  -= O_CH3P_HCOP_H2_2;
	b[ipCH3P]      -= b_O_CH3P_HCOP_H2;
		
	c[ipCH3P][ipATO] -= O_CH3P_HCOP_H2_1;
	c[ipATO][ipATO]  -= O_CH3P_HCOP_H2_2;
	b[ipATO]      -= b_O_CH3P_HCOP_H2;
	
	c[ipCH3P][ipHCOP] += O_CH3P_HCOP_H2_1;
	c[ipATO][ipHCOP]  += O_CH3P_HCOP_H2_2;
	b[ipHCOP]      += b_O_CH3P_HCOP_H2;
	
	/* >>refer	mol	O	CO+	=>	CO	O+	UMIST	*/
		
	c[ipCOP][ipCOP] -= O_COP_CO_OP_1;
	c[ipATO][ipCOP] -= O_COP_CO_OP_2;
	b[ipCOP]     -= b_O_COP_CO_OP;
		
	c[ipCOP][ipATO] -= O_COP_CO_OP_1;
	c[ipATO][ipATO] -= O_COP_CO_OP_2;
	b[ipATO]     -= b_O_COP_CO_OP;

	c[ipCOP][ipCO]  += O_COP_CO_OP_1;
	c[ipATO][ipCO]  += O_COP_CO_OP_2;
	b[ipCO]      += b_O_COP_CO_OP;

	c[ipCOP][ipOP]  += O_COP_CO_OP_1;
	c[ipATO][ipOP]  += O_COP_CO_OP_2;
	b[ipOP]      += b_O_COP_CO_OP;
	
	/* >>refer	mol	O	H2O	=>	OH	OH	UMIST	*/
		
	c[ipH2O][ipH2O] -= O_H2O_OH_OH_1;
	c[ipATO][ipH2O] -= O_H2O_OH_OH_2;
	b[ipH2O]     -= b_O_H2O_OH_OH;
		
	c[ipH2O][ipATO] -= O_H2O_OH_OH_1;
	c[ipATO][ipATO] -= O_H2O_OH_OH_2;
	b[ipATO]     -= b_O_H2O_OH_OH;
	
	c[ipH2O][ipOH]  += 2*O_H2O_OH_OH_1;
	c[ipATO][ipOH]  += 2*O_H2O_OH_OH_2;
	b[ipOH]      += 2*b_O_H2O_OH_OH;
	
	/* >>refer	mol	O	H2O+	=>	O2+	H2	UMIST	*/
		
	c[ipH2OP][ipH2OP] -= O_H2OP_O2P_H2_1;
	c[ipATO][ipH2OP]  -= O_H2OP_O2P_H2_2;
	b[ipH2OP]      -= b_O_H2OP_O2P_H2;
		
	c[ipH2OP][ipATO] -= O_H2OP_O2P_H2_1;
	c[ipATO][ipATO]  -= O_H2OP_O2P_H2_2;
	b[ipATO]      -= b_O_H2OP_O2P_H2;
	
	c[ipH2OP][ipO2P]  += O_H2OP_O2P_H2_1;
	c[ipATO][ipO2P]   += O_H2OP_O2P_H2_2;
	b[ipO2P]       += b_O_H2OP_O2P_H2;
	
	/* >>refer	mol	O	O	=>	O2	PHOTON	UMIST	*/
	
	c[ipATO][ipO2] += 0.5*O_O_O2_nu_1;
	c[ipATO][ipO2] += 0.5*O_O_O2_nu_2;
	b[ipO2]     += 0.5*b_O_O_O2_nu;
	
	c[ipATO][ipATO] -= O_O_O2_nu_1;
	c[ipATO][ipATO] -= O_O_O2_nu_2;
	b[ipATO]     -= b_O_O_O2_nu;
	
	/* >>refer	mol	O	OH	=>	O2	H	UMIST	*/
		
	c[ipOH][ipOH]    -= O_OH_O2_H_1;
	c[ipATO][ipOH]   -= O_OH_O2_H_2;
	b[ipOH]       -= b_O_OH_O2_H;
		
	c[ipOH][ipATO]    -= O_OH_O2_H_1;
	c[ipATO][ipATO]   -= O_OH_O2_H_2;
	b[ipATO]       -= b_O_OH_O2_H;
	
	c[ipOH][ipO2]    += O_OH_O2_H_1;
	c[ipATO][ipO2]   += O_OH_O2_H_2;
	b[ipO2]       += b_O_OH_O2_H;
	
	/* >>refer	mol	O	OH+	=>	O2+	H	UMIST	*/
		
	c[ipOHP][ipOHP] -= O_OHP_O2P_H_1;
	c[ipATO][ipOHP] -= O_OHP_O2P_H_2;
	b[ipOHP]     -= b_O_OHP_O2P_H;
		
	c[ipOHP][ipATO] -= O_OHP_O2P_H_1;
	c[ipATO][ipATO] -= O_OHP_O2P_H_2;
	b[ipATO]     -= b_O_OHP_O2P_H;
	
	c[ipOHP][ipO2P] += O_OHP_O2P_H_1;
	c[ipATO][ipO2P] += O_OHP_O2P_H_2;
	b[ipO2P]     += b_O_OHP_O2P_H;
	
	/* >>refer	mol	O	Si	=>	SiO	PHOTON	UMIST	*/
		
	c[ipATSI][ipATO] -= O_Si_SiO_nu_1;
	c[ipATO][ipATO]   -= O_Si_SiO_nu_2;
	b[ipATO]      -= b_O_Si_SiO_nu;
		
	c[ipATSI][ipATSI] -= O_Si_SiO_nu_1;
	c[ipATO][ipATSI]   -= O_Si_SiO_nu_2;
	b[ipATSI]      -= b_O_Si_SiO_nu;
		
	c[ipATSI][ipSIO] += O_Si_SiO_nu_1;
	c[ipATO][ipSIO]   += O_Si_SiO_nu_2;
	b[ipSIO]      += b_O_Si_SiO_nu;
	
	/* >>refer	mol	O	Si+	=>	SiO+	PHOTON	UMIST	*/
			
	c[ipSIP][ipSIP] -= O_SiP_SiOP_nu_1;
	c[ipATO][ipSIP] -= O_SiP_SiOP_nu_2;
	b[ipSIP]     -= b_O_SiP_SiOP_nu;
			
	c[ipSIP][ipATO] -= O_SiP_SiOP_nu_1;
	c[ipATO][ipATO] -= O_SiP_SiOP_nu_2;
	b[ipATO]     -= b_O_SiP_SiOP_nu;
			
	c[ipSIP][ipSIOP] += O_SiP_SiOP_nu_1;
	c[ipATO][ipSIOP] += O_SiP_SiOP_nu_2;
	b[ipSIOP]     += b_O_SiP_SiOP_nu;
	
	/* >>refer	mol	O	SiH	=>	SiO	H	UMIST	*/
		
	c[ipSIH][ipSIH] -= O_SiH_SiO_H_1;
	c[ipATO][ipSIH] -= O_SiH_SiO_H_2;
	b[ipSIH]     -= b_O_SiH_SiO_H;
		
	c[ipSIH][ipATO] -= O_SiH_SiO_H_1;
	c[ipATO][ipATO] -= O_SiH_SiO_H_2;
	b[ipATO]     -= b_O_SiH_SiO_H;
	
	c[ipSIH][ipSIO] += O_SiH_SiO_H_1;
	c[ipATO][ipSIO] += O_SiH_SiO_H_2;
	b[ipSIO]     += b_O_SiH_SiO_H;
	
	/* >>refer	mol	O	SiH2+	=>	SiOH+	H	UMIST	*/
			
	c[ipSIH2P][ipSIH2P] -= O_SiH2P_SiOHP_H_1;
	c[ipATO][ipSIH2P]   -= O_SiH2P_SiOHP_H_2;
	b[ipSIH2P]       -= b_O_SiH2P_SiOHP_H;
			
	c[ipSIH2P][ipATO] -= O_SiH2P_SiOHP_H_1;
	c[ipATO][ipATO]   -= O_SiH2P_SiOHP_H_2;
	b[ipATO]       -= b_O_SiH2P_SiOHP_H;
	
	c[ipSIH2P][ipHOSIP] += O_SiH2P_SiOHP_H_1;
	c[ipATO][ipHOSIP]   += O_SiH2P_SiOHP_H_2;
	b[ipHOSIP]       += b_O_SiH2P_SiOHP_H;
	
	/* >>refer	mol	O	SiO+	=>	O2	Si+	UMIST	*/
			
	c[ipSIOP][ipSIOP] -= O_SiOP_O2_SiP_1;
	c[ipATO][ipSIOP]  -= O_SiOP_O2_SiP_2;
	b[ipSIOP]      -= b_O_SiOP_O2_SiP;
			
	c[ipSIOP][ipATO] -= O_SiOP_O2_SiP_1;
	c[ipATO][ipATO]  -= O_SiOP_O2_SiP_2;
	b[ipATO]      -= b_O_SiOP_O2_SiP;
	
	c[ipSIOP][ipO2]   += O_SiOP_O2_SiP_1;
	c[ipATO][ipO2]    += O_SiOP_O2_SiP_2;
	b[ipO2]        += b_O_SiOP_O2_SiP;
	
	c[ipSIOP][ipSIP]   += O_SiOP_O2_SiP_1;
	c[ipATO][ipSIP]    += O_SiOP_O2_SiP_2;
	b[ipSIP]        += b_O_SiOP_O2_SiP;
	
	/* >>refer	mol	O+	CH	=>	O	CH+	UMIST	*/
		
	
	c[ipCH][ipOP] -= OP_CH_O_CHP_1;
	c[ipOP][ipOP] -= OP_CH_O_CHP_2;
	b[ipOP] -= b_OP_CH_O_CHP;
	
	c[ipCH][ipCH] -= OP_CH_O_CHP_1;
	c[ipOP][ipCH] -= OP_CH_O_CHP_2;
	b[ipCH] -= b_OP_CH_O_CHP;
	
	c[ipCH][ipCHP] += OP_CH_O_CHP_1;
	c[ipOP][ipCHP] += OP_CH_O_CHP_2;
	b[ipCHP] += b_OP_CH_O_CHP;
	
	c[ipCH][ipATO] += OP_CH_O_CHP_1;
	c[ipOP][ipATO] += OP_CH_O_CHP_2;
	b[ipATO] += b_OP_CH_O_CHP;
	
	/* >>refer	mol	O+	CH	=>	CO+	H	UMIST	*/
		
	
	c[ipCH][ipOP] -= OP_CH_COP_H_1;
	c[ipOP][ipOP] -= OP_CH_COP_H_2;
	b[ipOP] -= b_OP_CH_COP_H;
	
	c[ipCH][ipCH] -= OP_CH_COP_H_1;
	c[ipOP][ipCH] -= OP_CH_COP_H_2;
	b[ipCH] -= b_OP_CH_COP_H;
	
	c[ipCH][ipCOP] += OP_CH_COP_H_1;
	c[ipOP][ipCOP] += OP_CH_COP_H_2;
	b[ipCOP] += b_OP_CH_COP_H;
	
	/* >>refer	mol	O+	CH2	=>	O	CH2+	UMIST	*/
		
	
	c[ipCH2][ipOP] -= OP_CH2_O_CH2P_1;
	c[ipOP][ipOP] -= OP_CH2_O_CH2P_2;
	b[ipOP] -= b_OP_CH2_O_CH2P;

	c[ipCH2][ipCH2] -= OP_CH2_O_CH2P_1;
	c[ipOP][ipCH2] -= OP_CH2_O_CH2P_2;
	b[ipCH2] -= b_OP_CH2_O_CH2P;
	
	c[ipCH2][ipCH2P] += OP_CH2_O_CH2P_1;
	c[ipOP][ipCH2P] += OP_CH2_O_CH2P_2;
	b[ipCH2P] += b_OP_CH2_O_CH2P;
	
	c[ipCH2][ipATO] += OP_CH2_O_CH2P_1;
	c[ipOP][ipATO] += OP_CH2_O_CH2P_2;
	b[ipATO] += b_OP_CH2_O_CH2P;
	
	/* >>refer	mol	O+	CO	=>	CO+	O	UMIST	*/
		
	c[ipCO][ipOP] -= OP_CO_COP_O_1;
	c[ipOP][ipOP] -= OP_CO_COP_O_2;
	b[ipOP] -= b_OP_CO_COP_O;
	
	c[ipCO][ipCO] -= OP_CO_COP_O_1;
	c[ipOP][ipCO] -= OP_CO_COP_O_2;
	b[ipCO] -= b_OP_CO_COP_O;
	
	c[ipCO][ipCOP] += OP_CO_COP_O_1;
	c[ipOP][ipCOP] += OP_CO_COP_O_2;
	b[ipCOP] += b_OP_CO_COP_O;
	
	c[ipCO][ipATO] += OP_CO_COP_O_1;
	c[ipOP][ipATO] += OP_CO_COP_O_2;
	b[ipATO] += b_OP_CO_COP_O;
	
	/* >>refer	mol	O+	H2O	=>	H2O+	O	UMIST	*/
		
	
	c[ipH2O][ipOP] -= OP_H2O_H2OP_O_1;
	c[ipOP][ipOP] -= OP_H2O_H2OP_O_2;
	b[ipOP] -= b_OP_H2O_H2OP_O;
	
	c[ipH2O][ipH2O] -= OP_H2O_H2OP_O_1;
	c[ipOP][ipH2O] -= OP_H2O_H2OP_O_2;
	b[ipH2O] -= b_OP_H2O_H2OP_O;
	
	c[ipH2O][ipH2OP] += OP_H2O_H2OP_O_1;
	c[ipOP][ipH2OP] += OP_H2O_H2OP_O_2;
	b[ipH2OP] += b_OP_H2O_H2OP_O;
	
	c[ipH2O][ipATO] += OP_H2O_H2OP_O_1;
	c[ipOP][ipATO] += OP_H2O_H2OP_O_2;
	b[ipATO] += b_OP_H2O_H2OP_O;
	
	/* >>refer	mol	O+	O2	=>	O2+	O	UMIST	*/
		
	
	c[ipO2][ipOP] -= OP_O2_O2P_O_1;
	c[ipOP][ipOP] -= OP_O2_O2P_O_2;
	b[ipOP] -= b_OP_O2_O2P_O;
	
	c[ipO2][ipO2] -= OP_O2_O2P_O_1;
	c[ipOP][ipO2] -= OP_O2_O2P_O_2;
	b[ipO2] -= b_OP_O2_O2P_O;
	
	c[ipO2][ipO2P] += OP_O2_O2P_O_1;
	c[ipOP][ipO2P] += OP_O2_O2P_O_2;
	b[ipO2P] += b_OP_O2_O2P_O;
	
	c[ipO2][ipATO] += OP_O2_O2P_O_1;
	c[ipOP][ipATO] += OP_O2_O2P_O_2;
	b[ipATO] += b_OP_O2_O2P_O;
	
	/* >>refer	mol	O+	OH	=>	O2+	H	UMIST	*/
		
	
	c[ipOH][ipOP] -= OP_OH_O2P_H_1;
	c[ipOP][ipOP] -= OP_OH_O2P_H_2;
	b[ipOP] -= b_OP_OH_O2P_H;
	
	c[ipOH][ipOH] -= OP_OH_O2P_H_1;
	c[ipOP][ipOH] -= OP_OH_O2P_H_2;
	b[ipOH] -= b_OP_OH_O2P_H;
	
	
	c[ipOH][ipO2P] += OP_OH_O2P_H_1;
	c[ipOP][ipO2P] += OP_OH_O2P_H_2;
	b[ipO2P] += b_OP_OH_O2P_H;
	
	/* >>refer	mol	O+	OH	=>	OH+	O	UMIST	*/
		
	
	c[ipOH][ipOP] -= OP_OH_OHP_O_1;
	c[ipOP][ipOP] -= OP_OH_OHP_O_2;
	b[ipOP] -= b_OP_OH_OHP_O;
	
	c[ipOH][ipOH] -= OP_OH_OHP_O_1;
	c[ipOP][ipOH] -= OP_OH_OHP_O_2;
	b[ipOH] -= b_OP_OH_OHP_O;
	
	c[ipOH][ipOHP] += OP_OH_OHP_O_1;
	c[ipOP][ipOHP] += OP_OH_OHP_O_2;
	b[ipOHP] += b_OP_OH_OHP_O;
	
	c[ipOH][ipATO] += OP_OH_OHP_O_1;
	c[ipOP][ipATO] += OP_OH_OHP_O_2;
	b[ipATO] += b_OP_OH_OHP_O;
	
	/* >>refer	mol	Si	CH+	=>	Si+	CH	UMIST	*/
		
	c[ipCHP][ipCHP] -= Si_CHP_SiP_CH_1;
	c[ipATSI][ipCHP] -= Si_CHP_SiP_CH_2;
	b[ipCHP] -= b_Si_CHP_SiP_CH;
		
	c[ipCHP][ipATSI] -= Si_CHP_SiP_CH_1;
	c[ipATSI][ipATSI] -= Si_CHP_SiP_CH_2;
	b[ipATSI] -= b_Si_CHP_SiP_CH;
	
	c[ipCHP][ipCH] += Si_CHP_SiP_CH_1;
	c[ipATSI][ipCH] += Si_CHP_SiP_CH_2;
	b[ipCH] += b_Si_CHP_SiP_CH;
	
	c[ipCHP][ipSIP] += Si_CHP_SiP_CH_1;
	c[ipATSI][ipSIP] += Si_CHP_SiP_CH_2;
	b[ipSIP] += b_Si_CHP_SiP_CH;
	
	/* >>refer	mol	Si	H2O+	=>	Si+	H2O	UMIST	*/
			
	c[ipH2OP][ipH2OP] -= Si_H2OP_SiP_H2O_1;
	c[ipATSI][ipH2OP] -= Si_H2OP_SiP_H2O_2;
	b[ipH2OP] -= b_Si_H2OP_SiP_H2O;
			
	c[ipH2OP][ipATSI] -= Si_H2OP_SiP_H2O_1;
	c[ipATSI][ipATSI] -= Si_H2OP_SiP_H2O_2;
	b[ipATSI] -= b_Si_H2OP_SiP_H2O;
	
	c[ipH2OP][ipSIP] += Si_H2OP_SiP_H2O_1;
	c[ipATSI][ipSIP] += Si_H2OP_SiP_H2O_2;
	b[ipSIP] += b_Si_H2OP_SiP_H2O;

	c[ipH2OP][ipH2O] += Si_H2OP_SiP_H2O_1;
	c[ipATSI][ipH2O] += Si_H2OP_SiP_H2O_2;
	b[ipH2O] += b_Si_H2OP_SiP_H2O;
	
	/* >>refer	mol	Si	OH	=>	SiO	H	UMIST	*/
		
	c[ipOH][ipOH] -= Si_OH_SiO_H_1;
	c[ipATSI][ipOH] -= Si_OH_SiO_H_2;
	b[ipOH] -= b_Si_OH_SiO_H;
		
	c[ipOH][ipATSI] -= Si_OH_SiO_H_1;
	c[ipATSI][ipATSI] -= Si_OH_SiO_H_2;
	b[ipATSI] -= b_Si_OH_SiO_H;
	
	c[ipOH][ipSIO] += Si_OH_SiO_H_1;
	c[ipATSI][ipSIO] += Si_OH_SiO_H_2;
	b[ipSIO] += b_Si_OH_SiO_H;
	
	/* >>refer	mol	Si	O2+	=>	O2	Si+	UMIST	*/
		
	c[ipO2P][ipO2P] -= Si_O2P_O2_SiP_1;
	c[ipATSI][ipO2P] -= Si_O2P_O2_SiP_2;
	b[ipO2P] -= b_Si_O2P_O2_SiP;
			
	c[ipO2P][ipATSI] -= Si_O2P_O2_SiP_1;
	c[ipATSI][ipATSI] -= Si_O2P_O2_SiP_2;
	b[ipATSI] -= b_Si_O2P_O2_SiP;
	
	c[ipO2P][ipO2] += Si_O2P_O2_SiP_1;
	c[ipATSI][ipO2] += Si_O2P_O2_SiP_2;
	b[ipO2] += b_Si_O2P_O2_SiP;
	
	c[ipO2P][ipSIP] += Si_O2P_O2_SiP_1;
	c[ipATSI][ipSIP] += Si_O2P_O2_SiP_2;
	b[ipSIP] += b_Si_O2P_O2_SiP;
	
	/* >>refer	mol	Si+	H2O	=>	SiOH+	H	UMIST	*/
		
	c[ipH2O][ipSIP] -= SiP_H2O_SiOHP_H_1;
	c[ipSIP][ipSIP] -= SiP_H2O_SiOHP_H_2;
	b[ipSIP] -= b_SiP_H2O_SiOHP_H;
	
	c[ipH2O][ipH2O] -= SiP_H2O_SiOHP_H_1;
	c[ipSIP][ipH2O] -= SiP_H2O_SiOHP_H_2;
	b[ipH2O] -= b_SiP_H2O_SiOHP_H;
	
	c[ipH2O][ipHOSIP] += SiP_H2O_SiOHP_H_1;
	c[ipSIP][ipHOSIP] += SiP_H2O_SiOHP_H_2;
	b[ipHOSIP] += b_SiP_H2O_SiOHP_H;
	
	/* >>refer	mol	Si+	OH	=>	SiO+	H	UMIST	*/
		
	c[ipOH][ipSIP] -= SiP_OH_SiOP_H_1;
	c[ipSIP][ipSIP] -= SiP_OH_SiOP_H_2;
	b[ipSIP] -= b_SiP_OH_SiOP_H;
	
	c[ipOH][ipOH] -= SiP_OH_SiOP_H_1;
	c[ipSIP][ipOH] -= SiP_OH_SiOP_H_2;
	b[ipOH] -= b_SiP_OH_SiOP_H;
	
	c[ipOH][ipSIOP] += SiP_OH_SiOP_H_1;
	c[ipSIP][ipSIOP] += SiP_OH_SiOP_H_2;
	b[ipSIOP] += b_SiP_OH_SiOP_H;
	
	/* >>refer	mol	Si+	O2	=>	SiO+	O	UMIST	*/
			
	c[ipO2][ipSIP] -= SiP_O2_SiOP_O_1; 
	c[ipSIP][ipSIP] -= SiP_O2_SiOP_O_2;
	b[ipSIP] -= b_SiP_O2_SiOP_O;
	
	c[ipO2][ipO2] -= SiP_O2_SiOP_O_1; 
	c[ipSIP][ipO2] -= SiP_O2_SiOP_O_2;
	b[ipO2] -= b_SiP_O2_SiOP_O;
	
	c[ipO2][ipSIOP] += SiP_O2_SiOP_O_1; 
	c[ipSIP][ipSIOP] += SiP_O2_SiOP_O_2;
	b[ipSIOP] += b_SiP_O2_SiOP_O;
		
	c[ipO2][ipATO] += SiP_O2_SiOP_O_1; 
	c[ipSIP][ipATO] += SiP_O2_SiOP_O_2;
	b[ipATO] += b_SiP_O2_SiOP_O;
	
	/* >>refer	mol	CH	CO+	=>	HCO+	C	UMIST	*/
			
	c[ipCOP][ipCH] -= CH_COP_HCOP_C_1; 
	c[ipCH][ipCH] -= CH_COP_HCOP_C_2;
	b[ipCH] -= b_CH_COP_HCOP_C;
	
	c[ipCOP][ipCOP] -= CH_COP_HCOP_C_1; 
	c[ipCH][ipCOP] -= CH_COP_HCOP_C_2;
	b[ipCOP] -= b_CH_COP_HCOP_C;
	
	c[ipCOP][ipHCOP] += CH_COP_HCOP_C_1; 
	c[ipCH][ipHCOP] += CH_COP_HCOP_C_2;
	b[ipHCOP] += b_CH_COP_HCOP_C;
	
	c[ipCOP][ipATC] += CH_COP_HCOP_C_1; 
	c[ipCH][ipATC] += CH_COP_HCOP_C_2;
	b[ipATC] += b_CH_COP_HCOP_C;
	
	/* >>refer	mol	CH	CO+	=>	CO	CH+	UMIST	*/
	
	c[ipCOP][ipCH] -= CH_COP_CO_CHP_1; 
	c[ipCH][ipCH] -= CH_COP_CO_CHP_2;
	b[ipCH] -= b_CH_COP_CO_CHP;
	
	c[ipCOP][ipCOP] -= CH_COP_CO_CHP_1; 
	c[ipCH][ipCOP] -= CH_COP_CO_CHP_2;
	b[ipCOP] -= b_CH_COP_CO_CHP;
	
	c[ipCOP][ipCO] += CH_COP_CO_CHP_1; 
	c[ipCH][ipCO] += CH_COP_CO_CHP_2;
	b[ipCO] += b_CH_COP_CO_CHP;
	
	c[ipCOP][ipCHP] += CH_COP_CO_CHP_1; 
	c[ipCH][ipCHP] += CH_COP_CO_CHP_2;
	b[ipCHP] += b_CH_COP_CO_CHP;
	
	/* >>refer	mol	CH	H2O+	=>	H2O	CH+	UMIST	*/
		
	c[ipH2OP][ipCH] -= CH_H2OP_H2O_CHP_1; 
	c[ipCH][ipCH] -= CH_H2OP_H2O_CHP_2;
	b[ipCH] -= b_CH_H2OP_H2O_CHP;
	
	c[ipH2OP][ipH2OP] -= CH_H2OP_H2O_CHP_1; 
	c[ipCH][ipH2OP] -= CH_H2OP_H2O_CHP_2;
	b[ipH2OP] -= b_CH_H2OP_H2O_CHP;
	
	c[ipH2OP][ipH2O] += CH_H2OP_H2O_CHP_1; 
	c[ipCH][ipH2O] += CH_H2OP_H2O_CHP_2;
	b[ipH2O] += b_CH_H2OP_H2O_CHP;
	
	c[ipH2OP][ipCHP] += CH_H2OP_H2O_CHP_1; 
	c[ipCH][ipCHP] += CH_H2OP_H2O_CHP_2;
	b[ipCHP] += b_CH_H2OP_H2O_CHP;
	
	/* >>refer	mol	CH	H2O+	=>	OH	CH2+	UMIST	*/
		
	c[ipH2OP][ipCH] -= CH_H2OP_OH_CH2P_1; 
	c[ipCH][ipCH] -= CH_H2OP_OH_CH2P_2;
	b[ipCH] -= b_CH_H2OP_OH_CH2P;
	
	c[ipH2OP][ipH2OP] -= CH_H2OP_OH_CH2P_1; 
	c[ipCH][ipH2OP] -= CH_H2OP_OH_CH2P_2;
	b[ipH2OP] -= b_CH_H2OP_OH_CH2P;
	
	c[ipH2OP][ipOH] += CH_H2OP_OH_CH2P_1; 
	c[ipCH][ipOH] += CH_H2OP_OH_CH2P_2;
	b[ipOH] += b_CH_H2OP_OH_CH2P;
	
	c[ipH2OP][ipCH2P] += CH_H2OP_OH_CH2P_1; 
	c[ipCH][ipCH2P] += CH_H2OP_OH_CH2P_2;
	b[ipCH2P] += b_CH_H2OP_OH_CH2P;
	
	
	/* >>refer	mol	CH	H3O+	=>	H2O	CH2+	UMIST	*/
		
	c[ipH3OP][ipCH] -= CH_H3OP_H2O_CH2P_1; 
	c[ipCH][ipCH] -= CH_H3OP_H2O_CH2P_2;
	b[ipCH] -= b_CH_H3OP_H2O_CH2P;
	
	c[ipH3OP][ipH3OP] -= CH_H3OP_H2O_CH2P_1; 
	c[ipCH][ipH3OP] -= CH_H3OP_H2O_CH2P_2;
	b[ipH3OP] -= b_CH_H3OP_H2O_CH2P;
	
	c[ipH3OP][ipH2O] += CH_H3OP_H2O_CH2P_1; 
	c[ipCH][ipH2O] += CH_H3OP_H2O_CH2P_2;
	b[ipH2O] += b_CH_H3OP_H2O_CH2P;
	
	c[ipH3OP][ipCH2P] += CH_H3OP_H2O_CH2P_1; 
	c[ipCH][ipCH2P] += CH_H3OP_H2O_CH2P_2;
	b[ipCH2P] += b_CH_H3OP_H2O_CH2P;
	
	/* >>refer	mol	CH	HCO+	=>	CO	CH2+	UMIST	*/
		
	c[ipHCOP][ipCH] -= CH_HCOP_CO_CH2P_1; 
	c[ipCH][ipCH] -= CH_HCOP_CO_CH2P_2;
	b[ipCH] -= b_CH_HCOP_CO_CH2P;
	
	c[ipHCOP][ipHCOP] -= CH_HCOP_CO_CH2P_1; 
	c[ipCH][ipHCOP] -= CH_HCOP_CO_CH2P_2;
	b[ipHCOP] -= b_CH_HCOP_CO_CH2P;
	
	c[ipHCOP][ipCO] += CH_HCOP_CO_CH2P_1; 
	c[ipCH][ipCO] += CH_HCOP_CO_CH2P_2;
	b[ipCO] += b_CH_HCOP_CO_CH2P;
	
	c[ipHCOP][ipCH2P] += CH_HCOP_CO_CH2P_1; 
	c[ipCH][ipCH2P] += CH_HCOP_CO_CH2P_2;
	b[ipCH2P] += b_CH_HCOP_CO_CH2P;
	
	/* >>refer	mol	CH	O2	=>	CO	OH	UMIST	*/
		
	c[ipO2][ipCH] -= CH_O2_CO_OH_1; 
	c[ipCH][ipCH] -= CH_O2_CO_OH_2;
	b[ipCH] -= b_CH_O2_CO_OH;
	
	c[ipO2][ipO2] -= CH_O2_CO_OH_1; 
	c[ipCH][ipO2] -= CH_O2_CO_OH_2;
	b[ipO2] -= b_CH_O2_CO_OH;
	
	c[ipO2][ipCO] += CH_O2_CO_OH_1; 
	c[ipCH][ipCO] += CH_O2_CO_OH_2;
	b[ipCO] += b_CH_O2_CO_OH;
	
	c[ipO2][ipOH] += CH_O2_CO_OH_1; 
	c[ipCH][ipOH] += CH_O2_CO_OH_2;
	b[ipOH] += b_CH_O2_CO_OH;
	
	/* >>refer	mol	CH	O2+	=>	O2	CH+	UMIST	*/
		
	c[ipO2P][ipCH] -= CH_O2P_O2_CHP_1; 
	c[ipCH][ipCH] -= CH_O2P_O2_CHP_2;
	b[ipCH] -= b_CH_O2P_O2_CHP;
	
	c[ipO2P][ipO2P] -= CH_O2P_O2_CHP_1; 
	c[ipCH][ipO2P] -= CH_O2P_O2_CHP_2;
	b[ipO2P] -= b_CH_O2P_O2_CHP;
	
	c[ipO2P][ipO2] += CH_O2P_O2_CHP_1; 
	c[ipCH][ipO2] += CH_O2P_O2_CHP_2;
	b[ipO2] += b_CH_O2P_O2_CHP;
	
	c[ipO2P][ipCHP] += CH_O2P_O2_CHP_1; 
	c[ipCH][ipCHP] += CH_O2P_O2_CHP_2;
	b[ipCHP] += b_CH_O2P_O2_CHP;
	
	/* >>refer	mol	CH	O2+	=>	HCO+	O	UMIST	*/
		
	c[ipO2P][ipCH] -= CH_O2P_HCOP_O_1; 
	c[ipCH][ipCH] -= CH_O2P_HCOP_O_2;
	b[ipCH] -= b_CH_O2P_HCOP_O;
	
	c[ipO2P][ipO2P] -= CH_O2P_HCOP_O_1; 
	c[ipCH][ipO2P] -= CH_O2P_HCOP_O_2;
	b[ipO2P] -= b_CH_O2P_HCOP_O;
	
	c[ipO2P][ipHCOP] += CH_O2P_HCOP_O_1; 
	c[ipCH][ipHCOP] += CH_O2P_HCOP_O_2;
	b[ipHCOP] += b_CH_O2P_HCOP_O;
	
	c[ipO2P][ipATO] += CH_O2P_HCOP_O_1; 
	c[ipCH][ipATO] += CH_O2P_HCOP_O_2;
	b[ipATO] += b_CH_O2P_HCOP_O;
	
	/* >>refer	mol	CH	OH+	=>	OH	CH+	UMIST	*/
		
	c[ipOHP][ipCH] -= CH_OHP_OH_CHP_1; 
	c[ipCH][ipCH] -= CH_OHP_OH_CHP_2;
	b[ipCH] -= b_CH_OHP_OH_CHP;
	
	c[ipOHP][ipOHP] -= CH_OHP_OH_CHP_1; 
	c[ipCH][ipOHP] -= CH_OHP_OH_CHP_2;
	b[ipOHP] -= b_CH_OHP_OH_CHP;
	
	c[ipOHP][ipOH] += CH_OHP_OH_CHP_1; 
	c[ipCH][ipOH] += CH_OHP_OH_CHP_2;
	b[ipOH] += b_CH_OHP_OH_CHP;
	
	c[ipOHP][ipCHP] += CH_OHP_OH_CHP_1; 
	c[ipCH][ipCHP] += CH_OHP_OH_CHP_2;
	b[ipCHP] += b_CH_OHP_OH_CHP;
	
	/* >>refer	mol	CH	OH+	=>	O	CH2+	UMIST	*/
	
	c[ipOHP][ipCH] -= CH_OHP_O_CH2P_1; 
	c[ipCH][ipCH] -= CH_OHP_O_CH2P_2;
	b[ipCH] -= b_CH_OHP_O_CH2P;
	
	c[ipOHP][ipOHP] -= CH_OHP_O_CH2P_1; 
	c[ipCH][ipOHP] -= CH_OHP_O_CH2P_2;
	b[ipOHP] -= b_CH_OHP_O_CH2P;
	
	c[ipOHP][ipCH2P] += CH_OHP_O_CH2P_1; 
	c[ipCH][ipCH2P] += CH_OHP_O_CH2P_2;
	b[ipCH2P] += b_CH_OHP_O_CH2P;
	
	c[ipOHP][ipATO] += CH_OHP_O_CH2P_1; 
	c[ipCH][ipATO] += CH_OHP_O_CH2P_2;
	b[ipATO] += b_CH_OHP_O_CH2P;
	
	/* >>refer	mol	CH	SiO+	=>	HCO+	Si	UMIST	*/
	
	c[ipSIOP][ipCH] -= CH_SiOP_HCOP_Si_1; 
	c[ipCH][ipCH] -= CH_SiOP_HCOP_Si_2;
	b[ipCH] -= b_CH_SiOP_HCOP_Si;
	
	c[ipSIOP][ipSIOP] -= CH_SiOP_HCOP_Si_1; 
	c[ipCH][ipSIOP] -= CH_SiOP_HCOP_Si_2;
	b[ipSIOP] -= b_CH_SiOP_HCOP_Si;
	
	c[ipSIOP][ipHCOP] += CH_SiOP_HCOP_Si_1; 
	c[ipCH][ipHCOP] += CH_SiOP_HCOP_Si_2;
	b[ipHCOP] += b_CH_SiOP_HCOP_Si;
		
	c[ipSIOP][ipATSI] += CH_SiOP_HCOP_Si_1; 
	c[ipCH][ipATSI] += CH_SiOP_HCOP_Si_2;
	b[ipATSI] += b_CH_SiOP_HCOP_Si;
	
	/* >>refer	mol	CH+	H2O	=>	H3O+	C	UMIST	*/
	
	c[ipH2O][ipCHP] -= CHP_H2O_H3OP_C_1; 
	c[ipCHP][ipCHP] -= CHP_H2O_H3OP_C_2;
	b[ipCHP] -= b_CHP_H2O_H3OP_C;
	
	c[ipH2O][ipH2O] -= CHP_H2O_H3OP_C_1; 
	c[ipCHP][ipH2O] -= CHP_H2O_H3OP_C_2;
	b[ipH2O] -= b_CHP_H2O_H3OP_C;
	
	c[ipH2O][ipH3OP] += CHP_H2O_H3OP_C_1; 
	c[ipCHP][ipH3OP] += CHP_H2O_H3OP_C_2;
	b[ipH3OP] += b_CHP_H2O_H3OP_C;
	
	c[ipH2O][ipATC] += CHP_H2O_H3OP_C_1; 
	c[ipCHP][ipATC] += CHP_H2O_H3OP_C_2;
	b[ipATC] += b_CHP_H2O_H3OP_C;
	
	/* >>refer	mol	CH+	H2O	=>	HCO+	H2	UMIST	*/
		
	c[ipH2O][ipCHP] -= CHP_H2O_HCOP_H2_1; 
	c[ipCHP][ipCHP] -= CHP_H2O_HCOP_H2_2;
	b[ipCHP] -= b_CHP_H2O_HCOP_H2;
	
	c[ipH2O][ipH2O] -= CHP_H2O_HCOP_H2_1; 
	c[ipCHP][ipH2O] -= CHP_H2O_HCOP_H2_2;
	b[ipH2O] -= b_CHP_H2O_HCOP_H2;
	
	c[ipH2O][ipHCOP] += CHP_H2O_HCOP_H2_1; 
	c[ipCHP][ipHCOP] += CHP_H2O_HCOP_H2_2;
	b[ipHCOP] += b_CHP_H2O_HCOP_H2;
	
	/* >>refer	mol	CH+	O2	=>	HCO+	O	UMIST	*/
	
	c[ipO2][ipCHP] -= CHP_O2_HCOP_O_1; 
	c[ipCHP][ipCHP] -= CHP_O2_HCOP_O_2;
	b[ipCHP] -= b_CHP_O2_HCOP_O;
	
	c[ipO2][ipO2] -= CHP_O2_HCOP_O_1; 
	c[ipCHP][ipO2] -= CHP_O2_HCOP_O_2;
	b[ipO2] -= b_CHP_O2_HCOP_O;
	
	c[ipO2][ipHCOP] += CHP_O2_HCOP_O_1; 
	c[ipCHP][ipHCOP] += CHP_O2_HCOP_O_2;
	b[ipHCOP] += b_CHP_O2_HCOP_O;
	
	c[ipO2][ipATO] += CHP_O2_HCOP_O_1; 
	c[ipCHP][ipATO] += CHP_O2_HCOP_O_2;
	b[ipATO] += b_CHP_O2_HCOP_O;


	/* >>refer	mol	CH+	OH	=>	CO+	H2	UMIST	*/
	
	c[ipOH][ipCHP] -= CHP_OH_COP_H2_1; 
	c[ipCHP][ipCHP] -= CHP_OH_COP_H2_2;
	b[ipCHP] -= b_CHP_OH_COP_H2;
	
	c[ipOH][ipOH] -= CHP_OH_COP_H2_1; 
	c[ipCHP][ipOH] -= CHP_OH_COP_H2_2;
	b[ipOH] -= b_CHP_OH_COP_H2;
	
	c[ipOH][ipCOP] += CHP_OH_COP_H2_1; 
	c[ipCHP][ipCOP] += CHP_OH_COP_H2_2;
	b[ipCOP] += b_CHP_OH_COP_H2;

	
	/* >>refer	mol	CH+	O2	=>	CO+	OH	UMIST	*/
	
	c[ipO2][ipCHP] -= CHP_O2_COP_OH_1; 
	c[ipCHP][ipCHP] -= CHP_O2_COP_OH_2;
	b[ipCHP] -= b_CHP_O2_COP_OH;
	
	c[ipO2][ipO2] -= CHP_O2_COP_OH_1; 
	c[ipCHP][ipO2] -= CHP_O2_COP_OH_2;
	b[ipO2] -= b_CHP_O2_COP_OH;
	
	c[ipO2][ipCOP] += CHP_O2_COP_OH_1; 
	c[ipCHP][ipCOP] += CHP_O2_COP_OH_2;
	b[ipCOP] += b_CHP_O2_COP_OH;
	
	c[ipO2][ipOH] += CHP_O2_COP_OH_1; 
	c[ipCHP][ipOH] += CHP_O2_COP_OH_2;
	b[ipOH] += b_CHP_O2_COP_OH;
	
	/* >>refer	mol	CH2	CO+	=>	HCO+	CH	UMIST	*/
	
	c[ipCOP][ipCH2] -= CH2_COP_HCOP_CH_1; 
	c[ipCH2][ipCH2] -= CH2_COP_HCOP_CH_2;
	b[ipCH2] -= b_CH2_COP_HCOP_CH;
	
	c[ipCOP][ipCOP] -= CH2_COP_HCOP_CH_1; 
	c[ipCH2][ipCOP] -= CH2_COP_HCOP_CH_2;
	b[ipCOP] -= b_CH2_COP_HCOP_CH;
	
	c[ipCOP][ipHCOP] += CH2_COP_HCOP_CH_1; 
	c[ipCH2][ipHCOP] += CH2_COP_HCOP_CH_2;
	b[ipHCOP] += b_CH2_COP_HCOP_CH;
	
	c[ipCOP][ipCH] += CH2_COP_HCOP_CH_1; 
	c[ipCH2][ipCH] += CH2_COP_HCOP_CH_2;
	b[ipCH] += b_CH2_COP_HCOP_CH;
	
	/* >>refer	mol	CH2	CO+	=>	CO	CH2+	UMIST	*/
	
	c[ipCOP][ipCH2] -= CH2_COP_CO_CH2P_1; 
	c[ipCH2][ipCH2] -= CH2_COP_CO_CH2P_2;
	b[ipCH2] -= b_CH2_COP_CO_CH2P;
	
	c[ipCOP][ipCOP] -= CH2_COP_CO_CH2P_1; 
	c[ipCH2][ipCOP] -= CH2_COP_CO_CH2P_2;
	b[ipCOP] -= b_CH2_COP_CO_CH2P;
	
	c[ipCOP][ipCO] += CH2_COP_CO_CH2P_1; 
	c[ipCH2][ipCO] += CH2_COP_CO_CH2P_2;
	b[ipCO] += b_CH2_COP_CO_CH2P;
	
	c[ipCOP][ipCH2P] += CH2_COP_CO_CH2P_1; 
	c[ipCH2][ipCH2P] += CH2_COP_CO_CH2P_2;
	b[ipCH2P] += b_CH2_COP_CO_CH2P;

	/* >>refer	mol	CH2	H2O+	=>	H2O	CH2+	UMIST	*/
	
	c[ipH2OP][ipCH2] -= CH2_H2OP_H2O_CH2P_1; 
	c[ipCH2][ipCH2] -= CH2_H2OP_H2O_CH2P_2;
	b[ipCH2] -= b_CH2_H2OP_H2O_CH2P;
	
	c[ipH2OP][ipH2OP] -= CH2_H2OP_H2O_CH2P_1; 
	c[ipCH2][ipH2OP] -= CH2_H2OP_H2O_CH2P_2;
	b[ipH2OP] -= b_CH2_H2OP_H2O_CH2P;
	
	c[ipH2OP][ipH2O] += CH2_H2OP_H2O_CH2P_1; 
	c[ipCH2][ipH2O] += CH2_H2OP_H2O_CH2P_2;
	b[ipH2O] += b_CH2_H2OP_H2O_CH2P;
	
	c[ipH2OP][ipCH2P] += CH2_H2OP_H2O_CH2P_1; 
	c[ipCH2][ipCH2P] += CH2_H2OP_H2O_CH2P_2;
	b[ipCH2P] += b_CH2_H2OP_H2O_CH2P;
	
	/* >>refer	mol	CH2	H2O+	=>	OH	CH3+	UMIST	*/
	
	c[ipH2OP][ipCH2] -= CH2_H2OP_OH_CH3P_1; 
	c[ipCH2][ipCH2] -= CH2_H2OP_OH_CH3P_2;
	b[ipCH2] -= b_CH2_H2OP_OH_CH3P;
	
	c[ipH2OP][ipH2OP] -= CH2_H2OP_OH_CH3P_1; 
	c[ipCH2][ipH2OP] -= CH2_H2OP_OH_CH3P_2;
	b[ipH2OP] -= b_CH2_H2OP_OH_CH3P;
	
	c[ipH2OP][ipOH] += CH2_H2OP_OH_CH3P_1; 
	c[ipCH2][ipOH] += CH2_H2OP_OH_CH3P_2;
	b[ipOH] += b_CH2_H2OP_OH_CH3P;
	
	c[ipH2OP][ipCH3P] += CH2_H2OP_OH_CH3P_1; 
	c[ipCH2][ipCH3P] += CH2_H2OP_OH_CH3P_2;
	b[ipCH3P] += b_CH2_H2OP_OH_CH3P;
	
	/* >>refer	mol	CH2	H3O+	=>	H2O	CH3+	UMIST	*/
	
	c[ipH3OP][ipCH2] -= CH2_H3OP_H2O_CH3P_1; 
	c[ipCH2][ipCH2] -= CH2_H3OP_H2O_CH3P_2;
	b[ipCH2] -= b_CH2_H3OP_H2O_CH3P;
	
	c[ipH3OP][ipH3OP] -= CH2_H3OP_H2O_CH3P_1; 
	c[ipCH2][ipH3OP] -= CH2_H3OP_H2O_CH3P_2;
	b[ipH3OP] -= b_CH2_H3OP_H2O_CH3P;
	
	c[ipH3OP][ipH2O] += CH2_H3OP_H2O_CH3P_1; 
	c[ipCH2][ipH2O] += CH2_H3OP_H2O_CH3P_2;
	b[ipH2O] += b_CH2_H3OP_H2O_CH3P;
	
	c[ipH3OP][ipCH3P] += CH2_H3OP_H2O_CH3P_1; 
	c[ipCH2][ipCH3P] += CH2_H3OP_H2O_CH3P_2;
	b[ipCH3P] += b_CH2_H3OP_H2O_CH3P;
	
	/* >>refer	mol	CH2	HCO+	=>	CO	CH3+	UMIST	*/
	
	c[ipHCOP][ipCH2] -= CH2_HCOP_CO_CH3P_1; 
	c[ipCH2][ipCH2] -= CH2_HCOP_CO_CH3P_2;
	b[ipCH2] -= b_CH2_HCOP_CO_CH3P;
	
	c[ipHCOP][ipHCOP] -= CH2_HCOP_CO_CH3P_1; 
	c[ipCH2][ipHCOP] -= CH2_HCOP_CO_CH3P_2;
	b[ipHCOP] -= b_CH2_HCOP_CO_CH3P;
	
	c[ipHCOP][ipCO] += CH2_HCOP_CO_CH3P_1; 
	c[ipCH2][ipCO] += CH2_HCOP_CO_CH3P_2;
	b[ipCO] += b_CH2_HCOP_CO_CH3P;
	
	c[ipHCOP][ipCH3P] += CH2_HCOP_CO_CH3P_1; 
	c[ipCH2][ipCH3P] += CH2_HCOP_CO_CH3P_2;
	b[ipCH3P] += b_CH2_HCOP_CO_CH3P;
	
	/* >>refer	mol	CH2	O2	=>	CO	H2O	UMIST	*/
	
	c[ipO2][ipCH2] -= CH2_O2_CO_H2O_1; 
	c[ipCH2][ipCH2] -= CH2_O2_CO_H2O_2;
	b[ipCH2] -= b_CH2_O2_CO_H2O;
	
	c[ipO2][ipO2] -= CH2_O2_CO_H2O_1; 
	c[ipCH2][ipO2] -= CH2_O2_CO_H2O_2;
	b[ipO2] -= b_CH2_O2_CO_H2O;
	
	c[ipO2][ipCO] += CH2_O2_CO_H2O_1; 
	c[ipCH2][ipCO] += CH2_O2_CO_H2O_2;
	b[ipCO] += b_CH2_O2_CO_H2O;
	
	c[ipO2][ipH2O] += CH2_O2_CO_H2O_1; 
	c[ipCH2][ipH2O] += CH2_O2_CO_H2O_2;
	b[ipH2O] += b_CH2_O2_CO_H2O;
	
	/* >>refer	mol	CH2	O2+	=>	O2	CH2+	UMIST	*/
	
	c[ipO2P][ipCH2] -= CH2_O2P_O2_CH2P_1; 
	c[ipCH2][ipCH2] -= CH2_O2P_O2_CH2P_2;
	b[ipCH2] -= b_CH2_O2P_O2_CH2P;
	
	c[ipO2P][ipO2P] -= CH2_O2P_O2_CH2P_1; 
	c[ipCH2][ipO2P] -= CH2_O2P_O2_CH2P_2;
	b[ipO2P] -= b_CH2_O2P_O2_CH2P;
	
	c[ipO2P][ipO2] += CH2_O2P_O2_CH2P_1; 
	c[ipCH2][ipO2] += CH2_O2P_O2_CH2P_2;
	b[ipO2] += b_CH2_O2P_O2_CH2P;
	
	c[ipO2P][ipCH2P] += CH2_O2P_O2_CH2P_1; 
	c[ipCH2][ipCH2P] += CH2_O2P_O2_CH2P_2;
	b[ipCH2P] += b_CH2_O2P_O2_CH2P;
	
	/* >>refer	mol	CH2	OH	=>	H2O	CH	UMIST	*/
	
	c[ipOH][ipCH2] -= CH2_OH_H2O_CH_1; 
	c[ipCH2][ipCH2] -= CH2_OH_H2O_CH_2;
	b[ipCH2] -= b_CH2_OH_H2O_CH;
	
	c[ipOH][ipOH] -= CH2_OH_H2O_CH_1; 
	c[ipCH2][ipOH] -= CH2_OH_H2O_CH_2;
	b[ipOH] -= b_CH2_OH_H2O_CH;
	
	c[ipOH][ipH2O] += CH2_OH_H2O_CH_1; 
	c[ipCH2][ipH2O] += CH2_OH_H2O_CH_2;
	b[ipH2O] += b_CH2_OH_H2O_CH;
	
	c[ipOH][ipCH] += CH2_OH_H2O_CH_1; 
	c[ipCH2][ipCH] += CH2_OH_H2O_CH_2;
	b[ipCH] += b_CH2_OH_H2O_CH;
	
	
	/* >>refer	mol	CH2	OH+	=>	OH	CH2+	UMIST	*/
		
	c[ipOHP][ipCH2] -= CH2_OHP_OH_CH2P_1; 
	c[ipCH2][ipCH2] -= CH2_OHP_OH_CH2P_2;
	b[ipCH2] -= b_CH2_OHP_OH_CH2P;
	
	c[ipOHP][ipOHP] -= CH2_OHP_OH_CH2P_1; 
	c[ipCH2][ipOHP] -= CH2_OHP_OH_CH2P_2;
	b[ipOHP] -= b_CH2_OHP_OH_CH2P;
	
	c[ipOHP][ipOH] += CH2_OHP_OH_CH2P_1; 
	c[ipCH2][ipOH] += CH2_OHP_OH_CH2P_2;
	b[ipOH] += b_CH2_OHP_OH_CH2P;
	
	c[ipOHP][ipCH2P] += CH2_OHP_OH_CH2P_1; 
	c[ipCH2][ipCH2P] += CH2_OHP_OH_CH2P_2;
	b[ipCH2P] += b_CH2_OHP_OH_CH2P;
	
	/* >>refer	mol	CH2	OH+	=>	O	CH3+	UMIST	*/
	
	c[ipOHP][ipCH2] -= CH2_OHP_O_CH3P_1; 
	c[ipCH2][ipCH2] -= CH2_OHP_O_CH3P_2;
	b[ipCH2] -= b_CH2_OHP_O_CH3P;
	
	c[ipOHP][ipOHP] -= CH2_OHP_O_CH3P_1; 
	c[ipCH2][ipOHP] -= CH2_OHP_O_CH3P_2;
	b[ipOHP] -= b_CH2_OHP_O_CH3P;
	
	c[ipOHP][ipCH3P] += CH2_OHP_O_CH3P_1; 
	c[ipCH2][ipCH3P] += CH2_OHP_O_CH3P_2;
	b[ipCH3P] += b_CH2_OHP_O_CH3P;
	
	c[ipOHP][ipATO] += CH2_OHP_O_CH3P_1; 
	c[ipCH2][ipATO] += CH2_OHP_O_CH3P_2;
	b[ipATO] += b_CH2_OHP_O_CH3P;
	
	/* >>refer	mol	CH2+	O2+	=>	HCO+	OH	UMIST	*/
		
	c[ipO2P][ipCH2P] -= CH2P_O2_HCOP_OH_1; 
	c[ipCH2P][ipCH2P] -= CH2P_O2_HCOP_OH_2;
	b[ipCH2P] -= b_CH2P_O2_HCOP_OH;
	
	c[ipO2P][ipO2P] -= CH2P_O2_HCOP_OH_1; 
	c[ipCH2P][ipO2P] -= CH2P_O2_HCOP_OH_2;
	b[ipO2P] -= b_CH2P_O2_HCOP_OH;
	
	c[ipO2P][ipHCOP] += CH2P_O2_HCOP_OH_1; 
	c[ipCH2P][ipHCOP] += CH2P_O2_HCOP_OH_2;
	b[ipHCOP] += b_CH2P_O2_HCOP_OH;
	
	c[ipO2P][ipOH] += CH2P_O2_HCOP_OH_1; 
	c[ipCH2P][ipOH] += CH2P_O2_HCOP_OH_2;
	b[ipOH] += b_CH2P_O2_HCOP_OH;
	
	/* >>refer	mol	CO+	O2	=>	O2+	CO	UMIST	*/
	
	c[ipO2][ipCOP] -= COP_O2_O2P_CO_1; 
	c[ipCOP][ipCOP] -= COP_O2_O2P_CO_2;
	b[ipCOP] -= b_COP_O2_O2P_CO;
	
	c[ipO2][ipO2] -= COP_O2_O2P_CO_1; 
	c[ipCOP][ipO2] -= COP_O2_O2P_CO_2;
	b[ipO2] -= b_COP_O2_O2P_CO;
	
	c[ipO2][ipO2P] += COP_O2_O2P_CO_1; 
	c[ipCOP][ipO2P] += COP_O2_O2P_CO_2;
	b[ipO2P] += b_COP_O2_O2P_CO;
	
	c[ipO2][ipCO] += COP_O2_O2P_CO_1; 
	c[ipCOP][ipCO] += COP_O2_O2P_CO_2;
	b[ipCO] += b_COP_O2_O2P_CO;
	
	/* >>refer	mol	H2O	CO+	=>	CO	H2O+	UMIST	*/
		
	c[ipCOP][ipH2O] -= H2O_COP_CO_H2OP_1; 
	c[ipH2O][ipH2O] -= H2O_COP_CO_H2OP_2;
	b[ipH2O] -= b_H2O_COP_CO_H2OP;
	
	c[ipCOP][ipCOP] -= H2O_COP_CO_H2OP_1; 
	c[ipH2O][ipCOP] -= H2O_COP_CO_H2OP_2;
	b[ipCOP] -= b_H2O_COP_CO_H2OP;
	
	c[ipCOP][ipCO] += H2O_COP_CO_H2OP_1; 
	c[ipH2O][ipCO] += H2O_COP_CO_H2OP_2;
	b[ipCO] += b_H2O_COP_CO_H2OP;
	
	c[ipCOP][ipH2OP] += H2O_COP_CO_H2OP_1; 
	c[ipH2O][ipH2OP] += H2O_COP_CO_H2OP_2;
	b[ipH2OP] += b_H2O_COP_CO_H2OP;
	
	/* >>refer	mol	H2O	CO+	=>	HCO+	OH	UMIST	*/
	
	c[ipCOP][ipH2O] -= H2O_COP_HCOP_OH_1; 
	c[ipH2O][ipH2O] -= H2O_COP_HCOP_OH_2;
	b[ipH2O] -= b_H2O_COP_HCOP_OH;
	
	c[ipCOP][ipCOP] -= H2O_COP_HCOP_OH_1; 
	c[ipH2O][ipCOP] -= H2O_COP_HCOP_OH_2;
	b[ipCOP] -= b_H2O_COP_HCOP_OH;
	
	c[ipCOP][ipHCOP] += H2O_COP_HCOP_OH_1; 
	c[ipH2O][ipHCOP] += H2O_COP_HCOP_OH_2;
	b[ipHCOP] += b_H2O_COP_HCOP_OH;
	
	c[ipCOP][ipOH] += H2O_COP_HCOP_OH_1; 
	c[ipH2O][ipOH] += H2O_COP_HCOP_OH_2;
	b[ipOH] += b_H2O_COP_HCOP_OH;
	
	/* >>refer	mol	H2O	HCO+	=>	CO	H3O+	UMIST	*/
	
	c[ipHCOP][ipH2O] -= H2O_HCOP_CO_H3OP_1; 
	c[ipH2O][ipH2O] -= H2O_HCOP_CO_H3OP_2;
	b[ipH2O] -= b_H2O_HCOP_CO_H3OP;
	
	c[ipHCOP][ipHCOP] -= H2O_HCOP_CO_H3OP_1; 
	c[ipH2O][ipHCOP] -= H2O_HCOP_CO_H3OP_2;
	b[ipHCOP] -= b_H2O_HCOP_CO_H3OP;
	
	c[ipHCOP][ipCO] += H2O_HCOP_CO_H3OP_1; 
	c[ipH2O][ipCO] += H2O_HCOP_CO_H3OP_2;
	b[ipCO] += b_H2O_HCOP_CO_H3OP;
	
	c[ipHCOP][ipH3OP] += H2O_HCOP_CO_H3OP_1; 
	c[ipH2O][ipH3OP] += H2O_HCOP_CO_H3OP_2;
	b[ipH3OP] += b_H2O_HCOP_CO_H3OP;
	
	
	/* >>refer	mol	H2O+	CO	=>	HCO+	OH	UMIST	*/
		
	c[ipCO][ipH2OP] -= H2OP_CO_HCOP_OH_1; 
	c[ipH2OP][ipH2OP] -= H2OP_CO_HCOP_OH_2;
	b[ipH2OP] -= b_H2OP_CO_HCOP_OH;
	
	c[ipCO][ipCO] -= H2OP_CO_HCOP_OH_1; 
	c[ipH2OP][ipCO] -= H2OP_CO_HCOP_OH_2;
	b[ipCO] -= b_H2OP_CO_HCOP_OH;
	
	c[ipCO][ipHCOP] += H2OP_CO_HCOP_OH_1; 
	c[ipH2OP][ipHCOP] += H2OP_CO_HCOP_OH_2;
	b[ipHCOP] += b_H2OP_CO_HCOP_OH;
	
	c[ipCO][ipOH] += H2OP_CO_HCOP_OH_1; 
	c[ipH2OP][ipOH] += H2OP_CO_HCOP_OH_2;
	b[ipOH] += b_H2OP_CO_HCOP_OH;
	
	
	/* >>refer	mol	H2O+	H2O	=>	H3O+	OH	UMIST	*/
		
	c[ipH2O][ipH2OP] -= H2OP_H2O_H3OP_OH_1; 
	c[ipH2OP][ipH2OP] -= H2OP_H2O_H3OP_OH_2;
	b[ipH2OP] -= b_H2OP_H2O_H3OP_OH;
	
	c[ipH2O][ipH2O] -= H2OP_H2O_H3OP_OH_1; 
	c[ipH2OP][ipH2O] -= H2OP_H2O_H3OP_OH_2;
	b[ipH2O] -= b_H2OP_H2O_H3OP_OH;
	
	c[ipH2O][ipH3OP] += H2OP_H2O_H3OP_OH_1; 
	c[ipH2OP][ipH3OP] += H2OP_H2O_H3OP_OH_2;
	b[ipH3OP] += b_H2OP_H2O_H3OP_OH;
	
	c[ipH2O][ipOH] += H2OP_H2O_H3OP_OH_1; 
	c[ipH2OP][ipOH] += H2OP_H2O_H3OP_OH_2;
	b[ipOH] += b_H2OP_H2O_H3OP_OH;
	
	/* >>refer	mol	H2O+	O2	=>	O2+	H2O	UMIST	*/
		
	
	c[ipO2][ipH2OP] -= H2OP_O2_O2P_H2O_1; 
	c[ipH2OP][ipH2OP] -= H2OP_O2_O2P_H2O_2;
	b[ipH2OP] -= b_H2OP_O2_O2P_H2O;
	
	c[ipO2][ipO2] -= H2OP_O2_O2P_H2O_1; 
	c[ipH2OP][ipO2] -= H2OP_O2_O2P_H2O_2;
	b[ipO2] -= b_H2OP_O2_O2P_H2O;
	
	c[ipO2][ipO2P] += H2OP_O2_O2P_H2O_1; 
	c[ipH2OP][ipO2P] += H2OP_O2_O2P_H2O_2;
	b[ipO2P] += b_H2OP_O2_O2P_H2O;
	
	c[ipO2][ipH2O] += H2OP_O2_O2P_H2O_1; 
	c[ipH2OP][ipH2O] += H2OP_O2_O2P_H2O_2;
	b[ipH2O] += b_H2OP_O2_O2P_H2O;
	
	/* >>refer	mol	H3O+	SiH	=>	SiH2+	H2O	UMIST	*/
		
	
	c[ipSIH][ipH3OP] -= H3OP_SiH_SiH2P_H2O_1; 
	c[ipH3OP][ipH3OP] -= H3OP_SiH_SiH2P_H2O_2;
	b[ipH3OP] -= b_H3OP_SiH_SiH2P_H2O;
	
	c[ipSIH][ipSIH] -= H3OP_SiH_SiH2P_H2O_1; 
	c[ipH3OP][ipSIH] -= H3OP_SiH_SiH2P_H2O_2;
	b[ipSIH] -= b_H3OP_SiH_SiH2P_H2O;
	
	c[ipSIH][ipSIH2P] += H3OP_SiH_SiH2P_H2O_1; 
	c[ipH3OP][ipSIH2P] += H3OP_SiH_SiH2P_H2O_2;
	b[ipSIH2P] += b_H3OP_SiH_SiH2P_H2O;
	
	c[ipSIH][ipH2O] += H3OP_SiH_SiH2P_H2O_1; 
	c[ipH3OP][ipH2O] += H3OP_SiH_SiH2P_H2O_2;
	b[ipH2O] += b_H3OP_SiH_SiH2P_H2O;
	
	/* >>refer	mol	H3O+	SiO	=>	SiOH+	H2O	UMIST	*/
		
	
	c[ipSIO][ipH3OP] -= H3OP_SiO_SiOHP_H2O_1; 
	c[ipH3OP][ipH3OP] -= H3OP_SiO_SiOHP_H2O_2;
	b[ipH3OP] -= b_H3OP_SiO_SiOHP_H2O;
	
	c[ipSIO][ipSIO] -= H3OP_SiO_SiOHP_H2O_1; 
	c[ipH3OP][ipSIO] -= H3OP_SiO_SiOHP_H2O_2;
	b[ipSIO] -= b_H3OP_SiO_SiOHP_H2O;
	
	c[ipSIO][ipHOSIP] += H3OP_SiO_SiOHP_H2O_1; 
	c[ipH3OP][ipHOSIP] += H3OP_SiO_SiOHP_H2O_2;
	b[ipHOSIP] += b_H3OP_SiO_SiOHP_H2O;
	
	c[ipSIO][ipH2O] += H3OP_SiO_SiOHP_H2O_1; 
	c[ipH3OP][ipH2O] += H3OP_SiO_SiOHP_H2O_2;
	b[ipH2O] += b_H3OP_SiO_SiOHP_H2O;
	
	/* >>refer	mol	HCO+	SiH	=>	SiH2+	CO	UMIST	*/
		
	
	c[ipSIH][ipHCOP] -= HCOP_SiH_SiH2P_CO_1; 
	c[ipHCOP][ipHCOP] -= HCOP_SiH_SiH2P_CO_2;
	b[ipHCOP] -= b_HCOP_SiH_SiH2P_CO;
	
	c[ipSIH][ipSIH] -= HCOP_SiH_SiH2P_CO_1; 
	c[ipHCOP][ipSIH] -= HCOP_SiH_SiH2P_CO_2;
	b[ipSIH] -= b_HCOP_SiH_SiH2P_CO;
	
	c[ipSIH][ipSIH2P] += HCOP_SiH_SiH2P_CO_1; 
	c[ipHCOP][ipSIH2P] += HCOP_SiH_SiH2P_CO_2;
	b[ipSIH2P] += b_HCOP_SiH_SiH2P_CO;
	
	c[ipSIH][ipCO] += HCOP_SiH_SiH2P_CO_1; 
	c[ipHCOP][ipCO] += HCOP_SiH_SiH2P_CO_2;
	b[ipCO] += b_HCOP_SiH_SiH2P_CO;
	
	/* >>refer	mol	HCO+	SiO	=>	SiOH+	CO	UMIST	*/
	
	c[ipSIO][ipHCOP] -= HCOP_SiO_SiOHP_CO_1; 
	c[ipHCOP][ipHCOP] -= HCOP_SiO_SiOHP_CO_2;
	b[ipHCOP] -= b_HCOP_SiO_SiOHP_CO;
	
	c[ipSIO][ipSIO] -= HCOP_SiO_SiOHP_CO_1; 
	c[ipHCOP][ipSIO] -= HCOP_SiO_SiOHP_CO_2;
	b[ipSIO] -= b_HCOP_SiO_SiOHP_CO;
	
	c[ipSIO][ipHOSIP] += HCOP_SiO_SiOHP_CO_1; 
	c[ipHCOP][ipHOSIP] += HCOP_SiO_SiOHP_CO_2;
	b[ipHOSIP] += b_HCOP_SiO_SiOHP_CO;
	
	c[ipSIO][ipCO] += HCOP_SiO_SiOHP_CO_1; 
	c[ipHCOP][ipCO] += HCOP_SiO_SiOHP_CO_2;
	b[ipCO] += b_HCOP_SiO_SiOHP_CO;
	
	/* >>refer	mol	OH	CO+	=>	HCO+	O	UMIST	*/
	
	c[ipCOP][ipOH] -= OH_COP_HCOP_O_1; 
	c[ipOH][ipOH] -= OH_COP_HCOP_O_2;
	b[ipOH] -= b_OH_COP_HCOP_O;
	
	c[ipCOP][ipCOP] -= OH_COP_HCOP_O_1; 
	c[ipOH][ipCOP] -= OH_COP_HCOP_O_2;
	b[ipCOP] -= b_OH_COP_HCOP_O;
	
	c[ipCOP][ipHCOP] += OH_COP_HCOP_O_1; 
	c[ipOH][ipHCOP] += OH_COP_HCOP_O_2;
	b[ipHCOP] += b_OH_COP_HCOP_O;
	
	c[ipCOP][ipATO] += OH_COP_HCOP_O_1; 
	c[ipOH][ipATO] += OH_COP_HCOP_O_2;
	b[ipATO] += b_OH_COP_HCOP_O;
	
	/* >>refer	mol	OH	CO+	=>	CO	OH+	UMIST	*/
	
	c[ipCOP][ipOH] -= OH_COP_CO_OHP_1; 
	c[ipOH][ipOH] -= OH_COP_CO_OHP_2;
	b[ipOH] -= b_OH_COP_CO_OHP;
	
	c[ipCOP][ipCOP] -= OH_COP_CO_OHP_1; 
	c[ipOH][ipCOP] -= OH_COP_CO_OHP_2;
	b[ipCOP] -= b_OH_COP_CO_OHP;
	
	c[ipCOP][ipCO] += OH_COP_CO_OHP_1; 
	c[ipOH][ipCO] += OH_COP_CO_OHP_2;
	b[ipCO] += b_OH_COP_CO_OHP;
	
	c[ipCOP][ipOHP] += OH_COP_CO_OHP_1; 
	c[ipOH][ipOHP] += OH_COP_CO_OHP_2;
	b[ipOHP] += b_OH_COP_CO_OHP;
	
	/* >>refer	mol	OH	H2O+	=>	H3O+	O	UMIST	*/
	
	c[ipH2OP][ipOH] -= OH_H2OP_H3OP_O_1; 
	c[ipOH][ipOH] -= OH_H2OP_H3OP_O_2;
	b[ipOH] -= b_OH_H2OP_H3OP_O;
	
	c[ipH2OP][ipH2OP] -= OH_H2OP_H3OP_O_1; 
	c[ipOH][ipH2OP] -= OH_H2OP_H3OP_O_2;
	b[ipH2OP] -= b_OH_H2OP_H3OP_O;
	
	c[ipH2OP][ipH3OP] += OH_H2OP_H3OP_O_1; 
	c[ipOH][ipH3OP] += OH_H2OP_H3OP_O_2;
	b[ipH3OP] += b_OH_H2OP_H3OP_O;
	
	c[ipH2OP][ipATO] += OH_H2OP_H3OP_O_1; 
	c[ipOH][ipATO] += OH_H2OP_H3OP_O_2;
	b[ipATO] += b_OH_H2OP_H3OP_O;
	
	/* >>refer	mol	OH	HCO+	=>	CO	H2O+	UMIST	*/
	
	c[ipHCOP][ipOH] -= OH_HCOP_CO_H2OP_1; 
	c[ipOH][ipOH] -= OH_HCOP_CO_H2OP_2;
	b[ipOH] -= b_OH_HCOP_CO_H2OP;
	
	c[ipHCOP][ipHCOP] -= OH_HCOP_CO_H2OP_1; 
	c[ipOH][ipHCOP] -= OH_HCOP_CO_H2OP_2;
	b[ipHCOP] -= b_OH_HCOP_CO_H2OP;
	
	c[ipHCOP][ipCO] += OH_HCOP_CO_H2OP_1; 
	c[ipOH][ipCO] += OH_HCOP_CO_H2OP_2;
	b[ipCO] += b_OH_HCOP_CO_H2OP;
	
	c[ipHCOP][ipH2OP] += OH_HCOP_CO_H2OP_1; 
	c[ipOH][ipH2OP] += OH_HCOP_CO_H2OP_2;
	b[ipH2OP] += b_OH_HCOP_CO_H2OP;
	
	/* >>refer	mol	OH	OH	=>	H2O	H	UMIST	*/
	
	c[ipOH][ipOH] -= OH_OH_H2O_O_1; 
	c[ipOH][ipOH] -= OH_OH_H2O_O_2;
	b[ipOH] -= b_OH_OH_H2O_O;
	
	c[ipOH][ipH2O] += 0.5*OH_OH_H2O_O_1; 
	c[ipOH][ipH2O] += 0.5*OH_OH_H2O_O_2;
	b[ipH2O] += 0.5*b_OH_OH_H2O_O;
	
	/* >>refer	mol	OH+	CO	=>	HCO+	O	UMIST	*/
		
	c[ipCO][ipOHP] -= OHP_CO_HCOP_O_1; 
	c[ipOHP][ipOHP] -= OHP_CO_HCOP_O_2;
	b[ipOHP] -= b_OHP_CO_HCOP_O;
	
	c[ipCO][ipCO] -= OHP_CO_HCOP_O_1; 
	c[ipOHP][ipCO] -= OHP_CO_HCOP_O_2;
	b[ipCO] -= b_OHP_CO_HCOP_O;
	
	c[ipCO][ipHCOP] += OHP_CO_HCOP_O_1; 
	c[ipOHP][ipHCOP] += OHP_CO_HCOP_O_2;
	b[ipHCOP] += b_OHP_CO_HCOP_O;
	
	c[ipCO][ipATO] += OHP_CO_HCOP_O_1; 
	c[ipOHP][ipATO] += OHP_CO_HCOP_O_2;
	b[ipATO] += b_OHP_CO_HCOP_O;
	
	/* >>refer	mol	OH+	H2O	=>	H3O+	O	UMIST	*/
	
	c[ipH2O][ipOHP] -= OHP_H2O_H3OP_O_1; 
	c[ipOHP][ipOHP] -= OHP_H2O_H3OP_O_2;
	b[ipOHP] -= b_OHP_H2O_H3OP_O;
	
	c[ipH2O][ipH2O] -= OHP_H2O_H3OP_O_1; 
	c[ipOHP][ipH2O] -= OHP_H2O_H3OP_O_2;
	b[ipH2O] -= b_OHP_H2O_H3OP_O;
	
	c[ipH2O][ipH3OP] += OHP_H2O_H3OP_O_1; 
	c[ipOHP][ipH3OP] += OHP_H2O_H3OP_O_2;
	b[ipH3OP] += b_OHP_H2O_H3OP_O;
	
	c[ipH2O][ipATO] += OHP_H2O_H3OP_O_1; 
	c[ipOHP][ipATO] += OHP_H2O_H3OP_O_2;
	b[ipATO] += b_OHP_H2O_H3OP_O;
	
	/* >>refer	mol	OH+	H2O	=>	H2O+	OH	UMIST	*/
	
	c[ipH2O][ipOHP] -= OHP_H2O_H2OP_OH_1; 
	c[ipOHP][ipOHP] -= OHP_H2O_H2OP_OH_2;
	b[ipOHP] -= b_OHP_H2O_H2OP_OH;
	
	c[ipH2O][ipH2O] -= OHP_H2O_H2OP_OH_1; 
	c[ipOHP][ipH2O] -= OHP_H2O_H2OP_OH_2;
	b[ipH2O] -= b_OHP_H2O_H2OP_OH;
	
	c[ipH2O][ipH2OP] += OHP_H2O_H2OP_OH_1; 
	c[ipOHP][ipH2OP] += OHP_H2O_H2OP_OH_2;
	b[ipH2OP] += b_OHP_H2O_H2OP_OH;
	
	c[ipH2O][ipOH] += OHP_H2O_H2OP_OH_1; 
	c[ipOHP][ipOH] += OHP_H2O_H2OP_OH_2;
	b[ipOH] += b_OHP_H2O_H2OP_OH;
	
	/* >>refer	mol	OH+	O2	=>	O2+	OH	UMIST	*/
	
	c[ipO2][ipOHP] -= OHP_O2_O2P_OH_1; 
	c[ipOHP][ipOHP] -= OHP_O2_O2P_OH_2;
	b[ipOHP] -= b_OHP_O2_O2P_OH;
	
	c[ipO2][ipO2] -= OHP_O2_O2P_OH_1; 
	c[ipOHP][ipO2] -= OHP_O2_O2P_OH_2;
	b[ipO2] -= b_OHP_O2_O2P_OH;
	
	c[ipO2][ipO2P] += OHP_O2_O2P_OH_1; 
	c[ipOHP][ipO2P] += OHP_O2_O2P_OH_2;
	b[ipO2P] += b_OHP_O2_O2P_OH;
	
	c[ipO2][ipOH] += OHP_O2_O2P_OH_1; 
	c[ipOHP][ipOH] += OHP_O2_O2P_OH_2;
	b[ipOH] += b_OHP_O2_O2P_OH;
	
	/* >>refer	mol	OH+	OH	=>	H2O+	O	UMIST	*/
	
	c[ipOH][ipOHP] -= OHP_OH_H2OP_O_1; 
	c[ipOHP][ipOHP] -= OHP_OH_H2OP_O_2;
	b[ipOHP] -= b_OHP_OH_H2OP_O;
	
	c[ipOH][ipOH] -= OHP_OH_H2OP_O_1; 
	c[ipOHP][ipOH] -= OHP_OH_H2OP_O_2;
	b[ipOH] -= b_OHP_OH_H2OP_O;
	
	c[ipOH][ipH2OP] += OHP_OH_H2OP_O_1; 
	c[ipOHP][ipH2OP] += OHP_OH_H2OP_O_2;
	b[ipH2OP] += b_OHP_OH_H2OP_O;
	
	c[ipOH][ipATO] += OHP_OH_H2OP_O_1; 
	c[ipOHP][ipATO] += OHP_OH_H2OP_O_2;
	b[ipATO] += b_OHP_OH_H2OP_O;
	
	/* >>refer	mol	OH+	SiH	=>	SiH2+	O	UMIST	*/
	
	c[ipSIH][ipOHP] -= OHP_SiH_SiH2P_O_1; 
	c[ipOHP][ipOHP] -= OHP_SiH_SiH2P_O_2;
	b[ipOHP] -= b_OHP_SiH_SiH2P_O;
	
	c[ipSIH][ipSIH] -= OHP_SiH_SiH2P_O_1; 
	c[ipOHP][ipSIH] -= OHP_SiH_SiH2P_O_2;
	b[ipSIH] -= b_OHP_SiH_SiH2P_O;
	
	c[ipSIH][ipSIH2P] += OHP_SiH_SiH2P_O_1; 
	c[ipOHP][ipSIH2P] += OHP_SiH_SiH2P_O_2;
	b[ipSIH2P] += b_OHP_SiH_SiH2P_O;
	
	c[ipSIH][ipATO] += OHP_SiH_SiH2P_O_1; 
	c[ipOHP][ipATO] += OHP_SiH_SiH2P_O_2;
	b[ipATO] += b_OHP_SiH_SiH2P_O;
	
	/* >>refer	mol	OH+	SiO	=>	SiOH+	O	UMIST	*/
	
	c[ipSIO][ipOHP] -= OHP_SiO_SiOHP_O_1; 
	c[ipOHP][ipOHP] -= OHP_SiO_SiOHP_O_2;
	b[ipOHP] -= b_OHP_SiO_SiOHP_O;
	
	c[ipSIO][ipSIO] -= OHP_SiO_SiOHP_O_1; 
	c[ipOHP][ipSIO] -= OHP_SiO_SiOHP_O_2;
	b[ipSIO] -= b_OHP_SiO_SiOHP_O;
	
	c[ipSIO][ipHOSIP] += OHP_SiO_SiOHP_O_1; 
	c[ipOHP][ipHOSIP] += OHP_SiO_SiOHP_O_2;
	b[ipHOSIP] += b_OHP_SiO_SiOHP_O;
	
	c[ipSIO][ipATO] += OHP_SiO_SiOHP_O_1; 
	c[ipOHP][ipATO] += OHP_SiO_SiOHP_O_2;
	b[ipATO] += b_OHP_SiO_SiOHP_O;


	/* new coupled reactions involving CH3, CH4, CH4+, and CH5+ */

	/* >>refer mol   C CH5P  =>  CH4 CHP   UMIST    */
	
	c[ipATC][ipCH5P] -= C_CH5P_CH4_CHP_1;
	c[ipCH5P][ipCH5P] -= C_CH5P_CH4_CHP_2;
	b[ipCH5P] -= b_C_CH5P_CH4_CHP;
	
	c[ipATC][ipATC] -= C_CH5P_CH4_CHP_1;
	c[ipCH5P][ipATC] -= C_CH5P_CH4_CHP_2;
	b[ipATC] -= b_C_CH5P_CH4_CHP;
	
	c[ipATC][ipCH4] += C_CH5P_CH4_CHP_1;
	c[ipCH5P][ipCH4] += C_CH5P_CH4_CHP_2;
	b[ipCH4] += b_C_CH5P_CH4_CHP;
	
	c[ipATC][ipCHP] += C_CH5P_CH4_CHP_1;
	c[ipCH5P][ipCHP] += C_CH5P_CH4_CHP_2;
	b[ipCHP] += b_C_CH5P_CH4_CHP;
	
	/* >>refer mol   O CH4  =>  OH CH3   UMIST    */
	
	c[ipATO][ipCH4] -= O_CH4_OH_CH3_1;
	c[ipCH4][ipCH4] -= O_CH4_OH_CH3_2;
	b[ipCH4] -= b_O_CH4_OH_CH3;
		
	c[ipATO][ipATO] -= O_CH4_OH_CH3_1;
	c[ipCH4][ipATO] -= O_CH4_OH_CH3_2;
	b[ipATO] -= b_O_CH4_OH_CH3;
	
	c[ipATO][ipOH] += O_CH4_OH_CH3_1;
	c[ipCH4][ipOH] += O_CH4_OH_CH3_2;
	b[ipOH] += b_O_CH4_OH_CH3;
	
	c[ipATO][ipCH3] += O_CH4_OH_CH3_1;
	c[ipCH4][ipCH3] += O_CH4_OH_CH3_2;
	b[ipCH3] += b_O_CH4_OH_CH3;
	
	
	/* >>refer mol   O CH4P  =>  OH CH3P   UMIST    */
	
	c[ipATO][ipCH4P] -= O_CH4P_OH_CH3P_1;
	c[ipCH4P][ipCH4P] -= O_CH4P_OH_CH3P_2;
	b[ipCH4P] -= b_O_CH4P_OH_CH3P;
	
	c[ipATO][ipATO] -= O_CH4P_OH_CH3P_1;
	c[ipCH4P][ipATO] -= O_CH4P_OH_CH3P_2;
	b[ipATO] -= b_O_CH4P_OH_CH3P;
	
	c[ipATO][ipOH] += O_CH4P_OH_CH3P_1;
	c[ipCH4P][ipOH] += O_CH4P_OH_CH3P_2;
	b[ipOH] += b_O_CH4P_OH_CH3P;
	
	c[ipATO][ipCH3P] += O_CH4P_OH_CH3P_1;
	c[ipCH4P][ipCH3P] += O_CH4P_OH_CH3P_2;
	b[ipCH3P] += b_O_CH4P_OH_CH3P;
			
	/* >>refer mol   O CH5P  =>  H3OP CH2   UMIST    */
	
	c[ipATO][ipCH5P] -= O_CH5P_H3OP_CH2_1;
	c[ipCH5P][ipCH5P] -= O_CH5P_H3OP_CH2_2;
	b[ipCH5P] -= b_O_CH5P_H3OP_CH2;
		
	c[ipATO][ipATO] -= O_CH5P_H3OP_CH2_1;
	c[ipCH5P][ipATO] -= O_CH5P_H3OP_CH2_2;
	b[ipATO] -= b_O_CH5P_H3OP_CH2;
	
	c[ipATO][ipH3OP] += O_CH5P_H3OP_CH2_1;
	c[ipCH5P][ipH3OP] += O_CH5P_H3OP_CH2_2;
	b[ipH3OP] += b_O_CH5P_H3OP_CH2;
	
	c[ipATO][ipCH2] += O_CH5P_H3OP_CH2_1;
	c[ipCH5P][ipCH2] += O_CH5P_H3OP_CH2_2;
	b[ipCH2] += b_O_CH5P_H3OP_CH2;
	
	
	/* >>refer mol   OP CH4  =>  OH CH3P   UMIST    */
	
	c[ipOP][ipOP] -= OP_CH4_OH_CH3P_1;
	c[ipCH4][ipOP] -= OP_CH4_OH_CH3P_2;
	b[ipOP] -= b_OP_CH4_OH_CH3P;
	
	c[ipOP][ipCH4] -= OP_CH4_OH_CH3P_1;
	c[ipCH4][ipCH4] -= OP_CH4_OH_CH3P_2;
	b[ipCH4] -= b_OP_CH4_OH_CH3P;

	c[ipOP][ipOH] += OP_CH4_OH_CH3P_1;
	c[ipCH4][ipOH] += OP_CH4_OH_CH3P_2;
	b[ipOH] += b_OP_CH4_OH_CH3P;
	
	c[ipOP][ipCH3P] += OP_CH4_OH_CH3P_1;
	c[ipCH4][ipCH3P] += OP_CH4_OH_CH3P_2;
	b[ipCH3P] += b_OP_CH4_OH_CH3P;
	
	
	/* >>refer mol   OP CH4  =>  CH4P O   UMIST    */
	
	c[ipOP][ipOP] -= OP_CH4_CH4P_O_1;
	c[ipCH4][ipOP] -= OP_CH4_CH4P_O_2;
	b[ipOP] -= b_OP_CH4_CH4P_O;
	
	c[ipOP][ipCH4] -= OP_CH4_CH4P_O_1;
	c[ipCH4][ipCH4] -= OP_CH4_CH4P_O_2;
	b[ipCH4] -= b_OP_CH4_CH4P_O;
	
	c[ipOP][ipCH4P] += OP_CH4_CH4P_O_1;
	c[ipCH4][ipCH4P] += OP_CH4_CH4P_O_2;
	b[ipCH4P] += b_OP_CH4_CH4P_O;
		
	c[ipOP][ipATO] += OP_CH4_CH4P_O_1;
	c[ipCH4][ipATO] += OP_CH4_CH4P_O_2;
	b[ipATO] += b_OP_CH4_CH4P_O;
	
	
	/* >>refer mol   CH4 CH  =>  CH3 CH2   UMIST    */
	
	c[ipCH4][ipCH4] -= CH4_CH_CH3_CH2_1;
	c[ipCH][ipCH4] -= CH4_CH_CH3_CH2_2;
	b[ipCH4] -= b_CH4_CH_CH3_CH2;
	
	c[ipCH4][ipCH] -= CH4_CH_CH3_CH2_1;
	c[ipCH][ipCH] -= CH4_CH_CH3_CH2_2;
	b[ipCH] -= b_CH4_CH_CH3_CH2;
	
	c[ipCH4][ipCH3] += CH4_CH_CH3_CH2_1;
	c[ipCH][ipCH3] += CH4_CH_CH3_CH2_2;
	b[ipCH3] += b_CH4_CH_CH3_CH2;
	
	c[ipCH4][ipCH2] += CH4_CH_CH3_CH2_1;
	c[ipCH][ipCH2] += CH4_CH_CH3_CH2_2;
	b[ipCH2] += b_CH4_CH_CH3_CH2;
	
	
	/* >>refer mol   CH5P CH  =>  CH4 CH2P   UMIST    */

	c[ipCH5P][ipCH5P] -= CH5P_CH_CH4_CH2P_1;
	c[ipCH][ipCH5P] -= CH5P_CH_CH4_CH2P_2;
	b[ipCH5P] -= b_CH5P_CH_CH4_CH2P;
	
	c[ipCH5P][ipCH] -= CH5P_CH_CH4_CH2P_1;
	c[ipCH][ipCH] -= CH5P_CH_CH4_CH2P_2;
	b[ipCH] -= b_CH5P_CH_CH4_CH2P;
	
	c[ipCH5P][ipCH4] += CH5P_CH_CH4_CH2P_1;
	c[ipCH][ipCH4] += CH5P_CH_CH4_CH2P_2;
	b[ipCH4] += b_CH5P_CH_CH4_CH2P;
	
	c[ipCH5P][ipCH2P] += CH5P_CH_CH4_CH2P_1;
	c[ipCH][ipCH2P] += CH5P_CH_CH4_CH2P_2;
	b[ipCH2P] += b_CH5P_CH_CH4_CH2P;
	
	
	/* >>refer mol   CH2 CH2  =>  CH3 CH   UMIST    */
	
	c[ipCH2][ipCH2] -= CH2_CH2_CH3_CH_1;
	c[ipCH2][ipCH2] -= CH2_CH2_CH3_CH_2;
	b[ipCH2] -= b_CH2_CH2_CH3_CH;
	
	c[ipCH2][ipCH3] += 0.5*CH2_CH2_CH3_CH_1;
	c[ipCH2][ipCH3] += 0.5*CH2_CH2_CH3_CH_2;
	b[ipCH3] += 0.5*b_CH2_CH2_CH3_CH;
	
	c[ipCH2][ipCH] += 0.5*CH2_CH2_CH3_CH_1;
	c[ipCH2][ipCH] += 0.5*CH2_CH2_CH3_CH_2;
	b[ipCH] += 0.5*b_CH2_CH2_CH3_CH;
	
	
	/* >>refer mol   CH4 CH2  =>  CH3 CH3   UMIST    */
	
	c[ipCH4][ipCH4] -= CH4_CH2_CH3_CH3_1;
	c[ipCH2][ipCH4] -= CH4_CH2_CH3_CH3_2;
	b[ipCH4] -= b_CH4_CH2_CH3_CH3;
	
	c[ipCH4][ipCH2] -= CH4_CH2_CH3_CH3_1;
	c[ipCH2][ipCH2] -= CH4_CH2_CH3_CH3_2;
	b[ipCH2] -= b_CH4_CH2_CH3_CH3;
	
	c[ipCH4][ipCH3] += CH4_CH2_CH3_CH3_1;
	c[ipCH2][ipCH3] += CH4_CH2_CH3_CH3_2;
	b[ipCH3] += b_CH4_CH2_CH3_CH3;
	
	c[ipCH4][ipCH3] += CH4_CH2_CH3_CH3_1;
	c[ipCH2][ipCH3] += CH4_CH2_CH3_CH3_2;
	b[ipCH3] += b_CH4_CH2_CH3_CH3;
	
	
	/* >>refer mol   OH CH2  =>  O CH3   UMIST    */
	
	c[ipOH][ipOH] -= OH_CH2_O_CH3_1;	
	c[ipCH2][ipOH] -= OH_CH2_O_CH3_2;
	b[ipOH] -= b_OH_CH2_O_CH3;
	
	c[ipOH][ipCH2] -= OH_CH2_O_CH3_1;
	c[ipCH2][ipCH2] -= OH_CH2_O_CH3_2;
	b[ipCH2] -= b_OH_CH2_O_CH3;
	
	c[ipOH][ipCH3] += OH_CH2_O_CH3_1;
	c[ipCH2][ipCH3] += OH_CH2_O_CH3_2;
	b[ipCH3] += b_OH_CH2_O_CH3;
	
	c[ipOH][ipATO] += OH_CH2_O_CH3_1;
	c[ipCH2][ipATO] += OH_CH2_O_CH3_2;
	b[ipATO] += b_OH_CH2_O_CH3;
	
	
	/* >>refer mol   CH5P CH2  =>  CH4 CH3P   UMIST    */
	
	c[ipCH5P][ipCH5P] -= CH5P_CH2_CH4_CH3P_1;
	c[ipCH2][ipCH5P] -= CH5P_CH2_CH4_CH3P_2;
	b[ipCH5P] -= b_CH5P_CH2_CH4_CH3P;
	
	c[ipCH5P][ipCH2] -= CH5P_CH2_CH4_CH3P_1;
	c[ipCH2][ipCH2] -= CH5P_CH2_CH4_CH3P_2;
	b[ipCH2] -= b_CH5P_CH2_CH4_CH3P;
	
	c[ipCH5P][ipCH4] += CH5P_CH2_CH4_CH3P_1;
	c[ipCH2][ipCH4] += CH5P_CH2_CH4_CH3P_2;
	b[ipCH4] += b_CH5P_CH2_CH4_CH3P;
	
	c[ipCH5P][ipCH3P] += CH5P_CH2_CH4_CH3P_1;
	c[ipCH2][ipCH3P] += CH5P_CH2_CH4_CH3P_2;
	b[ipCH3P] += b_CH5P_CH2_CH4_CH3P;
	
	
	/* >>refer mol   OH CH3  =>  CH4 O   UMIST    */
	
	c[ipOH][ipOH] -= OH_CH3_CH4_O_1;
	c[ipCH3][ipOH] -= OH_CH3_CH4_O_2;
	b[ipOH] -= b_OH_CH3_CH4_O;
	
	c[ipOH][ipCH3] -= OH_CH3_CH4_O_1;
	c[ipCH3][ipCH3] -= OH_CH3_CH4_O_2;
	b[ipCH3] -= b_OH_CH3_CH4_O;
	
	c[ipOH][ipCH4] += OH_CH3_CH4_O_1;
	c[ipCH3][ipCH4] += OH_CH3_CH4_O_2;
	b[ipCH4] += b_OH_CH3_CH4_O;
		
	c[ipOH][ipATO] += OH_CH3_CH4_O_1;
	c[ipCH3][ipATO] += OH_CH3_CH4_O_2;
	b[ipATO] += b_OH_CH3_CH4_O;
	
	
	/* >>refer mol   OH CH3  =>  H2O CH2   UMIST    */
	
	c[ipOH][ipOH] -= OH_CH3_H2O_CH2_1;
	c[ipCH3][ipOH] -= OH_CH3_H2O_CH2_2;
	b[ipOH] -= b_OH_CH3_H2O_CH2;

	c[ipOH][ipCH3] -= OH_CH3_H2O_CH2_1;
	c[ipCH3][ipCH3] -= OH_CH3_H2O_CH2_2;
	b[ipCH3] -= b_OH_CH3_H2O_CH2;
	
	c[ipOH][ipH2O] += OH_CH3_H2O_CH2_1;
	c[ipCH3][ipH2O] += OH_CH3_H2O_CH2_2;
	b[ipH2O] += b_OH_CH3_H2O_CH2;
	
	c[ipOH][ipCH2] += OH_CH3_H2O_CH2_1;
	c[ipCH3][ipCH2] += OH_CH3_H2O_CH2_2;
	b[ipCH2] += b_OH_CH3_H2O_CH2;
	
	
	/* >>refer mol   H2O CH3  =>  OH CH4   UMIST    */
	
	c[ipH2O][ipH2O] -= H2O_CH3_OH_CH4_1;
	c[ipCH3][ipH2O] -= H2O_CH3_OH_CH4_2;
	b[ipH2O] -= b_H2O_CH3_OH_CH4;
	
	c[ipH2O][ipCH3] -= H2O_CH3_OH_CH4_1;
	c[ipCH3][ipCH3] -= H2O_CH3_OH_CH4_2;
	b[ipCH3] -= b_H2O_CH3_OH_CH4;
	
	c[ipH2O][ipOH] += H2O_CH3_OH_CH4_1;
	c[ipCH3][ipOH] += H2O_CH3_OH_CH4_2;
	b[ipOH] += b_H2O_CH3_OH_CH4;
	
	c[ipH2O][ipCH4] += H2O_CH3_OH_CH4_1;
	c[ipCH3][ipCH4] += H2O_CH3_OH_CH4_2;
	b[ipCH4] += b_H2O_CH3_OH_CH4;


	/* >>refer mol   CH3 CH3  =>  CH4 CH2   UMIST    */
	
	c[ipCH3][ipCH3] -= CH3_CH3_CH4_CH2_1;
	c[ipCH3][ipCH3] -= CH3_CH3_CH4_CH2_2;
	b[ipCH3] -= b_CH3_CH3_CH4_CH2;
	
	c[ipCH3][ipCH4] += 0.5*CH3_CH3_CH4_CH2_1;
	c[ipCH3][ipCH4] += 0.5*CH3_CH3_CH4_CH2_2;
	b[ipCH4] += 0.5*b_CH3_CH3_CH4_CH2;
	
	c[ipCH3][ipCH2] += 0.5*CH3_CH3_CH4_CH2_1;
	c[ipCH3][ipCH2] += 0.5*CH3_CH3_CH4_CH2_2;
	b[ipCH2] += 0.5*b_CH3_CH3_CH4_CH2;
	
	
	/* >>refer mol   OH CH4  =>  H2O CH3   UMIST    */

	c[ipOH][ipOH] -= OH_CH4_H2O_CH3_1;
	c[ipCH4][ipOH] -= OH_CH4_H2O_CH3_2;
	b[ipOH] -= b_OH_CH4_H2O_CH3;
	
	c[ipOH][ipCH4] -= OH_CH4_H2O_CH3_1;
	c[ipCH4][ipCH4] -= OH_CH4_H2O_CH3_2;
	b[ipCH4] -= b_OH_CH4_H2O_CH3;
	
	c[ipOH][ipH2O] += OH_CH4_H2O_CH3_1;
	c[ipCH4][ipH2O] += OH_CH4_H2O_CH3_2;
	b[ipH2O] += b_OH_CH4_H2O_CH3;
	
	c[ipOH][ipCH3] += OH_CH4_H2O_CH3_1;
	c[ipCH4][ipCH3] += OH_CH4_H2O_CH3_2;
	b[ipCH3] += b_OH_CH4_H2O_CH3;
	
	
	/* >>refer mol   OHP CH4  =>  CH5P O   UMIST    */
	
	c[ipOHP][ipOHP] -= OHP_CH4_CH5P_O_1;
	c[ipCH4][ipOHP] -= OHP_CH4_CH5P_O_2;
	b[ipOHP] -= b_OHP_CH4_CH5P_O;
	
	c[ipOHP][ipCH4] -= OHP_CH4_CH5P_O_1;
	c[ipCH4][ipCH4] -= OHP_CH4_CH5P_O_2;
	b[ipCH4] -= b_OHP_CH4_CH5P_O;
	
	c[ipOHP][ipCH5P] += OHP_CH4_CH5P_O_1;
	c[ipCH4][ipCH5P] += OHP_CH4_CH5P_O_2;
	b[ipCH5P] += b_OHP_CH4_CH5P_O;
	
	c[ipOHP][ipATO] += OHP_CH4_CH5P_O_1;
	c[ipCH4][ipATO] += OHP_CH4_CH5P_O_2;
	b[ipATO] += b_OHP_CH4_CH5P_O;
	
	
	/* >>refer mol   OHP CH4  =>  H3OP CH2   UMIST    */
	
	c[ipOHP][ipOHP] -= OHP_CH4_H3OP_CH2_1;
	c[ipCH4][ipOHP] -= OHP_CH4_H3OP_CH2_2;
	b[ipOHP] -= b_OHP_CH4_H3OP_CH2;
	
	c[ipOHP][ipCH4] -= OHP_CH4_H3OP_CH2_1;
	c[ipCH4][ipCH4] -= OHP_CH4_H3OP_CH2_2;
	b[ipCH4] -= b_OHP_CH4_H3OP_CH2;
	
	c[ipOHP][ipH3OP] += OHP_CH4_H3OP_CH2_1;
	c[ipCH4][ipH3OP] += OHP_CH4_H3OP_CH2_2;
	b[ipH3OP] += b_OHP_CH4_H3OP_CH2;
	
	c[ipOHP][ipCH2] += OHP_CH4_H3OP_CH2_1;
	c[ipCH4][ipCH2] += OHP_CH4_H3OP_CH2_2;
	b[ipCH2] += b_OHP_CH4_H3OP_CH2;
	
	
	/* >>refer mol   H2OP CH4  =>  H3OP CH3   UMIST    */
	
	c[ipH2OP][ipH2OP] -= H2OP_CH4_H3OP_CH3_1;
	c[ipCH4][ipH2OP] -= H2OP_CH4_H3OP_CH3_2;
	b[ipH2OP] -= b_H2OP_CH4_H3OP_CH3;
	
	c[ipH2OP][ipCH4] -= H2OP_CH4_H3OP_CH3_1;
	c[ipCH4][ipCH4] -= H2OP_CH4_H3OP_CH3_2;
	b[ipCH4] -= b_H2OP_CH4_H3OP_CH3;
	
	c[ipH2OP][ipH3OP] += H2OP_CH4_H3OP_CH3_1;
	c[ipCH4][ipH3OP] += H2OP_CH4_H3OP_CH3_2;
	b[ipH3OP] += b_H2OP_CH4_H3OP_CH3;
	
	c[ipH2OP][ipCH3] += H2OP_CH4_H3OP_CH3_1;
	c[ipCH4][ipCH3] += H2OP_CH4_H3OP_CH3_2;
	b[ipCH3] += b_H2OP_CH4_H3OP_CH3;
	
	
	/* >>refer mol   COP CH4  =>  HCOP CH3   UMIST    */
	
	c[ipCOP][ipCOP] -= COP_CH4_HCOP_CH3_1;
	c[ipCH4][ipCOP] -= COP_CH4_HCOP_CH3_2;
	b[ipCOP] -= b_COP_CH4_HCOP_CH3;
	
	c[ipCOP][ipCH4] -= COP_CH4_HCOP_CH3_1;
	c[ipCH4][ipCH4] -= COP_CH4_HCOP_CH3_2;
	b[ipCH4] -= b_COP_CH4_HCOP_CH3;
	
	c[ipCOP][ipHCOP] += COP_CH4_HCOP_CH3_1;
	c[ipCH4][ipHCOP] += COP_CH4_HCOP_CH3_2;
	b[ipHCOP] += b_COP_CH4_HCOP_CH3;
	
	c[ipCOP][ipCH3] += COP_CH4_HCOP_CH3_1;
	c[ipCH4][ipCH3] += COP_CH4_HCOP_CH3_2;
	b[ipCH3] += b_COP_CH4_HCOP_CH3;
	
	
	/* >>refer mol   COP CH4  =>  CO CH4P   UMIST    */
	
	c[ipCOP][ipCOP] -= COP_CH4_CO_CH4P_1;
	c[ipCH4][ipCOP] -= COP_CH4_CO_CH4P_2;
	b[ipCOP] -= b_COP_CH4_CO_CH4P;
	
	c[ipCOP][ipCH4] -= COP_CH4_CO_CH4P_1;
	c[ipCH4][ipCH4] -= COP_CH4_CO_CH4P_2;
	b[ipCH4] -= b_COP_CH4_CO_CH4P;
	
	c[ipCOP][ipCO] += COP_CH4_CO_CH4P_1;
	c[ipCH4][ipCO] += COP_CH4_CO_CH4P_2;
	b[ipCO] += b_COP_CH4_CO_CH4P;
	
	c[ipCOP][ipCH4P] += COP_CH4_CO_CH4P_1;
	c[ipCH4][ipCH4P] += COP_CH4_CO_CH4P_2;
	b[ipCH4P] += b_COP_CH4_CO_CH4P;
	
	
	/* >>refer mol   CH4 CH4P  =>  CH5P CH3   UMIST    */
	
	c[ipCH4][ipCH4] -= CH4_CH4P_CH5P_CH3_1;
	c[ipCH4P][ipCH4] -= CH4_CH4P_CH5P_CH3_2;
	b[ipCH4] -= b_CH4_CH4P_CH5P_CH3;
	
	c[ipCH4][ipCH4P] -= CH4_CH4P_CH5P_CH3_1;
	c[ipCH4P][ipCH4P] -= CH4_CH4P_CH5P_CH3_2;
	b[ipCH4P] -= b_CH4_CH4P_CH5P_CH3;
	
	c[ipCH4][ipCH5P] += CH4_CH4P_CH5P_CH3_1;
	c[ipCH4P][ipCH5P] += CH4_CH4P_CH5P_CH3_2;
	b[ipCH5P] += b_CH4_CH4P_CH5P_CH3;
	
	c[ipCH4][ipCH3] += CH4_CH4P_CH5P_CH3_1;
	c[ipCH4P][ipCH3] += CH4_CH4P_CH5P_CH3_2;
	b[ipCH3] += b_CH4_CH4P_CH5P_CH3;
	
	
	/* >>refer mol   H2O CH4P  =>  H3OP CH3   UMIST    */
	
	c[ipH2O][ipH2O] -= H2O_CH4P_H3OP_CH3_1;
	c[ipCH4P][ipH2O] -= H2O_CH4P_H3OP_CH3_2;
	b[ipH2O] -= b_H2O_CH4P_H3OP_CH3;
	
	c[ipH2O][ipCH4P] -= H2O_CH4P_H3OP_CH3_1;
	c[ipCH4P][ipCH4P] -= H2O_CH4P_H3OP_CH3_2;
	b[ipCH4P] -= b_H2O_CH4P_H3OP_CH3;
	
	c[ipH2O][ipH3OP] += H2O_CH4P_H3OP_CH3_1;
	c[ipCH4P][ipH3OP] += H2O_CH4P_H3OP_CH3_2;
	b[ipH3OP] += b_H2O_CH4P_H3OP_CH3;
	
	c[ipH2O][ipCH3] += H2O_CH4P_H3OP_CH3_1;
	c[ipCH4P][ipCH3] += H2O_CH4P_H3OP_CH3_2;
	b[ipCH3] += b_H2O_CH4P_H3OP_CH3;
	
	
	/* >>refer mol   CO CH4P  =>  HCOP CH3   UMIST    */
	
	c[ipCO][ipCO] -= CO_CH4P_HCOP_CH3_1;
	c[ipCH4P][ipCO] -= CO_CH4P_HCOP_CH3_2;
	b[ipCO] -= b_CO_CH4P_HCOP_CH3;
	
	c[ipCO][ipCH4P] -= CO_CH4P_HCOP_CH3_1;
	c[ipCH4P][ipCH4P] -= CO_CH4P_HCOP_CH3_2;
	b[ipCH4P] -= b_CO_CH4P_HCOP_CH3;
	
	c[ipCO][ipHCOP] += CO_CH4P_HCOP_CH3_1;
	c[ipCH4P][ipHCOP] += CO_CH4P_HCOP_CH3_2;
	b[ipHCOP] += b_CO_CH4P_HCOP_CH3;
	
	c[ipCO][ipCH3] += CO_CH4P_HCOP_CH3_1;
	c[ipCH4P][ipCH3] += CO_CH4P_HCOP_CH3_2;
	b[ipCH3] += b_CO_CH4P_HCOP_CH3;
	
	
	/* >>refer mol   O2 CH4P  =>  O2P CH4   UMIST    */
	
	c[ipO2][ipO2] -= O2_CH4P_O2P_CH4_1;
	c[ipCH4P][ipO2] -= O2_CH4P_O2P_CH4_2;
	b[ipO2] -= b_O2_CH4P_O2P_CH4;
	
	c[ipO2][ipCH4P] -= O2_CH4P_O2P_CH4_1;
	c[ipCH4P][ipCH4P] -= O2_CH4P_O2P_CH4_2;
	b[ipCH4P] -= b_O2_CH4P_O2P_CH4;
	
	c[ipO2][ipO2P] += O2_CH4P_O2P_CH4_1;
	c[ipCH4P][ipO2P] += O2_CH4P_O2P_CH4_2;
	b[ipO2P] += b_O2_CH4P_O2P_CH4;
	
	c[ipO2][ipCH4] += O2_CH4P_O2P_CH4_1;
	c[ipCH4P][ipCH4] += O2_CH4P_O2P_CH4_2;
	b[ipCH4] += b_O2_CH4P_O2P_CH4;
	
	/* >>refer mol   H2O CH5P  =>  H3OP CH4   UMIST    */
	
	c[ipH2O][ipH2O] -= H2O_CH5P_H3OP_CH4_1;
	c[ipCH5P][ipH2O] -= H2O_CH5P_H3OP_CH4_2;
	b[ipH2O] -= b_H2O_CH5P_H3OP_CH4;
	
	c[ipH2O][ipCH5P] -= H2O_CH5P_H3OP_CH4_1;
	c[ipCH5P][ipCH5P] -= H2O_CH5P_H3OP_CH4_2;
	b[ipCH5P] -= b_H2O_CH5P_H3OP_CH4;

	c[ipH2O][ipH3OP] += H2O_CH5P_H3OP_CH4_1;
	c[ipCH5P][ipH3OP] += H2O_CH5P_H3OP_CH4_2;
	b[ipH3OP] += b_H2O_CH5P_H3OP_CH4;
	
	c[ipH2O][ipCH4] += H2O_CH5P_H3OP_CH4_1;
	c[ipCH5P][ipCH4] += H2O_CH5P_H3OP_CH4_2;
	b[ipCH4] += b_H2O_CH5P_H3OP_CH4;
	
	
	/* >>refer mol   CO CH5P  =>  HCOP CH4   UMIST    */
	
	c[ipCO][ipCO] -= CO_CH5P_HCOP_CH4_1;
	c[ipCH5P][ipCO] -= CO_CH5P_HCOP_CH4_2;
	b[ipCO] -= b_CO_CH5P_HCOP_CH4;
	
	c[ipCO][ipCH5P] -= CO_CH5P_HCOP_CH4_1;
	c[ipCH5P][ipCH5P] -= CO_CH5P_HCOP_CH4_2;
	b[ipCH5P] -= b_CO_CH5P_HCOP_CH4;
	
	c[ipCO][ipHCOP] += CO_CH5P_HCOP_CH4_1;
	c[ipCH5P][ipHCOP] += CO_CH5P_HCOP_CH4_2;
	b[ipHCOP] += b_CO_CH5P_HCOP_CH4;
	
	c[ipCO][ipCH4] += CO_CH5P_HCOP_CH4_1;
	c[ipCH5P][ipCH4] += CO_CH5P_HCOP_CH4_2;
	b[ipCH4] += b_CO_CH5P_HCOP_CH4;
	
	
	/* >>refer mol   CH5P OH  =>  H2OP CH4   UMIST    */
	
	c[ipCH5P][ipCH5P] -= CH5P_OH_H2OP_CH4_1;
	c[ipOH][ipCH5P] -= CH5P_OH_H2OP_CH4_2;
	b[ipCH5P] -= b_CH5P_OH_H2OP_CH4;
	
	c[ipCH5P][ipOH] -= CH5P_OH_H2OP_CH4_1;
	c[ipOH][ipOH] -= CH5P_OH_H2OP_CH4_2;
	b[ipOH] -= b_CH5P_OH_H2OP_CH4;
	
	c[ipCH5P][ipH2OP] += CH5P_OH_H2OP_CH4_1;
	c[ipOH][ipH2OP] += CH5P_OH_H2OP_CH4_2;
	b[ipH2OP] += b_CH5P_OH_H2OP_CH4;
	
	c[ipCH5P][ipCH4] += CH5P_OH_H2OP_CH4_1;
	c[ipOH][ipCH4] += CH5P_OH_H2OP_CH4_2;
	b[ipCH4] += b_CH5P_OH_H2OP_CH4;



	/* >>refer	mol	C+ + e => C	 Done internally by Cloudy*/
	/* Recombination of C+ */
	c[ipCP][ipCP] -= (ionbal.RateRecomTot[ipCARBON][0]+gv.GrainChTrRate[ipCARBON][1][0]);
	c[ipCP][ipATC] += (ionbal.RateRecomTot[ipCARBON][0]+gv.GrainChTrRate[ipCARBON][1][0]);


	/* >>refer	mol	C + hnu => C+	e	 Done internally by Cloudy*/

	/* Photoionization and grain ct ionization of Co */
	c[ipATC][ipCP] += (ionbal.RateIonizTot[ipCARBON][0]+gv.GrainChTrRate[ipCARBON][0][1]);
	c[ipATC][ipATC] -= (ionbal.RateIonizTot[ipCARBON][0]+gv.GrainChTrRate[ipCARBON][0][1]);

	/* >>refer	mol	O+ + e => O	 Done internally by Cloudy*/

	/* Recombination of O+ */
	c[ipOP][ipOP] -= (ionbal.RateRecomTot[ipOXYGEN][0]+gv.GrainChTrRate[ipOXYGEN][1][0]);
	c[ipOP][ipATO] += (ionbal.RateRecomTot[ipOXYGEN][0]+gv.GrainChTrRate[ipOXYGEN][1][0]);

	/* >>refer	mol	O + hnu => O+	e	Done internally by Cloudy*/

	/* Photoionization and grain ct ionization of Oo */
	c[ipATO][ipOP] += (ionbal.RateIonizTot[ipOXYGEN][0]+gv.GrainChTrRate[ipOXYGEN][0][1]);
	c[ipATO][ipATO] -= (ionbal.RateIonizTot[ipOXYGEN][0]+gv.GrainChTrRate[ipOXYGEN][0][1]);

	/* >>refer	mol	Si + hnu => Si+	e	Done internally by Cloudy*/

	/* Photoionization of Si */

	c[ipATSI][ipSIP] += (ionbal.RateIonizTot[ipSILICON][0]+gv.GrainChTrRate[ipSILICON][0][1]);
	c[ipATSI][ipATSI] -= (ionbal.RateIonizTot[ipSILICON][0]+gv.GrainChTrRate[ipSILICON][0][1]);

	/* >>refer	mol	Si+ + e => Si	 Done internally by Cloudy*/

	/* Recombination of Si+ */
	c[ipSIP][ipSIP] -= (ionbal.RateRecomTot[ipSILICON][0]+gv.GrainChTrRate[ipSILICON][1][0]);
	c[ipSIP][ipATSI] += (ionbal.RateRecomTot[ipSILICON][0]+gv.GrainChTrRate[ipSILICON][1][0]);
#	if 0

#	endif
	/****************************Add all terms with H2star from TH85*********************/

	
	c[ipCH][ipCH]  -= H2s_CH_C_H2_H; 
	c[ipCH][ipATC]  += H2s_CH_C_H2_H; 
	
	c[ipOH][ipOH]  -= H2s_OH_O_H2_H; 
	c[ipOH][ipATO]  += H2s_OH_O_H2_H; 
	
	c[ipH2O][ipH2O] -= H2s_H2O_OH_H2_H;
	c[ipH2O][ipOH]  += H2s_H2O_OH_H2_H; 
	
	c[ipO2][ipO2]   -= H2s_O2_O_O_H2;
	c[ipO2][ipATO]   += 2*H2s_O2_O_O_H2;
	
	c[ipCH2][ipCH2]  -= H2s_CH2_CH3_H;  
	c[ipCH2][ipCH3]  += H2s_CH2_CH3_H; 
	
	c[ipCH][ipCH]   -= H2s_CH_CH3_nu;
	c[ipCH][ipCH3]  += H2s_CH_CH3_nu;

	c[ipCH3][ipCH3]  -= H2s_CH3_CH4_H;
	c[ipCH3][ipCH4]  += H2s_CH3_CH4_H;
	
	 
	/* >>refer	mol	H2* + CH => CH2 + H	 Tielens & Hollenbach 1985, Ap. J. 291, 722 Table 9*/

	/* Formation of CH2 */
	c[ipCH][ipCH2] += h2s_ch_ch2_h;

	/* Destruction of CH */
	c[ipCH][ipCH] -=h2s_ch_ch2_h;

	/* >>refer	mol	H2* + O => OH + H	 Tielens & Hollenbach 1985, Ap. J. 291, 722 Table 9*/

	/* Formation of OH */
	
	c[ipATO][ipOH] += h2s_o_oh_h;
	c[ipATO][ipATO] -= h2s_o_oh_h;

	/* >>refer	mol	H2* + OH => H2O + H	 Tielens & Hollenbach 1985, Ap. J. 291, 722 Table 9*/

	/* Formation of H2O */
	/* >>chng 03 sep 12, sec par had been 0.67, was supposed to
	 * be zero - Gargi Shaw */
	c[ipOH][ipH2O] += h2s_o_oh_h;

	/* Destruction of OH */
	c[ipOH][ipOH] -= h2s_oh_h2o_h;

	/* >>refer	mol	H2* + O2 => 2OH	 Tielens & Hollenbach 1985, Ap. J. 291, 722 Table 9*/

	/* >>chng 03 sep 12, sec par was supposed to be 0, was
	 * 0.67, caught by Gargi Shaw */
	/* Formation of OH */
	c[ipO2][ipOH] += 2*h2s_o2_twooh;

	/* Destruction of O2 */
	c[ipO2][ipO2] -= h2s_o2_twooh;

	/* >>refer	mol	H2* + C => CH + H	 Tielens & Hollenbach 1985, Ap. J. 291, 722 Table 9*/

	/* Formation of CH */
	c[ipATC][ipCH] += h2s_c_ch_h;
	c[ipATC][ipATC] -= h2s_c_ch_h;

	/* >>refer	mol	H2* + C+ => CH+ + H	Tielens & Hollenbach 1985, Ap. J. 291, 722 Table 9*/

	/* Formation of CH+ */
	c[ipCP][ipCHP] += h2s_cp_chp_h;

	/* Destruction of C+ */
	c[ipCP][ipCP] -= h2s_cp_chp_h;

	co_e_c_o_e = 0;

	/* inner shell photoionization of CO, assume rates are same as K 1s and 2s
	 * shell of C and O */
	/* >>chng 04 may 26, upper limit should be ns<2, had been ns<2 so picked up
	 * valence shell, which was incorrect */
	for ( ns=0; ns<2; ++ns )
	{
		ion = 0;
		co_e_c_o_e += ionbal.PhotoRate_Shell[ipCARBON][ion][ns][0];
		co_e_c_o_e += ionbal.PhotoRate_Shell[ipOXYGEN][ion][ns][0];
	}
	c[ipCO][ipCO] -= co_e_c_o_e*co.lgUMISTrates;
	c[ipCO][ipATO] += co_e_c_o_e*co.lgUMISTrates;
	c[ipCO][ipATC] += co_e_c_o_e*co.lgUMISTrates;



		/*  >>refer   mol  H2     +  N     =>   NH    +     H     UMIST  */

			c[ipATN][ipNH]  += H2_N_NH_H;
			c[ipATN][ipATN]  -= H2_N_NH_H;

		/*  >>refer   mol  H2     +  NH    =>   NH2   +     H	    UMIST  */
			
			c[ipNH][ipNH]  -= H2_NH_NH2_H;
			c[ipNH][ipNH2] += H2_NH_NH2_H;

		/*  >>refer   mol  H2     +  NH2   =>   NH3   +     H	    UMIST  */

			c[ipNH2][ipNH2] -= H2_NH2_NH3_H;
			c[ipNH2][ipNH3] += H2_NH2_NH3_H;

		/*  >>refer   mol  H2     +  CN    =>   HCN   +     H	    UMIST  */

			c[ipCN][ipCN]  -= H2_CN_HCN_H;
			c[ipCN][ipHCN] += H2_CN_HCN_H;

		/*  >>refer   mol  C      +  NH    =>   N     +     CH    UMIST  */

			c[ipNH][ipNH] -= C_NH_N_CH_1;
			c[ipATC][ipNH] -= C_NH_N_CH_2;
			b[ipNH] -= b_C_NH_N_CH;

			c[ipNH][ipATC] -= C_NH_N_CH_1;
			c[ipATC][ipATC] -= C_NH_N_CH_2;
			b[ipATC] -= b_C_NH_N_CH;
			
			c[ipNH][ipCH] += C_NH_N_CH_1;
			c[ipATC][ipCH] += C_NH_N_CH_2;
			b[ipCH] += b_C_NH_N_CH;
			
			c[ipNH][ipATN] += C_NH_N_CH_1;
			c[ipATC][ipATN] += C_NH_N_CH_2;
			b[ipATN] += b_C_NH_N_CH;

		/*  >>refer   mol  C      +  NH    =>   CN    +     H	    UMIST  */

			c[ipNH][ipNH] -= C_NH_CN_H_1;
			c[ipATC][ipNH] -= C_NH_CN_H_2;
			b[ipNH] -= b_C_NH_CN_H;

			c[ipNH][ipATC] -= C_NH_CN_H_1;
			c[ipATC][ipATC] -= C_NH_CN_H_2;
			b[ipATC] -= b_C_NH_CN_H;
			
			c[ipNH][ipCN] += C_NH_CN_H_1;
			c[ipATC][ipCN] += C_NH_CN_H_2;
			b[ipCN] += b_C_NH_CN_H;

		/*  >>refer   mol  C      +  N2    =>   CN    +     N	    UMIST  */

			c[ipN2][ipN2] -= C_N2_CN_N_1;
			c[ipATC][ipN2] -= C_N2_CN_N_2;
			b[ipN2] -= b_C_N2_CN_N;

			c[ipN2][ipATC] -= C_N2_CN_N_1;
			c[ipATC][ipATC] -= C_N2_CN_N_2;
			b[ipATC] -= b_C_N2_CN_N;
			
			c[ipN2][ipCN] += C_N2_CN_N_1;
			c[ipATC][ipCN] += C_N2_CN_N_2;
			b[ipCN] += b_C_N2_CN_N;
			
			c[ipN2][ipATN] += C_N2_CN_N_1;
			c[ipATC][ipATN] += C_N2_CN_N_2;
			b[ipATN] += b_C_N2_CN_N;

		/*  >>refer   mol  C      +  NO    =>   CN    +     O	    UMIST  */

			c[ipNO][ipNO] -= C_NO_CN_O_1;
			c[ipATC][ipNO] -= C_NO_CN_O_2;
			b[ipNO] -= b_C_NO_CN_O;
			
			c[ipNO][ipATC] -= C_NO_CN_O_1;
			c[ipATC][ipATC] -= C_NO_CN_O_2;
			b[ipATC] -= b_C_NO_CN_O;
			
			c[ipNO][ipCN] += C_NO_CN_O_1;
			c[ipATC][ipCN] += C_NO_CN_O_2;
			b[ipCN] += b_C_NO_CN_O;

			c[ipNO][ipATO] += C_NO_CN_O_1;
			c[ipATC][ipATO] += C_NO_CN_O_2;
			b[ipATO] += b_C_NO_CN_O;

		/*  >>refer   mol  C      +  NO    =>   CO    +     N	    UMIST  */

			c[ipNO][ipNO] -= C_NO_CO_N_1;
			c[ipATC][ipNO] -= C_NO_CO_N_2;
			b[ipNO] -= b_C_NO_CO_N;

			c[ipNO][ipATC] -= C_NO_CO_N_1;
			c[ipATC][ipATC] -= C_NO_CO_N_2;
			b[ipATC] -= b_C_NO_CO_N;
			
			c[ipNO][ipCO] += C_NO_CO_N_1;
			c[ipATC][ipCO] += C_NO_CO_N_2;
			b[ipCO] += b_C_NO_CO_N;
			
			c[ipNO][ipATN] += C_NO_CO_N_1;
			c[ipATC][ipATN] += C_NO_CO_N_2;
			b[ipATN] += b_C_NO_CO_N;

		/*  >>refer   mol  C      +  HS    =>   S     +     CH    UMIST  */

			c[ipHS][ipHS] -= C_HS_S_CH_1;
			c[ipATC][ipHS] -= C_HS_S_CH_2;
			b[ipHS] -= b_C_HS_S_CH;
	
			c[ipHS][ipATC] -= C_HS_S_CH_1;
			c[ipATC][ipATC] -= C_HS_S_CH_2;
			b[ipATC] -= b_C_HS_S_CH;
			
			c[ipHS][ipCH] += C_HS_S_CH_1;
			c[ipATC][ipCH] += C_HS_S_CH_2;
			b[ipCH] += b_C_HS_S_CH;
		
			c[ipHS][ipATS] += C_HS_S_CH_1;
			c[ipATC][ipATS] += C_HS_S_CH_2;
			b[ipATS] += b_C_HS_S_CH;

		/*  >>refer   mol  C      +  HS    =>   CS    +     H	    UMIST  */

			c[ipHS][ipHS] -= C_HS_CS_H_1;
			c[ipATC][ipHS] -= C_HS_CS_H_2;
			b[ipHS] -= b_C_HS_CS_H;

			c[ipHS][ipATC] -= C_HS_CS_H_1;
			c[ipATC][ipATC] -= C_HS_CS_H_2;
			b[ipATC] -= b_C_HS_CS_H;
			
			c[ipHS][ipCS] += C_HS_CS_H_1;
			c[ipATC][ipCS] += C_HS_CS_H_2;
			b[ipCS] += b_C_HS_CS_H;

		/*  >>refer   mol  C      +  OCN   =>   CO    +     CN    UMIST  */

			c[ipOCN][ipOCN] -= C_OCN_CO_CN_1;
			c[ipATC][ipOCN] -= C_OCN_CO_CN_2;
			b[ipOCN] -= b_C_OCN_CO_CN;

			c[ipOCN][ipATC] -= C_OCN_CO_CN_1;
			c[ipATC][ipATC] -= C_OCN_CO_CN_2;
			b[ipATC] -= b_C_OCN_CO_CN;
			
			c[ipOCN][ipCO] += C_OCN_CO_CN_1;
			c[ipATC][ipCO] += C_OCN_CO_CN_2;
			b[ipCO] += b_C_OCN_CO_CN;
			
			c[ipOCN][ipCN] += C_OCN_CO_CN_1;
			c[ipATC][ipCN] += C_OCN_CO_CN_2;
			b[ipCN] += b_C_OCN_CO_CN;

		/*  >>refer   mol  C      +  NS    =>   S     +     CN    UMIST  */

			c[ipNS][ipATC] -= C_NS_S_CN_1;
			c[ipATC][ipATC] -= C_NS_S_CN_2;
			b[ipATC] -= b_C_NS_S_CN;
			
			c[ipNS][ipNS] -= C_NS_S_CN_1;
			c[ipATC][ipNS] -= C_NS_S_CN_2;
			b[ipNS] -= b_C_NS_S_CN;
			
			c[ipNS][ipCN] += C_NS_S_CN_1;
			c[ipATC][ipCN] += C_NS_S_CN_2;
			b[ipCN] += b_C_NS_S_CN;

			c[ipNS][ipATS] += C_NS_S_CN_1;
			c[ipATC][ipATS] += C_NS_S_CN_2;
			b[ipATS] += b_C_NS_S_CN;

		/*  >>refer   mol  C      +  NS    =>   CS    +     N	    UMIST  */

			c[ipNS][ipATC] -= C_NS_CS_N_1;
			c[ipATC][ipATC] -= C_NS_CS_N_2;
			b[ipATC] -= b_C_NS_CS_N;
			
			c[ipNS][ipNS] -= C_NS_CS_N_1;
			c[ipATC][ipNS] -= C_NS_CS_N_2;
			b[ipNS] -= b_C_NS_CS_N;
			
			c[ipNS][ipCS] += C_NS_CS_N_1;
			c[ipATC][ipCS] += C_NS_CS_N_2;
			b[ipCS] += b_C_NS_CS_N;

			c[ipNS][ipATN] += C_NS_CS_N_1;
			c[ipATC][ipATN] += C_NS_CS_N_2;
			b[ipATN] += b_C_NS_CS_N;

		/*  >>refer   mol  C      +  SO    =>   S     +     CO    UMIST  */

			c[ipSO][ipATC] -= C_SO_S_CO_1;
			c[ipATC][ipATC] -= C_SO_S_CO_2;
			b[ipATC] -= b_C_SO_S_CO;
			
			c[ipSO][ipSO] -= C_SO_S_CO_1;
			c[ipATC][ipSO] -= C_SO_S_CO_2;
			b[ipSO] -= b_C_SO_S_CO;
			
			c[ipSO][ipCO] += C_SO_S_CO_1;
			c[ipATC][ipCO] += C_SO_S_CO_2;
			b[ipCO] += b_C_SO_S_CO;

			c[ipSO][ipATS] += C_SO_S_CO_1;
			c[ipATC][ipATS] += C_SO_S_CO_2;
			b[ipATS] += b_C_SO_S_CO;

		/*  >>refer   mol  C      +  S2    =>   CS    +     S	    UMIST  */

			c[ipS2][ipATC] -= C_S2_CS_S_1;
			c[ipATC][ipATC] -= C_S2_CS_S_2;
			b[ipATC] -= b_C_S2_CS_S;
			
			c[ipS2][ipS2] -= C_S2_CS_S_1;
			c[ipATC][ipS2] -= C_S2_CS_S_2;
			b[ipS2] -= b_C_S2_CS_S;
			
			c[ipS2][ipCS] += C_S2_CS_S_1;
			c[ipATC][ipCS] += C_S2_CS_S_2;
			b[ipCS] += b_C_S2_CS_S;

			c[ipS2][ipATS] += C_S2_CS_S_1;
			c[ipATC][ipATS] += C_S2_CS_S_2;
			b[ipATS] += b_C_S2_CS_S;

		/*  >>refer   mol  CH     +  N     =>   NH    +     C	    UMIST  */

			c[ipATN][ipCH] -= CH_N_NH_C_1;
			c[ipCH][ipCH] -= CH_N_NH_C_2;
			b[ipCH] -= b_CH_N_NH_C;
			
			c[ipATN][ipATN] -= CH_N_NH_C_1;
			c[ipCH][ipATN] -= CH_N_NH_C_2;
			b[ipATN] -= b_CH_N_NH_C;
			
			c[ipATN][ipNH] += CH_N_NH_C_1;
			c[ipCH][ipNH] += CH_N_NH_C_2;
			b[ipNH] += b_CH_N_NH_C;

			c[ipATN][ipATC] += CH_N_NH_C_1;
			c[ipCH][ipATC] += CH_N_NH_C_2;
			b[ipATC] += b_CH_N_NH_C;

		/*  >>refer   mol  CH     +  N     =>   CN    +     H	    UMIST  */

			c[ipATN][ipCH] -= CH_N_CN_H_1;
			c[ipCH][ipCH] -= CH_N_CN_H_2;
			b[ipCH] -= b_CH_N_CN_H;

			c[ipATN][ipATN] -= CH_N_CN_H_1;
			c[ipCH][ipATN] -= CH_N_CN_H_2;
			b[ipATN] -= b_CH_N_CN_H;
			
			c[ipATN][ipCN] += CH_N_CN_H_1;
			c[ipCH][ipCN] += CH_N_CN_H_2;
			b[ipCN] += b_CH_N_CN_H;

		/*  >>refer   mol  CH     +  N2    =>   HCN   +     N	    UMIST  */

			c[ipN2][ipCH] -= CH_N2_HCN_N_1;
			c[ipCH][ipCH] -= CH_N2_HCN_N_2;
			b[ipCH] -= b_CH_N2_HCN_N;
			
			c[ipN2][ipN2] -= CH_N2_HCN_N_1;
			c[ipCH][ipN2] -= CH_N2_HCN_N_2;
			b[ipN2] -= b_CH_N2_HCN_N;
			
			c[ipN2][ipHCN] += CH_N2_HCN_N_1;
			c[ipCH][ipHCN] += CH_N2_HCN_N_2;
			b[ipHCN] += b_CH_N2_HCN_N;
			
			c[ipN2][ipATN] += CH_N2_HCN_N_1;
			c[ipCH][ipATN] += CH_N2_HCN_N_2;
			b[ipATN] += b_CH_N2_HCN_N;

		/*  >>refer   mol  CH     +  NO    =>   HCN   +     O	    UMIST  */

			c[ipNO][ipCH] -= CH_NO_HCN_O_1;
			c[ipCH][ipCH] -= CH_NO_HCN_O_2;
			b[ipCH] -= b_CH_NO_HCN_O;
			
			c[ipNO][ipNO] -= CH_NO_HCN_O_1;
			c[ipCH][ipNO] -= CH_NO_HCN_O_2;
			b[ipNO] -= b_CH_NO_HCN_O;
			
			c[ipNO][ipHCN] += CH_NO_HCN_O_1;
			c[ipCH][ipHCN] += CH_NO_HCN_O_2;
			b[ipHCN] += b_CH_NO_HCN_O;
			
			c[ipNO][ipATO] += CH_NO_HCN_O_1;
			c[ipCH][ipATO] += CH_NO_HCN_O_2;
			b[ipATO] += b_CH_NO_HCN_O;

		/*  >>refer   mol  CH     +  NO    =>   CN    +     OH    UMIST  */

			c[ipNO][ipCH] -= CH_NO_CN_OH_1;
			c[ipCH][ipCH] -= CH_NO_CN_OH_2;
			b[ipCH] -= b_CH_NO_CN_OH;
			
			c[ipNO][ipNO] -= CH_NO_CN_OH_1;
			c[ipCH][ipNO] -= CH_NO_CN_OH_2;
			b[ipNO] -= b_CH_NO_CN_OH;
			
			c[ipNO][ipCN] += CH_NO_CN_OH_1;
			c[ipCH][ipCN] += CH_NO_CN_OH_2;
			b[ipCN] += b_CH_NO_CN_OH;
			
			c[ipNO][ipOH] += CH_NO_CN_OH_1;
			c[ipCH][ipOH] += CH_NO_CN_OH_2;
			b[ipOH] += b_CH_NO_CN_OH;

		/*  >>refer   mol  CH     +  NO    =>   OCN   +     H	    UMIST  */

			c[ipNO][ipCH] -= CH_NO_OCN_H_1;
			c[ipCH][ipCH] -= CH_NO_OCN_H_2;
			b[ipCH] -= b_CH_NO_OCN_H;
			
			c[ipNO][ipNO] -= CH_NO_OCN_H_1;
			c[ipCH][ipNO] -= CH_NO_OCN_H_2;
			b[ipNO] -= b_CH_NO_OCN_H;
			
			c[ipNO][ipOCN] += CH_NO_OCN_H_1;
			c[ipCH][ipOCN] += CH_NO_OCN_H_2;
			b[ipOCN] += b_CH_NO_OCN_H;

		/*  >>refer   mol  CH     +  HNO   =>   NO    +     CH2   UMIST  */

			c[ipHNO][ipCH] -= CH_HNO_NO_CH2_1;
			c[ipCH][ipCH] -= CH_HNO_NO_CH2_2;
			b[ipCH] -= b_CH_HNO_NO_CH2;
			
			c[ipHNO][ipHNO] -= CH_HNO_NO_CH2_1;
			c[ipCH][ipHNO] -= CH_HNO_NO_CH2_2;
			b[ipHNO] -= b_CH_HNO_NO_CH2;
			
			c[ipHNO][ipNO] += CH_HNO_NO_CH2_1;
			c[ipCH][ipNO] += CH_HNO_NO_CH2_2;
			b[ipNO] += b_CH_HNO_NO_CH2;
			
			c[ipHNO][ipCH2] += CH_HNO_NO_CH2_1;
			c[ipCH][ipCH2] += CH_HNO_NO_CH2_2;
			b[ipCH2] += b_CH_HNO_NO_CH2;

		/*  >>refer   mol  CH     +  S     =>   HS    +     C	    UMIST  */

			c[ipATS][ipCH] -= CH_S_HS_C_1;
			c[ipCH][ipCH] -= CH_S_HS_C_2;
			b[ipCH] -= b_CH_S_HS_C;
			
			c[ipATS][ipATS] -= CH_S_HS_C_1;
			c[ipCH][ipATS] -= CH_S_HS_C_2;
			b[ipATS] -= b_CH_S_HS_C;
			
			c[ipATS][ipHS] += CH_S_HS_C_1;
			c[ipCH][ipHS] += CH_S_HS_C_2;
			b[ipHS] += b_CH_S_HS_C;

			c[ipATS][ipATC] += CH_S_HS_C_1;
			c[ipCH][ipATC] += CH_S_HS_C_2;
			b[ipATC] += b_CH_S_HS_C;

		/*  >>refer   mol  CH     +  S     =>   CS    +     H	    UMIST  */

			c[ipATS][ipCH] -= CH_S_CS_H_1;
			c[ipCH][ipCH] -= CH_S_CS_H_2;
			b[ipCH] -= b_CH_S_CS_H;

			c[ipATS][ipATS] -= CH_S_CS_H_1;
			c[ipCH][ipATS] -= CH_S_CS_H_2;
			b[ipATS] -= b_CH_S_CS_H;
			
			c[ipATS][ipCS] += CH_S_CS_H_1;
			c[ipCH][ipCS] += CH_S_CS_H_2;
			b[ipCS] += b_CH_S_CS_H;

		/*  >>refer   mol  N      +  NH    =>   N2    +     H	    UMIST  */

			c[ipNH][ipNH] -= N_NH_N2_H_1;
			c[ipATN][ipNH] -= N_NH_N2_H_2;
			b[ipNH] -= b_N_NH_N2_H;

			c[ipNH][ipATN] -= N_NH_N2_H_1;
			c[ipATN][ipATN] -= N_NH_N2_H_2;
			b[ipATN] -= b_N_NH_N2_H;
			
			c[ipNH][ipN2] += N_NH_N2_H_1;
			c[ipATN][ipN2] += N_NH_N2_H_2;
			b[ipN2] += b_N_NH_N2_H;

		/*  >>refer   mol  N      +  CH3   =>   HCN   +     H  +  H      UMIST  */

			c[ipCH3][ipCH3] -= N_CH3_HCN_H_H_1;
			c[ipATN][ipCH3] -= N_CH3_HCN_H_H_2;
			b[ipCH3] -= b_N_CH3_HCN_H_H;

			c[ipCH3][ipATN] -= N_CH3_HCN_H_H_1;
			c[ipATN][ipATN] -= N_CH3_HCN_H_H_2;
			b[ipATN] -= b_N_CH3_HCN_H_H;
			
			c[ipCH3][ipHCN] += N_CH3_HCN_H_H_1;
			c[ipATN][ipHCN] += N_CH3_HCN_H_H_2;
			b[ipHCN] += b_N_CH3_HCN_H_H;

		/*  >>refer   mol  N      +  CH3   =>   HCN   +     H2    UMIST  */

			c[ipCH3][ipCH3] -= N_CH3_HCN_H2_1;
			c[ipATN][ipCH3] -= N_CH3_HCN_H2_2;
			b[ipCH3] -= b_N_CH3_HCN_H2;

			c[ipCH3][ipATN] -= N_CH3_HCN_H2_1;
			c[ipATN][ipATN] -= N_CH3_HCN_H2_2;
			b[ipATN] -= b_N_CH3_HCN_H2;
			
			c[ipCH3][ipHCN] += N_CH3_HCN_H2_1;
			c[ipATN][ipHCN] += N_CH3_HCN_H2_2;
			b[ipHCN] += b_N_CH3_HCN_H2;

		/*  >>refer   mol  N      +  OH    =>   O     +     NH    UMIST  */

			c[ipOH][ipATN] -= N_OH_O_NH_1;
			c[ipATN][ipATN] -= N_OH_O_NH_2;
			b[ipATN] -= b_N_OH_O_NH;
			
			c[ipOH][ipOH] -= N_OH_O_NH_1;
			c[ipATN][ipOH] -= N_OH_O_NH_2;
			b[ipOH] -= b_N_OH_O_NH;
			
			c[ipOH][ipNH] += N_OH_O_NH_1;
			c[ipATN][ipNH] += N_OH_O_NH_2;
			b[ipNH] += b_N_OH_O_NH;

			c[ipOH][ipATO] += N_OH_O_NH_1;
			c[ipATN][ipATO] += N_OH_O_NH_2;
			b[ipATO] += b_N_OH_O_NH;

		/*  >>refer   mol  N      +  OH    =>   NO    +     H	    UMIST  */

			c[ipOH][ipOH] -= N_OH_NO_H_1;
			c[ipATN][ipOH] -= N_OH_NO_H_2;
			b[ipOH] -= b_N_OH_NO_H;

			c[ipOH][ipATN] -= N_OH_NO_H_1;
			c[ipATN][ipATN] -= N_OH_NO_H_2;
			b[ipATN] -= b_N_OH_NO_H;
			
			c[ipOH][ipNO] += N_OH_NO_H_1;
			c[ipATN][ipNO] += N_OH_NO_H_2;
			b[ipNO] += b_N_OH_NO_H;

		/*  >>refer   mol  N      +  CN    =>   N2    +     C	    UMIST  */

			c[ipCN][ipCN] -= N_CN_N2_C_1;
			c[ipATN][ipCN] -= N_CN_N2_C_2;
			b[ipCN] -= b_N_CN_N2_C;
			
			c[ipCN][ipATN] -= N_CN_N2_C_1;
			c[ipATN][ipATN] -= N_CN_N2_C_2;
			b[ipATN] -= b_N_CN_N2_C;
			
			c[ipCN][ipN2] += N_CN_N2_C_1;
			c[ipATN][ipN2] += N_CN_N2_C_2;
			b[ipN2] += b_N_CN_N2_C;

			c[ipCN][ipATC] += N_CN_N2_C_1;
			c[ipATN][ipATC] += N_CN_N2_C_2;
			b[ipATC] += b_N_CN_N2_C;

		/*  >>refer   mol  N      +  SIH   =>   SIN   +     H	    UMIST  */

			c[ipSIH][ipSIH] -= N_SIH_SIN_H_1;
			c[ipATN][ipSIH] -= N_SIH_SIN_H_2;
			b[ipSIH] -= b_N_SIH_SIN_H;

			c[ipSIH][ipATN] -= N_SIH_SIN_H_1;
			c[ipATN][ipATN] -= N_SIH_SIN_H_2;
			b[ipATN] -= b_N_SIH_SIN_H;
			
			c[ipSIH][ipSIN] += N_SIH_SIN_H_1;
			c[ipATN][ipSIN] += N_SIH_SIN_H_2;
			b[ipSIN] += b_N_SIH_SIN_H;

		/*  >>refer   mol  N      +  NO    =>   N2    +     O	    UMIST  */

			c[ipNO][ipNO] -= N_NO_N2_O_1;
			c[ipATN][ipNO] -= N_NO_N2_O_2;
			b[ipNO] -= b_N_NO_N2_O;
			
			c[ipNO][ipATN] -= N_NO_N2_O_1;
			c[ipATN][ipATN] -= N_NO_N2_O_2;
			b[ipATN] -= b_N_NO_N2_O;
			
			c[ipNO][ipN2] += N_NO_N2_O_1;
			c[ipATN][ipN2] += N_NO_N2_O_2;
			b[ipN2] += b_N_NO_N2_O;

			c[ipNO][ipATO] += N_NO_N2_O_1;
			c[ipATN][ipATO] += N_NO_N2_O_2;
			b[ipATO] += b_N_NO_N2_O;

		/*  >>refer   mol  N      +  HNO   =>   NO    +     NH    UMIST  */

			c[ipHNO][ipHNO] -= N_HNO_NO_NH_1;
			c[ipATN][ipHNO] -= N_HNO_NO_NH_2;
			b[ipHNO] -= b_N_HNO_NO_NH;
			
			c[ipHNO][ipATN] -= N_HNO_NO_NH_1;
			c[ipATN][ipATN] -= N_HNO_NO_NH_2;
			b[ipATN] -= b_N_HNO_NO_NH;

			c[ipHNO][ipNO] += N_HNO_NO_NH_1;
			c[ipATN][ipNO] += N_HNO_NO_NH_2;
			b[ipNO] += b_N_HNO_NO_NH;
			
			c[ipHNO][ipNH] += N_HNO_NO_NH_1;
			c[ipATN][ipNH] += N_HNO_NO_NH_2;
			b[ipNH] += b_N_HNO_NO_NH;

		/*  >>refer   mol  N      +  HNO   =>   N2O   +     H	    UMIST  */

			c[ipHNO][ipHNO] -= N_HNO_N2O_H_1;
			c[ipATN][ipHNO] -= N_HNO_N2O_H_2;
			b[ipHNO] -= b_N_HNO_N2O_H;
			
			c[ipHNO][ipATN] -= N_HNO_N2O_H_1;
			c[ipATN][ipATN] -= N_HNO_N2O_H_2;
			b[ipATN] -= b_N_HNO_N2O_H;

			c[ipHNO][ipN2O] += N_HNO_N2O_H_1;
			c[ipATN][ipN2O] += N_HNO_N2O_H_2;
			b[ipN2O] += b_N_HNO_N2O_H;

		/*  >>refer   mol  N      +  O2    =>   NO    +     O	    UMIST  */

			c[ipO2][ipO2] -= N_O2_NO_O_1;
			c[ipATN][ipO2] -= N_O2_NO_O_2;
			b[ipO2] -= b_N_O2_NO_O;
			
			c[ipO2][ipATN] -= N_O2_NO_O_1;
			c[ipATN][ipATN] -= N_O2_NO_O_2;
			b[ipATN] -= b_N_O2_NO_O;
			
			c[ipO2][ipNO] += N_O2_NO_O_1;
			c[ipATN][ipNO] += N_O2_NO_O_2;
			b[ipNO] += b_N_O2_NO_O;

			c[ipO2][ipATO] += N_O2_NO_O_1;
			c[ipATN][ipATO] += N_O2_NO_O_2;
			b[ipATO] += b_N_O2_NO_O;

		/*  >>refer   mol  N      +  HS    =>   NS    +     H     UMIST  */

			c[ipHS][ipHS] -= N_HS_NS_H_1;
			c[ipATN][ipHS] -= N_HS_NS_H_2;
			b[ipHS] -= b_N_HS_NS_H;

			c[ipHS][ipATN] -= N_HS_NS_H_1;
			c[ipATN][ipATN] -= N_HS_NS_H_2;
			b[ipATN] -= b_N_HS_NS_H;
			
			c[ipHS][ipNS] += N_HS_NS_H_1;
			c[ipATN][ipNS] += N_HS_NS_H_2;
			b[ipNS] += b_N_HS_NS_H;

		/*  >>refer   mol  N      +  HS    =>   S     +     NH    UMIST  */

			c[ipHS][ipATN] -= N_HS_S_NH_1;
			c[ipATN][ipATN] -= N_HS_S_NH_2;
			b[ipATN] -= b_N_HS_S_NH;
			
			c[ipHS][ipHS] -= N_HS_S_NH_1;
			c[ipATN][ipHS] -= N_HS_S_NH_2;
			b[ipHS] -= b_N_HS_S_NH;
			
			c[ipHS][ipNH] += N_HS_S_NH_1;
			c[ipATN][ipNH] += N_HS_S_NH_2;
			b[ipNH] += b_N_HS_S_NH;

			c[ipHS][ipATS] += N_HS_S_NH_1;
			c[ipATN][ipATS] += N_HS_S_NH_2;
			b[ipATS] += b_N_HS_S_NH;

		/*  >>refer   mol  N      +  CS    =>   S     +     CN    UMIST  */

			c[ipCS][ipATN] -= N_CS_S_CN_1;
			c[ipATN][ipATN] -= N_CS_S_CN_2;
			b[ipATN] -= b_N_CS_S_CN;
			
			c[ipCS][ipCS] -= N_CS_S_CN_1;
			c[ipATN][ipCS] -= N_CS_S_CN_2;
			b[ipCS] -= b_N_CS_S_CN;
			
			c[ipCS][ipCN] += N_CS_S_CN_1;
			c[ipATN][ipCN] += N_CS_S_CN_2;
			b[ipCN] += b_N_CS_S_CN;

			c[ipCS][ipATS] += N_CS_S_CN_1;
			c[ipATN][ipATS] += N_CS_S_CN_2;
			b[ipATS] += b_N_CS_S_CN;

		/*  >>refer   mol  N      +  NS    =>   S     +     N2    UMIST  */

			c[ipNS][ipNS] -= N_NS_S_N2_1;
			c[ipATN][ipNS] -= N_NS_S_N2_2;
			b[ipNS] -= b_N_NS_S_N2;
			
			c[ipNS][ipATN] -= N_NS_S_N2_1;
			c[ipATN][ipATN] -= N_NS_S_N2_2;
			b[ipATN] -= b_N_NS_S_N2;
			
			c[ipNS][ipN2] += N_NS_S_N2_1;
			c[ipATN][ipN2] += N_NS_S_N2_2;
			b[ipN2] += b_N_NS_S_N2;

			c[ipNS][ipATS] += N_NS_S_N2_1;
			c[ipATN][ipATS] += N_NS_S_N2_2;
			b[ipATS] += b_N_NS_S_N2;

		/*  >>refer   mol  N      +  SO    =>   NS    +     O	    UMIST  */

			c[ipSO][ipSO] -= N_SO_NS_O_1;
			c[ipATN][ipSO] -= N_SO_NS_O_2;
			b[ipSO] -= b_N_SO_NS_O;

			c[ipSO][ipATN] -= N_SO_NS_O_1;
			c[ipATN][ipATN] -= N_SO_NS_O_2;
			b[ipATN] -= b_N_SO_NS_O;

			c[ipSO][ipNS] += N_SO_NS_O_1;
			c[ipATN][ipNS] += N_SO_NS_O_2;
			b[ipNS] += b_N_SO_NS_O;

			c[ipSO][ipATO] += N_SO_NS_O_1;
			c[ipATN][ipATO] += N_SO_NS_O_2;
			b[ipATO] += b_N_SO_NS_O;

		/*  >>refer   mol  N      +  SO    =>   S     +     NO    UMIST  */

			c[ipSO][ipSO] -= N_SO_S_NO_1;
			c[ipATN][ipSO] -= N_SO_S_NO_2;
			b[ipSO] -= b_N_SO_S_NO;
			
			c[ipSO][ipATN] -= N_SO_S_NO_1;
			c[ipATN][ipATN] -= N_SO_S_NO_2;
			b[ipATN] -= b_N_SO_S_NO;
			
			c[ipSO][ipNO] += N_SO_S_NO_1;
			c[ipATN][ipNO] += N_SO_S_NO_2;
			b[ipNO] += b_N_SO_S_NO;

			c[ipSO][ipATS] += N_SO_S_NO_1;
			c[ipATN][ipATS] += N_SO_S_NO_2;
			b[ipATS] += b_N_SO_S_NO;

		/*  >>refer   mol  N      +  S2    =>   NS    +     S     UMIST  */

			c[ipS2][ipS2] -= N_S2_NS_S_1;
			c[ipATN][ipS2] -= N_S2_NS_S_2;
			b[ipS2] -= b_N_S2_NS_S;
			
			c[ipS2][ipATN] -= N_S2_NS_S_1;
			c[ipATN][ipATN] -= N_S2_NS_S_2;
			b[ipATN] -= b_N_S2_NS_S;
			
			c[ipS2][ipNS] += N_S2_NS_S_1;
			c[ipATN][ipNS] += N_S2_NS_S_2;
			b[ipNS] += b_N_S2_NS_S;

			c[ipS2][ipATS] += N_S2_NS_S_1;
			c[ipATN][ipATS] += N_S2_NS_S_2;
			b[ipATS] += b_N_S2_NS_S;

		/*  >>refer   mol  CH2    +  CN    =>   HCN   +     CH    UMIST  */

			c[ipCN][ipCH2] -= CH2_CN_HCN_CH_1;
			c[ipCH2][ipCH2] -= CH2_CN_HCN_CH_2;
			b[ipCH2] -= b_CH2_CN_HCN_CH;
			
			c[ipCN][ipCN] -= CH2_CN_HCN_CH_1;
			c[ipCH2][ipCN] -= CH2_CN_HCN_CH_2;
			b[ipCN] -= b_CH2_CN_HCN_CH;
			
			c[ipCN][ipHCN] += CH2_CN_HCN_CH_1;
			c[ipCH2][ipHCN] += CH2_CN_HCN_CH_2;
			b[ipHCN] += b_CH2_CN_HCN_CH;
			
			c[ipCN][ipCH] += CH2_CN_HCN_CH_1;
			c[ipCH2][ipCH] += CH2_CN_HCN_CH_2;
			b[ipCH] += b_CH2_CN_HCN_CH;

		/*  >>refer   mol  CH2    +  NO    =>   HCN   +     OH    UMIST  */

			c[ipNO][ipCH2] -= CH2_NO_HCN_OH_1;
			c[ipCH2][ipCH2] -= CH2_NO_HCN_OH_2;
			b[ipCH2] -= b_CH2_NO_HCN_OH;
			
			c[ipNO][ipNO] -= CH2_NO_HCN_OH_1;
			c[ipCH2][ipNO] -= CH2_NO_HCN_OH_2;
			b[ipNO] -= b_CH2_NO_HCN_OH;
			
			c[ipNO][ipHCN] += CH2_NO_HCN_OH_1;
			c[ipCH2][ipHCN] += CH2_NO_HCN_OH_2;
			b[ipHCN] += b_CH2_NO_HCN_OH;
			
			c[ipNO][ipOH] += CH2_NO_HCN_OH_1;
			c[ipCH2][ipOH] += CH2_NO_HCN_OH_2;
			b[ipOH] += b_CH2_NO_HCN_OH;

		/*  >>refer   mol  CH2    +  HNO   =>   NO    +     CH3   UMIST  */

			c[ipHNO][ipHNO] -= CH2_HNO_NO_CH3_1;
			c[ipCH2][ipHNO] -= CH2_HNO_NO_CH3_2;
			b[ipHNO] -= b_CH2_HNO_NO_CH3;
			
			c[ipHNO][ipCH2] -= CH2_HNO_NO_CH3_1;
			c[ipCH2][ipCH2] -= CH2_HNO_NO_CH3_2;
			b[ipCH2] -= b_CH2_HNO_NO_CH3;
			
			c[ipHNO][ipNO] += CH2_HNO_NO_CH3_1;
			c[ipCH2][ipNO] += CH2_HNO_NO_CH3_2;
			b[ipNO] += b_CH2_HNO_NO_CH3;
			
			c[ipHNO][ipCH3] += CH2_HNO_NO_CH3_1;
			c[ipCH2][ipCH3] += CH2_HNO_NO_CH3_2;
			b[ipCH3] += b_CH2_HNO_NO_CH3;

		/*  >>refer   mol  NH     +  NH    =>   N2    +     H  +  H      UMIST  */

			c[ipNH][ipNH] -= NH_NH_N2_H_H_1;
			c[ipNH][ipNH] -= NH_NH_N2_H_H_2;
			b[ipNH] -= b_NH_NH_N2_H_H;
			
			c[ipNH][ipN2] += 0.5*NH_NH_N2_H_H_1;
			c[ipNH][ipN2] += 0.5*NH_NH_N2_H_H_2;
			b[ipN2] += 0.5*b_NH_NH_N2_H_H;
			
		/*  >>refer   mol  NH     +  NH    =>   N2    +     H2    UMIST  */

			c[ipNH][ipNH] -= NH_NH_N2_H2_1;
			c[ipNH][ipNH] -= NH_NH_N2_H2_2;
			b[ipNH] -= b_NH_NH_N2_H2;
			
			c[ipNH][ipN2] += 0.5*NH_NH_N2_H2_1;
			c[ipNH][ipN2] += 0.5*NH_NH_N2_H2_2;
			b[ipN2] += 0.5*b_NH_NH_N2_H2;
			
		/*  >>refer   mol  NH     +  O     =>   OH    +     N     UMIST  */

			c[ipATO][ipNH] -= NH_O_OH_N_1;
			c[ipNH][ipNH] -= NH_O_OH_N_2;
			b[ipNH] -= b_NH_O_OH_N;
			
			c[ipATO][ipATO] -= NH_O_OH_N_1;
			c[ipNH][ipATO] -= NH_O_OH_N_2;
			b[ipATO] -= b_NH_O_OH_N;
			
			c[ipATO][ipOH] += NH_O_OH_N_1;
			c[ipNH][ipOH] += NH_O_OH_N_2;
			b[ipOH] += b_NH_O_OH_N;

			c[ipATO][ipATN] += NH_O_OH_N_1;
			c[ipNH][ipATN] += NH_O_OH_N_2;
			b[ipATN] += b_NH_O_OH_N;

		/*  >>refer   mol  NH     +  O     =>   NO    +     H     UMIST  */

			c[ipATO][ipNH] -= NH_O_NO_H_1;
			c[ipNH][ipNH] -= NH_O_NO_H_2;
			b[ipNH] -= b_NH_O_NO_H;

			c[ipATO][ipATO] -= NH_O_NO_H_1;
			c[ipNH][ipATO] -= NH_O_NO_H_2;
			b[ipATO] -= b_NH_O_NO_H;
			
			c[ipATO][ipNO] += NH_O_NO_H_1;
			c[ipNH][ipNO] += NH_O_NO_H_2;
			b[ipNO] += b_NH_O_NO_H;

		/*  >>refer   mol  NH     +  OH    =>   HNO   +     H     UMIST  */

			c[ipOH][ipNH] -= NH_OH_HNO_H_1;
			c[ipNH][ipNH] -= NH_OH_HNO_H_2;
			b[ipNH] -= b_NH_OH_HNO_H;
			
			c[ipOH][ipOH] -= NH_OH_HNO_H_1;
			c[ipNH][ipOH] -= NH_OH_HNO_H_2;
			b[ipOH] -= b_NH_OH_HNO_H;
			
			c[ipOH][ipHNO] += NH_OH_HNO_H_1;
			c[ipNH][ipHNO] += NH_OH_HNO_H_2;
			b[ipHNO] += b_NH_OH_HNO_H;

		/*  >>refer   mol  NH     +  OH    =>   NH2   +     O     UMIST  */

			c[ipOH][ipOH] -= NH_OH_NH2_O_1;
			c[ipNH][ipOH] -= NH_OH_NH2_O_2;
			b[ipOH] -= b_NH_OH_NH2_O;
			
			c[ipOH][ipNH] -= NH_OH_NH2_O_1;
			c[ipNH][ipNH] -= NH_OH_NH2_O_2;
			b[ipNH] -= b_NH_OH_NH2_O;
			
			c[ipOH][ipNH2] += NH_OH_NH2_O_1;
			c[ipNH][ipNH2] += NH_OH_NH2_O_2;
			b[ipNH2] += b_NH_OH_NH2_O;
			
			c[ipOH][ipATO] += NH_OH_NH2_O_1;
			c[ipNH][ipATO] += NH_OH_NH2_O_2;
			b[ipATO] += b_NH_OH_NH2_O;
			
		/*  >>refer   mol  NH     +  OH    =>   H2O   +     N     UMIST  */

			c[ipOH][ipOH] -= NH_OH_H2O_N_1;
			c[ipNH][ipOH] -= NH_OH_H2O_N_2;
			b[ipOH] -= b_NH_OH_H2O_N;
			
			c[ipOH][ipNH] -= NH_OH_H2O_N_1;
			c[ipNH][ipNH] -= NH_OH_H2O_N_2;
			b[ipNH] -= b_NH_OH_H2O_N;
			
			c[ipOH][ipH2O] += NH_OH_H2O_N_1;
			c[ipNH][ipH2O] += NH_OH_H2O_N_2;
			b[ipH2O] += b_NH_OH_H2O_N;
			
			c[ipOH][ipATN] += NH_OH_H2O_N_1;
			c[ipNH][ipATN] += NH_OH_H2O_N_2;
			b[ipATN] += b_NH_OH_H2O_N;

		/*  >>refer   mol  NH     +  H2O   =>   OH    +     NH2   UMIST  */

			c[ipH2O][ipNH] -= NH_H2O_OH_NH2_1;
			c[ipNH][ipNH] -= NH_H2O_OH_NH2_2;
			b[ipNH] -= b_NH_H2O_OH_NH2;
			
			c[ipH2O][ipH2O] -= NH_H2O_OH_NH2_1;
			c[ipNH][ipH2O] -= NH_H2O_OH_NH2_2;
			b[ipH2O] -= b_NH_H2O_OH_NH2;
			
			c[ipH2O][ipOH] += NH_H2O_OH_NH2_1;
			c[ipNH][ipOH] += NH_H2O_OH_NH2_2;
			b[ipOH] += b_NH_H2O_OH_NH2;
			
			c[ipH2O][ipNH2] += NH_H2O_OH_NH2_1;
			c[ipNH][ipNH2] += NH_H2O_OH_NH2_2;
			b[ipNH2] += b_NH_H2O_OH_NH2;

		/*  >>refer   mol  NH     +  CN    =>   HCN   +     N     UMIST  */

			c[ipCN][ipNH] -= NH_CN_HCN_N_1;
			c[ipNH][ipNH] -= NH_CN_HCN_N_2;
			b[ipNH] -= b_NH_CN_HCN_N;
			
			c[ipCN][ipCN] -= NH_CN_HCN_N_1;
			c[ipNH][ipCN] -= NH_CN_HCN_N_2;
			b[ipCN] -= b_NH_CN_HCN_N;
			
			c[ipCN][ipHCN] += NH_CN_HCN_N_1;
			c[ipNH][ipHCN] += NH_CN_HCN_N_2;
			b[ipHCN] += b_NH_CN_HCN_N;
			
			c[ipCN][ipATN] += NH_CN_HCN_N_1;
			c[ipNH][ipATN] += NH_CN_HCN_N_2;
			b[ipATN] += b_NH_CN_HCN_N;
			
		/*  >>refer   mol  NH     +  NO    =>   N2O   +     H     UMIST  */

			c[ipNO][ipNH] -= NH_NO_N2O_H_1;
			c[ipNH][ipNH] -= NH_NO_N2O_H_2;
			b[ipNH] -= b_NH_NO_N2O_H;
			
			c[ipNO][ipNO] -= NH_NO_N2O_H_1;
			c[ipNH][ipNO] -= NH_NO_N2O_H_2;
			b[ipNO] -= b_NH_NO_N2O_H;
			
			c[ipNO][ipN2O] += NH_NO_N2O_H_1;
			c[ipNH][ipN2O] += NH_NO_N2O_H_2;
			b[ipN2O] += b_NH_NO_N2O_H;

		/*  >>refer   mol  NH     +  NO    =>   N2    +     O  +  H      UMIST  */

			c[ipNO][ipNH] -= NH_NO_N2_O_H_1;
			c[ipNH][ipNH] -= NH_NO_N2_O_H_2;
			b[ipNH] -= b_NH_NO_N2_O_H;
			
			c[ipNO][ipNO] -= NH_NO_N2_O_H_1;
			c[ipNH][ipNO] -= NH_NO_N2_O_H_2;
			b[ipNO] -= b_NH_NO_N2_O_H;
			
			c[ipNO][ipN2] += NH_NO_N2_O_H_1;
			c[ipNH][ipN2] += NH_NO_N2_O_H_2;
			b[ipN2] += b_NH_NO_N2_O_H;
			
			c[ipNO][ipATO] += NH_NO_N2_O_H_1;
			c[ipNH][ipATO] += NH_NO_N2_O_H_2;
			b[ipATO] += b_NH_NO_N2_O_H;

		/*  >>refer   mol  NH     +  NO    =>   N2    +     OH    UMIST  */

			c[ipNO][ipNH] -= NH_NO_N2_OH_1;
			c[ipNH][ipNH] -= NH_NO_N2_OH_2;
			b[ipNH] -= b_NH_NO_N2_OH;
			
			c[ipNO][ipNO] -= NH_NO_N2_OH_1;
			c[ipNH][ipNO] -= NH_NO_N2_OH_2;
			b[ipNO] -= b_NH_NO_N2_OH;
			
			c[ipNO][ipN2] += NH_NO_N2_OH_1;
			c[ipNH][ipN2] += NH_NO_N2_OH_2;
			b[ipN2] += b_NH_NO_N2_OH;
			
			c[ipNO][ipOH] += NH_NO_N2_OH_1;
			c[ipNH][ipOH] += NH_NO_N2_OH_2;
			b[ipOH] += b_NH_NO_N2_OH;

		/*  >>refer   mol  NH     +  S     =>   HS    +     N     UMIST  */

			c[ipATS][ipNH] -= NH_S_HS_N_1;
			c[ipNH][ipNH] -= NH_S_HS_N_2;
			b[ipNH] -= b_NH_S_HS_N;
			
			c[ipATS][ipATS] -= NH_S_HS_N_1;
			c[ipNH][ipATS] -= NH_S_HS_N_2;
			b[ipATS] -= b_NH_S_HS_N;
			
			c[ipATS][ipHS] += NH_S_HS_N_1;
			c[ipNH][ipHS] += NH_S_HS_N_2;
			b[ipHS] += b_NH_S_HS_N;

			c[ipATS][ipATN] += NH_S_HS_N_1;
			c[ipNH][ipATN] += NH_S_HS_N_2;
			b[ipATN] += b_NH_S_HS_N;

		/*  >>refer   mol  NH     +  S     =>   NS    +     H     UMIST  */

			c[ipATS][ipNH] -= NH_S_NS_H_1;
			c[ipNH][ipNH] -= NH_S_NS_H_2;
			b[ipNH] -= b_NH_S_NS_H;

			c[ipATS][ipATS] -= NH_S_NS_H_1;
			c[ipNH][ipATS] -= NH_S_NS_H_2;
			b[ipATS] -= b_NH_S_NS_H;
			
			c[ipATS][ipNS] += NH_S_NS_H_1;
			c[ipNH][ipNS] += NH_S_NS_H_2;
			b[ipNS] += b_NH_S_NS_H;

		/*  >>refer   mol  NH     +  NO2   =>   HNO   +     NO    UMIST  */

			c[ipNO2][ipNH] -= NH_NO2_HNO_NO_1;
			c[ipNH][ipNH] -= NH_NO2_HNO_NO_2;
			b[ipNH] -= b_NH_NO2_HNO_NO;
			
			c[ipNO2][ipNO2] -= NH_NO2_HNO_NO_1;
			c[ipNH][ipNO2] -= NH_NO2_HNO_NO_2;
			b[ipNO2] -= b_NH_NO2_HNO_NO;
			
			c[ipNO2][ipHNO] += NH_NO2_HNO_NO_1;
			c[ipNH][ipHNO] += NH_NO2_HNO_NO_2;
			b[ipHNO] += b_NH_NO2_HNO_NO;
			
			c[ipNO2][ipNO] += NH_NO2_HNO_NO_1;
			c[ipNH][ipNO] += NH_NO2_HNO_NO_2;
			b[ipNO] += b_NH_NO2_HNO_NO;

		/*  >>refer   mol  NH     +  NO2   =>   N2O   +     OH    UMIST  */

			c[ipNO2][ipNH] -= NH_NO2_N2O_OH_1;
			c[ipNH][ipNH] -= NH_NO2_N2O_OH_2;
			b[ipNH] -= b_NH_NO2_N2O_OH;
			
			c[ipNO2][ipNO2] -= NH_NO2_N2O_OH_1;
			c[ipNH][ipNO2] -= NH_NO2_N2O_OH_2;
			b[ipNO2] -= b_NH_NO2_N2O_OH;
			
			c[ipNO2][ipN2O] += NH_NO2_N2O_OH_1;
			c[ipNH][ipN2O] += NH_NO2_N2O_OH_2;
			b[ipN2O] += b_NH_NO2_N2O_OH;
			
			c[ipNO2][ipOH] += NH_NO2_N2O_OH_1;
			c[ipNH][ipOH] += NH_NO2_N2O_OH_2;
			b[ipOH] += b_NH_NO2_N2O_OH;

		/*  >>refer   mol  CH3    +  NH3   =>   CH4   +     NH2   UMIST  */

			c[ipNH3][ipCH3] -= CH3_NH3_CH4_NH2_1;
			c[ipCH3][ipCH3] -= CH3_NH3_CH4_NH2_2;
			b[ipCH3] -= b_CH3_NH3_CH4_NH2;
			
			c[ipNH3][ipNH3] -= CH3_NH3_CH4_NH2_1;
			c[ipCH3][ipNH3] -= CH3_NH3_CH4_NH2_2;
			b[ipNH3] -= b_CH3_NH3_CH4_NH2;
			
			c[ipNH3][ipCH4] += CH3_NH3_CH4_NH2_1;
			c[ipCH3][ipCH4] += CH3_NH3_CH4_NH2_2;
			b[ipCH4] += b_CH3_NH3_CH4_NH2;
			
			c[ipNH3][ipNH2] += CH3_NH3_CH4_NH2_1;
			c[ipCH3][ipNH2] += CH3_NH3_CH4_NH2_2;
			b[ipNH2] += b_CH3_NH3_CH4_NH2;

		/*  >>refer   mol  CH3    +  CN    =>   HCN   +     CH2   UMIST  */

			c[ipCN][ipCH3] -= CH3_CN_HCN_CH2_1;
			c[ipCH3][ipCH3] -= CH3_CN_HCN_CH2_2;
			b[ipCH3] -= b_CH3_CN_HCN_CH2;
			
			c[ipCN][ipCN] -= CH3_CN_HCN_CH2_1;
			c[ipCH3][ipCN] -= CH3_CN_HCN_CH2_2;
			b[ipCN] -= b_CH3_CN_HCN_CH2;
			
			c[ipCN][ipCH2] += CH3_CN_HCN_CH2_1;
			c[ipCH3][ipCH2] += CH3_CN_HCN_CH2_2;
			b[ipCH2] += b_CH3_CN_HCN_CH2;
			
			c[ipCN][ipHCN] += CH3_CN_HCN_CH2_1;
			c[ipCH3][ipHCN] += CH3_CN_HCN_CH2_2;
			b[ipHCN] += b_CH3_CN_HCN_CH2;

		/*  >>refer   mol  CH3    +  HNO   =>   NO    +     CH4   UMIST  */

			c[ipHNO][ipCH3] -= CH3_HNO_NO_CH4_1;
			c[ipCH3][ipCH3] -= CH3_HNO_NO_CH4_2;
			b[ipCH3] -= b_CH3_HNO_NO_CH4;
			
			c[ipHNO][ipHNO] -= CH3_HNO_NO_CH4_1;
			c[ipCH3][ipHNO] -= CH3_HNO_NO_CH4_2;
			b[ipHNO] -= b_CH3_HNO_NO_CH4;
			
			c[ipHNO][ipNO] += CH3_HNO_NO_CH4_1;
			c[ipCH3][ipNO] += CH3_HNO_NO_CH4_2;
			b[ipNO] += b_CH3_HNO_NO_CH4;
			
			c[ipHNO][ipCH4] += CH3_HNO_NO_CH4_1;
			c[ipCH3][ipCH4] += CH3_HNO_NO_CH4_2;
			b[ipCH4] += b_CH3_HNO_NO_CH4;

		/*  >>refer   mol  O      +  NH2   =>   OH    +     NH    UMIST  */

			c[ipNH2][ipNH2] -= O_NH2_OH_NH_1;
			c[ipATO][ipNH2] -= O_NH2_OH_NH_2;
			b[ipNH2] -= b_O_NH2_OH_NH;

			c[ipNH2][ipATO] -= O_NH2_OH_NH_1;
			c[ipATO][ipATO] -= O_NH2_OH_NH_2;
			b[ipATO] -= b_O_NH2_OH_NH;
			
			c[ipNH2][ipOH] += O_NH2_OH_NH_1;
			c[ipATO][ipOH] += O_NH2_OH_NH_2;
			b[ipOH] += b_O_NH2_OH_NH;
			
			c[ipNH2][ipNH] += O_NH2_OH_NH_1;
			c[ipATO][ipNH] += O_NH2_OH_NH_2;
			b[ipNH] += b_O_NH2_OH_NH;

		/*  >>refer   mol  O      +  NH2   =>   NO    +     H2    UMIST  */
			
			c[ipNH2][ipNH2] -= O_NH2_NO_H2_1;
			c[ipATO][ipNH2] -= O_NH2_NO_H2_2;
			b[ipNH2] -= b_O_NH2_NO_H2;
			
			c[ipNH2][ipATO] -= O_NH2_NO_H2_1;
			c[ipATO][ipATO] -= O_NH2_NO_H2_2;
			b[ipATO] -= b_O_NH2_NO_H2;
			
			c[ipNH2][ipNO] += O_NH2_NO_H2_1;
			c[ipATO][ipNO] += O_NH2_NO_H2_2;
			b[ipNO] += b_O_NH2_NO_H2;
			
		/*  >>refer   mol  O      +  NH3   =>   OH    +     NH2   UMIST  */

			c[ipNH3][ipNH3] -= O_NH3_OH_NH2_1;
			c[ipATO][ipNH3] -= O_NH3_OH_NH2_2;
			b[ipNH3] -= b_O_NH3_OH_NH2;

			c[ipNH3][ipATO] -= O_NH3_OH_NH2_1;
			c[ipATO][ipATO] -= O_NH3_OH_NH2_2;
			b[ipATO] -= b_O_NH3_OH_NH2;
			
			c[ipNH3][ipOH] += O_NH3_OH_NH2_1;
			c[ipATO][ipOH] += O_NH3_OH_NH2_2;
			b[ipOH] += b_O_NH3_OH_NH2;
			
			c[ipNH3][ipNH2] += O_NH3_OH_NH2_1;
			c[ipATO][ipNH2] += O_NH3_OH_NH2_2;
			b[ipNH2] += b_O_NH3_OH_NH2;

		/*  >>refer   mol  O      +  CN    =>   NO    +     C     UMIST  */
			
			c[ipCN][ipCN] -= O_CN_NO_C_1;
			c[ipATO][ipCN] -= O_CN_NO_C_2;
			b[ipCN] -= b_O_CN_NO_C;
						
			c[ipCN][ipATO] -= O_CN_NO_C_1;
			c[ipATO][ipATO] -= O_CN_NO_C_2;
			b[ipATO] -= b_O_CN_NO_C;
			
			c[ipCN][ipNO] += O_CN_NO_C_1;
			c[ipATO][ipNO] += O_CN_NO_C_2;
			b[ipNO] += b_O_CN_NO_C;

			c[ipCN][ipATC] += O_CN_NO_C_1;
			c[ipATO][ipATC] += O_CN_NO_C_2;
			b[ipATC] += b_O_CN_NO_C;

		/*  >>refer   mol  O      +  CN    =>   CO    +     N     UMIST  */

			c[ipCN][ipCN] -= O_CN_CO_N_1;
			c[ipATO][ipCN] -= O_CN_CO_N_2;
			b[ipCN] -= b_O_CN_CO_N;
			
			c[ipCN][ipATO] -= O_CN_CO_N_1;
			c[ipATO][ipATO] -= O_CN_CO_N_2;
			b[ipATO] -= b_O_CN_CO_N;
			
			c[ipCN][ipCO] += O_CN_CO_N_1;
			c[ipATO][ipCO] += O_CN_CO_N_2;
			b[ipCO] += b_O_CN_CO_N;

			c[ipCN][ipATN] += O_CN_CO_N_1;
			c[ipATO][ipATN] += O_CN_CO_N_2;
			b[ipATN] += b_O_CN_CO_N;

		/*  >>refer   mol  O      +  HCN   =>   CN    +     OH    UMIST  */

			c[ipHCN][ipHCN] -= O_HCN_CN_OH_1;
			c[ipATO][ipHCN] -= O_HCN_CN_OH_2;
			b[ipHCN] -= b_O_HCN_CN_OH;
			
			c[ipHCN][ipATO] -= O_HCN_CN_OH_1;
			c[ipATO][ipATO] -= O_HCN_CN_OH_2;
			b[ipATO] -= b_O_HCN_CN_OH;

			c[ipHCN][ipCN] += O_HCN_CN_OH_1;
			c[ipATO][ipCN] += O_HCN_CN_OH_2;
			b[ipCN] += b_O_HCN_CN_OH;
			
			c[ipHCN][ipOH] += O_HCN_CN_OH_1;
			c[ipATO][ipOH] += O_HCN_CN_OH_2;
			b[ipOH] += b_O_HCN_CN_OH;

		/*  >>refer   mol  O      +  HCN   =>   CO    +     NH    UMIST  */

			c[ipHCN][ipHCN] -= O_HCN_CO_NH_1;
			c[ipATO][ipHCN] -= O_HCN_CO_NH_2;
			b[ipHCN] -= b_O_HCN_CO_NH;
			
			c[ipHCN][ipATO] -= O_HCN_CO_NH_1;
			c[ipATO][ipATO] -= O_HCN_CO_NH_2;
			b[ipATO] -= b_O_HCN_CO_NH;

			c[ipHCN][ipCO] += O_HCN_CO_NH_1;
			c[ipATO][ipCO] += O_HCN_CO_NH_2;
			b[ipCO] += b_O_HCN_CO_NH;
			
			c[ipHCN][ipNH] += O_HCN_CO_NH_1;
			c[ipATO][ipNH] += O_HCN_CO_NH_2;
			b[ipNH] += b_O_HCN_CO_NH;

		/*  >>refer   mol  O      +  HCN   =>   OCN   +     H     UMIST  */
			
			c[ipHCN][ipHCN] -= O_HCN_OCN_H_1;
			c[ipATO][ipHCN] -= O_HCN_OCN_H_2;
			b[ipHCN] -= b_O_HCN_OCN_H;
				
			c[ipHCN][ipATO] -= O_HCN_OCN_H_1;
			c[ipATO][ipATO] -= O_HCN_OCN_H_2;
			b[ipATO] -= b_O_HCN_OCN_H;
			
			c[ipHCN][ipOCN] += O_HCN_OCN_H_1;
			c[ipATO][ipOCN] += O_HCN_OCN_H_2;
			b[ipOCN] += b_O_HCN_OCN_H;

		/*  >>refer   mol  O      +  N2    =>   NO    +     N     UMIST  */

			c[ipN2][ipN2] -= O_N2_NO_N_1;
			c[ipATO][ipN2] -= O_N2_NO_N_2;
			b[ipN2] -= b_O_N2_NO_N;
			
			c[ipN2][ipATO] -= O_N2_NO_N_1;
			c[ipATO][ipATO] -= O_N2_NO_N_2;
			b[ipATO] -= b_O_N2_NO_N;
			
			c[ipN2][ipNO] += O_N2_NO_N_1;
			c[ipATO][ipNO] += O_N2_NO_N_2;
			b[ipNO] += b_O_N2_NO_N;

			c[ipN2][ipATN] += O_N2_NO_N_1;
			c[ipATO][ipATN] += O_N2_NO_N_2;
			b[ipATN] += b_O_N2_NO_N;

		/*  >>refer   mol  O      +  NO    =>   O2    +     N     UMIST  */

			c[ipNO][ipNO] -= O_NO_O2_N_1;
			c[ipATO][ipNO] -= O_NO_O2_N_2;
			b[ipNO] -= b_O_NO_O2_N;
			
			c[ipNO][ipATO] -= O_NO_O2_N_1;
			c[ipATO][ipATO] -= O_NO_O2_N_2;
			b[ipATO] -= b_O_NO_O2_N;
			
			c[ipNO][ipO2] += O_NO_O2_N_1;
			c[ipATO][ipO2] += O_NO_O2_N_2;
			b[ipO2] += b_O_NO_O2_N;

			c[ipNO][ipATN] += O_NO_O2_N_1;
			c[ipATO][ipATN] += O_NO_O2_N_2;
			b[ipATN] += b_O_NO_O2_N;

		/*  >>refer   mol  O      +  HNO   =>   NO    +     OH    UMIST  */
			
			c[ipHNO][ipHNO] -= O_HNO_NO_OH_1;
			c[ipATO][ipHNO] -= O_HNO_NO_OH_2;
			b[ipHNO] -= b_O_HNO_NO_OH;
						
			c[ipHNO][ipATO] -= O_HNO_NO_OH_1;
			c[ipATO][ipATO] -= O_HNO_NO_OH_2;
			b[ipATO] -= b_O_HNO_NO_OH;

			c[ipHNO][ipNO] += O_HNO_NO_OH_1;
			c[ipATO][ipNO] += O_HNO_NO_OH_2;
			b[ipNO] += b_O_HNO_NO_OH;
			
			c[ipHNO][ipOH] += O_HNO_NO_OH_1;
			c[ipATO][ipOH] += O_HNO_NO_OH_2;
			b[ipOH] += b_O_HNO_NO_OH;

		/*  >>refer   mol  O      +  HNO   =>   O2    +     NH    UMIST  */
			
			c[ipHNO][ipHNO] -= O_HNO_O2_NH_1;
			c[ipATO][ipHNO] -= O_HNO_O2_NH_2;
			b[ipHNO] -= b_O_HNO_O2_NH;
			
			c[ipHNO][ipATO] -= O_HNO_O2_NH_1;
			c[ipATO][ipATO] -= O_HNO_O2_NH_2;
			b[ipATO] -= b_O_HNO_O2_NH;
			
			c[ipHNO][ipO2] += O_HNO_O2_NH_1;
			c[ipATO][ipO2] += O_HNO_O2_NH_2;
			b[ipO2] += b_O_HNO_O2_NH;
			
			c[ipHNO][ipNH] += O_HNO_O2_NH_1;
			c[ipATO][ipNH] += O_HNO_O2_NH_2;
			b[ipNH] += b_O_HNO_O2_NH;

		/*  >>refer   mol  O      +  HNO   =>   NO2   +     H     UMIST  */
			
			c[ipHNO][ipHNO] -= O_HNO_NO2_H_1;
			c[ipATO][ipHNO] -= O_HNO_NO2_H_2;
			b[ipHNO] -= b_O_HNO_NO2_H;
						
			c[ipHNO][ipATO] -= O_HNO_NO2_H_1;
			c[ipATO][ipATO] -= O_HNO_NO2_H_2;
			b[ipATO] -= b_O_HNO_NO2_H;

			c[ipHNO][ipNO2] += O_HNO_NO2_H_1;
			c[ipATO][ipNO2] += O_HNO_NO2_H_2;
			b[ipNO2] += b_O_HNO_NO2_H;

		/*  >>refer   mol  O      +  HS    =>   S     +     OH    UMIST  */

			c[ipHS][ipHS] -= O_HS_S_OH_1;
			c[ipATO][ipHS] -= O_HS_S_OH_2;
			b[ipHS] -= b_O_HS_S_OH;
			
			c[ipHS][ipATO] -= O_HS_S_OH_1;
			c[ipATO][ipATO] -= O_HS_S_OH_2;
			b[ipATO] -= b_O_HS_S_OH;
			
			c[ipHS][ipOH] += O_HS_S_OH_1;
			c[ipATO][ipOH] += O_HS_S_OH_2;
			b[ipOH] += b_O_HS_S_OH;

			c[ipHS][ipATS] += O_HS_S_OH_1;
			c[ipATO][ipATS] += O_HS_S_OH_2;
			b[ipATS] += b_O_HS_S_OH;

		/*  >>refer   mol  O      +  HS    =>   SO    +     H     UMIST  */
			
			c[ipHS][ipHS] -= O_HS_SO_H_1;
			c[ipATO][ipHS] -= O_HS_SO_H_2;
			b[ipHS] -= b_O_HS_SO_H;
			
			c[ipHS][ipATO] -= O_HS_SO_H_1;
			c[ipATO][ipATO] -= O_HS_SO_H_2;
			b[ipATO] -= b_O_HS_SO_H;
			
			c[ipHS][ipSO] += O_HS_SO_H_1;
			c[ipATO][ipSO] += O_HS_SO_H_2;
			b[ipSO] += b_O_HS_SO_H;

		/*  >>refer   mol  O      +  OCN   =>   NO    +     CO    UMIST  */
			
			c[ipOCN][ipOCN] -= O_OCN_NO_CO_1;
			c[ipATO][ipOCN] -= O_OCN_NO_CO_2;
			b[ipOCN] -= b_O_OCN_NO_CO;
					
			c[ipOCN][ipATO] -= O_OCN_NO_CO_1;
			c[ipATO][ipATO] -= O_OCN_NO_CO_2;
			b[ipATO] -= b_O_OCN_NO_CO;
			
			c[ipOCN][ipNO] += O_OCN_NO_CO_1;
			c[ipATO][ipNO] += O_OCN_NO_CO_2;
			b[ipNO] += b_O_OCN_NO_CO;
			
			c[ipOCN][ipCO] += O_OCN_NO_CO_1;
			c[ipATO][ipCO] += O_OCN_NO_CO_2;
			b[ipCO] += b_O_OCN_NO_CO;

		/*  >>refer   mol  O      +  OCN   =>   O2    +     CN    UMIST  */
			
			c[ipOCN][ipOCN] -= O_OCN_O2_CN_1;
			c[ipATO][ipOCN] -= O_OCN_O2_CN_2;
			b[ipOCN] -= b_O_OCN_O2_CN;
			
			c[ipOCN][ipATO] -= O_OCN_O2_CN_1;
			c[ipATO][ipATO] -= O_OCN_O2_CN_2;
			b[ipATO] -= b_O_OCN_O2_CN;
			
			c[ipOCN][ipO2] += O_OCN_O2_CN_1;
			c[ipATO][ipO2] += O_OCN_O2_CN_2;
			b[ipO2] += b_O_OCN_O2_CN;
			
			c[ipOCN][ipCN] += O_OCN_O2_CN_1;
			c[ipATO][ipCN] += O_OCN_O2_CN_2;
			b[ipCN] += b_O_OCN_O2_CN;

		/*  >>refer   mol  O      +  SIN   =>   NO    +     SI    UMIST  */

			c[ipSIN][ipSIN] -= O_SIN_NO_SI_1;
			c[ipATO][ipSIN] -= O_SIN_NO_SI_2;
			b[ipSIN] -= b_O_SIN_NO_SI;
			
			c[ipSIN][ipATO] -= O_SIN_NO_SI_1;
			c[ipATO][ipATO] -= O_SIN_NO_SI_2;
			b[ipATO] -= b_O_SIN_NO_SI;
			
			c[ipSIN][ipNO] += O_SIN_NO_SI_1;
			c[ipATO][ipNO] += O_SIN_NO_SI_2;
			b[ipNO] += b_O_SIN_NO_SI;

			c[ipSIN][ipATSI] += O_SIN_NO_SI_1;
			c[ipATO][ipATSI] += O_SIN_NO_SI_2;
			b[ipATSI] += b_O_SIN_NO_SI;
			
		/*  >>refer   mol  O      +  SIN   =>   SIO   +     N     UMIST  */

			c[ipSIN][ipSIN] -= O_SIN_SIO_N_1;
			c[ipATO][ipSIN] -= O_SIN_SIO_N_2;
			b[ipSIN] -= b_O_SIN_SIO_N;
			
			c[ipSIN][ipATO] -= O_SIN_SIO_N_1;
			c[ipATO][ipATO] -= O_SIN_SIO_N_2;
			b[ipATO] -= b_O_SIN_SIO_N;
			
			c[ipSIN][ipSIO] += O_SIN_SIO_N_1;
			c[ipATO][ipSIO] += O_SIN_SIO_N_2;
			b[ipSIO] += b_O_SIN_SIO_N;

			c[ipSIN][ipATN] += O_SIN_SIO_N_1;
			c[ipATO][ipATN] += O_SIN_SIO_N_2;
			b[ipATN] += b_O_SIN_SIO_N;

		/*  >>refer   mol  O      +  N2O   =>   NO    +     NO    UMIST  */

			c[ipN2O][ipN2O] -= O_N2O_NO_NO_1;
			c[ipATO][ipN2O] -= O_N2O_NO_NO_2;
			b[ipN2O] -= b_O_N2O_NO_NO;

			c[ipN2O][ipATO] -= O_N2O_NO_NO_1;
			c[ipATO][ipATO] -= O_N2O_NO_NO_2;
			b[ipATO] -= b_O_N2O_NO_NO;
			
			c[ipN2O][ipNO] += O_N2O_NO_NO_1;
			c[ipATO][ipNO] += O_N2O_NO_NO_2;
			b[ipNO] += b_O_N2O_NO_NO;
			
			c[ipN2O][ipNO] += O_N2O_NO_NO_1;
			c[ipATO][ipNO] += O_N2O_NO_NO_2;
			b[ipNO] += b_O_N2O_NO_NO;

		/*  >>refer   mol  O      +  N2O   =>   O2    +     N2    UMIST  */

			c[ipN2O][ipN2O] -= O_N2O_O2_N2_1;
			c[ipATO][ipN2O] -= O_N2O_O2_N2_2;
			b[ipN2O] -= b_O_N2O_O2_N2;

			c[ipN2O][ipATO] -= O_N2O_O2_N2_1;
			c[ipATO][ipATO] -= O_N2O_O2_N2_2;
			b[ipATO] -= b_O_N2O_O2_N2;
			
			c[ipN2O][ipO2] += O_N2O_O2_N2_1;
			c[ipATO][ipO2] += O_N2O_O2_N2_2;
			b[ipO2] += b_O_N2O_O2_N2;
			
			c[ipN2O][ipN2] += O_N2O_O2_N2_1;
			c[ipATO][ipN2] += O_N2O_O2_N2_2;
			b[ipN2] += b_O_N2O_O2_N2;

		/*  >>refer   mol  O      +  CS    =>   S     +     CO    UMIST  */

			c[ipCS][ipCS] -= O_CS_S_CO_1;
			c[ipATO][ipCS] -= O_CS_S_CO_2;
			b[ipCS] -= b_O_CS_S_CO;
			
			c[ipCS][ipATO] -= O_CS_S_CO_1;
			c[ipATO][ipATO] -= O_CS_S_CO_2;
			b[ipATO] -= b_O_CS_S_CO;
			
			c[ipCS][ipCO] += O_CS_S_CO_1;
			c[ipATO][ipCO] += O_CS_S_CO_2;
			b[ipCO] += b_O_CS_S_CO;

			c[ipCS][ipATS] += O_CS_S_CO_1;
			c[ipATO][ipATS] += O_CS_S_CO_2;
			b[ipATS] += b_O_CS_S_CO;

		/*  >>refer   mol  O      +  CS    =>   SO    +     C     UMIST  */

			c[ipCS][ipCS] -= O_CS_SO_C_1;
			c[ipATO][ipCS] -= O_CS_SO_C_2;
			b[ipCS] -= b_O_CS_SO_C;
			
			c[ipCS][ipATO] -= O_CS_SO_C_1;
			c[ipATO][ipATO] -= O_CS_SO_C_2;
			b[ipATO] -= b_O_CS_SO_C;
			
			c[ipCS][ipSO] += O_CS_SO_C_1;
			c[ipATO][ipSO] += O_CS_SO_C_2;
			b[ipSO] += b_O_CS_SO_C;

			c[ipCS][ipATC] += O_CS_SO_C_1;
			c[ipATO][ipATC] += O_CS_SO_C_2;
			b[ipATC] += b_O_CS_SO_C;

		/*  >>refer   mol  O      +  NS    =>   S     +     NO    UMIST  */

			c[ipNS][ipNS] -= O_NS_S_NO_1;
			c[ipATO][ipNS] -= O_NS_S_NO_2;
			b[ipNS] -= b_O_NS_S_NO;
			
			c[ipNS][ipATO] -= O_NS_S_NO_1;
			c[ipATO][ipATO] -= O_NS_S_NO_2;
			b[ipATO] -= b_O_NS_S_NO;
			
			c[ipNS][ipNO] += O_NS_S_NO_1;
			c[ipATO][ipNO] += O_NS_S_NO_2;
			b[ipNO] += b_O_NS_S_NO;

			c[ipNS][ipATS] += O_NS_S_NO_1;
			c[ipATO][ipATS] += O_NS_S_NO_2;
			b[ipATS] += b_O_NS_S_NO;

		/*  >>refer   mol  O      +  NS    =>   SO    +     N     UMIST  */

			c[ipNS][ipNS] -= O_NS_SO_N_1;
			c[ipATO][ipNS] -= O_NS_SO_N_2;
			b[ipNS] -= b_O_NS_SO_N;
			
			c[ipNS][ipATO] -= O_NS_SO_N_1;
			c[ipATO][ipATO] -= O_NS_SO_N_2;
			b[ipATO] -= b_O_NS_SO_N;
			
			c[ipNS][ipSO] += O_NS_SO_N_1;
			c[ipATO][ipSO] += O_NS_SO_N_2;
			b[ipSO] += b_O_NS_SO_N;

			c[ipNS][ipATN] += O_NS_SO_N_1;
			c[ipATO][ipATN] += O_NS_SO_N_2;
			b[ipATN] += b_O_NS_SO_N;

		/*  >>refer   mol  O      +  SO    =>   S     +     O2    UMIST  */

			c[ipSO][ipSO] -= O_SO_S_O2_1;
			c[ipATO][ipSO] -= O_SO_S_O2_2;
			b[ipSO] -= b_O_SO_S_O2;
			
			c[ipSO][ipATO] -= O_SO_S_O2_1;
			c[ipATO][ipATO] -= O_SO_S_O2_2;
			b[ipATO] -= b_O_SO_S_O2;
			
			c[ipSO][ipO2] += O_SO_S_O2_1;
			c[ipATO][ipO2] += O_SO_S_O2_2;
			b[ipO2] += b_O_SO_S_O2;

			c[ipSO][ipATS] += O_SO_S_O2_1;
			c[ipATO][ipATS] += O_SO_S_O2_2;
			b[ipATS] += b_O_SO_S_O2;

		/*  >>refer   mol  O      +  OCS   =>   SO    +     CO    UMIST  */

			c[ipOCS][ipOCS] -= O_OCS_SO_CO_1;
			c[ipATO][ipOCS] -= O_OCS_SO_CO_2;
			b[ipOCS] -= b_O_OCS_SO_CO;
			
			c[ipOCS][ipATO] -= O_OCS_SO_CO_1;
			c[ipATO][ipATO] -= O_OCS_SO_CO_2;
			b[ipATO] -= b_O_OCS_SO_CO;

			c[ipOCS][ipSO] += O_OCS_SO_CO_1;
			c[ipATO][ipSO] += O_OCS_SO_CO_2;
			b[ipSO] += b_O_OCS_SO_CO;
			
			c[ipOCS][ipCO] += O_OCS_SO_CO_1;
			c[ipATO][ipCO] += O_OCS_SO_CO_2;
			b[ipCO] += b_O_OCS_SO_CO;

		/*  >>refer   mol  O      +  S2    =>   SO    +     S     UMIST  */

			c[ipS2][ipS2] -= O_S2_SO_S_1;
			c[ipATO][ipS2] -= O_S2_SO_S_2;
			b[ipS2] -= b_O_S2_SO_S;
			
			c[ipS2][ipATO] -= O_S2_SO_S_1;
			c[ipATO][ipATO] -= O_S2_SO_S_2;
			b[ipATO] -= b_O_S2_SO_S;
			
			c[ipS2][ipSO] += O_S2_SO_S_1;
			c[ipATO][ipSO] += O_S2_SO_S_2;
			b[ipSO] += b_O_S2_SO_S;

			c[ipS2][ipATS] += O_S2_SO_S_1;
			c[ipATO][ipATS] += O_S2_SO_S_2;
			b[ipATS] += b_O_S2_SO_S;

		/*  >>refer   mol  NH2    +  OH    =>   NH3   +     O     UMIST  */
			
			c[ipOH][ipNH2] -= NH2_OH_NH3_O_1;
			c[ipNH2][ipNH2] -= NH2_OH_NH3_O_2;
			b[ipNH2] -= b_NH2_OH_NH3_O;

			c[ipOH][ipOH] -= NH2_OH_NH3_O_1;
			c[ipNH2][ipOH] -= NH2_OH_NH3_O_2;
			b[ipOH] -= b_NH2_OH_NH3_O;
			
			c[ipOH][ipNH3] += NH2_OH_NH3_O_1;
			c[ipNH2][ipNH3] += NH2_OH_NH3_O_2;
			b[ipNH3] += b_NH2_OH_NH3_O;
			
			c[ipOH][ipATO] += NH2_OH_NH3_O_1;
			c[ipNH2][ipATO] += NH2_OH_NH3_O_2;
			b[ipATO] += b_NH2_OH_NH3_O;

		/*  >>refer   mol  NH2    +  OH    =>   H2O   +     NH    UMIST  */
			
			c[ipOH][ipNH2] -= NH2_OH_H2O_NH_1;
			c[ipNH2][ipNH2] -= NH2_OH_H2O_NH_2;
			b[ipNH2] -= b_NH2_OH_H2O_NH;
			
			c[ipOH][ipOH] -= NH2_OH_H2O_NH_1;
			c[ipNH2][ipOH] -= NH2_OH_H2O_NH_2;
			b[ipOH] -= b_NH2_OH_H2O_NH;
			
			c[ipOH][ipH2O] += NH2_OH_H2O_NH_1;
			c[ipNH2][ipH2O] += NH2_OH_H2O_NH_2;
			b[ipH2O] += b_NH2_OH_H2O_NH;
			
			c[ipOH][ipNH] += NH2_OH_H2O_NH_1;
			c[ipNH2][ipNH] += NH2_OH_H2O_NH_2;
			b[ipNH] += b_NH2_OH_H2O_NH;

		/*  >>refer   mol  NH2    +  NO    =>   N2    +     OH +  H      UMIST  */

			c[ipNO][ipNH2] -= NH2_NO_N2_OH_H_1;
			c[ipNH2][ipNH2] -= NH2_NO_N2_OH_H_2;
			b[ipNH2] -= b_NH2_NO_N2_OH_H;
			
			c[ipNO][ipNO] -= NH2_NO_N2_OH_H_1;
			c[ipNH2][ipNO] -= NH2_NO_N2_OH_H_2;
			b[ipNO] -= b_NH2_NO_N2_OH_H;
			
			c[ipNO][ipN2] += NH2_NO_N2_OH_H_1;
			c[ipNH2][ipN2] += NH2_NO_N2_OH_H_2;
			b[ipN2] += b_NH2_NO_N2_OH_H;
			
			c[ipNO][ipOH] += NH2_NO_N2_OH_H_1;
			c[ipNH2][ipOH] += NH2_NO_N2_OH_H_2;
			b[ipOH] += b_NH2_NO_N2_OH_H;

		/*  >>refer   mol  NH2    +  NO    =>   N2    +     H2O   UMIST  */

			c[ipNO][ipNH2] -= NH2_NO_N2_H2O_1;
			c[ipNH2][ipNH2] -= NH2_NO_N2_H2O_2;
			b[ipNH2] -= b_NH2_NO_N2_H2O;
			
			c[ipNO][ipNO] -= NH2_NO_N2_H2O_1;
			c[ipNH2][ipNO] -= NH2_NO_N2_H2O_2;
			b[ipNO] -= b_NH2_NO_N2_H2O;
			
			c[ipNO][ipN2] += NH2_NO_N2_H2O_1;
			c[ipNH2][ipN2] += NH2_NO_N2_H2O_2;
			b[ipN2] += b_NH2_NO_N2_H2O;
			
			c[ipNO][ipH2O] += NH2_NO_N2_H2O_1;
			c[ipNH2][ipH2O] += NH2_NO_N2_H2O_2;
			b[ipH2O] += b_NH2_NO_N2_H2O;

		/*  >>refer   mol  CH4    +  CN    =>   HCN   +     CH3   UMIST  */

			c[ipCN][ipCH4] -= CH4_CN_HCN_CH3_1;
			c[ipCH4][ipCH4] -= CH4_CN_HCN_CH3_2;
			b[ipCH4] -= b_CH4_CN_HCN_CH3;
			
			c[ipCN][ipCN] -= CH4_CN_HCN_CH3_1;
			c[ipCH4][ipCN] -= CH4_CN_HCN_CH3_2;
			b[ipCN] -= b_CH4_CN_HCN_CH3;
			
			c[ipCN][ipHCN] += CH4_CN_HCN_CH3_1;
			c[ipCH4][ipHCN] += CH4_CN_HCN_CH3_2;
			b[ipHCN] += b_CH4_CN_HCN_CH3;
			
			c[ipCN][ipCH3] += CH4_CN_HCN_CH3_1;
			c[ipCH4][ipCH3] += CH4_CN_HCN_CH3_2;
			b[ipCH3] += b_CH4_CN_HCN_CH3;

		/*  >>refer   mol  OH     +  NH3   =>   H2O   +     NH2   UMIST  */

			c[ipNH3][ipOH] -= OH_NH3_H2O_NH2_1;
			c[ipOH][ipOH] -= OH_NH3_H2O_NH2_2;
			b[ipOH] -= b_OH_NH3_H2O_NH2;
			
			c[ipNH3][ipNH3] -= OH_NH3_H2O_NH2_1;
			c[ipOH][ipNH3] -= OH_NH3_H2O_NH2_2;
			b[ipNH3] -= b_OH_NH3_H2O_NH2;
			
			c[ipNH3][ipH2O] += OH_NH3_H2O_NH2_1;
			c[ipOH][ipH2O] += OH_NH3_H2O_NH2_2;
			b[ipH2O] += b_OH_NH3_H2O_NH2;
			
			c[ipNH3][ipNH2] += OH_NH3_H2O_NH2_1;
			c[ipOH][ipNH2] += OH_NH3_H2O_NH2_2;
			b[ipNH2] += b_OH_NH3_H2O_NH2;

		/*  >>refer   mol  OH     +  CN    =>   HCN   +     O     UMIST  */

			c[ipCN][ipOH] -= OH_CN_HCN_O_1;
			c[ipOH][ipOH] -= OH_CN_HCN_O_2;
			b[ipOH] -= b_OH_CN_HCN_O;
			
			c[ipCN][ipCN] -= OH_CN_HCN_O_1;
			c[ipOH][ipCN] -= OH_CN_HCN_O_2;
			b[ipCN] -= b_OH_CN_HCN_O;
			
			c[ipCN][ipHCN] += OH_CN_HCN_O_1;
			c[ipOH][ipHCN] += OH_CN_HCN_O_2;
			b[ipHCN] += b_OH_CN_HCN_O;
			
			c[ipCN][ipATO] += OH_CN_HCN_O_1;
			c[ipOH][ipATO] += OH_CN_HCN_O_2;
			b[ipATO] += b_OH_CN_HCN_O;
			
		/*  >>refer   mol  OH     +  CN    =>   OCN   +     H     UMIST  */

			c[ipCN][ipOH] -= OH_CN_OCN_H_1;
			c[ipOH][ipOH] -= OH_CN_OCN_H_2;
			b[ipOH] -= b_OH_CN_OCN_H;
			
			c[ipCN][ipCN] -= OH_CN_OCN_H_1;
			c[ipOH][ipCN] -= OH_CN_OCN_H_2;
			b[ipCN] -= b_OH_CN_OCN_H;
			
			c[ipCN][ipOCN] += OH_CN_OCN_H_1;
			c[ipOH][ipOCN] += OH_CN_OCN_H_2;
			b[ipOCN] += b_OH_CN_OCN_H;

		/*  >>refer   mol  OH     +  HCN   =>   CN    +     H2O   UMIST  */

			c[ipHCN][ipOH] -= OH_HCN_CN_H2O_1;
			c[ipOH][ipOH] -= OH_HCN_CN_H2O_2;
			b[ipOH] -= b_OH_HCN_CN_H2O;
			
			c[ipHCN][ipHCN] -= OH_HCN_CN_H2O_1;
			c[ipOH][ipHCN] -= OH_HCN_CN_H2O_2;
			b[ipHCN] -= b_OH_HCN_CN_H2O;
			
			c[ipHCN][ipCN] += OH_HCN_CN_H2O_1;
			c[ipOH][ipCN] += OH_HCN_CN_H2O_2;
			b[ipCN] += b_OH_HCN_CN_H2O;
			
			c[ipHCN][ipH2O] += OH_HCN_CN_H2O_1;
			c[ipOH][ipH2O] += OH_HCN_CN_H2O_2;
			b[ipH2O] += b_OH_HCN_CN_H2O;

		/*  >>refer   mol  OH     +  HCN   =>   CO    +     NH2   UMIST  */

			c[ipHCN][ipOH] -= OH_HCN_CO_NH2_1;
			c[ipOH][ipOH] -= OH_HCN_CO_NH2_2;
			b[ipOH] -= b_OH_HCN_CO_NH2;
			
			c[ipHCN][ipHCN] -= OH_HCN_CO_NH2_1;
			c[ipOH][ipHCN] -= OH_HCN_CO_NH2_2;
			b[ipHCN] -= b_OH_HCN_CO_NH2;
			
			c[ipHCN][ipCO] += OH_HCN_CO_NH2_1;
			c[ipOH][ipCO] += OH_HCN_CO_NH2_2;
			b[ipCO] += b_OH_HCN_CO_NH2;
			
			c[ipHCN][ipNH2] += OH_HCN_CO_NH2_1;
			c[ipOH][ipNH2] += OH_HCN_CO_NH2_2;
			b[ipNH2] += b_OH_HCN_CO_NH2;

		/*  >>refer   mol  OH     +  NO    =>   NO2   +     H     UMIST  */

			c[ipNO][ipOH] -= OH_NO_NO2_H_1;
			c[ipOH][ipOH] -= OH_NO_NO2_H_2;
			b[ipOH] -= b_OH_NO_NO2_H;
			
			c[ipNO][ipNO] -= OH_NO_NO2_H_1;
			c[ipOH][ipNO] -= OH_NO_NO2_H_2;
			b[ipNO] -= b_OH_NO_NO2_H;
			
			c[ipNO][ipNO2] += OH_NO_NO2_H_1;
			c[ipOH][ipNO2] += OH_NO_NO2_H_2;
			b[ipNO2] += b_OH_NO_NO2_H;
			
		/*  >>refer   mol  OH     +  S     =>   HS    +     O     UMIST  */

			c[ipATS][ipOH] -= OH_S_HS_O_1;
			c[ipOH][ipOH] -= OH_S_HS_O_2;
			b[ipOH] -= b_OH_S_HS_O;
			
			c[ipATS][ipATS] -= OH_S_HS_O_1;
			c[ipOH][ipATS] -= OH_S_HS_O_2;
			b[ipATS] -= b_OH_S_HS_O;
			
			c[ipATS][ipHS] += OH_S_HS_O_1;
			c[ipOH][ipHS] += OH_S_HS_O_2;
			b[ipHS] += b_OH_S_HS_O;

			c[ipATS][ipATO] += OH_S_HS_O_1;
			c[ipOH][ipATO] += OH_S_HS_O_2;
			b[ipATO] += b_OH_S_HS_O;

		/*  >>refer   mol  OH     +  S     =>   SO    +     H     UMIST  */

			c[ipATS][ipOH] -= OH_S_SO_H_1;
			c[ipOH][ipOH] -= OH_S_SO_H_2;
			b[ipOH] -= b_OH_S_SO_H;

			c[ipATS][ipATS] -= OH_S_SO_H_1;
			c[ipOH][ipATS] -= OH_S_SO_H_2;
			b[ipATS] -= b_OH_S_SO_H;
			
			c[ipATS][ipSO] += OH_S_SO_H_1;
			c[ipOH][ipSO] += OH_S_SO_H_2;
			b[ipSO] += b_OH_S_SO_H;

		/*  >>refer   mol  OH     +  N2O   =>   HNO   +     NO    UMIST  */

			c[ipN2O][ipOH] -= OH_N2O_HNO_NO_1;
			c[ipOH][ipOH] -= OH_N2O_HNO_NO_2;
			b[ipOH] -= b_OH_N2O_HNO_NO;
			
			c[ipN2O][ipN2O] -= OH_N2O_HNO_NO_1;
			c[ipOH][ipN2O] -= OH_N2O_HNO_NO_2;
			b[ipN2O] -= b_OH_N2O_HNO_NO;
			
			c[ipN2O][ipHNO] += OH_N2O_HNO_NO_1;
			c[ipOH][ipHNO] += OH_N2O_HNO_NO_2;
			b[ipHNO] += b_OH_N2O_HNO_NO;
			
			c[ipN2O][ipNO] += OH_N2O_HNO_NO_1;
			c[ipOH][ipNO] += OH_N2O_HNO_NO_2;
			b[ipNO] += b_OH_N2O_HNO_NO;

		/*  >>refer   mol  OH     +  CS    =>   OCS   +     H     UMIST  */
			
			c[ipCS][ipOH] -= OH_CS_OCS_H_1;
			c[ipOH][ipOH] -= OH_CS_OCS_H_2;
			b[ipOH] -= b_OH_CS_OCS_H;

			c[ipCS][ipCS] -= OH_CS_OCS_H_1;
			c[ipOH][ipCS] -= OH_CS_OCS_H_2;
			b[ipCS] -= b_OH_CS_OCS_H;
			
			c[ipCS][ipOCS] += OH_CS_OCS_H_1;
			c[ipOH][ipOCS] += OH_CS_OCS_H_2;
			b[ipOCS] += b_OH_CS_OCS_H;
			
		/*  >>refer   mol  NH3    +  CN    =>   HCN   +     NH2   UMIST  */

			c[ipCN][ipNH3] -= NH3_CN_HCN_NH2_1;
			c[ipNH3][ipNH3] -= NH3_CN_HCN_NH2_2;
			b[ipNH3] -= b_NH3_CN_HCN_NH2;
			
			c[ipCN][ipCN] -= NH3_CN_HCN_NH2_1;
			c[ipNH3][ipCN] -= NH3_CN_HCN_NH2_2;
			b[ipCN] -= b_NH3_CN_HCN_NH2;
			
			c[ipCN][ipHCN] += NH3_CN_HCN_NH2_1;
			c[ipNH3][ipHCN] += NH3_CN_HCN_NH2_2;
			b[ipHCN] += b_NH3_CN_HCN_NH2;
			
			c[ipCN][ipNH2] += NH3_CN_HCN_NH2_1;
			c[ipNH3][ipNH2] += NH3_CN_HCN_NH2_2;
			b[ipNH2] += b_NH3_CN_HCN_NH2;

		/*  >>refer   mol  CN     +  NO    =>   N2    +     CO    UMIST  */

			c[ipNO][ipCN] -= CN_NO_N2_CO_1;
			c[ipCN][ipCN] -= CN_NO_N2_CO_2;
			b[ipCN] -= b_CN_NO_N2_CO;
			
			c[ipNO][ipNO] -= CN_NO_N2_CO_1;
			c[ipCN][ipNO] -= CN_NO_N2_CO_2;
			b[ipNO] -= b_CN_NO_N2_CO;
			
			c[ipNO][ipN2] += CN_NO_N2_CO_1;
			c[ipCN][ipN2] += CN_NO_N2_CO_2;
			b[ipN2] += b_CN_NO_N2_CO;
			
			c[ipNO][ipCO] += CN_NO_N2_CO_1;
			c[ipCN][ipCO] += CN_NO_N2_CO_2;
			b[ipCO] += b_CN_NO_N2_CO;

		/*  >>refer   mol  CN     +  NO    =>   OCN   +     N     UMIST  */

			c[ipNO][ipCN] -= CN_NO_OCN_N_1;
			c[ipCN][ipCN] -= CN_NO_OCN_N_2;
			b[ipCN] -= b_CN_NO_OCN_N;
			
			c[ipNO][ipNO] -= CN_NO_OCN_N_1;
			c[ipCN][ipNO] -= CN_NO_OCN_N_2;
			b[ipNO] -= b_CN_NO_OCN_N;
			
			c[ipNO][ipOCN] += CN_NO_OCN_N_1;
			c[ipCN][ipOCN] += CN_NO_OCN_N_2;
			b[ipOCN] += b_CN_NO_OCN_N;
			
			c[ipNO][ipATN] += CN_NO_OCN_N_1;
			c[ipCN][ipATN] += CN_NO_OCN_N_2;
			b[ipATN] += b_CN_NO_OCN_N;
			
		/*  >>refer   mol  CN     +  HNO   =>   NO    +     HCN   UMIST  */

			c[ipHNO][ipCN] -= CN_HNO_NO_HCN_1;
			c[ipCN][ipCN] -= CN_HNO_NO_HCN_2;
			b[ipCN] -= b_CN_HNO_NO_HCN;
			
			c[ipHNO][ipHNO] -= CN_HNO_NO_HCN_1;
			c[ipCN][ipHNO] -= CN_HNO_NO_HCN_2;
			b[ipHNO] -= b_CN_HNO_NO_HCN;
			
			c[ipHNO][ipNO] += CN_HNO_NO_HCN_1;
			c[ipCN][ipNO] += CN_HNO_NO_HCN_2;
			b[ipNO] += b_CN_HNO_NO_HCN;
			
			c[ipHNO][ipHCN] += CN_HNO_NO_HCN_1;
			c[ipCN][ipHCN] += CN_HNO_NO_HCN_2;
			b[ipHCN] += b_CN_HNO_NO_HCN;

		/*  >>refer   mol  CN     +  O2    =>   NO    +     CO    UMIST  */

			c[ipO2][ipCN] -= CN_O2_NO_CO_1;
			c[ipCN][ipCN] -= CN_O2_NO_CO_2;
			b[ipCN] -= b_CN_O2_NO_CO;
			
			c[ipO2][ipO2] -= CN_O2_NO_CO_1;
			c[ipCN][ipO2] -= CN_O2_NO_CO_2;
			b[ipO2] -= b_CN_O2_NO_CO;
			
			c[ipO2][ipNO] += CN_O2_NO_CO_1;
			c[ipCN][ipNO] += CN_O2_NO_CO_2;
			b[ipNO] += b_CN_O2_NO_CO;
			
			c[ipO2][ipCO] += CN_O2_NO_CO_1;
			c[ipCN][ipCO] += CN_O2_NO_CO_2;
			b[ipCO] += b_CN_O2_NO_CO;

		/*  >>refer   mol  CN     +  S     =>   NS    +     C     UMIST  */

			c[ipATS][ipCN] -= CN_S_NS_C_1;
			c[ipCN][ipCN] -= CN_S_NS_C_2;
			b[ipCN] -= b_CN_S_NS_C;
			
			c[ipATS][ipATS] -= CN_S_NS_C_1;
			c[ipCN][ipATS] -= CN_S_NS_C_2;
			b[ipATS] -= b_CN_S_NS_C;
			
			c[ipATS][ipNS] += CN_S_NS_C_1;
			c[ipCN][ipNS] += CN_S_NS_C_2;
			b[ipNS] += b_CN_S_NS_C;

			c[ipATS][ipATC] += CN_S_NS_C_1;
			c[ipCN][ipATC] += CN_S_NS_C_2;
			b[ipATC] += b_CN_S_NS_C;
			
		/*  >>refer   mol  CN     +  S     =>   CS    +     N     UMIST  */

			c[ipATS][ipCN] -= CN_S_CS_N_1;
			c[ipCN][ipCN] -= CN_S_CS_N_2;
			b[ipCN] -= b_CN_S_CS_N;
			
			c[ipATS][ipATS] -= CN_S_CS_N_1;
			c[ipCN][ipATS] -= CN_S_CS_N_2;
			b[ipATS] -= b_CN_S_CS_N;
			
			c[ipATS][ipCS] += CN_S_CS_N_1;
			c[ipCN][ipCS] += CN_S_CS_N_2;
			b[ipCS] += b_CN_S_CS_N;

			c[ipATS][ipATN] += CN_S_CS_N_1;
			c[ipCN][ipATN] += CN_S_CS_N_2;
			b[ipATN] += b_CN_S_CS_N;

		/*  >>refer   mol  CO     +  HS    =>   OCS   +     H     UMIST  */

			c[ipHS][ipCO] -= CO_HS_OCS_H_1;
			c[ipCO][ipCO] -= CO_HS_OCS_H_2;
			b[ipCO] -= b_CO_HS_OCS_H;
			
			c[ipHS][ipHS] -= CO_HS_OCS_H_1;
			c[ipCO][ipHS] -= CO_HS_OCS_H_2;
			b[ipHS] -= b_CO_HS_OCS_H;
			
			c[ipHS][ipOCS] += CO_HS_OCS_H_1;
			c[ipCO][ipOCS] += CO_HS_OCS_H_2;
			b[ipOCS] += b_CO_HS_OCS_H;
			
		/*  >>refer   mol  N2     +  O2    =>   N2O   +     O     UMIST  */

			c[ipO2][ipN2] -= N2_O2_N2O_O_1;
			c[ipN2][ipN2] -= N2_O2_N2O_O_2;
			b[ipN2] -= b_N2_O2_N2O_O;
			
			c[ipO2][ipO2] -= N2_O2_N2O_O_1;
			c[ipN2][ipO2] -= N2_O2_N2O_O_2;
			b[ipO2] -= b_N2_O2_N2O_O;
			
			c[ipO2][ipN2O] += N2_O2_N2O_O_1;
			c[ipN2][ipN2O] += N2_O2_N2O_O_2;
			b[ipN2O] += b_N2_O2_N2O_O;
			
			c[ipO2][ipATO] += N2_O2_N2O_O_1;
			c[ipN2][ipATO] += N2_O2_N2O_O_2;
			b[ipATO] += b_N2_O2_N2O_O;
			
		/*  >>refer   mol  NO     +  NO    =>   O2    +     N2    UMIST  */

			c[ipNO][ipNO] -= NO_NO_O2_N2_1;
			c[ipNO][ipNO] -= NO_NO_O2_N2_2;
			b[ipNO] -= b_NO_NO_O2_N2;
			
			c[ipNO][ipO2] += 0.5*NO_NO_O2_N2_1;
			c[ipNO][ipO2] += 0.5*NO_NO_O2_N2_2;
			b[ipO2] += 0.5*b_NO_NO_O2_N2;
			
			c[ipNO][ipN2] += 0.5*NO_NO_O2_N2_1;
			c[ipNO][ipN2] += 0.5*NO_NO_O2_N2_2;
			b[ipN2] += 0.5*b_NO_NO_O2_N2;

		/*  >>refer   mol  NO     +  NO    =>   N2O   +     O     UMIST  */

			c[ipNO][ipNO] -= NO_NO_N2O_O_1;
			c[ipNO][ipNO] -= NO_NO_N2O_O_2;
			b[ipNO] -= b_NO_NO_N2O_O;
			
			c[ipNO][ipN2O] += 0.5*NO_NO_N2O_O_1;
			c[ipNO][ipN2O] += 0.5*NO_NO_N2O_O_2;
			b[ipN2O] += 0.5*b_NO_NO_N2O_O;
			
			c[ipNO][ipATO] += 0.5*NO_NO_N2O_O_1;
			c[ipNO][ipATO] += 0.5*NO_NO_N2O_O_2;
			b[ipATO] += 0.5*b_NO_NO_N2O_O;
			
		/*  >>refer   mol  NO     +  HNO   =>   N2O   +     OH    UMIST  */

			c[ipHNO][ipNO] -= NO_HNO_N2O_OH_1;
			c[ipNO][ipNO] -= NO_HNO_N2O_OH_2;
			b[ipNO] -= b_NO_HNO_N2O_OH;
			
			c[ipHNO][ipHNO] -= NO_HNO_N2O_OH_1;
			c[ipNO][ipHNO] -= NO_HNO_N2O_OH_2;
			b[ipHNO] -= b_NO_HNO_N2O_OH;
			
			c[ipHNO][ipN2O] += NO_HNO_N2O_OH_1;
			c[ipNO][ipN2O] += NO_HNO_N2O_OH_2;
			b[ipN2O] += b_NO_HNO_N2O_OH;
			
			c[ipHNO][ipOH] += NO_HNO_N2O_OH_1;
			c[ipNO][ipOH] += NO_HNO_N2O_OH_2;
			b[ipOH] += b_NO_HNO_N2O_OH;

		/*  >>refer   mol  NO     +  S     =>   NS    +     O     UMIST  */

			c[ipATS][ipNO] -= NO_S_NS_O_1;
			c[ipNO][ipNO] -= NO_S_NS_O_2;
			b[ipNO] -= b_NO_S_NS_O;
			
			c[ipATS][ipATS] -= NO_S_NS_O_1;
			c[ipNO][ipATS] -= NO_S_NS_O_2;
			b[ipATS] -= b_NO_S_NS_O;
			
			c[ipATS][ipNS] += NO_S_NS_O_1;
			c[ipNO][ipNS] += NO_S_NS_O_2;
			b[ipNS] += b_NO_S_NS_O;
	
			c[ipATS][ipATO] += NO_S_NS_O_1;
			c[ipNO][ipATO] += NO_S_NS_O_2;
			b[ipATO] += b_NO_S_NS_O;
			
		/*  >>refer   mol  NO     +  S     =>   SO    +     N     UMIST  */

			c[ipATS][ipNO] -= NO_S_SO_N_1;
			c[ipNO][ipNO] -= NO_S_SO_N_2;
			b[ipNO] -= b_NO_S_SO_N;
			
			c[ipATS][ipATS] -= NO_S_SO_N_1;
			c[ipNO][ipATS] -= NO_S_SO_N_2;
			b[ipATS] -= b_NO_S_SO_N;
			
			c[ipATS][ipSO] += NO_S_SO_N_1;
			c[ipNO][ipSO] += NO_S_SO_N_2;
			b[ipSO] += b_NO_S_SO_N;

			c[ipATS][ipATN] += NO_S_SO_N_1;
			c[ipNO][ipATN] += NO_S_SO_N_2;
			b[ipATN] += b_NO_S_SO_N;
			
		/*  >>refer   mol  O2     +  S     =>   SO    +     O     UMIST  */

			c[ipATS][ipO2] -= O2_S_SO_O_1;
			c[ipO2][ipO2] -= O2_S_SO_O_2;
			b[ipO2] -= b_O2_S_SO_O;
			
			c[ipATS][ipATS] -= O2_S_SO_O_1;
			c[ipO2][ipATS] -= O2_S_SO_O_2;
			b[ipATS] -= b_O2_S_SO_O;
			
			c[ipATS][ipSO] += O2_S_SO_O_1;
			c[ipO2][ipSO] += O2_S_SO_O_2;
			b[ipSO] += b_O2_S_SO_O;

			c[ipATS][ipATO] += O2_S_SO_O_1;
			c[ipO2][ipATO] += O2_S_SO_O_2;
			b[ipATO] += b_O2_S_SO_O;
			
		/*  >>refer   mol  S      +  SO    =>   S2    +     O     UMIST  */
			
			c[ipSO][ipSO] -= S_SO_S2_O_1;
			c[ipATS][ipSO] -= S_SO_S2_O_2;
			b[ipSO] -= b_S_SO_S2_O;
						
			c[ipSO][ipATS] -= S_SO_S2_O_1;
			c[ipATS][ipATS] -= S_SO_S2_O_2;
			b[ipATS] -= b_S_SO_S2_O;
			
			c[ipSO][ipS2] += S_SO_S2_O_1;
			c[ipATS][ipS2] += S_SO_S2_O_2;
			b[ipS2] += b_S_SO_S2_O;

			c[ipSO][ipATO] += S_SO_S2_O_1;
			c[ipATS][ipATO] += S_SO_S2_O_2;
			b[ipATO] += b_S_SO_S2_O;
			
		/*  >>refer   mol  H+     +  HNO   =>   NO+   +     H2    UMIST  */

			c[ipHNO][ipHNO] -= HP_HNO_NOP_H2;
			c[ipHNO][ipNOP] += HP_HNO_NOP_H2;

		/*  >>refer   mol  H+     +  HS    =>   S+    +     H2    UMIST  */

			c[ipHS][ipHS] -= HP_HS_SP_H2;
			c[ipHS][ipSP] += HP_HS_SP_H2;

		/*  >>refer   mol  H      +  HS+   =>   S+    +     H2    UMIST  */

			c[ipHSP][ipHSP] -= H_HSP_SP_H2;
			c[ipHSP][ipSP] += H_HSP_SP_H2;

		/*  >>refer   mol  H+     +  OCS   =>   HS+   +     CO    UMIST  */

			c[ipOCS][ipOCS] -= HP_OCS_HSP_CO;
			c[ipOCS][ipHSP] += HP_OCS_HSP_CO;
			c[ipOCS][ipCO] += HP_OCS_HSP_CO;

		/*  >>refer   mol  H2+    +  N     =>   NH+   +     H     UMIST  */

			c[ipATN][ipNHP] += H2P_N_NHP_H;
			c[ipATN][ipATN] -= H2P_N_NHP_H;

		/*  >>refer   mol  H2     +  N+    =>   NH+   +     H     UMIST  */

			c[ipNP][ipNP] -= H2_NP_NHP_H;
			c[ipNP][ipNHP] += H2_NP_NHP_H;

		/*  >>refer   mol  H2     +  NH+   =>   N     +     H3+   UMIST  */

			c[ipNHP][ipNHP] -= H2_NHP_N_H3P;
			c[ipNHP][ipATN] += H2_NHP_N_H3P;

		/*  >>refer   mol  H2+    +  NH    =>   NH2+  +     H     UMIST  */

			c[ipNH][ipNH] -= H2P_NH_NH2P_H;
			c[ipNH][ipNH2P] += H2P_NH_NH2P_H;

		/*  >>refer   mol  H2     +  NH+   =>   NH2+  +     H     UMIST  */

			c[ipNHP][ipNHP] -= H2_NHP_NH2P_H;
			c[ipNHP][ipNH2P] += H2_NHP_NH2P_H;

		/*  >>refer   mol  H2     +  NH2+  =>   NH3+  +     H     UMIST  */

			c[ipNH2P][ipNH2P] -= H2_NH2P_NH3P_H;
			c[ipNH2P][ipNH3P] += H2_NH2P_NH3P_H;

		/*  >>refer   mol  H2     +  NH3+  =>   NH4+  +     H     UMIST  */

			c[ipNH3P][ipNH3P] -= H2_NH3P_NH4P_H;
			c[ipNH3P][ipNH4P] += H2_NH3P_NH4P_H;

		/*  >>refer   mol  H2+    +  CN    =>   HCN+  +     H     UMIST  */

			c[ipCN][ipCN] -= H2P_CN_HCNP_H;
			c[ipCN][ipHCNP] += H2P_CN_HCNP_H;

		/*  >>refer   mol  H2     +  CN+   =>   HCN+  +     H     UMIST  */

			c[ipCNP][ipCNP] -= H2_CNP_HCNP_H;
			c[ipCNP][ipHCNP] += H2_CNP_HCNP_H;

		/*  >>refer   mol  H2+    +  NO    =>   HNO+  +     H     UMIST  */

			c[ipNO][ipNO] -= H2P_NO_HNOP_H;
			c[ipNO][ipHNOP] += H2P_NO_HNOP_H;

		/*  >>refer   mol  H2     +  S+    =>   HS+   +     H     UMIST  */

			c[ipSP][ipSP] -= H2_SP_HSP_H;
			c[ipSP][ipHSP] += H2_SP_HSP_H;

		/*  >>refer   mol  H2     +  CS+   =>   HCS+  +     H     UMIST  */

			c[ipCSP][ipCSP] -= H2_CSP_HCSP_H;
			c[ipCSP][ipHCSP] += H2_CSP_HCSP_H;

		/*  >>refer   mol  H2     +  NO2+  =>   NO+   +     H2O   UMIST  */

			c[ipNO2P][ipNO2P] -= H2_NO2P_NOP_H2O;
			c[ipNO2P][ipNOP] += H2_NO2P_NOP_H2O;
			c[ipNO2P][ipH2O] += H2_NO2P_NOP_H2O;

		/*  >>refer   mol  H3+    +  NH    =>   NH2+  +     H2    UMIST  */

			c[ipNH][ipNH] -= H3P_NH_NH2P_H2;
			c[ipNH][ipNH2P] += H3P_NH_NH2P_H2;

		/*  >>refer   mol  H3+    +  NH2   =>   NH3+  +     H2    UMIST  */

			c[ipNH2][ipNH2] -= H3P_NH2_NH3P_H2;
			c[ipNH2][ipNH3P] += H3P_NH2_NH3P_H2;

		/*  >>refer   mol  H3+    +  NH3   =>   NH4+  +     H2    UMIST  */

			c[ipNH3][ipNH3] -= H3P_NH3_NH4P_H2;
			c[ipNH3][ipNH4P] += H3P_NH3_NH4P_H2;

		/*  >>refer   mol  H3+    +  CN    =>   HCN+  +     H2    UMIST  */

			c[ipCN][ipCN] -= H3P_CN_HCNP_H2;
			c[ipCN][ipHCNP] += H3P_CN_HCNP_H2;

		/*  >>refer   mol  H3+    +  NO    =>   HNO+  +     H2    UMIST  */

			c[ipNO][ipNO] -= H3P_NO_HNOP_H2;
			c[ipNO][ipHNOP] += H3P_NO_HNOP_H2;

		/*  >>refer   mol  H3+    +  S     =>   HS+   +     H2    UMIST  */

			c[ipATS][ipHSP] += H3P_S_HSP_H2;
			c[ipATS][ipATS] -= H3P_S_HSP_H2;

		/*  >>refer   mol  H3+    +  CS    =>   HCS+  +     H2    UMIST  */

			c[ipCS][ipCS] -= H3P_CS_HCSP_H2;
			c[ipCS][ipHCSP] += H3P_CS_HCSP_H2;

		/*  >>refer   mol  H3+    +  NO2   =>   NO+   +     OH +  H2     UMIST  */

			c[ipNO2][ipNO2] -= H3P_NO2_NOP_OH_H2;
			c[ipNO2][ipNOP] += H3P_NO2_NOP_OH_H2;

		/*  >>refer   mol  He+    +  NH    =>   N+    +     He +  H      UMIST  */

			c[ipNH][ipNH] -= HeP_NH_NP_He_H;
			c[ipNH][ipNP] += HeP_NH_NP_He_H;

		/*  >>refer   mol  He+    +  NH2   =>   N+    +     He +  H2     UMIST  */

			c[ipNH2][ipNH2] -= HeP_NH2_NP_He_H2;
			c[ipNH2][ipNP] += HeP_NH2_NP_He_H2;

		/*  >>refer   mol  He+    +  NH2   =>   NH+   +     He +  H      UMIST  */

			c[ipNH2][ipNH2] -= HeP_NH2_NHP_He_H;
			c[ipNH2][ipNHP] += HeP_NH2_NHP_He_H;

		/*  >>refer   mol  He+    +  NH3   =>   NH+   +     He +  H2     UMIST  */

			c[ipNH3][ipNH3] -= HeP_NH3_NHP_He_H2;
			c[ipNH3][ipNHP] += HeP_NH3_NHP_He_H2;

		/*  >>refer   mol  He+    +  NH3   =>   NH2+  +     He +  H      UMIST  */

			c[ipNH3][ipNH3] -= HeP_NH3_NH2P_He_H;
			c[ipNH3][ipNH2P] += HeP_NH3_NH2P_He_H;

		/*  >>refer   mol  He+    +  CN    =>   N     +     C+ +  He     UMIST  */

			c[ipCN][ipCN] -= HeP_CN_N_CP_He;
			c[ipCN][ipCP] += HeP_CN_N_CP_He;
			c[ipCN][ipATN] += HeP_CN_N_CP_He;

		/*  >>refer   mol  He+    +  CN    =>   N+    +     C  +  He     UMIST  */

			c[ipCN][ipCN] -= HeP_CN_NP_C_He;
			c[ipCN][ipNP] += HeP_CN_NP_C_He;
			c[ipCN][ipATC] += HeP_CN_NP_C_He;

		/*  >>refer   mol  He+    +  HCN   =>   N     +     CH+ +  He     UMIST  */

			c[ipHCN][ipHCN] -= HeP_HCN_N_CHP_He;
			c[ipHCN][ipCHP] += HeP_HCN_N_CHP_He;
			c[ipHCN][ipATN] += HeP_HCN_N_CHP_He;

		/*  >>refer   mol  He+    +  HCN   =>   N+    +     CH +  He     UMIST  */

			c[ipHCN][ipHCN] -= HeP_HCN_NP_CH_He;
			c[ipHCN][ipNP] += HeP_HCN_NP_CH_He;
			c[ipHCN][ipCH] += HeP_HCN_NP_CH_He;

		/*  >>refer   mol  He+    +  HCN   =>   N     +     C+ +  He +  H       UMIST  */

			c[ipHCN][ipHCN] -= HeP_HCN_N_CP_He_H;
			c[ipHCN][ipCP] += HeP_HCN_N_CP_He_H;
			c[ipHCN][ipATN] += HeP_HCN_CNP_He_H;

		/*  >>refer   mol  He+    +  HCN   =>   CN+   +     He +  H      UMIST  */

			c[ipHCN][ipHCN] -= HeP_HCN_CNP_He_H;
			c[ipHCN][ipCNP]  += HeP_HCN_CNP_He_H;

		/*  >>refer   mol  He+    +  N2    =>   N+    +     N  +  He     UMIST  */

			c[ipN2][ipN2] -= HeP_N2_NP_N_He;
			c[ipN2][ipNP] += HeP_N2_NP_N_He;
			c[ipN2][ipATN] += HeP_N2_NP_N_He;

		/*  >>refer   mol  He+    +  NO    =>   O+    +     N  +  He     UMIST  */

			c[ipNO][ipNO] -= HeP_NO_OP_N_He;
			c[ipNO][ipOP] += HeP_NO_OP_N_He;
			c[ipNO][ipATN] += HeP_NO_OP_N_He;

		/*  >>refer   mol  He+    +  NO    =>   O     +     N+ +  He     UMIST  */

			c[ipNO][ipNO] -= HeP_NO_O_NP_He;
			c[ipNO][ipNP] += HeP_NO_O_NP_He;
			c[ipNO][ipATO] += HeP_NO_O_NP_He;

		/*  >>refer   mol  He+    +  HNO   =>   NO+   +     He +  H      UMIST  */

			c[ipHNO][ipHNO] -= HeP_HNO_NOP_He_H;
			c[ipHNO][ipNOP] += HeP_HNO_NOP_He_H;

		/*  >>refer   mol  He+    +  HNO   =>   NO    +     He +  H+     UMIST  */

			c[ipHNO][ipHNO] -= HeP_HNO_NO_He_HP;
			c[ipHNO][ipNO] += HeP_HNO_NO_He_HP;

		/*  >>refer   mol  He+    +  HS    =>   S+    +     He +  H      UMIST  */

			c[ipHS][ipHS] -= HeP_HS_SP_He_H;
			c[ipHS][ipSP] += HeP_HS_SP_He_H;

		/*  >>refer   mol  He+    +  OCN   =>   CN    +     O+ +  He     UMIST  */

			c[ipOCN][ipOCN] -= HeP_OCN_CN_OP_He;
			c[ipOCN][ipCN] += HeP_OCN_CN_OP_He;
			c[ipOCN][ipOP] += HeP_OCN_CN_OP_He;

		/*  >>refer   mol  He+    +  OCN   =>   CN+   +     O  +  He     UMIST  */

			c[ipOCN][ipOCN] -= HeP_OCN_CNP_O_He;
			c[ipOCN][ipCNP] += HeP_OCN_CNP_O_He;
			c[ipOCN][ipATO] += HeP_OCN_CNP_O_He;

		/*  >>refer   mol  He+    +  SIN   =>   SI+   +     N  +  He     UMIST  */

			c[ipSIN][ipSIN] -= HeP_SIN_SIP_N_He;
			c[ipSIN][ipSIP] += HeP_SIN_SIP_N_He;
			c[ipSIN][ipATN] += HeP_SIN_SIP_N_He;

		/*  >>refer   mol  He+    +  N2O   =>   N2    +     O+ +  He     UMIST  */

			c[ipN2O][ipN2O] -= HeP_N2O_N2_OP_He;
			c[ipN2O][ipN2] += HeP_N2O_N2_OP_He;
			c[ipN2O][ipOP] += HeP_N2O_N2_OP_He;

		/*  >>refer   mol  He+    +  N2O   =>   N2+   +     O  +  He     UMIST  */

			c[ipN2O][ipN2O] -= HeP_N2O_N2P_O_He;
			c[ipN2O][ipN2P] += HeP_N2O_N2P_O_He;
			c[ipN2O][ipATO] += HeP_N2O_N2P_O_He;

		/*  >>refer   mol  He+    +  N2O   =>   NO    +     N+ +  He     UMIST  */

			c[ipN2O][ipN2O] -= HeP_N2O_NO_NP_He;
			c[ipN2O][ipNO] += HeP_N2O_NO_NP_He;
			c[ipN2O][ipNP] += HeP_N2O_NO_NP_He;

		/*  >>refer   mol  He+    +  N2O   =>   NO+   +     N  +  He     UMIST  */

			c[ipN2O][ipN2O] -= HeP_N2O_NOP_N_He;
			c[ipN2O][ipNOP] += HeP_N2O_NOP_N_He;
			c[ipN2O][ipATN] += HeP_N2O_NOP_N_He;

		/*  >>refer   mol  He+    +  CS    =>   S+    +     C  +  He     UMIST  */

			c[ipCS][ipCS] -= HeP_CS_SP_C_He;
			c[ipCS][ipSP] += HeP_CS_SP_C_He;
			c[ipCS][ipATC] += HeP_CS_SP_C_He;

		/*  >>refer   mol  He+    +  CS    =>   S     +     C+ +  He     UMIST  */

			c[ipCS][ipCS] -= HeP_CS_S_CP_He;
			c[ipCS][ipCP] += HeP_CS_S_CP_He;
			c[ipCS][ipATS] += HeP_CS_S_CP_He;

		/*  >>refer   mol  He+    +  NS    =>   S     +     N+ +  He     UMIST  */

			c[ipNS][ipNS] -= HeP_NS_S_NP_He;
			c[ipNS][ipNP] += HeP_NS_S_NP_He;
			c[ipNS][ipATS] += HeP_NS_S_NP_He;

		/*  >>refer   mol  He+    +  NS    =>   S+    +     N  +  He     UMIST  */

			c[ipNS][ipNS] -= HeP_NS_SP_N_He;
			c[ipNS][ipSP] += HeP_NS_SP_N_He;
			c[ipNS][ipATN] += HeP_NS_SP_N_He;

		/*  >>refer   mol  He+    +  SO    =>   S     +     O+ +  He     UMIST  */

			c[ipSO][ipSO] -= HeP_SO_S_OP_He;
			c[ipSO][ipOP] += HeP_SO_S_OP_He;
			c[ipSO][ipATS] += HeP_SO_S_OP_He;

		/*  >>refer   mol  He+    +  SO    =>   S+    +     O  +  He     UMIST  */

			c[ipSO][ipSO] -= HeP_SO_SP_O_He;
			c[ipSO][ipSP] += HeP_SO_SP_O_He;
			c[ipSO][ipATO] += HeP_SO_SP_O_He;

		/*  >>refer   mol  He+    +  OCS   =>   S+    +     CO +  He     UMIST  */

			c[ipOCS][ipOCS] -= HeP_OCS_SP_CO_He;
			c[ipOCS][ipSP] += HeP_OCS_SP_CO_He;
			c[ipOCS][ipCO] += HeP_OCS_SP_CO_He;

		/*  >>refer   mol  He+    +  OCS   =>   S     +     CO+ +  He     UMIST  */

			c[ipOCS][ipOCS] -= HeP_OCS_S_COP_He;
			c[ipOCS][ipCOP] += HeP_OCS_S_COP_He;
			c[ipOCS][ipATS] += HeP_OCS_S_COP_He;

		/*  >>refer   mol  He+    +  OCS   =>   CS+   +     O  +  He     UMIST  */

			c[ipOCS][ipOCS] -= HeP_OCS_CSP_O_He;
			c[ipOCS][ipCSP] += HeP_OCS_CSP_O_He;
			c[ipOCS][ipATO] += HeP_OCS_CSP_O_He;

		/*  >>refer   mol  He+    +  OCS   =>   CS    +     O+ +  He     UMIST  */

			c[ipOCS][ipOCS] -= HeP_OCS_CS_OP_He;
			c[ipOCS][ipCS] += HeP_OCS_CS_OP_He;
			c[ipOCS][ipOP] += HeP_OCS_CS_OP_He;

		/*  >>refer   mol  He+    +  S2    =>   S+    +     S  +  He     UMIST  */

			c[ipS2][ipS2] -= HeP_S2_SP_S_He;
			c[ipS2][ipSP] += HeP_S2_SP_S_He;
			c[ipS2][ipATS] += HeP_S2_SP_S_He;

		/*  >>refer   mol  C      +  NH+   =>   N     +     CH+   UMIST  */
			
			c[ipNHP][ipNHP] -= C_NHP_N_CHP_1;
			c[ipATC][ipNHP] -= C_NHP_N_CHP_2;
			b[ipNHP] -= b_C_NHP_N_CHP;
						
			c[ipNHP][ipATC] -= C_NHP_N_CHP_1;
			c[ipATC][ipATC] -= C_NHP_N_CHP_2;
			b[ipATC] -= b_C_NHP_N_CHP;
			
			c[ipNHP][ipCHP] += C_NHP_N_CHP_1;
			c[ipATC][ipCHP] += C_NHP_N_CHP_2;
			b[ipCHP] += b_C_NHP_N_CHP;

			c[ipNHP][ipATN] += C_NHP_N_CHP_1;
			c[ipATC][ipATN] += C_NHP_N_CHP_2;
			b[ipATN] += b_C_NHP_N_CHP;

		/*  >>refer   mol  C+     +  NH    =>   CN+   +     H     UMIST  */

			c[ipNH][ipCP] -= CP_NH_CNP_H_1;
			c[ipCP][ipCP] -= CP_NH_CNP_H_2;
			b[ipCP] -= b_CP_NH_CNP_H;
			
			c[ipNH][ipNH] -= CP_NH_CNP_H_1;
			c[ipCP][ipNH] -= CP_NH_CNP_H_2;
			b[ipNH] -= b_CP_NH_CNP_H;
			
			c[ipNH][ipCNP] += CP_NH_CNP_H_1;
			c[ipCP][ipCNP] += CP_NH_CNP_H_2;
			b[ipCNP] += b_CP_NH_CNP_H;

			/*  >>refer   mol  C+     +  NH    =>   CH+   +     N     Zsargo & Federman 2003  */

			c[ipNH][ipCP] -= CP_NH_CHP_N_1;
			c[ipCP][ipCP] -= CP_NH_CHP_N_2;
			b[ipCP] -= b_CP_NH_CHP_N;
			
			c[ipNH][ipNH] -= CP_NH_CHP_N_1;
			c[ipCP][ipNH] -= CP_NH_CHP_N_2;
			b[ipNH] -= b_CP_NH_CHP_N;
			
			c[ipNH][ipCHP] += CP_NH_CHP_N_1;
			c[ipCP][ipCHP] += CP_NH_CHP_N_2;
			b[ipCHP] += b_CP_NH_CHP_N;

			c[ipNH][ipATN] += CP_NH_CHP_N_1;
			c[ipCP][ipATN] += CP_NH_CHP_N_2;
			b[ipATN] += b_CP_NH_CHP_N;


		/*  >>refer   mol  C+     +  NH2   =>   HCN+  +     H     UMIST  */

			c[ipNH2][ipCP] -= CP_NH2_HCNP_H_1;
			c[ipCP][ipCP] -= CP_NH2_HCNP_H_2;
			b[ipCP] -= b_CP_NH2_HCNP_H;
			
			c[ipNH2][ipNH2] -= CP_NH2_HCNP_H_1;
			c[ipCP][ipNH2] -= CP_NH2_HCNP_H_2;
			b[ipNH2] -= b_CP_NH2_HCNP_H;
			
			c[ipNH2][ipHCNP] += CP_NH2_HCNP_H_1;
			c[ipCP][ipHCNP] += CP_NH2_HCNP_H_2;
			b[ipHCNP] += b_CP_NH2_HCNP_H;

		/*  >>refer   mol  C      +  NH2+  =>   NH    +     CH+   UMIST  */

			c[ipNH2P][ipNH2P] -= C_NH2P_NH_CHP_1;
			c[ipATC][ipNH2P] -= C_NH2P_NH_CHP_2;
			b[ipNH2P] -= b_C_NH2P_NH_CHP;

			c[ipNH2P][ipATC] -= C_NH2P_NH_CHP_1;
			c[ipATC][ipATC] -= C_NH2P_NH_CHP_2;
			b[ipATC] -= b_C_NH2P_NH_CHP;
			
			c[ipNH2P][ipNH] += C_NH2P_NH_CHP_1;
			c[ipATC][ipNH] += C_NH2P_NH_CHP_2;
			b[ipNH] += b_C_NH2P_NH_CHP;
			
			c[ipNH2P][ipCHP] += C_NH2P_NH_CHP_1;
			c[ipATC][ipCHP] += C_NH2P_NH_CHP_2;
			b[ipCHP] += b_C_NH2P_NH_CHP;

		/*  >>refer   mol  C+     +  NH3   =>   HCN+  +     H2    UMIST  */

			c[ipNH3][ipCP] -= CP_NH3_HCNP_H2_1;
			c[ipCP][ipCP] -= CP_NH3_HCNP_H2_2;
			b[ipCP] -= b_CP_NH3_HCNP_H2;
			
			c[ipNH3][ipNH3] -= CP_NH3_HCNP_H2_1;
			c[ipCP][ipNH3] -= CP_NH3_HCNP_H2_2;
			b[ipNH3] -= b_CP_NH3_HCNP_H2;
			
			c[ipNH3][ipHCNP] += CP_NH3_HCNP_H2_1;
			c[ipCP][ipHCNP] += CP_NH3_HCNP_H2_2;
			b[ipHCNP] += b_CP_NH3_HCNP_H2;

		/*  >>refer   mol  C      +  NH3+  =>   NH    +     CH2+  UMIST  */

			c[ipNH3P][ipNH3P] -= C_NH3P_NH_CH2P_1;
			c[ipATC][ipNH3P] -= C_NH3P_NH_CH2P_2;
			b[ipNH3P] -= b_C_NH3P_NH_CH2P;

			c[ipNH3P][ipATC] -= C_NH3P_NH_CH2P_1;
			c[ipATC][ipATC] -= C_NH3P_NH_CH2P_2;
			b[ipATC] -= b_C_NH3P_NH_CH2P;
			
			c[ipNH3P][ipNH] += C_NH3P_NH_CH2P_1;
			c[ipATC][ipNH] += C_NH3P_NH_CH2P_2;
			b[ipNH] += b_C_NH3P_NH_CH2P;
			
			c[ipNH3P][ipCH2P] += C_NH3P_NH_CH2P_1;
			c[ipATC][ipCH2P] += C_NH3P_NH_CH2P_2;
			b[ipCH2P] += b_C_NH3P_NH_CH2P;

		/*  >>refer   mol  C      +  HCN+  =>   CN    +     CH+   UMIST  */

			c[ipHCNP][ipHCNP] -= C_HCNP_CN_CHP_1;
			c[ipATC][ipHCNP] -= C_HCNP_CN_CHP_2;
			b[ipHCNP] -= b_C_HCNP_CN_CHP;

			c[ipHCNP][ipATC] -= C_HCNP_CN_CHP_1;
			c[ipATC][ipATC] -= C_HCNP_CN_CHP_2;
			b[ipATC] -= b_C_HCNP_CN_CHP;
			
			c[ipHCNP][ipCN] += C_HCNP_CN_CHP_1;
			c[ipATC][ipCN] += C_HCNP_CN_CHP_2;
			b[ipCN] += b_C_HCNP_CN_CHP;
			
			c[ipHCNP][ipCHP] += C_HCNP_CN_CHP_1;
			c[ipATC][ipCHP] += C_HCNP_CN_CHP_2;
			b[ipCHP] += b_C_HCNP_CN_CHP;

		/*  >>refer   mol  C      +  HNO+  =>   NO    +     CH+   UMIST  */

			c[ipHNOP][ipHNOP] -= C_HNOP_NO_CHP_1;
			c[ipATC][ipHNOP] -= C_HNOP_NO_CHP_2;
			b[ipHNOP] -= b_C_HNOP_NO_CHP;

			c[ipHNOP][ipATC] -= C_HNOP_NO_CHP_1;
			c[ipATC][ipATC] -= C_HNOP_NO_CHP_2;
			b[ipATC] -= b_C_HNOP_NO_CHP;
			
			c[ipHNOP][ipNO] += C_HNOP_NO_CHP_1;
			c[ipATC][ipNO] += C_HNOP_NO_CHP_2;
			b[ipNO] += b_C_HNOP_NO_CHP;
			
			c[ipHNOP][ipCHP] += C_HNOP_NO_CHP_1;
			c[ipATC][ipCHP] += C_HNOP_NO_CHP_2;
			b[ipCHP] += b_C_HNOP_NO_CHP;

		/*  >>refer   mol  C+     +  HS    =>   CS+   +     H     UMIST  */

			c[ipHS][ipCP] -= CP_HS_CSP_H_1;
			c[ipCP][ipCP] -= CP_HS_CSP_H_2;
			b[ipCP] -= b_CP_HS_CSP_H;
			
			c[ipHS][ipHS] -= CP_HS_CSP_H_1;
			c[ipCP][ipHS] -= CP_HS_CSP_H_2;
			b[ipHS] -= b_CP_HS_CSP_H;
			
			c[ipHS][ipCSP] += CP_HS_CSP_H_1;
			c[ipCP][ipCSP] += CP_HS_CSP_H_2;
			b[ipCSP] += b_CP_HS_CSP_H;

		/*  >>refer   mol  C      +  HS+   =>   CS+   +     H     UMIST  */

			c[ipHSP][ipHSP] -= C_HSP_CSP_H_1;
			c[ipATC][ipHSP] -= C_HSP_CSP_H_2;
			b[ipHSP] -= b_C_HSP_CSP_H;

			c[ipHSP][ipATC] -= C_HSP_CSP_H_1;
			c[ipATC][ipATC] -= C_HSP_CSP_H_2;
			b[ipATC] -= b_C_HSP_CSP_H;
			
			c[ipHSP][ipCSP] += C_HSP_CSP_H_1;
			c[ipATC][ipCSP] += C_HSP_CSP_H_2;
			b[ipCSP] += b_C_HSP_CSP_H;

		/*  >>refer   mol  C+     +  OCN   =>   CO+   +     CN    UMIST  */

			c[ipOCN][ipCP] -= CP_OCN_COP_CN_1;
			c[ipCP][ipCP] -= CP_OCN_COP_CN_2;
			b[ipCP] -= b_CP_OCN_COP_CN;
			
			c[ipOCN][ipOCN] -= CP_OCN_COP_CN_1;
			c[ipCP][ipOCN] -= CP_OCN_COP_CN_2;
			b[ipOCN] -= b_CP_OCN_COP_CN;
			
			c[ipOCN][ipCOP] += CP_OCN_COP_CN_1;
			c[ipCP][ipCOP] += CP_OCN_COP_CN_2;
			b[ipCOP] += b_CP_OCN_COP_CN;
			
			c[ipOCN][ipCN] += CP_OCN_COP_CN_1;
			c[ipCP][ipCN] += CP_OCN_COP_CN_2;
			b[ipCN] += b_CP_OCN_COP_CN;

		/*  >>refer   mol  C+     +  NS    =>   CS+   +     N     UMIST  */

			c[ipNS][ipCP] -= CP_NS_CSP_N_1;
			c[ipCP][ipCP] -= CP_NS_CSP_N_2;
			b[ipCP] -= b_CP_NS_CSP_N;
			
			c[ipNS][ipNS] -= CP_NS_CSP_N_1;
			c[ipCP][ipNS] -= CP_NS_CSP_N_2;
			b[ipNS] -= b_CP_NS_CSP_N;
			
			c[ipNS][ipCSP] += CP_NS_CSP_N_1;
			c[ipCP][ipCSP] += CP_NS_CSP_N_2;
			b[ipCSP] += b_CP_NS_CSP_N;
			
			c[ipNS][ipATN] += CP_NS_CSP_N_1;
			c[ipCP][ipATN] += CP_NS_CSP_N_2;
			b[ipATN] += b_CP_NS_CSP_N;

		/*  >>refer   mol  C+     +  SO    =>   S+    +     CO    UMIST  */

			c[ipSO][ipCP] -= CP_SO_SP_CO_1;
			c[ipCP][ipCP] -= CP_SO_SP_CO_2;
			b[ipCP] -= b_CP_SO_SP_CO;
			
			c[ipSO][ipSO] -= CP_SO_SP_CO_1;
			c[ipCP][ipSO] -= CP_SO_SP_CO_2;
			b[ipSO] -= b_CP_SO_SP_CO;
			
			c[ipSO][ipSP] += CP_SO_SP_CO_1;
			c[ipCP][ipSP] += CP_SO_SP_CO_2;
			b[ipSP] += b_CP_SO_SP_CO;
			
			c[ipSO][ipCO] += CP_SO_SP_CO_1;
			c[ipCP][ipCO] += CP_SO_SP_CO_2;
			b[ipCO] += b_CP_SO_SP_CO;

		/*  >>refer   mol  C+     +  SO    =>   S     +     CO+   UMIST  */

			c[ipSO][ipCP] -= CP_SO_S_COP_1;
			c[ipCP][ipCP] -= CP_SO_S_COP_2;
			b[ipCP] -= b_CP_SO_S_COP;
			
			c[ipSO][ipSO] -= CP_SO_S_COP_1;
			c[ipCP][ipSO] -= CP_SO_S_COP_2;
			b[ipSO] -= b_CP_SO_S_COP;
			
			c[ipSO][ipCOP] += CP_SO_S_COP_1;
			c[ipCP][ipCOP] += CP_SO_S_COP_2;
			b[ipCOP] += b_CP_SO_S_COP;
			
			c[ipSO][ipATS] += CP_SO_S_COP_1;
			c[ipCP][ipATS] += CP_SO_S_COP_2;
			b[ipATS] += b_CP_SO_S_COP;

		/*  >>refer   mol  C+     +  SO    =>   CS+   +     O     UMIST  */

			c[ipSO][ipCP] -= CP_SO_CSP_O_1;
			c[ipCP][ipCP] -= CP_SO_CSP_O_2;
			b[ipCP] -= b_CP_SO_CSP_O;
			
			c[ipSO][ipSO] -= CP_SO_CSP_O_1;
			c[ipCP][ipSO] -= CP_SO_CSP_O_2;
			b[ipSO] -= b_CP_SO_CSP_O;
			
			c[ipSO][ipCSP] += CP_SO_CSP_O_1;
			c[ipCP][ipCSP] += CP_SO_CSP_O_2;
			b[ipCSP] += b_CP_SO_CSP_O;
			
			c[ipSO][ipATO] += CP_SO_CSP_O_1;
			c[ipCP][ipATO] += CP_SO_CSP_O_2;
			b[ipATO] += b_CP_SO_CSP_O;
			
		/*  >>refer   mol  C+     +  OCS   =>   CS+   +     CO    UMIST  */

			c[ipOCS][ipCP] -= CP_OCS_CSP_CO_1;
			c[ipCP][ipCP] -= CP_OCS_CSP_CO_2;
			b[ipCP] -= b_CP_OCS_CSP_CO;
			
			c[ipOCS][ipOCS] -= CP_OCS_CSP_CO_1;
			c[ipCP][ipOCS] -= CP_OCS_CSP_CO_2;
			b[ipOCS] -= b_CP_OCS_CSP_CO;
			
			c[ipOCS][ipCSP] += CP_OCS_CSP_CO_1;
			c[ipCP][ipCSP] += CP_OCS_CSP_CO_2;
			b[ipCSP] += b_CP_OCS_CSP_CO;
			
			c[ipOCS][ipCO] += CP_OCS_CSP_CO_1;
			c[ipCP][ipCO] += CP_OCS_CSP_CO_2;
			b[ipCO] += b_CP_OCS_CSP_CO;

		/*  >>refer   mol  CH+    +  N     =>   CN+   +     H     UMIST  */

			c[ipATN][ipCHP] -= CHP_N_CNP_H_1;
			c[ipCHP][ipCHP] -= CHP_N_CNP_H_2;
			b[ipCHP] -= b_CHP_N_CNP_H;

			c[ipATN][ipATN] -= CHP_N_CNP_H_1;
			c[ipCHP][ipATN] -= CHP_N_CNP_H_2;
			b[ipATN] -= b_CHP_N_CNP_H;
			
			c[ipATN][ipCNP] += CHP_N_CNP_H_1;
			c[ipCHP][ipCNP] += CHP_N_CNP_H_2;
			b[ipCNP] += b_CHP_N_CNP_H;

		/*  >>refer   mol  CH     +  N+    =>   CN+   +     H     UMIST  */

			c[ipNP][ipCH] -= CH_NP_CNP_H_1;
			c[ipCH][ipCH] -= CH_NP_CNP_H_2;
			b[ipCH] -= b_CH_NP_CNP_H;
			
			c[ipNP][ipNP] -= CH_NP_CNP_H_1;
			c[ipCH][ipNP] -= CH_NP_CNP_H_2;
			b[ipNP] -= b_CH_NP_CNP_H;
			
			c[ipNP][ipCNP] += CH_NP_CNP_H_1;
			c[ipCH][ipCNP] += CH_NP_CNP_H_2;
			b[ipCNP] += b_CH_NP_CNP_H;
			
		/*  >>refer   mol  CH     +  NH+   =>   CH2+  +     N     UMIST  */

			c[ipNHP][ipCH] -= CH_NHP_CH2P_N_1;
			c[ipCH][ipCH] -= CH_NHP_CH2P_N_2;
			b[ipCH] -= b_CH_NHP_CH2P_N;
			
			c[ipNHP][ipNHP] -= CH_NHP_CH2P_N_1;
			c[ipCH][ipNHP] -= CH_NHP_CH2P_N_2;
			b[ipNHP] -= b_CH_NHP_CH2P_N;
			
			c[ipNHP][ipCH2P] += CH_NHP_CH2P_N_1;
			c[ipCH][ipCH2P] += CH_NHP_CH2P_N_2;
			b[ipCH2P] += b_CH_NHP_CH2P_N;
			
			c[ipNHP][ipATN] += CH_NHP_CH2P_N_1;
			c[ipCH][ipATN] += CH_NHP_CH2P_N_2;
			b[ipATN] += b_CH_NHP_CH2P_N;
			
		/*  >>refer   mol  CH+    +  NH    =>   CN+   +     H2    UMIST  */

			c[ipNH][ipCHP] -= CHP_NH_CNP_H2_1;
			c[ipCHP][ipCHP] -= CHP_NH_CNP_H2_2;
			b[ipCHP] -= b_CHP_NH_CNP_H2;
			
			c[ipNH][ipNH] -= CHP_NH_CNP_H2_1;
			c[ipCHP][ipNH] -= CHP_NH_CNP_H2_2;
			b[ipNH] -= b_CHP_NH_CNP_H2;
			
			c[ipNH][ipCNP] += CHP_NH_CNP_H2_1;
			c[ipCHP][ipCNP] += CHP_NH_CNP_H2_2;
			b[ipCNP] += b_CHP_NH_CNP_H2;
			
		/*  >>refer   mol  CH+    +  NH2   =>   HCN+  +     H2    UMIST  */

			c[ipNH2][ipCHP] -= CHP_NH2_HCNP_H2_1;
			c[ipCHP][ipCHP] -= CHP_NH2_HCNP_H2_2;
			b[ipCHP] -= b_CHP_NH2_HCNP_H2;
			
			c[ipNH2][ipNH2] -= CHP_NH2_HCNP_H2_1;
			c[ipCHP][ipNH2] -= CHP_NH2_HCNP_H2_2;
			b[ipNH2] -= b_CHP_NH2_HCNP_H2;
			
			c[ipNH2][ipHCNP] += CHP_NH2_HCNP_H2_1;
			c[ipCHP][ipHCNP] += CHP_NH2_HCNP_H2_2;
			b[ipHCNP] += b_CHP_NH2_HCNP_H2;

		/*  >>refer   mol  CH     +  NH2+  =>   NH    +     CH2+  UMIST  */

			c[ipNH2P][ipCH] -= CH_NH2P_NH_CH2P_1;
			c[ipCH][ipCH] -= CH_NH2P_NH_CH2P_2;
			b[ipCH] -= b_CH_NH2P_NH_CH2P;
			
			c[ipNH2P][ipNH2P] -= CH_NH2P_NH_CH2P_1;
			c[ipCH][ipNH2P] -= CH_NH2P_NH_CH2P_2;
			b[ipNH2P] -= b_CH_NH2P_NH_CH2P;
			
			c[ipNH2P][ipNH] += CH_NH2P_NH_CH2P_1;
			c[ipCH][ipNH] += CH_NH2P_NH_CH2P_2;
			b[ipNH] += b_CH_NH2P_NH_CH2P;
			
			c[ipNH2P][ipCH2P] += CH_NH2P_NH_CH2P_1;
			c[ipCH][ipCH2P] += CH_NH2P_NH_CH2P_2;
			b[ipCH2P] += b_CH_NH2P_NH_CH2P;

		/*  >>refer   mol  CH+    +  NH3   =>   NH4+  +     C     UMIST  */

			c[ipNH3][ipCHP] -= CHP_NH3_NH4P_C_1;
			c[ipCHP][ipCHP] -= CHP_NH3_NH4P_C_2;
			b[ipCHP] -= b_CHP_NH3_NH4P_C;
			
			c[ipNH3][ipNH3] -= CHP_NH3_NH4P_C_1;
			c[ipCHP][ipNH3] -= CHP_NH3_NH4P_C_2;
			b[ipNH3] -= b_CHP_NH3_NH4P_C;
			
			c[ipNH3][ipNH4P] += CHP_NH3_NH4P_C_1;
			c[ipCHP][ipNH4P] += CHP_NH3_NH4P_C_2;
			b[ipNH4P] += b_CHP_NH3_NH4P_C;
			
			c[ipNH3][ipATC] += CHP_NH3_NH4P_C_1;
			c[ipCHP][ipATC] += CHP_NH3_NH4P_C_2;
			b[ipATC] += b_CHP_NH3_NH4P_C;

		/*  >>refer   mol  CH     +  NH3+  =>   NH4+  +     C     UMIST  */

			c[ipNH3P][ipCH] -= CH_NH3P_NH4P_C_1;
			c[ipCH][ipCH] -= CH_NH3P_NH4P_C_2;
			b[ipCH] -= b_CH_NH3P_NH4P_C;
			
			c[ipNH3P][ipNH3P] -= CH_NH3P_NH4P_C_1;
			c[ipCH][ipNH3P] -= CH_NH3P_NH4P_C_2;
			b[ipNH3P] -= b_CH_NH3P_NH4P_C;
			
			c[ipNH3P][ipNH4P] += CH_NH3P_NH4P_C_1;
			c[ipCH][ipNH4P] += CH_NH3P_NH4P_C_2;
			b[ipNH4P] += b_CH_NH3P_NH4P_C;
			
			c[ipNH3P][ipATC] += CH_NH3P_NH4P_C_1;
			c[ipCH][ipATC] += CH_NH3P_NH4P_C_2;
			b[ipATC] += b_CH_NH3P_NH4P_C;

		/*  >>refer   mol  CH     +  HCN+  =>   CN    +     CH2+  UMIST  */

			c[ipHCNP][ipCH] -= CH_HCNP_CN_CH2P_1;
			c[ipCH][ipCH] -= CH_HCNP_CN_CH2P_2;
			b[ipCH] -= b_CH_HCNP_CN_CH2P;
			
			c[ipHCNP][ipHCNP] -= CH_HCNP_CN_CH2P_1;
			c[ipCH][ipHCNP] -= CH_HCNP_CN_CH2P_2;
			b[ipHCNP] -= b_CH_HCNP_CN_CH2P;
			
			c[ipHCNP][ipCN] += CH_HCNP_CN_CH2P_1;
			c[ipCH][ipCN] += CH_HCNP_CN_CH2P_2;
			b[ipCN] += b_CH_HCNP_CN_CH2P;
			
			c[ipHCNP][ipCH2P] += CH_HCNP_CN_CH2P_1;
			c[ipCH][ipCH2P] += CH_HCNP_CN_CH2P_2;
			b[ipCH2P] += b_CH_HCNP_CN_CH2P;

		/*  >>refer   mol  CH     +  HNO+  =>   NO    +     CH2+  UMIST  */

			c[ipHNOP][ipCH] -= CH_HNOP_NO_CH2P_1;
			c[ipCH][ipCH] -= CH_HNOP_NO_CH2P_2;
			b[ipCH] -= b_CH_HNOP_NO_CH2P;
			
			c[ipHNOP][ipHNOP] -= CH_HNOP_NO_CH2P_1;
			c[ipCH][ipHNOP] -= CH_HNOP_NO_CH2P_2;
			b[ipHNOP] -= b_CH_HNOP_NO_CH2P;
			
			c[ipHNOP][ipNO] += CH_HNOP_NO_CH2P_1;
			c[ipCH][ipNO] += CH_HNOP_NO_CH2P_2;
			b[ipNO] += b_CH_HNOP_NO_CH2P;
			
			c[ipHNOP][ipCH2P] += CH_HNOP_NO_CH2P_1;
			c[ipCH][ipCH2P] += CH_HNOP_NO_CH2P_2;
			b[ipCH2P] += b_CH_HNOP_NO_CH2P;

		/*  >>refer   mol  CH+    +  S     =>   HS+   +     C     UMIST  */
			
			c[ipATS][ipCHP] -= CHP_S_HSP_C_1;
			c[ipCHP][ipCHP] -= CHP_S_HSP_C_2;
			b[ipCHP] -= b_CHP_S_HSP_C;
						
			c[ipATS][ipATS] -= CHP_S_HSP_C_1;
			c[ipCHP][ipATS] -= CHP_S_HSP_C_2;
			b[ipATS] -= b_CHP_S_HSP_C;
			
			c[ipATS][ipHSP] += CHP_S_HSP_C_1;
			c[ipCHP][ipHSP] += CHP_S_HSP_C_2;
			b[ipHSP] += b_CHP_S_HSP_C;

			c[ipATS][ipATC] += CHP_S_HSP_C_1;
			c[ipCHP][ipATC] += CHP_S_HSP_C_2;
			b[ipATC] += b_CHP_S_HSP_C;

		/*  >>refer   mol  CH+    +  S     =>   CS+   +     H     UMIST  */
			
			c[ipATS][ipCHP] -= CHP_S_CSP_H_1;
			c[ipCHP][ipCHP] -= CHP_S_CSP_H_2;
			b[ipCHP] -= b_CHP_S_CSP_H;
			
			c[ipATS][ipATS] -= CHP_S_CSP_H_1;
			c[ipCHP][ipATS] -= CHP_S_CSP_H_2;
			b[ipATS] -= b_CHP_S_CSP_H;
			
			c[ipATS][ipCSP] += CHP_S_CSP_H_1;
			c[ipCHP][ipCSP] += CHP_S_CSP_H_2;
			b[ipCSP] += b_CHP_S_CSP_H;

		/*  >>refer   mol  CH     +  S+    =>   CS+   +     H     UMIST  */

			c[ipSP][ipCH] -= CH_SP_CSP_H_1;
			c[ipCH][ipCH] -= CH_SP_CSP_H_2;
			b[ipCH] -= b_CH_SP_CSP_H;
			
			c[ipSP][ipSP] -= CH_SP_CSP_H_1;
			c[ipCH][ipSP] -= CH_SP_CSP_H_2;
			b[ipSP] -= b_CH_SP_CSP_H;
			
			c[ipSP][ipCSP] += CH_SP_CSP_H_1;
			c[ipCH][ipCSP] += CH_SP_CSP_H_2;
			b[ipCSP] += b_CH_SP_CSP_H;

		/*  >>refer   mol  CH     +  HS+   =>   S     +     CH2+  UMIST  */

			c[ipHSP][ipCH] -= CH_HSP_S_CH2P_1;
			c[ipCH][ipCH] -= CH_HSP_S_CH2P_2;
			b[ipCH] -= b_CH_HSP_S_CH2P;
			
			c[ipHSP][ipHSP] -= CH_HSP_S_CH2P_1;
			c[ipCH][ipHSP] -= CH_HSP_S_CH2P_2;
			b[ipHSP] -= b_CH_HSP_S_CH2P;
			
			c[ipHSP][ipCH2P] += CH_HSP_S_CH2P_1;
			c[ipCH][ipCH2P] += CH_HSP_S_CH2P_2;
			b[ipCH2P] += b_CH_HSP_S_CH2P;
			
			c[ipHSP][ipATS] += CH_HSP_S_CH2P_1;
			c[ipCH][ipATS] += CH_HSP_S_CH2P_2;
			b[ipATS] += b_CH_HSP_S_CH2P;

		/*  >>refer   mol  CH+    +  OCS   =>   HCS+  +     CO    UMIST  */

			c[ipOCS][ipCHP] -= CHP_OCS_HCSP_CO_1;
			c[ipCHP][ipCHP] -= CHP_OCS_HCSP_CO_2;
			b[ipCHP] -= b_CHP_OCS_HCSP_CO;
			
			c[ipOCS][ipOCS] -= CHP_OCS_HCSP_CO_1;
			c[ipCHP][ipOCS] -= CHP_OCS_HCSP_CO_2;
			b[ipOCS] -= b_CHP_OCS_HCSP_CO;
			
			c[ipOCS][ipHCSP] += CHP_OCS_HCSP_CO_1;
			c[ipCHP][ipHCSP] += CHP_OCS_HCSP_CO_2;
			b[ipHCSP] += b_CHP_OCS_HCSP_CO;
			
			c[ipOCS][ipCO] += CHP_OCS_HCSP_CO_1;
			c[ipCHP][ipCO] += CHP_OCS_HCSP_CO_2;
			b[ipCO] += b_CHP_OCS_HCSP_CO;

		/*  >>refer   mol  N      +  CH2+  =>   HCN+  +     H     UMIST  */

			c[ipCH2P][ipCH2P] -= N_CH2P_HCNP_H_1;
			c[ipATN][ipCH2P] -= N_CH2P_HCNP_H_2;
			b[ipCH2P] -= b_N_CH2P_HCNP_H;

			c[ipCH2P][ipATN] -= N_CH2P_HCNP_H_1;
			c[ipATN][ipATN] -= N_CH2P_HCNP_H_2;
			b[ipATN] -= b_N_CH2P_HCNP_H;
			
			c[ipCH2P][ipHCNP] += N_CH2P_HCNP_H_1;
			c[ipATN][ipHCNP] += N_CH2P_HCNP_H_2;
			b[ipHCNP] += b_N_CH2P_HCNP_H;

		/*  >>refer   mol  N+     +  NH    =>   N2+   +     H     UMIST  */

			c[ipNH][ipNP] -= NP_NH_N2P_H_1;
			c[ipNP][ipNP] -= NP_NH_N2P_H_2;
			b[ipNP] -= b_NP_NH_N2P_H;
			
			c[ipNH][ipNH] -= NP_NH_N2P_H_1;
			c[ipNP][ipNH] -= NP_NH_N2P_H_2;
			b[ipNH] -= b_NP_NH_N2P_H;
			
			c[ipNH][ipN2P] += NP_NH_N2P_H_1;
			c[ipNP][ipN2P] += NP_NH_N2P_H_2;
			b[ipN2P] += b_NP_NH_N2P_H;
			
		/*  >>refer   mol  N      +  NH+   =>   N2+   +     H     UMIST  */

			c[ipNHP][ipNHP] -= N_NHP_N2P_H_1;
			c[ipATN][ipNHP] -= N_NHP_N2P_H_2;
			b[ipNHP] -= b_N_NHP_N2P_H;

			c[ipNHP][ipATN] -= N_NHP_N2P_H_1;
			c[ipATN][ipATN] -= N_NHP_N2P_H_2;
			b[ipATN] -= b_N_NHP_N2P_H;
			
			c[ipNHP][ipN2P] += N_NHP_N2P_H_1;
			c[ipATN][ipN2P] += N_NHP_N2P_H_2;
			b[ipN2P] += b_N_NHP_N2P_H;

		/*  >>refer   mol  N+     +  CH4   =>   HCN+  +     H2 +  H      UMIST  */

			c[ipCH4][ipNP] -= NP_CH4_HCNP_H2_H_1;
			c[ipNP][ipNP] -= NP_CH4_HCNP_H2_H_2;
			b[ipNP] -= b_NP_CH4_HCNP_H2_H;
			
			c[ipCH4][ipCH4] -= NP_CH4_HCNP_H2_H_1;
			c[ipNP][ipCH4] -= NP_CH4_HCNP_H2_H_2;
			b[ipCH4] -= b_NP_CH4_HCNP_H2_H;
			
			c[ipCH4][ipHCNP] += NP_CH4_HCNP_H2_H_1;
			c[ipNP][ipHCNP] += NP_CH4_HCNP_H2_H_2;
			b[ipHCNP] += b_NP_CH4_HCNP_H2_H;
			
		/*  >>refer   mol  N      +  OH+   =>   NO+   +     H     UMIST  */

			c[ipOHP][ipOHP] -= N_OHP_NOP_H_1;
			c[ipATN][ipOHP] -= N_OHP_NOP_H_2;
			b[ipOHP] -= b_N_OHP_NOP_H;

			c[ipOHP][ipATN] -= N_OHP_NOP_H_1;
			c[ipATN][ipATN] -= N_OHP_NOP_H_2;
			b[ipATN] -= b_N_OHP_NOP_H;
			
			c[ipOHP][ipNOP] += N_OHP_NOP_H_1;
			c[ipATN][ipNOP] += N_OHP_NOP_H_2;
			b[ipNOP] += b_N_OHP_NOP_H;

		/*  >>refer   mol  N+     +  NH3   =>   NH2+  +     NH    UMIST  */

			c[ipNH3][ipNP] -= NP_NH3_NH2P_NH_1;
			c[ipNP][ipNP] -= NP_NH3_NH2P_NH_2;
			b[ipNP] -= b_NP_NH3_NH2P_NH;
			
			c[ipNH3][ipNH3] -= NP_NH3_NH2P_NH_1;
			c[ipNP][ipNH3] -= NP_NH3_NH2P_NH_2;
			b[ipNH3] -= b_NP_NH3_NH2P_NH;
			
			c[ipNH3][ipNH2P] += NP_NH3_NH2P_NH_1;
			c[ipNP][ipNH2P] += NP_NH3_NH2P_NH_2;
			b[ipNH2P] += b_NP_NH3_NH2P_NH;
			
			c[ipNH3][ipNH] += NP_NH3_NH2P_NH_1;
			c[ipNP][ipNH] += NP_NH3_NH2P_NH_2;
			b[ipNH] += b_NP_NH3_NH2P_NH;

		/*  >>refer   mol  N      +  H2O+  =>   HNO+  +     H     UMIST  */

			c[ipH2OP][ipH2OP] -= N_H2OP_HNOP_H_1;
			c[ipATN][ipH2OP] -= N_H2OP_HNOP_H_2;
			b[ipH2OP] -= b_N_H2OP_HNOP_H;

			c[ipH2OP][ipATN] -= N_H2OP_HNOP_H_1;
			c[ipATN][ipATN] -= N_H2OP_HNOP_H_2;
			b[ipATN] -= b_N_H2OP_HNOP_H;
			
			c[ipH2OP][ipHNOP] += N_H2OP_HNOP_H_1;
			c[ipATN][ipHNOP] += N_H2OP_HNOP_H_2;
			b[ipHNOP] += b_N_H2OP_HNOP_H;

		/*  >>refer   mol  N+     +  CO    =>   NO+   +     C     UMIST  */

			c[ipCO][ipNP] -= NP_CO_NOP_C_1;
			c[ipNP][ipNP] -= NP_CO_NOP_C_2;
			b[ipNP] -= b_NP_CO_NOP_C;
			
			c[ipCO][ipCO] -= NP_CO_NOP_C_1;
			c[ipNP][ipCO] -= NP_CO_NOP_C_2;
			b[ipCO] -= b_NP_CO_NOP_C;
			
			c[ipCO][ipNOP] += NP_CO_NOP_C_1;
			c[ipNP][ipNOP] += NP_CO_NOP_C_2;
			b[ipNOP] += b_NP_CO_NOP_C;
			
			c[ipCO][ipATC] += NP_CO_NOP_C_1;
			c[ipNP][ipATC] += NP_CO_NOP_C_2;
			b[ipATC] += b_NP_CO_NOP_C;
			
		/*  >>refer   mol  N+     +  NO    =>   N2+   +     O     UMIST  */

			c[ipNO][ipNP] -= NP_NO_N2P_O_1;
			c[ipNP][ipNP] -= NP_NO_N2P_O_2;
			b[ipNP] -= b_NP_NO_N2P_O;
			
			c[ipNO][ipNO] -= NP_NO_N2P_O_1;
			c[ipNP][ipNO] -= NP_NO_N2P_O_2;
			b[ipNO] -= b_NP_NO_N2P_O;
			
			c[ipNO][ipN2P] += NP_NO_N2P_O_1;
			c[ipNP][ipN2P] += NP_NO_N2P_O_2;
			b[ipN2P] += b_NP_NO_N2P_O;
			
			c[ipNO][ipATO] += NP_NO_N2P_O_1;
			c[ipNP][ipATO] += NP_NO_N2P_O_2;
			b[ipATO] += b_NP_NO_N2P_O;
			
		/*  >>refer   mol  N+     +  O2    =>   NO    +     O+    UMIST  */

			c[ipO2][ipNP] -= NP_O2_NO_OP_1;
			c[ipNP][ipNP] -= NP_O2_NO_OP_2;
			b[ipNP] -= b_NP_O2_NO_OP;
			
			c[ipO2][ipO2] -= NP_O2_NO_OP_1;
			c[ipNP][ipO2] -= NP_O2_NO_OP_2;
			b[ipO2] -= b_NP_O2_NO_OP;
			
			c[ipO2][ipNO] += NP_O2_NO_OP_1;
			c[ipNP][ipNO] += NP_O2_NO_OP_2;
			b[ipNO] += b_NP_O2_NO_OP;
			
			c[ipO2][ipOP] += NP_O2_NO_OP_1;
			c[ipNP][ipOP] += NP_O2_NO_OP_2;
			b[ipOP] += b_NP_O2_NO_OP;

		/*  >>refer   mol  N+     +  O2    =>   NO+   +     O     UMIST  */

			c[ipO2][ipNP] -= NP_O2_NOP_O_1;
			c[ipNP][ipNP] -= NP_O2_NOP_O_2;
			b[ipNP] -= b_NP_O2_NOP_O;
			
			c[ipO2][ipO2] -= NP_O2_NOP_O_1;
			c[ipNP][ipO2] -= NP_O2_NOP_O_2;
			b[ipO2] -= b_NP_O2_NOP_O;
			
			c[ipO2][ipNOP] += NP_O2_NOP_O_1;
			c[ipNP][ipNOP] += NP_O2_NOP_O_2;
			b[ipNOP] += b_NP_O2_NOP_O;
			
			c[ipO2][ipATO] += NP_O2_NOP_O_1;
			c[ipNP][ipATO] += NP_O2_NOP_O_2;
			b[ipATO] += b_NP_O2_NOP_O;

		/*  >>refer   mol  N      +  O2+   =>   NO+   +     O     UMIST  */

			c[ipO2P][ipO2P] -= N_O2P_NOP_O_1;
			c[ipATN][ipO2P] -= N_O2P_NOP_O_2;
			b[ipO2P] -= b_N_O2P_NOP_O;
			
			c[ipO2P][ipATN] -= N_O2P_NOP_O_1;
			c[ipATN][ipATN] -= N_O2P_NOP_O_2;
			b[ipATN] -= b_N_O2P_NOP_O;
			
			c[ipO2P][ipNOP] += N_O2P_NOP_O_1;
			c[ipATN][ipNOP] += N_O2P_NOP_O_2;
			b[ipNOP] += b_N_O2P_NOP_O;

			c[ipO2P][ipATO] += N_O2P_NOP_O_1;
			c[ipATN][ipATO] += N_O2P_NOP_O_2;
			b[ipATO] += b_N_O2P_NOP_O;

		/*  >>refer   mol  N      +  HS+   =>   NS+   +     H     UMIST  */

			c[ipHSP][ipHSP] -= N_HSP_NSP_H_1;
			c[ipATN][ipHSP] -= N_HSP_NSP_H_2;
			b[ipHSP] -= b_N_HSP_NSP_H;

			c[ipHSP][ipATN] -= N_HSP_NSP_H_1;
			c[ipATN][ipATN] -= N_HSP_NSP_H_2;
			b[ipATN] -= b_N_HSP_NSP_H;
			
			c[ipHSP][ipNSP] += N_HSP_NSP_H_1;
			c[ipATN][ipNSP] += N_HSP_NSP_H_2;
			b[ipNSP] += b_N_HSP_NSP_H;

		/*  >>refer   mol  N      +  SIO+  =>   NO+   +     Si    UMIST  */

			c[ipSIOP][ipSIOP] -= N_SIOP_NOP_SI_1;
			c[ipATN][ipSIOP] -= N_SIOP_NOP_SI_2;
			b[ipSIOP] -= b_N_SIOP_NOP_SI;
			
			c[ipSIOP][ipATN] -= N_SIOP_NOP_SI_1;
			c[ipATN][ipATN] -= N_SIOP_NOP_SI_2;
			b[ipATN] -= b_N_SIOP_NOP_SI;
			
			c[ipSIOP][ipNOP] += N_SIOP_NOP_SI_1;
			c[ipATN][ipNOP] += N_SIOP_NOP_SI_2;
			b[ipNOP] += b_N_SIOP_NOP_SI;

			c[ipSIOP][ipATSI] += N_SIOP_NOP_SI_1;
			c[ipATN][ipATSI] += N_SIOP_NOP_SI_2;
			b[ipATSI] += b_N_SIOP_NOP_SI;

		/*  >>refer   mol  N      +  SIO+  =>   NO    +     SI+   UMIST  */

			c[ipSIOP][ipSIOP] -= N_SIOP_NO_SIP_1;
			c[ipATN][ipSIOP] -= N_SIOP_NO_SIP_2;
			b[ipSIOP] -= b_N_SIOP_NO_SIP;

			c[ipSIOP][ipATN] -= N_SIOP_NO_SIP_1;
			c[ipATN][ipATN] -= N_SIOP_NO_SIP_2;
			b[ipATN] -= b_N_SIOP_NO_SIP;
			
			c[ipSIOP][ipNO] += N_SIOP_NO_SIP_1;
			c[ipATN][ipNO] += N_SIOP_NO_SIP_2;
			b[ipNO] += b_N_SIOP_NO_SIP;
			
			c[ipSIOP][ipSIP] += N_SIOP_NO_SIP_1;
			c[ipATN][ipSIP] += N_SIOP_NO_SIP_2;
			b[ipSIP] += b_N_SIOP_NO_SIP;

		/*  >>refer   mol  N      +  SO+   =>   NS+   +     O     UMIST  */

			c[ipSOP][ipSOP] -= N_SOP_NSP_O_1;
			c[ipATN][ipSOP] -= N_SOP_NSP_O_2;
			b[ipSOP] -= b_N_SOP_NSP_O;
			
			c[ipSOP][ipATN] -= N_SOP_NSP_O_1;
			c[ipATN][ipATN] -= N_SOP_NSP_O_2;
			b[ipATN] -= b_N_SOP_NSP_O;
			
			c[ipSOP][ipNSP] += N_SOP_NSP_O_1;
			c[ipATN][ipNSP] += N_SOP_NSP_O_2;
			b[ipNSP] += b_N_SOP_NSP_O;
			
			c[ipSOP][ipATO] += N_SOP_NSP_O_1;
			c[ipATN][ipATO] += N_SOP_NSP_O_2;
			b[ipATO] += b_N_SOP_NSP_O;

		/*  >>refer   mol  N+     +  OCS   =>   S+    +     CO +  N      UMIST  */

			c[ipOCS][ipNP] -= NP_OCS_SP_CO_N_1;
			c[ipNP][ipNP] -= NP_OCS_SP_CO_N_2;
			b[ipNP] -= b_NP_OCS_SP_CO_N;
			
			c[ipOCS][ipOCS] -= NP_OCS_SP_CO_N_1;
			c[ipNP][ipOCS] -= NP_OCS_SP_CO_N_2;
			b[ipOCS] -= b_NP_OCS_SP_CO_N;
			
			c[ipOCS][ipSP] += NP_OCS_SP_CO_N_1;
			c[ipNP][ipSP] += NP_OCS_SP_CO_N_2;
			b[ipSP] += b_NP_OCS_SP_CO_N;
			
			c[ipOCS][ipCO] += NP_OCS_SP_CO_N_1;
			c[ipNP][ipCO] += NP_OCS_SP_CO_N_2;
			b[ipCO] += b_NP_OCS_SP_CO_N;
			
			c[ipOCS][ipATN] += NP_OCS_SP_CO_N_1;
			c[ipNP][ipATN] += NP_OCS_SP_CO_N_2;
			b[ipATN] += b_NP_OCS_SP_CO_N;

		/*  >>refer   mol  N+     +  OCS   =>   CS+   +     NO    UMIST  */

			c[ipOCS][ipNP] -= NP_OCS_CSP_NO_1;
			c[ipNP][ipNP] -= NP_OCS_CSP_NO_2;
			b[ipNP] -= b_NP_OCS_CSP_NO;
			
			c[ipOCS][ipOCS] -= NP_OCS_CSP_NO_1;
			c[ipNP][ipOCS] -= NP_OCS_CSP_NO_2;
			b[ipOCS] -= b_NP_OCS_CSP_NO;
			
			c[ipOCS][ipCSP] += NP_OCS_CSP_NO_1;
			c[ipNP][ipCSP] += NP_OCS_CSP_NO_2;
			b[ipCSP] += b_NP_OCS_CSP_NO;
			
			c[ipOCS][ipNO] += NP_OCS_CSP_NO_1;
			c[ipNP][ipNO] += NP_OCS_CSP_NO_2;
			b[ipNO] += b_NP_OCS_CSP_NO;

		/*  >>refer   mol  CH2    +  NH+   =>   CH3+  +     N     UMIST  */

			c[ipNHP][ipCH2] -= CH2_NHP_CH3P_N_1;
			c[ipCH2][ipCH2] -= CH2_NHP_CH3P_N_2;
			b[ipCH2] -= b_CH2_NHP_CH3P_N;
			
			c[ipNHP][ipNHP] -= CH2_NHP_CH3P_N_1;
			c[ipCH2][ipNHP] -= CH2_NHP_CH3P_N_2;
			b[ipNHP] -= b_CH2_NHP_CH3P_N;
			
			c[ipNHP][ipCH3P] += CH2_NHP_CH3P_N_1;
			c[ipCH2][ipCH3P] += CH2_NHP_CH3P_N_2;
			b[ipCH3P] += b_CH2_NHP_CH3P_N;
			
			c[ipNHP][ipATN] += CH2_NHP_CH3P_N_1;
			c[ipCH2][ipATN] += CH2_NHP_CH3P_N_2;
			b[ipATN] += b_CH2_NHP_CH3P_N;

		/*  >>refer   mol  CH2    +  NH2+  =>   CH3+  +     NH    UMIST  */

			c[ipNH2P][ipCH2] -= CH2_NH2P_CH3P_NH_1;
			c[ipCH2][ipCH2] -= CH2_NH2P_CH3P_NH_2;
			b[ipCH2] -= b_CH2_NH2P_CH3P_NH;
			
			c[ipNH2P][ipNH2P] -= CH2_NH2P_CH3P_NH_1;
			c[ipCH2][ipNH2P] -= CH2_NH2P_CH3P_NH_2;
			b[ipNH2P] -= b_CH2_NH2P_CH3P_NH;
			
			c[ipNH2P][ipCH3P] += CH2_NH2P_CH3P_NH_1;
			c[ipCH2][ipCH3P] += CH2_NH2P_CH3P_NH_2;
			b[ipCH3P] += b_CH2_NH2P_CH3P_NH;
			
			c[ipNH2P][ipNH] += CH2_NH2P_CH3P_NH_1;
			c[ipCH2][ipNH] += CH2_NH2P_CH3P_NH_2;
			b[ipNH] += b_CH2_NH2P_CH3P_NH;

		/*  >>refer   mol  CH2+   +  NH3   =>   NH4+  +     CH    UMIST  */

			c[ipNH3][ipCH2P] -= CH2P_NH3_NH4P_CH_1;
			c[ipCH2P][ipCH2P] -= CH2P_NH3_NH4P_CH_2;
			b[ipCH2P] -= b_CH2P_NH3_NH4P_CH;
			
			c[ipNH3][ipNH3] -= CH2P_NH3_NH4P_CH_1;
			c[ipCH2P][ipNH3] -= CH2P_NH3_NH4P_CH_2;
			b[ipNH3] -= b_CH2P_NH3_NH4P_CH;
			
			c[ipNH3][ipNH4P] += CH2P_NH3_NH4P_CH_1;
			c[ipCH2P][ipNH4P] += CH2P_NH3_NH4P_CH_2;
			b[ipNH4P] += b_CH2P_NH3_NH4P_CH;
			
			c[ipNH3][ipCH] += CH2P_NH3_NH4P_CH_1;
			c[ipCH2P][ipCH] += CH2P_NH3_NH4P_CH_2;
			b[ipCH] += b_CH2P_NH3_NH4P_CH;

		/*  >>refer   mol  CH2    +  NH3+  =>   NH2   +     CH3+  UMIST  */

			c[ipNH3P][ipCH2] -= CH2_NH3P_NH2_CH3P_1;
			c[ipCH2][ipCH2] -= CH2_NH3P_NH2_CH3P_2;
			b[ipCH2] -= b_CH2_NH3P_NH2_CH3P;
			
			c[ipNH3P][ipNH3P] -= CH2_NH3P_NH2_CH3P_1;
			c[ipCH2][ipNH3P] -= CH2_NH3P_NH2_CH3P_2;
			b[ipNH3P] -= b_CH2_NH3P_NH2_CH3P;
			
			c[ipNH3P][ipNH2] += CH2_NH3P_NH2_CH3P_1;
			c[ipCH2][ipNH2] += CH2_NH3P_NH2_CH3P_2;
			b[ipNH2] += b_CH2_NH3P_NH2_CH3P;
			
			c[ipNH3P][ipCH3P] += CH2_NH3P_NH2_CH3P_1;
			c[ipCH2][ipCH3P] += CH2_NH3P_NH2_CH3P_2;
			b[ipCH3P] += b_CH2_NH3P_NH2_CH3P;

		/*  >>refer   mol  CH2    +  HCN+  =>   CN    +     CH3+  UMIST  */

			c[ipHCNP][ipCH2] -= CH2_HCNP_CN_CH3P_1;
			c[ipCH2][ipCH2] -= CH2_HCNP_CN_CH3P_2;
			b[ipCH2] -= b_CH2_HCNP_CN_CH3P;
			
			c[ipHCNP][ipHCNP] -= CH2_HCNP_CN_CH3P_1;
			c[ipCH2][ipHCNP] -= CH2_HCNP_CN_CH3P_2;
			b[ipHCNP] -= b_CH2_HCNP_CN_CH3P;
			
			c[ipHCNP][ipCN] += CH2_HCNP_CN_CH3P_1;
			c[ipCH2][ipCN] += CH2_HCNP_CN_CH3P_2;
			b[ipCN] += b_CH2_HCNP_CN_CH3P;
			
			c[ipHCNP][ipCH3P] += CH2_HCNP_CN_CH3P_1;
			c[ipCH2][ipCH3P] += CH2_HCNP_CN_CH3P_2;
			b[ipCH3P] += b_CH2_HCNP_CN_CH3P;

		/*  >>refer   mol  CH2    +  HNO+  =>   NO    +     CH3+  UMIST  */

			c[ipHNOP][ipCH2] -= CH2_HNOP_NO_CH3P_1;
			c[ipCH2][ipCH2] -= CH2_HNOP_NO_CH3P_2;
			b[ipCH2] -= b_CH2_HNOP_NO_CH3P;
			
			c[ipHNOP][ipHNOP] -= CH2_HNOP_NO_CH3P_1;
			c[ipCH2][ipHNOP] -= CH2_HNOP_NO_CH3P_2;
			b[ipHNOP] -= b_CH2_HNOP_NO_CH3P;
			
			c[ipHNOP][ipNO] += CH2_HNOP_NO_CH3P_1;
			c[ipCH2][ipNO] += CH2_HNOP_NO_CH3P_2;
			b[ipNO] += b_CH2_HNOP_NO_CH3P;
			
			c[ipHNOP][ipCH3P] += CH2_HNOP_NO_CH3P_1;
			c[ipCH2][ipCH3P] += CH2_HNOP_NO_CH3P_2;
			b[ipCH3P] += b_CH2_HNOP_NO_CH3P;

		/*  >>refer   mol  CH2+   +  S     =>   HCS+  +     H     UMIST  */

			c[ipATS][ipCH2P] -= CH2P_S_HCSP_H_1;
			c[ipCH2P][ipCH2P] -= CH2P_S_HCSP_H_2;
			b[ipCH2P] -= b_CH2P_S_HCSP_H;
			
			c[ipATS][ipATS] -= CH2P_S_HCSP_H_1;
			c[ipCH2P][ipATS] -= CH2P_S_HCSP_H_2;
			b[ipATS] -= b_CH2P_S_HCSP_H;
			
			c[ipATS][ipHCSP] += CH2P_S_HCSP_H_1;
			c[ipCH2P][ipHCSP] += CH2P_S_HCSP_H_2;
			b[ipHCSP] += b_CH2P_S_HCSP_H;

		/*  >>refer   mol  CH2    +  S+    =>   HCS+  +     H     UMIST  */

			c[ipSP][ipCH2] -= CH2_SP_HCSP_H_1;
			c[ipCH2][ipCH2] -= CH2_SP_HCSP_H_2;
			b[ipCH2] -= b_CH2_SP_HCSP_H;
			
			c[ipSP][ipSP] -= CH2_SP_HCSP_H_1;
			c[ipCH2][ipSP] -= CH2_SP_HCSP_H_2;
			b[ipSP] -= b_CH2_SP_HCSP_H;
			
			c[ipSP][ipHCSP] += CH2_SP_HCSP_H_1;
			c[ipCH2][ipHCSP] += CH2_SP_HCSP_H_2;
			b[ipHCSP] += b_CH2_SP_HCSP_H;
			
		/*  >>refer   mol  NH+    +  NH    =>   NH2+  +     N     UMIST  */

			c[ipNH][ipNHP] -= NHP_NH_NH2P_N_1;
			c[ipNHP][ipNHP] -= NHP_NH_NH2P_N_2;
			b[ipNHP] -= b_NHP_NH_NH2P_N;
			
			c[ipNH][ipNH] -= NHP_NH_NH2P_N_1;
			c[ipNHP][ipNH] -= NHP_NH_NH2P_N_2;
			b[ipNH] -= b_NHP_NH_NH2P_N;
			
			c[ipNH][ipNH2P] += NHP_NH_NH2P_N_1;
			c[ipNHP][ipNH2P] += NHP_NH_NH2P_N_2;
			b[ipNH2P] += b_NHP_NH_NH2P_N;
			
			c[ipNH][ipATN] += NHP_NH_NH2P_N_1;
			c[ipNHP][ipATN] += NHP_NH_NH2P_N_2;
			b[ipATN] += b_NHP_NH_NH2P_N;
			
		/*  >>refer   mol  NH+    +  O     =>   OH+   +     N     UMIST  */

			c[ipATO][ipNHP] -= NHP_O_OHP_N_1;
			c[ipNHP][ipNHP] -= NHP_O_OHP_N_2;
			b[ipNHP] -= b_NHP_O_OHP_N;
			
			c[ipATO][ipATO] -= NHP_O_OHP_N_1;
			c[ipNHP][ipATO] -= NHP_O_OHP_N_2;
			b[ipATO] -= b_NHP_O_OHP_N;
			
			c[ipATO][ipOHP] += NHP_O_OHP_N_1;
			c[ipNHP][ipOHP] += NHP_O_OHP_N_2;
			b[ipOHP] += b_NHP_O_OHP_N;

			c[ipATO][ipATN] += NHP_O_OHP_N_1;
			c[ipNHP][ipATN] += NHP_O_OHP_N_2;
			b[ipATN] += b_NHP_O_OHP_N;

		/*  >>refer   mol  NH     +  O+    =>   NO+   +     H     UMIST  */

			c[ipOP][ipNH] -= NH_OP_NOP_H_1;
			c[ipNH][ipNH] -= NH_OP_NOP_H_2;
			b[ipNH] -= b_NH_OP_NOP_H;
			
			c[ipOP][ipOP] -= NH_OP_NOP_H_1;
			c[ipNH][ipOP] -= NH_OP_NOP_H_2;
			b[ipOP] -= b_NH_OP_NOP_H;
			
			c[ipOP][ipNOP] += NH_OP_NOP_H_1;
			c[ipNH][ipNOP] += NH_OP_NOP_H_2;
			b[ipNOP] += b_NH_OP_NOP_H;

		/*  >>refer   mol  NH+    +  NH2   =>   NH3+  +     N     UMIST  */

			c[ipNH2][ipNHP] -= NHP_NH2_NH3P_N_1;
			c[ipNHP][ipNHP] -= NHP_NH2_NH3P_N_2;
			b[ipNHP] -= b_NHP_NH2_NH3P_N;
			
			c[ipNH2][ipNH2] -= NHP_NH2_NH3P_N_1;
			c[ipNHP][ipNH2] -= NHP_NH2_NH3P_N_2;
			b[ipNH2] -= b_NHP_NH2_NH3P_N;
			
			c[ipNH2][ipNH3P] += NHP_NH2_NH3P_N_1;
			c[ipNHP][ipNH3P] += NHP_NH2_NH3P_N_2;
			b[ipNH3P] += b_NHP_NH2_NH3P_N;
			
			c[ipNH2][ipATN] += NHP_NH2_NH3P_N_1;
			c[ipNHP][ipATN] += NHP_NH2_NH3P_N_2;
			b[ipATN] += b_NHP_NH2_NH3P_N;
			
		/*  >>refer   mol  NH     +  NH2+  =>   NH3+  +     N     UMIST  */

			c[ipNH2P][ipNH] -= NH_NH2P_NH3P_N_1;
			c[ipNH][ipNH] -= NH_NH2P_NH3P_N_2;
			b[ipNH] -= b_NH_NH2P_NH3P_N;
			
			c[ipNH2P][ipNH2P] -= NH_NH2P_NH3P_N_1;
			c[ipNH][ipNH2P] -= NH_NH2P_NH3P_N_2;
			b[ipNH2P] -= b_NH_NH2P_NH3P_N;
			
			c[ipNH2P][ipNH3P] += NH_NH2P_NH3P_N_1;
			c[ipNH][ipNH3P] += NH_NH2P_NH3P_N_2;
			b[ipNH3P] += b_NH_NH2P_NH3P_N;
			
			c[ipNH2P][ipATN] += NH_NH2P_NH3P_N_1;
			c[ipNH][ipATN] += NH_NH2P_NH3P_N_2;
			b[ipATN] += b_NH_NH2P_NH3P_N;
			
		/*  >>refer   mol  NH+    +  OH    =>   H2O+  +     N     UMIST  */
			
			c[ipOH][ipNHP] -= NHP_OH_H2OP_N_1;
			c[ipNHP][ipNHP] -= NHP_OH_H2OP_N_2;
			b[ipNHP] -= b_NHP_OH_H2OP_N;

			c[ipOH][ipOH] -= NHP_OH_H2OP_N_1;
			c[ipNHP][ipOH] -= NHP_OH_H2OP_N_2;
			b[ipOH] -= b_NHP_OH_H2OP_N;
			
			c[ipOH][ipH2OP] += NHP_OH_H2OP_N_1;
			c[ipNHP][ipH2OP] += NHP_OH_H2OP_N_2;
			b[ipH2OP] += b_NHP_OH_H2OP_N;
			
			c[ipOH][ipATN] += NHP_OH_H2OP_N_1;
			c[ipNHP][ipATN] += NHP_OH_H2OP_N_2;
			b[ipATN] += b_NHP_OH_H2OP_N;
			
		/*  >>refer   mol  NH     +  OH+   =>   NH2+  +     O     UMIST  */

			c[ipOHP][ipNH] -= NH_OHP_NH2P_O_1;
			c[ipNH][ipNH] -= NH_OHP_NH2P_O_2;
			b[ipNH] -= b_NH_OHP_NH2P_O;
			
			c[ipOHP][ipOHP] -= NH_OHP_NH2P_O_1;
			c[ipNH][ipOHP] -= NH_OHP_NH2P_O_2;
			b[ipOHP] -= b_NH_OHP_NH2P_O;
			
			c[ipOHP][ipNH2P] += NH_OHP_NH2P_O_1;
			c[ipNH][ipNH2P] += NH_OHP_NH2P_O_2;
			b[ipNH2P] += b_NH_OHP_NH2P_O;
			
			c[ipOHP][ipATO] += NH_OHP_NH2P_O_1;
			c[ipNH][ipATO] += NH_OHP_NH2P_O_2;
			b[ipATO] += b_NH_OHP_NH2P_O;
			
		/*  >>refer   mol  NH+    +  NH3   =>   NH4+  +     N     UMIST  */

			c[ipNH3][ipNHP] -= NHP_NH3_NH4P_N_1;
			c[ipNHP][ipNHP] -= NHP_NH3_NH4P_N_2;
			b[ipNHP] -= b_NHP_NH3_NH4P_N;
			
			c[ipNH3][ipNH3] -= NHP_NH3_NH4P_N_1;
			c[ipNHP][ipNH3] -= NHP_NH3_NH4P_N_2;
			b[ipNH3] -= b_NHP_NH3_NH4P_N;
			
			c[ipNH3][ipNH4P] += NHP_NH3_NH4P_N_1;
			c[ipNHP][ipNH4P] += NHP_NH3_NH4P_N_2;
			b[ipNH4P] += b_NHP_NH3_NH4P_N;
			
			c[ipNH3][ipATN] += NHP_NH3_NH4P_N_1;
			c[ipNHP][ipATN] += NHP_NH3_NH4P_N_2;
			b[ipATN] += b_NHP_NH3_NH4P_N;
			
		/*  >>refer   mol  NH     +  NH3+  =>   NH4+  +     N     UMIST  */

			c[ipNH3P][ipNH] -= NH_NH3P_NH4P_N_1;
			c[ipNH][ipNH] -= NH_NH3P_NH4P_N_2;
			b[ipNH] -= b_NH_NH3P_NH4P_N;
			
			c[ipNH3P][ipNH3P] -= NH_NH3P_NH4P_N_1;
			c[ipNH][ipNH3P] -= NH_NH3P_NH4P_N_2;
			b[ipNH3P] -= b_NH_NH3P_NH4P_N;
			
			c[ipNH3P][ipNH4P] += NH_NH3P_NH4P_N_1;
			c[ipNH][ipNH4P] += NH_NH3P_NH4P_N_2;
			b[ipNH4P] += b_NH_NH3P_NH4P_N;
			
			c[ipNH3P][ipATN] += NH_NH3P_NH4P_N_1;
			c[ipNH][ipATN] += NH_NH3P_NH4P_N_2;
			b[ipATN] += b_NH_NH3P_NH4P_N;

		/*  >>refer   mol  NH     +  CH5+  =>   CH4   +     NH2+  UMIST  */

			c[ipCH5P][ipNH] -= NH_CH5P_CH4_NH2P_1;
			c[ipNH][ipNH] -= NH_CH5P_CH4_NH2P_2;
			b[ipNH] -= b_NH_CH5P_CH4_NH2P;
			
			c[ipCH5P][ipCH5P] -= NH_CH5P_CH4_NH2P_1;
			c[ipNH][ipCH5P] -= NH_CH5P_CH4_NH2P_2;
			b[ipCH5P] -= b_NH_CH5P_CH4_NH2P;
			
			c[ipCH5P][ipCH4] += NH_CH5P_CH4_NH2P_1;
			c[ipNH][ipCH4] += NH_CH5P_CH4_NH2P_2;
			b[ipCH4] += b_NH_CH5P_CH4_NH2P;
			
			c[ipCH5P][ipNH2P] += NH_CH5P_CH4_NH2P_1;
			c[ipNH][ipNH2P] += NH_CH5P_CH4_NH2P_2;
			b[ipNH2P] += b_NH_CH5P_CH4_NH2P;

		/*  >>refer   mol  NH+    +  H2O   =>   NH3+  +     O     UMIST  */

			c[ipH2O][ipNHP] -= NHP_H2O_NH3P_O_1;
			c[ipNHP][ipNHP] -= NHP_H2O_NH3P_O_2;
			b[ipNHP] -= b_NHP_H2O_NH3P_O;
			
			c[ipH2O][ipH2O] -= NHP_H2O_NH3P_O_1;
			c[ipNHP][ipH2O] -= NHP_H2O_NH3P_O_2;
			b[ipH2O] -= b_NHP_H2O_NH3P_O;
			
			c[ipH2O][ipNH3P] += NHP_H2O_NH3P_O_1;
			c[ipNHP][ipNH3P] += NHP_H2O_NH3P_O_2;
			b[ipNH3P] += b_NHP_H2O_NH3P_O;		
			
			c[ipH2O][ipATO] += NHP_H2O_NH3P_O_1;
			c[ipNHP][ipATO] += NHP_H2O_NH3P_O_2;
			b[ipATO] += b_NHP_H2O_NH3P_O;		

		/*  >>refer   mol  NH+    +  H2O   =>   H3O+  +     N     UMIST  */

			c[ipH2O][ipNHP] -= NHP_H2O_H3OP_N_1;
			c[ipNHP][ipNHP] -= NHP_H2O_H3OP_N_2;
			b[ipNHP] -= b_NHP_H2O_H3OP_N;
			
			c[ipH2O][ipH2O] -= NHP_H2O_H3OP_N_1;
			c[ipNHP][ipH2O] -= NHP_H2O_H3OP_N_2;
			b[ipH2O] -= b_NHP_H2O_H3OP_N;
			
			c[ipH2O][ipH3OP] += NHP_H2O_H3OP_N_1;
			c[ipNHP][ipH3OP] += NHP_H2O_H3OP_N_2;
			b[ipH3OP] += b_NHP_H2O_H3OP_N;
			
			c[ipH2O][ipATN] += NHP_H2O_H3OP_N_1;
			c[ipNHP][ipATN] += NHP_H2O_H3OP_N_2;
			b[ipATN] += b_NHP_H2O_H3OP_N;

		/*  >>refer   mol  NH+    +  H2O   =>   HNO+  +     H2    UMIST  */

			c[ipH2O][ipNHP] -= NHP_H2O_HNOP_H2_1;
			c[ipNHP][ipNHP] -= NHP_H2O_HNOP_H2_2;
			b[ipNHP] -= b_NHP_H2O_HNOP_H2;
			
			c[ipH2O][ipH2O] -= NHP_H2O_HNOP_H2_1;
			c[ipNHP][ipH2O] -= NHP_H2O_HNOP_H2_2;
			b[ipH2O] -= b_NHP_H2O_HNOP_H2;
			
			c[ipH2O][ipHNOP] += NHP_H2O_HNOP_H2_1;
			c[ipNHP][ipHNOP] += NHP_H2O_HNOP_H2_2;
			b[ipHNOP] += b_NHP_H2O_HNOP_H2;
			
		/*  >>refer   mol  NH     +  H2O+  =>   H3O+  +     N     UMIST  */

			c[ipH2OP][ipNH] -= NH_H2OP_H3OP_N_1;
			c[ipNH][ipNH] -= NH_H2OP_H3OP_N_2;
			b[ipNH] -= b_NH_H2OP_H3OP_N;
			
			c[ipH2OP][ipH2OP] -= NH_H2OP_H3OP_N_1;
			c[ipNH][ipH2OP] -= NH_H2OP_H3OP_N_2;
			b[ipH2OP] -= b_NH_H2OP_H3OP_N;
			
			c[ipH2OP][ipH3OP] += NH_H2OP_H3OP_N_1;
			c[ipNH][ipH3OP] += NH_H2OP_H3OP_N_2;
			b[ipH3OP] += b_NH_H2OP_H3OP_N;
			
			c[ipH2OP][ipATN] += NH_H2OP_H3OP_N_1;
			c[ipNH][ipATN] += NH_H2OP_H3OP_N_2;
			b[ipATN] += b_NH_H2OP_H3OP_N;

		/*  >>refer   mol  NH+    +  H2O   =>   OH    +     NH2+  UMIST  */

			c[ipH2O][ipNHP] -= NHP_H2O_OH_NH2P_1;
			c[ipNHP][ipNHP] -= NHP_H2O_OH_NH2P_2;
			b[ipNHP] -= b_NHP_H2O_OH_NH2P;
			
			c[ipH2O][ipH2O] -= NHP_H2O_OH_NH2P_1;
			c[ipNHP][ipH2O] -= NHP_H2O_OH_NH2P_2;
			b[ipH2O] -= b_NHP_H2O_OH_NH2P;
			
			c[ipH2O][ipOH] += NHP_H2O_OH_NH2P_1;
			c[ipNHP][ipOH] += NHP_H2O_OH_NH2P_2;
			b[ipOH] += b_NHP_H2O_OH_NH2P;
			
			c[ipH2O][ipNH2P] += NHP_H2O_OH_NH2P_1;
			c[ipNHP][ipNH2P] += NHP_H2O_OH_NH2P_2;
			b[ipNH2P] += b_NHP_H2O_OH_NH2P;

		/*  >>refer   mol  NH+    +  CN    =>   HCN+  +     N     UMIST  */

			c[ipCN][ipCN] -= NHP_CN_HCNP_N_1;
			c[ipNHP][ipCN] -= NHP_CN_HCNP_N_2;
			b[ipCN] -= b_NHP_CN_HCNP_N;
			
			c[ipCN][ipNHP] -= NHP_CN_HCNP_N_1;
			c[ipNHP][ipNHP] -= NHP_CN_HCNP_N_2;
			b[ipNHP] -= b_NHP_CN_HCNP_N;
			
			c[ipCN][ipHCNP] += NHP_CN_HCNP_N_1;
			c[ipNHP][ipHCNP] += NHP_CN_HCNP_N_2;
			b[ipHCNP] += b_NHP_CN_HCNP_N;
			
			c[ipCN][ipATN] += NHP_CN_HCNP_N_1;
			c[ipNHP][ipATN] += NHP_CN_HCNP_N_2;
			b[ipATN] += b_NHP_CN_HCNP_N;
			
		/*  >>refer   mol  NH     +  HCN+  =>   CN    +     NH2+  UMIST  */

			c[ipHCNP][ipNH] -= NH_HCNP_CN_NH2P_1;
			c[ipNH][ipNH] -= NH_HCNP_CN_NH2P_2;
			b[ipNH] -= b_NH_HCNP_CN_NH2P;
			
			c[ipHCNP][ipHCNP] -= NH_HCNP_CN_NH2P_1;
			c[ipNH][ipHCNP] -= NH_HCNP_CN_NH2P_2;
			b[ipHCNP] -= b_NH_HCNP_CN_NH2P;
			
			c[ipHCNP][ipCN] += NH_HCNP_CN_NH2P_1;
			c[ipNH][ipCN] += NH_HCNP_CN_NH2P_2;
			b[ipCN] += b_NH_HCNP_CN_NH2P;
			
			c[ipHCNP][ipNH2P] += NH_HCNP_CN_NH2P_1;
			c[ipNH][ipNH2P] += NH_HCNP_CN_NH2P_2;
			b[ipNH2P] += b_NH_HCNP_CN_NH2P;

		/*  >>refer   mol  NH+    +  CO    =>   HCO+  +     N     UMIST  */

			c[ipCO][ipNHP] -= NHP_CO_HCOP_N_1;
			c[ipNHP][ipNHP] -= NHP_CO_HCOP_N_2;
			b[ipNHP] -= b_NHP_CO_HCOP_N;
			
			c[ipCO][ipCO] -= NHP_CO_HCOP_N_1;
			c[ipNHP][ipCO] -= NHP_CO_HCOP_N_2;
			b[ipCO] -= b_NHP_CO_HCOP_N;
			
			c[ipCO][ipHCOP] += NHP_CO_HCOP_N_1;
			c[ipNHP][ipHCOP] += NHP_CO_HCOP_N_2;
			b[ipHCOP] += b_NHP_CO_HCOP_N;
			
			c[ipCO][ipATN] += NHP_CO_HCOP_N_1;
			c[ipNHP][ipATN] += NHP_CO_HCOP_N_2;
			b[ipATN] += b_NHP_CO_HCOP_N;
			
		/*  >>refer   mol  NH     +  CO+   =>   HCO+  +     N     UMIST  */

			c[ipCOP][ipNH] -= NH_COP_HCOP_N_1;
			c[ipNH][ipNH] -= NH_COP_HCOP_N_2;
			b[ipNH] -= b_NH_COP_HCOP_N;
			
			c[ipCOP][ipCOP] -= NH_COP_HCOP_N_1;
			c[ipNH][ipCOP] -= NH_COP_HCOP_N_2;
			b[ipCOP] -= b_NH_COP_HCOP_N;
			
			c[ipCOP][ipHCOP] += NH_COP_HCOP_N_1;
			c[ipNH][ipHCOP] += NH_COP_HCOP_N_2;
			b[ipHCOP] += b_NH_COP_HCOP_N;
			
			c[ipCOP][ipATN] += NH_COP_HCOP_N_1;
			c[ipNH][ipATN] += NH_COP_HCOP_N_2;
			b[ipATN] += b_NH_COP_HCOP_N;
			
		/*  >>refer   mol  NH+    +  CO    =>   OCN+  +     H     UMIST  */

			c[ipCO][ipNHP] -= NHP_CO_OCNP_H_1;
			c[ipNHP][ipNHP] -= NHP_CO_OCNP_H_2;
			b[ipNHP] -= b_NHP_CO_OCNP_H;
			
			c[ipCO][ipCO] -= NHP_CO_OCNP_H_1;
			c[ipNHP][ipCO] -= NHP_CO_OCNP_H_2;
			b[ipCO] -= b_NHP_CO_OCNP_H;
			
			c[ipCO][ipOCNP] += NHP_CO_OCNP_H_1;
			c[ipNHP][ipOCNP] += NHP_CO_OCNP_H_2;
			b[ipOCNP] += b_NHP_CO_OCNP_H;

		/*  >>refer   mol  NH     +  SI+   =>   SIN+  +     H     UMIST  */

			c[ipSIP][ipNH] -= NH_SIP_SINP_H_1;
			c[ipNH][ipNH] -= NH_SIP_SINP_H_2;
			b[ipNH] -= b_NH_SIP_SINP_H;
			
			c[ipSIP][ipSIP] -= NH_SIP_SINP_H_1;
			c[ipNH][ipSIP] -= NH_SIP_SINP_H_2;
			b[ipSIP] -= b_NH_SIP_SINP_H;
			
			c[ipSIP][ipSINP] += NH_SIP_SINP_H_1;
			c[ipNH][ipSINP] += NH_SIP_SINP_H_2;
			b[ipSINP] += b_NH_SIP_SINP_H;
			
		/*  >>refer   mol  NH     +  HCO+  =>   CO    +     NH2+  UMIST  */

			c[ipHCOP][ipNH] -= NH_HCOP_CO_NH2P_1;
			c[ipNH][ipNH] -= NH_HCOP_CO_NH2P_2;
			b[ipNH] -= b_NH_HCOP_CO_NH2P;
			
			c[ipHCOP][ipHCOP] -= NH_HCOP_CO_NH2P_1;
			c[ipNH][ipHCOP] -= NH_HCOP_CO_NH2P_2;
			b[ipHCOP] -= b_NH_HCOP_CO_NH2P;
			
			c[ipHCOP][ipCO] += NH_HCOP_CO_NH2P_1;
			c[ipNH][ipCO] += NH_HCOP_CO_NH2P_2;
			b[ipCO] += b_NH_HCOP_CO_NH2P;
			
			c[ipHCOP][ipNH2P] += NH_HCOP_CO_NH2P_1;
			c[ipNH][ipNH2P] += NH_HCOP_CO_NH2P_2;
			b[ipNH2P] += b_NH_HCOP_CO_NH2P;

		/*  >>refer   mol  NH     +  HNO+  =>   NO    +     NH2+  UMIST  */

			c[ipHNOP][ipNH] -= NH_HNOP_NO_NH2P_1;
			c[ipNH][ipNH] -= NH_HNOP_NO_NH2P_2;
			b[ipNH] -= b_NH_HNOP_NO_NH2P;
			
			c[ipHNOP][ipHNOP] -= NH_HNOP_NO_NH2P_1;
			c[ipNH][ipHNOP] -= NH_HNOP_NO_NH2P_2;
			b[ipHNOP] -= b_NH_HNOP_NO_NH2P;
			
			c[ipHNOP][ipNO] += NH_HNOP_NO_NH2P_1;
			c[ipNH][ipNO] += NH_HNOP_NO_NH2P_2;
			b[ipNO] += b_NH_HNOP_NO_NH2P;
			
			c[ipHNOP][ipNH2P] += NH_HNOP_NO_NH2P_1;
			c[ipNH][ipNH2P] += NH_HNOP_NO_NH2P_2;
			b[ipNH2P] += b_NH_HNOP_NO_NH2P;

		/*  >>refer   mol  NH     +  O2+   =>   HNO+  +     O     UMIST  */
			
			c[ipO2P][ipO2P] -= NH_O2P_HNOP_O_1;
			c[ipNH][ipO2P] -= NH_O2P_HNOP_O_2;
			b[ipO2P] -= b_NH_O2P_HNOP_O;

			c[ipO2P][ipNH] -= NH_O2P_HNOP_O_1;
			c[ipNH][ipNH] -= NH_O2P_HNOP_O_2;
			b[ipNH] -= b_NH_O2P_HNOP_O;
			
			c[ipO2P][ipHNOP] += NH_O2P_HNOP_O_1;
			c[ipNH][ipHNOP] += NH_O2P_HNOP_O_2;
			b[ipHNOP] += b_NH_O2P_HNOP_O;
			
			c[ipO2P][ipATO] += NH_O2P_HNOP_O_1;
			c[ipNH][ipATO] += NH_O2P_HNOP_O_2;
			b[ipATO] += b_NH_O2P_HNOP_O;
			
		/*  >>refer   mol  NH+    +  O2    =>   NO+   +     OH    UMIST  */

			c[ipO2][ipNHP] -= NHP_O2_NOP_OH_1;
			c[ipNHP][ipNHP] -= NHP_O2_NOP_OH_2;
			b[ipNHP] -= b_NHP_O2_NOP_OH;
			
			c[ipO2][ipO2] -= NHP_O2_NOP_OH_1;
			c[ipNHP][ipO2] -= NHP_O2_NOP_OH_2;
			b[ipO2] -= b_NHP_O2_NOP_OH;
			
			c[ipO2][ipNOP] += NHP_O2_NOP_OH_1;
			c[ipNHP][ipNOP] += NHP_O2_NOP_OH_2;
			b[ipNOP] += b_NHP_O2_NOP_OH;
			
			c[ipO2][ipOH] += NHP_O2_NOP_OH_1;
			c[ipNHP][ipOH] += NHP_O2_NOP_OH_2;
			b[ipOH] += b_NHP_O2_NOP_OH;

		/*  >>refer   mol  NH     +  O2+   =>   NO2+  +     H     UMIST  */

			c[ipO2P][ipNH] -= NH_O2P_NO2P_H_1;
			c[ipNH][ipNH] -= NH_O2P_NO2P_H_2;
			b[ipNH] -= b_NH_O2P_NO2P_H;
			
			c[ipO2P][ipO2P] -= NH_O2P_NO2P_H_1;
			c[ipNH][ipO2P] -= NH_O2P_NO2P_H_2;
			b[ipO2P] -= b_NH_O2P_NO2P_H;
			
			c[ipO2P][ipNO2P] += NH_O2P_NO2P_H_1;
			c[ipNH][ipNO2P] += NH_O2P_NO2P_H_2;
			b[ipNO2P] += b_NH_O2P_NO2P_H;
			
		/*  >>refer   mol  NH+    +  S     =>   HS+   +     N     UMIST  */

			c[ipATS][ipNHP] -= NHP_S_HSP_N_1;
			c[ipNHP][ipNHP] -= NHP_S_HSP_N_2;
			b[ipNHP] -= b_NHP_S_HSP_N;
			
			c[ipATS][ipATS] -= NHP_S_HSP_N_1;
			c[ipNHP][ipATS] -= NHP_S_HSP_N_2;
			b[ipATS] -= b_NHP_S_HSP_N;
			
			c[ipATS][ipHSP] += NHP_S_HSP_N_1;
			c[ipNHP][ipHSP] += NHP_S_HSP_N_2;
			b[ipHSP] += b_NHP_S_HSP_N;

			c[ipATS][ipATN] += NHP_S_HSP_N_1;
			c[ipNHP][ipATN] += NHP_S_HSP_N_2;
			b[ipATN] += b_NHP_S_HSP_N;
			
		/*  >>refer   mol  NH+    +  S     =>   NS+   +     H     UMIST  */

			c[ipATS][ipNHP] -= NHP_S_NSP_H_1;
			c[ipNHP][ipNHP] -= NHP_S_NSP_H_2;
			b[ipNHP] -= b_NHP_S_NSP_H;

			c[ipATS][ipATS] -= NHP_S_NSP_H_1;
			c[ipNHP][ipATS] -= NHP_S_NSP_H_2;
			b[ipATS] -= b_NHP_S_NSP_H;
			
			c[ipATS][ipNSP] += NHP_S_NSP_H_1;
			c[ipNHP][ipNSP] += NHP_S_NSP_H_2;
			b[ipNSP] += b_NHP_S_NSP_H;

		/*  >>refer   mol  NH     +  S+    =>   NS+   +     H     UMIST  */

			c[ipSP][ipNH] -= NH_SP_NSP_H_1;
			c[ipNH][ipNH] -= NH_SP_NSP_H_2;
			b[ipNH] -= b_NH_SP_NSP_H;
			
			c[ipSP][ipSP] -= NH_SP_NSP_H_1;
			c[ipNH][ipSP] -= NH_SP_NSP_H_2;
			b[ipSP] -= b_NH_SP_NSP_H;
			
			c[ipSP][ipNSP] += NH_SP_NSP_H_1;
			c[ipNH][ipNSP] += NH_SP_NSP_H_2;
			b[ipNSP] += b_NH_SP_NSP_H;

		/*  >>refer   mol  CH3+   +  NH3   =>   NH4+  +     CH2   UMIST  */

			c[ipNH3][ipCH3P] -= CH3P_NH3_NH4P_CH2_1;
			c[ipCH3P][ipCH3P] -= CH3P_NH3_NH4P_CH2_2;
			b[ipCH3P] -= b_CH3P_NH3_NH4P_CH2;
			
			c[ipNH3][ipNH3] -= CH3P_NH3_NH4P_CH2_1;
			c[ipCH3P][ipNH3] -= CH3P_NH3_NH4P_CH2_2;
			b[ipNH3] -= b_CH3P_NH3_NH4P_CH2;
			
			c[ipNH3][ipNH4P] += CH3P_NH3_NH4P_CH2_1;
			c[ipCH3P][ipNH4P] += CH3P_NH3_NH4P_CH2_2;
			b[ipNH4P] += b_CH3P_NH3_NH4P_CH2;
			
			c[ipNH3][ipCH2] += CH3P_NH3_NH4P_CH2_1;
			c[ipCH3P][ipCH2] += CH3P_NH3_NH4P_CH2_2;
			b[ipCH2] += b_CH3P_NH3_NH4P_CH2;

		/*  >>refer   mol  CH3+   +  S     =>   HCS+  +     H2    UMIST  */

			c[ipATS][ipCH3P] -= CH3P_S_HCSP_H2_1;
			c[ipCH3P][ipCH3P] -= CH3P_S_HCSP_H2_2;
			b[ipCH3P] -= b_CH3P_S_HCSP_H2;

			c[ipATS][ipATS] -= CH3P_S_HCSP_H2_1;
			c[ipCH3P][ipATS] -= CH3P_S_HCSP_H2_2;
			b[ipATS] -= b_CH3P_S_HCSP_H2;
			
			c[ipATS][ipHCSP] += CH3P_S_HCSP_H2_1;
			c[ipCH3P][ipHCSP] += CH3P_S_HCSP_H2_2;
			b[ipHCSP] += b_CH3P_S_HCSP_H2;

		/*  >>refer   mol  O      +  NH2+  =>   HNO+  +     H     UMIST  */

			c[ipNH2P][ipNH2P] -= O_NH2P_HNOP_H_1;
			c[ipATO][ipNH2P] -= O_NH2P_HNOP_H_2;
			b[ipNH2P] -= b_O_NH2P_HNOP_H;

			c[ipNH2P][ipATO] -= O_NH2P_HNOP_H_1;
			c[ipATO][ipATO] -= O_NH2P_HNOP_H_2;
			b[ipATO] -= b_O_NH2P_HNOP_H;
			
			c[ipNH2P][ipHNOP] += O_NH2P_HNOP_H_1;
			c[ipATO][ipHNOP] += O_NH2P_HNOP_H_2;
			b[ipHNOP] += b_O_NH2P_HNOP_H;

		/*  >>refer   mol  O      +  NH3+  =>   HNO+  +     H2    UMIST  */

			c[ipNH3P][ipNH3P] -= O_NH3P_HNOP_H2_1;
			c[ipATO][ipNH3P] -= O_NH3P_HNOP_H2_2;
			b[ipNH3P] -= b_O_NH3P_HNOP_H2;

			c[ipNH3P][ipATO] -= O_NH3P_HNOP_H2_1;
			c[ipATO][ipATO] -= O_NH3P_HNOP_H2_2;
			b[ipATO] -= b_O_NH3P_HNOP_H2;
			
			c[ipNH3P][ipHNOP] += O_NH3P_HNOP_H2_1;
			c[ipATO][ipHNOP] += O_NH3P_HNOP_H2_2;
			b[ipHNOP] += b_O_NH3P_HNOP_H2;

		/*  >>refer   mol  O+     +  CN    =>   NO+   +     C     UMIST  */

			c[ipCN][ipOP] -= OP_CN_NOP_C_1;
			c[ipOP][ipOP] -= OP_CN_NOP_C_2;
			b[ipOP] -= b_OP_CN_NOP_C;
			
			c[ipCN][ipCN] -= OP_CN_NOP_C_1;
			c[ipOP][ipCN] -= OP_CN_NOP_C_2;
			b[ipCN] -= b_OP_CN_NOP_C;
			
			c[ipCN][ipNOP] += OP_CN_NOP_C_1;
			c[ipOP][ipNOP] += OP_CN_NOP_C_2;
			b[ipNOP] += b_OP_CN_NOP_C;
			
			c[ipCN][ipATC] += OP_CN_NOP_C_1;
			c[ipOP][ipATC] += OP_CN_NOP_C_2;
			b[ipATC] += b_OP_CN_NOP_C;

		/*  >>refer   mol  O+     +  HCN   =>   CO+   +     NH    UMIST  */

			c[ipHCN][ipOP] -= OP_HCN_COP_NH_1;
			c[ipOP][ipOP] -= OP_HCN_COP_NH_2;
			b[ipOP] -= b_OP_HCN_COP_NH;
			
			c[ipHCN][ipHCN] -= OP_HCN_COP_NH_1;
			c[ipOP][ipHCN] -= OP_HCN_COP_NH_2;
			b[ipHCN] -= b_OP_HCN_COP_NH;
			
			c[ipHCN][ipCOP] += OP_HCN_COP_NH_1;
			c[ipOP][ipCOP] += OP_HCN_COP_NH_2;
			b[ipCOP] += b_OP_HCN_COP_NH;
			
			c[ipHCN][ipNH] += OP_HCN_COP_NH_1;
			c[ipOP][ipNH] += OP_HCN_COP_NH_2;
			b[ipNH] += b_OP_HCN_COP_NH;

		/*  >>refer   mol  O+     +  HCN   =>   NO+   +     CH    UMIST  */

			c[ipHCN][ipOP] -= OP_HCN_NOP_CH_1;
			c[ipOP][ipOP] -= OP_HCN_NOP_CH_2;
			b[ipOP] -= b_OP_HCN_NOP_CH;
			
			c[ipHCN][ipHCN] -= OP_HCN_NOP_CH_1;
			c[ipOP][ipHCN] -= OP_HCN_NOP_CH_2;
			b[ipHCN] -= b_OP_HCN_NOP_CH;
			
			c[ipHCN][ipNOP] += OP_HCN_NOP_CH_1;
			c[ipOP][ipNOP] += OP_HCN_NOP_CH_2;
			b[ipNOP] += b_OP_HCN_NOP_CH;
			
			c[ipHCN][ipCH] += OP_HCN_NOP_CH_1;
			c[ipOP][ipCH] += OP_HCN_NOP_CH_2;
			b[ipCH] += b_OP_HCN_NOP_CH;

		/*  >>refer   mol  O+     +  HCN   =>   HCO+  +     N     UMIST  */

			c[ipHCN][ipOP] -= OP_HCN_HCOP_N_1;
			c[ipOP][ipOP] -= OP_HCN_HCOP_N_2;
			b[ipOP] -= b_OP_HCN_HCOP_N;
			
			c[ipHCN][ipHCN] -= OP_HCN_HCOP_N_1;
			c[ipOP][ipHCN] -= OP_HCN_HCOP_N_2;
			b[ipHCN] -= b_OP_HCN_HCOP_N;
			
			c[ipHCN][ipHCOP] += OP_HCN_HCOP_N_1;
			c[ipOP][ipHCOP] += OP_HCN_HCOP_N_2;
			b[ipHCOP] += b_OP_HCN_HCOP_N;
			
			c[ipHCN][ipATN] += OP_HCN_HCOP_N_1;
			c[ipOP][ipATN] += OP_HCN_HCOP_N_2;
			b[ipATN] += b_OP_HCN_HCOP_N;

		/*  >>refer   mol  O+     +  N2    =>   NO+   +     N     UMIST  */

			c[ipN2][ipOP] -= OP_N2_NOP_N_1;
			c[ipOP][ipOP] -= OP_N2_NOP_N_2;
			b[ipOP] -= b_OP_N2_NOP_N;
			
			c[ipN2][ipN2] -= OP_N2_NOP_N_1;
			c[ipOP][ipN2] -= OP_N2_NOP_N_2;
			b[ipN2] -= b_OP_N2_NOP_N;
			
			c[ipN2][ipNOP] += OP_N2_NOP_N_1;
			c[ipOP][ipNOP] += OP_N2_NOP_N_2;
			b[ipNOP] += b_OP_N2_NOP_N;
			
			c[ipN2][ipATN] += OP_N2_NOP_N_1;
			c[ipOP][ipATN] += OP_N2_NOP_N_2;
			b[ipATN] += b_OP_N2_NOP_N;
			
		/*  >>refer   mol  O      +  N2+   =>   NO+   +     N     UMIST  */

			c[ipN2P][ipN2P] -= O_N2P_NOP_N_1;
			c[ipATO][ipN2P] -= O_N2P_NOP_N_2;
			b[ipN2P] -= b_O_N2P_NOP_N;
			
			c[ipN2P][ipATO] -= O_N2P_NOP_N_1;
			c[ipATO][ipATO] -= O_N2P_NOP_N_2;
			b[ipATO] -= b_O_N2P_NOP_N;
			
			c[ipN2P][ipNOP] += O_N2P_NOP_N_1;
			c[ipATO][ipNOP] += O_N2P_NOP_N_2;
			b[ipNOP] += b_O_N2P_NOP_N;

			c[ipN2P][ipATN] += O_N2P_NOP_N_1;
			c[ipATO][ipATN] += O_N2P_NOP_N_2;
			b[ipATN] += b_O_N2P_NOP_N;

		/*  >>refer   mol  O      +  HNO+  =>   NO2+  +     H     UMIST  */

			c[ipHNOP][ipHNOP] -= O_HNOP_NO2P_H_1;
			c[ipATO][ipHNOP] -= O_HNOP_NO2P_H_2;
			b[ipHNOP] -= b_O_HNOP_NO2P_H;

			c[ipHNOP][ipATO] -= O_HNOP_NO2P_H_1;
			c[ipATO][ipATO] -= O_HNOP_NO2P_H_2;
			b[ipATO] -= b_O_HNOP_NO2P_H;
			
			c[ipHNOP][ipNO2P] += O_HNOP_NO2P_H_1;
			c[ipATO][ipNO2P] += O_HNOP_NO2P_H_2;
			b[ipNO2P] += b_O_HNOP_NO2P_H;

		/*  >>refer   mol  O      +  HS+   =>   S+    +     OH    UMIST  */

			c[ipHSP][ipHSP] -= O_HSP_SP_OH_1;
			c[ipATO][ipHSP] -= O_HSP_SP_OH_2;
			b[ipHSP] -= b_O_HSP_SP_OH;

			c[ipHSP][ipATO] -= O_HSP_SP_OH_1;
			c[ipATO][ipATO] -= O_HSP_SP_OH_2;
			b[ipATO] -= b_O_HSP_SP_OH;
			
			c[ipHSP][ipSP] += O_HSP_SP_OH_1;
			c[ipATO][ipSP] += O_HSP_SP_OH_2;
			b[ipSP] += b_O_HSP_SP_OH;
			
			c[ipHSP][ipOH] += O_HSP_SP_OH_1;
			c[ipATO][ipOH] += O_HSP_SP_OH_2;
			b[ipOH] += b_O_HSP_SP_OH;

		/*  >>refer   mol  O      +  HS+   =>   SO+   +     H     UMIST  */

			c[ipHSP][ipHSP] -= O_HSP_SOP_H_1;
			c[ipATO][ipHSP] -= O_HSP_SOP_H_2;
			b[ipHSP] -= b_O_HSP_SOP_H;

			c[ipHSP][ipATO] -= O_HSP_SOP_H_1;
			c[ipATO][ipATO] -= O_HSP_SOP_H_2;
			b[ipATO] -= b_O_HSP_SOP_H;
			
			c[ipHSP][ipSOP] += O_HSP_SOP_H_1;
			c[ipATO][ipSOP] += O_HSP_SOP_H_2;
			b[ipSOP] += b_O_HSP_SOP_H;

		/*  >>refer   mol  O      +  SIN+  =>   SIO+  +     N     UMIST  */

			c[ipSINP][ipSINP] -= O_SINP_SIOP_N_1;
			c[ipATO][ipSINP] -= O_SINP_SIOP_N_2;
			b[ipSINP] -= b_O_SINP_SIOP_N;
			
			c[ipSINP][ipATO] -= O_SINP_SIOP_N_1;
			c[ipATO][ipATO] -= O_SINP_SIOP_N_2;
			b[ipATO] -= b_O_SINP_SIOP_N;
			
			c[ipSINP][ipSIOP] += O_SINP_SIOP_N_1;
			c[ipATO][ipSIOP] += O_SINP_SIOP_N_2;
			b[ipSIOP] += b_O_SINP_SIOP_N;

			c[ipSINP][ipATN] += O_SINP_SIOP_N_1;
			c[ipATO][ipATN] += O_SINP_SIOP_N_2;
			b[ipATN] += b_O_SINP_SIOP_N;

		/*  >>refer   mol  O+     +  N2O   =>   NO+   +     NO    UMIST  */

			c[ipN2O][ipOP] -= OP_N2O_NOP_NO_1;
			c[ipOP][ipOP] -= OP_N2O_NOP_NO_2;
			b[ipOP] -= b_OP_N2O_NOP_NO;
			
			c[ipN2O][ipN2O] -= OP_N2O_NOP_NO_1;
			c[ipOP][ipN2O] -= OP_N2O_NOP_NO_2;
			b[ipN2O] -= b_OP_N2O_NOP_NO;
			
			c[ipN2O][ipNOP] += OP_N2O_NOP_NO_1;
			c[ipOP][ipNOP] += OP_N2O_NOP_NO_2;
			b[ipNOP] += b_OP_N2O_NOP_NO;
			
			c[ipN2O][ipNO] += OP_N2O_NOP_NO_1;
			c[ipOP][ipNO] += OP_N2O_NOP_NO_2;
			b[ipNO] += b_OP_N2O_NOP_NO;

		/*  >>refer   mol  O      +  CS+   =>   S     +     CO+   UMIST  */

			c[ipCSP][ipCSP] -= O_CSP_S_COP_1;
			c[ipATO][ipCSP] -= O_CSP_S_COP_2;
			b[ipCSP] -= b_O_CSP_S_COP;
			
			c[ipCSP][ipATO] -= O_CSP_S_COP_1;
			c[ipATO][ipATO] -= O_CSP_S_COP_2;
			b[ipATO] -= b_O_CSP_S_COP;
			
			c[ipCSP][ipCOP] += O_CSP_S_COP_1;
			c[ipATO][ipCOP] += O_CSP_S_COP_2;
			b[ipCOP] += b_O_CSP_S_COP;

			c[ipCSP][ipATS] += O_CSP_S_COP_1;
			c[ipATO][ipATS] += O_CSP_S_COP_2;
			b[ipATS] += b_O_CSP_S_COP;

		/*  >>refer   mol  O      +  HCS+  =>   S     +     HCO+  UMIST  */

			c[ipHCSP][ipHCSP] -= O_HCSP_S_HCOP_1;
			c[ipATO][ipHCSP] -= O_HCSP_S_HCOP_2;
			b[ipHCSP] -= b_O_HCSP_S_HCOP;
			
			c[ipHCSP][ipATO] -= O_HCSP_S_HCOP_1;
			c[ipATO][ipATO] -= O_HCSP_S_HCOP_2;
			b[ipATO] -= b_O_HCSP_S_HCOP;
			
			c[ipHCSP][ipHCOP] += O_HCSP_S_HCOP_1;
			c[ipATO][ipHCOP] += O_HCSP_S_HCOP_2;
			b[ipHCOP] += b_O_HCSP_S_HCOP;

			c[ipHCSP][ipATS] += O_HCSP_S_HCOP_1;
			c[ipATO][ipATS] += O_HCSP_S_HCOP_2;
			b[ipATS] += b_O_HCSP_S_HCOP;

		/*  >>refer   mol  O      +  HCS+  =>   OCS+  +     H     UMIST  */

			c[ipHCSP][ipHCSP] -= O_HCSP_OCSP_H_1;
			c[ipATO][ipHCSP] -= O_HCSP_OCSP_H_2;
			b[ipHCSP] -= b_O_HCSP_OCSP_H;

			c[ipHCSP][ipATO] -= O_HCSP_OCSP_H_1;
			c[ipATO][ipATO] -= O_HCSP_OCSP_H_2;
			b[ipATO] -= b_O_HCSP_OCSP_H;
			
			c[ipHCSP][ipOCSP] += O_HCSP_OCSP_H_1;
			c[ipATO][ipOCSP] += O_HCSP_OCSP_H_2;
			b[ipOCSP] += b_O_HCSP_OCSP_H;

		/*  >>refer   mol  O+     +  NO2   =>   O2    +     NO+   UMIST  */

			c[ipNO2][ipOP] -= OP_NO2_O2_NOP_1;
			c[ipOP][ipOP] -= OP_NO2_O2_NOP_2;
			b[ipOP] -= b_OP_NO2_O2_NOP;
			
			c[ipNO2][ipNO2] -= OP_NO2_O2_NOP_1;
			c[ipOP][ipNO2] -= OP_NO2_O2_NOP_2;
			b[ipNO2] -= b_OP_NO2_O2_NOP;
			
			c[ipNO2][ipO2] += OP_NO2_O2_NOP_1;
			c[ipOP][ipO2] += OP_NO2_O2_NOP_2;
			b[ipO2] += b_OP_NO2_O2_NOP;
			
			c[ipNO2][ipNOP] += OP_NO2_O2_NOP_1;
			c[ipOP][ipNOP] += OP_NO2_O2_NOP_2;
			b[ipNOP] += b_OP_NO2_O2_NOP;

		/*  >>refer   mol  O      +  NS+   =>   S     +     NO+   UMIST  */

			c[ipNSP][ipATO] -= O_NSP_S_NOP_1;
			c[ipATO][ipATO] -= O_NSP_S_NOP_2;
			b[ipATO] -= b_O_NSP_S_NOP;
			
			c[ipNSP][ipNSP] -= O_NSP_S_NOP_1;
			c[ipATO][ipNSP] -= O_NSP_S_NOP_2;
			b[ipNSP] -= b_O_NSP_S_NOP;
			
			c[ipNSP][ipNOP] += O_NSP_S_NOP_1;
			c[ipATO][ipNOP] += O_NSP_S_NOP_2;
			b[ipNOP] += b_O_NSP_S_NOP;

			c[ipNSP][ipATS] += O_NSP_S_NOP_1;
			c[ipATO][ipATS] += O_NSP_S_NOP_2;
			b[ipATS] += b_O_NSP_S_NOP;

		/*  >>refer   mol  NH2+   +  NH2   =>   NH3+  +     NH    UMIST  */

			c[ipNH2][ipNH2P] -= NH2P_NH2_NH3P_NH_1;
			c[ipNH2P][ipNH2P] -= NH2P_NH2_NH3P_NH_2;
			b[ipNH2P] -= b_NH2P_NH2_NH3P_NH;
			
			c[ipNH2][ipNH2] -= NH2P_NH2_NH3P_NH_1;
			c[ipNH2P][ipNH2] -= NH2P_NH2_NH3P_NH_2;
			b[ipNH2] -= b_NH2P_NH2_NH3P_NH;
			
			c[ipNH2][ipNH3P] += NH2P_NH2_NH3P_NH_1;
			c[ipNH2P][ipNH3P] += NH2P_NH2_NH3P_NH_2;
			b[ipNH3P] += b_NH2P_NH2_NH3P_NH;
			
			c[ipNH2][ipNH] += NH2P_NH2_NH3P_NH_1;
			c[ipNH2P][ipNH] += NH2P_NH2_NH3P_NH_2;
			b[ipNH] += b_NH2P_NH2_NH3P_NH;

		/*  >>refer   mol  NH2    +  OH+   =>   NH3+  +     O     UMIST  */

			c[ipOHP][ipNH2] -= NH2_OHP_NH3P_O_1;
			c[ipNH2][ipNH2] -= NH2_OHP_NH3P_O_2;
			b[ipNH2] -= b_NH2_OHP_NH3P_O;
			
			c[ipOHP][ipOHP] -= NH2_OHP_NH3P_O_1;
			c[ipNH2][ipOHP] -= NH2_OHP_NH3P_O_2;
			b[ipOHP] -= b_NH2_OHP_NH3P_O;
			
			c[ipOHP][ipNH3P] += NH2_OHP_NH3P_O_1;
			c[ipNH2][ipNH3P] += NH2_OHP_NH3P_O_2;
			b[ipNH3P] += b_NH2_OHP_NH3P_O;
			
			c[ipOHP][ipATO] += NH2_OHP_NH3P_O_1;
			c[ipNH2][ipATO] += NH2_OHP_NH3P_O_2;
			b[ipATO] += b_NH2_OHP_NH3P_O;

		/*  >>refer   mol  NH2+   +  OH    =>   H2O+  +     NH    UMIST  */

			c[ipOH][ipNH2P] -= NH2P_OH_H2OP_NH_1;
			c[ipNH2P][ipNH2P] -= NH2P_OH_H2OP_NH_2;
			b[ipNH2P] -= b_NH2P_OH_H2OP_NH;
			
			c[ipOH][ipOH] -= NH2P_OH_H2OP_NH_1;
			c[ipNH2P][ipOH] -= NH2P_OH_H2OP_NH_2;
			b[ipOH] -= b_NH2P_OH_H2OP_NH;
						
			c[ipOH][ipH2OP] += NH2P_OH_H2OP_NH_1;
			c[ipNH2P][ipH2OP] += NH2P_OH_H2OP_NH_2;
			b[ipH2OP] += b_NH2P_OH_H2OP_NH;
			
			c[ipOH][ipNH] += NH2P_OH_H2OP_NH_1;
			c[ipNH2P][ipNH] += NH2P_OH_H2OP_NH_2;
			b[ipNH] += b_NH2P_OH_H2OP_NH;

		/*  >>refer   mol  NH2+   +  NH3   =>   NH4+  +     NH    UMIST  */

			c[ipNH3][ipNH2P] -= NH2P_NH3_NH4P_NH_1;
			c[ipNH2P][ipNH2P] -= NH2P_NH3_NH4P_NH_2;
			b[ipNH2P] -= b_NH2P_NH3_NH4P_NH;
			
			c[ipNH3][ipNH3] -= NH2P_NH3_NH4P_NH_1;
			c[ipNH2P][ipNH3] -= NH2P_NH3_NH4P_NH_2;
			b[ipNH3] -= b_NH2P_NH3_NH4P_NH;
			
			c[ipNH3][ipNH4P] += NH2P_NH3_NH4P_NH_1;
			c[ipNH2P][ipNH4P] += NH2P_NH3_NH4P_NH_2;
			b[ipNH4P] += b_NH2P_NH3_NH4P_NH;
			
			c[ipNH3][ipNH] += NH2P_NH3_NH4P_NH_1;
			c[ipNH2P][ipNH] += NH2P_NH3_NH4P_NH_2;
			b[ipNH] += b_NH2P_NH3_NH4P_NH;

		/*  >>refer   mol  NH2    +  NH3+  =>   NH4+  +     NH    UMIST  */

			c[ipNH3P][ipNH2] -= NH2_NH3P_NH4P_NH_1;
			c[ipNH2][ipNH2] -= NH2_NH3P_NH4P_NH_2;
			b[ipNH2] -= b_NH2_NH3P_NH4P_NH;
			
			c[ipNH3P][ipNH3P] -= NH2_NH3P_NH4P_NH_1;
			c[ipNH2][ipNH3P] -= NH2_NH3P_NH4P_NH_2;
			b[ipNH3P] -= b_NH2_NH3P_NH4P_NH;
			
			c[ipNH3P][ipNH4P] += NH2_NH3P_NH4P_NH_1;
			c[ipNH2][ipNH4P] += NH2_NH3P_NH4P_NH_2;
			b[ipNH4P] += b_NH2_NH3P_NH4P_NH;
			
			c[ipNH3P][ipNH] += NH2_NH3P_NH4P_NH_1;
			c[ipNH2][ipNH] += NH2_NH3P_NH4P_NH_2;
			b[ipNH] += b_NH2_NH3P_NH4P_NH;

		/*  >>refer   mol  NH2    +  CH5+  =>   NH3+  +     CH4   UMIST  */

			c[ipCH5P][ipNH2] -= NH2_CH5P_NH3P_CH4_1;
			c[ipNH2][ipNH2] -= NH2_CH5P_NH3P_CH4_2;
			b[ipNH2] -= b_NH2_CH5P_NH3P_CH4;
			
			c[ipCH5P][ipCH5P] -= NH2_CH5P_NH3P_CH4_1;
			c[ipNH2][ipCH5P] -= NH2_CH5P_NH3P_CH4_2;
			b[ipCH5P] -= b_NH2_CH5P_NH3P_CH4;
			
			c[ipCH5P][ipNH3P] += NH2_CH5P_NH3P_CH4_1;
			c[ipNH2][ipNH3P] += NH2_CH5P_NH3P_CH4_2;
			b[ipNH3P] += b_NH2_CH5P_NH3P_CH4;
			
			c[ipCH5P][ipCH4] += NH2_CH5P_NH3P_CH4_1;
			c[ipNH2][ipCH4] += NH2_CH5P_NH3P_CH4_2;
			b[ipCH4] += b_NH2_CH5P_NH3P_CH4;

		/*  >>refer   mol  NH2+   +  H2O   =>   NH3+  +     OH    UMIST  */

			c[ipH2O][ipNH2P] -= NH2P_H2O_NH3P_OH_1;
			c[ipNH2P][ipNH2P] -= NH2P_H2O_NH3P_OH_2;
			b[ipNH2P] -= b_NH2P_H2O_NH3P_OH;
			
			c[ipH2O][ipH2O] -= NH2P_H2O_NH3P_OH_1;
			c[ipNH2P][ipH2O] -= NH2P_H2O_NH3P_OH_2;
			b[ipH2O] -= b_NH2P_H2O_NH3P_OH;
			
			c[ipH2O][ipNH3P] += NH2P_H2O_NH3P_OH_1;
			c[ipNH2P][ipNH3P] += NH2P_H2O_NH3P_OH_2;
			b[ipNH3P] += b_NH2P_H2O_NH3P_OH;
			
			c[ipH2O][ipOH] += NH2P_H2O_NH3P_OH_1;
			c[ipNH2P][ipOH] += NH2P_H2O_NH3P_OH_2;
			b[ipOH] += b_NH2P_H2O_NH3P_OH;

		/*  >>refer   mol  NH2+   +  H2O   =>   H3O+  +     NH    UMIST  */

			c[ipH2O][ipNH2P] -= NH2P_H2O_H3OP_NH_1;
			c[ipNH2P][ipNH2P] -= NH2P_H2O_H3OP_NH_2;
			b[ipNH2P] -= b_NH2P_H2O_H3OP_NH;
			
			c[ipH2O][ipH2O] -= NH2P_H2O_H3OP_NH_1;
			c[ipNH2P][ipH2O] -= NH2P_H2O_H3OP_NH_2;
			b[ipH2O] -= b_NH2P_H2O_H3OP_NH;
			
			c[ipH2O][ipH3OP] += NH2P_H2O_H3OP_NH_1;
			c[ipNH2P][ipH3OP] += NH2P_H2O_H3OP_NH_2;
			b[ipH3OP] += b_NH2P_H2O_H3OP_NH;
			
			c[ipH2O][ipNH] += NH2P_H2O_H3OP_NH_1;
			c[ipNH2P][ipNH] += NH2P_H2O_H3OP_NH_2;
			b[ipNH] += b_NH2P_H2O_H3OP_NH;

		/*  >>refer   mol  NH2    +  H2O+  =>   NH3+  +     OH    UMIST  */

			c[ipH2OP][ipNH2] -= NH2_H2OP_NH3P_OH_1;
			c[ipNH2][ipNH2] -= NH2_H2OP_NH3P_OH_2;
			b[ipNH2] -= b_NH2_H2OP_NH3P_OH;
			
			c[ipH2OP][ipH2OP] -= NH2_H2OP_NH3P_OH_1;
			c[ipNH2][ipH2OP] -= NH2_H2OP_NH3P_OH_2;
			b[ipH2OP] -= b_NH2_H2OP_NH3P_OH;
			
			c[ipH2OP][ipNH3P] += NH2_H2OP_NH3P_OH_1;
			c[ipNH2][ipNH3P] += NH2_H2OP_NH3P_OH_2;
			b[ipNH3P] += b_NH2_H2OP_NH3P_OH;
			
			c[ipH2OP][ipOH] += NH2_H2OP_NH3P_OH_1;
			c[ipNH2][ipOH] += NH2_H2OP_NH3P_OH_2;
			b[ipOH] += b_NH2_H2OP_NH3P_OH;

		/*  >>refer   mol  NH2+   +  H2O   =>   NH4+  +     O     UMIST  */

			c[ipH2O][ipNH2P] -= NH2P_H2O_NH4P_O_1;
			c[ipNH2P][ipNH2P] -= NH2P_H2O_NH4P_O_2;
			b[ipNH2P] -= b_NH2P_H2O_NH4P_O;
			
			c[ipH2O][ipH2O] -= NH2P_H2O_NH4P_O_1;
			c[ipNH2P][ipH2O] -= NH2P_H2O_NH4P_O_2;
			b[ipH2O] -= b_NH2P_H2O_NH4P_O;
			
			c[ipH2O][ipNH4P] += NH2P_H2O_NH4P_O_1;
			c[ipNH2P][ipNH4P] += NH2P_H2O_NH4P_O_2;
			b[ipNH4P] += b_NH2P_H2O_NH4P_O;
			
			c[ipH2O][ipATO] += NH2P_H2O_NH4P_O_1;
			c[ipNH2P][ipATO] += NH2P_H2O_NH4P_O_2;
			b[ipATO] += b_NH2P_H2O_NH4P_O;
			
		/*  >>refer   mol  NH2    +  H3O+  =>   H2O   +     NH3+  UMIST  */

			c[ipH3OP][ipNH2] -= NH2_H3OP_H2O_NH3P_1;
			c[ipNH2][ipNH2] -= NH2_H3OP_H2O_NH3P_2;
			b[ipNH2] -= b_NH2_H3OP_H2O_NH3P;
			
			c[ipH3OP][ipH3OP] -= NH2_H3OP_H2O_NH3P_1;
			c[ipNH2][ipH3OP] -= NH2_H3OP_H2O_NH3P_2;
			b[ipH3OP] -= b_NH2_H3OP_H2O_NH3P;
			
			c[ipH3OP][ipH2O] += NH2_H3OP_H2O_NH3P_1;
			c[ipNH2][ipH2O] += NH2_H3OP_H2O_NH3P_2;
			b[ipH2O] += b_NH2_H3OP_H2O_NH3P;
			
			c[ipH3OP][ipNH3P] += NH2_H3OP_H2O_NH3P_1;
			c[ipNH2][ipNH3P] += NH2_H3OP_H2O_NH3P_2;
			b[ipNH3P] += b_NH2_H3OP_H2O_NH3P;

		/*  >>refer   mol  NH2    +  HCN+  =>   CN    +     NH3+  UMIST  */

			c[ipHCNP][ipNH2] -= NH2_HCNP_CN_NH3P_1;
			c[ipNH2][ipNH2] -= NH2_HCNP_CN_NH3P_2;
			b[ipNH2] -= b_NH2_HCNP_CN_NH3P;
			
			c[ipHCNP][ipHCNP] -= NH2_HCNP_CN_NH3P_1;
			c[ipNH2][ipHCNP] -= NH2_HCNP_CN_NH3P_2;
			b[ipHCNP] -= b_NH2_HCNP_CN_NH3P;
			
			c[ipHCNP][ipCN] += NH2_HCNP_CN_NH3P_1;
			c[ipNH2][ipCN] += NH2_HCNP_CN_NH3P_2;
			b[ipCN] += b_NH2_HCNP_CN_NH3P;
			
			c[ipHCNP][ipNH3P] += NH2_HCNP_CN_NH3P_1;
			c[ipNH2][ipNH3P] += NH2_HCNP_CN_NH3P_2;
			b[ipNH3P] += b_NH2_HCNP_CN_NH3P;

		/*  >>refer   mol  NH2    +  CO+   =>   HCO+  +     NH    UMIST  */

			c[ipCOP][ipNH2] -= NH2_COP_HCOP_NH_1;
			c[ipNH2][ipNH2] -= NH2_COP_HCOP_NH_2;
			b[ipNH2] -= b_NH2_COP_HCOP_NH;
			
			c[ipCOP][ipCOP] -= NH2_COP_HCOP_NH_1;
			c[ipNH2][ipCOP] -= NH2_COP_HCOP_NH_2;
			b[ipCOP] -= b_NH2_COP_HCOP_NH;
			
			c[ipCOP][ipHCOP] += NH2_COP_HCOP_NH_1;
			c[ipNH2][ipHCOP] += NH2_COP_HCOP_NH_2;
			b[ipHCOP] += b_NH2_COP_HCOP_NH;
			
			c[ipCOP][ipNH] += NH2_COP_HCOP_NH_1;
			c[ipNH2][ipNH] += NH2_COP_HCOP_NH_2;
			b[ipNH] += b_NH2_COP_HCOP_NH;

		/*  >>refer   mol  NH2    +  HCO+  =>   CO    +     NH3+  UMIST  */

			c[ipHCOP][ipNH2] -= NH2_HCOP_CO_NH3P_1;
			c[ipNH2][ipNH2] -= NH2_HCOP_CO_NH3P_2;
			b[ipNH2] -= b_NH2_HCOP_CO_NH3P;
			
			c[ipHCOP][ipHCOP] -= NH2_HCOP_CO_NH3P_1;
			c[ipNH2][ipHCOP] -= NH2_HCOP_CO_NH3P_2;
			b[ipHCOP] -= b_NH2_HCOP_CO_NH3P;
			
			c[ipHCOP][ipCO] += NH2_HCOP_CO_NH3P_1;
			c[ipNH2][ipCO] += NH2_HCOP_CO_NH3P_2;
			b[ipCO] += b_NH2_HCOP_CO_NH3P;
			
			c[ipHCOP][ipNH3P] += NH2_HCOP_CO_NH3P_1;
			c[ipNH2][ipNH3P] += NH2_HCOP_CO_NH3P_2;
			b[ipNH3P] += b_NH2_HCOP_CO_NH3P;

		/*  >>refer   mol  NH2    +  HNO+  =>   NO    +     NH3+  UMIST  */

			c[ipHNOP][ipNH2] -= NH2_HNOP_NO_NH3P_1;
			c[ipNH2][ipNH2] -= NH2_HNOP_NO_NH3P_2;
			b[ipNH2] -= b_NH2_HNOP_NO_NH3P;
			
			c[ipHNOP][ipHNOP] -= NH2_HNOP_NO_NH3P_1;
			c[ipNH2][ipHNOP] -= NH2_HNOP_NO_NH3P_2;
			b[ipHNOP] -= b_NH2_HNOP_NO_NH3P;
			
			c[ipHNOP][ipNO] += NH2_HNOP_NO_NH3P_1;
			c[ipNH2][ipNO] += NH2_HNOP_NO_NH3P_2;
			b[ipNO] += b_NH2_HNOP_NO_NH3P;
			
			c[ipHNOP][ipNH3P] += NH2_HNOP_NO_NH3P_1;
			c[ipNH2][ipNH3P] += NH2_HNOP_NO_NH3P_2;
			b[ipNH3P] += b_NH2_HNOP_NO_NH3P;

		/*  >>refer   mol  NH2+   +  O2    =>   HNO+  +     OH    UMIST  */

			c[ipO2][ipNH2P] -= NH2P_O2_HNOP_OH_1;
			c[ipNH2P][ipNH2P] -= NH2P_O2_HNOP_OH_2;
			b[ipNH2P] -= b_NH2P_O2_HNOP_OH;
			
			c[ipO2][ipO2] -= NH2P_O2_HNOP_OH_1;
			c[ipNH2P][ipO2] -= NH2P_O2_HNOP_OH_2;
			b[ipO2] -= b_NH2P_O2_HNOP_OH;
			
			c[ipO2][ipHNOP] += NH2P_O2_HNOP_OH_1;
			c[ipNH2P][ipHNOP] += NH2P_O2_HNOP_OH_2;
			b[ipHNOP] += b_NH2P_O2_HNOP_OH;
			
			c[ipO2][ipOH] += NH2P_O2_HNOP_OH_1;
			c[ipNH2P][ipOH] += NH2P_O2_HNOP_OH_2;
			b[ipOH] += b_NH2P_O2_HNOP_OH;

		/*  >>refer   mol  NH2+   +  S     =>   HS+   +     NH    UMIST  */

			c[ipATS][ipNH2P] -= NH2P_S_HSP_NH_1;
			c[ipNH2P][ipNH2P] -= NH2P_S_HSP_NH_2;
			b[ipNH2P] -= b_NH2P_S_HSP_NH;
			
			c[ipATS][ipATS] -= NH2P_S_HSP_NH_1;
			c[ipNH2P][ipATS] -= NH2P_S_HSP_NH_2;
			b[ipATS] -= b_NH2P_S_HSP_NH;
			
			c[ipATS][ipHSP] += NH2P_S_HSP_NH_1;
			c[ipNH2P][ipHSP] += NH2P_S_HSP_NH_2;
			b[ipHSP] += b_NH2P_S_HSP_NH;
			
			c[ipATS][ipNH] += NH2P_S_HSP_NH_1;
			c[ipNH2P][ipNH] += NH2P_S_HSP_NH_2;
			b[ipNH] += b_NH2P_S_HSP_NH;

		/*  >>refer   mol  CH4+   +  NH3   =>   NH4+  +     CH3   UMIST  */

			c[ipNH3][ipCH4P] -= CH4P_NH3_NH4P_CH3_1;
			c[ipCH4P][ipCH4P] -= CH4P_NH3_NH4P_CH3_2;
			b[ipCH4P] -= b_CH4P_NH3_NH4P_CH3;
			
			c[ipNH3][ipNH3] -= CH4P_NH3_NH4P_CH3_1;
			c[ipCH4P][ipNH3] -= CH4P_NH3_NH4P_CH3_2;
			b[ipNH3] -= b_CH4P_NH3_NH4P_CH3;
			
			c[ipNH3][ipNH4P] += CH4P_NH3_NH4P_CH3_1;
			c[ipCH4P][ipNH4P] += CH4P_NH3_NH4P_CH3_2;
			b[ipNH4P] += b_CH4P_NH3_NH4P_CH3;
			
			c[ipNH3][ipCH3] += CH4P_NH3_NH4P_CH3_1;
			c[ipCH4P][ipCH3] += CH4P_NH3_NH4P_CH3_2;
			b[ipCH3] += b_CH4P_NH3_NH4P_CH3;

		/*  >>refer   mol  CH4    +  NH3+  =>   NH4+  +     CH3   UMIST  */

			c[ipNH3P][ipCH4] -= CH4_NH3P_NH4P_CH3_1;
			c[ipCH4][ipCH4] -= CH4_NH3P_NH4P_CH3_2;
			b[ipCH4] -= b_CH4_NH3P_NH4P_CH3;
			
			c[ipNH3P][ipNH3P] -= CH4_NH3P_NH4P_CH3_1;
			c[ipCH4][ipNH3P] -= CH4_NH3P_NH4P_CH3_2;
			b[ipNH3P] -= b_CH4_NH3P_NH4P_CH3;
			
			c[ipNH3P][ipNH4P] += CH4_NH3P_NH4P_CH3_1;
			c[ipCH4][ipNH4P] += CH4_NH3P_NH4P_CH3_2;
			b[ipNH4P] += b_CH4_NH3P_NH4P_CH3;
			
			c[ipNH3P][ipCH3] += CH4_NH3P_NH4P_CH3_1;
			c[ipCH4][ipCH3] += CH4_NH3P_NH4P_CH3_2;
			b[ipCH3] += b_CH4_NH3P_NH4P_CH3;

		/*  >>refer   mol  CH4    +  N2+   =>   N2    +     CH2+ + H2    UMIST  */

			c[ipN2P][ipCH4] -= CH4_N2P_N2_CH2P_H2_1;
			c[ipCH4][ipCH4] -= CH4_N2P_N2_CH2P_H2_2;
			b[ipCH4] -= b_CH4_N2P_N2_CH2P_H2;
			
			c[ipN2P][ipN2P] -= CH4_N2P_N2_CH2P_H2_1;
			c[ipCH4][ipN2P] -= CH4_N2P_N2_CH2P_H2_2;
			b[ipN2P] -= b_CH4_N2P_N2_CH2P_H2;
			
			c[ipN2P][ipN2] += CH4_N2P_N2_CH2P_H2_1;
			c[ipCH4][ipN2] += CH4_N2P_N2_CH2P_H2_2;
			b[ipN2] += b_CH4_N2P_N2_CH2P_H2;
			
			c[ipN2P][ipCH2P] += CH4_N2P_N2_CH2P_H2_1;
			c[ipCH4][ipCH2P] += CH4_N2P_N2_CH2P_H2_2;
			b[ipCH2P] += b_CH4_N2P_N2_CH2P_H2;

		/*  >>refer   mol  CH4    +  N2+   =>   N2    +     CH3+ + H     UMIST  */

			c[ipN2P][ipCH4] -= CH4_N2P_N2_CH3P_H_1;
			c[ipCH4][ipCH4] -= CH4_N2P_N2_CH3P_H_2;
			b[ipCH4] -= b_CH4_N2P_N2_CH3P_H;
			
			c[ipN2P][ipN2P] -= CH4_N2P_N2_CH3P_H_1;
			c[ipCH4][ipN2P] -= CH4_N2P_N2_CH3P_H_2;
			b[ipN2P] -= b_CH4_N2P_N2_CH3P_H;
			
			c[ipN2P][ipN2] += CH4_N2P_N2_CH3P_H_1;
			c[ipCH4][ipN2] += CH4_N2P_N2_CH3P_H_2;
			b[ipN2] += b_CH4_N2P_N2_CH3P_H;
			
			c[ipN2P][ipCH3P] += CH4_N2P_N2_CH3P_H_1;
			c[ipCH4][ipCH3P] += CH4_N2P_N2_CH3P_H_2;
			b[ipCH3P] += b_CH4_N2P_N2_CH3P_H;

		/*  >>refer   mol  CH4    +  HNO+  =>   NO    +     CH5+  UMIST  */

			c[ipHNOP][ipCH4] -= CH4_HNOP_NO_CH5P_1;
			c[ipCH4][ipCH4] -= CH4_HNOP_NO_CH5P_2;
			b[ipCH4] -= b_CH4_HNOP_NO_CH5P;
			
			c[ipHNOP][ipHNOP] -= CH4_HNOP_NO_CH5P_1;
			c[ipCH4][ipHNOP] -= CH4_HNOP_NO_CH5P_2;
			b[ipHNOP] -= b_CH4_HNOP_NO_CH5P;
			
			c[ipHNOP][ipNO] += CH4_HNOP_NO_CH5P_1;
			c[ipCH4][ipNO] += CH4_HNOP_NO_CH5P_2;
			b[ipNO] += b_CH4_HNOP_NO_CH5P;
			
			c[ipHNOP][ipCH5P] += CH4_HNOP_NO_CH5P_1;
			c[ipCH4][ipCH5P] += CH4_HNOP_NO_CH5P_2;
			b[ipCH5P] += b_CH4_HNOP_NO_CH5P;

		/*  >>refer   mol  CH4    +  S+    =>   HCS+  +     H2   + H     UMIST  */

			c[ipSP][ipCH4] -= CH4_SP_HCSP_H2_H_1;
			c[ipCH4][ipCH4] -= CH4_SP_HCSP_H2_H_2;
			b[ipCH4] -= b_CH4_SP_HCSP_H2_H;
			
			c[ipSP][ipSP] -= CH4_SP_HCSP_H2_H_1;
			c[ipCH4][ipSP] -= CH4_SP_HCSP_H2_H_2;
			b[ipSP] -= b_CH4_SP_HCSP_H2_H;
			
			c[ipSP][ipHCSP] += CH4_SP_HCSP_H2_H_1;
			c[ipCH4][ipHCSP] += CH4_SP_HCSP_H2_H_2;
			b[ipHCSP] += b_CH4_SP_HCSP_H2_H;
			
		/*  >>refer   mol  CH4    +  CS+   =>   HCS+  +     CH3   UMIST  */

			c[ipCSP][ipCH4] -= CH4_CSP_HCSP_CH3_1;
			c[ipCH4][ipCH4] -= CH4_CSP_HCSP_CH3_2;
			b[ipCH4] -= b_CH4_CSP_HCSP_CH3;
			
			c[ipCSP][ipCSP] -= CH4_CSP_HCSP_CH3_1;
			c[ipCH4][ipCSP] -= CH4_CSP_HCSP_CH3_2;
			b[ipCSP] -= b_CH4_CSP_HCSP_CH3;
			
			c[ipCSP][ipHCSP] += CH4_CSP_HCSP_CH3_1;
			c[ipCH4][ipHCSP] += CH4_CSP_HCSP_CH3_2;
			b[ipHCSP] += b_CH4_CSP_HCSP_CH3;
			
			c[ipCSP][ipCH3] += CH4_CSP_HCSP_CH3_1;
			c[ipCH4][ipCH3] += CH4_CSP_HCSP_CH3_2;
			b[ipCH3] += b_CH4_CSP_HCSP_CH3;

		/*  >>refer   mol  OH+    +  NH3   =>   NH4+  +     O     UMIST  */

			c[ipNH3][ipOHP] -= OHP_NH3_NH4P_O_1;
			c[ipOHP][ipOHP] -= OHP_NH3_NH4P_O_2;
			b[ipOHP] -= b_OHP_NH3_NH4P_O;
			
			c[ipNH3][ipNH3] -= OHP_NH3_NH4P_O_1;
			c[ipOHP][ipNH3] -= OHP_NH3_NH4P_O_2;
			b[ipNH3] -= b_OHP_NH3_NH4P_O;
			
			c[ipNH3][ipNH4P] += OHP_NH3_NH4P_O_1;
			c[ipOHP][ipNH4P] += OHP_NH3_NH4P_O_2;
			b[ipNH4P] += b_OHP_NH3_NH4P_O;
			
			c[ipNH3][ipATO] += OHP_NH3_NH4P_O_1;
			c[ipOHP][ipATO] += OHP_NH3_NH4P_O_2;
			b[ipATO] += b_OHP_NH3_NH4P_O;
			
		/*  >>refer   mol  OH     +  NH3+  =>   NH4+  +     O     UMIST  */

			c[ipNH3P][ipOH] -= OH_NH3P_NH4P_O_1;
			c[ipOH][ipOH] -= OH_NH3P_NH4P_O_2;
			b[ipOH] -= b_OH_NH3P_NH4P_O;
			
			c[ipNH3P][ipNH3P] -= OH_NH3P_NH4P_O_1;
			c[ipOH][ipNH3P] -= OH_NH3P_NH4P_O_2;
			b[ipNH3P] -= b_OH_NH3P_NH4P_O;
			
			c[ipNH3P][ipNH4P] += OH_NH3P_NH4P_O_1;
			c[ipOH][ipNH4P] += OH_NH3P_NH4P_O_2;
			b[ipNH4P] += b_OH_NH3P_NH4P_O;
			
			c[ipNH3P][ipATO] += OH_NH3P_NH4P_O_1;
			c[ipOH][ipATO] += OH_NH3P_NH4P_O_2;
			b[ipATO] += b_OH_NH3P_NH4P_O;

		/*  >>refer   mol  OH+    +  CN    =>   HCN+  +     O     UMIST  */

			c[ipCN][ipOHP] -= OHP_CN_HCNP_O_1;
			c[ipOHP][ipOHP] -= OHP_CN_HCNP_O_2;
			b[ipOHP] -= b_OHP_CN_HCNP_O;
			
			c[ipCN][ipCN] -= OHP_CN_HCNP_O_1;
			c[ipOHP][ipCN] -= OHP_CN_HCNP_O_2;
			b[ipCN] -= b_OHP_CN_HCNP_O;
			
			c[ipCN][ipHCNP] += OHP_CN_HCNP_O_1;
			c[ipOHP][ipHCNP] += OHP_CN_HCNP_O_2;
			b[ipHCNP] += b_OHP_CN_HCNP_O;
			
			c[ipCN][ipATO] += OHP_CN_HCNP_O_1;
			c[ipOHP][ipATO] += OHP_CN_HCNP_O_2;
			b[ipATO] += b_OHP_CN_HCNP_O;

		/*  >>refer   mol  OH     +  HCN+  =>   CN    +     H2O+  UMIST  */

			c[ipHCNP][ipOH] -= OH_HCNP_CN_H2OP_1;
			c[ipOH][ipOH] -= OH_HCNP_CN_H2OP_2;
			b[ipOH] -= b_OH_HCNP_CN_H2OP;
			
			c[ipHCNP][ipHCNP] -= OH_HCNP_CN_H2OP_1;
			c[ipOH][ipHCNP] -= OH_HCNP_CN_H2OP_2;
			b[ipHCNP] -= b_OH_HCNP_CN_H2OP;
			
			c[ipHCNP][ipCN] += OH_HCNP_CN_H2OP_1;
			c[ipOH][ipCN] += OH_HCNP_CN_H2OP_2;
			b[ipCN] += b_OH_HCNP_CN_H2OP;
			
			c[ipHCNP][ipH2OP] += OH_HCNP_CN_H2OP_1;
			c[ipOH][ipH2OP] += OH_HCNP_CN_H2OP_2;
			b[ipH2OP] += b_OH_HCNP_CN_H2OP;

		/*  >>refer   mol  OH+    +  NO    =>   HNO+  +     O     UMIST  */

			c[ipNO][ipOHP] -= OHP_NO_HNOP_O_1;
			c[ipOHP][ipOHP] -= OHP_NO_HNOP_O_2;
			b[ipOHP] -= b_OHP_NO_HNOP_O;
			
			c[ipNO][ipNO] -= OHP_NO_HNOP_O_1;
			c[ipOHP][ipNO] -= OHP_NO_HNOP_O_2;
			b[ipNO] -= b_OHP_NO_HNOP_O;
			
			c[ipNO][ipHNOP] += OHP_NO_HNOP_O_1;
			c[ipOHP][ipHNOP] += OHP_NO_HNOP_O_2;
			b[ipHNOP] += b_OHP_NO_HNOP_O;
			
			c[ipNO][ipATO] += OHP_NO_HNOP_O_1;
			c[ipOHP][ipATO] += OHP_NO_HNOP_O_2;
			b[ipATO] += b_OHP_NO_HNOP_O;
			
		/*  >>refer   mol  OH     +  HNO+  =>   NO    +     H2O+  UMIST  */

			c[ipHNOP][ipOH] -= OH_HNOP_NO_H2OP_1;
			c[ipOH][ipOH] -= OH_HNOP_NO_H2OP_2;
			b[ipOH] -= b_OH_HNOP_NO_H2OP;
			
			c[ipHNOP][ipHNOP] -= OH_HNOP_NO_H2OP_1;
			c[ipOH][ipHNOP] -= OH_HNOP_NO_H2OP_2;
			b[ipHNOP] -= b_OH_HNOP_NO_H2OP;
			
			c[ipHNOP][ipNO] += OH_HNOP_NO_H2OP_1;
			c[ipOH][ipNO] += OH_HNOP_NO_H2OP_2;
			b[ipNO] += b_OH_HNOP_NO_H2OP;
			
			c[ipHNOP][ipH2OP] += OH_HNOP_NO_H2OP_1;
			c[ipOH][ipH2OP] += OH_HNOP_NO_H2OP_2;
			b[ipH2OP] += b_OH_HNOP_NO_H2OP;

		/*  >>refer   mol  OH+    +  S     =>   HS+   +     O     UMIST  */

			c[ipATS][ipOHP] -= OHP_S_HSP_O_1;
			c[ipOHP][ipOHP] -= OHP_S_HSP_O_2;
			b[ipOHP] -= b_OHP_S_HSP_O;

			c[ipATS][ipATS] -= OHP_S_HSP_O_1;
			c[ipOHP][ipATS] -= OHP_S_HSP_O_2;
			b[ipATS] -= b_OHP_S_HSP_O;
			
			c[ipATS][ipHSP] += OHP_S_HSP_O_1;
			c[ipOHP][ipHSP] += OHP_S_HSP_O_2;
			b[ipHSP] += b_OHP_S_HSP_O;
			
			c[ipATS][ipATO] += OHP_S_HSP_O_1;
			c[ipOHP][ipATO] += OHP_S_HSP_O_2;
			b[ipATO] += b_OHP_S_HSP_O;

		/*  >>refer   mol  OH+    +  S     =>   SO+   +     H     UMIST  */

			c[ipATS][ipOHP] -= OHP_S_SOP_H_1;
			c[ipOHP][ipOHP] -= OHP_S_SOP_H_2;
			b[ipOHP] -= b_OHP_S_SOP_H;

			c[ipATS][ipATS] -= OHP_S_SOP_H_1;
			c[ipOHP][ipATS] -= OHP_S_SOP_H_2;
			b[ipATS] -= b_OHP_S_SOP_H;
			
			c[ipATS][ipSOP] += OHP_S_SOP_H_1;
			c[ipOHP][ipSOP] += OHP_S_SOP_H_2;
			b[ipSOP] += b_OHP_S_SOP_H;

		/*  >>refer   mol  OH     +  S+    =>   SO+   +     H     UMIST  */

			c[ipSP][ipOH] -= OH_SP_SOP_H_1;
			c[ipOH][ipOH] -= OH_SP_SOP_H_2;
			b[ipOH] -= b_OH_SP_SOP_H;
			
			c[ipSP][ipSP] -= OH_SP_SOP_H_1;
			c[ipOH][ipSP] -= OH_SP_SOP_H_2;
			b[ipSP] -= b_OH_SP_SOP_H;
			
			c[ipSP][ipSOP] += OH_SP_SOP_H_1;
			c[ipOH][ipSOP] += OH_SP_SOP_H_2;
			b[ipSOP] += b_OH_SP_SOP_H;
			
		/*  >>refer   mol  NH3+   +  NH3   =>   NH4+  +     NH2   UMIST  */

			c[ipNH3][ipNH3P] -= NH3P_NH3_NH4P_NH2_1;
			c[ipNH3P][ipNH3P] -= NH3P_NH3_NH4P_NH2_2;
			b[ipNH3P] -= b_NH3P_NH3_NH4P_NH2;
			
			c[ipNH3][ipNH3] -= NH3P_NH3_NH4P_NH2_1;
			c[ipNH3P][ipNH3] -= NH3P_NH3_NH4P_NH2_2;
			b[ipNH3] -= b_NH3P_NH3_NH4P_NH2;
			
			c[ipNH3][ipNH4P] += NH3P_NH3_NH4P_NH2_1;
			c[ipNH3P][ipNH4P] += NH3P_NH3_NH4P_NH2_2;
			b[ipNH4P] += b_NH3P_NH3_NH4P_NH2;
			
			c[ipNH3][ipNH2] += NH3P_NH3_NH4P_NH2_1;
			c[ipNH3P][ipNH2] += NH3P_NH3_NH4P_NH2_2;
			b[ipNH2] += b_NH3P_NH3_NH4P_NH2;

		/*  >>refer   mol  NH3    +  CH5+  =>   NH4+  +     CH4   UMIST  */

			c[ipCH5P][ipNH3] -= NH3_CH5P_NH4P_CH4_1;
			c[ipNH3][ipNH3] -= NH3_CH5P_NH4P_CH4_2;
			b[ipNH3] -= b_NH3_CH5P_NH4P_CH4;
			
			c[ipCH5P][ipCH5P] -= NH3_CH5P_NH4P_CH4_1;
			c[ipNH3][ipCH5P] -= NH3_CH5P_NH4P_CH4_2;
			b[ipCH5P] -= b_NH3_CH5P_NH4P_CH4;
			
			c[ipCH5P][ipNH4P] += NH3_CH5P_NH4P_CH4_1;
			c[ipNH3][ipNH4P] += NH3_CH5P_NH4P_CH4_2;
			b[ipNH4P] += b_NH3_CH5P_NH4P_CH4;
			
			c[ipCH5P][ipCH4] += NH3_CH5P_NH4P_CH4_1;
			c[ipNH3][ipCH4] += NH3_CH5P_NH4P_CH4_2;
			b[ipCH4] += b_NH3_CH5P_NH4P_CH4;

		/*  >>refer   mol  NH3+   +  H2O   =>   NH4+  +     OH    UMIST  */

			c[ipH2O][ipNH3P] -= NH3P_H2O_NH4P_OH_1;
			c[ipNH3P][ipNH3P] -= NH3P_H2O_NH4P_OH_2;
			b[ipNH3P] -= b_NH3P_H2O_NH4P_OH;
			
			c[ipH2O][ipH2O] -= NH3P_H2O_NH4P_OH_1;
			c[ipNH3P][ipH2O] -= NH3P_H2O_NH4P_OH_2;
			b[ipH2O] -= b_NH3P_H2O_NH4P_OH;
			
			c[ipH2O][ipNH4P] += NH3P_H2O_NH4P_OH_1;
			c[ipNH3P][ipNH4P] += NH3P_H2O_NH4P_OH_2;
			b[ipNH4P] += b_NH3P_H2O_NH4P_OH;
			
			c[ipH2O][ipOH] += NH3P_H2O_NH4P_OH_1;
			c[ipNH3P][ipOH] += NH3P_H2O_NH4P_OH_2;
			b[ipOH] += b_NH3P_H2O_NH4P_OH;

		/*  >>refer   mol  NH3    +  H2O+  =>   NH4+  +     OH    UMIST  */

			c[ipH2OP][ipNH3] -= NH3_H2OP_NH4P_OH_1;
			c[ipNH3][ipNH3] -= NH3_H2OP_NH4P_OH_2;
			b[ipNH3] -= b_NH3_H2OP_NH4P_OH;
			
			c[ipH2OP][ipH2OP] -= NH3_H2OP_NH4P_OH_1;
			c[ipNH3][ipH2OP] -= NH3_H2OP_NH4P_OH_2;
			b[ipH2OP] -= b_NH3_H2OP_NH4P_OH;
			
			c[ipH2OP][ipNH4P] += NH3_H2OP_NH4P_OH_1;
			c[ipNH3][ipNH4P] += NH3_H2OP_NH4P_OH_2;
			b[ipNH4P] += b_NH3_H2OP_NH4P_OH;
			
			c[ipH2OP][ipOH] += NH3_H2OP_NH4P_OH_1;
			c[ipNH3][ipOH] += NH3_H2OP_NH4P_OH_2;
			b[ipOH] += b_NH3_H2OP_NH4P_OH;

		/*  >>refer   mol  NH3    +  H3O+  =>   NH4+  +     H2O   UMIST  */

			c[ipH3OP][ipNH3] -= NH3_H3OP_NH4P_H2O_1;
			c[ipNH3][ipNH3] -= NH3_H3OP_NH4P_H2O_2;
			b[ipNH3] -= b_NH3_H3OP_NH4P_H2O;
			
			c[ipH3OP][ipH3OP] -= NH3_H3OP_NH4P_H2O_1;
			c[ipNH3][ipH3OP] -= NH3_H3OP_NH4P_H2O_2;
			b[ipH3OP] -= b_NH3_H3OP_NH4P_H2O;
			
			c[ipH3OP][ipNH4P] += NH3_H3OP_NH4P_H2O_1;
			c[ipNH3][ipNH4P] += NH3_H3OP_NH4P_H2O_2;
			b[ipNH4P] += b_NH3_H3OP_NH4P_H2O;
			
			c[ipH3OP][ipH2O] += NH3_H3OP_NH4P_H2O_1;
			c[ipNH3][ipH2O] += NH3_H3OP_NH4P_H2O_2;
			b[ipH2O] += b_NH3_H3OP_NH4P_H2O;

		/*  >>refer   mol  NH3    +  CO+   =>   HCO+  +     NH2   UMIST  */

			c[ipCOP][ipNH3] -= NH3_COP_HCOP_NH2_1;
			c[ipNH3][ipNH3] -= NH3_COP_HCOP_NH2_2;
			b[ipNH3] -= b_NH3_COP_HCOP_NH2;
			
			c[ipCOP][ipCOP] -= NH3_COP_HCOP_NH2_1;
			c[ipNH3][ipCOP] -= NH3_COP_HCOP_NH2_2;
			b[ipCOP] -= b_NH3_COP_HCOP_NH2;
			
			c[ipCOP][ipHCOP] += NH3_COP_HCOP_NH2_1;
			c[ipNH3][ipHCOP] += NH3_COP_HCOP_NH2_2;
			b[ipHCOP] += b_NH3_COP_HCOP_NH2;
			
			c[ipCOP][ipNH2] += NH3_COP_HCOP_NH2_1;
			c[ipNH3][ipNH2] += NH3_COP_HCOP_NH2_2;
			b[ipNH2] += b_NH3_COP_HCOP_NH2;

		/*  >>refer   mol  NH3    +  HCO+  =>   CO    +     NH4+  UMIST  */

			c[ipHCOP][ipNH3] -= NH3_HCOP_CO_NH4P_1;
			c[ipNH3][ipNH3] -= NH3_HCOP_CO_NH4P_2;
			b[ipNH3] -= b_NH3_HCOP_CO_NH4P;
			
			c[ipHCOP][ipHCOP] -= NH3_HCOP_CO_NH4P_1;
			c[ipNH3][ipHCOP] -= NH3_HCOP_CO_NH4P_2;
			b[ipHCOP] -= b_NH3_HCOP_CO_NH4P;
			
			c[ipHCOP][ipCO] += NH3_HCOP_CO_NH4P_1;
			c[ipNH3][ipCO] += NH3_HCOP_CO_NH4P_2;
			b[ipCO] += b_NH3_HCOP_CO_NH4P;
			
			c[ipHCOP][ipNH4P] += NH3_HCOP_CO_NH4P_1;
			c[ipNH3][ipNH4P] += NH3_HCOP_CO_NH4P_2;
			b[ipNH4P] += b_NH3_HCOP_CO_NH4P;

		/*  >>refer   mol  NH3    +  HNO+  =>   NO    +     NH4+  UMIST  */

			c[ipHNOP][ipNH3] -= NH3_HNOP_NO_NH4P_1;
			c[ipNH3][ipNH3] -= NH3_HNOP_NO_NH4P_2;
			b[ipNH3] -= b_NH3_HNOP_NO_NH4P;
			
			c[ipHNOP][ipHNOP] -= NH3_HNOP_NO_NH4P_1;
			c[ipNH3][ipHNOP] -= NH3_HNOP_NO_NH4P_2;
			b[ipHNOP] -= b_NH3_HNOP_NO_NH4P;
			
			c[ipHNOP][ipNO] += NH3_HNOP_NO_NH4P_1;
			c[ipNH3][ipNO] += NH3_HNOP_NO_NH4P_2;
			b[ipNO] += b_NH3_HNOP_NO_NH4P;
			
			c[ipHNOP][ipNH4P] += NH3_HNOP_NO_NH4P_1;
			c[ipNH3][ipNH4P] += NH3_HNOP_NO_NH4P_2;
			b[ipNH4P] += b_NH3_HNOP_NO_NH4P;

		/*  >>refer   mol  NH3    +  HS+   =>   S     +     NH4+  UMIST  */

			c[ipHSP][ipNH3] -= NH3_HSP_S_NH4P_1;
			c[ipNH3][ipNH3] -= NH3_HSP_S_NH4P_2;
			b[ipNH3] -= b_NH3_HSP_S_NH4P;
			
			c[ipHSP][ipHSP] -= NH3_HSP_S_NH4P_1;
			c[ipNH3][ipHSP] -= NH3_HSP_S_NH4P_2;
			b[ipHSP] -= b_NH3_HSP_S_NH4P;
			
			c[ipHSP][ipNH4P] += NH3_HSP_S_NH4P_1;
			c[ipNH3][ipNH4P] += NH3_HSP_S_NH4P_2;
			b[ipNH4P] += b_NH3_HSP_S_NH4P;
			
			c[ipHSP][ipATS] += NH3_HSP_S_NH4P_1;
			c[ipNH3][ipATS] += NH3_HSP_S_NH4P_2;
			b[ipATS] += b_NH3_HSP_S_NH4P;

		/*  >>refer   mol  NH3    +  HCS+  =>   CS    +     NH4+  UMIST  */

			c[ipHCSP][ipNH3] -= NH3_HCSP_CS_NH4P_1;
			c[ipNH3][ipNH3] -= NH3_HCSP_CS_NH4P_2;
			b[ipNH3] -= b_NH3_HCSP_CS_NH4P;
			
			c[ipHCSP][ipHCSP] -= NH3_HCSP_CS_NH4P_1;
			c[ipNH3][ipHCSP] -= NH3_HCSP_CS_NH4P_2;
			b[ipHCSP] -= b_NH3_HCSP_CS_NH4P;
			
			c[ipHCSP][ipCS] += NH3_HCSP_CS_NH4P_1;
			c[ipNH3][ipCS] += NH3_HCSP_CS_NH4P_2;
			b[ipCS] += b_NH3_HCSP_CS_NH4P;
			
			c[ipHCSP][ipNH4P] += NH3_HCSP_CS_NH4P_1;
			c[ipNH3][ipNH4P] += NH3_HCSP_CS_NH4P_2;
			b[ipNH4P] += b_NH3_HCSP_CS_NH4P;

		/*  >>refer   mol  CH5+   +  S     =>   HS+   +     CH4   UMIST  */

			c[ipATS][ipCH5P] -= CH5P_S_HSP_CH4_1;
			c[ipCH5P][ipCH5P] -= CH5P_S_HSP_CH4_2;
			b[ipCH5P] -= b_CH5P_S_HSP_CH4;

			c[ipATS][ipATS] -= CH5P_S_HSP_CH4_1;
			c[ipCH5P][ipATS] -= CH5P_S_HSP_CH4_2;
			b[ipATS] -= b_CH5P_S_HSP_CH4;
			
			c[ipATS][ipHSP] += CH5P_S_HSP_CH4_1;
			c[ipCH5P][ipHSP] += CH5P_S_HSP_CH4_2;
			b[ipHSP] += b_CH5P_S_HSP_CH4;
			
			c[ipATS][ipCH4] += CH5P_S_HSP_CH4_1;
			c[ipCH5P][ipCH4] += CH5P_S_HSP_CH4_2;
			b[ipCH4] += b_CH5P_S_HSP_CH4;

		/*  >>refer   mol  H2O    +  CN+   =>   HCN+  +     OH    UMIST  */

			c[ipCNP][ipH2O] -= H2O_CNP_HCNP_OH_1;
			c[ipH2O][ipH2O] -= H2O_CNP_HCNP_OH_2;
			b[ipH2O] -= b_H2O_CNP_HCNP_OH;
			
			c[ipCNP][ipCNP] -= H2O_CNP_HCNP_OH_1;
			c[ipH2O][ipCNP] -= H2O_CNP_HCNP_OH_2;
			b[ipCNP] -= b_H2O_CNP_HCNP_OH;
			
			c[ipCNP][ipHCNP] += H2O_CNP_HCNP_OH_1;
			c[ipH2O][ipHCNP] += H2O_CNP_HCNP_OH_2;
			b[ipHCNP] += b_H2O_CNP_HCNP_OH;
			
			c[ipCNP][ipOH] += H2O_CNP_HCNP_OH_1;
			c[ipH2O][ipOH] += H2O_CNP_HCNP_OH_2;
			b[ipOH] += b_H2O_CNP_HCNP_OH;

		/*  >>refer   mol  H2O    +  CN+   =>   HCO+  +     NH    UMIST  */

			c[ipCNP][ipH2O] -= H2O_CNP_HCOP_NH_1;
			c[ipH2O][ipH2O] -= H2O_CNP_HCOP_NH_2;
			b[ipH2O] -= b_H2O_CNP_HCOP_NH;
			
			c[ipCNP][ipCNP] -= H2O_CNP_HCOP_NH_1;
			c[ipH2O][ipCNP] -= H2O_CNP_HCOP_NH_2;
			b[ipCNP] -= b_H2O_CNP_HCOP_NH;
			
			c[ipCNP][ipHCOP] += H2O_CNP_HCOP_NH_1;
			c[ipH2O][ipHCOP] += H2O_CNP_HCOP_NH_2;
			b[ipHCOP] += b_H2O_CNP_HCOP_NH;
			
			c[ipCNP][ipNH] += H2O_CNP_HCOP_NH_1;
			c[ipH2O][ipNH] += H2O_CNP_HCOP_NH_2;
			b[ipNH] += b_H2O_CNP_HCOP_NH;

		/*  >>refer   mol  H2O    +  HCN+  =>   CN    +     H3O+  UMIST  */

			c[ipHCNP][ipH2O] -= H2O_HCNP_CN_H3OP_1;
			c[ipH2O][ipH2O] -= H2O_HCNP_CN_H3OP_2;
			b[ipH2O] -= b_H2O_HCNP_CN_H3OP;
			
			c[ipHCNP][ipHCNP] -= H2O_HCNP_CN_H3OP_1;
			c[ipH2O][ipHCNP] -= H2O_HCNP_CN_H3OP_2;
			b[ipHCNP] -= b_H2O_HCNP_CN_H3OP;
			
			c[ipHCNP][ipCN] += H2O_HCNP_CN_H3OP_1;
			c[ipH2O][ipCN] += H2O_HCNP_CN_H3OP_2;
			b[ipCN] += b_H2O_HCNP_CN_H3OP;
			
			c[ipHCNP][ipH3OP] += H2O_HCNP_CN_H3OP_1;
			c[ipH2O][ipH3OP] += H2O_HCNP_CN_H3OP_2;
			b[ipH3OP] += b_H2O_HCNP_CN_H3OP;

		/*  >>refer   mol  H2O    +  HNO+  =>   NO    +     H3O+  UMIST  */

			c[ipHNOP][ipH2O] -= H2O_HNOP_NO_H3OP_1;
			c[ipH2O][ipH2O] -= H2O_HNOP_NO_H3OP_2;
			b[ipH2O] -= b_H2O_HNOP_NO_H3OP;
			
			c[ipHNOP][ipHNOP] -= H2O_HNOP_NO_H3OP_1;
			c[ipH2O][ipHNOP] -= H2O_HNOP_NO_H3OP_2;
			b[ipHNOP] -= b_H2O_HNOP_NO_H3OP;
			
			c[ipHNOP][ipNO] += H2O_HNOP_NO_H3OP_1;
			c[ipH2O][ipNO] += H2O_HNOP_NO_H3OP_2;
			b[ipNO] += b_H2O_HNOP_NO_H3OP;
			
			c[ipHNOP][ipH3OP] += H2O_HNOP_NO_H3OP_1;
			c[ipH2O][ipH3OP] += H2O_HNOP_NO_H3OP_2;
			b[ipH3OP] += b_H2O_HNOP_NO_H3OP;

		/*  >>refer   mol  H2O+   +  S     =>   HS+   +     OH    UMIST  */

			c[ipATS][ipH2OP] -= H2OP_S_HSP_OH_1;
			c[ipH2OP][ipH2OP] -= H2OP_S_HSP_OH_2;
			b[ipH2OP] -= b_H2OP_S_HSP_OH;

			c[ipATS][ipATS] -= H2OP_S_HSP_OH_1;
			c[ipH2OP][ipATS] -= H2OP_S_HSP_OH_2;
			b[ipATS] -= b_H2OP_S_HSP_OH;
			
			c[ipATS][ipHSP] += H2OP_S_HSP_OH_1;
			c[ipH2OP][ipHSP] += H2OP_S_HSP_OH_2;
			b[ipHSP] += b_H2OP_S_HSP_OH;
			
			c[ipATS][ipOH] += H2OP_S_HSP_OH_1;
			c[ipH2OP][ipOH] += H2OP_S_HSP_OH_2;
			b[ipOH] += b_H2OP_S_HSP_OH;

		/*  >>refer   mol  H2O    +  HS+   =>   S     +     H3O+  UMIST  */

			c[ipHSP][ipH2O] -= H2O_HSP_S_H3OP_1;
			c[ipH2O][ipH2O] -= H2O_HSP_S_H3OP_2;
			b[ipH2O] -= b_H2O_HSP_S_H3OP;
			
			c[ipHSP][ipHSP] -= H2O_HSP_S_H3OP_1;
			c[ipH2O][ipHSP] -= H2O_HSP_S_H3OP_2;
			b[ipHSP] -= b_H2O_HSP_S_H3OP;
			
			c[ipHSP][ipH3OP] += H2O_HSP_S_H3OP_1;
			c[ipH2O][ipH3OP] += H2O_HSP_S_H3OP_2;
			b[ipH3OP] += b_H2O_HSP_S_H3OP;
			
			c[ipHSP][ipATS] += H2O_HSP_S_H3OP_1;
			c[ipH2O][ipATS] += H2O_HSP_S_H3OP_2;
			b[ipATS] += b_H2O_HSP_S_H3OP;
			
		/*  >>refer   mol  H3O+   +  CS    =>   HCS+  +     H2O   UMIST  */

			c[ipCS][ipH3OP] -= H3OP_CS_HCSP_H2O_1;
			c[ipH3OP][ipH3OP] -= H3OP_CS_HCSP_H2O_2;
			b[ipH3OP] -= b_H3OP_CS_HCSP_H2O;
			
			c[ipCS][ipCS] -= H3OP_CS_HCSP_H2O_1;
			c[ipH3OP][ipCS] -= H3OP_CS_HCSP_H2O_2;
			b[ipCS] -= b_H3OP_CS_HCSP_H2O;
			
			c[ipCS][ipHCSP] += H3OP_CS_HCSP_H2O_1;
			c[ipH3OP][ipHCSP] += H3OP_CS_HCSP_H2O_2;
			b[ipHCSP] += b_H3OP_CS_HCSP_H2O;
			
			c[ipCS][ipH2O] += H3OP_CS_HCSP_H2O_1;
			c[ipH3OP][ipH2O] += H3OP_CS_HCSP_H2O_2;
			b[ipH2O] += b_H3OP_CS_HCSP_H2O;

		/*  >>refer   mol  CN+    +  NO    =>   OCN+  +     N     UMIST  */

			c[ipNO][ipCNP] -= CNP_NO_OCNP_N_1;
			c[ipCNP][ipCNP] -= CNP_NO_OCNP_N_2;
			b[ipCNP] -= b_CNP_NO_OCNP_N;
			
			c[ipNO][ipNO] -= CNP_NO_OCNP_N_1;
			c[ipCNP][ipNO] -= CNP_NO_OCNP_N_2;
			b[ipNO] -= b_CNP_NO_OCNP_N;
			
			c[ipNO][ipOCNP] += CNP_NO_OCNP_N_1;
			c[ipCNP][ipOCNP] += CNP_NO_OCNP_N_2;
			b[ipOCNP] += b_CNP_NO_OCNP_N;
			
			c[ipNO][ipATN] += CNP_NO_OCNP_N_1;
			c[ipCNP][ipATN] += CNP_NO_OCNP_N_2;
			b[ipATN] += b_CNP_NO_OCNP_N;

		/*  >>refer   mol  CN     +  HNO+  =>   NO    +     HCN+  UMIST  */

			c[ipHNOP][ipCN] -= CN_HNOP_NO_HCNP_1;
			c[ipCN][ipCN] -= CN_HNOP_NO_HCNP_2;
			b[ipCN] -= b_CN_HNOP_NO_HCNP;
			
			c[ipHNOP][ipHNOP] -= CN_HNOP_NO_HCNP_1;
			c[ipCN][ipHNOP] -= CN_HNOP_NO_HCNP_2;
			b[ipHNOP] -= b_CN_HNOP_NO_HCNP;
			
			c[ipHNOP][ipNO] += CN_HNOP_NO_HCNP_1;
			c[ipCN][ipNO] += CN_HNOP_NO_HCNP_2;
			b[ipNO] += b_CN_HNOP_NO_HCNP;
			
			c[ipHNOP][ipHCNP] += CN_HNOP_NO_HCNP_1;
			c[ipCN][ipHCNP] += CN_HNOP_NO_HCNP_2;
			b[ipHCNP] += b_CN_HNOP_NO_HCNP;

		/*  >>refer   mol  CN+    +  O2    =>   NO+   +     CO    UMIST  */

			c[ipO2][ipCNP] -= CNP_O2_NOP_CO_1;
			c[ipCNP][ipCNP] -= CNP_O2_NOP_CO_2;
			b[ipCNP] -= b_CNP_O2_NOP_CO;
			
			c[ipO2][ipO2] -= CNP_O2_NOP_CO_1;
			c[ipCNP][ipO2] -= CNP_O2_NOP_CO_2;
			b[ipO2] -= b_CNP_O2_NOP_CO;
			
			c[ipO2][ipNOP] += CNP_O2_NOP_CO_1;
			c[ipCNP][ipNOP] += CNP_O2_NOP_CO_2;
			b[ipNOP] += b_CNP_O2_NOP_CO;
			
			c[ipO2][ipCO] += CNP_O2_NOP_CO_1;
			c[ipCNP][ipCO] += CNP_O2_NOP_CO_2;
			b[ipCO] += b_CNP_O2_NOP_CO;

		/*  >>refer   mol  CN+    +  O2    =>   OCN+  +     O     UMIST  */

			c[ipO2][ipCNP] -= CNP_O2_OCNP_O_1;
			c[ipCNP][ipCNP] -= CNP_O2_OCNP_O_2;
			b[ipCNP] -= b_CNP_O2_OCNP_O;
			
			c[ipO2][ipO2] -= CNP_O2_OCNP_O_1;
			c[ipCNP][ipO2] -= CNP_O2_OCNP_O_2;
			b[ipO2] -= b_CNP_O2_OCNP_O;
			
			c[ipO2][ipOCNP] += CNP_O2_OCNP_O_1;
			c[ipCNP][ipOCNP] += CNP_O2_OCNP_O_2;
			b[ipOCNP] += b_CNP_O2_OCNP_O;
			
			c[ipO2][ipATO] += CNP_O2_OCNP_O_1;
			c[ipCNP][ipATO] += CNP_O2_OCNP_O_2;
			b[ipATO] += b_CNP_O2_OCNP_O;
			
		/*  >>refer   mol  HCN+   +  CO    =>   HCO+  +     CN    UMIST  */

			c[ipCO][ipHCNP] -= HCNP_CO_HCOP_CN_1;
			c[ipHCNP][ipHCNP] -= HCNP_CO_HCOP_CN_2;
			b[ipHCNP] -= b_HCNP_CO_HCOP_CN;
			
			c[ipCO][ipCO] -= HCNP_CO_HCOP_CN_1;
			c[ipHCNP][ipCO] -= HCNP_CO_HCOP_CN_2;
			b[ipCO] -= b_HCNP_CO_HCOP_CN;
			
			c[ipCO][ipHCOP] += HCNP_CO_HCOP_CN_1;
			c[ipHCNP][ipHCOP] += HCNP_CO_HCOP_CN_2;
			b[ipHCOP] += b_HCNP_CO_HCOP_CN;
			
			c[ipCO][ipCN] += HCNP_CO_HCOP_CN_1;
			c[ipHCNP][ipCN] += HCNP_CO_HCOP_CN_2;
			b[ipCN] += b_HCNP_CO_HCOP_CN;

		/*  >>refer   mol  HCN+   +  S     =>   HS+   +     CN    UMIST  */

			c[ipATS][ipHCNP] -= HCNP_S_HSP_CN_1;
			c[ipHCNP][ipHCNP] -= HCNP_S_HSP_CN_2;
			b[ipHCNP] -= b_HCNP_S_HSP_CN;

			c[ipATS][ipATS] -= HCNP_S_HSP_CN_1;
			c[ipHCNP][ipATS] -= HCNP_S_HSP_CN_2;
			b[ipATS] -= b_HCNP_S_HSP_CN;
			
			c[ipATS][ipHSP] += HCNP_S_HSP_CN_1;
			c[ipHCNP][ipHSP] += HCNP_S_HSP_CN_2;
			b[ipHSP] += b_HCNP_S_HSP_CN;
			
			c[ipATS][ipCN] += HCNP_S_HSP_CN_1;
			c[ipHCNP][ipCN] += HCNP_S_HSP_CN_2;
			b[ipCN] += b_HCNP_S_HSP_CN;

		/*  >>refer   mol  CO     +  HNO+  =>   NO    +     HCO+  UMIST  */

			c[ipHNOP][ipCO] -= CO_HNOP_NO_HCOP_1;
			c[ipCO][ipCO] -= CO_HNOP_NO_HCOP_2;
			b[ipCO] -= b_CO_HNOP_NO_HCOP;
			
			c[ipHNOP][ipHNOP] -= CO_HNOP_NO_HCOP_1;
			c[ipCO][ipHNOP] -= CO_HNOP_NO_HCOP_2;
			b[ipHNOP] -= b_CO_HNOP_NO_HCOP;
			
			c[ipHNOP][ipNO] += CO_HNOP_NO_HCOP_1;
			c[ipCO][ipNO] += CO_HNOP_NO_HCOP_2;
			b[ipNO] += b_CO_HNOP_NO_HCOP;
			
			c[ipHNOP][ipHCOP] += CO_HNOP_NO_HCOP_1;
			c[ipCO][ipHCOP] += CO_HNOP_NO_HCOP_2;
			b[ipHCOP] += b_CO_HNOP_NO_HCOP;

		/*  >>refer   mol  N2+    +  OCS   =>   S+    +     N2  + CO     UMIST  */

			c[ipOCS][ipN2P] -= N2P_OCS_SP_N2_CO_1;
			c[ipN2P][ipN2P] -= N2P_OCS_SP_N2_CO_2;
			b[ipN2P] -= b_N2P_OCS_SP_N2_CO;
			
			c[ipOCS][ipOCS] -= N2P_OCS_SP_N2_CO_1;
			c[ipN2P][ipOCS] -= N2P_OCS_SP_N2_CO_2;
			b[ipOCS] -= b_N2P_OCS_SP_N2_CO;
			
			c[ipOCS][ipSP] += N2P_OCS_SP_N2_CO_1;
			c[ipN2P][ipSP] += N2P_OCS_SP_N2_CO_2;
			b[ipSP] += b_N2P_OCS_SP_N2_CO;
			
			c[ipOCS][ipN2] += N2P_OCS_SP_N2_CO_1;
			c[ipN2P][ipN2] += N2P_OCS_SP_N2_CO_2;
			b[ipN2] += b_N2P_OCS_SP_N2_CO;

			c[ipOCS][ipCO] += N2P_OCS_SP_N2_CO_1;
			c[ipN2P][ipCO] += N2P_OCS_SP_N2_CO_2;
			b[ipCO] += b_N2P_OCS_SP_N2_CO;

		/*  >>refer   mol  HCO+   +  S     =>   HS+   +     CO    UMIST  */

			c[ipATS][ipHCOP] -= HCOP_S_HSP_CO_1;
			c[ipHCOP][ipHCOP] -= HCOP_S_HSP_CO_2;
			b[ipHCOP] -= b_HCOP_S_HSP_CO;

			c[ipATS][ipATS] -= HCOP_S_HSP_CO_1;
			c[ipHCOP][ipATS] -= HCOP_S_HSP_CO_2;
			b[ipATS] -= b_HCOP_S_HSP_CO;
			
			c[ipATS][ipHSP] += HCOP_S_HSP_CO_1;
			c[ipHCOP][ipHSP] += HCOP_S_HSP_CO_2;
			b[ipHSP] += b_HCOP_S_HSP_CO;
			
			c[ipATS][ipCO] += HCOP_S_HSP_CO_1;
			c[ipHCOP][ipCO] += HCOP_S_HSP_CO_2;
			b[ipCO] += b_HCOP_S_HSP_CO;

		/*  >>refer   mol  HCO+   +  CS    =>   HCS+  +     CO    UMIST  */

			c[ipCS][ipHCOP] -= HCOP_CS_HCSP_CO_1;
			c[ipHCOP][ipHCOP] -= HCOP_CS_HCSP_CO_2;
			b[ipHCOP] -= b_HCOP_CS_HCSP_CO;
			
			c[ipCS][ipCS] -= HCOP_CS_HCSP_CO_1;
			c[ipHCOP][ipCS] -= HCOP_CS_HCSP_CO_2;
			b[ipCS] -= b_HCOP_CS_HCSP_CO;
			
			c[ipCS][ipHCSP] += HCOP_CS_HCSP_CO_1;
			c[ipHCOP][ipHCSP] += HCOP_CS_HCSP_CO_2;
			b[ipHCSP] += b_HCOP_CS_HCSP_CO;
			
			c[ipCS][ipCO] += HCOP_CS_HCSP_CO_1;
			c[ipHCOP][ipCO] += HCOP_CS_HCSP_CO_2;
			b[ipCO] += b_HCOP_CS_HCSP_CO;

		/*  >>refer   mol  HNO+   +  S     =>   HS+   +     NO    UMIST  */
			
			c[ipATS][ipHNOP] -= HNOP_S_HSP_NO_1;
			c[ipHNOP][ipHNOP] -= HNOP_S_HSP_NO_2;
			b[ipHNOP] -= b_HNOP_S_HSP_NO;
			
			c[ipATS][ipATS] -= HNOP_S_HSP_NO_1;
			c[ipHNOP][ipATS] -= HNOP_S_HSP_NO_2;
			b[ipATS] -= b_HNOP_S_HSP_NO;
			
			c[ipATS][ipHSP] += HNOP_S_HSP_NO_1;
			c[ipHNOP][ipHSP] += HNOP_S_HSP_NO_2;
			b[ipHSP] += b_HNOP_S_HSP_NO;
			
			c[ipATS][ipNO] += HNOP_S_HSP_NO_1;
			c[ipHNOP][ipNO] += HNOP_S_HSP_NO_2;
			b[ipNO] += b_HNOP_S_HSP_NO;

		/*  >>refer   mol  O2     +  S+    =>   SO+   +     O     UMIST  */

			c[ipSP][ipO2] -= O2_SP_SOP_O_1;
			c[ipO2][ipO2] -= O2_SP_SOP_O_2;
			b[ipO2] -= b_O2_SP_SOP_O;
			
			c[ipSP][ipSP] -= O2_SP_SOP_O_1;
			c[ipO2][ipSP] -= O2_SP_SOP_O_2;
			b[ipSP] -= b_O2_SP_SOP_O;
			
			c[ipSP][ipSOP] += O2_SP_SOP_O_1;
			c[ipO2][ipSOP] += O2_SP_SOP_O_2;
			b[ipSOP] += b_O2_SP_SOP_O;
			
			c[ipSP][ipATO] += O2_SP_SOP_O_1;
			c[ipO2][ipATO] += O2_SP_SOP_O_2;
			b[ipATO] += b_O2_SP_SOP_O;

		/*  >>refer   mol  O2+    +  S     =>   SO+   +     O     UMIST  */

			c[ipATS][ipO2P] -= O2P_S_SOP_O_1;
			c[ipO2P][ipO2P] -= O2P_S_SOP_O_2;
			b[ipO2P] -= b_O2P_S_SOP_O;
			
			c[ipATS][ipATS] -= O2P_S_SOP_O_1;
			c[ipO2P][ipATS] -= O2P_S_SOP_O_2;
			b[ipATS] -= b_O2P_S_SOP_O;
			
			c[ipATS][ipSOP] += O2P_S_SOP_O_1;
			c[ipO2P][ipSOP] += O2P_S_SOP_O_2;
			b[ipSOP] += b_O2P_S_SOP_O;

			c[ipATS][ipATO] += O2P_S_SOP_O_1;
			c[ipO2P][ipATO] += O2P_S_SOP_O_2;
			b[ipATO] += b_O2P_S_SOP_O;

		/*  >>refer   mol  O2     +  CS+   =>   OCS+  +     O     UMIST  */

			c[ipCSP][ipO2] -= O2_CSP_OCSP_O_1;
			c[ipO2][ipO2] -= O2_CSP_OCSP_O_2;
			b[ipO2] -= b_O2_CSP_OCSP_O;
			
			c[ipCSP][ipCSP] -= O2_CSP_OCSP_O_1;
			c[ipO2][ipCSP] -= O2_CSP_OCSP_O_2;
			b[ipCSP] -= b_O2_CSP_OCSP_O;
			
			c[ipCSP][ipOCSP] += O2_CSP_OCSP_O_1;
			c[ipO2][ipOCSP] += O2_CSP_OCSP_O_2;
			b[ipOCSP] += b_O2_CSP_OCSP_O;
			
			c[ipCSP][ipATO] += O2_CSP_OCSP_O_1;
			c[ipO2][ipATO] += O2_CSP_OCSP_O_2;
			b[ipATO] += b_O2_CSP_OCSP_O;
			
		/*  >>refer   mol  S      +  SIO+  =>   SO    +     SI+   UMIST  */
			
			c[ipSIOP][ipSIOP] -= S_SIOP_SO_SIP_1;
			c[ipATS][ipSIOP] -= S_SIOP_SO_SIP_2;
			b[ipSIOP] -= b_S_SIOP_SO_SIP;
			
			c[ipSIOP][ipATS] -= S_SIOP_SO_SIP_1;
			c[ipATS][ipATS] -= S_SIOP_SO_SIP_2;
			b[ipATS] -= b_S_SIOP_SO_SIP;
			
			c[ipSIOP][ipSO] += S_SIOP_SO_SIP_1;
			c[ipATS][ipSO] += S_SIOP_SO_SIP_2;
			b[ipSO] += b_S_SIOP_SO_SIP;
			
			c[ipSIOP][ipSIP] += S_SIOP_SO_SIP_1;
			c[ipATS][ipSIP] += S_SIOP_SO_SIP_2;
			b[ipSIP] += b_S_SIOP_SO_SIP;

		/*  >>refer   mol  H+     +  NH    =>   NH+   +     H     UMIST  */

			c[ipNH][ipNH] -= HP_NH_NHP_H;
			c[ipNH][ipNHP] += HP_NH_NHP_H;

		/*  >>refer   mol  H+     +  NH2   =>   NH2+  +     H     UMIST  */

			c[ipNH2][ipNH2] -= HP_NH2_NH2P_H;
			c[ipNH2][ipNH2P] += HP_NH2_NH2P_H;

		/*  >>refer   mol  H+     +  NH3   =>   NH3+  +     H     UMIST  */

			c[ipNH3][ipNH3] -= HP_NH3_NH3P_H;
			c[ipNH3][ipNH3P] += HP_NH3_NH3P_H;

		/*  >>refer   mol  H      +  CN+   =>   CN    +     H+    UMIST  */

			c[ipCNP][ipCNP] -= H_CNP_CN_HP;
			c[ipCNP][ipCN] += H_CNP_CN_HP;

		/*  >>refer   mol  H+     +  HCN   =>   HCN+  +     H     UMIST  */

			c[ipHCN][ipHCN] -= HP_HCN_HCNP_H;
			c[ipHCN][ipHCNP] += HP_HCN_HCNP_H;

		/*  >>refer   mol  H      +  HCN+  =>   HCN   +     H+    UMIST  */

			c[ipHCNP][ipHCNP] -= H_HCNP_HCN_HP;
			c[ipHCNP][ipHCN] += H_HCNP_HCN_HP;

		/*  >>refer   mol  H      +  N2+   =>   N2    +     H+    UMIST  */

			c[ipN2P][ipN2P] -= H_N2P_N2_HP;
			c[ipN2P][ipN2] += H_N2P_N2_HP;

		/*  >>refer   mol  H+     +  NO    =>   NO+   +     H     UMIST  */

			c[ipNO][ipNO] -= HP_NO_NOP_H;
			c[ipNO][ipNOP] += HP_NO_NOP_H;

		/*  >>refer   mol  H+     +  HS    =>   HS+   +     H     UMIST  */

			c[ipHS][ipHS] -= HP_HS_HSP_H;
			c[ipHS][ipHSP] += HP_HS_HSP_H;

		/*  >>refer   mol  H+     +  SIN   =>   SIN+  +     H     UMIST  */

			c[ipSIN][ipSIN] -= HP_SIN_SINP_H;
			c[ipSIN][ipSINP] += HP_SIN_SINP_H;

		/*  >>refer   mol  H+     +  CS    =>   CS+   +     H     UMIST  */

			c[ipCS][ipCS] -= HP_CS_CSP_H;
			c[ipCS][ipCSP] += HP_CS_CSP_H;

		/*  >>refer   mol  H+     +  NS    =>   NS+   +     H     UMIST  */

			c[ipNS][ipNS] -= HP_NS_NSP_H;
			c[ipNS][ipNSP] += HP_NS_NSP_H;

		/*  >>refer   mol  H+     +  SO    =>   SO+   +     H     UMIST  */

			c[ipSO][ipSO] -= HP_SO_SOP_H;
			c[ipSO][ipSOP] += HP_SO_SOP_H;

		/*  >>refer   mol  H+     +  OCS   =>   OCS+  +     H     UMIST  */

			c[ipOCS][ipOCS] -= HP_OCS_OCSP_H;
			c[ipOCS][ipOCSP] += HP_OCS_OCSP_H;

		/*  >>refer   mol  H+     +  S2    =>   S2+   +     H     UMIST  */

			c[ipS2][ipS2] -= HP_S2_S2P_H;
			c[ipS2][ipS2P] += HP_S2_S2P_H;

		/*  >>refer   mol  H2+    +  NH    =>   NH+   +     H2    UMIST  */

			c[ipNH][ipNH] -= H2P_NH_NHP_H2;
			c[ipNH][ipNHP] += H2P_NH_NHP_H2;

		/*  >>refer   mol  H2+    +  NH2   =>   NH2+  +     H2    UMIST  */

			c[ipNH2][ipNH2] -= H2P_NH2_NH2P_H2;
			c[ipNH2][ipNH2P] += H2P_NH2_NH2P_H2;

		/*  >>refer   mol  H2+    +  NH3   =>   NH3+  +     H2    UMIST  */

			c[ipNH3][ipNH3] -= H2P_NH3_NH3P_H2;
			c[ipNH3][ipNH3P] += H2P_NH3_NH3P_H2;

		/*  >>refer   mol  H2+    +  CN    =>   CN+   +     H2    UMIST  */

			c[ipCN][ipCN] -= H2P_CN_CNP_H2;
			c[ipCN][ipCNP] += H2P_CN_CNP_H2;

		/*  >>refer   mol  H2+    +  HCN   =>   HCN+  +     H2    UMIST  */

			c[ipHCN][ipHCN] -= H2P_HCN_HCNP_H2;
			c[ipHCN][ipHCNP] += H2P_HCN_HCNP_H2;

		/*  >>refer   mol  H2+    +  NO    =>   NO+   +     H2    UMIST  */

			c[ipNO][ipNO] -= H2P_NO_NOP_H2;
			c[ipNO][ipNOP] += H2P_NO_NOP_H2;

		/*  >>refer   mol  He+    +  NH3   =>   NH3+  +     He    UMIST  */

			c[ipNH3][ipNH3] -= HeP_NH3_NH3P_He;
			c[ipNH3][ipNH3P] += HeP_NH3_NH3P_He;

		/*  >>refer   mol  He+    +  N2    =>   N2+   +     He    UMIST  */

			c[ipN2][ipN2] -= HeP_N2_N2P_He;
			c[ipN2][ipN2P] += HeP_N2_N2P_He;

		/*  >>refer   mol  C+     +  NH3   =>   NH3+  +     C     UMIST  */

			c[ipNH3][ipCP] -= CP_NH3_NH3P_C_1;
			c[ipCP][ipCP] -= CP_NH3_NH3P_C_2;
			b[ipCP] -= b_CP_NH3_NH3P_C;
			
			c[ipNH3][ipNH3] -= CP_NH3_NH3P_C_1;
			c[ipCP][ipNH3] -= CP_NH3_NH3P_C_2;
			b[ipNH3] -= b_CP_NH3_NH3P_C;
			
			c[ipNH3][ipNH3P] += CP_NH3_NH3P_C_1;
			c[ipCP][ipNH3P] += CP_NH3_NH3P_C_2;
			b[ipNH3P] += b_CP_NH3_NH3P_C;
			
			c[ipNH3][ipATC] += CP_NH3_NH3P_C_1;
			c[ipCP][ipATC] += CP_NH3_NH3P_C_2;
			b[ipATC] += b_CP_NH3_NH3P_C;

		/*  >>refer   mol  C      +  CN+   =>   CN    +     C+    UMIST  */
			
			c[ipCNP][ipCNP] -= C_CNP_CN_CP_1;
			c[ipATC][ipCNP] -= C_CNP_CN_CP_2;
			b[ipCNP] -= b_C_CNP_CN_CP;
			
			c[ipCNP][ipATC] -= C_CNP_CN_CP_1;
			c[ipATC][ipATC] -= C_CNP_CN_CP_2;
			b[ipATC] -= b_C_CNP_CN_CP;
			
			c[ipCNP][ipCN] += C_CNP_CN_CP_1;
			c[ipATC][ipCN] += C_CNP_CN_CP_2;
			b[ipCN] += b_C_CNP_CN_CP;
			
			c[ipCNP][ipCP] += C_CNP_CN_CP_1;
			c[ipATC][ipCP] += C_CNP_CN_CP_2;
			b[ipCP] += b_C_CNP_CN_CP;

		/*  >>refer   mol  C      +  N2+   =>   N2    +     C+    UMIST  */

			c[ipN2P][ipN2P] -= C_N2P_N2_CP_1;
			c[ipATC][ipN2P] -= C_N2P_N2_CP_2;
			b[ipN2P] -= b_C_N2P_N2_CP;

			c[ipN2P][ipATC] -= C_N2P_N2_CP_1;
			c[ipATC][ipATC] -= C_N2P_N2_CP_2;
			b[ipATC] -= b_C_N2P_N2_CP;
			
			c[ipN2P][ipN2] += C_N2P_N2_CP_1;
			c[ipATC][ipN2] += C_N2P_N2_CP_2;
			b[ipN2] += b_C_N2P_N2_CP;
			
			c[ipN2P][ipCP] += C_N2P_N2_CP_1;
			c[ipATC][ipCP] += C_N2P_N2_CP_2;
			b[ipCP] += b_C_N2P_N2_CP;

		/*  >>refer   mol  C+     +  NO    =>   NO+   +     C     UMIST  */

			c[ipNO][ipCP] -= CP_NO_NOP_C_1;
			c[ipCP][ipCP] -= CP_NO_NOP_C_2;
			b[ipCP] -= b_CP_NO_NOP_C;
			
			c[ipNO][ipNO] -= CP_NO_NOP_C_1;
			c[ipCP][ipNO] -= CP_NO_NOP_C_2;
			b[ipNO] -= b_CP_NO_NOP_C;
			
			c[ipNO][ipNOP] += CP_NO_NOP_C_1;
			c[ipCP][ipNOP] += CP_NO_NOP_C_2;
			b[ipNOP] += b_CP_NO_NOP_C;
			
			c[ipNO][ipATC] += CP_NO_NOP_C_1;
			c[ipCP][ipATC] += CP_NO_NOP_C_2;
			b[ipATC] += b_CP_NO_NOP_C;
			
		/*  >>refer   mol  C+     +  SIN   =>   SIN+  +     C     UMIST  */

			c[ipSIN][ipCP] -= CP_SIN_SINP_C_1;
			c[ipCP][ipCP] -= CP_SIN_SINP_C_2;
			b[ipCP] -= b_CP_SIN_SINP_C;
			
			c[ipSIN][ipSIN] -= CP_SIN_SINP_C_1;
			c[ipCP][ipSIN] -= CP_SIN_SINP_C_2;
			b[ipSIN] -= b_CP_SIN_SINP_C;
			
			c[ipSIN][ipSINP] += CP_SIN_SINP_C_1;
			c[ipCP][ipSINP] += CP_SIN_SINP_C_2;
			b[ipSINP] += b_CP_SIN_SINP_C;
			
			c[ipSIN][ipATC] += CP_SIN_SINP_C_1;
			c[ipCP][ipATC] += CP_SIN_SINP_C_2;
			b[ipATC] += b_CP_SIN_SINP_C;
			
		/*  >>refer   mol  C      +  CS+   =>   CS    +     C+    UMIST  */

			c[ipCSP][ipCSP] -= C_CSP_CS_CP_1;
			c[ipATC][ipCSP] -= C_CSP_CS_CP_2;
			b[ipCSP] -= b_C_CSP_CS_CP;

			c[ipCSP][ipATC] -= C_CSP_CS_CP_1;
			c[ipATC][ipATC] -= C_CSP_CS_CP_2;
			b[ipATC] -= b_C_CSP_CS_CP;
			
			c[ipCSP][ipCS] += C_CSP_CS_CP_1;
			c[ipATC][ipCS] += C_CSP_CS_CP_2;
			b[ipCS] += b_C_CSP_CS_CP;
			
			c[ipCSP][ipCP] += C_CSP_CS_CP_1;
			c[ipATC][ipCP] += C_CSP_CS_CP_2;
			b[ipCP] += b_C_CSP_CS_CP;

		/*  >>refer   mol  C+     +  NS    =>   NS+   +     C     UMIST  */

			c[ipNS][ipCP] -= CP_NS_NSP_C_1;
			c[ipCP][ipCP] -= CP_NS_NSP_C_2;
			b[ipCP] -= b_CP_NS_NSP_C;
			
			c[ipNS][ipNS] -= CP_NS_NSP_C_1;
			c[ipCP][ipNS] -= CP_NS_NSP_C_2;
			b[ipNS] -= b_CP_NS_NSP_C;
			
			c[ipNS][ipNSP] += CP_NS_NSP_C_1;
			c[ipCP][ipNSP] += CP_NS_NSP_C_2;
			b[ipNSP] += b_CP_NS_NSP_C;
			
			c[ipNS][ipATC] += CP_NS_NSP_C_1;
			c[ipCP][ipATC] += CP_NS_NSP_C_2;
			b[ipATC] += b_CP_NS_NSP_C;
			
		/*  >>refer   mol  C+     +  SO    =>   SO+   +     C     UMIST  */

			c[ipSO][ipCP] -= CP_SO_SOP_C_1;
			c[ipCP][ipCP] -= CP_SO_SOP_C_2;
			b[ipCP] -= b_CP_SO_SOP_C;
			
			c[ipSO][ipSO] -= CP_SO_SOP_C_1;
			c[ipCP][ipSO] -= CP_SO_SOP_C_2;
			b[ipSO] -= b_CP_SO_SOP_C;
			
			c[ipSO][ipSOP] += CP_SO_SOP_C_1;
			c[ipCP][ipSOP] += CP_SO_SOP_C_2;
			b[ipSOP] += b_CP_SO_SOP_C;
			
			c[ipSO][ipATC] += CP_SO_SOP_C_1;
			c[ipCP][ipATC] += CP_SO_SOP_C_2;
			b[ipATC] += b_CP_SO_SOP_C;
			
		/*  >>refer   mol  C+     +  OCS   =>   OCS+  +     C     UMIST  */

			c[ipOCS][ipCP] -= CP_OCS_OCSP_C_1;
			c[ipCP][ipCP] -= CP_OCS_OCSP_C_2;
			b[ipCP] -= b_CP_OCS_OCSP_C;
			
			c[ipOCS][ipOCS] -= CP_OCS_OCSP_C_1;
			c[ipCP][ipOCS] -= CP_OCS_OCSP_C_2;
			b[ipOCS] -= b_CP_OCS_OCSP_C;
			
			c[ipOCS][ipOCSP] += CP_OCS_OCSP_C_1;
			c[ipCP][ipOCSP] += CP_OCS_OCSP_C_2;
			b[ipOCSP] += b_CP_OCS_OCSP_C;
				
			c[ipOCS][ipATC] += CP_OCS_OCSP_C_1;
			c[ipCP][ipATC] += CP_OCS_OCSP_C_2;
			b[ipATC] += b_CP_OCS_OCSP_C;
			
		/*  >>refer   mol  CH     +  NH2+  =>   NH2   +     CH+   UMIST  */

			c[ipNH2P][ipCH] -= CH_NH2P_NH2_CHP_1;
			c[ipCH][ipCH] -= CH_NH2P_NH2_CHP_2;
			b[ipCH] -= b_CH_NH2P_NH2_CHP;
			
			c[ipNH2P][ipNH2P] -= CH_NH2P_NH2_CHP_1;
			c[ipCH][ipNH2P] -= CH_NH2P_NH2_CHP_2;
			b[ipNH2P] -= b_CH_NH2P_NH2_CHP;
			
			c[ipNH2P][ipNH2] += CH_NH2P_NH2_CHP_1;
			c[ipCH][ipNH2] += CH_NH2P_NH2_CHP_2;
			b[ipNH2] += b_CH_NH2P_NH2_CHP;
			
			c[ipNH2P][ipCHP] += CH_NH2P_NH2_CHP_1;
			c[ipCH][ipCHP] += CH_NH2P_NH2_CHP_2;
			b[ipCHP] += b_CH_NH2P_NH2_CHP;

		/*  >>refer   mol  CH+    +  NH3   =>   NH3+  +     CH    UMIST  */

			c[ipNH3][ipCHP] -= CHP_NH3_NH3P_CH_1;
			c[ipCHP][ipCHP] -= CHP_NH3_NH3P_CH_2;
			b[ipCHP] -= b_CHP_NH3_NH3P_CH;
			
			c[ipNH3][ipNH3] -= CHP_NH3_NH3P_CH_1;
			c[ipCHP][ipNH3] -= CHP_NH3_NH3P_CH_2;
			b[ipNH3] -= b_CHP_NH3_NH3P_CH;
			
			c[ipNH3][ipNH3P] += CHP_NH3_NH3P_CH_1;
			c[ipCHP][ipNH3P] += CHP_NH3_NH3P_CH_2;
			b[ipNH3P] += b_CHP_NH3_NH3P_CH;
			
			c[ipNH3][ipCH] += CHP_NH3_NH3P_CH_1;
			c[ipCHP][ipCH] += CHP_NH3_NH3P_CH_2;
			b[ipCH] += b_CHP_NH3_NH3P_CH;

		/*  >>refer   mol  CH     +  CN+   =>   CN    +     CH+   UMIST  */

			c[ipCNP][ipCH] -= CH_CNP_CN_CHP_1;
			c[ipCH][ipCH] -= CH_CNP_CN_CHP_2;
			b[ipCH] -= b_CH_CNP_CN_CHP;
			
			c[ipCNP][ipCNP] -= CH_CNP_CN_CHP_1;
			c[ipCH][ipCNP] -= CH_CNP_CN_CHP_2;
			b[ipCNP] -= b_CH_CNP_CN_CHP;
			
			c[ipCNP][ipCN] += CH_CNP_CN_CHP_1;
			c[ipCH][ipCN] += CH_CNP_CN_CHP_2;
			b[ipCN] += b_CH_CNP_CN_CHP;
			
			c[ipCNP][ipCHP] += CH_CNP_CN_CHP_1;
			c[ipCH][ipCHP] += CH_CNP_CN_CHP_2;
			b[ipCHP] += b_CH_CNP_CN_CHP;

		/*  >>refer   mol  CH     +  N2+   =>   N2    +     CH+   UMIST  */

			c[ipN2P][ipCH] -= CH_N2P_N2_CHP_1;
			c[ipCH][ipCH] -= CH_N2P_N2_CHP_2;
			b[ipCH] -= b_CH_N2P_N2_CHP;
			
			c[ipN2P][ipN2P] -= CH_N2P_N2_CHP_1;
			c[ipCH][ipN2P] -= CH_N2P_N2_CHP_2;
			b[ipN2P] -= b_CH_N2P_N2_CHP;
			
			c[ipN2P][ipN2] += CH_N2P_N2_CHP_1;
			c[ipCH][ipN2] += CH_N2P_N2_CHP_2;
			b[ipN2] += b_CH_N2P_N2_CHP;
			
			c[ipN2P][ipCHP] += CH_N2P_N2_CHP_1;
			c[ipCH][ipCHP] += CH_N2P_N2_CHP_2;
			b[ipCHP] += b_CH_N2P_N2_CHP;

		/*  >>refer   mol  CH+    +  NO    =>   NO+   +     CH    UMIST  */

			c[ipNO][ipCHP] -= CHP_NO_NOP_CH_1;
			c[ipCHP][ipCHP] -= CHP_NO_NOP_CH_2;
			b[ipCHP] -= b_CHP_NO_NOP_CH;
			
			c[ipNO][ipNO] -= CHP_NO_NOP_CH_1;
			c[ipCHP][ipNO] -= CHP_NO_NOP_CH_2;
			b[ipNO] -= b_CHP_NO_NOP_CH;
			
			c[ipNO][ipNOP] += CHP_NO_NOP_CH_1;
			c[ipCHP][ipNOP] += CHP_NO_NOP_CH_2;
			b[ipNOP] += b_CHP_NO_NOP_CH;
			
			c[ipNO][ipCH] += CHP_NO_NOP_CH_1;
			c[ipCHP][ipCH] += CHP_NO_NOP_CH_2;
			b[ipCH] += b_CHP_NO_NOP_CH;

		/*  >>refer   mol  N+     +  NH    =>   NH+   +     N     UMIST  */

			c[ipNH][ipNP] -= NP_NH_NHP_N_1;
			c[ipNP][ipNP] -= NP_NH_NHP_N_2;
			b[ipNP] -= b_NP_NH_NHP_N;
			
			c[ipNH][ipNH] -= NP_NH_NHP_N_1;
			c[ipNP][ipNH] -= NP_NH_NHP_N_2;
			b[ipNH] -= b_NP_NH_NHP_N;
			
			c[ipNH][ipNHP] += NP_NH_NHP_N_1;
			c[ipNP][ipNHP] += NP_NH_NHP_N_2;
			b[ipNHP] += b_NP_NH_NHP_N;
			
			c[ipNH][ipATN] += NP_NH_NHP_N_1;
			c[ipNP][ipATN] += NP_NH_NHP_N_2;
			b[ipATN] += b_NP_NH_NHP_N;
			
		/*  >>refer   mol  N+     +  NH2   =>   NH2+  +     N     UMIST  */

			c[ipNH2][ipNP] -= NP_NH2_NH2P_N_1;
			c[ipNP][ipNP] -= NP_NH2_NH2P_N_2;
			b[ipNP] -= b_NP_NH2_NH2P_N;
			
			c[ipNH2][ipNH2] -= NP_NH2_NH2P_N_1;
			c[ipNP][ipNH2] -= NP_NH2_NH2P_N_2;
			b[ipNH2] -= b_NP_NH2_NH2P_N;
			
			c[ipNH2][ipNH2P] += NP_NH2_NH2P_N_1;
			c[ipNP][ipNH2P] += NP_NH2_NH2P_N_2;
			b[ipNH2P] += b_NP_NH2_NH2P_N;
			
			c[ipNH2][ipATN] += NP_NH2_NH2P_N_1;
			c[ipNP][ipATN] += NP_NH2_NH2P_N_2;
			b[ipATN] += b_NP_NH2_NH2P_N;

		/*  >>refer   mol  N+     +  NH3   =>   NH3+  +     N     UMIST  */

			c[ipNH3][ipNP] -= NP_NH3_NH3P_N_1;
			c[ipNP][ipNP] -= NP_NH3_NH3P_N_2;
			b[ipNP] -= b_NP_NH3_NH3P_N;
			
			c[ipNH3][ipNH3] -= NP_NH3_NH3P_N_1;
			c[ipNP][ipNH3] -= NP_NH3_NH3P_N_2;
			b[ipNH3] -= b_NP_NH3_NH3P_N;
			
			c[ipNH3][ipNH3P] += NP_NH3_NH3P_N_1;
			c[ipNP][ipNH3P] += NP_NH3_NH3P_N_2;
			b[ipNH3P] += b_NP_NH3_NH3P_N;
			
			c[ipNH3][ipATN] += NP_NH3_NH3P_N_1;
			c[ipNP][ipATN] += NP_NH3_NH3P_N_2;
			b[ipATN] += b_NP_NH3_NH3P_N;

		/*  >>refer   mol  N+     +  CN    =>   CN+   +     N     UMIST  */

			c[ipCN][ipNP] -= NP_CN_CNP_N_1;
			c[ipNP][ipNP] -= NP_CN_CNP_N_2;
			b[ipNP] -= b_NP_CN_CNP_N;
			
			c[ipCN][ipCN] -= NP_CN_CNP_N_1;
			c[ipNP][ipCN] -= NP_CN_CNP_N_2;
			b[ipCN] -= b_NP_CN_CNP_N;
			
			c[ipCN][ipCNP] += NP_CN_CNP_N_1;
			c[ipNP][ipCNP] += NP_CN_CNP_N_2;
			b[ipCNP] += b_NP_CN_CNP_N;
			
			c[ipCN][ipATN] += NP_CN_CNP_N_1;
			c[ipNP][ipATN] += NP_CN_CNP_N_2;
			b[ipATN] += b_NP_CN_CNP_N;

		/*  >>refer   mol  N+     +  HCN   =>   HCN+  +     N     UMIST  */

			c[ipHCN][ipNP] -= NP_HCN_HCNP_N_1;
			c[ipNP][ipNP] -= NP_HCN_HCNP_N_2;
			b[ipNP] -= b_NP_HCN_HCNP_N;
			
			c[ipHCN][ipHCN] -= NP_HCN_HCNP_N_1;
			c[ipNP][ipHCN] -= NP_HCN_HCNP_N_2;
			b[ipHCN] -= b_NP_HCN_HCNP_N;
			
			c[ipHCN][ipHCNP] += NP_HCN_HCNP_N_1;
			c[ipNP][ipHCNP] += NP_HCN_HCNP_N_2;
			b[ipHCNP] += b_NP_HCN_HCNP_N;
			
			c[ipHCN][ipATN] += NP_HCN_HCNP_N_1;
			c[ipNP][ipATN] += NP_HCN_HCNP_N_2;
			b[ipATN] += b_NP_HCN_HCNP_N;

		/*  >>refer   mol  N      +  N2+   =>   N2    +     N+    UMIST  */

			c[ipN2P][ipN2P] -= N_N2P_N2_NP_1;
			c[ipATN][ipN2P] -= N_N2P_N2_NP_2;
			b[ipN2P] -= b_N_N2P_N2_NP;

			c[ipN2P][ipATN] -= N_N2P_N2_NP_1;
			c[ipATN][ipATN] -= N_N2P_N2_NP_2;
			b[ipATN] -= b_N_N2P_N2_NP;
			
			c[ipN2P][ipN2] += N_N2P_N2_NP_1;
			c[ipATN][ipN2] += N_N2P_N2_NP_2;
			b[ipN2] += b_N_N2P_N2_NP;
			
			c[ipN2P][ipNP] += N_N2P_N2_NP_1;
			c[ipATN][ipNP] += N_N2P_N2_NP_2;
			b[ipNP] += b_N_N2P_N2_NP;

		/*  >>refer   mol  N+     +  NO    =>   NO+   +     N     UMIST  */

			c[ipNO][ipNP] -= NP_NO_NOP_N_1;
			c[ipNP][ipNP] -= NP_NO_NOP_N_2;
			b[ipNP] -= b_NP_NO_NOP_N;
			
			c[ipNO][ipNO] -= NP_NO_NOP_N_1;
			c[ipNP][ipNO] -= NP_NO_NOP_N_2;
			b[ipNO] -= b_NP_NO_NOP_N;
			
			c[ipNO][ipNOP] += NP_NO_NOP_N_1;
			c[ipNP][ipNOP] += NP_NO_NOP_N_2;
			b[ipNOP] += b_NP_NO_NOP_N;
			
			c[ipNO][ipATN] += NP_NO_NOP_N_1;
			c[ipNP][ipATN] += NP_NO_NOP_N_2;
			b[ipATN] += b_NP_NO_NOP_N;
			
		/*  >>refer   mol  N+     +  OCS   =>   OCS+  +     N     UMIST  */

			c[ipOCS][ipNP] -= NP_OCS_OCSP_N_1;
			c[ipNP][ipNP] -= NP_OCS_OCSP_N_2;
			b[ipNP] -= b_NP_OCS_OCSP_N;
			
			c[ipOCS][ipOCS] -= NP_OCS_OCSP_N_1;
			c[ipNP][ipOCS] -= NP_OCS_OCSP_N_2;
			b[ipOCS] -= b_NP_OCS_OCSP_N;
			
			c[ipOCS][ipOCSP] += NP_OCS_OCSP_N_1;
			c[ipNP][ipOCSP] += NP_OCS_OCSP_N_2;
			b[ipOCSP] += b_NP_OCS_OCSP_N;
			
			c[ipOCS][ipATN] += NP_OCS_OCSP_N_1;
			c[ipNP][ipATN] += NP_OCS_OCSP_N_2;
			b[ipATN] += b_NP_OCS_OCSP_N;

		/*  >>refer   mol  CH2    +  NH2+  =>   NH2   +     CH2+  UMIST  */

			c[ipNH2P][ipCH2] -= CH2_NH2P_NH2_CH2P_1;
			c[ipCH2][ipCH2] -= CH2_NH2P_NH2_CH2P_2;
			b[ipCH2] -= b_CH2_NH2P_NH2_CH2P;
			
			c[ipNH2P][ipNH2P] -= CH2_NH2P_NH2_CH2P_1;
			c[ipCH2][ipNH2P] -= CH2_NH2P_NH2_CH2P_2;
			b[ipNH2P] -= b_CH2_NH2P_NH2_CH2P;
			
			c[ipNH2P][ipNH2] += CH2_NH2P_NH2_CH2P_1;
			c[ipCH2][ipNH2] += CH2_NH2P_NH2_CH2P_2;
			b[ipNH2] += b_CH2_NH2P_NH2_CH2P;
			
			c[ipNH2P][ipCH2P] += CH2_NH2P_NH2_CH2P_1;
			c[ipCH2][ipCH2P] += CH2_NH2P_NH2_CH2P_2;
			b[ipCH2P] += b_CH2_NH2P_NH2_CH2P;

		/*  >>refer   mol  CH2    +  CN+   =>   CN    +     CH2+  UMIST  */

			c[ipCNP][ipCH2] -= CH2_CNP_CN_CH2P_1;
			c[ipCH2][ipCH2] -= CH2_CNP_CN_CH2P_2;
			b[ipCH2] -= b_CH2_CNP_CN_CH2P;
			
			c[ipCNP][ipCNP] -= CH2_CNP_CN_CH2P_1;
			c[ipCH2][ipCNP] -= CH2_CNP_CN_CH2P_2;
			b[ipCNP] -= b_CH2_CNP_CN_CH2P;
			
			c[ipCNP][ipCN] += CH2_CNP_CN_CH2P_1;
			c[ipCH2][ipCN] += CH2_CNP_CN_CH2P_2;
			b[ipCN] += b_CH2_CNP_CN_CH2P;
			
			c[ipCNP][ipCH2P] += CH2_CNP_CN_CH2P_1;
			c[ipCH2][ipCH2P] += CH2_CNP_CN_CH2P_2;
			b[ipCH2P] += b_CH2_CNP_CN_CH2P;

		/*  >>refer   mol  CH2    +  N2+   =>   N2    +     CH2+  UMIST  */

			c[ipN2P][ipCH2] -= CH2_N2P_N2_CH2P_1;
			c[ipCH2][ipCH2] -= CH2_N2P_N2_CH2P_2;
			b[ipCH2] -= b_CH2_N2P_N2_CH2P;
			
			c[ipN2P][ipN2P] -= CH2_N2P_N2_CH2P_1;
			c[ipCH2][ipN2P] -= CH2_N2P_N2_CH2P_2;
			b[ipN2P] -= b_CH2_N2P_N2_CH2P;
			
			c[ipN2P][ipN2] += CH2_N2P_N2_CH2P_1;
			c[ipCH2][ipN2] += CH2_N2P_N2_CH2P_2;
			b[ipN2] += b_CH2_N2P_N2_CH2P;
			
			c[ipN2P][ipCH2P] += CH2_N2P_N2_CH2P_1;
			c[ipCH2][ipCH2P] += CH2_N2P_N2_CH2P_2;
			b[ipCH2P] += b_CH2_N2P_N2_CH2P;

		/*  >>refer   mol  CH2+   +  NO    =>   NO+   +     CH2   UMIST  */

			c[ipNO][ipCH2P] -= CH2P_NO_NOP_CH2_1;
			c[ipCH2P][ipCH2P] -= CH2P_NO_NOP_CH2_2;
			b[ipCH2P] -= b_CH2P_NO_NOP_CH2;
			
			c[ipNO][ipNO] -= CH2P_NO_NOP_CH2_1;
			c[ipCH2P][ipNO] -= CH2P_NO_NOP_CH2_2;
			b[ipNO] -= b_CH2P_NO_NOP_CH2;
			
			c[ipNO][ipNOP] += CH2P_NO_NOP_CH2_1;
			c[ipCH2P][ipNOP] += CH2P_NO_NOP_CH2_2;
			b[ipNOP] += b_CH2P_NO_NOP_CH2;
			
			c[ipNO][ipCH2] += CH2P_NO_NOP_CH2_1;
			c[ipCH2P][ipCH2] += CH2P_NO_NOP_CH2_2;
			b[ipCH2] += b_CH2P_NO_NOP_CH2;

		/*  >>refer   mol  NH     +  O+    =>   O     +     NH+   UMIST  */

			c[ipOP][ipOP] -= NH_OP_O_NHP_1;
			c[ipNH][ipOP] -= NH_OP_O_NHP_2;
			b[ipOP] -= b_NH_OP_O_NHP;
			
			c[ipOP][ipNH] -= NH_OP_O_NHP_1;
			c[ipNH][ipNH] -= NH_OP_O_NHP_2;
			b[ipNH] -= b_NH_OP_O_NHP;
			
			c[ipOP][ipNHP] += NH_OP_O_NHP_1;
			c[ipNH][ipNHP] += NH_OP_O_NHP_2;
			b[ipNHP] += b_NH_OP_O_NHP;
			
			c[ipOP][ipATO] += NH_OP_O_NHP_1;
			c[ipNH][ipATO] += NH_OP_O_NHP_2;
			b[ipATO] += b_NH_OP_O_NHP;
			
		/*  >>refer   mol  NH     +  OH+   =>   OH    +     NH+   UMIST  */

			c[ipOHP][ipOHP] -= NH_OHP_OH_NHP_1;
			c[ipNH][ipOHP] -= NH_OHP_OH_NHP_2;
			b[ipOHP] -= b_NH_OHP_OH_NHP;
			
			c[ipOHP][ipNH] -= NH_OHP_OH_NHP_1;
			c[ipNH][ipNH] -= NH_OHP_OH_NHP_2;
			b[ipNH] -= b_NH_OHP_OH_NHP;
			
			c[ipOHP][ipOH] += NH_OHP_OH_NHP_1;
			c[ipNH][ipOH] += NH_OHP_OH_NHP_2;
			b[ipOH] += b_NH_OHP_OH_NHP;
			
			c[ipOHP][ipNHP] += NH_OHP_OH_NHP_1;
			c[ipNH][ipNHP] += NH_OHP_OH_NHP_2;
			b[ipNHP] += b_NH_OHP_OH_NHP;

		/*  >>refer   mol  NH+    +  NH3   =>   NH3+  +     NH    UMIST  */

			c[ipNH3][ipNHP] -= NHP_NH3_NH3P_NH_1;
			c[ipNHP][ipNHP] -= NHP_NH3_NH3P_NH_2;
			b[ipNHP] -= b_NHP_NH3_NH3P_NH;
			
			c[ipNH3][ipNH3] -= NHP_NH3_NH3P_NH_1;
			c[ipNHP][ipNH3] -= NHP_NH3_NH3P_NH_2;
			b[ipNH3] -= b_NHP_NH3_NH3P_NH;
			
			c[ipNH3][ipNH3P] += NHP_NH3_NH3P_NH_1;
			c[ipNHP][ipNH3P] += NHP_NH3_NH3P_NH_2;
			b[ipNH3P] += b_NHP_NH3_NH3P_NH;
			
			c[ipNH3][ipNH] += NHP_NH3_NH3P_NH_1;
			c[ipNHP][ipNH] += NHP_NH3_NH3P_NH_2;
			b[ipNH] += b_NHP_NH3_NH3P_NH;

		/*  >>refer   mol  NH+    +  H2O   =>   H2O+  +     NH    UMIST  */

			c[ipH2O][ipNHP] -= NHP_H2O_H2OP_NH_1;
			c[ipNHP][ipNHP] -= NHP_H2O_H2OP_NH_2;
			b[ipNHP] -= b_NHP_H2O_H2OP_NH;
			
			c[ipH2O][ipH2O] -= NHP_H2O_H2OP_NH_1;
			c[ipNHP][ipH2O] -= NHP_H2O_H2OP_NH_2;
			b[ipH2O] -= b_NHP_H2O_H2OP_NH;
			
			c[ipH2O][ipH2OP] += NHP_H2O_H2OP_NH_1;
			c[ipNHP][ipH2OP] += NHP_H2O_H2OP_NH_2;
			b[ipH2OP] += b_NHP_H2O_H2OP_NH;
			
			c[ipH2O][ipNH] += NHP_H2O_H2OP_NH_1;
			c[ipNHP][ipNH] += NHP_H2O_H2OP_NH_2;
			b[ipNH] += b_NHP_H2O_H2OP_NH;

		/*  >>refer   mol  NH     +  CN+   =>   CN    +     NH+   UMIST  */

			c[ipCNP][ipNH] -= NH_CNP_CN_NHP_1;
			c[ipNH][ipNH] -= NH_CNP_CN_NHP_2;
			b[ipNH] -= b_NH_CNP_CN_NHP;
			
			c[ipCNP][ipCNP] -= NH_CNP_CN_NHP_1;
			c[ipNH][ipCNP] -= NH_CNP_CN_NHP_2;
			b[ipCNP] -= b_NH_CNP_CN_NHP;
			
			c[ipCNP][ipCN] += NH_CNP_CN_NHP_1;
			c[ipNH][ipCN] += NH_CNP_CN_NHP_2;
			b[ipCN] += b_NH_CNP_CN_NHP;
			
			c[ipCNP][ipNHP] += NH_CNP_CN_NHP_1;
			c[ipNH][ipNHP] += NH_CNP_CN_NHP_2;
			b[ipNHP] += b_NH_CNP_CN_NHP;

		/*  >>refer   mol  NH     +  CO+   =>   CO    +     NH+   UMIST  */

			c[ipCOP][ipNH] -= NH_COP_CO_NHP_1;
			c[ipNH][ipNH] -= NH_COP_CO_NHP_2;
			b[ipNH] -= b_NH_COP_CO_NHP;
			
			c[ipCOP][ipCOP] -= NH_COP_CO_NHP_1;
			c[ipNH][ipCOP] -= NH_COP_CO_NHP_2;
			b[ipCOP] -= b_NH_COP_CO_NHP;
			
			c[ipCOP][ipCO] += NH_COP_CO_NHP_1;
			c[ipNH][ipCO] += NH_COP_CO_NHP_2;
			b[ipCO] += b_NH_COP_CO_NHP;
			
			c[ipCOP][ipNHP] += NH_COP_CO_NHP_1;
			c[ipNH][ipNHP] += NH_COP_CO_NHP_2;
			b[ipNHP] += b_NH_COP_CO_NHP;

		/*  >>refer   mol  NH     +  N2+   =>   N2    +     NH+   UMIST  */

			c[ipN2P][ipNH] -= NH_N2P_N2_NHP_1;
			c[ipNH][ipNH] -= NH_N2P_N2_NHP_2;
			b[ipNH] -= b_NH_N2P_N2_NHP;
			
			c[ipN2P][ipN2P] -= NH_N2P_N2_NHP_1;
			c[ipNH][ipN2P] -= NH_N2P_N2_NHP_2;
			b[ipN2P] -= b_NH_N2P_N2_NHP;
			
			c[ipN2P][ipN2] += NH_N2P_N2_NHP_1;
			c[ipNH][ipN2] += NH_N2P_N2_NHP_2;
			b[ipN2] += b_NH_N2P_N2_NHP;
			
			c[ipN2P][ipNHP] += NH_N2P_N2_NHP_1;
			c[ipNH][ipNHP] += NH_N2P_N2_NHP_2;
			b[ipNHP] += b_NH_N2P_N2_NHP;

		/*  >>refer   mol  NH+    +  NO    =>   NO+   +     NH    UMIST  */

			c[ipNO][ipNHP] -= NHP_NO_NOP_NH_1;
			c[ipNHP][ipNHP] -= NHP_NO_NOP_NH_2;
			b[ipNHP] -= b_NHP_NO_NOP_NH;
			
			c[ipNO][ipNO] -= NHP_NO_NOP_NH_1;
			c[ipNHP][ipNO] -= NHP_NO_NOP_NH_2;
			b[ipNO] -= b_NHP_NO_NOP_NH;
			
			c[ipNO][ipNOP] += NHP_NO_NOP_NH_1;
			c[ipNHP][ipNOP] += NHP_NO_NOP_NH_2;
			b[ipNOP] += b_NHP_NO_NOP_NH;
			
			c[ipNO][ipNH] += NHP_NO_NOP_NH_1;
			c[ipNHP][ipNH] += NHP_NO_NOP_NH_2;
			b[ipNH] += b_NHP_NO_NOP_NH;

		/*  >>refer   mol  NH+    +  O2    =>   O2+   +     NH    UMIST  */

			c[ipO2][ipNHP] -= NHP_O2_O2P_NH_1;
			c[ipNHP][ipNHP] -= NHP_O2_O2P_NH_2;
			b[ipNHP] -= b_NHP_O2_O2P_NH;
			
			c[ipO2][ipO2] -= NHP_O2_O2P_NH_1;
			c[ipNHP][ipO2] -= NHP_O2_O2P_NH_2;
			b[ipO2] -= b_NHP_O2_O2P_NH;
			
			c[ipO2][ipO2P] += NHP_O2_O2P_NH_1;
			c[ipNHP][ipO2P] += NHP_O2_O2P_NH_2;
			b[ipO2P] += b_NHP_O2_O2P_NH;
			
			c[ipO2][ipNH] += NHP_O2_O2P_NH_1;
			c[ipNHP][ipNH] += NHP_O2_O2P_NH_2;
			b[ipNH] += b_NHP_O2_O2P_NH;

		/*  >>refer   mol  NH+    +  S     =>   S+    +     NH    UMIST  */
			
			c[ipATS][ipNHP] -= NHP_S_SP_NH_1;
			c[ipNHP][ipNHP] -= NHP_S_SP_NH_2;
			b[ipNHP] -= b_NHP_S_SP_NH;
			
			c[ipATS][ipATS] -= NHP_S_SP_NH_1;
			c[ipNHP][ipATS] -= NHP_S_SP_NH_2;
			b[ipATS] -= b_NHP_S_SP_NH;
			
			c[ipATS][ipSP] += NHP_S_SP_NH_1;
			c[ipNHP][ipSP] += NHP_S_SP_NH_2;
			b[ipSP] += b_NHP_S_SP_NH;
			
			c[ipATS][ipNH] += NHP_S_SP_NH_1;
			c[ipNHP][ipNH] += NHP_S_SP_NH_2;
			b[ipNH] += b_NHP_S_SP_NH;

		/*  >>refer   mol  CH3+   +  NO    =>   NO+   +     CH3   UMIST  */

			c[ipNO][ipCH3P] -= CH3P_NO_NOP_CH3_1;
			c[ipCH3P][ipCH3P] -= CH3P_NO_NOP_CH3_2;
			b[ipCH3P] -= b_CH3P_NO_NOP_CH3;
			
			c[ipNO][ipNO] -= CH3P_NO_NOP_CH3_1;
			c[ipCH3P][ipNO] -= CH3P_NO_NOP_CH3_2;
			b[ipNO] -= b_CH3P_NO_NOP_CH3;
			
			c[ipNO][ipNOP] += CH3P_NO_NOP_CH3_1;
			c[ipCH3P][ipNOP] += CH3P_NO_NOP_CH3_2;
			b[ipNOP] += b_CH3P_NO_NOP_CH3;
			
			c[ipNO][ipCH3] += CH3P_NO_NOP_CH3_1;
			c[ipCH3P][ipCH3] += CH3P_NO_NOP_CH3_2;
			b[ipCH3] += b_CH3P_NO_NOP_CH3;

		/*  >>refer   mol  O+     +  NH2   =>   NH2+  +     O     UMIST  */

			c[ipNH2][ipOP] -= OP_NH2_NH2P_O_1;
			c[ipOP][ipOP] -= OP_NH2_NH2P_O_2;
			b[ipOP] -= b_OP_NH2_NH2P_O;
			
			c[ipNH2][ipNH2] -= OP_NH2_NH2P_O_1;
			c[ipOP][ipNH2] -= OP_NH2_NH2P_O_2;
			b[ipNH2] -= b_OP_NH2_NH2P_O;
			
			c[ipNH2][ipNH2P] += OP_NH2_NH2P_O_1;
			c[ipOP][ipNH2P] += OP_NH2_NH2P_O_2;
			b[ipNH2P] += b_OP_NH2_NH2P_O;
			
			c[ipNH2][ipATO] += OP_NH2_NH2P_O_1;
			c[ipOP][ipATO] += OP_NH2_NH2P_O_2;
			b[ipATO] += b_OP_NH2_NH2P_O;
			
		/*  >>refer   mol  O+     +  NH3   =>   NH3+  +     O     UMIST  */

			c[ipNH3][ipOP] -= OP_NH3_NH3P_O_1;
			c[ipOP][ipOP] -= OP_NH3_NH3P_O_2;
			b[ipOP] -= b_OP_NH3_NH3P_O;
			
			c[ipNH3][ipNH3] -= OP_NH3_NH3P_O_1;
			c[ipOP][ipNH3] -= OP_NH3_NH3P_O_2;
			b[ipNH3] -= b_OP_NH3_NH3P_O;
			
			c[ipNH3][ipNH3P] += OP_NH3_NH3P_O_1;
			c[ipOP][ipNH3P] += OP_NH3_NH3P_O_2;
			b[ipNH3P] += b_OP_NH3_NH3P_O;
			
			c[ipNH3][ipATO] += OP_NH3_NH3P_O_1;
			c[ipOP][ipATO] += OP_NH3_NH3P_O_2;
			b[ipATO] += b_OP_NH3_NH3P_O;
			
		/*  >>refer   mol  O      +  CN+   =>   CN    +     O+    UMIST  */

			c[ipCNP][ipCNP] -= O_CNP_CN_OP_1;
			c[ipATO][ipCNP] -= O_CNP_CN_OP_2;
			b[ipCNP] -= b_O_CNP_CN_OP;

			c[ipCNP][ipATO] -= O_CNP_CN_OP_1;
			c[ipATO][ipATO] -= O_CNP_CN_OP_2;
			b[ipATO] -= b_O_CNP_CN_OP;
			
			c[ipCNP][ipCN] += O_CNP_CN_OP_1;
			c[ipATO][ipCN] += O_CNP_CN_OP_2;
			b[ipCN] += b_O_CNP_CN_OP;
			
			c[ipCNP][ipOP] += O_CNP_CN_OP_1;
			c[ipATO][ipOP] += O_CNP_CN_OP_2;
			b[ipOP] += b_O_CNP_CN_OP;

		/*  >>refer   mol  O      +  HCN+  =>   HCN   +     O+    UMIST  */

			c[ipHCNP][ipHCNP] -= O_HCNP_HCN_OP_1;
			c[ipATO][ipHCNP] -= O_HCNP_HCN_OP_2;
			b[ipHCNP] -= b_O_HCNP_HCN_OP;

			c[ipHCNP][ipATO] -= O_HCNP_HCN_OP_1;
			c[ipATO][ipATO] -= O_HCNP_HCN_OP_2;
			b[ipATO] -= b_O_HCNP_HCN_OP;
			
			c[ipHCNP][ipHCN] += O_HCNP_HCN_OP_1;
			c[ipATO][ipHCN] += O_HCNP_HCN_OP_2;
			b[ipHCN] += b_O_HCNP_HCN_OP;
			
			c[ipHCNP][ipOP] += O_HCNP_HCN_OP_1;
			c[ipATO][ipOP] += O_HCNP_HCN_OP_2;
			b[ipOP] += b_O_HCNP_HCN_OP;

		/*  >>refer   mol  O      +  N2+   =>   N2    +     O+    UMIST  */

			c[ipN2P][ipN2P] -= O_N2P_N2_OP_1;
			c[ipATO][ipN2P] -= O_N2P_N2_OP_2;
			b[ipN2P] -= b_O_N2P_N2_OP;

			c[ipN2P][ipATO] -= O_N2P_N2_OP_1;
			c[ipATO][ipATO] -= O_N2P_N2_OP_2;
			b[ipATO] -= b_O_N2P_N2_OP;
			
			c[ipN2P][ipN2] += O_N2P_N2_OP_1;
			c[ipATO][ipN2] += O_N2P_N2_OP_2;
			b[ipN2] += b_O_N2P_N2_OP;
			
			c[ipN2P][ipOP] += O_N2P_N2_OP_1;
			c[ipATO][ipOP] += O_N2P_N2_OP_2;
			b[ipOP] += b_O_N2P_N2_OP;

		/*  >>refer   mol  O+     +  NO    =>   NO+   +     O     UMIST  */

			c[ipNO][ipOP] -= OP_NO_NOP_O_1;
			c[ipOP][ipOP] -= OP_NO_NOP_O_2;
			b[ipOP] -= b_OP_NO_NOP_O;
			
			c[ipNO][ipNO] -= OP_NO_NOP_O_1;
			c[ipOP][ipNO] -= OP_NO_NOP_O_2;
			b[ipNO] -= b_OP_NO_NOP_O;
			
			c[ipNO][ipNOP] += OP_NO_NOP_O_1;
			c[ipOP][ipNOP] += OP_NO_NOP_O_2;
			b[ipNOP] += b_OP_NO_NOP_O;
			
			c[ipNO][ipATO] += OP_NO_NOP_O_1;
			c[ipOP][ipATO] += OP_NO_NOP_O_2;
			b[ipATO] += b_OP_NO_NOP_O;
			
		/*  >>refer   mol  O+     +  OCS   =>   OCS+  +     O     UMIST  */

			c[ipOCS][ipOP] -= OP_OCS_OCSP_O_1;
			c[ipOP][ipOP] -= OP_OCS_OCSP_O_2;
			b[ipOP] -= b_OP_OCS_OCSP_O;
			
			c[ipOCS][ipOCS] -= OP_OCS_OCSP_O_1;
			c[ipOP][ipOCS] -= OP_OCS_OCSP_O_2;
			b[ipOCS] -= b_OP_OCS_OCSP_O;
			
			c[ipOCS][ipOCSP] += OP_OCS_OCSP_O_1;
			c[ipOP][ipOCSP] += OP_OCS_OCSP_O_2;
			b[ipOCSP] += b_OP_OCS_OCSP_O;
			
			c[ipOCS][ipATO] += OP_OCS_OCSP_O_1;
			c[ipOP][ipATO] += OP_OCS_OCSP_O_2;
			b[ipATO] += b_OP_OCS_OCSP_O;

		/*  >>refer   mol  NH2    +  OH+   =>   OH    +     NH2+  UMIST  */

			c[ipOHP][ipNH2] -= NH2_OHP_OH_NH2P_1;
			c[ipNH2][ipNH2] -= NH2_OHP_OH_NH2P_2;
			b[ipNH2] -= b_NH2_OHP_OH_NH2P;
			
			c[ipOHP][ipOHP] -= NH2_OHP_OH_NH2P_1;
			c[ipNH2][ipOHP] -= NH2_OHP_OH_NH2P_2;
			b[ipOHP] -= b_NH2_OHP_OH_NH2P;
			
			c[ipOHP][ipOH] += NH2_OHP_OH_NH2P_1;
			c[ipNH2][ipOH] += NH2_OHP_OH_NH2P_2;
			b[ipOH] += b_NH2_OHP_OH_NH2P;
			
			c[ipOHP][ipNH2P] += NH2_OHP_OH_NH2P_1;
			c[ipNH2][ipNH2P] += NH2_OHP_OH_NH2P_2;
			b[ipNH2P] += b_NH2_OHP_OH_NH2P;

		/*  >>refer   mol  NH2+   +  NH3   =>   NH3+  +     NH2   UMIST  */

			c[ipNH3][ipNH2P] -= NH2P_NH3_NH3P_NH2_1;
			c[ipNH2P][ipNH2P] -= NH2P_NH3_NH3P_NH2_2;
			b[ipNH2P] -= b_NH2P_NH3_NH3P_NH2;
			
			c[ipNH3][ipNH3] -= NH2P_NH3_NH3P_NH2_1;
			c[ipNH2P][ipNH3] -= NH2P_NH3_NH3P_NH2_2;
			b[ipNH3] -= b_NH2P_NH3_NH3P_NH2;
			
			c[ipNH3][ipNH3P] += NH2P_NH3_NH3P_NH2_1;
			c[ipNH2P][ipNH3P] += NH2P_NH3_NH3P_NH2_2;
			b[ipNH3P] += b_NH2P_NH3_NH3P_NH2;
			
			c[ipNH3][ipNH2] += NH2P_NH3_NH3P_NH2_1;
			c[ipNH2P][ipNH2] += NH2P_NH3_NH3P_NH2_2;
			b[ipNH2] += b_NH2P_NH3_NH3P_NH2;

		/*  >>refer   mol  NH2    +  H2O+  =>   H2O   +     NH2+  UMIST  */

			c[ipH2OP][ipNH2] -= NH2_H2OP_H2O_NH2P_1;
			c[ipNH2][ipNH2] -= NH2_H2OP_H2O_NH2P_2;
			b[ipNH2] -= b_NH2_H2OP_H2O_NH2P;
			
			c[ipH2OP][ipH2OP] -= NH2_H2OP_H2O_NH2P_1;
			c[ipNH2][ipH2OP] -= NH2_H2OP_H2O_NH2P_2;
			b[ipH2OP] -= b_NH2_H2OP_H2O_NH2P;
			
			c[ipH2OP][ipH2O] += NH2_H2OP_H2O_NH2P_1;
			c[ipNH2][ipH2O] += NH2_H2OP_H2O_NH2P_2;
			b[ipH2O] += b_NH2_H2OP_H2O_NH2P;
			
			c[ipH2OP][ipNH2P] += NH2_H2OP_H2O_NH2P_1;
			c[ipNH2][ipNH2P] += NH2_H2OP_H2O_NH2P_2;
			b[ipNH2P] += b_NH2_H2OP_H2O_NH2P;

		/*  >>refer   mol  NH2    +  CN+   =>   CN    +     NH2+  UMIST  */

			c[ipCNP][ipNH2] -= NH2_CNP_CN_NH2P_1;
			c[ipNH2][ipNH2] -= NH2_CNP_CN_NH2P_2;
			b[ipNH2] -= b_NH2_CNP_CN_NH2P;
			
			c[ipCNP][ipCNP] -= NH2_CNP_CN_NH2P_1;
			c[ipNH2][ipCNP] -= NH2_CNP_CN_NH2P_2;
			b[ipCNP] -= b_NH2_CNP_CN_NH2P;
			
			c[ipCNP][ipCN] += NH2_CNP_CN_NH2P_1;
			c[ipNH2][ipCN] += NH2_CNP_CN_NH2P_2;
			b[ipCN] += b_NH2_CNP_CN_NH2P;
			
			c[ipCNP][ipNH2P] += NH2_CNP_CN_NH2P_1;
			c[ipNH2][ipNH2P] += NH2_CNP_CN_NH2P_2;
			b[ipNH2P] += b_NH2_CNP_CN_NH2P;

		/*  >>refer   mol  NH2    +  CO+   =>   CO    +     NH2+  UMIST  */

			c[ipCOP][ipNH2] -= NH2_COP_CO_NH2P_1;
			c[ipNH2][ipNH2] -= NH2_COP_CO_NH2P_2;
			b[ipNH2] -= b_NH2_COP_CO_NH2P;
			
			c[ipCOP][ipCOP] -= NH2_COP_CO_NH2P_1;
			c[ipNH2][ipCOP] -= NH2_COP_CO_NH2P_2;
			b[ipCOP] -= b_NH2_COP_CO_NH2P;
			
			c[ipCOP][ipCO] += NH2_COP_CO_NH2P_1;
			c[ipNH2][ipCO] += NH2_COP_CO_NH2P_2;
			b[ipCO] += b_NH2_COP_CO_NH2P;
			
			c[ipCOP][ipNH2P] += NH2_COP_CO_NH2P_1;
			c[ipNH2][ipNH2P] += NH2_COP_CO_NH2P_2;
			b[ipNH2P] += b_NH2_COP_CO_NH2P;

		/*  >>refer   mol  NH2    +  N2+   =>   N2    +     NH2+  UMIST  */

			c[ipN2P][ipNH2] -= NH2_N2P_N2_NH2P_1;
			c[ipNH2][ipNH2] -= NH2_N2P_N2_NH2P_2;
			b[ipNH2] -= b_NH2_N2P_N2_NH2P;
			
			c[ipN2P][ipN2P] -= NH2_N2P_N2_NH2P_1;
			c[ipNH2][ipN2P] -= NH2_N2P_N2_NH2P_2;
			b[ipN2P] -= b_NH2_N2P_N2_NH2P;
			
			c[ipN2P][ipN2] += NH2_N2P_N2_NH2P_1;
			c[ipNH2][ipN2] += NH2_N2P_N2_NH2P_2;
			b[ipN2] += b_NH2_N2P_N2_NH2P;
			
			c[ipN2P][ipNH2P] += NH2_N2P_N2_NH2P_1;
			c[ipNH2][ipNH2P] += NH2_N2P_N2_NH2P_2;
			b[ipNH2P] += b_NH2_N2P_N2_NH2P;

		/*  >>refer   mol  NH2+   +  NO    =>   NO+   +     NH2   UMIST  */

			c[ipNO][ipNH2P] -= NH2P_NO_NOP_NH2_1;
			c[ipNH2P][ipNH2P] -= NH2P_NO_NOP_NH2_2;
			b[ipNH2P] -= b_NH2P_NO_NOP_NH2;
			
			c[ipNO][ipNO] -= NH2P_NO_NOP_NH2_1;
			c[ipNH2P][ipNO] -= NH2P_NO_NOP_NH2_2;
			b[ipNO] -= b_NH2P_NO_NOP_NH2;
			
			c[ipNO][ipNOP] += NH2P_NO_NOP_NH2_1;
			c[ipNH2P][ipNOP] += NH2P_NO_NOP_NH2_2;
			b[ipNOP] += b_NH2P_NO_NOP_NH2;
			
			c[ipNO][ipNH2] += NH2P_NO_NOP_NH2_1;
			c[ipNH2P][ipNH2] += NH2P_NO_NOP_NH2_2;
			b[ipNH2] += b_NH2P_NO_NOP_NH2;

		/*  >>refer   mol  NH2    +  O2+   =>   O2    +     NH2+  UMIST  */

			c[ipO2P][ipNH2] -= NH2_O2P_O2_NH2P_1;
			c[ipNH2][ipNH2] -= NH2_O2P_O2_NH2P_2;
			b[ipNH2] -= b_NH2_O2P_O2_NH2P;
			
			c[ipO2P][ipO2P] -= NH2_O2P_O2_NH2P_1;
			c[ipNH2][ipO2P] -= NH2_O2P_O2_NH2P_2;
			b[ipO2P] -= b_NH2_O2P_O2_NH2P;
			
			c[ipO2P][ipO2] += NH2_O2P_O2_NH2P_1;
			c[ipNH2][ipO2] += NH2_O2P_O2_NH2P_2;
			b[ipO2] += b_NH2_O2P_O2_NH2P;
			
			c[ipO2P][ipNH2P] += NH2_O2P_O2_NH2P_1;
			c[ipNH2][ipNH2P] += NH2_O2P_O2_NH2P_2;
			b[ipNH2P] += b_NH2_O2P_O2_NH2P;

		/*  >>refer   mol  NH2+   +  S     =>   S+    +     NH2   UMIST  */

			c[ipATS][ipNH2P] -= NH2P_S_SP_NH2_1;
			c[ipNH2P][ipNH2P] -= NH2P_S_SP_NH2_2;
			b[ipNH2P] -= b_NH2P_S_SP_NH2;

			c[ipATS][ipATS] -= NH2P_S_SP_NH2_1;
			c[ipNH2P][ipATS] -= NH2P_S_SP_NH2_2;
			b[ipATS] -= b_NH2P_S_SP_NH2;
			
			c[ipATS][ipSP] += NH2P_S_SP_NH2_1;
			c[ipNH2P][ipSP] += NH2P_S_SP_NH2_2;
			b[ipSP] += b_NH2P_S_SP_NH2;
			
			c[ipATS][ipNH2] += NH2P_S_SP_NH2_1;
			c[ipNH2P][ipNH2] += NH2P_S_SP_NH2_2;
			b[ipNH2] += b_NH2P_S_SP_NH2;

		/*  >>refer   mol  CH4+   +  NH3   =>   NH3+  +     CH4   UMIST  */

			c[ipNH3][ipCH4P] -= CH4P_NH3_NH3P_CH4_1;
			c[ipCH4P][ipCH4P] -= CH4P_NH3_NH3P_CH4_2;
			b[ipCH4P] -= b_CH4P_NH3_NH3P_CH4;
			
			c[ipNH3][ipNH3] -= CH4P_NH3_NH3P_CH4_1;
			c[ipCH4P][ipNH3] -= CH4P_NH3_NH3P_CH4_2;
			b[ipNH3] -= b_CH4P_NH3_NH3P_CH4;
			
			c[ipNH3][ipNH3P] += CH4P_NH3_NH3P_CH4_1;
			c[ipCH4P][ipNH3P] += CH4P_NH3_NH3P_CH4_2;
			b[ipNH3P] += b_CH4P_NH3_NH3P_CH4;
			
			c[ipNH3][ipCH4] += CH4P_NH3_NH3P_CH4_1;
			c[ipCH4P][ipCH4] += CH4P_NH3_NH3P_CH4_2;
			b[ipCH4] += b_CH4P_NH3_NH3P_CH4;

		/*  >>refer   mol  CH4+   +  OCS   =>   OCS+  +     CH4   UMIST  */

			c[ipOCS][ipCH4P] -= CH4P_OCS_OCSP_CH4_1;
			c[ipCH4P][ipCH4P] -= CH4P_OCS_OCSP_CH4_2;
			b[ipCH4P] -= b_CH4P_OCS_OCSP_CH4;
			
			c[ipOCS][ipOCS] -= CH4P_OCS_OCSP_CH4_1;
			c[ipCH4P][ipOCS] -= CH4P_OCS_OCSP_CH4_2;
			b[ipOCS] -= b_CH4P_OCS_OCSP_CH4;
			
			c[ipOCS][ipOCSP] += CH4P_OCS_OCSP_CH4_1;
			c[ipCH4P][ipOCSP] += CH4P_OCS_OCSP_CH4_2;
			b[ipOCSP] += b_CH4P_OCS_OCSP_CH4;
			
			c[ipOCS][ipCH4] += CH4P_OCS_OCSP_CH4_1;
			c[ipCH4P][ipCH4] += CH4P_OCS_OCSP_CH4_2;
			b[ipCH4] += b_CH4P_OCS_OCSP_CH4;

		/*  >>refer   mol  OH+    +  NH3   =>   NH3+  +     OH    UMIST  */

			c[ipNH3][ipOHP] -= OHP_NH3_NH3P_OH_1;
			c[ipOHP][ipOHP] -= OHP_NH3_NH3P_OH_2;
			b[ipOHP] -= b_OHP_NH3_NH3P_OH;
			
			c[ipNH3][ipNH3] -= OHP_NH3_NH3P_OH_1;
			c[ipOHP][ipNH3] -= OHP_NH3_NH3P_OH_2;
			b[ipNH3] -= b_OHP_NH3_NH3P_OH;
			
			c[ipNH3][ipNH3P] += OHP_NH3_NH3P_OH_1;
			c[ipOHP][ipNH3P] += OHP_NH3_NH3P_OH_2;
			b[ipNH3P] += b_OHP_NH3_NH3P_OH;
			
			c[ipNH3][ipOH] += OHP_NH3_NH3P_OH_1;
			c[ipOHP][ipOH] += OHP_NH3_NH3P_OH_2;
			b[ipOH] += b_OHP_NH3_NH3P_OH;

		/*  >>refer   mol  OH     +  CN+   =>   CN    +     OH+   UMIST  */

			c[ipCNP][ipOH] -= OH_CNP_CN_OHP_1;
			c[ipOH][ipOH] -= OH_CNP_CN_OHP_2;
			b[ipOH] -= b_OH_CNP_CN_OHP;
			
			c[ipCNP][ipCNP] -= OH_CNP_CN_OHP_1;
			c[ipOH][ipCNP] -= OH_CNP_CN_OHP_2;
			b[ipCNP] -= b_OH_CNP_CN_OHP;
			
			c[ipCNP][ipCN] += OH_CNP_CN_OHP_1;
			c[ipOH][ipCN] += OH_CNP_CN_OHP_2;
			b[ipCN] += b_OH_CNP_CN_OHP;
			
			c[ipCNP][ipOHP] += OH_CNP_CN_OHP_1;
			c[ipOH][ipOHP] += OH_CNP_CN_OHP_2;
			b[ipOHP] += b_OH_CNP_CN_OHP;

		/*  >>refer   mol  OH     +  N2+   =>   N2    +     OH+   UMIST  */

			c[ipN2P][ipOH] -= OH_N2P_N2_OHP_1;
			c[ipOH][ipOH] -= OH_N2P_N2_OHP_2;
			b[ipOH] -= b_OH_N2P_N2_OHP;
			
			c[ipN2P][ipN2P] -= OH_N2P_N2_OHP_1;
			c[ipOH][ipN2P] -= OH_N2P_N2_OHP_2;
			b[ipN2P] -= b_OH_N2P_N2_OHP;
			
			c[ipN2P][ipN2] += OH_N2P_N2_OHP_1;
			c[ipOH][ipN2] += OH_N2P_N2_OHP_2;
			b[ipN2] += b_OH_N2P_N2_OHP;
			
			c[ipN2P][ipOHP] += OH_N2P_N2_OHP_1;
			c[ipOH][ipOHP] += OH_N2P_N2_OHP_2;
			b[ipOHP] += b_OH_N2P_N2_OHP;

		/*  >>refer   mol  OH+    +  NO    =>   NO+   +     OH    UMIST  */

			c[ipNO][ipOHP] -= OHP_NO_NOP_OH_1;
			c[ipOHP][ipOHP] -= OHP_NO_NOP_OH_2;
			b[ipOHP] -= b_OHP_NO_NOP_OH;
			
			c[ipNO][ipNO] -= OHP_NO_NOP_OH_1;
			c[ipOHP][ipNO] -= OHP_NO_NOP_OH_2;
			b[ipNO] -= b_OHP_NO_NOP_OH;
			
			c[ipNO][ipNOP] += OHP_NO_NOP_OH_1;
			c[ipOHP][ipNOP] += OHP_NO_NOP_OH_2;
			b[ipNOP] += b_OHP_NO_NOP_OH;
			
			c[ipNO][ipOH] += OHP_NO_NOP_OH_1;
			c[ipOHP][ipOH] += OHP_NO_NOP_OH_2;
			b[ipOH] += b_OHP_NO_NOP_OH;

		/*  >>refer   mol  NH3    +  H2O+  =>   H2O   +     NH3+  UMIST  */

			c[ipH2OP][ipNH3] -= NH3_H2OP_H2O_NH3P_1;
			c[ipNH3][ipNH3] -= NH3_H2OP_H2O_NH3P_2;
			b[ipNH3] -= b_NH3_H2OP_H2O_NH3P;
			
			c[ipH2OP][ipH2OP] -= NH3_H2OP_H2O_NH3P_1;
			c[ipNH3][ipH2OP] -= NH3_H2OP_H2O_NH3P_2;
			b[ipH2OP] -= b_NH3_H2OP_H2O_NH3P;
			
			c[ipH2OP][ipH2O] += NH3_H2OP_H2O_NH3P_1;
			c[ipNH3][ipH2O] += NH3_H2OP_H2O_NH3P_2;
			b[ipH2O] += b_NH3_H2OP_H2O_NH3P;
			
			c[ipH2OP][ipNH3P] += NH3_H2OP_H2O_NH3P_1;
			c[ipNH3][ipNH3P] += NH3_H2OP_H2O_NH3P_2;
			b[ipNH3P] += b_NH3_H2OP_H2O_NH3P;

		/*  >>refer   mol  NH3    +  HCN+  =>   HCN   +     NH3+  UMIST  */

			c[ipHCNP][ipNH3] -= NH3_HCNP_HCN_NH3P_1;
			c[ipNH3][ipNH3] -= NH3_HCNP_HCN_NH3P_2;
			b[ipNH3] -= b_NH3_HCNP_HCN_NH3P;
			
			c[ipHCNP][ipHCNP] -= NH3_HCNP_HCN_NH3P_1;
			c[ipNH3][ipHCNP] -= NH3_HCNP_HCN_NH3P_2;
			b[ipHCNP] -= b_NH3_HCNP_HCN_NH3P;
			
			c[ipHCNP][ipHCN] += NH3_HCNP_HCN_NH3P_1;
			c[ipNH3][ipHCN] += NH3_HCNP_HCN_NH3P_2;
			b[ipHCN] += b_NH3_HCNP_HCN_NH3P;
			
			c[ipHCNP][ipNH3P] += NH3_HCNP_HCN_NH3P_1;
			c[ipNH3][ipNH3P] += NH3_HCNP_HCN_NH3P_2;
			b[ipNH3P] += b_NH3_HCNP_HCN_NH3P;

		/*  >>refer   mol  NH3    +  CO+   =>   CO    +     NH3+  UMIST  */

			c[ipCOP][ipNH3] -= NH3_COP_CO_NH3P_1;
			c[ipNH3][ipNH3] -= NH3_COP_CO_NH3P_2;
			b[ipNH3] -= b_NH3_COP_CO_NH3P;
			
			c[ipCOP][ipCOP] -= NH3_COP_CO_NH3P_1;
			c[ipNH3][ipCOP] -= NH3_COP_CO_NH3P_2;
			b[ipCOP] -= b_NH3_COP_CO_NH3P;
			
			c[ipCOP][ipCO] += NH3_COP_CO_NH3P_1;
			c[ipNH3][ipCO] += NH3_COP_CO_NH3P_2;
			b[ipCO] += b_NH3_COP_CO_NH3P;
			
			c[ipCOP][ipNH3P] += NH3_COP_CO_NH3P_1;
			c[ipNH3][ipNH3P] += NH3_COP_CO_NH3P_2;
			b[ipNH3P] += b_NH3_COP_CO_NH3P;

		/*  >>refer   mol  NH3    +  N2+   =>   N2    +     NH3+  UMIST  */

			c[ipN2P][ipNH3] -= NH3_N2P_N2_NH3P_1;
			c[ipNH3][ipNH3] -= NH3_N2P_N2_NH3P_2;
			b[ipNH3] -= b_NH3_N2P_N2_NH3P;
			
			c[ipN2P][ipN2P] -= NH3_N2P_N2_NH3P_1;
			c[ipNH3][ipN2P] -= NH3_N2P_N2_NH3P_2;
			b[ipN2P] -= b_NH3_N2P_N2_NH3P;
			
			c[ipN2P][ipN2] += NH3_N2P_N2_NH3P_1;
			c[ipNH3][ipN2] += NH3_N2P_N2_NH3P_2;
			b[ipN2] += b_NH3_N2P_N2_NH3P;
			
			c[ipN2P][ipNH3P] += NH3_N2P_N2_NH3P_1;
			c[ipNH3][ipNH3P] += NH3_N2P_N2_NH3P_2;
			b[ipNH3P] += b_NH3_N2P_N2_NH3P;

		/*  >>refer   mol  NH3+   +  SI    =>   SI+   +     NH3   UMIST  */

			c[ipATSI][ipNH3P] -= NH3P_SI_SIP_NH3_1;
			c[ipNH3P][ipNH3P] -= NH3P_SI_SIP_NH3_2;
			b[ipNH3P] -= b_NH3P_SI_SIP_NH3;

			c[ipATSI][ipATSI] -= NH3P_SI_SIP_NH3_1;
			c[ipNH3P][ipATSI] -= NH3P_SI_SIP_NH3_2;
			b[ipATSI] -= b_NH3P_SI_SIP_NH3;
			
			c[ipATSI][ipSIP] += NH3P_SI_SIP_NH3_1;
			c[ipNH3P][ipSIP] += NH3P_SI_SIP_NH3_2;
			b[ipSIP] += b_NH3P_SI_SIP_NH3;
			
			c[ipATSI][ipNH3] += NH3P_SI_SIP_NH3_1;
			c[ipNH3P][ipNH3] += NH3P_SI_SIP_NH3_2;
			b[ipNH3] += b_NH3P_SI_SIP_NH3;

		/*  >>refer   mol  NH3+   +  NO    =>   NO+   +     NH3   UMIST  */

			c[ipNO][ipNH3P] -= NH3P_SI_SIP_NH3_1;
			c[ipNH3P][ipNH3P] -= NH3P_SI_SIP_NH3_2;
			b[ipNH3P] -= b_NH3P_SI_SIP_NH3;
			
			c[ipNO][ipNO] -= NH3P_SI_SIP_NH3_1;
			c[ipNH3P][ipNO] -= NH3P_SI_SIP_NH3_2;
			b[ipNO] -= b_NH3P_SI_SIP_NH3;
			
			c[ipNO][ipNOP] += NH3P_SI_SIP_NH3_1;
			c[ipNH3P][ipNOP] += NH3P_SI_SIP_NH3_2;
			b[ipNOP] += b_NH3P_SI_SIP_NH3;
			
			c[ipNO][ipNH3] += NH3P_SI_SIP_NH3_1;
			c[ipNH3P][ipNH3] += NH3P_SI_SIP_NH3_2;
			b[ipNH3] += b_NH3P_SI_SIP_NH3;

		/*  >>refer   mol  NH3    +  O2+   =>   O2    +     NH3+  UMIST  */

			c[ipO2P][ipNH3] -= NH3_O2P_O2_NH3P_1;
			c[ipNH3][ipNH3] -= NH3_O2P_O2_NH3P_2;
			b[ipNH3] -= b_NH3_O2P_O2_NH3P;
			
			c[ipO2P][ipO2P] -= NH3_O2P_O2_NH3P_1;
			c[ipNH3][ipO2P] -= NH3_O2P_O2_NH3P_2;
			b[ipO2P] -= b_NH3_O2P_O2_NH3P;
			
			c[ipO2P][ipO2] += NH3_O2P_O2_NH3P_1;
			c[ipNH3][ipO2] += NH3_O2P_O2_NH3P_2;
			b[ipO2] += b_NH3_O2P_O2_NH3P;
			
			c[ipO2P][ipNH3P] += NH3_O2P_O2_NH3P_1;
			c[ipNH3][ipNH3P] += NH3_O2P_O2_NH3P_2;
			b[ipNH3P] += b_NH3_O2P_O2_NH3P;

		/*  >>refer   mol  NH3    +  S+    =>   S     +     NH3+  UMIST  */

			c[ipSP][ipNH3] -= NH3_SP_S_NH3P_1;
			c[ipNH3][ipNH3] -= NH3_SP_S_NH3P_2;
			b[ipNH3] -= b_NH3_SP_S_NH3P;
			
			c[ipSP][ipSP] -= NH3_SP_S_NH3P_1;
			c[ipNH3][ipSP] -= NH3_SP_S_NH3P_2;
			b[ipSP] -= b_NH3_SP_S_NH3P;
			
			c[ipSP][ipNH3P] += NH3_SP_S_NH3P_1;
			c[ipNH3][ipNH3P] += NH3_SP_S_NH3P_2;
			b[ipNH3P] += b_NH3_SP_S_NH3P;
			
			c[ipSP][ipATS] += NH3_SP_S_NH3P_1;
			c[ipNH3][ipATS] += NH3_SP_S_NH3P_2;
			b[ipATS] += b_NH3_SP_S_NH3P;
			
		/*  >>refer   mol  NH3    +  HS+   =>   HS    +     NH3+  UMIST  */

			c[ipHSP][ipNH3] -= NH3_HSP_HS_NH3P_1;
			c[ipNH3][ipNH3] -= NH3_HSP_HS_NH3P_2;
			b[ipNH3] -= b_NH3_HSP_HS_NH3P;
			
			c[ipHSP][ipHSP] -= NH3_HSP_HS_NH3P_1;
			c[ipNH3][ipHSP] -= NH3_HSP_HS_NH3P_2;
			b[ipHSP] -= b_NH3_HSP_HS_NH3P;
			
			c[ipHSP][ipHS] += NH3_HSP_HS_NH3P_1;
			c[ipNH3][ipHS] += NH3_HSP_HS_NH3P_2;
			b[ipHS] += b_NH3_HSP_HS_NH3P;
			
			c[ipHSP][ipNH3P] += NH3_HSP_HS_NH3P_1;
			c[ipNH3][ipNH3P] += NH3_HSP_HS_NH3P_2;
			b[ipNH3P] += b_NH3_HSP_HS_NH3P;

		/*  >>refer   mol  NH3    +  SO+   =>   SO    +     NH3+  UMIST  */

			c[ipSOP][ipNH3] -= NH3_SOP_SO_NH3P_1;
			c[ipNH3][ipNH3] -= NH3_SOP_SO_NH3P_2;
			b[ipNH3] -= b_NH3_SOP_SO_NH3P;
			
			c[ipSOP][ipSOP] -= NH3_SOP_SO_NH3P_1;
			c[ipNH3][ipSOP] -= NH3_SOP_SO_NH3P_2;
			b[ipSOP] -= b_NH3_SOP_SO_NH3P;
			
			c[ipSOP][ipSO] += NH3_SOP_SO_NH3P_1;
			c[ipNH3][ipSO] += NH3_SOP_SO_NH3P_2;
			b[ipSO] += b_NH3_SOP_SO_NH3P;
			
			c[ipSOP][ipNH3P] += NH3_SOP_SO_NH3P_1;
			c[ipNH3][ipNH3P] += NH3_SOP_SO_NH3P_2;
			b[ipNH3P] += b_NH3_SOP_SO_NH3P;

		/*  >>refer   mol  H2O    +  HCN+  =>   HCN   +     H2O+  UMIST  */

			c[ipHCNP][ipH2O] -= H2O_HCNP_HCN_H2OP_1;
			c[ipH2O][ipH2O] -= H2O_HCNP_HCN_H2OP_2;
			b[ipH2O] -= b_H2O_HCNP_HCN_H2OP;
			
			c[ipHCNP][ipHCNP] -= H2O_HCNP_HCN_H2OP_1;
			c[ipH2O][ipHCNP] -= H2O_HCNP_HCN_H2OP_2;
			b[ipHCNP] -= b_H2O_HCNP_HCN_H2OP;
			
			c[ipHCNP][ipHCN] += H2O_HCNP_HCN_H2OP_1;
			c[ipH2O][ipHCN] += H2O_HCNP_HCN_H2OP_2;
			b[ipHCN] += b_H2O_HCNP_HCN_H2OP;
			
			c[ipHCNP][ipH2OP] += H2O_HCNP_HCN_H2OP_1;
			c[ipH2O][ipH2OP] += H2O_HCNP_HCN_H2OP_2;
			b[ipH2OP] += b_H2O_HCNP_HCN_H2OP;

		/*  >>refer   mol  H2O    +  N2+   =>   N2    +     H2O+  UMIST  */

			c[ipN2P][ipH2O] -= H2O_N2P_N2_H2OP_1;
			c[ipH2O][ipH2O] -= H2O_N2P_N2_H2OP_2;
			b[ipH2O] -= b_H2O_N2P_N2_H2OP;
			
			c[ipN2P][ipN2P] -= H2O_N2P_N2_H2OP_1;
			c[ipH2O][ipN2P] -= H2O_N2P_N2_H2OP_2;
			b[ipN2P] -= b_H2O_N2P_N2_H2OP;
			
			c[ipN2P][ipN2] += H2O_N2P_N2_H2OP_1;
			c[ipH2O][ipN2] += H2O_N2P_N2_H2OP_2;
			b[ipN2] += b_H2O_N2P_N2_H2OP;
			
			c[ipN2P][ipH2OP] += H2O_N2P_N2_H2OP_1;
			c[ipH2O][ipH2OP] += H2O_N2P_N2_H2OP_2;
			b[ipH2OP] += b_H2O_N2P_N2_H2OP;

		/*  >>refer   mol  H2O+   +  NO    =>   NO+   +     H2O   UMIST  */

			c[ipNO][ipH2OP] -= H2OP_NO_NOP_H2O_1;
			c[ipH2OP][ipH2OP] -= H2OP_NO_NOP_H2O_2;
			b[ipH2OP] -= b_H2OP_NO_NOP_H2O;
			
			c[ipNO][ipNO] -= H2OP_NO_NOP_H2O_1;
			c[ipH2OP][ipNO] -= H2OP_NO_NOP_H2O_2;
			b[ipNO] -= b_H2OP_NO_NOP_H2O;
			
			c[ipNO][ipNOP] += H2OP_NO_NOP_H2O_1;
			c[ipH2OP][ipNOP] += H2OP_NO_NOP_H2O_2;
			b[ipNOP] += b_H2OP_NO_NOP_H2O;
			
			c[ipNO][ipH2O] += H2OP_NO_NOP_H2O_1;
			c[ipH2OP][ipH2O] += H2OP_NO_NOP_H2O_2;
			b[ipH2O] += b_H2OP_NO_NOP_H2O;

		/*  >>refer   mol  CN+    +  HCN   =>   HCN+  +     CN    UMIST  */

			c[ipHCN][ipCNP] -= CNP_HCN_HCNP_CN_1;
			c[ipCNP][ipCNP] -= CNP_HCN_HCNP_CN_2;
			b[ipCNP] -= b_CNP_HCN_HCNP_CN;
			
			c[ipHCN][ipHCN] -= CNP_HCN_HCNP_CN_1;
			c[ipCNP][ipHCN] -= CNP_HCN_HCNP_CN_2;
			b[ipHCN] -= b_CNP_HCN_HCNP_CN;
			
			c[ipHCN][ipHCNP] += CNP_HCN_HCNP_CN_1;
			c[ipCNP][ipHCNP] += CNP_HCN_HCNP_CN_2;
			b[ipHCNP] += b_CNP_HCN_HCNP_CN;
			
			c[ipHCN][ipCN] += CNP_HCN_HCNP_CN_1;
			c[ipCNP][ipCN] += CNP_HCN_HCNP_CN_2;
			b[ipCN] += b_CNP_HCN_HCNP_CN;

		/*  >>refer   mol  CN+    +  CO    =>   CO+   +     CN    UMIST  */

			c[ipCO][ipCNP] -= CNP_CO_COP_CN_1;
			c[ipCNP][ipCNP] -= CNP_CO_COP_CN_2;
			b[ipCNP] -= b_CNP_CO_COP_CN;
			
			c[ipCO][ipCO] -= CNP_CO_COP_CN_1;
			c[ipCNP][ipCO] -= CNP_CO_COP_CN_2;
			b[ipCO] -= b_CNP_CO_COP_CN;
			
			c[ipCO][ipCOP] += CNP_CO_COP_CN_1;
			c[ipCNP][ipCOP] += CNP_CO_COP_CN_2;
			b[ipCOP] += b_CNP_CO_COP_CN;
			
			c[ipCO][ipCN] += CNP_CO_COP_CN_1;
			c[ipCNP][ipCN] += CNP_CO_COP_CN_2;
			b[ipCN] += b_CNP_CO_COP_CN;

		/*  >>refer   mol  CN     +  N2+   =>   N2    +     CN+   UMIST  */

			c[ipN2P][ipCN] -= CN_N2P_N2_CNP_1;
			c[ipCN][ipCN] -= CN_N2P_N2_CNP_2;
			b[ipCN] -= b_CN_N2P_N2_CNP;
			
			c[ipN2P][ipN2P] -= CN_N2P_N2_CNP_1;
			c[ipCN][ipN2P] -= CN_N2P_N2_CNP_2;
			b[ipN2P] -= b_CN_N2P_N2_CNP;
			
			c[ipN2P][ipN2] += CN_N2P_N2_CNP_1;
			c[ipCN][ipN2] += CN_N2P_N2_CNP_2;
			b[ipN2] += b_CN_N2P_N2_CNP;
			
			c[ipN2P][ipCNP] += CN_N2P_N2_CNP_1;
			c[ipCN][ipCNP] += CN_N2P_N2_CNP_2;
			b[ipCNP] += b_CN_N2P_N2_CNP;

		/*  >>refer   mol  CN+    +  NO    =>   NO+   +     CN    UMIST  */

			c[ipNO][ipCNP] -= CNP_NO_NOP_CN_1;
			c[ipCNP][ipCNP] -= CNP_NO_NOP_CN_2;
			b[ipCNP] -= b_CNP_NO_NOP_CN;
			
			c[ipNO][ipNO] -= CNP_NO_NOP_CN_1;
			c[ipCNP][ipNO] -= CNP_NO_NOP_CN_2;
			b[ipNO] -= b_CNP_NO_NOP_CN;
			
			c[ipNO][ipNOP] += CNP_NO_NOP_CN_1;
			c[ipCNP][ipNOP] += CNP_NO_NOP_CN_2;
			b[ipNOP] += b_CNP_NO_NOP_CN;
			
			c[ipNO][ipCN] += CNP_NO_NOP_CN_1;
			c[ipCNP][ipCN] += CNP_NO_NOP_CN_2;
			b[ipCN] += b_CNP_NO_NOP_CN;

		/*  >>refer   mol  CN+    +  O2    =>   O2+   +     CN    UMIST  */

			c[ipO2][ipCNP] -= CNP_O2_O2P_CN_1;
			c[ipCNP][ipCNP] -= CNP_O2_O2P_CN_2;
			b[ipCNP] -= b_CNP_O2_O2P_CN;
			
			c[ipO2][ipO2] -= CNP_O2_O2P_CN_1;
			c[ipCNP][ipO2] -= CNP_O2_O2P_CN_2;
			b[ipO2] -= b_CNP_O2_O2P_CN;
			
			c[ipO2][ipO2P] += CNP_O2_O2P_CN_1;
			c[ipCNP][ipO2P] += CNP_O2_O2P_CN_2;
			b[ipO2P] += b_CNP_O2_O2P_CN;
			
			c[ipO2][ipCN] += CNP_O2_O2P_CN_1;
			c[ipCNP][ipCN] += CNP_O2_O2P_CN_2;
			b[ipCN] += b_CNP_O2_O2P_CN;

		/*  >>refer   mol  CN+    +  S     =>   S+    +     CN    UMIST  */

			c[ipATS][ipCNP] -= CNP_S_SP_CN_1;
			c[ipCNP][ipCNP] -= CNP_S_SP_CN_2;
			b[ipCNP] -= b_CNP_S_SP_CN;

			c[ipATS][ipATS] -= CNP_S_SP_CN_1;
			c[ipCNP][ipATS] -= CNP_S_SP_CN_2;
			b[ipATS] -= b_CNP_S_SP_CN;
			
			c[ipATS][ipSP] += CNP_S_SP_CN_1;
			c[ipCNP][ipSP] += CNP_S_SP_CN_2;
			b[ipSP] += b_CNP_S_SP_CN;
			
			c[ipATS][ipCN] += CNP_S_SP_CN_1;
			c[ipCNP][ipCN] += CNP_S_SP_CN_2;
			b[ipCN] += b_CNP_S_SP_CN;

		/*  >>refer   mol  HCN    +  CO+   =>   CO    +     HCN+  UMIST  */

			c[ipCOP][ipHCN] -= HCN_COP_CO_HCNP_1;
			c[ipHCN][ipHCN] -= HCN_COP_CO_HCNP_2;
			b[ipHCN] -= b_HCN_COP_CO_HCNP;
			
			c[ipCOP][ipCOP] -= HCN_COP_CO_HCNP_1;
			c[ipHCN][ipCOP] -= HCN_COP_CO_HCNP_2;
			b[ipCOP] -= b_HCN_COP_CO_HCNP;
			
			c[ipCOP][ipCO] += HCN_COP_CO_HCNP_1;
			c[ipHCN][ipCO] += HCN_COP_CO_HCNP_2;
			b[ipCO] += b_HCN_COP_CO_HCNP;
			
			c[ipCOP][ipHCNP] += HCN_COP_CO_HCNP_1;
			c[ipHCN][ipHCNP] += HCN_COP_CO_HCNP_2;
			b[ipHCNP] += b_HCN_COP_CO_HCNP;

		/*  >>refer   mol  HCN    +  N2+   =>   N2    +     HCN+  UMIST  */

			c[ipN2P][ipHCN] -= HCN_N2P_N2_HCNP_1;
			c[ipHCN][ipHCN] -= HCN_N2P_N2_HCNP_2;
			b[ipHCN] -= b_HCN_N2P_N2_HCNP;
			
			c[ipN2P][ipN2P] -= HCN_N2P_N2_HCNP_1;
			c[ipHCN][ipN2P] -= HCN_N2P_N2_HCNP_2;
			b[ipN2P] -= b_HCN_N2P_N2_HCNP;
			
			c[ipN2P][ipN2] += HCN_N2P_N2_HCNP_1;
			c[ipHCN][ipN2] += HCN_N2P_N2_HCNP_2;
			b[ipN2] += b_HCN_N2P_N2_HCNP;
			
			c[ipN2P][ipHCNP] += HCN_N2P_N2_HCNP_1;
			c[ipHCN][ipHCNP] += HCN_N2P_N2_HCNP_2;
			b[ipHCNP] += b_HCN_N2P_N2_HCNP;

		/*  >>refer   mol  HCN+   +  NO    =>   NO+   +     HCN   UMIST  */

			c[ipNO][ipHCNP] -= HCNP_NO_NOP_HCN_1;
			c[ipHCNP][ipHCNP] -= HCNP_NO_NOP_HCN_2;
			b[ipHCNP] -= b_HCNP_NO_NOP_HCN;
			
			c[ipNO][ipNO] -= HCNP_NO_NOP_HCN_1;
			c[ipHCNP][ipNO] -= HCNP_NO_NOP_HCN_2;
			b[ipNO] -= b_HCNP_NO_NOP_HCN;
			
			c[ipNO][ipNOP] += HCNP_NO_NOP_HCN_1;
			c[ipHCNP][ipNOP] += HCNP_NO_NOP_HCN_2;
			b[ipNOP] += b_HCNP_NO_NOP_HCN;
			
			c[ipNO][ipHCN] += HCNP_NO_NOP_HCN_1;
			c[ipHCNP][ipHCN] += HCNP_NO_NOP_HCN_2;
			b[ipHCN] += b_HCNP_NO_NOP_HCN;

		/*  >>refer   mol  HCN+   +  O2    =>   O2+   +     HCN   UMIST  */

			c[ipO2][ipHCNP] -= HCNP_O2_O2P_HCN_1;
			c[ipHCNP][ipHCNP] -= HCNP_O2_O2P_HCN_2;
			b[ipHCNP] -= b_HCNP_O2_O2P_HCN;
			
			c[ipO2][ipO2] -= HCNP_O2_O2P_HCN_1;
			c[ipHCNP][ipO2] -= HCNP_O2_O2P_HCN_2;
			b[ipO2] -= b_HCNP_O2_O2P_HCN;
			
			c[ipO2][ipO2P] += HCNP_O2_O2P_HCN_1;
			c[ipHCNP][ipO2P] += HCNP_O2_O2P_HCN_2;
			b[ipO2P] += b_HCNP_O2_O2P_HCN;
			
			c[ipO2][ipHCN] += HCNP_O2_O2P_HCN_1;
			c[ipHCNP][ipHCN] += HCNP_O2_O2P_HCN_2;
			b[ipHCN] += b_HCNP_O2_O2P_HCN;

		/*  >>refer   mol  HCN+   +  S     =>   S+    +     HCN   UMIST  */

			c[ipATS][ipHCNP] -= HCNP_S_SP_HCN_1;
			c[ipHCNP][ipHCNP] -= HCNP_S_SP_HCN_2;
			b[ipHCNP] -= b_HCNP_S_SP_HCN;

			c[ipATS][ipATS] -= HCNP_S_SP_HCN_1;
			c[ipHCNP][ipATS] -= HCNP_S_SP_HCN_2;
			b[ipATS] -= b_HCNP_S_SP_HCN;
			
			c[ipATS][ipSP] += HCNP_S_SP_HCN_1;
			c[ipHCNP][ipSP] += HCNP_S_SP_HCN_2;
			b[ipSP] += b_HCNP_S_SP_HCN;
			
			c[ipATS][ipHCN] += HCNP_S_SP_HCN_1;
			c[ipHCNP][ipHCN] += HCNP_S_SP_HCN_2;
			b[ipHCN] += b_HCNP_S_SP_HCN;

		/*  >>refer   mol  CO     +  N2+   =>   N2    +     CO+   UMIST  */

			c[ipN2P][ipCO] -= CO_N2P_N2_COP_1;
			c[ipCO][ipCO] -= CO_N2P_N2_COP_2;
			b[ipCO] -= b_CO_N2P_N2_COP;
			
			c[ipN2P][ipN2P] -= CO_N2P_N2_COP_1;
			c[ipCO][ipN2P] -= CO_N2P_N2_COP_2;
			b[ipN2P] -= b_CO_N2P_N2_COP;
			
			c[ipN2P][ipN2] += CO_N2P_N2_COP_1;
			c[ipCO][ipN2] += CO_N2P_N2_COP_2;
			b[ipN2] += b_CO_N2P_N2_COP;
			
			c[ipN2P][ipCOP] += CO_N2P_N2_COP_1;
			c[ipCO][ipCOP] += CO_N2P_N2_COP_2;
			b[ipCOP] += b_CO_N2P_N2_COP;

		/*  >>refer   mol  CO+    +  NO    =>   NO+   +     CO    UMIST  */

			c[ipNO][ipCOP] -= COP_NO_NOP_CO_1;
			c[ipCOP][ipCOP] -= COP_NO_NOP_CO_2;
			b[ipCOP] -= b_COP_NO_NOP_CO;
			
			c[ipNO][ipNO] -= COP_NO_NOP_CO_1;
			c[ipCOP][ipNO] -= COP_NO_NOP_CO_2;
			b[ipNO] -= b_COP_NO_NOP_CO;
			
			c[ipNO][ipNOP] += COP_NO_NOP_CO_1;
			c[ipCOP][ipNOP] += COP_NO_NOP_CO_2;
			b[ipNOP] += b_COP_NO_NOP_CO;
			
			c[ipNO][ipCO] += COP_NO_NOP_CO_1;
			c[ipCOP][ipCO] += COP_NO_NOP_CO_2;
			b[ipCO] += b_COP_NO_NOP_CO;

		/*  >>refer   mol  N2+    +  NO    =>   NO+   +     N2    UMIST  */

			c[ipNO][ipN2P] -= N2P_NO_NOP_N2_1;
			c[ipN2P][ipN2P] -= N2P_NO_NOP_N2_2;
			b[ipN2P] -= b_N2P_NO_NOP_N2;
			
			c[ipNO][ipNO] -= N2P_NO_NOP_N2_1;
			c[ipN2P][ipNO] -= N2P_NO_NOP_N2_2;
			b[ipNO] -= b_N2P_NO_NOP_N2;
			
			c[ipNO][ipNOP] += N2P_NO_NOP_N2_1;
			c[ipN2P][ipNOP] += N2P_NO_NOP_N2_2;
			b[ipNOP] += b_N2P_NO_NOP_N2;
			
			c[ipNO][ipN2] += N2P_NO_NOP_N2_1;
			c[ipN2P][ipN2] += N2P_NO_NOP_N2_2;
			b[ipN2] += b_N2P_NO_NOP_N2;

		/*  >>refer   mol  N2+    +  O2    =>   O2+   +     N2    UMIST  */

			c[ipO2][ipN2P] -= N2P_O2_O2P_N2_1;
			c[ipN2P][ipN2P] -= N2P_O2_O2P_N2_2;
			b[ipN2P] -= b_N2P_O2_O2P_N2;
			
			c[ipO2][ipO2] -= N2P_O2_O2P_N2_1;
			c[ipN2P][ipO2] -= N2P_O2_O2P_N2_2;
			b[ipO2] -= b_N2P_O2_O2P_N2;
			
			c[ipO2][ipO2P] += N2P_O2_O2P_N2_1;
			c[ipN2P][ipO2P] += N2P_O2_O2P_N2_2;
			b[ipO2P] += b_N2P_O2_O2P_N2;
			
			c[ipO2][ipN2] += N2P_O2_O2P_N2_1;
			c[ipN2P][ipN2] += N2P_O2_O2P_N2_2;
			b[ipN2] += b_N2P_O2_O2P_N2;

		/*  >>refer   mol  N2+    +  S     =>   S+    +     N2    UMIST  */

			c[ipATS][ipN2P] -= N2P_S_SP_N2_1;
			c[ipN2P][ipN2P] -= N2P_S_SP_N2_2;
			b[ipN2P] -= b_N2P_S_SP_N2;

			c[ipATS][ipATS] -= N2P_S_SP_N2_1;
			c[ipN2P][ipATS] -= N2P_S_SP_N2_2;
			b[ipATS] -= b_N2P_S_SP_N2;
			
			c[ipATS][ipSP] += N2P_S_SP_N2_1;
			c[ipN2P][ipSP] += N2P_S_SP_N2_2;
			b[ipSP] += b_N2P_S_SP_N2;
			
			c[ipATS][ipN2] += N2P_S_SP_N2_1;
			c[ipN2P][ipN2] += N2P_S_SP_N2_2;
			b[ipN2] += b_N2P_S_SP_N2;

		/*  >>refer   mol  SI     +  NO+   =>   NO    +     SI+   UMIST  */

			c[ipNOP][ipNOP] -= SI_NOP_NO_SIP_1;
			c[ipATSI][ipNOP] -= SI_NOP_NO_SIP_2;
			b[ipNOP] -= b_SI_NOP_NO_SIP;

			c[ipNOP][ipATSI] -= SI_NOP_NO_SIP_1;
			c[ipATSI][ipATSI] -= SI_NOP_NO_SIP_2;
			b[ipATSI] -= b_SI_NOP_NO_SIP;
			
			c[ipNOP][ipNO] += SI_NOP_NO_SIP_1;
			c[ipATSI][ipNO] += SI_NOP_NO_SIP_2;
			b[ipNO] += b_SI_NOP_NO_SIP;
			
			c[ipNOP][ipSIP] += SI_NOP_NO_SIP_1;
			c[ipATSI][ipSIP] += SI_NOP_NO_SIP_2;
			b[ipSIP] += b_SI_NOP_NO_SIP;

		/*  >>refer   mol  SI     +  HS+   =>   HS    +     SI+   UMIST  */

			c[ipHSP][ipHSP] -= SI_HSP_HS_SIP_1;
			c[ipATSI][ipHSP] -= SI_HSP_HS_SIP_2;
			b[ipHSP] -= b_SI_HSP_HS_SIP;

			c[ipHSP][ipATSI] -= SI_HSP_HS_SIP_1;
			c[ipATSI][ipATSI] -= SI_HSP_HS_SIP_2;
			b[ipATSI] -= b_SI_HSP_HS_SIP;
			
			c[ipHSP][ipHS] += SI_HSP_HS_SIP_1;
			c[ipATSI][ipHS] += SI_HSP_HS_SIP_2;
			b[ipHS] += b_SI_HSP_HS_SIP;
			
			c[ipHSP][ipSIP] += SI_HSP_HS_SIP_1;
			c[ipATSI][ipSIP] += SI_HSP_HS_SIP_2;
			b[ipSIP] += b_SI_HSP_HS_SIP;
			
		/*  >>refer   mol  SI     +  CS+   =>   CS    +     SI+   UMIST  */

			c[ipCSP][ipCSP] -= SI_CSP_CS_SIP_1;
			c[ipATSI][ipCSP] -= SI_CSP_CS_SIP_2;
			b[ipCSP] -= b_SI_CSP_CS_SIP;

			c[ipCSP][ipATSI] -= SI_CSP_CS_SIP_1;
			c[ipATSI][ipATSI] -= SI_CSP_CS_SIP_2;
			b[ipATSI] -= b_SI_CSP_CS_SIP;
			
			c[ipCSP][ipCS] += SI_CSP_CS_SIP_1;
			c[ipATSI][ipCS] += SI_CSP_CS_SIP_2;
			b[ipCS] += b_SI_CSP_CS_SIP;
			
			c[ipCSP][ipSIP] += SI_CSP_CS_SIP_1;
			c[ipATSI][ipSIP] += SI_CSP_CS_SIP_2;
			b[ipSIP] += b_SI_CSP_CS_SIP;

		/*  >>refer   mol  NO     +  HNO+  =>   HNO   +     NO+   UMIST  */

			c[ipHNOP][ipNO] -= NO_HNOP_HNO_NOP_1;
			c[ipNO][ipNO] -= NO_HNOP_HNO_NOP_2;
			b[ipNO] -= b_NO_HNOP_HNO_NOP;
			
			c[ipHNOP][ipHNOP] -= NO_HNOP_HNO_NOP_1;
			c[ipNO][ipHNOP] -= NO_HNOP_HNO_NOP_2;
			b[ipHNOP] -= b_NO_HNOP_HNO_NOP;
			
			c[ipHNOP][ipHNO] += NO_HNOP_HNO_NOP_1;
			c[ipNO][ipHNO] += NO_HNOP_HNO_NOP_2;
			b[ipHNO] += b_NO_HNOP_HNO_NOP;
			
			c[ipHNOP][ipNOP] += NO_HNOP_HNO_NOP_1;
			c[ipNO][ipNOP] += NO_HNOP_HNO_NOP_2;
			b[ipNOP] += b_NO_HNOP_HNO_NOP;

		/*  >>refer   mol  NO     +  O2+   =>   O2    +     NO+   UMIST  */

			c[ipO2P][ipNO] -= NO_O2P_O2_NOP_1;
			c[ipNO][ipNO] -= NO_O2P_O2_NOP_2;
			b[ipNO] -= b_NO_O2P_O2_NOP;
			
			c[ipO2P][ipO2P] -= NO_O2P_O2_NOP_1;
			c[ipNO][ipO2P] -= NO_O2P_O2_NOP_2;
			b[ipO2P] -= b_NO_O2P_O2_NOP;
			
			c[ipO2P][ipO2] += NO_O2P_O2_NOP_1;
			c[ipNO][ipO2] += NO_O2P_O2_NOP_2;
			b[ipO2] += b_NO_O2P_O2_NOP;
			
			c[ipO2P][ipNOP] += NO_O2P_O2_NOP_1;
			c[ipNO][ipNOP] += NO_O2P_O2_NOP_2;
			b[ipNOP] += b_NO_O2P_O2_NOP;

		/*  >>refer   mol  NO     +  S+    =>   S     +     NO+   UMIST  */

			c[ipSP][ipNO] -= NO_SP_S_NOP_1;
			c[ipNO][ipNO] -= NO_SP_S_NOP_2;
			b[ipNO] -= b_NO_SP_S_NOP;
			
			c[ipSP][ipSP] -= NO_SP_S_NOP_1;
			c[ipNO][ipSP] -= NO_SP_S_NOP_2;
			b[ipSP] -= b_NO_SP_S_NOP;
			
			c[ipSP][ipNOP] += NO_SP_S_NOP_1;
			c[ipNO][ipNOP] += NO_SP_S_NOP_2;
			b[ipNOP] += b_NO_SP_S_NOP;
			
			c[ipSP][ipATS] += NO_SP_S_NOP_1;
			c[ipNO][ipATS] += NO_SP_S_NOP_2;
			b[ipATS] += b_NO_SP_S_NOP;

		/*  >>refer   mol  NO     +  HS+   =>   HS    +     NO+   UMIST  */

			c[ipHSP][ipNO] -= NO_HSP_HS_NOP_1;
			c[ipNO][ipNO] -= NO_HSP_HS_NOP_2;
			b[ipNO] -= b_NO_HSP_HS_NOP;
			
			c[ipHSP][ipHSP] -= NO_HSP_HS_NOP_1;
			c[ipNO][ipHSP] -= NO_HSP_HS_NOP_2;
			b[ipHSP] -= b_NO_HSP_HS_NOP;
			
			c[ipHSP][ipHS] += NO_HSP_HS_NOP_1;
			c[ipNO][ipHS] += NO_HSP_HS_NOP_2;
			b[ipHS] += b_NO_HSP_HS_NOP;
			
			c[ipHSP][ipNOP] += NO_HSP_HS_NOP_1;
			c[ipNO][ipNOP] += NO_HSP_HS_NOP_2;
			b[ipNOP] += b_NO_HSP_HS_NOP;

		/*  >>refer   mol  NO     +  SIO+  =>   SIO   +     NO+   UMIST  */

			c[ipSIOP][ipNO] -= NO_SIOP_SIO_NOP_1;
			c[ipNO][ipNO] -= NO_SIOP_SIO_NOP_2;
			b[ipNO] -= b_NO_SIOP_SIO_NOP;
			
			c[ipSIOP][ipSIOP] -= NO_SIOP_SIO_NOP_1;
			c[ipNO][ipSIOP] -= NO_SIOP_SIO_NOP_2;
			b[ipSIOP] -= b_NO_SIOP_SIO_NOP;
			
			c[ipSIOP][ipSIO] += NO_SIOP_SIO_NOP_1;
			c[ipNO][ipSIO] += NO_SIOP_SIO_NOP_2;
			b[ipSIO] += b_NO_SIOP_SIO_NOP;
			
			c[ipSIOP][ipNOP] += NO_SIOP_SIO_NOP_1;
			c[ipNO][ipNOP] += NO_SIOP_SIO_NOP_2;
			b[ipNOP] += b_NO_SIOP_SIO_NOP;

		/*  >>refer   mol  NO     +  S2+   =>   S2    +     NO+   UMIST  */

			c[ipS2P][ipNO] -= NO_S2P_S2_NOP_1;
			c[ipNO][ipNO] -= NO_S2P_S2_NOP_2;
			b[ipNO] -= b_NO_S2P_S2_NOP;
			
			c[ipS2P][ipS2P] -= NO_S2P_S2_NOP_1;
			c[ipNO][ipS2P] -= NO_S2P_S2_NOP_2;
			b[ipS2P] -= b_NO_S2P_S2_NOP;
			
			c[ipS2P][ipS2] += NO_S2P_S2_NOP_1;
			c[ipNO][ipS2] += NO_S2P_S2_NOP_2;
			b[ipS2] += b_NO_S2P_S2_NOP;
			
			c[ipS2P][ipNOP] += NO_S2P_S2_NOP_1;
			c[ipNO][ipNOP] += NO_S2P_S2_NOP_2;
			b[ipNOP] += b_NO_S2P_S2_NOP;

		/*  >>refer   mol  O2+    +  NO2   =>   NO2+  +     O2    UMIST  */

			c[ipNO2][ipO2P] -= O2P_NO2_NO2P_O2_1;
			c[ipO2P][ipO2P] -= O2P_NO2_NO2P_O2_2;
			b[ipO2P] -= b_O2P_NO2_NO2P_O2;
			
			c[ipNO2][ipNO2] -= O2P_NO2_NO2P_O2_1;
			c[ipO2P][ipNO2] -= O2P_NO2_NO2P_O2_2;
			b[ipNO2] -= b_O2P_NO2_NO2P_O2;
			
			c[ipNO2][ipNO2P] += O2P_NO2_NO2P_O2_1;
			c[ipO2P][ipNO2P] += O2P_NO2_NO2P_O2_2;
			b[ipNO2P] += b_O2P_NO2_NO2P_O2;
			
			c[ipNO2][ipO2] += O2P_NO2_NO2P_O2_1;
			c[ipO2P][ipO2] += O2P_NO2_NO2P_O2_2;
			b[ipO2] += b_O2P_NO2_NO2P_O2;

		/*  >>refer   mol  S      +  HS+   =>   HS    +     S+    UMIST  */

			c[ipHSP][ipHSP] -= S_HSP_HS_SP_1;
			c[ipATS][ipHSP] -= S_HSP_HS_SP_2;
			b[ipHSP] -= b_S_HSP_HS_SP;

			c[ipHSP][ipATS] -= S_HSP_HS_SP_1;
			c[ipATS][ipATS] -= S_HSP_HS_SP_2;
			b[ipATS] -= b_S_HSP_HS_SP;
						
			c[ipHSP][ipHS] += S_HSP_HS_SP_1;
			c[ipATS][ipHS] += S_HSP_HS_SP_2;
			b[ipHS] += b_S_HSP_HS_SP;
			
			c[ipHSP][ipSP] += S_HSP_HS_SP_1;
			c[ipATS][ipSP] += S_HSP_HS_SP_2;
			b[ipSP] += b_S_HSP_HS_SP;
			
		/*  >>refer   mol  NH+    +  e     =>   N     +     H     UMIST  */

			c[ipNHP][ipNHP] -= NHP_e_N_H;
			c[ipNHP][ipATN] += NHP_e_N_H;

		/*  >>refer   mol  NH2+   +  e     =>   N     +     H  +  H      UMIST  */

			c[ipNH2P][ipNH2P] -= NH2P_e_N_H_H;
			c[ipNH2P][ipATN] += NH2P_e_N_H_H;

		/*  >>refer   mol  NH2+   +  e     =>   NH    +     H     UMIST  */

			c[ipNH2P][ipNH2P] -= NH2P_e_NH_H;
			c[ipNH2P][ipNH] += NH2P_e_NH_H;

		/*  >>refer   mol  NH3+   +  e     =>   NH    +     H  +  H      UMIST  */

			c[ipNH3P][ipNH3P] -= NH3P_e_NH_H_H;
			c[ipNH3P][ipNH] += NH3P_e_NH_H_H;

		/*  >>refer   mol  NH3+   +  e     =>   NH2   +     H     UMIST  */

			c[ipNH3P][ipNH3P] -= NH3P_e_NH2_H;
			c[ipNH3P][ipNH2] += NH3P_e_NH2_H;

		/*  >>refer   mol  NH4+   +  e     =>   NH2   +     H  +  H      UMIST  */

			c[ipNH4P][ipNH4P] -= NH4P_e_NH2_H_H;
			c[ipNH4P][ipNH2] += NH4P_e_NH2_H_H;

		/*  >>refer   mol  NH4+   +  e     =>   NH2   +     H2    UMIST  */

			c[ipNH4P][ipNH4P] -= NH4P_e_NH2_H2;
			c[ipNH4P][ipNH2] += NH4P_e_NH2_H2;

		/*  >>refer   mol  NH4+   +  e     =>   NH3   +     H     UMIST  */

			c[ipNH4P][ipNH4P] -= NH4P_e_NH3_H;
			c[ipNH4P][ipNH3] += NH4P_e_NH3_H;

		/*  >>refer   mol  CN+    +  e     =>   N     +     C     UMIST  */

			c[ipCNP][ipCNP] -= CNP_e_N_C;
			c[ipCNP][ipATN] += CNP_e_N_C;
			c[ipCNP][ipATO] += CNP_e_N_C;

		/*  >>refer   mol  HCN+   +  e     =>   CN    +     H     UMIST  */

			c[ipHCNP][ipHCNP] -= HCNP_e_CN_H;
			c[ipHCNP][ipCN] += HCNP_e_CN_H;

		/*  >>refer   mol  N2+    +  e     =>   N     +     N     UMIST  */

			c[ipN2P][ipN2P] -= N2P_e_N_N;
			c[ipN2P][ipATN] += 2*N2P_e_N_N;

		/*  >>refer   mol  NO+    +  e     =>   O     +     N     UMIST  */

			c[ipNOP][ipNOP] -= NOP_e_O_N;
			c[ipNOP][ipATO] += NOP_e_O_N;
			c[ipNOP][ipATN] += NOP_e_O_N;

		/*  >>refer   mol  HNO+   +  e     =>   NO    +     H     UMIST  */

			c[ipHNOP][ipHNOP] -= HNOP_e_NO_H;
			c[ipHNOP][ipNO] += HNOP_e_NO_H;

		/*  >>refer   mol  HS+    +  e     =>   S     +     H     UMIST  */

			c[ipHSP][ipHSP] -= HSP_e_S_H;
			c[ipHSP][ipATS] += HSP_e_S_H;

		/*  >>refer   mol  OCN+   +  e     =>   CO    +     N     UMIST  */

			c[ipOCNP][ipOCNP] -= OCNP_e_CO_N;
			c[ipOCNP][ipCO] += OCNP_e_CO_N;
			c[ipOCNP][ipATN] += OCNP_e_CO_N;

		/*  >>refer   mol  SIN+   +  e     =>   SI    +     N     UMIST  */

			c[ipSINP][ipSINP] -= SINP_e_SI_N;
			c[ipSINP][ipATSI] += SINP_e_SI_N;
			c[ipSINP][ipATN] += SINP_e_SI_N;

		/*  >>refer   mol  CS+    +  e     =>   S     +     C     UMIST  */

			c[ipCSP][ipCSP] -= CSP_e_S_C;
			c[ipCSP][ipATS] += CSP_e_S_C;
			c[ipCSP][ipATC] += CSP_e_S_C;

		/*  >>refer   mol  HCS+   +  e     =>   CS    +     H     UMIST  */

			c[ipHCSP][ipHCSP] -= HCSP_e_CS_H;
			c[ipHCSP][ipCS] += HCSP_e_CS_H;

		/*  >>refer   mol  NO2+   +  e     =>   NO    +     O     UMIST  */

			c[ipNO2P][ipNO2P] -= NO2P_e_NO_O;
			c[ipNO2P][ipNO] += NO2P_e_NO_O;
			c[ipNO2P][ipATO] += NO2P_e_NO_O;

		/*  >>refer   mol  NS+    +  e     =>   S     +     N     UMIST  */

			c[ipNSP][ipNSP] -= NSP_e_S_N;
			c[ipNSP][ipATO] += NSP_e_S_N;
			c[ipNSP][ipATS] += NSP_e_S_N;

		/*  >>refer   mol  SO+    +  e     =>   S     +     O     UMIST  */

			c[ipSOP][ipSOP] -= SOP_e_S_O;
			c[ipSOP][ipATO] += SOP_e_S_O;
			c[ipSOP][ipATS] += SOP_e_S_O;

		/*  >>refer   mol  OCS+   +  e     =>   S     +     CO    UMIST  */

			c[ipOCSP][ipOCSP] -= OCSP_e_S_CO;
			c[ipOCSP][ipCO] += OCSP_e_S_CO;
			c[ipOCSP][ipATS] += OCSP_e_S_CO;

		/*  >>refer   mol  OCS+   +  e     =>   CS    +     O     UMIST  */

			c[ipOCSP][ipOCSP] -= OCSP_e_CS_O;
			c[ipOCSP][ipCS] += OCSP_e_CS_O;
			c[ipOCSP][ipATO] += OCSP_e_CS_O;

		/*  >>refer   mol  S2+    +  e     =>   S     +     S     UMIST  */

			c[ipS2P][ipS2P] -= S2P_e_S_S;
			c[ipS2P][ipATS] += 2*S2P_e_S_S;

		/*  >>refer   mol  H-     +  N     =>   NH    +     e     UMIST  */

			c[ipATN][ipATN] -= Hminus_N_NH_e;
			c[ipATN][ipNH] += Hminus_N_NH_e;
			
		/*  >>refer   mol  H-     +  NH    =>   NH2   +     e     UMIST  */

			c[ipNH][ipNH] -= Hminus_NH_NH2_e;
			c[ipNH][ipNH2] += Hminus_NH_NH2_e;

		/*  >>refer   mol  H-     +  NH2   =>   NH3   +     e     UMIST  */

			c[ipNH2][ipNH2] -= Hminus_NH2_NH3_e;
			c[ipNH2][ipNH3] += Hminus_NH2_NH3_e;

		/*  >>refer   mol  H-     +  CN    =>   HCN   +     e     UMIST  */

			c[ipCN][ipCN] -= Hminus_CN_HCN_e;
			c[ipCN][ipHCN] += Hminus_CN_HCN_e;

		/*  >>refer   mol  C      +  N     =>   CN    +     NU    UMIST  */

			c[ipATN][ipATN]  -= C_N_CN_NU_1;
			c[ipATC][ipATN] -= C_N_CN_NU_2;
			b[ipATN] -= b_C_N_CN_NU;

			c[ipATN][ipATC]  -= C_N_CN_NU_1;
			c[ipATC][ipATC] -= C_N_CN_NU_2;
			b[ipATC] -= b_C_N_CN_NU;

			c[ipATN][ipCN]  += C_N_CN_NU_1;
			c[ipATC][ipCN] += C_N_CN_NU_2;
			b[ipCN] += b_C_N_CN_NU;

		/*  >>refer   mol  C      +  S     =>   CS    +     NU    UMIST  */

			c[ipATS][ipATC]  -= C_S_CS_NU_1;
			c[ipATC][ipATC]  -= C_S_CS_NU_2;
			b[ipATC]  -= b_C_S_CS_NU;

			c[ipATS][ipATS]  -= C_S_CS_NU_1;
			c[ipATC][ipATS] -= C_S_CS_NU_2;
			b[ipATS] -= b_C_S_CS_NU;

			c[ipATS][ipCS]  += C_S_CS_NU_1;
			c[ipATC][ipCS] += C_S_CS_NU_2;
			b[ipCS] += b_C_S_CS_NU;

		/*  >>refer   mol  C+     +  S     =>   CS+   +     NU    UMIST  */

			c[ipATS][ipCP]  -= CP_S_CSP_NU_1;
			c[ipCP][ipCP]  -= CP_S_CSP_NU_2;
			b[ipCP] -= b_CP_S_CSP_NU;

			c[ipATS][ipATS]  -= CP_S_CSP_NU_1;
			c[ipCP][ipATS]  -= CP_S_CSP_NU_2;
			b[ipATS] -= b_CP_S_CSP_NU;

			c[ipATN][ipCSP]  += CP_S_CSP_NU_1;
			c[ipCP][ipCSP]  += CP_S_CSP_NU_2;
			b[ipCSP] += b_CP_S_CSP_NU;

		/*  >>refer   mol  N+     +  N     =>   N2+   +     NU    UMIST  */

			c[ipATN][ipNP]  -= NP_N_N2P_NU_1;
			c[ipNP][ipNP]  -= NP_N_N2P_NU_2;
			b[ipNP] -= b_NP_N_N2P_NU;

			c[ipATN][ipATN]  -= NP_N_N2P_NU_1;
			c[ipNP][ipATN]  -= NP_N_N2P_NU_2;
			b[ipATN] -= b_NP_N_N2P_NU;

			c[ipATN][ipN2P]  += NP_N_N2P_NU_1;
			c[ipNP][ipN2P] += NP_N_N2P_NU_2;
			b[ipN2P] += b_NP_N_N2P_NU;

		/*  >>refer   mol  CO     +  S     =>   OCS   +     NU    UMIST  */

			c[ipATS][ipCO]  -= CO_S_OCS_NU_1;
			c[ipCO][ipCO] -= CO_S_OCS_NU_2;
			b[ipCO] -= b_CO_S_OCS_NU;

			c[ipATS][ipATS]  -= CO_S_OCS_NU_1;
			c[ipCO][ipATS] -= CO_S_OCS_NU_2;
			b[ipATS] -= b_CO_S_OCS_NU;

			c[ipATS][ipOCS]  += CO_S_OCS_NU_1;
			c[ipCO][ipOCS] += CO_S_OCS_NU_2;
			b[ipOCS] += b_CO_S_OCS_NU;

		/*  >>refer   mol  NH     +  NU    =>   NH+   +     e     UMIST  */

			c[ipNH][ipNH] -= NH_NU_NHP_e;
			c[ipNH][ipNHP] += NH_NU_NHP_e;

		/*  >>refer   mol  NH     +  NU    =>   N     +     H     UMIST  */

			c[ipNH][ipNH] -= NH_NU_N_H;
			c[ipNH][ipATN] += NH_NU_N_H;

		/*  >>refer   mol  NH2    +  NU    =>   NH    +     H     UMIST  */

			c[ipNH2][ipNH2] -= NH2_NU_NH_H;
			c[ipNH2][ipNH] += NH2_NU_NH_H;

		/*  >>refer   mol  NH2    +  NU    =>   NH2+  +     e     UMIST  */

			c[ipNH2][ipNH2] -= NH2_NU_NH2P_e;
			c[ipNH2][ipNH2P] += NH2_NU_NH2P_e;

		/*  >>refer   mol  NH3    +  NU    =>   NH3+  +     e     UMIST  */

			c[ipNH3][ipNH3] -= NH3_NU_NH3P_e;
			c[ipNH3][ipNH3P] += NH3_NU_NH3P_e;

		/*  >>refer   mol  NH3    +  NU    =>   NH    +     H2    UMIST  */

			c[ipNH3][ipNH3] -= NH3_NU_NH_H2;
			c[ipNH3][ipNH] += NH3_NU_NH_H2;

		/*  >>refer   mol  NH3    +  NU    =>   NH2   +     H     UMIST  */

			c[ipNH3][ipNH3] -= NH3_NU_NH2_H;
			c[ipNH3][ipNH2] += NH3_NU_NH2_H;

		/*  >>refer   mol  CN     +  NU    =>   N     +     C     UMIST  */

			c[ipCN][ipCN] -= CN_NU_N_C;
			c[ipCN][ipATN] += CN_NU_N_C;
			c[ipCN][ipATC] += CN_NU_N_C;

		/*  >>refer   mol  HCN    +  NU    =>   CN    +     H     UMIST  */

			c[ipHCN][ipHCN] -= HCN_NU_CN_H;
			c[ipHCN][ipCN] += HCN_NU_CN_H;

		/*  >>refer   mol  N2     +  NU    =>   N     +     N     UMIST  */

			c[ipN2][ipN2] -= N2_NU_N_N;
			c[ipN2][ipATN] += 2*N2_NU_N_N;

		/*  >>refer   mol  NO     +  NU    =>   O     +     N     UMIST  */

			c[ipNO][ipNO] -= NO_NU_O_N;
			c[ipNO][ipATN] += NO_NU_O_N;
			c[ipNO][ipATO] += NO_NU_O_N;

		/*  >>refer   mol  NO     +  NU    =>   NO+   +     e     UMIST  */

			c[ipNO][ipNO] -= NO_NU_NOP_e;
			c[ipNO][ipNOP] += NO_NU_NOP_e;

		/*  >>refer   mol  HNO    +  NU    =>   NO    +     H     UMIST  */

			c[ipHNO][ipHNO] -= HNO_NU_NO_H;
			c[ipHNO][ipNO] += HNO_NU_NO_H;

		/*  >>refer   mol  HS     +  NU    =>   S     +     H     UMIST  */

			c[ipHS][ipHS] -= HS_NU_S_H;
			c[ipHS][ipATS] += HS_NU_S_H;

		/*  >>refer   mol  HS+    +  NU    =>   S+    +     H     UMIST  */

			c[ipHSP][ipHSP] -= HSP_NU_SP_H;
			c[ipHSP][ipSP] += HSP_NU_SP_H;

		/*  >>refer   mol  OCN    +  NU    =>   CN    +     O     UMIST  */

			c[ipOCN][ipOCN] -= OCN_NU_CN_O;
			c[ipOCN][ipCN] += OCN_NU_CN_O;
			c[ipOCN][ipATO] += OCN_NU_CN_O;

		/*  >>refer   mol  CS     +  NU    =>   S     +     C     UMIST  */

			c[ipCS][ipCS] -= CS_NU_S_C;
			c[ipCS][ipATS] += CS_NU_S_C;
			c[ipCS][ipATO] += CS_NU_S_C;

		/*  >>refer   mol  CS+    +  NU    =>   S+    +     C     UMIST  */

			c[ipCSP][ipCSP] -= CSP_NU_SP_C;
			c[ipCSP][ipSP] += CSP_NU_SP_C;
			c[ipCSP][ipATC] += CSP_NU_SP_C;

		/*  >>refer   mol  CS     +  NU    =>   CS+   +     e     UMIST  */

			c[ipCS][ipCS] -= CS_NU_CSP_e;
			c[ipCS][ipCSP] += CS_NU_CSP_e;

		/*  >>refer   mol  NO2    +  NU    =>   NO    +     O     UMIST  */

			c[ipNO2][ipNO2] -= NO2_NU_NO_O;
			c[ipNO2][ipNO] += NO2_NU_NO_O;
			c[ipNO2][ipATO] += NO2_NU_NO_O;

		/*  >>refer   mol  NS     +  NU    =>   S     +     N     UMIST  */

			c[ipNS][ipNS] -= NS_NU_S_N;
			c[ipNS][ipATS] += NS_NU_S_N;
			c[ipNS][ipATN] += NS_NU_S_N;

		/*  >>refer   mol  SO     +  NU    =>   S     +     O     UMIST  */

			c[ipSO][ipSO] -= SO_NU_S_O;
			c[ipSO][ipATO] += SO_NU_S_O;
			c[ipSO][ipATS] += SO_NU_S_O;

		/*  >>refer   mol  SO     +  NU    =>   SO+   +     e     UMIST  */

			c[ipSO][ipSO] -= SO_NU_SOP_e;
			c[ipSO][ipSOP] += SO_NU_SOP_e;

		/*  >>refer   mol  OCS    +  NU    =>   S     +     CO    UMIST  */

			c[ipOCS][ipOCS] -= OCS_NU_S_CO;
			c[ipOCS][ipCO] += OCS_NU_S_CO;
			c[ipOCS][ipATS] += OCS_NU_S_CO;

		/*  >>refer   mol  OCS    +  NU    =>   OCS+  +     e     UMIST  */

			c[ipOCS][ipOCS] -= OCS_NU_OCSP_e;
			c[ipOCS][ipOCSP] += OCS_NU_OCSP_e;

		/*  >>refer   mol  NH     +  CRP   =>   NH+   +     e     UMIST  */

			c[ipNH][ipNH] -= NH_CRP_NHP_e;
			c[ipNH][ipNHP] += NH_CRP_NHP_e;

		/*  >>refer   mol  NH     +  CRP   =>   N     +     H     UMIST  */

			c[ipNH][ipNH] -= NH_CRP_N_H;
			c[ipNH][ipATN] += NH_CRP_N_H;

		/*  >>refer   mol  NH2    +  CRP   =>   NH    +     H     UMIST  */

			c[ipNH2][ipNH2] -= NH2_CRP_NH_H;
			c[ipNH2][ipNH] += NH2_CRP_NH_H;

		/*  >>refer   mol  NH2    +  CRP   =>   NH2+  +     e     UMIST  */

			c[ipNH2][ipNH2] -= NH2_CRP_NH2P_e;
			c[ipNH2][ipNH2P] += NH2_CRP_NH2P_e;

		/*  >>refer   mol  NH3    +  CRP   =>   NH3+  +     e     UMIST  */

			c[ipNH3][ipNH3] -= NH3_CRP_NH3P_e;
			c[ipNH3][ipNH3P] += NH3_CRP_NH3P_e;

		/*  >>refer   mol  NH3    +  CRP   =>   NH    +     H2    UMIST  */

			c[ipNH3][ipNH3] -= NH3_CRP_NH_H2;
			c[ipNH3][ipNH] += NH3_CRP_NH_H2;

		/*  >>refer   mol  NH3    +  CRP   =>   NH2   +     H     UMIST  */

			c[ipNH3][ipNH3] -= NH3_CRP_NH2_H;
			c[ipNH3][ipNH2] += NH3_CRP_NH2_H;

		/*  >>refer   mol  CN     +  CRP   =>   N     +     C     UMIST  */

			c[ipCN][ipCN] -= CN_CRP_N_C;
			c[ipCN][ipATN] += CN_CRP_N_C;
			c[ipCN][ipATC] += CN_CRP_N_C;

		/*  >>refer   mol  HCN    +  CRP   =>   CN    +     H     UMIST  */

			c[ipHCN][ipHCN] -= HCN_CRP_CN_H;
			c[ipHCN][ipCN] += HCN_CRP_CN_H;

		/*  >>refer   mol  N2     +  CRP   =>   N     +     N     UMIST  */

			c[ipN2][ipN2] -= N2_CRP_N_N;
			c[ipN2][ipATN] += 2*N2_CRP_N_N;

		/*  >>refer   mol  NO     +  CRP   =>   O     +     N     UMIST  */
			
			c[ipNO][ipNO] -= NO_CRP_O_N;
			c[ipNO][ipATO] += NO_CRP_O_N;
			c[ipNO][ipATN] += NO_CRP_O_N;

		/*  >>refer   mol  NO     +  CRP   =>   NO+   +     e     UMIST  */

			c[ipNO][ipNO] -= NO_CRP_NOP_e;
			c[ipNO][ipNOP] += NO_CRP_NOP_e;

		/*  >>refer   mol  HNO    +  CRP   =>   NO    +     H     UMIST  */

			c[ipHNO][ipHNO] -= HNO_CRP_NO_H;
			c[ipHNO][ipNO] += HNO_CRP_NO_H;

		/*  >>refer   mol  HS     +  CRP   =>   S     +     H     UMIST  */

			c[ipHS][ipHS] -= HS_CRP_S_H;
			c[ipHS][ipATS] += HS_CRP_S_H;

		/*  >>refer   mol  OCN    +  CRP   =>   CN    +     O     UMIST  */

			c[ipOCN][ipOCN] -= OCN_CRP_CN_O;
			c[ipOCN][ipCN] += OCN_CRP_CN_O;
			c[ipOCN][ipATO] += OCN_CRP_CN_O;

		/*  >>refer   mol  CS     +  CRP   =>   S     +     C     UMIST  */

			c[ipCS][ipCS] -= CS_CRP_S_C;
			c[ipCS][ipATS] += CS_CRP_S_C;
			c[ipCS][ipATC] += CS_CRP_S_C;

		/*  >>refer   mol  CS     +  CRP   =>   CS+   +     e     UMIST  */

			c[ipCS][ipCS] -= CS_CRP_CSP_e;
			c[ipCS][ipCSP] += CS_CRP_CSP_e;

		/*  >>refer   mol  NO2    +  CRP   =>   NO    +     O     UMIST  */

			c[ipNO2][ipNO2] -= NO2_CRP_NO_O;
			c[ipNO2][ipNO] += NO2_CRP_NO_O;
			c[ipNO2][ipATO] += NO2_CRP_NO_O;

		/*  >>refer   mol  NS     +  CRP   =>   S     +     N     UMIST  */

			c[ipNS][ipNS] -= NS_CRP_S_N;
			c[ipNS][ipATS] += NS_CRP_S_N;
			c[ipNS][ipATN] += NS_CRP_S_N;

		/*  >>refer   mol  SO     +  CRP   =>   S     +     O     UMIST  */

			c[ipSO][ipSO] -= SO_CRP_S_O;
			c[ipSO][ipATS] += SO_CRP_S_O;
			c[ipSO][ipATO] += SO_CRP_S_O;

		/*  >>refer   mol  SO     +  CRP   =>   SO+   +     e     UMIST  */

			c[ipSO][ipSO] -= SO_CRP_SOP_e;
			c[ipSO][ipSOP] += SO_CRP_SOP_e;

		/*  >>refer   mol  OCS    +  CRP   =>   S     +     CO    UMIST  */

			c[ipOCS][ipOCS] -= OCS_CRP_S_CO;
			c[ipOCS][ipCO] += OCS_CRP_S_CO;
			c[ipOCS][ipATS] += OCS_CRP_S_CO;

		/*  >>refer   mol  OCS    +  CRP   =>   OCS+  +     e     UMIST  */

			c[ipOCS][ipOCS] -= OCS_CRP_OCSP_e;
			c[ipOCS][ipOCSP] += OCS_CRP_OCSP_e;

		/*  >>refer   mol  H-     +  NH4+  =>   NH3   +     H2    UMIST  */

			c[ipNH4P][ipNH4P] -= Hminus_NH4P_NH3_H2;
			c[ipNH4P][ipNH3] += Hminus_NH4P_NH3_H2;

		/*  >>refer   mol  CH     +  N+    =>   N     +     CH+   UMIST  */

			c[ipNP][ipCH] -= CH_NP_N_CHP_1;
			c[ipCH][ipCH] -= CH_NP_N_CHP_2;
			b[ipCH] -= b_CH_NP_N_CHP;
			
			c[ipNP][ipNP] -= CH_NP_N_CHP_1;
			c[ipCH][ipNP] -= CH_NP_N_CHP_2;
			b[ipNP] -= b_CH_NP_N_CHP;
			
			c[ipNP][ipCHP] += CH_NP_N_CHP_1;
			c[ipCH][ipCHP] += CH_NP_N_CHP_2;
			b[ipCHP] += b_CH_NP_N_CHP;
			
			c[ipNP][ipATN] += CH_NP_N_CHP_1;
			c[ipCH][ipATN] += CH_NP_N_CHP_2;
			b[ipATN] += b_CH_NP_N_CHP;

		/*  >>refer   mol  CH+    +  S     =>   S+    +     CH    UMIST  */
			
			c[ipATS][ipCHP] -= CHP_S_SP_CH_1;
			c[ipCHP][ipCHP] -= CHP_S_SP_CH_2;
			b[ipCHP] -= b_CHP_S_SP_CH;
			
			c[ipATS][ipATS] -= CHP_S_SP_CH_1;
			c[ipCHP][ipATS] -= CHP_S_SP_CH_2;
			b[ipATS] -= b_CHP_S_SP_CH;
			
			c[ipATS][ipSP] += CHP_S_SP_CH_1;
			c[ipCHP][ipSP] += CHP_S_SP_CH_2;
			b[ipSP] += b_CHP_S_SP_CH;
			
			c[ipATS][ipCH] += CHP_S_SP_CH_1;
			c[ipCHP][ipCH] += CHP_S_SP_CH_2;
			b[ipCH] += b_CHP_S_SP_CH;

		/*  >>refer   mol  N+     +  CH2   =>   CH2+  +     N	    UMIST  */

			c[ipCH2][ipNP] -= NP_CH2_CH2P_N_1;
			c[ipNP][ipNP] -= NP_CH2_CH2P_N_2;
			b[ipNP] -= b_NP_CH2_CH2P_N;

			c[ipCH2][ipCH2] -= NP_CH2_CH2P_N_1;
			c[ipNP][ipCH2] -= NP_CH2_CH2P_N_2;
			b[ipCH2] -= b_NP_CH2_CH2P_N;
						
			c[ipCH2][ipCH2P] += NP_CH2_CH2P_N_1;
			c[ipNP][ipCH2P] += NP_CH2_CH2P_N_2;
			b[ipCH2P] += b_NP_CH2_CH2P_N;
						
			c[ipCH2][ipATN] += NP_CH2_CH2P_N_1;
			c[ipNP][ipATN] += NP_CH2_CH2P_N_2;
			b[ipATN] += b_NP_CH2_CH2P_N;
			
		/*  >>refer   mol  N+     +  CH4   =>   CH4+  +     N	    UMIST  */

			c[ipCH4][ipNP] -= NP_CH4_CH4P_N_1;
			c[ipNP][ipNP] -= NP_CH4_CH4P_N_2;
			b[ipNP] -= b_NP_CH4_CH4P_N;
			
			c[ipCH4][ipCH4] -= NP_CH4_CH4P_N_1;
			c[ipNP][ipCH4] -= NP_CH4_CH4P_N_2;
			b[ipCH4] -= b_NP_CH4_CH4P_N;
			
			c[ipCH4][ipCH4P] += NP_CH4_CH4P_N_1;
			c[ipNP][ipCH4P] += NP_CH4_CH4P_N_2;
			b[ipCH4P] += b_NP_CH4_CH4P_N;
			
			c[ipCH4][ipATN] += NP_CH4_CH4P_N_1;
			c[ipNP][ipATN] += NP_CH4_CH4P_N_2;
			b[ipATN] += b_NP_CH4_CH4P_N;
			
		/*  >>refer   mol  N+     +  OH    =>   OH+   +     N	    UMIST  */

			c[ipOH][ipNP] -= NP_OH_OHP_N_1;
			c[ipNP][ipNP] -= NP_OH_OHP_N_2;
			b[ipNP] -= b_NP_OH_OHP_N;
			
			c[ipOH][ipOH] -= NP_OH_OHP_N_1;
			c[ipNP][ipOH] -= NP_OH_OHP_N_2;
			b[ipOH] -= b_NP_OH_OHP_N;
			
			c[ipOH][ipOHP] += NP_OH_OHP_N_1;
			c[ipNP][ipOHP] += NP_OH_OHP_N_2;
			b[ipOHP] += b_NP_OH_OHP_N;
			
			c[ipOH][ipATN] += NP_OH_OHP_N_1;
			c[ipNP][ipATN] += NP_OH_OHP_N_2;
			b[ipATN] += b_NP_OH_OHP_N;
			
		/*  >>refer   mol  N+     +  H2O   =>   H2O+  +     N	    UMIST  */

			c[ipH2O][ipNP] -= NP_H2O_H2OP_N_1;
			c[ipNP][ipNP] -= NP_H2O_H2OP_N_2;
			b[ipNP] -= b_NP_H2O_H2OP_N;
			
			c[ipH2O][ipH2O] -= NP_H2O_H2OP_N_1;
			c[ipNP][ipH2O] -= NP_H2O_H2OP_N_2;
			b[ipH2O] -= b_NP_H2O_H2OP_N;
			
			c[ipH2O][ipH2OP] += NP_H2O_H2OP_N_1;
			c[ipNP][ipH2OP] += NP_H2O_H2OP_N_2;
			b[ipH2OP] += b_NP_H2O_H2OP_N;
			
			c[ipH2O][ipATN] += NP_H2O_H2OP_N_1;
			c[ipNP][ipATN] += NP_H2O_H2OP_N_2;
			b[ipATN] += b_NP_H2O_H2OP_N;
			
		/*  >>refer   mol  N+     +  CO    =>   CO+   +     N	    UMIST  */

			c[ipCO][ipNP] -= NP_CO_COP_N_1;
			c[ipNP][ipNP] -= NP_CO_COP_N_2;
			b[ipNP] -= b_NP_CO_COP_N;
			
			c[ipCO][ipCO] -= NP_CO_COP_N_1;
			c[ipNP][ipCO] -= NP_CO_COP_N_2;
			b[ipCO] -= b_NP_CO_COP_N;
			
			c[ipCO][ipCOP] += NP_CO_COP_N_1;
			c[ipNP][ipCOP] += NP_CO_COP_N_2;
			b[ipCOP] += b_NP_CO_COP_N;
			
			c[ipCO][ipATN] += NP_CO_COP_N_1;
			c[ipNP][ipATN] += NP_CO_COP_N_2;
			b[ipATN] += b_NP_CO_COP_N;

		/*  >>refer   mol  N+     +  O2    =>   O2+   +     N	    UMIST  */

			c[ipO2][ipNP] -= NP_O2_O2P_N_1;
			c[ipNP][ipNP] -= NP_O2_O2P_N_2;
			b[ipNP] -= b_NP_O2_O2P_N;
			
			c[ipO2][ipO2] -= NP_O2_O2P_N_1;
			c[ipNP][ipO2] -= NP_O2_O2P_N_2;
			b[ipO2] -= b_NP_O2_O2P_N;
			
			c[ipO2][ipO2P] += NP_O2_O2P_N_1;
			c[ipNP][ipO2P] += NP_O2_O2P_N_2;
			b[ipO2P] += b_NP_O2_O2P_N;
			
			c[ipO2][ipATN] += NP_O2_O2P_N_1;
			c[ipNP][ipATN] += NP_O2_O2P_N_2;
			b[ipATN] += b_NP_O2_O2P_N;

		/*  >>refer   mol  OH+    +  S     =>   S+    +     OH    UMIST  */

			c[ipATS][ipOHP] -= OHP_S_SP_OH_1;
			c[ipOHP][ipOHP] -= OHP_S_SP_OH_2;
			b[ipOHP] -= b_OHP_S_SP_OH;

			c[ipATS][ipATS] -= OHP_S_SP_OH_1;
			c[ipOHP][ipATS] -= OHP_S_SP_OH_2;
			b[ipATS] -= b_OHP_S_SP_OH;
			
			c[ipATS][ipSP] += OHP_S_SP_OH_1;
			c[ipOHP][ipSP] += OHP_S_SP_OH_2;
			b[ipSP] += b_OHP_S_SP_OH;
			
			c[ipATS][ipOH] += OHP_S_SP_OH_1;
			c[ipOHP][ipOH] += OHP_S_SP_OH_2;
			b[ipOH] += b_OHP_S_SP_OH;

		/*  >>refer   mol  H2O+   +  S     =>   S+    +     H2O   UMIST  */

			c[ipATS][ipH2OP] -= H2OP_S_SP_H2O_1;
			c[ipH2OP][ipH2OP] -= H2OP_S_SP_H2O_2;
			b[ipH2OP] -= b_H2OP_S_SP_H2O;

			c[ipATS][ipATS] -= H2OP_S_SP_H2O_1;
			c[ipH2OP][ipATS] -= H2OP_S_SP_H2O_2;
			b[ipATS] -= b_H2OP_S_SP_H2O;
			
			c[ipATS][ipSP] += H2OP_S_SP_H2O_1;
			c[ipH2OP][ipSP] += H2OP_S_SP_H2O_2;
			b[ipSP] += b_H2OP_S_SP_H2O;
			
			c[ipATS][ipH2O] += H2OP_S_SP_H2O_1;
			c[ipH2OP][ipH2O] += H2OP_S_SP_H2O_2;
			b[ipH2O] += b_H2OP_S_SP_H2O;

		/*  >>refer   mol  CO+    +  S     =>   S+    +     CO    UMIST  */

			c[ipATS][ipCOP] -= COP_S_SP_CO_1;
			c[ipCOP][ipCOP] -= COP_S_SP_CO_2;
			b[ipCOP] -= b_COP_S_SP_CO;

			c[ipATS][ipATS] -= COP_S_SP_CO_1;
			c[ipCOP][ipATS] -= COP_S_SP_CO_2;
			b[ipATS] -= b_COP_S_SP_CO;
			
			c[ipATS][ipSP] += COP_S_SP_CO_1;
			c[ipCOP][ipSP] += COP_S_SP_CO_2;
			b[ipSP] += b_COP_S_SP_CO;
			
			c[ipATS][ipCO] += COP_S_SP_CO_1;
			c[ipCOP][ipCO] += COP_S_SP_CO_2;
			b[ipCO] += b_COP_S_SP_CO;

		/*  >>refer   mol  SI     +  S+    =>   S     +     SI+   UMIST  */

			c[ipSP][ipATSI] -= SI_SP_S_SIP_1;
			c[ipATSI][ipATSI] -= SI_SP_S_SIP_2;
			b[ipATSI] -= b_SI_SP_S_SIP;
			
			c[ipSP][ipSP] -= SI_SP_S_SIP_1;
			c[ipATSI][ipSP] -= SI_SP_S_SIP_2;
			b[ipSP] -= b_SI_SP_S_SIP;
			
			c[ipSP][ipSIP] += SI_SP_S_SIP_1;
			c[ipATSI][ipSIP] += SI_SP_S_SIP_2;
			b[ipSIP] += b_SI_SP_S_SIP;

			c[ipSP][ipATS] += SI_SP_S_SIP_1;
			c[ipATSI][ipATS] += SI_SP_S_SIP_2;
			b[ipATS] += b_SI_SP_S_SIP;
			
		/*  >>refer   mol  O2+    +  S     =>   S+    +     O2    UMIST  */

			c[ipATS][ipO2P] -= O2P_S_SP_O2_1;
			c[ipO2P][ipO2P] -= O2P_S_SP_O2_2;
			b[ipO2P] -= b_O2P_S_SP_O2;

			c[ipATS][ipATS] -= O2P_S_SP_O2_1;
			c[ipO2P][ipATS] -= O2P_S_SP_O2_2;
			b[ipATS] -= b_O2P_S_SP_O2;
			
			c[ipATS][ipSP] += O2P_S_SP_O2_1;
			c[ipO2P][ipSP] += O2P_S_SP_O2_2;
			b[ipSP] += b_O2P_S_SP_O2;
			
			c[ipATS][ipO2] += O2P_S_SP_O2_1;
			c[ipO2P][ipO2] += O2P_S_SP_O2_2;
			b[ipO2] += b_O2P_S_SP_O2;

		/*  >>refer   mol  H-     +  N+    =>   N     +     H	    UMIST  */

			c[ipNP][ipNP] -= Hminus_NP_N_H;
			c[ipNP][ipATN] += Hminus_NP_N_H;

		/* >>refer	mol	N+ + e => N	 Done internally by Cloudy*/
		/* Recombination of N+ */
		c[ipNP][ipNP] -= (ionbal.RateRecomTot[ipNITROGEN][0]+gv.GrainChTrRate[ipNITROGEN][1][0]);
		c[ipNP][ipATN] += (ionbal.RateRecomTot[ipNITROGEN][0]+gv.GrainChTrRate[ipNITROGEN][1][0]);
	
		/* >>refer	mol	N + hnu => N+	e	 Done internally by Cloudy*/

		/* Photoionization of N */
		c[ipATN][ipNP] += (ionbal.RateIonizTot[ipNITROGEN][0]+gv.GrainChTrRate[ipNITROGEN][0][1]);
		c[ipATN][ipATN] -= (ionbal.RateIonizTot[ipNITROGEN][0]+gv.GrainChTrRate[ipNITROGEN][0][1]);


		/* >>refer	mol	S+ + e => S	 Done internally by Cloudy*/

		/* Recombination of S+ */
		c[ipSP][ipSP] -= (ionbal.RateRecomTot[ipSULPHUR][0]+gv.GrainChTrRate[ipSULPHUR][1][0]);
		c[ipSP][ipATS] += (ionbal.RateRecomTot[ipSULPHUR][0]+gv.GrainChTrRate[ipSULPHUR][1][0]);

		/* >>refer	mol	S + hnu => S+	e	 Done internally by Cloudy*/

		/* Photoionization of S */
		c[ipATS][ipSP] += (ionbal.RateIonizTot[ipSULPHUR][0]+gv.GrainChTrRate[ipSULPHUR][0][1]);
		c[ipATS][ipATS] -= (ionbal.RateIonizTot[ipSULPHUR][0]+gv.GrainChTrRate[ipSULPHUR][0][1]);

		/* >>chng Mar 23, 2005 -- Add Chlorine to network */

		/*  >>refer   mol  O    +  CCl     =>   Cl    +     CO    UMIST  */
		c[ipCCl][ipCCl] -= O_CCl_Cl_CO_2;
		c[ipATO][ipCCl] -= O_CCl_Cl_CO_1;
		b[ipCCl] -= b_O_CCl_Cl_CO;

		c[ipCCl][ipATO] -= O_CCl_Cl_CO_2;
		c[ipATO][ipATO] -= O_CCl_Cl_CO_1;
		b[ipATO] -= b_O_CCl_Cl_CO;

		c[ipCCl][ipATCl] += O_CCl_Cl_CO_2;
		c[ipATO][ipATCl] += O_CCl_Cl_CO_1;
		b[ipATCl] += b_O_CCl_Cl_CO;

		c[ipCCl][ipCO] += O_CCl_Cl_CO_2;
		c[ipATO][ipCO] += O_CCl_Cl_CO_1;
		b[ipCO] += b_O_CCl_Cl_CO;

		/*  >>refer   mol  O    +  CCl     =>   ClO    +     C    UMIST  */

		c[ipCCl][ipCCl] -= O_CCl_ClO_C_2;
		c[ipATO][ipCCl] -= O_CCl_ClO_C_1;
		b[ipCCl] -= b_O_CCl_ClO_C;

		c[ipCCl][ipATO] -= O_CCl_ClO_C_2;
		c[ipATO][ipATO] -= O_CCl_ClO_C_1;
		b[ipATO] -= b_O_CCl_ClO_C;

		c[ipCCl][ipClO] += O_CCl_ClO_C_2;
		c[ipATO][ipClO] += O_CCl_ClO_C_1;
		b[ipClO] += b_O_CCl_ClO_C;

		c[ipCCl][ipATC] += O_CCl_ClO_C_2;
		c[ipATO][ipATC] += O_CCl_ClO_C_1;
		b[ipATC] += b_O_CCl_ClO_C;

		/*  >>refer   mol  O    +  ClO     =>   Cl    +     O2    UMIST  */

		c[ipClO][ipClO] -= O_ClO_Cl_O2_2;
		c[ipATO][ipClO] -= O_ClO_Cl_O2_1;
		b[ipClO] -= b_O_ClO_Cl_O2;

		c[ipClO][ipATO] -= O_ClO_Cl_O2_2;
		c[ipATO][ipATO] -= O_ClO_Cl_O2_1;
		b[ipATO] -= b_O_ClO_Cl_O2;
		
		c[ipClO][ipATCl] += O_ClO_Cl_O2_2;
		c[ipATO][ipATCl] += O_ClO_Cl_O2_1;
		b[ipATCl] += b_O_ClO_Cl_O2;

		c[ipClO][ipO2] += O_ClO_Cl_O2_2;
		c[ipATO][ipO2] += O_ClO_Cl_O2_1;
		b[ipO2] += b_O_ClO_Cl_O2;

		/*  >>refer   mol  H2    +  Cl+     =>   HCl+    +     H    UMIST  */

		c[ipClP][ipClP] -= H2_ClP_HClP_H;
		c[ipClP][ipHClP] += H2_ClP_HClP_H;

		/*  >>refer   mol  H2    +  HCl+     =>   H2Cl+    +     H    UMIST  */

		c[ipHClP][ipHClP] -= H2_HClP_H2ClP_H;
		c[ipHClP][ipH2ClP] += H2_HClP_H2ClP_H;

		/*  >>refer   mol  H3+    +  Cl     =>   HCl+    +     H2    UMIST  */

		c[ipATCl][ipATCl] -= H3P_Cl_HClP_H2;
		c[ipATCl][ipHClP] += H3P_Cl_HClP_H2;

		/*  >>refer   mol  H3+    +  HCl     =>   H2Cl+    +     H2    UMIST  */

		c[ipHCl][ipHCl] -= H3P_HCl_H2ClP_H2;
		c[ipHCl][ipH2ClP] += H3P_HCl_H2ClP_H2;

		/*  >>refer   mol  He+    +  HCl     =>   Cl+    +     He	+	H    UMIST  */

		c[ipHCl][ipHCl] -= HeP_HCl_ClP_He_H;
		c[ipHCl][ipClP] += HeP_HCl_ClP_He_H;

		/*  >>refer   mol  He+    +  CCl     =>   Cl    +     C+	+     He   UMIST  */

		c[ipCCl][ipCCl] -= HeP_CCl_Cl_CP_He;
		c[ipCCl][ipATCl] += HeP_CCl_Cl_CP_He;
		c[ipCCl][ipCP] += HeP_CCl_Cl_CP_He;

		/*  >>refer   mol  He+    +  ClO     =>   Cl+    +     O	+     He   UMIST  */

		c[ipClO][ipClO] -= HeP_ClO_ClP_O_He;
		c[ipClO][ipClP] += HeP_ClO_ClP_O_He;
		c[ipClO][ipATO] += HeP_ClO_ClP_O_He;

		/*  >>refer   mol  C+    +  HCl     =>   CCl+    +     H    UMIST  */

		c[ipHCl][ipHCl] -= CP_HCl_CClP_H_2;
		c[ipCP][ipHCl] -= CP_HCl_CClP_H_1;
		b[ipHCl] -= b_CP_HCl_CClP_H;

		c[ipHCl][ipCP] -= CP_HCl_CClP_H_2;
		c[ipCP][ipCP] -= CP_HCl_CClP_H_1;
		b[ipCP] -= b_CP_HCl_CClP_H;

		c[ipHCl][ipCClP] += CP_HCl_CClP_H_2;
		c[ipCP][ipCClP] += CP_HCl_CClP_H_1;
		b[ipCClP] += b_CP_HCl_CClP_H;

		/*  >>refer   mol  CH3+    +  HCl     =>   H2CCl+    +     H2    UMIST  */

		c[ipHCl][ipHCl] -= CH3P_HCl_H2CClP_H2_2;
		c[ipCH3P][ipHCl] -= CH3P_HCl_H2CClP_H2_1;
		b[ipHCl] -= b_CH3P_HCl_H2CClP_H2;

		c[ipHCl][ipCH3P] += CH3P_HCl_H2CClP_H2_2;
		c[ipCH3P][ipCH3P] += CH3P_HCl_H2CClP_H2_1;
		b[ipCH3P] += b_CH3P_HCl_H2CClP_H2;

		c[ipHCl][ipH2CClP] += CH3P_HCl_H2CClP_H2_2;
		c[ipCH3P][ipH2CClP] += CH3P_HCl_H2CClP_H2_1;
		b[ipH2CClP] += b_CH3P_HCl_H2CClP_H2;

		/*  >>refer   mol  H2O    +  H2Cl+     =>   HCl    +     H3O+    UMIST  */

		c[ipH2ClP][ipH2ClP] -= H2O_H2ClP_HCl_H3OP_2;
		c[ipH2O][ipH2ClP] -= H2O_H2ClP_HCl_H3OP_1;
		b[ipH2ClP] -= b_H2O_H2ClP_HCl_H3OP;

		c[ipH2ClP][ipH2O] -= H2O_H2ClP_HCl_H3OP_2;
		c[ipH2O][ipH2O] -= H2O_H2ClP_HCl_H3OP_1;
		b[ipH2O] -= b_H2O_H2ClP_HCl_H3OP;

		c[ipH2ClP][ipHCl] += H2O_H2ClP_HCl_H3OP_2;
		c[ipH2O][ipHCl] += H2O_H2ClP_HCl_H3OP_1;
		b[ipHCl] += b_H2O_H2ClP_HCl_H3OP;

		c[ipH2ClP][ipH3OP] += H2O_H2ClP_HCl_H3OP_2;
		c[ipH2O][ipH3OP] += H2O_H2ClP_HCl_H3OP_1;
		b[ipH3OP] += b_H2O_H2ClP_HCl_H3OP;

		/*  >>refer   mol  CO    +  H2Cl+     =>   HCl    +     HCO+    UMIST  */

		c[ipH2ClP][ipH2ClP] -= CO_H2ClP_HCl_HCOP_2;
		c[ipCO][ipH2ClP] -= CO_H2ClP_HCl_HCOP_1;
		b[ipH2ClP] -= b_CO_H2ClP_HCl_HCOP;

		c[ipH2ClP][ipCO] -= CO_H2ClP_HCl_HCOP_2;
		c[ipCO][ipCO] -= CO_H2ClP_HCl_HCOP_1;
		b[ipCO] -= b_CO_H2ClP_HCl_HCOP;
	
		c[ipH2ClP][ipHCl] += CO_H2ClP_HCl_HCOP_2;
		c[ipCO][ipHCl] += CO_H2ClP_HCl_HCOP_1;
		b[ipHCl] += b_CO_H2ClP_HCl_HCOP;

		c[ipH2ClP][ipHCOP] += CO_H2ClP_HCl_HCOP_2;
		c[ipCO][ipHCOP] += CO_H2ClP_HCl_HCOP_1;
		b[ipHCOP] += b_CO_H2ClP_HCl_HCOP;

		/*  >>refer   mol  H+    +  HCl     =>   HCl+    +     H    UMIST  */

		c[ipHCl][ipHCl] -= HP_HCl_HClP_H;
		c[ipHCl][ipHClP] += HP_HCl_HClP_H;

		/*  >>refer   mol  C+    +  CCl     =>   CCl+    +     C    UMIST  */

		c[ipCCl][ipCCl] -= CP_CCl_CClP_C_2;
		c[ipCP][ipCCl] -= CP_CCl_CClP_C_1;
		b[ipCCl] -= b_CP_CCl_CClP_C,

		c[ipCCl][ipCP] -= CP_CCl_CClP_C_2;
		c[ipCP][ipCP] -= CP_CCl_CClP_C_1;
		b[ipCP] -= b_CP_CCl_CClP_C;

		c[ipCCl][ipCClP] += CP_CCl_CClP_C_2;
		c[ipCP][ipCClP] += CP_CCl_CClP_C_1;
		b[ipCClP] += b_CP_CCl_CClP_C;

		c[ipCCl][ipATC] += CP_CCl_CClP_C_2;
		c[ipCP][ipATC] += CP_CCl_CClP_C_1;
		b[ipATC] += b_CP_CCl_CClP_C;

		/*  >>refer   mol  C+    +  ClO     =>   ClO+    +     C    UMIST  */

		c[ipClO][ipClO] -= CP_ClO_ClOP_C_2;
		c[ipCP][ipClO] -= CP_ClO_ClOP_C_1;
		b[ipClO] -= b_CP_ClO_ClOP_C;

		c[ipClO][ipCP] -= CP_ClO_ClOP_C_2;
		c[ipCP][ipCP] -= CP_ClO_ClOP_C_1;
		b[ipCP] -= b_CP_ClO_ClOP_C;

		c[ipClO][ipClOP] += CP_ClO_ClOP_C_2;
		c[ipCP][ipClOP] += CP_ClO_ClOP_C_1;
		b[ipClOP] += b_CP_ClO_ClOP_C;

		c[ipClO][ipATC] += CP_ClO_ClOP_C_2;
		c[ipCP][ipATC] += CP_ClO_ClOP_C_1;
		b[ipATC] += b_CP_ClO_ClOP_C;

		/*  >>refer   mol  O2    +  Cl+     =>   Cl    +     O2+    UMIST  */

		c[ipClP][ipClP] -= O2_ClP_Cl_O2P_2;
		c[ipO2][ipClP] -= O2_ClP_Cl_O2P_1;
		b[ipClP] -= b_O2_ClP_Cl_O2P;

		c[ipClP][ipO2] -= O2_ClP_Cl_O2P_2;
		c[ipO2][ipO2] -= O2_ClP_Cl_O2P_1;
		b[ipO2] -= b_O2_ClP_Cl_O2P;

		c[ipClP][ipATCl] += O2_ClP_Cl_O2P_2;
		c[ipO2][ipATCl] += O2_ClP_Cl_O2P_1;
		b[ipATCl] += b_O2_ClP_Cl_O2P;

		c[ipClP][ipO2P] += O2_ClP_Cl_O2P_2;
		c[ipO2][ipO2P] += O2_ClP_Cl_O2P_1;
		b[ipO2P] += b_O2_ClP_Cl_O2P;

		/*  >>refer   mol  HCl+    +  e     =>   Cl    +     H    UMIST  */

		c[ipHClP][ipHClP] -= HClP_e_Cl_H;
		c[ipHClP][ipATCl] += HClP_e_Cl_H;

		/*  >>refer   mol  H2Cl+    +  e     =>   Cl    +     2H     UMIST  */

		c[ipH2ClP][ipH2ClP] -= H2ClP_e_Cl_H_H;
		c[ipH2ClP][ipATCl] += H2ClP_e_Cl_H_H;

		/*  >>refer   mol  H2Cl+    +  e     =>   HCl    +     H    UMIST  */

		c[ipH2ClP][ipH2ClP] -= H2ClP_e_HCl_H;
		c[ipH2ClP][ipHCl] += H2ClP_e_HCl_H;

		/*  >>refer   mol  CCl+    +  e     =>   Cl    +     C     UMIST  */

		c[ipCClP][ipCClP] -= CClP_e_Cl_C;
		c[ipCClP][ipATCl] += CClP_e_Cl_C;
		c[ipCClP][ipATC] += CClP_e_Cl_C;

		/*  >>refer   mol  H2CCl+    +  e     =>   CCl    +     2H     UMIST  */

		c[ipH2CClP][ipH2CClP] -= H2CClP_e_CCl_H_H;
		c[ipH2CClP][ipCCl] += H2CClP_e_CCl_H_H;

		/*  >>refer   mol  ClO+    +  e     =>   Cl    +     O    UMIST  */

		c[ipClOP][ipClOP] -= ClOP_e_Cl_O;
		c[ipClOP][ipATCl] += ClOP_e_Cl_O;
		c[ipClOP][ipATO] += ClOP_e_Cl_O;

		/*  >>refer   mol  HCl    +  NU     =>   Cl    +     H   UMIST  */

		c[ipHCl][ipHCl] -= HCl_NU_Cl_H;
		c[ipHCl][ipATCl] += HCl_NU_Cl_H;

		/*  >>refer   mol  CCl    +  NU     =>   Cl    +     C    UMIST  */

		c[ipCCl][ipCCl] -= CCl_NU_Cl_C;
		c[ipCCl][ipATCl] += CCl_NU_Cl_C;
		c[ipCCl][ipATC] += CCl_NU_Cl_C;

		/*  >>refer   mol  ClO    +  NU     =>   Cl    +     O    UMIST  */

		c[ipClO][ipClO] -= ClO_NU_Cl_O;
		c[ipClO][ipATCl] += ClO_NU_Cl_O;
		c[ipClO][ipATO] += ClO_NU_Cl_O;

		/*  >>refer   mol  HCl    +  crnu    =>   Cl    +     H    UMIST  */

		c[ipHCl][ipHCl] -= HCl_crnu_Cl_H;
		c[ipHCl][ipATCl] += HCl_crnu_Cl_H;

		/*  >>refer   mol  CCl    +  crnu    =>   Cl    +     C    UMIST  */

		c[ipCCl][ipCCl] -= CCl_crnu_Cl_C;
		c[ipCCl][ipATCl] += CCl_crnu_Cl_C;
		c[ipCCl][ipATC] += CCl_crnu_Cl_C;

		/*  >>refer   mol  ClO    +  crnu    =>   Cl    +     O    UMIST  */

		c[ipClO][ipClO] -= ClO_crnu_Cl_O;
		c[ipClO][ipATCl] += ClO_crnu_Cl_O;
		c[ipClO][ipATO] += ClO_crnu_Cl_O;

		/* >>refer	mol	Cl+ + e => Cl	 Done internally by Cloudy*/

		/* Recombination of S+ */
		c[ipClP][ipClP] -= (ionbal.RateRecomTot[ipCHLORINE][0]+gv.GrainChTrRate[ipCHLORINE][1][0]);
		c[ipClP][ipATCl] += (ionbal.RateRecomTot[ipCHLORINE][0]+gv.GrainChTrRate[ipCHLORINE][1][0]);

		/* >>refer	mol	Cl + hnu => Cl+	e	 Done internally by Cloudy*/

		/* Photoionization of S */
		c[ipATCl][ipClP] += (ionbal.RateIonizTot[ipCHLORINE][0]+gv.GrainChTrRate[ipCHLORINE][0][1]);
		c[ipATCl][ipATCl] -= (ionbal.RateIonizTot[ipCHLORINE][0]+gv.GrainChTrRate[ipCHLORINE][0][1]);

		/******************************** Gas-Grain Chemistry**********************************/

		/*  The Gas-Grain Chemistry rates are taken from: 
		>>refer	Hasegawa, T. I. & Herbst, E. 1993, MNRAS, 261, 83

		So far only CO depletion onto grains is considered, however, this code can be generalized 
		if desired to other molecules, using the data in the above paper.  There are three important reactions 
		to determine the abundance of a molecule on grain surfaces deep in molecular clouds.  The
		rate of accretion of molecule X onto grains is

		R(accretion) = PI*[grain_radius]^2*[thermal velocity]*[density_of_grains]

		Two processes remove molecule X from the grain surface, and that is thermal evaporation, due
		to the heat of the grain, and cosmic ray deabsorption.  The first of these rates come from the 
		above paper, and depends primarily on the dust temperature.  The cosmic ray rate is a constant,
		calculated in Hasegawa and Herbst.  

		For each molecule desired, I have created a new species which is the density of that molecule
		on the grain surface */

		
		/* evaporation rate of molecule on grain is:

		k(evap) = [vibrational absorption frequency]*exp[-binding_energy/dust_temperature]

		The binding energies come from Hasegawa and Herbst, Table 4.  The vibrational frequency comes from
		equation 3 of >>refer Hasegawa, T. I., Herbst, E., & Leung, C. M. 1992, ApJSS, 82, 167

		[vibrational absorption frequency] = 
		SQRT[[2*number_of_sites_for_grain_absorption*binding_energy]/[PI^2*mass_of_molecule]]

		**********************************************************************************************/

		/* number of sites on grain */
		number_of_sites = 1.5e15f;
		/* binding energy of CO */
		binding_energy = 1210.0f;
		/* exponential factor in evaporation rate */
		exponent = 0.0f;
		/* grain density times grain area */
		den_times_area = 0.0f;

		/* calculate the rates that are dependent on grain physics.  This includes grain density, 
		cross sectional area, and dust temperature of each constituent.  Note that 
		
		gv.bin[nd]->IntVol/gv.bin[nd]->AvVol*gv.bin[nd]->dstfactor*dense.gas_phase[ipHYDROGEN]
		
		is the grain density, according to abundances.c */

		for( nd=0; nd < gv.nBin; nd++ )

			{
				exponent += (double)exp(-binding_energy/gv.bin[nd]->tedust);
				den_times_area += (double)((gv.bin[nd]->IntVol/gv.bin[nd]->AvVol)*
					pow(gv.bin[nd]->AvRadius,2)*3.14159f*gv.bin[nd]->dstfactor*
					dense.gas_phase[ipHYDROGEN]);


			}
		/* If NO GRAIN MOLECULES command is on, or if grain physics is disabled, then there cannot be *
		 * molecular depletion.  Set the exponent to a large value, which assures that the abundance  *
		 * of molecules on grains is zero */ 
 
		if(!mole.lgGrain_mole_deplete ||!gv.lgGrainPhysicsOn )
		{
		}
	
		vib_freq = (float)(pow(2*number_of_sites*binding_energy/(3.14159f*3.14159f*28*ATOMIC_MASS_UNIT),0.5));
		kevap = vib_freq*exponent;


		/*	CO_grain	evaporation =>	CO	+	grain	*/

		c[ipCOGR][ipCOGR] -= kevap;
		c[ipCOGR][ipCO] += kevap;

		/*	CO_grain	+	cosmic_ray	=>	CO	+	grain	*/

		c[ipCOGR][ipCOGR] -= 9.8e-15;
		c[ipCOGR][ipCO] += 9.8e-15;

		/*	CO	+	grain	=>	CO_grain	*/

		c[ipCO][ipCO] -= den_times_area*pow((BOLTZMANN*phycon.te/(28*ATOMIC_MASS_UNIT)),0.5);
		c[ipCO][ipCOGR] += den_times_area*pow((BOLTZMANN*phycon.te/(28*ATOMIC_MASS_UNIT)),0.5);

		if(!mole.lgGrain_mole_deplete || !gv.lgDustOn )
		{
			c[ipCOGR][ipCOGR] = -1;
			c[ipCOGR][ipCO] = 1;
		}
	/********************************************************************************************/
	
	/*							DONE WITH BALANCE EQUATIONS									    */
	
	/********************************************************************************************/
	
		
		/* >>chng Oct. 21, 2004 -- Terms that contribute to the ionization balance of C, O, S, Si, and N
	   will now be inserted directly into the ion solver.  Before, we just corrected the recombination rate
	   by scaling saying that IONIZATION_RATE = RECOMBINATION_RATE * [n(X+)/n(X0)].  This code follows the logic
	   of hmole_step, written by Robin Williams. */

		
		/* sink terms should include only terms that either form or destroy an atomic or singly ionized
		   element in the network, but not both.  This means that terms that destroy X at the expense of 
		   forming X+ can't be included.  The rate of destruction of X or X+ is equal to the formation of 
		   the inverse.  We therefore add this rate to sink, which gets rid of this dependence */

	/* The following code is all generalized.  After all the molecules, the total number being NUM_HEAVY_MOLEC,
	   there are 2*NUM_ELEMENTS remaining.  The first set, equal to NUM_ELEMENTS, are the ionized elemental
	   species.  The second set, also equal to NUM_ELEMENTS, are the atomic species.  They are in order such that
		   
	   array number for X = array number for (X+) + NUM_ELEMENTS

	   In the future, if one wants to add another element to the network, such as Na the macros must be changed.
	   Also, the array number for Na and Na+ must obey the equation above.  If this is done, then the sources,
	   sinks, and molecular recombination terms are automatically added!!! */

	/* co.nelem_hevmol[i][j] is just the dominant element for a given molecule.  For an element this
	   string is just the element itself!! */

	for(i=NUM_HEAVY_MOLEC;i<NUM_HEAVY_MOLEC+NUM_ELEMENTS; ++i)
	{ 
		mole.sink[co.nelem_hevmol[i]][0] -= (c[i+NUM_ELEMENTS][i+NUM_ELEMENTS] + 
								c[i+NUM_ELEMENTS][i])*dense.lgElmtOn[co.nelem_hevmol[i]];

		mole.sink[co.nelem_hevmol[i]][1] -= (c[i][i] + c[i][i+NUM_ELEMENTS] )*dense.lgElmtOn[co.nelem_hevmol[i]];
	}

	/* source terms must be multiplied by the density of the corresponding matrix element */

	for(j=NUM_HEAVY_MOLEC;j < NUM_HEAVY_MOLEC+NUM_ELEMENTS; ++j)
	{
		for(i=0; i < NUM_COMOLE_CALC; ++i)
		{
			mole.source[co.nelem_hevmol[j]][1] += co.hevmol[i]*c[i][j]*dense.lgElmtOn[co.nelem_hevmol[j]];
			mole.source[co.nelem_hevmol[j]][0] += co.hevmol[i]*c[i][j+NUM_ELEMENTS]*dense.lgElmtOn[co.nelem_hevmol[j+NUM_ELEMENTS]];
		}
	}

	/* subtract out diagonal terms, they are already in sink.  Also subtract recombination terms,
	   as these are done by mole.xMoleChTrRate */

	for(i=NUM_HEAVY_MOLEC;i < NUM_HEAVY_MOLEC+NUM_ELEMENTS; ++i)
	{ 
		mole.source[co.nelem_hevmol[i]][1]   -= ( c[i][i]*co.hevmol[i] + 
			c[i+NUM_ELEMENTS][i]*co.hevmol[i+NUM_ELEMENTS])*dense.lgElmtOn[co.nelem_hevmol[i]] ;

		mole.source[co.nelem_hevmol[i]][0]   -= ( c[i+NUM_ELEMENTS][i+NUM_ELEMENTS]*co.hevmol[i+NUM_ELEMENTS] + 
			c[i][i+NUM_ELEMENTS]*co.hevmol[i])*dense.lgElmtOn[co.nelem_hevmol[i+NUM_ELEMENTS]] ;

		/* The source terms, as they are right now, include negative contributions.  This is because the 
		linearization "trick" creates source terms.   Take for example the reaction:
		C+        H2O      =>        HCO+   H          
		This reaction destroys C+, but in the act of linearizing, there will be the following term:
		c[ipH2O][ipCP]
		This is only a numerical "trick" and not a real matrix term.  All terms like this
		have to be removed to get the "true" source terms.  Fortunately, the sum of all these terms
		is just b[i].  To remove these terms, we have to add b[i] if the overall contribution from these terms
		was negative, subtract if their contribution was positive.  This is done by subtracting b[i] 
		from the current value of source. */

		mole.source[co.nelem_hevmol[i]][1] = mole.source[co.nelem_hevmol[i]][1] - b[i];
		mole.source[co.nelem_hevmol[i]][0] = mole.source[co.nelem_hevmol[i]][0] - b[i+NUM_ELEMENTS];

	}

	/* negative source terms are actually destruction mechanisms, so should go into the sink vector.
	   source and sinks have different units, so if source is negative divide by the density of
	   the corresponding element to get same units.  For example, if:

	   mole.source[ipCARBON][0]

	   is negative, we divide by dense.xIonDense[ipCARBON][0] to get rate in same units as mole.sink */

	for(i=2; i < LIMELM; ++i)
	{
		for(j=0; j < 2; ++j)
		{
			/*if source term is negative, make it a sink and then set to zero*/ 
			if(mole.source[i][j] < 0)
			{
				mole.sink[i][j] -= (mole.source[i][j]/SDIV(dense.xIonDense[i][j]));
				mole.source[i][j] = 0;
			}
		}
	}
	/* These are rates of recombination for atomic and singly ionized species that are due to molecular processes
	   This will be added to ion_solver to get correct ionization balance */

	for(i=NUM_HEAVY_MOLEC;i < NUM_HEAVY_MOLEC+NUM_ELEMENTS; ++i)
	{ 
		mole.xMoleChTrRate[co.nelem_hevmol[i]][1][0] = (float)(c[i][i+NUM_ELEMENTS] - 
				ionbal.RateRecomTot[co.nelem_hevmol[i]][0])*dense.lgElmtOn[co.nelem_hevmol[i]];

		mole.xMoleChTrRate[co.nelem_hevmol[i]][0][1] =  (float)(c[i+NUM_ELEMENTS][i] - 
				ionbal.RateIonizTot[co.nelem_hevmol[i+NUM_ELEMENTS]][0])*dense.lgElmtOn[co.nelem_hevmol[i+NUM_ELEMENTS]];
	}
	
	/* If rate for going from 1-0 or 0-1 is negative then it should be added to the inverse process */
	for(i=2; i < LIMELM; ++i)
	{
		for(j=0; j < 2; ++j)
		{
			for(k=0; k< 2; ++k)
			{
			/*only possible charge transfers are 0-1 or 1-0 */		
				if(j != k)
				{
					if( mole.xMoleChTrRate[i][j][k] < 0. )
					{
						mole.xMoleChTrRate[i][k][j] -= mole.xMoleChTrRate[i][j][k];
						mole.xMoleChTrRate[i][j][k] = 0.;
					}
				}
			}
		}
	}

	cartot_ion = dense.gas_phase[ipCARBON];
	oxytot_ion = dense.gas_phase[ipOXYGEN];
	siltot_ion = dense.gas_phase[ipSILICON];
	nitrotot_ion = dense.gas_phase[ipNITROGEN];
	sulptot_ion = dense.gas_phase[ipSULPHUR];
	chlotot_ion = dense.gas_phase[ipCHLORINE];


	for ( i=2; i < ipCARBON+2; i++ )
	{
		cartot_ion -= dense.xIonDense[ipCARBON][i];
	}
	for ( i=2; i < ipOXYGEN+2; i++ )
	{
		oxytot_ion -= dense.xIonDense[ipOXYGEN][i];
	}

	for ( i=2; i < ipSILICON+2; i++ )
	{
		siltot_ion -= dense.xIonDense[ipSILICON][i];
	}
	
	for ( i=2; i < ipNITROGEN+2; i++ )
	{
		nitrotot_ion -= dense.xIonDense[ipNITROGEN][i];
	}

	for ( i=2; i < ipSULPHUR+2; i++ )
	{
		sulptot_ion -= dense.xIonDense[ipSULPHUR][i];
	}

	for ( i=2; i < ipCHLORINE+2; i++ )
	{
		chlotot_ion -= dense.xIonDense[ipCHLORINE][i];
	}
	

	/* at this point the cartot_mol, sum of molecules and atom/first ion,
	 * should be equal to the gas_phase minus double and higher ions */
	/*fprintf(ioQQQ," dbuggggas\t %.2f\t%f\t%f\n",fnzone,
		cartot_mol/cartot_ion,
		oxytot_mol/oxytot_ion);*/

	/* these will be used in population equation in case of homogeneous equation */

	b[ipATC]  = 0;
	b[ipATO]  = 0;
	b[ipATSI] = 0;
	b[ipATN]  = 0;
	b[ipATS]  = 0;
	b[ipATCl]  = 0;

	b[ipATC]  = cartot_ion;
	b[ipATO]  = oxytot_ion;
	b[ipATSI] = siltot_ion;
	b[ipATN]  = nitrotot_ion;
	b[ipATS]  = sulptot_ion;
	b[ipATCl] = chlotot_ion;


	/* <<chng 03 Nov 11--Nick Abel,  Set up the non-zero matrix elements that go into the conservation equations
	   for atomic C, O, and Si, this is now set up by looping over the atomic species in
	   co.h and setting the number of atoms of C, O, and Si equal to the variable co.nCarb, 
	   co.nOxyg, and co.nSili respectively.  For every element (excluding Hydrogen) not in
	   the network an if statement will be needed */

	/* loop over last NUM_ELEMENTS elements in co vector - these are atoms of the NUM_ELEMENTS */
	for ( i=NUM_COMOLE_CALC - NUM_ELEMENTS; i < NUM_COMOLE_CALC; i++ )
	{
		for ( j=0; j < NUM_COMOLE_CALC; j++ )

		{
			if ( i == (NUM_COMOLE_CALC - 6) )	
				
			{	
				matrix_value = co.nCarb[j];
				c[j][i] = 0;
				c[j][i] = matrix_value;
			}

			else if ( i == (NUM_COMOLE_CALC - 5 ))	
				
			{
				matrix_value = co.nOxyg[j];
				c[j][i] = 0;
				c[j][i] = matrix_value;
			}
			
			else if ( i == (NUM_COMOLE_CALC - 4) )
				
			{	
				matrix_value = co.nSili[j];
				c[j][i] = 0;
				c[j][i] = matrix_value;
			}

			else if ( i == (NUM_COMOLE_CALC - 3) )
				
			{	
				matrix_value = co.nNitr[j];
				c[j][i] = 0;
				c[j][i] = matrix_value;
			}
		
			else if ( i == (NUM_COMOLE_CALC - 2) )
				
			{	
				matrix_value = co.nSulp[j];
				c[j][i] = 0;
				c[j][i] = matrix_value;
			}

			else if ( i == (NUM_COMOLE_CALC - 1) )
				
			{	
				matrix_value = co.nChlo[j];
				c[j][i] = 0;
				c[j][i] = matrix_value;
			}


		}

	}

	/*--------------------------------------------------------------------
		* */
	/*printf( " Here are all matrix elements\n ");

	for (i=0; i < NUM_COMOLE_CALC; i++)
		
	{
		

			for (j=0; j < NUM_COMOLE_CALC; j++)
			{
				printf( "%4.4s", co.chLab[i] );
				printf( "%4.4s\t", co.chLab[j] );
				printf( "%.8e,\n", c[j][i] );
			}
	}
	printf( "b's are:\n" );
	for (i=0; i < NUM_COMOLE_CALC; i++)
	{
		printf( "%.8e,\n", b[i] );
	}*/


	if ( trace.lgTrace && trace.lgTr_CO_Mole )
	{
		fprintf( ioQQQ, " COMOLE matrix\n           " );

#		if 0
		for ( i=0; i < MIN2(NUM_COMOLE_CALC,8); i++ )
		{
			fprintf( ioQQQ, "%4.4s", co.chLab[i] );
		}
		fprintf( ioQQQ, "     \n" );

		for ( j=0; j < NUM_COMOLE_CALC; j++ )
		{
			fprintf( ioQQQ, " %4.4s", co.chLab[j] );
			fprintf( ioQQQ, " " );
			for ( i=0; i < MIN2(NUM_COMOLE_CALC,8); i++ )
			{
				fprintf( ioQQQ, "%9.1e", c[i][j] );
			}
			fprintf( ioQQQ, "\n" );
		}
#		endif

		fprintf( ioQQQ, " COMOLE matrix\n           " );

		for ( i=0; i < NUM_COMOLE_CALC; i++ )
		{
			fprintf( ioQQQ, "%4.4s", co.chLab[i] );
		}
		fprintf( ioQQQ, "     \n" );

		for ( j=0; j < NUM_COMOLE_CALC; j++ )
		{
			fprintf( ioQQQ, " %4.4s", co.chLab[j] );
			fprintf( ioQQQ, " " );
			for ( i=0; i < NUM_COMOLE_CALC; i++ )
			{
				/*fprintf( ioQQQ, "%9.1e", c[i][j] );*/
				fprintf( ioQQQ, ", %.15e", c[i][j] );
			}
			fprintf( ioQQQ, "\n" );
		}

		fprintf( ioQQQ, " COMOLE b\n           " );
		for ( j=0; j < NUM_COMOLE_CALC; j++ )
		{
			fprintf( ioQQQ, " %4.4s", co.chLab[j] );
			fprintf( ioQQQ, ", %.15e\n", b[j] );
		}

		fprintf( ioQQQ, 
			" COMOLE H2 den:%10.3e, H2,3+=%10.2e%10.2e Carb sum=%10.3e Oxy sum=%10.3e\n", 
			hmi.H2_total, 
			hmi.Hmolec[ipMH2p], 
			hmi.Hmolec[ipMH3p], 
			c[NUM_HEAVY_MOLEC][ipATC], 
			c[NUM_HEAVY_MOLEC][ipATO] );

	}

	/* save matrix before we invert it */
	for ( i=0; i < NUM_COMOLE_CALC; i++ )
	{
		for ( j=0; j < NUM_COMOLE_CALC; j++ )
		{
			csav[j][i] = c[j][i];
		}
	}
	/* remember longest CO timescale */

	if ( -c[ipCO][ipCO] > SMALLFLOAT )
	{
		/* this is rate CO is destroyed, equal to formation rate in equilibrium */
		timesc.AgeCOMoleDest = -1./c[ipCO][ipCO];
		/* moved to radinc */
		/*timesc.BigCOMoleForm = (float)MAX2(timesc.BigCOMoleForm,timesc.AgeCOMoleForm);*/
	}
	else
	{
		timesc.AgeCOMoleDest = 0.;
	}

	/* this one may be more robust */
	for ( j=0; j < NUM_COMOLE_CALC; j++ )
	{
		for ( i=0; i < NUM_COMOLE_CALC; i++ )
		{
			amat[i][j] = c[i][j];
		}
	}

	merror = 0;

	/*lint -e740 unusual pointer case */
	getrf_wrapper(NUM_COMOLE_CALC,NUM_COMOLE_CALC,(double*)amat,NUM_COMOLE_CALC, ipiv,&merror);
	/*DGETRF(NUM_COMOLE_CALC,NUM_COMOLE_CALC,(double*)amat,NUM_COMOLE_CALC, ipiv,&merror);*/

	if ( merror != 0 )
	{
		fprintf( ioQQQ, " CO_step getrf_wrapper error\n" );
		puts( "[Stop in CO_step]" );
		cdEXIT(EXIT_FAILURE);
	}

	getrs_wrapper('N',NUM_COMOLE_CALC,1,(double*)amat,NUM_COMOLE_CALC,ipiv,b,NUM_COMOLE_CALC,&merror);
	/*lint +e740 unusual pointer case */

	if ( merror != 0 )
	{
		fprintf( ioQQQ, " CO_step: dgetrs finds singular or ill-conditioned matrix\n" );
		puts( "[Stop in CO_step]" );
		cdEXIT(EXIT_FAILURE);
	}

	/* check for negative populations, which happens when 100% co */
	*lgNegPop = FALSE;
	*lgZerPop = FALSE;
	co_denominator = 0;

	for ( i=0; i < NUM_COMOLE_CALC; i++ )
	{
		if ( b[i] < 0. )
		{
			/* >>chng 03 sep 11, the following*/
			/* comparison between the solution vector from 32 bit machines vs
			 * maple on a pc shows that very small negative numbers are produced by
			 * the linear algebra package used by the code, but is positive with
			 * the math package.  allow very small negative numbers to occur,
			 * and silently reset to a postive number. 
			 *
			 * >>chng 04 feb 25
			 * Here we want to check if the negative abundance is a significant
			 * fraction of any of the species in the network.  The code below
			 * checks to see which elements the molecule in question is made of,
			 * then finds which one has the least abundance.  The one with
			 * the least abundance will then be used as the divisor in checking 
			 * if the negative abundance is too large to be ignored.*/
			
			/*  >>chng 04 apr 21, when an element in the chemical network is 
			 *  turned off, the molecular abundance of a species containing that
			 *  element was not zero but rather a small number of order 1e-20. When 
			 *  these abundances went negative, the code thought the abundances were
			 *  important because it checks the ratio of the species to its gas phase
			 *  abundance.  When the gas phase abundance is zero, the denominator is set 
			 *  to SMALLFLOAT, which is ~1e-35.  This led to a ratio of molecule to gas
			 *	phase of ~1e15, clearly unphysical.  Here the value of co_denominator 
			 *  will be set to a high value if the gas phase abundance is zero.*/

			if( dense.lgElmtOn[co.nelem_hevmol[i]] )
			{
				co_denominator = dense.gas_phase[co.nelem_hevmol[i]];
			}
			else
			{
				/* >>chng 04 apr 20, set to zero if element is off */
				co_denominator = 1e10;
			}

			
			/* we must have a positive value by now, unlesss element is turned off */
			ASSERT( co_denominator > SMALLFLOAT || !dense.lgElmtOn[co.nelem_hevmol[i]] );

			/* >>chng 04 feb 28, threshold from -1e-10 to -1e-6, the 
			 * level of roundoff in a float */
			/*if( b[i] / MAX2 (co_denominator, SMALLFLOAT) > -1e-10) */
			/*if( b[i] / SDIV(co_denominator) > -1e-6 ) */
			/* >>chng 04 oct 31, change limit from -1e-6 to -1e-5, produced only
			 * a few comments in map.in */
			if( b[i] / SDIV(co_denominator) > -1e-5 ) 
			{
				/* this case, it is only slightly negative relative to
				 * the parent species - multiply by -1 to make positive
				 * and press on */
				b[i] *= -1.;
			}
			else
			{
				/*>>chng 04 mar 06 press on */
				*lgNegPop = TRUE;
				if( nzone>0 )
				{
					static long int nzFail=-2;
					long int nzFail2 = (long)fnzone*100;
					if( !conv.lgSearch )
						fprintf(ioQQQ," PROBLEM ");
					fprintf(ioQQQ,
						" CO_step neg pop for species %li %s, value is %.2e rel value is %.2e zone %.2f Te %.4e Search?%c\n",
						i , 
						co.chLab[i] , 
						/* abs value */
						b[i], 
						/* relative value */
						b[i] / SDIV(co_denominator) ,
						fnzone,
						phycon.te,TorF( conv.lgSearch ) );
					/* if well beyond search phase and still having problems, announce them */
					/* only announce one failure per sweep through solvers by ConvFail */
					if( nzFail2 !=nzFail )
					{
						nzFail = nzFail2;
						ConvFail( "pops" , "CO");
					}
				}
				b[i] *= -1;
				/*>>chng 04 jan 24 set to zero instead of *-1,
				 * h2_cr.in co density went to infinite, with some negative
				 * var getting to - inf 
				 *>>chng 04 nov 03, remove this - not needed h2_cr works without
				b[i] = SMALLFLOAT;*/
			} 
		}
		else if ( b[i] == 0. )
		{
			/* this is not used for anything in calling routine and 
			 * could be cleaned up */
			*lgZerPop = TRUE;
			/* >>chng 04 feb 28, zero pop not really a problem in very deep neutral
			 * gas - this happens */
#			if 0
			if( /*nzone>0 ||*/ CODEBUG>1 )
			{
				fprintf(ioQQQ," PROBLEM ");
				fprintf(ioQQQ,
					" CO_step zero pop for species %li %s, value is %.2e zone %li\n",
					i , co.chLab[i] , b[i], nzone);
			}
#			endif
			b[i] = SMALLFLOAT;
		}
		co.hevmol[i] = (float)b[i];
	}
	/* >>chng 04 mar 06 pass negative pop as a problem, but not a fatal one,
	 * also do not call this during 0th zone when search for conditions
	 * is underway */
	if( *lgNegPop && (nzone>0&&!conv.lgSearch)  )
	{
		conv.lgConvPops = FALSE;
		fprintf(ioQQQ," CO network negative population occurred, Te=%.4e, calling ConvFail. ",
			phycon.te);
		fprintf( ioQQQ, " CO/C=%9.1e", co.hevmol[ipCO]/SDIV(dense.gas_phase[ipCARBON]) );
		fprintf( ioQQQ, "\n" );
		/*ConvFail("pops", "CO");*/
		*lgNegPop = FALSE;
	}
	/* if negative pops were present need to renorm b to get
	 * proper sum rule */
	if( 0 && *lgNegPop )
	{

		for(j=0; j<2; ++j )
		{
			double sumcar = 0.;
			double sumoxy = 0.;
			double renorm;
			for ( i=0; i < NUM_COMOLE_CALC; i++ )
			{
				/* this case, different molecules */
				sumcar += co.hevmol[i]*co.nCarb[i];
				sumoxy += co.hevmol[i]*co.nOxyg[i];
			}
			renorm = (cartot_ion + oxytot_ion) / SDIV( sumcar+sumoxy);
			if(j)
				fprintf(ioQQQ,"\t%f\n", renorm);
			else
				fprintf(ioQQQ,"renormco\t%.2f\t%f", fnzone, renorm);
			for ( i=0; i < NUM_COMOLE_CALC; i++ )
			{
				co.hevmol[i] *= (float)renorm;
			}
		}
	}

	if ( merror != 0 )
	{
		fprintf( ioQQQ, " COMOLE matrix inversion error, MERROR=%5ld zone=%5ld\n", 
				 (long)merror, nzone );
		ShowMe();
		fprintf( ioQQQ, " Product matrix\n           " );

		for ( i=0; i < MIN2(NUM_COMOLE_CALC,8); i++ )
		{
			fprintf( ioQQQ, "%4.4s", co.chLab[i] );
		}
		fprintf( ioQQQ, "     \n" );

		for ( j=0; j < NUM_COMOLE_CALC; j++ )
		{
			fprintf( ioQQQ, " %4.4s", co.chLab[j] );
			fprintf( ioQQQ, " " );

			for ( i=0; i < MIN2(NUM_COMOLE_CALC,8); i++ )
			{
				fprintf( ioQQQ, "%9.1e", csav[i][j]*
						 co.hevmol[i] );
			}
			fprintf( ioQQQ, "\n" );
		}

		if ( NUM_COMOLE_CALC >= 9 )
		{
			fprintf( ioQQQ, " COMOLE matrix\n           " );
			for ( i=8; i < NUM_COMOLE_CALC; i++ )
			{
				fprintf( ioQQQ, "%4.4s", co.chLab[i] );
			}
			fprintf( ioQQQ, "     \n" );

			for ( j=0; j < NUM_COMOLE_CALC; j++ )
			{
				fprintf( ioQQQ, " %4.4s", co.chLab[j] );
				fprintf( ioQQQ, " " );
				for ( i=8; i < NUM_COMOLE_CALC; i++ )
				{
					fprintf( ioQQQ, "%9.1e", 
							 csav[i][j]* co.hevmol[i] );
				}
				fprintf( ioQQQ, "\n" );
			}
		}

		fprintf( ioQQQ, " Mole dens:" );
		for ( j=0; j < NUM_COMOLE_CALC; j++ )
		{
			fprintf( ioQQQ, " %4.4s:%.2e", co.chLab[j]
					 , co.hevmol[j] );
		}
		fprintf( ioQQQ, " \n" );

		puts( "[Stop in CO_step]" );
		cdEXIT(EXIT_FAILURE);
	}

	if ( trace.lgTrace && trace.lgTr_CO_Mole )
	{
		fprintf( ioQQQ, " Product matrix\n           " );

		for ( i=0; i < MIN2(NUM_COMOLE_CALC,8); i++ )
		{
			fprintf( ioQQQ, "%4.4s", co.chLab[i] );
		}
		fprintf( ioQQQ, "     \n" );

		for ( j=0; j < NUM_COMOLE_CALC; j++ )
		{
			fprintf( ioQQQ, " %4.4s", co.chLab[j] );
			fprintf( ioQQQ, " " );
			for ( i=0; i < MIN2(NUM_COMOLE_CALC,8); i++ )
			{
				fprintf( ioQQQ, "%9.1e", csav[i][j]*
						 co.hevmol[i] );
			}
			fprintf( ioQQQ, "\n" );

		}

		if ( NUM_COMOLE_CALC >= 9 )
		{
			fprintf( ioQQQ, " COMOLE matrix\n           " );
			for ( i=8; i < NUM_COMOLE_CALC; i++ )
			{
				fprintf( ioQQQ, "%4.4s", co.chLab[i] );
			}
			fprintf( ioQQQ, "     \n" );

			for ( j=0; j < NUM_COMOLE_CALC; j++ )
			{
				fprintf( ioQQQ, " %4.4s", co.chLab[j] );
				fprintf( ioQQQ, " " );

				for ( i=8; i < NUM_COMOLE_CALC; i++ )
				{
					fprintf( ioQQQ, "%9.1e", csav[i][j]* co.hevmol[i] );
				}
				fprintf( ioQQQ, "\n" );
			}
		}

		fprintf( ioQQQ, " Mole dens:" );
		for ( j=0; j < NUM_COMOLE_CALC; j++ )
		{
			fprintf( ioQQQ, " %4.4s:%.2e", co.chLab[j]
					 , co.hevmol[j] );
		}
		fprintf( ioQQQ, " \n" );
	}

	/* heating due to CO photodissociation */
	co.CODissHeat = (float)(co.CO_photo_dissoc_rate*co.hevmol[ipCO]*1e-12);

	thermal.heating[0][9] = co.CODissHeat;

	/* the real multi-level model molecule */
	abundan = co.hevmol[ipCO];
	/* IonStg and nelem were set to 2, 0 in makelevlines */
	ASSERT( C12O16Rotate[0].IonStg < LIMELM+2 );
	ASSERT( C12O16Rotate[0].nelem-1 < LIMELM+2 );
	dense.xIonDense[ C12O16Rotate[0].nelem-1][C12O16Rotate[0].IonStg-1] = abundan;

	CO_PopsEmisCool(&C12O16Rotate, nCORotate,abundan , "12CO",
		&CoolHeavy.C12O16Rot,&CoolHeavy.dC12O16Rot );
	/*if( nzone>400 )
		fprintf(ioQQQ,"DEBUG co cool\t%.2f\t%.4e\t%li\t%.4e\t%.4e\n",
			fnzone,CoolHeavy.C12O16Rot,nCORotate,abundan,phycon.te );*/

	abundan = co.hevmol[ipCO]/co.RatioC12O16_2_C13O16;
	/* IonStg and nelem were set to 3, 0 in makelevlines */
	ASSERT( C13O16Rotate[0].IonStg < LIMELM+2 );
	ASSERT( C13O16Rotate[0].nelem-1 < LIMELM+2 );
	dense.xIonDense[ C13O16Rotate[0].nelem-1][C13O16Rotate[0].IonStg-1] = abundan;

	CO_PopsEmisCool(&C13O16Rotate, nCORotate,abundan ,"13CO",
		&CoolHeavy.C13O16Rot,&CoolHeavy.dC13O16Rot );

	/* now set total density of each element locked in gas phase */
	for ( i=0;i<LIMELM; ++i )
	{
		dense.xMolecules[i] = 0.;
	}

	/* total number of H per unit vol in molecules,
	 * of course not including H0/H+ */
	for (i=0;i<N_H_MOLEC;i++) 
	{
		dense.xMolecules[ipHYDROGEN] += hmi.Hmolec[i]*hmi.nProton[i];
	}
	dense.xMolecules[ipHYDROGEN] -= (hmi.Hmolec[ipMH] + hmi.Hmolec[ipMHp]);

	/* >>chng 02 sep 05, dense.xMolecules[[ipHYDROGEN] is the part of H 
	 * that is not computed in hmole
	 * add in gas phase abundances locked in molecules
	 * H is special since treated in bidiag with all hmole molecules
	 * xMolecules are the densities of these species that are done in co,
	 * this sum is only up to NUM_HEAVY_MOLEC and so does not include the atoms/ions */
	for(i=0; i<NUM_HEAVY_MOLEC; ++i)
	{
		/* these were all initialized to zero above */
		dense.xMolecules[ipHYDROGEN] +=co.nHydr[i]*co.hevmol[i];
		dense.xMolecules[ipCARBON]   +=co.nCarb[i]*co.hevmol[i];
		dense.xMolecules[ipOXYGEN]   +=co.nOxyg[i]*co.hevmol[i];
		dense.xMolecules[ipSILICON]  +=co.nSili[i]*co.hevmol[i];
		dense.xMolecules[ipSULPHUR]  +=co.nSulp[i]*co.hevmol[i];
		dense.xMolecules[ipNITROGEN] +=co.nNitr[i]*co.hevmol[i];
		dense.xMolecules[ipCHLORINE] +=co.nChlo[i]*co.hevmol[i];
	}

	/* This takes the average of an element's atomic and singly ionized density from ion_solver 
	   and comole and sets the solution to the updated co.hevmol[i] value.  The macro NUM_HEAVY_MOLEC
	   is the number of heavy element molecules in the network.  In addition there are 2*NUM_ELEMENTS
	   in the network, half atomic elements and half ionized elements.  Finally, the total number of species
	   in the network, including elements, equals NUM_COMOLE_CALC.  The following will first take the average of 
	   the ionized species (between NUM_HEAVY_MOLEC and NUM_HEAVY_MOLEC+NUM_ELEMENTS) and the other statement will
	   take the average of the atomic species (between NUM_HEAVY_MOLEC+NUM_ELEMENTS and NUM_COMOLE_CALC).  
	   This is generalized, so that if one wants to insert, for example, a sodium chemistry, once the macro's
	   are updated this if statement immediately works */

	for(i=NUM_HEAVY_MOLEC;i<NUM_HEAVY_MOLEC+NUM_ELEMENTS; ++i)
	{ 
		/*printf("MOL\t%3e\tION\t%e\tSPECIES\t%s\n", co.hevmol[i],dense.xIonDense[co.nelem_hevmol[i]][1],co.chLab[i]);*/
		co.hevmol[i] = ((dense.xIonDense[co.nelem_hevmol[i]][1]+ co.hevmol[i])/2)*dense.lgElmtOn[co.nelem_hevmol[i]];
	}

	for(i=NUM_HEAVY_MOLEC+NUM_ELEMENTS;i<NUM_COMOLE_CALC; ++i)
	{ 
		/*printf("MOL\t%3e\tION\t%e\tSPECIES\t%s\n", co.hevmol[i],dense.xIonDense[co.nelem_hevmol[i]][0],co.chLab[i]);*/
		co.hevmol[i] = ((dense.xIonDense[co.nelem_hevmol[i]][0]+ co.hevmol[i])/2)*dense.lgElmtOn[co.nelem_hevmol[i]];
	}


	/* check whether ion and chem solvers agree yet */
	if( dense.lgElmtOn[ipCARBON] &&
		fabs(dense.xIonDense[ipCARBON][0]- co.hevmol[ipATC])/SDIV(dense.gas_phase[ipCARBON]) >0.001 )
	{
		conv.lgConvIoniz = FALSE;
		sprintf( conv.chConvIoniz, "CO C0 con");
		conv.BadConvIoniz[0] = dense.xIonDense[ipCARBON][0];
		conv.BadConvIoniz[1] = co.hevmol[ipATC];
	}
	else if( dense.lgElmtOn[ipCARBON] &&
		fabs(dense.xIonDense[ipCARBON][1]- co.hevmol[ipCP])/SDIV(dense.gas_phase[ipCARBON]) >0.001 )
	{
		conv.lgConvIoniz = FALSE;
		sprintf( conv.chConvIoniz, "CO C1 con");
		conv.BadConvIoniz[0] = dense.xIonDense[ipCARBON][1];
		conv.BadConvIoniz[1] = co.hevmol[ipCP];
	}
	else if( dense.lgElmtOn[ipOXYGEN] &&
		fabs(dense.xIonDense[ipOXYGEN][0]- co.hevmol[ipATO])/SDIV(dense.gas_phase[ipOXYGEN]) >0.001 )
	{
		conv.lgConvIoniz = FALSE;
		sprintf( conv.chConvIoniz, "CO O0 con");
		conv.BadConvIoniz[0] = dense.xIonDense[ipOXYGEN][0];
		conv.BadConvIoniz[1] = co.hevmol[ipATO];
	}
	else if( dense.lgElmtOn[ipOXYGEN] &&
		fabs(dense.xIonDense[ipOXYGEN][1]- co.hevmol[ipOP])/SDIV(dense.gas_phase[ipOXYGEN]) >0.001 )
	{
		conv.lgConvIoniz = FALSE;
		sprintf( conv.chConvIoniz, "CO O1 con");
		conv.BadConvIoniz[0] = dense.xIonDense[ipOXYGEN][1];
		conv.BadConvIoniz[1] = co.hevmol[ipOP];
	}

	/* now update ionization distribution of the elements we just did */
	dense.xIonDense[ipCARBON][0] = co.hevmol[ipATC]*dense.lgElmtOn[ipCARBON] ; 
	dense.xIonDense[ipCARBON][1] = co.hevmol[ipCP]*dense.lgElmtOn[ipCARBON]; 
	dense.IonLow[ipCARBON] = 0;
	dense.IonHigh[ipCARBON] = MAX2( dense.IonHigh[ipCARBON] , 1 );

	dense.xIonDense[ipOXYGEN][0] = co.hevmol[ipATO]*dense.lgElmtOn[ipOXYGEN] ; 
	dense.xIonDense[ipOXYGEN][1] = co.hevmol[ipOP]*dense.lgElmtOn[ipOXYGEN]; 
	dense.IonLow[ipOXYGEN] = 0;
	dense.IonHigh[ipOXYGEN] = MAX2( dense.IonHigh[ipOXYGEN] , 1 );

	dense.xIonDense[ipSILICON][0] = co.hevmol[ipATSI]*dense.lgElmtOn[ipSILICON] ; 
	dense.xIonDense[ipSILICON][1] = co.hevmol[ipSIP]*dense.lgElmtOn[ipSILICON]; 
	dense.IonLow[ipSILICON] = 0;
	dense.IonHigh[ipSILICON] = MAX2( dense.IonHigh[ipSILICON] , 1 );

	dense.xIonDense[ipSULPHUR][0] = co.hevmol[ipATS]*dense.lgElmtOn[ipSULPHUR] ; 
	dense.xIonDense[ipSULPHUR][1] = co.hevmol[ipSP]*dense.lgElmtOn[ipSULPHUR]; 
	dense.IonLow[ipSULPHUR] = 0;
	dense.IonHigh[ipSULPHUR] = MAX2( dense.IonHigh[ipSULPHUR] , 1 );

	dense.xIonDense[ipNITROGEN][0] = co.hevmol[ipATN]*dense.lgElmtOn[ipNITROGEN] ; 
	dense.xIonDense[ipNITROGEN][1] = co.hevmol[ipNP]*dense.lgElmtOn[ipNITROGEN]; 
	dense.IonLow[ipNITROGEN] = 0;
	dense.IonHigh[ipNITROGEN] = MAX2( dense.IonHigh[ipNITROGEN] , 1 );
	
	dense.xIonDense[ipCHLORINE][0] = co.hevmol[ipATCl]*dense.lgElmtOn[ipCHLORINE] ; 
	dense.xIonDense[ipCHLORINE][1] = co.hevmol[ipClP]*dense.lgElmtOn[ipCHLORINE]; 
	dense.IonLow[ipCHLORINE] = 0;
	dense.IonHigh[ipCHLORINE] = MAX2( dense.IonHigh[ipCHLORINE] , 1 );


	/* if populations not conserved then not converged */
#	define EPS_MOLE	0.1
	if( dense.lgElmtOn[ipHYDROGEN] &&
		dense.xMolecules[ipHYDROGEN] > dense.gas_phase[ipHYDROGEN]*(1.+EPS_MOLE) )
	{
		conv.lgConvIoniz = FALSE;
		sprintf( conv.chConvIoniz, "COcon%2i",ipHYDROGEN );
		conv.BadConvIoniz[0] = dense.xMolecules[ipHYDROGEN];
		conv.BadConvIoniz[1] = dense.gas_phase[ipHYDROGEN];
	}
	else if( dense.lgElmtOn[ipCARBON] &&
		dense.xMolecules[ipCARBON] > dense.gas_phase[ipCARBON]*(1.+EPS_MOLE) )
	{
		conv.lgConvIoniz = FALSE;
		sprintf( conv.chConvIoniz, "COcon%2i",ipCARBON );
		conv.BadConvIoniz[0] = dense.xMolecules[ipCARBON];
		conv.BadConvIoniz[1] = dense.gas_phase[ipCARBON];
	}
	else if( dense.lgElmtOn[ipOXYGEN] &&
		dense.xMolecules[ipOXYGEN] > dense.gas_phase[ipOXYGEN]*(1.+EPS_MOLE) )
	{
		conv.lgConvIoniz = FALSE;
		sprintf( conv.chConvIoniz, "COcon%2i",ipOXYGEN );
		conv.BadConvIoniz[0] = dense.xMolecules[ipOXYGEN];
		conv.BadConvIoniz[1] = dense.gas_phase[ipOXYGEN];
	}
	else if( dense.lgElmtOn[ipSILICON] &&
		dense.xMolecules[ipSILICON] > dense.gas_phase[ipSILICON]*(1.+EPS_MOLE) )
	{
		conv.lgConvIoniz = FALSE;
		sprintf( conv.chConvIoniz, "COcon%2i",ipSILICON );
		conv.BadConvIoniz[0] = dense.xMolecules[ipSILICON];
		conv.BadConvIoniz[1] = dense.gas_phase[ipSILICON];
	}
	else if( dense.lgElmtOn[ipSULPHUR] &&
		dense.xMolecules[ipSULPHUR] > dense.gas_phase[ipSULPHUR]*(1.+EPS_MOLE) )
	{
		conv.lgConvIoniz = FALSE;
		sprintf( conv.chConvIoniz, "COcon%2i",ipSULPHUR );
		conv.BadConvIoniz[0] = dense.xMolecules[ipSULPHUR];
		conv.BadConvIoniz[1] = dense.gas_phase[ipSULPHUR];
	}
	else if( dense.lgElmtOn[ipNITROGEN] &&
		dense.xMolecules[ipNITROGEN] > dense.gas_phase[ipNITROGEN]*(1.+EPS_MOLE) )
	{
		conv.lgConvIoniz = FALSE;
		sprintf( conv.chConvIoniz, "COcon%2i",ipNITROGEN );
		conv.BadConvIoniz[0] = dense.xMolecules[ipNITROGEN];
		conv.BadConvIoniz[1] = dense.gas_phase[ipNITROGEN];
	}

	else if( dense.lgElmtOn[ipCHLORINE] &&
		dense.xMolecules[ipCHLORINE] > dense.gas_phase[ipCHLORINE]*(1.+EPS_MOLE) )
	{
		conv.lgConvIoniz = FALSE;
		sprintf( conv.chConvIoniz, "COcon%2i",ipCHLORINE );
		conv.BadConvIoniz[0] = dense.xMolecules[ipCHLORINE];
		conv.BadConvIoniz[1] = dense.gas_phase[ipCHLORINE];
	}
#	undef EPS_MOLE

#	ifdef DEBUG_FUN
	fputs( " <->CO_step()\n", debug_fp );
#	endif
	return;
}

/*lint +e550 */


/*lint +e778 constant expression evaluatess to 0 in operation '-' */

